Package: shard
Type: Package
Title: Deterministic, Zero-Copy Parallel Execution for R
Version: 0.1.1
Authors@R: 
    person("Bradley", "Buchsbaum", email = "brad.buchsbaum@gmail.com", role = c("aut", "cre", "cph"))
Description: Provides a parallel execution runtime for R that emphasizes
    deterministic memory behavior and efficient handling of large shared inputs.
    'shard' enables zero-copy parallel reads via shared/memory-mapped segments,
    encourages explicit output buffers to avoid large result aggregation, and
    supervises worker processes to mitigate memory drift via controlled recycling.
    Diagnostics report peak memory usage, end-of-run memory return, and hidden
    copy/materialization events to support reproducible performance benchmarking.
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
Depends: R (>= 4.1.0)
Imports: methods, parallel, stats, tools, utils
Suggests: knitr, pkgload, rmarkdown, testthat (>= 3.0.0), ps, jsonlite,
        tibble, withr
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: yes
URL: https://bbuchsbaum.github.io/shard/,
        https://github.com/bbuchsbaum/shard
BugReports: https://github.com/bbuchsbaum/shard/issues
SystemRequirements: POSIX shared memory (optional), memory-mapped files
Config/testthat/edition: 3
Packaged: 2026-04-05 14:02:34 UTC; bbuchsbaum
Author: Bradley Buchsbaum [aut, cre, cph]
Maintainer: Bradley Buchsbaum <brad.buchsbaum@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-05 20:40:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 22:18:38 UTC; unix
Archs: shard.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",shard_buffer)
S3method("[",shard_descriptor)
S3method("[",shard_descriptor_lazy)
S3method("[",shard_results_placeholder)
S3method("[<-",shard_buffer)
S3method("[<-",shard_shared_vector)
S3method("[[",shard_descriptor)
S3method("[[",shard_descriptor_lazy)
S3method("[[",shard_results_placeholder)
S3method("[[<-",shard_shared_vector)
S3method("attr<-",shard_shared_vector)
S3method("attributes<-",shard_shared_vector)
S3method("dim<-",shard_shared_vector)
S3method("dimnames<-",shard_shared_vector)
S3method("names<-",shard_shared_vector)
S3method(as.array,shard_buffer)
S3method(as.double,shard_buffer)
S3method(as.integer,shard_buffer)
S3method(as.logical,shard_buffer)
S3method(as.matrix,shard_buffer)
S3method(as.raw,shard_buffer)
S3method(as.vector,shard_buffer)
S3method(as_tibble,shard_dataset)
S3method(as_tibble,shard_row_groups)
S3method(as_tibble,shard_table_buffer)
S3method(as_tibble,shard_table_handle)
S3method(close,shard_deep_shared)
S3method(close,shard_shared)
S3method(close,shard_shared_vector)
S3method(collect,shard_dataset)
S3method(collect,shard_row_groups)
S3method(collect,shard_table_handle)
S3method(dim,shard_buffer)
S3method(fetch,default)
S3method(fetch,shard_deep_shared)
S3method(fetch,shard_shared)
S3method(length,shard_buffer)
S3method(length,shard_descriptor)
S3method(length,shard_descriptor_lazy)
S3method(length,shard_results_placeholder)
S3method(materialize,default)
S3method(materialize,shard_shared)
S3method(materialize,shard_view_block)
S3method(materialize,shard_view_gather)
S3method(print,arena_result)
S3method(print,shard_apply_policy)
S3method(print,shard_buffer)
S3method(print,shard_deep_shared)
S3method(print,shard_descriptor)
S3method(print,shard_descriptor_lazy)
S3method(print,shard_dispatch_result)
S3method(print,shard_health_report)
S3method(print,shard_idx_range)
S3method(print,shard_pool)
S3method(print,shard_queue)
S3method(print,shard_reduce_result)
S3method(print,shard_report)
S3method(print,shard_result)
S3method(print,shard_results_placeholder)
S3method(print,shard_segment)
S3method(print,shard_shared)
S3method(print,shard_shared_vector)
S3method(print,shard_tiles)
S3method(print,shard_view_block)
S3method(print,shard_view_gather)
S3method(print,shard_worker)
S3method(shard_share_hook,default)
S3method(stream_map,shard_dataset)
S3method(stream_map,shard_row_groups)
S3method(stream_reduce,shard_dataset)
S3method(stream_reduce,shard_row_groups)
S3method(table_finalize,shard_table_buffer)
S3method(table_finalize,shard_table_sink)
S3method(table_write,shard_table_buffer)
S3method(table_write,shard_table_sink)
export(affinity_supported)
export(arena)
export(arena_depth)
export(as_shared)
export(as_tibble)
export(available_backings)
export(bool)
export(buffer)
export(buffer_advise)
export(buffer_close)
export(buffer_diagnostics)
export(buffer_info)
export(buffer_open)
export(buffer_path)
export(collect)
export(copy_report)
export(cow_report)
export(dispatch_chunks)
export(factor_col)
export(fetch)
export(float64)
export(idx_range)
export(in_arena)
export(int32)
export(is_block_view)
export(is_shared)
export(is_shared_vector)
export(is_view)
export(is_windows)
export(iterate_row_groups)
export(list_kernels)
export(materialize)
export(mem_report)
export(pin_workers)
export(pool_create)
export(pool_dispatch)
export(pool_get)
export(pool_health_check)
export(pool_lapply)
export(pool_sapply)
export(pool_status)
export(pool_stop)
export(raw_col)
export(recommendations)
export(register_kernel)
export(report)
export(results)
export(row_layout)
export(schema)
export(scratch_diagnostics)
export(scratch_matrix)
export(scratch_pool_config)
export(segment_advise)
export(segment_close)
export(segment_create)
export(segment_info)
export(segment_open)
export(segment_path)
export(segment_protect)
export(segment_read)
export(segment_report)
export(segment_size)
export(segment_write)
export(set_affinity)
export(shard_apply_matrix)
export(shard_apply_policy)
export(shard_crossprod)
export(shard_get_adapter)
export(shard_lapply_shared)
export(shard_list_adapters)
export(shard_map)
export(shard_reduce)
export(shard_register_adapter)
export(shard_share_hook)
export(shard_unregister_adapter)
export(shards)
export(shards_list)
export(share)
export(share_open)
export(shared_advise)
export(shared_diagnostics)
export(shared_info)
export(shared_reset_diagnostics)
export(shared_segment)
export(shared_vector)
export(shared_view)
export(stream_count)
export(stream_filter)
export(stream_group_count)
export(stream_group_sum)
export(stream_map)
export(stream_reduce)
export(stream_sum)
export(stream_top_k)
export(string_col)
export(succeeded)
export(table_buffer)
export(table_diagnostics)
export(table_finalize)
export(table_sink)
export(table_write)
export(task_report)
export(view)
export(view_block)
export(view_diagnostics)
export(view_gather)
export(view_info)
useDynLib(shard, .registration = TRUE)

# shard

<!-- badges: start -->
[![pkgdown](https://github.com/bbuchsbaum/shard/actions/workflows/pkgdown.yaml/badge.svg)](https://bbuchsbaum.github.io/shard/)
[![R-CMD-check](https://github.com/bbuchsbaum/shard/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/bbuchsbaum/shard/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

**Deterministic, zero-copy parallel execution for R.**

`shard` is a parallel runtime for workloads that look like:
- "run the same numeric kernel over many slices of big data"
- "thousands of independent tasks over a shared dataset"
- "parallel GLM / simulation / bootstrap / feature screening"

It focuses on three things that are often painful in R parallelism:
1) **Shared immutable inputs** (avoid duplicating large objects across workers)
2) **Explicit output buffers** (avoid huge result-gather lists)
3) **Deterministic cleanup** (supervise workers and recycle on memory drift)

---

## Installation

From CRAN (once released):

```r
install.packages("shard")
```

Development version:

```r
# install.packages("pak")
pak::pak("bbuchsbaum/shard")
```

---

## Core concepts

### 1) Share large inputs (read-only by default)

```r
X <- shard::share(X)   # matrix/array/vector
Y <- shard::share(Y)
```

Shared objects are designed for zero-copy parallel reads (where the OS allows)
and are treated as immutable by default inside parallel tasks.

### 2) Allocate explicit output buffers

Instead of returning giant objects from each worker, write to a preallocated buffer:

```r
out <- shard::buffer("double", dim = c(1e6))   # example: 1M outputs
```

### 3) Run shard_map() over shards

```r
blocks <- shard::shards(1e6, block_size = "auto")

run <- shard::shard_map(
  blocks,
  borrow = list(X = X, Y = Y),
  out = list(out = out),
  workers = 8,
  fun = function(block, X, Y, out) {
    # block contains indices
    idx <- block$idx
    out[idx] <- colMeans(Y[, idx, drop = FALSE])
  }
)

shard::report(run)
```

---

## Safety defaults (and why they matter)

### Borrowed inputs are immutable

By default, trying to mutate borrowed/shared inputs is treated as a bug:
- `cow = "deny"` (default): mutation triggers an error
- `cow = "audit"`: detect and flag (best-effort; platform dependent)
- `cow = "allow"`: allow copy-on-write, track it, and enforce budgets

**Why default is deny:**
- Prevents silent memory blowups from accidental wide writes
- Prevents subtle correctness bugs (changes are private to a worker)
- Keeps behavior predictable across platforms

---

## Deterministic cleanup via worker supervision

R's GC and allocator behavior can lead to memory drift in long-running workers.

`shard` monitors per-worker memory usage and can recycle workers when drift
exceeds thresholds, keeping end-of-run memory close to baseline.

---

## Diagnostics

After a run, `shard` can report:
- total and per-worker peak RSS
- end RSS vs baseline
- materialized bytes (hidden copies)
- recycling events, retries, timing

```r
rep <- shard::report(run)
print(rep)

shard::mem_report(run)
shard::copy_report(run)
```

---

## Convenience wrappers

If your workload is “apply a function over columns” or “lapply over a list”,
`shard` provides convenience wrappers that handle sharing and buffering
automatically while still running through the supervised runtime.

### Column-wise apply (scalar return)

```r
X <- matrix(rnorm(1e6), nrow = 1000)

scores <- shard::shard_apply_matrix(
  X,
  MARGIN = 2,
  FUN = function(v, y) cor(v, y),
  VARS = list(y = rnorm(nrow(X))),
  workers = 8
)
```

### List lapply (guarded gather)

```r
xs <- lapply(1:1000, function(i) rnorm(100))

out <- shard::shard_lapply_shared(
  xs,
  FUN = function(el) mean(el),
  workers = 8
)
```

For large outputs (big vectors/data.frames per element), prefer `buffer()`, `table_sink()`,
or `shard_reduce()` instead of gathering everything to the master.

---

---

## License

MIT

# =============================================================================
# Quick Benchmark: shard vs memshare
# =============================================================================
# Fair comparison using pre-warmed workers for both packages.
#
# Run: Rscript inst/bench/shard_vs_memshare_quick.R
# =============================================================================

suppressPackageStartupMessages({
  library(shard)
  if (!requireNamespace("memshare", quietly = TRUE)) {
    stop("Install memshare: install.packages('memshare')")
  }
  library(memshare)
  library(parallel)
})

n_workers <- 4L
# Keep this short: memshare enforces a <32-char UID limit on macOS.
bench_id <- paste0("sv", paste(sample(c(letters, 0:9), 6, replace = TRUE), collapse = ""))

cat("============================================================\n")
cat("Quick Benchmark: shard vs memshare (", n_workers, "workers)\n")
cat("============================================================\n\n")
cat("shard version: ", as.character(utils::packageVersion("shard")), "\n", sep = "")
cat("shard path:    ", find.package("shard"), "\n", sep = "")
cat("NOTE: This benchmark uses the installed 'shard'. If you're iterating locally, run `R CMD INSTALL .` first.\n\n")

results <- list()

run_test <- function(name, desc, mem_fn, shard_fn) {
  cat(sprintf("Test %s: %s\n", name, desc))
  ns <- paste0(bench_id, "_t", name)

  # Create fresh workers for each test (memshare has cluster state issues)
  cl <- makeCluster(n_workers)
  pool_stop()
  pool_create(n_workers)
  # Warm shard pool
  invisible(shard_map(n_workers, function(sh) NULL, workers = n_workers, diagnostics = FALSE))

  t_mem <- tryCatch({
    system.time(mem_fn(cl, ns))[["elapsed"]]
  }, error = function(e) {
    cat("  memshare error:", conditionMessage(e), "\n")
    NA
  })

  t_shard <- tryCatch({
    system.time(shard_fn())[["elapsed"]]
  }, error = function(e) {
    cat("  shard error:", conditionMessage(e), "\n")
    NA
  })

  stopCluster(cl)
  pool_stop()

  if (!is.na(t_mem) && !is.na(t_shard)) {
    ratio <- t_mem / t_shard
    cat(sprintf("  memshare: %.3fs | shard: %.3fs | ratio: %.2fx\n\n",
        t_mem, t_shard, ratio))
  } else {
    ratio <- NA
    cat("\n")
  }

  list(mem = t_mem, shard = t_shard, ratio = ratio)
}

frobenius_norm <- function(m) {
  # Fast, generic Frobenius norm (allocates, but typically cheaper than BLAS
  # call overhead for small matrices in this benchmark).
  sqrt(sum(m * m))
}

auto_chunk_size <- function(n, workers, target_chunks_per_worker = 4L, min = 1L, max = 32L) {
  n <- as.integer(n)
  workers <- as.integer(workers)
  target_chunks_per_worker <- as.integer(target_chunks_per_worker)

  if (is.na(n) || n <= 0L) return(min)
  if (is.na(workers) || workers <= 0L) return(min)
  if (is.na(target_chunks_per_worker) || target_chunks_per_worker <= 0L) return(min)

  # Heuristic: make each worker handle a few chunks to amortize dispatch overhead,
  # but keep chunk sizes bounded so work still load-balances well.
  chunk <- as.integer(ceiling(n / (workers * target_chunks_per_worker)))
  chunk <- max(min, min(max, chunk))
  chunk
}

# =============================================================================
# Test 1: Column means
# =============================================================================
set.seed(42)
X1 <- matrix(rnorm(2000 * 500), nrow = 2000, ncol = 500)
expected1 <- colMeans(X1)

results$test1 <- run_test("1", "Column means (2000x500 matrix)",
  mem_fn = function(cl, ns) {
    tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    on.exit({
      tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    }, add = TRUE)

    res <- memApply(X1, 2, mean, CLUSTER = cl, NAMESPACE = ns, MAX.CORES = n_workers)
    stopifnot(all.equal(unlist(res), expected1, tolerance = 1e-10))
  },
  shard_fn = function() {
    out <- buffer("double", dim = 500)
    shard_map(
      500,
      borrow = list(X = X1),
      out = list(out = out),
      kernel = "col_means",
      workers = n_workers,
      profile = "speed",
      diagnostics = FALSE
    )
    stopifnot(all.equal(as.numeric(out[]), expected1, tolerance = 1e-10))
  }
)

# =============================================================================
# Test 2: List of matrices (Frobenius norms)
# =============================================================================
set.seed(456)
mat_list <- lapply(1:100, function(i) matrix(rnorm(50*50), 50, 50))
expected2 <- vapply(mat_list, frobenius_norm, numeric(1))
mat_list_shard <- lapply(mat_list, function(m) share(m, backing = "mmap"))
n2 <- length(mat_list)

results$test2 <- run_test("2", "Frobenius norms (100 matrices, 50x50)",
  mem_fn = function(cl, ns) {
    tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    on.exit({
      tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    }, add = TRUE)

    res <- memLapply(
      X = mat_list,
      FUN = frobenius_norm,
      CLUSTER = cl,
      NAMESPACE = ns,
      MAX.CORES = n_workers
    )
    stopifnot(all.equal(unlist(res), expected2, tolerance = 1e-10))
  },
  shard_fn = function() {
    out <- buffer("double", dim = 100)
    shard_map(
      n2,
      borrow = list(mats = mat_list_shard),
      out = list(out = out),
      fun = function(sh, mats, out) {
        for (i in sh$idx) {
          m <- mats[[i]]
          out[i] <- sqrt(sum(m * m))
        }
        NULL
      },
      workers = n_workers,
      # General lever: amortize per-item dispatch overhead for many-small-items workloads.
      chunk_size = auto_chunk_size(n2, n_workers, target_chunks_per_worker = 1L),
      profile = "speed",
      diagnostics = FALSE
    )
    stopifnot(all.equal(as.numeric(out[]), expected2, tolerance = 1e-10))
  }
)

# =============================================================================
# Test 3: Crossprod X'Y
# =============================================================================
set.seed(789)
X3 <- matrix(rnorm(2000 * 64), nrow = 2000, ncol = 64)
Y3 <- matrix(rnorm(2000 * 128), nrow = 2000, ncol = 128)
expected3 <- crossprod(X3, Y3)

tile_spec <- matrix(
  c(
    1, 32, 1, 64,
    1, 32, 65, 128,
    33, 64, 1, 64,
    33, 64, 65, 128
  ),
  ncol = 4,
  byrow = TRUE
)
colnames(tile_spec) <- c("i_start", "i_end", "j_start", "j_end")
tile_list <- lapply(seq_len(nrow(tile_spec)), function(k) as.double(tile_spec[k, ]))

results$test3 <- run_test("3", "Crossprod X'Y (2000x64) x (2000x128)",
  mem_fn = function(cl, ns) {
    tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    on.exit({
      tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    }, add = TRUE)

    registerVariables(ns, list(X = X3, Y = Y3))
    res_tiles <- memLapply(
      X = tile_list,
      FUN = function(tile, X, Y) {
        i <- as.integer(tile[1]):as.integer(tile[2])
        j <- as.integer(tile[3]):as.integer(tile[4])
        crossprod(X[, i, drop = FALSE], Y[, j, drop = FALSE])
      },
      CLUSTER = cl, NAMESPACE = ns, VARS = c("X", "Y"), MAX.CORES = n_workers
    )
    out_mem <- matrix(0, 64, 128)
    for (k in seq_len(nrow(tile_spec))) {
      tile <- tile_spec[k, ]
      out_mem[tile[1]:tile[2], tile[3]:tile[4]] <- res_tiles[[k]]
    }
    stopifnot(all.equal(out_mem, expected3, tolerance = 1e-8))
  },
  shard_fn = function() {
    res <- shard_crossprod(X3, Y3, workers = n_workers, block_x = 32, block_y = 64,
                           backing = "mmap", materialize = "always", diagnostics = FALSE)
    stopifnot(all.equal(res$value, expected3, tolerance = 1e-8))
  }
)

# =============================================================================
# Test 4: Column variance (more complex per-column operation)
# =============================================================================
set.seed(321)
X4 <- matrix(rnorm(3000 * 300), nrow = 3000, ncol = 300)
expected4 <- apply(X4, 2, var)

results$test4 <- run_test("4", "Column variance (3000x300 matrix)",
  mem_fn = function(cl, ns) {
    tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    on.exit({
      tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    }, add = TRUE)

    res <- memApply(X4, 2, var, CLUSTER = cl, NAMESPACE = ns, MAX.CORES = n_workers)
    stopifnot(all.equal(unlist(res), expected4, tolerance = 1e-10))
  },
  shard_fn = function() {
    out <- buffer("double", dim = 300)
    shard_map(
      300,
      borrow = list(X = X4),
      out = list(out = out),
      kernel = "col_vars",
      workers = n_workers,
      profile = "speed",
      diagnostics = FALSE
    )
    stopifnot(all.equal(as.numeric(out[]), expected4, tolerance = 1e-10))
  }
)

# =============================================================================
# Summary
# =============================================================================
cat("============================================================\n")
cat("SUMMARY (ratio > 1 means shard faster)\n")
cat("============================================================\n\n")

cat(sprintf("%-35s %10s %10s %10s\n", "Test", "memshare", "shard", "Ratio"))
cat(strrep("-", 67), "\n")

test_names <- c(
  test1 = "1. Column means",
  test2 = "2. Matrix Frobenius norms",
  test3 = "3. Crossprod X'Y",
  test4 = "4. Column variance"
)

for (name in names(results)) {
  r <- results[[name]]
  if (!is.na(r$mem) && !is.na(r$shard)) {
    ratio <- r$mem / r$shard
    winner <- if(ratio > 1.1) "SHARD" else if(ratio < 0.9) "memshare" else "~tie"
    cat(sprintf("%-35s %9.3fs %9.3fs %9.2fx  %s\n",
        test_names[name], r$mem, r$shard, ratio, winner))
  } else {
    cat(sprintf("%-35s %9s %9s %9s\n",
        test_names[name],
        if(is.na(r$mem)) "FAILED" else sprintf("%.3fs", r$mem),
        if(is.na(r$shard)) "FAILED" else sprintf("%.3fs", r$shard),
        "N/A"))
  }
}

cat("\n")
cat("Notes:\n")
cat("- Each test uses fresh workers for fair comparison\n")
cat("- shard uses explicit output buffers (no result gathering)\n")
cat("- memshare gathers results into lists\n")
cat("- shard includes shard_crossprod() for built-in tiled matrix ops\n")

cat("\nBenchmark complete.\n")

# Benchmark / example: view-enabled t(X_block) %*% Y_gather using a gather view.
#
# This demonstrates a searchlight-style hot path:
# - Y is shared once
# - each shard uses an indexed (gather) view over columns of Y
# - compute packs the gathered columns into scratch explicitly and uses BLAS-3
# - view materialization remains 0; packing bytes are reported separately
#
# Run from repo root:
#   R -q -f inst/bench/view_xTy_gather_shard_map.R

suppressPackageStartupMessages(library(shard))

set.seed(1)
n <- 4000L
p <- 24L
v <- 2000L

X <- matrix(rnorm(n * p), nrow = n)
Y <- matrix(rnorm(n * v), nrow = n)

colnames(X) <- paste0("x", seq_len(ncol(X)))
colnames(Y) <- paste0("y", seq_len(ncol(Y)))

Xsh <- share(X, backing = "mmap")
Ysh <- share(Y, backing = "mmap")

# 200 searchlights, each with 64 (non-contiguous) columns.
idxs <- replicate(200, sort(sample.int(ncol(Y), 64)), simplify = FALSE)
shards_g <- shards_list(idxs)

res <- shard_map(
  shards_g,
  borrow = list(X = Xsh, Y = Ysh),
  fun = function(shard, X, Y) {
    vY <- view(Y, cols = shard$idx, type = "auto")
    shard:::view_xTy(X, vY)
  },
  workers = 4,
  diagnostics = TRUE,
  chunk_size = 2
)

print(task_report(res))
print(copy_report(res))

mats <- results(res)
cat("\nNum results:", length(mats), "\n")
cat("First result dim:", paste(dim(mats[[1]]), collapse = "x"), "\n")

# Benchmark / example: view-enabled t(X_block) %*% Y_block using two views
#
# This demonstrates a "two-view" hot path:
# - X and Y are shared once
# - shard fun constructs block views for BOTH operands
# - the compute uses BLAS-3 (dgemm) without materializing either view
#
# Run from repo root:
#   R -q -f inst/bench/view_xTy_2views_shard_map.R

suppressPackageStartupMessages(library(shard))

set.seed(1)
n <- 5000L
p <- 32L
v <- 256L

X <- matrix(rnorm(n * p), nrow = n)
Y <- matrix(rnorm(n * v), nrow = n)

colnames(X) <- paste0("x", seq_len(ncol(X)))
colnames(Y) <- paste0("y", seq_len(ncol(Y)))

Xsh <- share(X, backing = "mmap")
Ysh <- share(Y, backing = "mmap")

blocks <- shards(ncol(Y), block_size = 32, workers = 4)

res <- shard_map(
  blocks,
  borrow = list(X = Xsh, Y = Ysh),
  fun = function(shard, X, Y) {
    vX <- view_block(X, cols = idx_range(1, ncol(X)))
    vY <- view_block(Y, cols = idx_range(shard$start, shard$end))
    shard:::view_crossprod(vX, vY)
  },
  workers = 4,
  diagnostics = TRUE
)

print(task_report(res))
print(copy_report(res))

Z <- do.call(cbind, results(res))
cat("\nResult dim:", paste(dim(Z), collapse = "x"), "\n")
cat("Max abs error vs base crossprod:", max(abs(Z - crossprod(X, Y))), "\n")

# Benchmark / example: view-enabled X'Y block computation with shard_map
#
# This script demonstrates an end-to-end pipeline where:
# - X and Y are shared once (no per-worker duplication)
# - each shard operates on a block view of Y (no Y[, idx] slice allocation)
# - the hot path uses a BLAS-3 kernel (t(X) %*% Y_block) without view materialization
#
# Run from repo root:
#   R -q -f inst/bench/view_xTy_shard_map.R

suppressPackageStartupMessages(library(shard))

set.seed(1)
n <- 5000L
p <- 32L
v <- 256L

X <- matrix(rnorm(n * p), nrow = n)
Y <- matrix(rnorm(n * v), nrow = n)

colnames(X) <- paste0("x", seq_len(ncol(X)))
colnames(Y) <- paste0("y", seq_len(ncol(Y)))

Xsh <- share(X, backing = "mmap")
Ysh <- share(Y, backing = "mmap")

blocks <- shards(ncol(Y), block_size = 32, workers = 4)

res <- shard_map(
  blocks,
  borrow = list(X = Xsh, Y = Ysh),
  fun = function(shard, X, Y) {
    vY <- view_block(Y, cols = idx_range(shard$start, shard$end))
    shard:::view_xTy(X, vY)
  },
  workers = 4,
  diagnostics = TRUE
)

print(task_report(res))
print(copy_report(res))

# Materialize the full matrix for a quick correctness check.
Z <- do.call(cbind, results(res))
cat("\nResult dim:", paste(dim(Z), collapse = "x"), "\n")
cat("Max abs error vs base crossprod:", max(abs(Z - crossprod(X, Y))), "\n")

# =============================================================================
# Benchmark: shard vs memshare (overhead + shared-big-data access)
# =============================================================================
#
# Run: Rscript inst/bench/shard_vs_memshare_overhead.R
#
# Notes:
# - Uses fresh PSOCK workers per test for fair comparison.
# - shard writes results into explicit output buffers (no result gathering).
# - memshare gathers results into lists.
# =============================================================================

suppressPackageStartupMessages({
  library(shard)
  if (!requireNamespace("memshare", quietly = TRUE)) {
    stop("Install memshare: install.packages('memshare')")
  }
  library(memshare)
  library(parallel)
})

n_workers <- 4L
bench_id <- paste0("sv", paste(sample(c(letters, 0:9), 8, replace = TRUE), collapse = ""))

auto_chunk_size <- function(n, workers, target_chunks_per_worker = 2L, min = 1L, max = 256L) {
  n <- as.integer(n)
  workers <- as.integer(workers)
  target_chunks_per_worker <- as.integer(target_chunks_per_worker)

  if (is.na(n) || n <= 0L) return(min)
  if (is.na(workers) || workers <= 0L) return(min)
  if (is.na(target_chunks_per_worker) || target_chunks_per_worker <= 0L) return(min)

  chunk <- as.integer(ceiling(n / (workers * target_chunks_per_worker)))
  chunk <- max(min, min(max, chunk))
  chunk
}

env_int <- function(name, default) {
  x <- Sys.getenv(name, unset = "")
  if (!nzchar(x)) return(as.integer(default))
  suppressWarnings(as.integer(x))
}

run_memshare <- function(desc, mem_fn, ns) {
  cl <- NULL
  t_mem <- tryCatch({
    cl <- makeCluster(n_workers)
    on.exit(tryCatch(stopCluster(cl), error = function(e) NULL), add = TRUE)
    system.time(mem_fn(cl, ns))[["elapsed"]]
  }, error = function(e) {
    cat("  memshare error:", conditionMessage(e), "\n")
    NA_real_
  })

  # Ensure workers are torn down even if memshare poisoned the cluster.
  if (!is.null(cl)) {
    tryCatch(stopCluster(cl), error = function(e) NULL)
  }
  t_mem
}

run_shard <- function(desc, shard_fn) {
  pool_stop()
  pool_create(n_workers)
  # Warm shard pool
  invisible(shard_map(n_workers, function(sh) NULL, workers = n_workers, diagnostics = FALSE))
  t_shard <- tryCatch({
    system.time(shard_fn())[["elapsed"]]
  }, error = function(e) {
    cat("  shard error:", conditionMessage(e), "\n")
    NA_real_
  })
  pool_stop()
  t_shard
}

run_test <- function(name, desc, mem_fn, shard_fn) {
  cat(sprintf("Test %s: %s\n", name, desc))
  ns <- paste0(bench_id, "_t", name)

  # Run memshare and shard in separate worker lifecycles so a failure in one
  # (common when experimenting with memshare namespaces) doesn't corrupt the other.
  t_mem <- run_memshare(desc, mem_fn, ns)
  t_shard <- run_shard(desc, shard_fn)

  if (!is.na(t_mem) && !is.na(t_shard)) {
    ratio <- t_mem / t_shard
    cat(sprintf("  memshare: %.3fs | shard: %.3fs | ratio: %.2fx\n\n", t_mem, t_shard, ratio))
  } else {
    ratio <- NA_real_
    cat("\n")
  }

  list(mem = t_mem, shard = t_shard, ratio = ratio)
}

cat("============================================================\n")
cat("Overhead Benchmark: shard vs memshare (", n_workers, "workers)\n", sep = "")
cat("============================================================\n\n")
cat("shard version: ", as.character(utils::packageVersion("shard")), "\n", sep = "")
cat("shard path:    ", find.package("shard"), "\n\n", sep = "")

results <- list()

# =============================================================================
# Test 1: Dispatch throughput (many tiny tasks)
# =============================================================================
set.seed(1)
n_tasks1 <- env_int("SHARD_BENCH_N_TASKS1", 10000L)
idx1 <- seq_len(n_tasks1)
idx1_d <- as.double(idx1)

results$test1 <- run_test("1", sprintf("Dispatch throughput (%d tiny tasks)", n_tasks1),
  mem_fn = function(cl, ns) {
    tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    on.exit({
      tryCatch(releaseVariables(ns, c("idx")), error = function(e) NULL)
      tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    }, add = TRUE)

    registerVariables(ns, list(idx = as.list(idx1_d)))

    res <- memLapply(
      X = "idx",
      FUN = function(i) i,
      CLUSTER = cl,
      NAMESPACE = ns,
      MAX.CORES = n_workers
    )
    stopifnot(identical(unlist(res, use.names = FALSE), idx1_d))
  },
  shard_fn = function() {
    out <- buffer("int", dim = n_tasks1)
    shard_map(
      n_tasks1,
      borrow = list(idx = idx1),
      out = list(out = out),
      fun = function(sh, idx, out) {
        for (j in sh$idx) out[j] <- idx[j]
        NULL
      },
      workers = n_workers,
      chunk_size = auto_chunk_size(n_tasks1, n_workers, target_chunks_per_worker = 1L, max = 2048L),
      profile = "speed",
      diagnostics = FALSE
    )
    stopifnot(identical(as.integer(out[]), idx1))
  }
)

# =============================================================================
# Test 2: Shared-big-matrix access (random row sums)
# =============================================================================
set.seed(2)
n_rows <- 20000L
n_cols <- 256L
X2 <- matrix(rnorm(n_rows * n_cols), nrow = n_rows, ncol = n_cols)
X2_shard <- share(X2, backing = "mmap")

n_tasks2 <- env_int("SHARD_BENCH_N_TASKS2", 50000L)
rows2 <- sample.int(n_rows, n_tasks2, replace = TRUE)
rows2_d <- as.double(rows2)
expected2 <- rowSums(X2[rows2, , drop = FALSE])

results$test2 <- run_test("2", sprintf("Shared big X + row sums (%dx%d, %d queries)", n_rows, n_cols, n_tasks2),
  mem_fn = function(cl, ns) {
    tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    on.exit({
      tryCatch(releaseVariables(ns, c("X", "rows")), error = function(e) NULL)
      tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    }, add = TRUE)

    registerVariables(ns, list(X = X2, rows = as.list(rows2_d)))
    res <- memLapply(
      X = "rows",
      FUN = function(r, X) sum(X[as.integer(r), ]),
      CLUSTER = cl,
      NAMESPACE = ns,
      VARS = c("X"),
      MAX.CORES = n_workers
    )
    stopifnot(all.equal(unlist(res, use.names = FALSE), expected2, tolerance = 1e-10))
  },
  shard_fn = function() {
    out <- buffer("double", dim = n_tasks2)
    shard_map(
      n_tasks2,
      borrow = list(X = X2_shard, rows = rows2),
      out = list(out = out),
      fun = function(sh, X, rows, out) {
        for (j in sh$idx) out[j] <- sum(X[rows[j], ])
        NULL
      },
      workers = n_workers,
      chunk_size = auto_chunk_size(n_tasks2, n_workers, target_chunks_per_worker = 2L, max = 8192L),
      profile = "speed",
      diagnostics = FALSE
    )
    stopifnot(all.equal(as.numeric(out[]), expected2, tolerance = 1e-10))
  }
)

# =============================================================================
# Test 3: Many small objects (list of small numeric vectors)
# =============================================================================
set.seed(3)
n_vecs <- env_int("SHARD_BENCH_N_VECS", 20000L)
vec_len <- env_int("SHARD_BENCH_VEC_LEN", 16L)
vec_list <- replicate(n_vecs, rnorm(vec_len), simplify = FALSE)
expected3 <- vapply(vec_list, sum, numeric(1))

results$test3 <- run_test("3", sprintf("Many small vectors (list of %d x len=%d)", n_vecs, vec_len),
  mem_fn = function(cl, ns) {
    tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    on.exit({
      tryCatch(releaseVariables(ns, c("vecs")), error = function(e) NULL)
      tryCatch(memshare_gc(namespace = ns, cluster = cl), error = function(e) NULL)
    }, add = TRUE)

    registerVariables(ns, list(vecs = vec_list))

    res <- memLapply(
      X = "vecs",
      # Avoid primitives here: some parallel/memshare paths end up touching
      # function environments, and primitives have a NULL environment.
      FUN = function(v) sum(v),
      CLUSTER = cl,
      NAMESPACE = ns,
      MAX.CORES = n_workers
    )
    stopifnot(all.equal(unlist(res, use.names = FALSE), expected3, tolerance = 1e-12))
  },
  shard_fn = function() {
    out <- buffer("double", dim = n_vecs)
    shard_map(
      n_vecs,
      # Intentionally unshared list input: this measures the cost of shipping
      # many small objects to workers when they can't be represented as one
      # contiguous shared slab.
      borrow = list(vecs = vec_list),
      out = list(out = out),
      fun = function(sh, vecs, out) {
        for (j in sh$idx) {
          out[j] <- sum(vecs[[j]])
        }
        NULL
      },
      workers = n_workers,
      chunk_size = auto_chunk_size(n_vecs, n_workers, target_chunks_per_worker = 2L, max = 4096L),
      profile = "speed",
      diagnostics = FALSE
    )
    stopifnot(all.equal(as.numeric(out[]), expected3, tolerance = 1e-12))
  }
)

# =============================================================================
# Summary
# =============================================================================
cat("============================================================\n")
cat("SUMMARY (ratio > 1 means shard faster)\n")
cat("============================================================\n\n")

cat(sprintf("%-40s %10s %10s %10s\n", "Test", "memshare", "shard", "Ratio"))
cat(strrep("-", 74), "\n")

test_names <- c(
  test1 = "1. Dispatch throughput",
  test2 = "2. Shared big X row sums",
  test3 = "3. Many small vectors"
)

for (name in names(results)) {
  r <- results[[name]]
  if (!is.na(r$mem) && !is.na(r$shard)) {
    ratio <- r$mem / r$shard
    winner <- if (ratio > 1.1) "SHARD" else if (ratio < 0.9) "memshare" else "~tie"
    cat(sprintf("%-40s %9.3fs %9.3fs %9.2fx  %s\n",
      test_names[name], r$mem, r$shard, ratio, winner
    ))
  } else {
    cat(sprintf("%-40s %9s %9s %9s\n",
      test_names[name],
      if (is.na(r$mem)) "FAILED" else sprintf("%.3fs", r$mem),
      if (is.na(r$shard)) "FAILED" else sprintf("%.3fs", r$shard),
      "N/A"
    ))
  }
}

cat("\nNotes:\n")
cat("- Each test uses fresh workers for fair comparison\n")
cat("- shard uses explicit output buffers (no result gathering)\n")
cat("- memshare gathers results into lists\n")
cat("\nBenchmark complete.\n")

# Benchmark / example: full crossprod(X, Y) via 2D tiles, block views, and an
# explicit output buffer.
#
# This exercises the "wow" path:
# - X and Y are shared once
# - each task constructs two block views (vX, vY)
# - compute uses BLAS-3 (dgemm) without materializing either view
# - results write directly into a shared output buffer (no gather + bind)
# - kernel dispatch avoids boilerplate: shard_map(..., kernel="crossprod_tile")
#
# Run from repo root:
#   R -q -f inst/bench/tiled_crossprod_buffer_shard_map.R

suppressPackageStartupMessages(library(shard))

set.seed(1)
n <- 6000L
p <- 128L
v <- 256L

X <- matrix(rnorm(n * p), nrow = n)
Y <- matrix(rnorm(n * v), nrow = n)

colnames(X) <- paste0("x", seq_len(ncol(X)))
colnames(Y) <- paste0("y", seq_len(ncol(Y)))

res <- shard_crossprod(
  X,
  Y,
  workers = 4,
  block_x = 32L,
  block_y = 64L,
  backing = "mmap",
  materialize = "always",
  diagnostics = TRUE
)

print(task_report(res$run))
print(copy_report(res$run))

out <- res$value
cat("\nResult dim:", paste(dim(out), collapse = "x"), "\n")
cat("Max abs error vs base crossprod:", max(abs(out - crossprod(X, Y))), "\n")

# Benchmark: shm_queue vs rpc_chunked overhead for tiny tasks
#
# This is a microbenchmark intended to exercise dispatch overhead, not compute.
# It uses a simple out-buffer write so results are not gathered.
#
# Run:
#   R -q -f inst/bench/shm_queue_overhead_shard_map.R

library(shard)

pool_stop()
on.exit(pool_stop(), add = TRUE)

workers <- 4L
n <- 20000L

tiny_kernel <- function(sh, out) {
  out[sh$idx] <- sh$idx
  NULL
}

run_rpc <- function() {
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")
  blocks <- shards(n, block_size = 1L, workers = workers)
  t <- system.time(
    shard_map(
      blocks,
      out = list(out = out),
      fun = tiny_kernel,
      workers = workers,
      chunk_size = 1L,
      autotune = "none",
      diagnostics = FALSE
    )
  )
  list(time = t[["elapsed"]], out = out)
}

run_shm <- function() {
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")
  t <- system.time(
    shard_map(
      n,
      out = list(out = out),
      fun = tiny_kernel,
      workers = workers,
      chunk_size = 1L,
      dispatch_mode = "shm_queue",
      dispatch_opts = list(block_size = 1L),
      diagnostics = FALSE
    )
  )
  list(time = t[["elapsed"]], out = out)
}

rpc <- run_rpc()
shm <- run_shm()

cat("n=", n, "workers=", workers, "\n")
cat("rpc_chunked elapsed:", rpc$time, "sec  (tasks/sec=", n / rpc$time, ")\n")
cat("shm_queue  elapsed:", shm$time, "sec  (tasks/sec=", n / shm$time, ")\n")

stopifnot(identical(as.integer(rpc$out[]), seq_len(n)))
stopifnot(identical(as.integer(shm$out[]), seq_len(n)))

# =============================================================================
# Benchmark: shard vs memshare - Apples-to-Apples Comparison
# =============================================================================
#
# This benchmark compares shard and memshare on equivalent operations where
# both packages offer similar functionality.
#
# Prerequisites:
#   install.packages("memshare")
#   install.packages("bench")  # optional, for better timing
#
# Run:
#   Rscript inst/bench/shard_vs_memshare_benchmark.R
#
# =============================================================================

suppressPackageStartupMessages({
  library(shard)
  skip_memshare <- tolower(Sys.getenv("SHARD_BENCH_SKIP_MEMSHARE", "")) %in% c("1", "true", "yes")
  if (!skip_memshare) {
    if (!requireNamespace("memshare", quietly = TRUE)) {
      stop("Please install memshare: install.packages('memshare')")
    }
    library(memshare)
  }
  library(parallel)
})

# -----------------------------------------------------------------------------
# Configuration
# -----------------------------------------------------------------------------

n_workers <- 4L
set.seed(42)
skip_memshare <- tolower(Sys.getenv("SHARD_BENCH_SKIP_MEMSHARE", "")) %in% c("1", "true", "yes")
bench_only_raw <- trimws(Sys.getenv("SHARD_BENCH_ONLY", ""))
bench_only_ids <- if (nzchar(bench_only_raw)) {
  suppressWarnings(as.integer(strsplit(bench_only_raw, ",", fixed = TRUE)[[1]]))
} else integer()
bench_only_ids <- bench_only_ids[is.finite(bench_only_ids)]
should_run <- function(id) length(bench_only_ids) == 0L || (id %in% bench_only_ids)

cat("=============================================================================\n")
cat("BENCHMARK: shard vs memshare\n")
cat("=============================================================================\n")
cat("Workers:", n_workers, "\n")
cat("Platform:", R.version$platform, "\n")
cat("R version:", R.version.string, "\n\n")
cat("shard version: ", as.character(utils::packageVersion("shard")), "\n", sep = "")
cat("shard path:    ", find.package("shard"), "\n\n", sep = "")
if (skip_memshare) {
  cat("NOTE: SHARD_BENCH_SKIP_MEMSHARE is set; memshare benchmarks will be skipped.\n\n")
}
if (length(bench_only_ids) > 0) {
  cat("NOTE: SHARD_BENCH_ONLY is set; running only benchmarks: ",
      paste(bench_only_ids, collapse = ", "),
      "\n\n", sep = "")
}

# Cleanup function
cleanup <- function() {
  suppressWarnings({
    tryCatch(pool_stop(), error = function(e) NULL)
    gc(verbose = FALSE)
  })
}

on.exit(cleanup(), add = TRUE)

# -----------------------------------------------------------------------------
# Helper: Run benchmark with timing
# -----------------------------------------------------------------------------

run_bench <- function(name, fn, times = 3, min_elapsed = 0.05, max_reps = 4096L) {
  cat("  Running", name, "...\n")
  if (!is.function(fn)) stop("run_bench: fn must be a function()", call. = FALSE)
  timings <- numeric(times)
  result <- NULL
  for (i in seq_len(times)) {
    gc(verbose = FALSE)
    reps <- 1L
    elapsed <- 0
    repeat {
      t <- system.time({
        for (k in seq_len(reps)) {
          result <- fn()
        }
      })
      elapsed <- t[["elapsed"]]
      if (!is.finite(elapsed)) break
      if (elapsed >= min_elapsed) break
      if (reps >= as.integer(max_reps)) break
      reps <- reps * 2L
    }
    timings[i] <- elapsed / max(1L, reps)
  }
  list(
    name = name,
    median = median(timings),
    min = min(timings),
    max = max(timings),
    result = result
  )
}

fmt_s <- function(x) sprintf("%.6f", x)

fmt_speedup <- function(mem, shard) {
  if (is.null(mem) || is.null(shard)) return("NA")
  if (!is.finite(mem) || !is.finite(shard) || shard <= 0) return("NA")
  sprintf("%.2fx", mem / shard)
}

unique_ns <- function(prefix) {
  suffix <- paste(sample(c(letters, 0:9), 6, replace = TRUE), collapse = "")
  paste0(prefix, "_", Sys.getpid(), "_", suffix)
}

memshare_result_ok <- function(x, expected_len) {
  if (is.null(x)) return(FALSE)
  if (is.null(x$result)) return(FALSE)
  if (!is.numeric(x$result)) return(FALSE)
  if (length(x$result) != expected_len) return(FALSE)
  if (anyNA(x$result)) return(FALSE)
  TRUE
}

check_equal <- function(label, actual, expected, tolerance = 1e-10) {
  ok <- isTRUE(all.equal(actual, expected, tolerance = tolerance))
  if (!ok) {
    delta <- tryCatch(
      mean(abs(as.numeric(actual) - as.numeric(expected))),
      error = function(e) NA_real_
    )
    cat("  NOTE: validation failed for ", label, ".\n", sep = "")
    if (is.finite(delta)) {
      cat("    Mean absolute difference: ", delta, "\n", sep = "")
    }
    if (!skip_memshare) {
      cat("    If memshare is built for a different R version, try re-running with:\n")
      cat("      SHARD_BENCH_SKIP_MEMSHARE=1 Rscript inst/bench/shard_vs_memshare_benchmark.R\n")
    }
  }
  ok
}

# =============================================================================
# BENCHMARK 1: Parallel Column-wise Apply (Column Means)
# =============================================================================
# memshare: memApply(X, MARGIN=2, FUN)
# shard:    shard_map with column views or shard_apply_matrix
# =============================================================================

cat("\n-----------------------------------------------------------------------------\n")
cat("BENCHMARK 1: Parallel Column-wise Apply (Column Means)\n")
cat("-----------------------------------------------------------------------------\n")

bench1_run <- function(n_rows, n_cols) {
  cat("\nMatrix size:", n_rows, "x", n_cols, "\n")

  X <- matrix(rnorm(n_rows * n_cols), nrow = n_rows, ncol = n_cols)

  # Ground truth
  expected <- colMeans(X)

  # --- memshare ---
  memshare_result <- NULL
  if (!skip_memshare) {
    cleanup()
    cl <- makeCluster(n_workers)
    on.exit(tryCatch(stopCluster(cl), error = function(e) NULL), add = TRUE)
    ns <- unique_ns("b1")
    memshare_result <- tryCatch({
      # memshare::memApply has been flaky on some macOS/R setups (registerVariables
      # collisions and cleanup errors). Use a stable equivalent based on
      # memLapply over column indices + a registered X.
      registerVariables(ns, list(X = X))
      run_bench("memshare::memLapply (colMeans)", function() {
        res <- memLapply(
          X = as.list(as.double(seq_len(n_cols))),
          FUN = function(j, X) {
            j <- as.integer(j)
            mean(X[, j])
          },
          CLUSTER = cl,
          NAMESPACE = ns,
          VARS = "X",
          MAX.CORES = n_workers
        )
        out <- unlist(res, use.names = FALSE)
        if (length(out) != n_cols || anyNA(out)) stop("memshare returned invalid result", call. = FALSE)
        out
      }, times = 1)
    },
    error = function(e) {
      cat("  NOTE: memshare failed; skipping memshare timings.\n")
      cat("    ", conditionMessage(e), "\n", sep = "")
      NULL
    },
    finally = {
      tryCatch(releaseVariables(ns, c("X")), error = function(e) NULL)
      tryCatch(memshare_gc(ns, cluster = cl), error = function(e) NULL)
      tryCatch(stopCluster(cl), error = function(e) NULL)
    })
  }

  # --- shard (shard_map with column iteration) ---
  cleanup()

  X_shared <- share(X)
  out <- buffer("double", dim = n_cols)

  shard_result <- run_bench("shard::shard_map", function() {
    shard_map(
      n_cols,
      borrow = list(X = X_shared),
      out = list(out = out),
      kernel = "col_means",
      workers = n_workers,
      profile = "speed",
      diagnostics = FALSE
    )
    as.numeric(out[])
  })

  # --- Validate ---
  if (memshare_result_ok(memshare_result, n_cols)) {
    check_equal("memshare bench1", memshare_result$result, expected, tolerance = 1e-10)
  }
  check_equal("shard bench1", shard_result$result, expected, tolerance = 1e-10)

  cat("\n  Results:\n")
  if (memshare_result_ok(memshare_result, n_cols)) {
    cat("    memshare:  ", fmt_s(memshare_result$median), "s (median)\n", sep = "")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
    cat("    Speedup:   ", fmt_speedup(memshare_result$median, shard_result$median), "\n", sep = "")
  } else {
    cat("    memshare:  SKIPPED\n")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
  }

  list(memshare = memshare_result, shard = shard_result)
}

# Run with different sizes
bench1_small <- if (should_run(1)) bench1_run(1000, 500) else NULL
bench1_large <- if (should_run(1)) bench1_run(10000, 1000) else NULL


# =============================================================================
# BENCHMARK 2: Parallel Lapply Over List of Matrices
# =============================================================================
# memshare: memLapply(list, FUN)
# shard:    shard_map with shards_list
# =============================================================================

cat("\n-----------------------------------------------------------------------------\n")
cat("BENCHMARK 2: Parallel Lapply Over List of Matrices\n")
cat("-----------------------------------------------------------------------------\n")

bench2_run <- function(n_items, mat_size) {
  cat("\nList length:", n_items, ", Matrix size:", mat_size, "x", mat_size, "\n")

  # Create list of matrices
  mat_list <- lapply(seq_len(n_items), function(i) {
    matrix(rnorm(mat_size * mat_size), nrow = mat_size, ncol = mat_size)
  })

  # Ground truth: Frobenius norm of each matrix
  expected <- vapply(mat_list, function(m) sqrt(sum(m * m)), numeric(1))

  # --- memshare ---
  cleanup()
  cl <- makeCluster(n_workers)

  ns <- unique_ns("b2")
  memshare_result <- NULL
  if (!skip_memshare) {
    memshare_result <- tryCatch({
      registerVariables(ns, list(mats = mat_list))
      run_bench("memshare::memLapply", function() {
        res <- memLapply(
          X = "mats",
          FUN = function(m) sqrt(sum(m * m)),
          CLUSTER = cl,
          NAMESPACE = ns,
          MAX.CORES = n_workers
        )
        out <- unlist(res)
        if (length(out) != n_items || anyNA(out)) stop("memshare::memLapply returned invalid result", call. = FALSE)
        out
      }, times = 1)
    },
    error = function(e) {
      cat("  NOTE: memshare failed; skipping memshare timings.\n")
      cat("    ", conditionMessage(e), "\n", sep = "")
      NULL
    },
    finally = {
      tryCatch(releaseVariables(ns, c("mats")), error = function(e) NULL)
      tryCatch(memshare_gc(ns, cluster = cl), error = function(e) NULL)
    }
  )

    stopCluster(cl)
    tryCatch(memshare_gc(ns, cluster = NULL), error = function(e) NULL)
  }

  # --- shard ---
  cleanup()

  # Sharing each matrix as an ALTREP-backed segment does not scale: with many
  # matrices it can exhaust worker file descriptors / segment handles during
  # unserialization (macOS often fails with "Failed to open ... ALTREP
  # unserialize"). For this benchmark we share the list container as ONE
  # serialized object and fetch it once per shard.
  mat_list_shared <- share(mat_list)
  out <- buffer("double", dim = n_items)

  shard_result <- run_bench("shard::shard_map", function() {
    shard_map(
      shards(n_items, block_size = ceiling(n_items / n_workers), workers = n_workers),
      borrow = list(mats = mat_list_shared),
      out = list(out = out),
      fun = function(sh, mats, out) {
        mats0 <- fetch(mats)
        for (i in sh$idx) {
          m <- mats0[[i]]
          out[i] <- sqrt(sum(m * m))
        }
        NULL
      },
      workers = n_workers,
      profile = "speed",
      diagnostics = FALSE
    )
    as.numeric(out[])
  })

  # --- Validate ---
  if (memshare_result_ok(memshare_result, n_items)) {
    check_equal("memshare bench2", memshare_result$result, expected, tolerance = 1e-10)
  }
  check_equal("shard bench2", shard_result$result, expected, tolerance = 1e-10)

  cat("\n  Results:\n")
  if (memshare_result_ok(memshare_result, n_items)) {
    cat("    memshare:  ", fmt_s(memshare_result$median), "s (median)\n", sep = "")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
    cat("    Speedup:   ", fmt_speedup(memshare_result$median, shard_result$median), "\n", sep = "")
  } else {
    cat("    memshare:  FAILED\n")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
  }

  list(memshare = memshare_result, shard = shard_result)
}

bench2_small <- if (should_run(2)) bench2_run(100, 50) else NULL
bench2_large <- if (should_run(2)) bench2_run(500, 100) else NULL


# =============================================================================
# BENCHMARK 3: Many Small Tasks (Dispatch Overhead)
# =============================================================================
# This tests the overhead of dispatching many tiny tasks
# =============================================================================

cat("\n-----------------------------------------------------------------------------\n")
cat("BENCHMARK 3: Many Small Tasks (Dispatch Overhead)\n")
cat("-----------------------------------------------------------------------------\n")

bench3_run <- function(n_tasks) {
  cat("\nNumber of tasks:", n_tasks, "\n")

  # Ground truth
  expected <- seq_len(n_tasks)^2

  # --- memshare ---
  cleanup()
  cl <- makeCluster(n_workers)

  ns <- unique_ns("b3")
  memshare_result <- NULL
  if (!skip_memshare) {
    memshare_result <- tryCatch(
      run_bench("memshare::memLapply", function() {
        res <- memLapply(
          X = as.list(as.double(seq_len(n_tasks))),
          FUN = function(x) x^2,
          CLUSTER = cl,
          NAMESPACE = ns,
          MAX.CORES = n_workers
        )
        out <- unlist(res)
        if (length(out) != n_tasks || anyNA(out)) stop("memshare::memLapply returned invalid result", call. = FALSE)
        out
      }, times = 1),
      error = function(e) {
        cat("  NOTE: memshare failed; skipping memshare timings.\n")
        cat("    ", conditionMessage(e), "\n", sep = "")
        NULL
      },
      finally = {
        tryCatch(memshare_gc(ns, cluster = cl), error = function(e) NULL)
      }
    )

    stopCluster(cl)
    tryCatch(memshare_gc(ns, cluster = NULL), error = function(e) NULL)
  }

  # --- shard (rpc_chunked mode) ---
  cleanup()

  out <- buffer("double", dim = n_tasks)

  shard_rpc_result <- run_bench("shard::shard_map (rpc_chunked)", function() {
    shard_map(
      shards(n_tasks, workers = n_workers),
      out = list(out = out),
      fun = function(sh, out) {
        for (i in sh$idx) {
          out[i] <- i^2
        }
        NULL
      },
      workers = n_workers,
      diagnostics = FALSE
    )
    as.numeric(out[])
  })

  # --- shard (shm_queue mode - if tiny tasks) ---
  shard_shm_result <- NULL
  if (n_tasks >= 1000) {
    cleanup()

    out2 <- buffer("integer", dim = n_tasks, init = 0L)

    shard_shm_result <- run_bench("shard::shard_map (shm_queue)", function() {
      shard_map(
        n_tasks,
        out = list(out = out2),
        fun = function(sh, out) {
          out[sh$idx] <- as.integer(sh$idx^2)
          NULL
        },
        workers = n_workers,
        chunk_size = 1L,
        dispatch_mode = "shm_queue",
        dispatch_opts = list(block_size = 1L),
        diagnostics = FALSE
      )
      as.numeric(out2[])
    })
  }

  # --- Validate ---
  if (memshare_result_ok(memshare_result, n_tasks)) {
    check_equal("memshare bench3", memshare_result$result, expected, tolerance = 1e-10)
  }
  check_equal("shard bench3 rpc", shard_rpc_result$result, expected, tolerance = 1e-10)
  if (!is.null(shard_shm_result)) {
    check_equal("shard bench3 shm_queue", shard_shm_result$result, expected, tolerance = 1e-10)
  }

  cat("\n  Results:\n")
  if (memshare_result_ok(memshare_result, n_tasks)) {
    cat("    memshare:           ", fmt_s(memshare_result$median), "s (median)\n", sep = "")
  } else {
    cat("    memshare:           FAILED\n")
  }
  cat("    shard (rpc):        ", fmt_s(shard_rpc_result$median), "s (median)\n", sep = "")
  if (!is.null(shard_shm_result)) {
    cat("    shard (shm_queue):  ", fmt_s(shard_shm_result$median), "s (median)\n", sep = "")
    if (memshare_result_ok(memshare_result, n_tasks)) {
      cat("    Speedup vs memshare:", fmt_speedup(memshare_result$median, shard_shm_result$median), "\n")
    }
  }

  list(memshare = memshare_result, shard_rpc = shard_rpc_result, shard_shm = shard_shm_result)
}

bench3_small <- if (should_run(3)) bench3_run(500) else NULL
bench3_large <- if (should_run(3)) bench3_run(5000) else NULL


# =============================================================================
# BENCHMARK 4: Large Shared Matrix Access Pattern
# =============================================================================
# Tests zero-copy efficiency when multiple workers read from shared matrix
# =============================================================================

cat("\n-----------------------------------------------------------------------------\n")
cat("BENCHMARK 4: Large Shared Matrix - Row Subsetting\n")
cat("-----------------------------------------------------------------------------\n")

bench4_run <- function(n_rows, n_cols, n_subsets) {
  cat("\nMatrix:", n_rows, "x", n_cols, ", Subsets:", n_subsets, "\n")

  X <- matrix(rnorm(n_rows * n_cols), nrow = n_rows, ncol = n_cols)

  # Random row subsets
  subset_size <- n_rows %/% 10
  subsets <- lapply(seq_len(n_subsets), function(i) {
    sample(n_rows, subset_size)
  })

  # Ground truth
  expected <- sapply(subsets, function(idx) sum(X[idx, ]))

  # --- memshare ---
  memshare_result <- NULL
  if (!skip_memshare) {
    cleanup()
    ns <- unique_ns("b4")
    cl <- NULL
    memshare_result <- tryCatch({
      cl <- makeCluster(n_workers)
      registerVariables(ns, list(X = X))
      run_bench("memshare (registered)", function() {
        res <- memLapply(
          X = lapply(subsets, as.double),
          FUN = function(idx, X) {
            sum(X[as.integer(idx), ])
          },
          CLUSTER = cl,
          NAMESPACE = ns,
          VARS = "X",
          MAX.CORES = n_workers
        )
        out <- unlist(res)
        if (length(out) != n_subsets || anyNA(out)) stop("memshare::memLapply returned invalid result", call. = FALSE)
        out
      }, times = 1)
    }, error = function(e) {
      cat("  NOTE: memshare failed; skipping memshare timings.\n")
      cat("    ", conditionMessage(e), "\n", sep = "")
      NULL
    }, finally = {
      if (!is.null(cl)) {
        tryCatch(memshare_gc(ns, cluster = cl), error = function(e) NULL)
        tryCatch(stopCluster(cl), error = function(e) NULL)
      }
      tryCatch(memshare_gc(ns, cluster = NULL), error = function(e) NULL)
    })
  }

  # --- shard ---
  cleanup()

  X_shared <- share(X)
  out <- buffer("double", dim = n_subsets)

  # Share the subsets list as a single serialized object (one segment). Deep
  # sharing turns this into many components and is unnecessary here.
  subsets_shared <- share(subsets)

  shard_result <- run_bench("shard::share + shard_map", function() {
    shard_map(
      shards(n_subsets, block_size = ceiling(n_subsets / n_workers), workers = n_workers),
      borrow = list(X = X_shared, subsets = subsets_shared),
      out = list(out = out),
      fun = function(sh, X, subsets, out) {
        subs <- fetch(subsets)
        for (i in sh$idx) {
          out[i] <- sum(X[subs[[i]], ])
        }
        NULL
      },
      workers = n_workers,
      diagnostics = FALSE
    )
    as.numeric(out[])
  })

  # --- Validate ---
  if (memshare_result_ok(memshare_result, n_subsets)) {
    check_equal("memshare bench4", memshare_result$result, expected, tolerance = 1e-10)
  }
  check_equal("shard bench4", shard_result$result, expected, tolerance = 1e-10)

  cat("\n  Results:\n")
  if (memshare_result_ok(memshare_result, n_subsets)) {
    cat("    memshare:  ", fmt_s(memshare_result$median), "s (median)\n", sep = "")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
    cat("    Speedup:   ", fmt_speedup(memshare_result$median, shard_result$median), "\n", sep = "")
  } else {
    cat("    memshare:  FAILED\n")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
  }

  list(memshare = memshare_result, shard = shard_result)
}

bench4_result <- if (should_run(4)) bench4_run(5000, 200, 100) else NULL


# =============================================================================
# BENCHMARK 5: Cross-product Style Operation
# =============================================================================
# X'Y computation - tests BLAS-3 efficiency with shared data
# =============================================================================

cat("\n-----------------------------------------------------------------------------\n")
cat("BENCHMARK 5: Cross-product (X'Y) with Shared Matrices\n")
cat("-----------------------------------------------------------------------------\n")

bench5_run <- function(n, p, v) {
  cat("\nX:", n, "x", p, ", Y:", n, "x", v, "\n")

  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  Y <- matrix(rnorm(n * v), nrow = n, ncol = v)

  # Ground truth
  expected <- crossprod(X, Y)

  # Tile sizes (used by both implementations)
  block_p <- ceiling(p / 2)
  block_v <- ceiling(v / 2)

  tiles <- list()
  for (i in seq(1, p, by = block_p)) {
    for (j in seq(1, v, by = block_v)) {
      i_end <- min(i + block_p - 1, p)
      j_end <- min(j + block_v - 1, v)
      tiles <- c(tiles, list(as.double(c(i, i_end, j, j_end))))
    }
  }

  # --- memshare: Manual tiled crossprod ---
  memshare_result <- NULL
  if (!skip_memshare) {
    cleanup()
    ns <- unique_ns("b5")
    cl <- NULL
    memshare_result <- tryCatch({
      cl <- makeCluster(n_workers)
      registerVariables(ns, list(X = X, Y = Y))

      run_bench("memshare (manual tiles)", function() {
        res <- memLapply(
          X = tiles,
          FUN = function(tile, X, Y) {
            i_idx <- as.integer(tile[1]):as.integer(tile[2])
            j_idx <- as.integer(tile[3]):as.integer(tile[4])
            crossprod(X[, i_idx, drop = FALSE], Y[, j_idx, drop = FALSE])
          },
          CLUSTER = cl,
          NAMESPACE = ns,
          VARS = c("X", "Y"),
          MAX.CORES = n_workers
        )

        # Assemble result
        out <- matrix(0, nrow = p, ncol = v)
        for (k in seq_along(tiles)) {
          tile <- tiles[[k]]
          i_idx <- as.integer(tile[1]):as.integer(tile[2])
          j_idx <- as.integer(tile[3]):as.integer(tile[4])
          out[i_idx, j_idx] <- res[[k]]
        }
        out
      }, times = 1, min_elapsed = 0.1)
    }, error = function(e) {
      cat("  NOTE: memshare failed; skipping memshare timings.\n")
      cat("    ", conditionMessage(e), "\n", sep = "")
      NULL
    }, finally = {
      if (!is.null(cl)) {
        tryCatch(memshare_gc(ns, cluster = cl), error = function(e) NULL)
        tryCatch(stopCluster(cl), error = function(e) NULL)
      }
      tryCatch(memshare_gc(ns, cluster = NULL), error = function(e) NULL)
    })
  }

  # --- shard: shard_crossprod ---
  cleanup()

  # Share inputs once so the pool never holds stale references to segments that
  # have been GC'd on the master across repeated runs.
  X_shared <- share(X, backing = "mmap")
  Y_shared <- share(Y, backing = "mmap")

  shard_result <- run_bench("shard::shard_crossprod", function() {
    res <- shard_crossprod(
      X_shared, Y_shared,
      workers = n_workers,
      block_x = block_p,
      block_y = block_v,
      backing = "mmap",
      materialize = "always",
      diagnostics = FALSE
    )
    res$value
  })

  # --- Validate ---
  if (memshare_result_ok(memshare_result, p * v)) {
    check_equal("memshare bench5", memshare_result$result, expected, tolerance = 1e-8)
  }
  check_equal("shard bench5", shard_result$result, expected, tolerance = 1e-8)

  cat("\n  Results:\n")
  if (memshare_result_ok(memshare_result, p * v)) {
    cat("    memshare:  ", fmt_s(memshare_result$median), "s (median)\n", sep = "")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
    cat("    Speedup:   ", fmt_speedup(memshare_result$median, shard_result$median), "\n", sep = "")
  } else {
    cat("    memshare:  FAILED\n")
    cat("    shard:     ", fmt_s(shard_result$median), "s (median)\n", sep = "")
  }

  list(memshare = memshare_result, shard = shard_result)
}

bench5_result <- if (should_run(5)) bench5_run(3000, 64, 128) else NULL


# =============================================================================
# BENCHMARK 6: Memory Efficiency (Peak RSS)
# =============================================================================
# Compare peak memory usage during parallel operations
# =============================================================================

cat("\n-----------------------------------------------------------------------------\n")
cat("BENCHMARK 6: Memory Efficiency During Parallel Operations\n")
cat("-----------------------------------------------------------------------------\n")

bench6_run <- function(n_rows, n_cols) {
  cat("\nMatrix size:", n_rows, "x", n_cols,
      "(", round(n_rows * n_cols * 8 / 1024^2, 1), "MB)\n")

  X <- matrix(rnorm(n_rows * n_cols), nrow = n_rows, ncol = n_cols)
  mat_size_mb <- n_rows * n_cols * 8 / 1024^2

  # --- shard with memory reporting ---
  cleanup()
  gc(verbose = FALSE)

  cat("\n  Running shard with diagnostics...\n")

  tryCatch({
    X_shared <- share(X)
    out <- buffer("double", dim = n_cols)

    res <- shard_map(
      shards(n_cols, workers = n_workers),
      borrow = list(X = X_shared),
      out = list(out = out),
      fun = function(sh, X, out) {
        # Keep a stable worker set for diagnostics runs: if a worker is recycled
        # mid-call, it can miss the borrowed inputs and fail chunks.
        for (j in sh$idx) {
          out[j] <- sum(X[, j]^2)
        }
        NULL
      },
      workers = n_workers,
      profile = "speed",
      recycle = FALSE,
      diagnostics = TRUE
    )

    mem <- tryCatch(suppressWarnings(mem_report()), error = function(e) NULL)
    copy <- tryCatch(suppressWarnings(copy_report(res)), error = function(e) NULL)

    main_peak_mb <- NA_real_
    worker_peak_mb <- NA_real_
    if (!is.null(mem)) {
      main_peak_mb <- suppressWarnings(as.double(mem$main$peak_mb %||% NA_real_))
      peaks <- tryCatch(
        vapply(mem$workers, function(w) suppressWarnings(as.double(w$peak_mb %||% NA_real_)), numeric(1)),
        error = function(e) numeric()
      )
      peaks <- peaks[is.finite(peaks)]
      if (length(peaks) > 0) worker_peak_mb <- mean(peaks)
    }

    if (is.finite(main_peak_mb) && is.finite(worker_peak_mb) && !is.null(copy)) {
      cat("\n  shard Memory Report:\n")
      cat("    Main process peak RSS:  ", round(main_peak_mb, 1), "MB\n")
      cat("    Worker peak RSS (mean): ", round(worker_peak_mb, 1), "MB\n")
      cat("    View materialized bytes:", copy$view_materialized_bytes, "\n")
      cat("    Buffer write bytes:     ", copy$buffer_write_bytes, "\n")
    } else {
      cat("\n  NOTE: Memory diagnostics unavailable for this run.\n")
    }

    cat("\n  Note: memshare doesn't provide equivalent memory diagnostics.\n")
    cat("  shard's explicit buffer approach avoids result gathering overhead.\n")

    list(
      shard_main_peak_mb = main_peak_mb,
      shard_worker_peak_mb = worker_peak_mb,
      matrix_size_mb = mat_size_mb
    )
  }, error = function(e) {
    cat("\n  NOTE: shard diagnostics run failed.\n")
    cat("    ", conditionMessage(e), "\n", sep = "")
    NULL
  })
}

bench6_result <- if (should_run(6)) bench6_run(5000, 500) else NULL


# =============================================================================
# SUMMARY
# =============================================================================

cat("\n=============================================================================\n")
cat("SUMMARY\n")
cat("=============================================================================\n\n")

cat("Benchmark results show relative performance (speedup > 1 means shard is faster):\n\n")

cat("1. Column-wise Apply:\n")
cat("   - Small matrix: shard ", fmt_speedup(bench1_small$memshare$median, bench1_small$shard$median), "\n", sep = "")
cat("   - Large matrix: shard ", fmt_speedup(bench1_large$memshare$median, bench1_large$shard$median), "\n", sep = "")

cat("\n2. Lapply Over List:\n")
cat("   - Small list:   shard ", fmt_speedup(bench2_small$memshare$median, bench2_small$shard$median), "\n", sep = "")
cat("   - Large list:   shard ", fmt_speedup(bench2_large$memshare$median, bench2_large$shard$median), "\n", sep = "")

cat("\n3. Many Small Tasks:\n")
cat("   - 500 tasks:    shard (rpc) ", fmt_speedup(bench3_small$memshare$median, bench3_small$shard_rpc$median), "\n", sep = "")
if (!is.null(bench3_large$shard_shm)) {
  cat("   - 5000 tasks:   shard (shm_queue) ", fmt_speedup(bench3_large$memshare$median, bench3_large$shard_shm$median), "\n", sep = "")
}

cat("\n4. Large Shared Matrix Access:\n")
cat("   - shard ", fmt_speedup(bench4_result$memshare$median, bench4_result$shard$median), "\n", sep = "")

cat("\n5. Cross-product (X'Y):\n")
cat("   - shard ", fmt_speedup(bench5_result$memshare$median, bench5_result$shard$median), "\n", sep = "")

cat("\n6. Memory Efficiency:\n")
cat("   - shard provides detailed diagnostics (mem_report, copy_report)\n")
cat("   - memshare lacks equivalent memory tracking\n")

cat("\n=============================================================================\n")
cat("KEY DIFFERENTIATORS:\n")
cat("=============================================================================\n")
cat("
shard advantages:
  - Explicit output buffers eliminate result gathering overhead
  - shm_queue mode provides 5-20x speedup for tiny tasks
  - Zero-copy views for matrix operations
  - Comprehensive diagnostics (mem_report, copy_report, recommendations)
  - Worker supervision with automatic recycling
  - Immutability by default prevents silent copies

memshare advantages:
  - Simpler API (fewer parameters)
  - Direct integration with parallel::makeCluster
  - Allows mutable shared data
")

cleanup()
cat("\nBenchmark complete.\n")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)

## ----load-shard---------------------------------------------------------------
library(shard)

## ----first-example------------------------------------------------------------
set.seed(42)
X <- matrix(rnorm(5000), nrow = 100, ncol = 50)

# Share the matrix (zero-copy for workers)
X_shared <- share(X)

# Allocate an output buffer
out <- buffer("double", dim = ncol(X))

# Define column shards and run
blocks <- shards(ncol(X), workers = 2)
run <- shard_map(
  blocks,
  borrow = list(X = X_shared),
  out    = list(out = out),
  workers = 2,
  fun = function(shard, X, out) {
    for (j in shard$idx) {
      out[j] <- mean(X[, j])
    }
  }
)

# Read results from the buffer
result <- out[]
head(result)

## ----share-input--------------------------------------------------------------
X_shared <- share(X)
is_shared(X_shared)
shared_info(X_shared)

## ----buffer-demo--------------------------------------------------------------
buf <- buffer("double", dim = c(10, 5))
buf[1:5, 1] <- rnorm(5)
buf[6:10, 1] <- rnorm(5)
buf[, 1]

## ----buffer-types-------------------------------------------------------------
int_buf <- buffer("integer", dim = 100)
mat_buf <- buffer("double", dim = c(50, 20))

## ----shards-demo--------------------------------------------------------------
blocks <- shards(1000, workers = 4)
blocks

## ----shard-indices------------------------------------------------------------
blocks[[1]]$idx[1:10]  # first 10 indices of shard 1

## ----shard-map-example--------------------------------------------------------
set.seed(1)
X <- matrix(rnorm(2000), nrow = 100, ncol = 20)
X_shared <- share(X)
col_sds <- buffer("double", dim = ncol(X))

blocks <- shards(ncol(X), workers = 2)
run <- shard_map(
  blocks,
  borrow  = list(X = X_shared),
  out     = list(col_sds = col_sds),
  workers = 2,
  fun = function(shard, X, col_sds) {
    for (j in shard$idx) {
      col_sds[j] <- sd(X[, j])
    }
  }
)

# Results are already in the buffer
sd_values <- col_sds[]

# Verify against base R
all.equal(sd_values, apply(X, 2, sd))

## ----return-values------------------------------------------------------------
blocks <- shards(10, workers = 2)
run <- shard_map(
  blocks,
  workers = 2,
  fun = function(shard) {
    sum(shard$idx)
  }
)
results(run)

## ----apply-matrix-------------------------------------------------------------
set.seed(1)
X <- matrix(rnorm(2000), nrow = 100, ncol = 20)
y <- rnorm(100)

# Correlate each column of X with y
cors <- shard_apply_matrix(
  X,
  MARGIN = 2,
  FUN = function(v, y) cor(v, y),
  VARS = list(y = y),
  workers = 2
)
head(cors)

## ----lapply-shared------------------------------------------------------------
chunks <- lapply(1:10, function(i) rnorm(100))

means <- shard_lapply_shared(
  chunks,
  FUN = mean,
  workers = 2
)
unlist(means)

## ----diagnostics--------------------------------------------------------------
report(result = run)

## ----pool-management, eval = FALSE--------------------------------------------
# # Create a pool with 4 workers and a 1GB memory cap
# pool_create(n = 4, rss_limit = "1GB")
# 
# # Check pool health
# pool_status()
# 
# # Run multiple shard_map() calls (reuses the same pool)
# run1 <- shard_map(shards(1000), workers = 4, fun = function(s) sum(s$idx))
# run2 <- shard_map(shards(500),  workers = 4, fun = function(s) mean(s$idx))
# 
# # Shut down workers when done
# pool_stop()

## ----cow-deny, eval = FALSE---------------------------------------------------
# shard_map(
#   shards(10),
#   borrow = list(X = share(matrix(1:100, 10, 10))),
#   workers = 2,
#   cow = "deny",
#   fun = function(shard, X) {
#     X[1, 1] <- 999  # Error: mutation denied
#   }
# )

## ----cleanup, include = FALSE-------------------------------------------------
try(pool_stop(), silent = TRUE)

## ----cleanup-show, eval = FALSE-----------------------------------------------
# pool_stop()

#' @title RSS Monitoring Utilities
#' @description Cross-platform utilities for monitoring process memory usage.
#' @name rss
NULL

#' Get RSS for a Process ID
#'
#' Retrieves the Resident Set Size (RSS) for a process using the best
#' available method for the current platform.
#'
#' @param pid Integer. Process ID to query.
#' @return Numeric. RSS in bytes, or NA if unavailable.
#' @keywords internal
#' @noRd
rss_get_pid <- function(pid) {
  if (is.na(pid)) {
    return(NA_real_)
  }


  # Try ps package first (most reliable, cross-platform)
  if (requireNamespace("ps", quietly = TRUE)) {
    rss <- rss_via_ps(pid)
    if (!is.na(rss)) return(rss)
  }

  # Platform-specific fallbacks
  os <- .Platform$OS.type
  sysname <- Sys.info()[["sysname"]]

  if (sysname == "Linux") {
    rss <- rss_via_proc(pid)
    if (!is.na(rss)) return(rss)
  }

  if (sysname %in% c("Darwin", "Linux")) {
    rss <- rss_via_ps_cmd(pid)
    if (!is.na(rss)) return(rss)
  }

  if (os == "windows") {
    rss <- rss_via_wmic(pid)
    if (!is.na(rss)) return(rss)
  }

  NA_real_
}

#' Get RSS via ps Package
#'
#' @param pid Process ID.
#' @return RSS in bytes or NA.
#' @keywords internal
#' @noRd
rss_via_ps <- function(pid) {
  tryCatch({
    p <- ps::ps_handle(pid)
    if (!ps::ps_is_running(p)) {
      return(NA_real_)
    }
    mem <- ps::ps_memory_info(p)
    as.numeric(mem[["rss"]])
  }, error = function(e) NA_real_)
}

#' Get RSS via /proc filesystem (Linux)
#'
#' @param pid Process ID.
#' @return RSS in bytes or NA.
#' @keywords internal
#' @noRd
rss_via_proc <- function(pid) {
  statm_path <- file.path("/proc", pid, "statm")

  if (!file.exists(statm_path)) {
    return(NA_real_)
  }

  tryCatch({
    # statm format: size resident shared text lib data dt
    # All values in pages
    statm <- scan(statm_path, what = integer(), quiet = TRUE, n = 2)
    if (length(statm) < 2) return(NA_real_)

    # Get page size
    page_size <- tryCatch(
      as.integer(system("getconf PAGESIZE", intern = TRUE)),
      error = function(e) 4096L  # Default to 4KB
    )

    as.numeric(statm[2]) * page_size
  }, error = function(e) NA_real_)
}

#' Get RSS via ps Command (Unix)
#'
#' @param pid Process ID.
#' @return RSS in bytes or NA.
#' @keywords internal
#' @noRd
rss_via_ps_cmd <- function(pid) {
  tryCatch({
    # ps -o rss= gives RSS in KB
    output <- withCallingHandlers(
      system2(
        "ps",
        args = c("-o", "rss=", "-p", as.character(pid)),
        stdout = TRUE,
        stderr = FALSE
      ),
      warning = function(w) invokeRestart("muffleWarning")
    )
    status <- attr(output, "status", exact = TRUE)
    if (!is.null(status) && !identical(as.integer(status), 0L)) {
      return(NA_real_)
    }
    if (length(output) == 0 || nchar(trimws(output[1])) == 0) {
      return(NA_real_)
    }
    rss_kb <- as.numeric(trimws(output[1]))
    if (is.na(rss_kb)) return(NA_real_)
    rss_kb * 1024  # Convert to bytes
  }, error = function(e) NA_real_)
}

#' Get RSS via WMIC (Windows)
#'
#' @param pid Process ID.
#' @return RSS in bytes or NA.
#' @keywords internal
#' @noRd
rss_via_wmic <- function(pid) {
  tryCatch({
    cmd <- sprintf("wmic process where ProcessId=%d get WorkingSetSize /value", pid)
    output <- shell(cmd, intern = TRUE, mustWork = FALSE)
    # Parse WorkingSetSize=NNNN
    ws_line <- grep("WorkingSetSize=", output, value = TRUE)
    if (length(ws_line) == 0) return(NA_real_)
    ws <- sub("WorkingSetSize=", "", ws_line[1])
    as.numeric(trimws(ws))
  }, error = function(e) NA_real_)
}

#' Get Current Process RSS
#'
#' @return RSS in bytes for the current R process.
#' @keywords internal
#' @noRd
rss_self <- function() {
  rss_get_pid(Sys.getpid())
}

#' Monitor RSS Over Time
#'
#' Creates a closure that records RSS samples with timestamps.
#'
#' @param pid Process ID to monitor.
#' @return A list with `sample()` and `history()` functions.
#' @keywords internal
#' @noRd
rss_monitor <- function(pid) {
  samples <- list()

  list(
    sample = function() {
      rss <- rss_get_pid(pid)
      entry <- list(timestamp = Sys.time(), rss = rss)
      samples <<- c(samples, list(entry))
      entry
    },
    history = function() {
      if (length(samples) == 0) {
        return(data.frame(
          timestamp = as.POSIXct(character(0)),
          rss = numeric(0)
        ))
      }
      data.frame(
        timestamp = do.call(c, lapply(samples, `[[`, "timestamp")),
        rss = vapply(samples, `[[`, numeric(1), "rss")
      )
    },
    peak = function() {
      if (length(samples) == 0) return(NA_real_)
      max(vapply(samples, `[[`, numeric(1), "rss"), na.rm = TRUE)
    },
    drift = function() {
      if (length(samples) < 2) return(NA_real_)
      rss_vals <- vapply(samples, `[[`, numeric(1), "rss")
      baseline <- rss_vals[1]
      current <- rss_vals[length(rss_vals)]
      if (is.na(baseline) || baseline == 0) return(NA_real_)
      (current - baseline) / baseline
    }
  )
}

test_that("rss_self returns current process RSS", {
  rss <- rss_self()

  # Should be either a positive number or NA (if monitoring unavailable)
  if (!is.na(rss)) {
    expect_true(rss > 0)
    # R process should use at least some memory (10MB minimum)
    expect_gt(rss, 10 * 1024^2)
  }
})

test_that("rss_get_pid returns NA for invalid PID", {
  expect_true(is.na(rss_get_pid(NA)))
  # On Windows, wmic may be missing and produce a warning
  rss <- suppressWarnings(rss_get_pid(999999999)) # Unlikely to exist
  expect_true(is.na(rss))
})

test_that("rss_get_pid works for current process", {
  expect_silent({
    rss <- rss_get_pid(Sys.getpid())
  })

  # Should match rss_self
  rss_self_val <- rss_self()

  if (!is.na(rss) && !is.na(rss_self_val)) {
    # Allow some variance (10%)
    expect_true(abs(rss - rss_self_val) / rss_self_val < 0.1)
  }
})

test_that("rss_monitor tracks samples over time", {
  mon <- rss_monitor(Sys.getpid())

  # Take some samples
  mon$sample()
  Sys.sleep(0.05)
  mon$sample()
  Sys.sleep(0.05)
  mon$sample()

  hist <- mon$history()
  expect_equal(nrow(hist), 3L)
  expect_true("timestamp" %in% names(hist))
  expect_true("rss" %in% names(hist))

  # Peak should be valid
  peak <- mon$peak()
  if (!is.na(peak)) {
    expect_gt(peak, 0)
  }
})

test_that("rss_monitor drift calculation", {
  mon <- rss_monitor(Sys.getpid())

  # Single sample - drift should be NA
  mon$sample()
  expect_true(is.na(mon$drift()))

  # Second sample - drift should be calculable
  mon$sample()
  drift <- mon$drift()

  # Drift should be small for stable process
  if (!is.na(drift)) {
    expect_true(abs(drift) < 0.5)  # Less than 50% change
  }
})

#' @title Shard Descriptor Creation
#' @description Create shard descriptors for parallel execution with autotuning.
#' @name shards
NULL

#' Create Shard Descriptors
#'
#' Produces shard descriptors (index ranges) for use with `shard_map()`.
#' Supports autotuning based on worker count and memory constraints.
#'
#' @param n Integer. Total number of items to shard.
#' @param block_size Block size specification. Can be:
#'   - `"auto"` (default): Autotune based on worker count
#'   - Integer: Explicit number of items per shard
#'   - Character: Human-readable like `"1K"`, `"10K"`
#' @param workers Integer. Number of workers for autotuning (default: pool size or detectCores - 1).
#' @param strategy Sharding strategy: `"contiguous"` (default) or `"strided"`.
#' @param min_shards_per_worker Integer. Minimum shards per worker for load balancing (default 4).
#' @param max_shards_per_worker Integer. Maximum shards per worker to limit overhead (default 64).
#' @param scratch_bytes_per_item Numeric. Expected scratch memory per item for memory budgeting.
#' @param scratch_budget Character or numeric. Total scratch memory budget (e.g., "1GB").
#'
#' @return A `shard_descriptor` object containing:
#'   - `n`: Total items
#'   - `block_size`: Computed block size
#'   - `strategy`: Strategy used
#'   - `shards`: List of shard descriptors with `id`, `start`, `end`, `idx` fields
#'
#' @export
#' @examples
#' blocks <- shards(1e6, workers = 8)
#' length(blocks$shards)
#'
#' blocks <- shards(1000, block_size = 100)
#'
#' blocks$shards[[1]]$idx
shards <- function(n,
                   block_size = "auto",
                   workers = NULL,
                   strategy = c("contiguous", "strided"),
                   min_shards_per_worker = 4L,
                   max_shards_per_worker = 64L,
                   scratch_bytes_per_item = 0,
                   scratch_budget = 0) {
  # Validate n
  n <- as.integer(n)
  if (is.na(n) || n < 1L) {
    stop("shards: n must be a positive integer", call. = FALSE)
  }

  strategy <- match.arg(strategy)

  # Determine worker count for autotuning
  if (is.null(workers)) {
    pool <- pool_get()
    if (!is.null(pool)) {
      workers <- pool$n
    } else {
      dc <- parallel::detectCores()
      if (is.na(dc) || dc < 1L) dc <- 1L
      workers <- dc - 1L
    }
  }
  workers <- as.integer(workers)
  if (is.na(workers) || workers < 1L) workers <- 1L

  # Parse block_size
  if (identical(block_size, "auto")) {
    block_size <- autotune_block_size(
      n = n,
      workers = workers,
      min_shards_per_worker = min_shards_per_worker,
      max_shards_per_worker = max_shards_per_worker,
      scratch_bytes_per_item = scratch_bytes_per_item,
      scratch_budget = scratch_budget
    )
  } else if (is.character(block_size)) {
    block_size <- parse_count(block_size)
  } else {
    block_size <- as.integer(block_size)
  }

  if (is.na(block_size) || block_size < 1L) {
    stop("shards: block_size must be a positive integer", call. = FALSE)
  }

  # Create shards based on strategy
  if (strategy == "contiguous") {
    shard_list <- create_contiguous_shards(n, block_size)
  } else {
    num_shards <- ceiling(n / block_size)
    shard_list <- create_strided_shards(n, num_shards)
  }

  structure(
    list(
      n = n,
      block_size = block_size,
      strategy = strategy,
      num_shards = length(shard_list),
      shards = shard_list
    ),
    class = "shard_descriptor"
  )
}

#' Create Shards from an Explicit Index List
#'
#' Constructs a `shard_descriptor` from a user-supplied list of index vectors.
#' This is useful for non-contiguous workloads like searchlights/feature sets
#' where each shard operates on an arbitrary subset.
#'
#' @param idxs List of integer vectors (1-based indices). Each element becomes
#'   one shard with fields `id`, `idx`, and `len`.
#' @return A \code{shard_descriptor} list describing the chunk layout.
#' @examples
#' sh <- shards_list(list(1:10, 11:20, 21:30))
#' length(sh)
#' @export
shards_list <- function(idxs) {
  if (!is.list(idxs)) stop("idxs must be a list", call. = FALSE)

  shards <- vector("list", length(idxs))
  for (i in seq_along(idxs)) {
    idx <- idxs[[i]]
    if (!is.numeric(idx) || is.null(idx)) stop("idxs[[", i, "]] must be a numeric/integer vector", call. = FALSE)
    idx <- as.integer(idx)
    if (length(idx) < 1L) stop("idxs[[", i, "]] must be non-empty", call. = FALSE)
    if (anyNA(idx)) stop("idxs[[", i, "]] must not contain NA", call. = FALSE)
    if (any(idx < 1L)) stop("idxs[[", i, "]] must be >= 1", call. = FALSE)

    shards[[i]] <- list(id = i, idx = idx, len = length(idx))
  }

  structure(
    list(
      n = length(shards),
      block_size = NA_integer_,
      strategy = "list",
      num_shards = length(shards),
      shards = shards
    ),
    class = "shard_descriptor"
  )
}

#' Autotune Block Size
#'
#' Determines optimal block size based on worker count and constraints.
#'
#' @param n Total items.
#' @param workers Number of workers.
#' @param min_shards_per_worker Minimum shards per worker.
#' @param max_shards_per_worker Maximum shards per worker.
#' @param scratch_bytes_per_item Scratch bytes per item.
#' @param scratch_budget Total scratch budget.
#' @return Integer block size.
#' @keywords internal
#' @noRd
autotune_block_size <- function(n, workers,
                                min_shards_per_worker = 4L,
                                max_shards_per_worker = 64L,
                                scratch_bytes_per_item = 0,
                                scratch_budget = 0) {
  if (workers < 1L) workers <- 1L

  # Target shard counts

  min_shards <- workers * min_shards_per_worker
  max_shards <- workers * max_shards_per_worker

  # Block size from target shard counts
  block_from_min <- ceiling(n / min_shards)
  block_from_max <- max(floor(n / max_shards), 1L)

  # Start with block size for good parallelism

  block_size <- max(block_from_min, 1L)

  # Apply memory budget constraint if specified
  if (scratch_bytes_per_item > 0 && scratch_budget > 0) {
    if (is.character(scratch_budget)) {
      scratch_budget <- parse_bytes(scratch_budget)
    }
    max_items_per_budget <- scratch_budget / scratch_bytes_per_item
    memory_constrained_block <- floor(max_items_per_budget / workers)
    if (memory_constrained_block > 0 && memory_constrained_block < block_size) {
      block_size <- memory_constrained_block
    }
  }

  # Ensure shard count is in range
  num_shards <- ceiling(n / block_size)
  if (num_shards < min_shards && block_from_min > 1L) {
    block_size <- block_from_min
  }
  if (num_shards > max_shards && block_from_max > 0L) {
    block_size <- max(block_from_max, block_size)
  }

  max(as.integer(block_size), 1L)
}

#' Create Contiguous Shards
#'
#' Creates shards with consecutive indices.
#'
#' @param n Total items.
#' @param block_size Items per shard.
#' @return List of shard descriptors.
#' @keywords internal
#' @noRd
create_contiguous_shards <- function(n, block_size) {
  num_shards <- ceiling(n / block_size)
  shards <- vector("list", num_shards)

  start <- 1L
  for (i in seq_len(num_shards)) {
    end <- min(start + block_size - 1L, n)
    shards[[i]] <- list(
      id = i,
      start = start,
      end = end,
      idx = start:end,
      len = end - start + 1L
    )
    start <- end + 1L
  }

  shards
}

create_contiguous_shards_window_ <- function(start, end, block_size, start_id = 1L) {
  start <- as.integer(start)
  end <- as.integer(end)
  block_size <- as.integer(block_size)
  start_id <- as.integer(start_id)

  if (is.na(start) || is.na(end) || start < 1L || end < start) {
    stop("Invalid shard window (start/end)", call. = FALSE)
  }
  if (is.na(block_size) || block_size < 1L) stop("block_size must be >= 1", call. = FALSE)
  if (is.na(start_id) || start_id < 1L) stop("start_id must be >= 1", call. = FALSE)

  n <- end - start + 1L
  num_shards <- ceiling(n / block_size)
  shards <- vector("list", num_shards)

  cur <- start
  id <- start_id
  for (i in seq_len(num_shards)) {
    e <- min(cur + block_size - 1L, end)
    shards[[i]] <- list(
      id = id,
      start = cur,
      end = e,
      idx = cur:e,
      len = e - cur + 1L
    )
    cur <- e + 1L
    id <- id + 1L
  }
  shards
}

#' Create Strided Shards
#'
#' Creates shards with interleaved indices.
#'
#' @param n Total items.
#' @param num_shards Number of shards.
#' @return List of shard descriptors.
#' @keywords internal
#' @noRd
create_strided_shards <- function(n, num_shards) {
  num_shards <- max(min(num_shards, n), 1L)
  shards <- vector("list", num_shards)

  for (i in seq_len(num_shards)) {
    # Indices: i, i+num_shards, i+2*num_shards, ...
    idx <- seq(from = i, to = n, by = num_shards)
    shards[[i]] <- list(
      id = i,
      start = i,
      stride = num_shards,
      idx = idx,
      len = length(idx)
    )
  }

  shards
}

#' Parse Count String
#'
#' Parses strings like "1K", "10K", "1M" to integers.
#'
#' @param x Character string.
#' @return Integer value.
#' @keywords internal
#' @noRd
parse_count <- function(x) {
  if (is.numeric(x)) return(as.integer(x))

  x <- toupper(trimws(x))
  match <- regexec("^([0-9.]+)\\s*(K|M|B)?$", x)
  parts <- regmatches(x, match)[[1]]

  if (length(parts) < 2) {
    stop("Cannot parse count string: ", x, call. = FALSE)
  }

  value <- as.numeric(parts[2])
  unit <- if (length(parts) >= 3) parts[3] else ""

  multiplier <- switch(
    unit,
    "K" = 1000L,
    "M" = 1000000L,
    "B" = 1000000000L,
    1L
  )

  as.integer(value * multiplier)
}

#' Print a shard_descriptor Object
#'
#' @param x A \code{shard_descriptor} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @examples
#' sh <- shards(100, block_size = 25)
#' print(sh)
#' @export
print.shard_descriptor <- function(x, ...) {
  cat("shard descriptor\n")
  cat("  Items:", format(x$n, big.mark = ","), "\n")
  cat("  Block size:", format(x$block_size, big.mark = ","), "\n")
  cat("  Strategy:", x$strategy, "\n")
  cat("  Shards:", x$num_shards, "\n")

  # Show size distribution
  sizes <- vapply(x$shards, function(s) s$len, integer(1))
  if (length(unique(sizes)) == 1L) {
    cat("  Shard size:", sizes[1], "(uniform)\n")
  } else {
    cat("  Shard sizes:", min(sizes), "-", max(sizes), "\n")
  }

  invisible(x)
}

#' Length of a shard_descriptor Object
#'
#' @param x A \code{shard_descriptor} object.
#' @return An integer scalar giving the number of shards.
#' @examples
#' sh <- shards(100, block_size = 25)
#' length(sh)
#' @export
length.shard_descriptor <- function(x) {
  x$num_shards
}

#' Subset Shard Descriptor
#'
#' @param x A shard_descriptor object.
#' @param i Index or indices.
#' @return A subset of the object.
#' @examples
#' sh <- shards(100, block_size = 25)
#' sh[1:2]
#' @export
`[.shard_descriptor` <- function(x, i) {
  x$shards[i]
}

#' Get Single Shard
#'
#' @param x A shard_descriptor object.
#' @param i Index.
#' @return A subset of the object.
#' @examples
#' sh <- shards(100, block_size = 25)
#' sh[[1]]
#' @export
`[[.shard_descriptor` <- function(x, i) {
  x$shards[[i]]
}

test_that("shards_list builds a shard_descriptor with idx vectors", {
  idxs <- list(1:3, c(2, 5, 7), integer())
  expect_error(shards_list(idxs), "non-empty", fixed = FALSE)

  idxs <- list(1:3, c(2, 5, 7))
  s <- shards_list(idxs)
  expect_true(inherits(s, "shard_descriptor"))
  expect_equal(s$num_shards, 2L)
  expect_equal(s$shards[[1]]$idx, 1:3)
  expect_equal(s$shards[[2]]$len, 3L)
})

test_that("shards() creates correct contiguous shards", {
  desc <- shards(100, block_size = 25, strategy = "contiguous")

  expect_s3_class(desc, "shard_descriptor")
  expect_equal(desc$n, 100)
  expect_equal(desc$block_size, 25)
  expect_equal(desc$num_shards, 4)
  expect_equal(length(desc), 4)

  # Check each shard
  expect_equal(desc[[1]]$idx, 1:25)
  expect_equal(desc[[2]]$idx, 26:50)
  expect_equal(desc[[3]]$idx, 51:75)
  expect_equal(desc[[4]]$idx, 76:100)

  # All indices covered exactly once
  all_idx <- unlist(lapply(desc$shards, function(s) s$idx))
  expect_equal(sort(all_idx), 1:100)
})

test_that("shards() creates correct strided shards", {
  desc <- shards(10, block_size = 3, strategy = "strided")

  expect_s3_class(desc, "shard_descriptor")
  expect_equal(desc$strategy, "strided")

  # With block_size=3, we get ceil(10/3)=4 shards
  expect_equal(desc$num_shards, 4)

  # Check strided pattern
  expect_equal(desc[[1]]$idx, c(1, 5, 9))
  expect_equal(desc[[2]]$idx, c(2, 6, 10))
  expect_equal(desc[[3]]$idx, c(3, 7))
  expect_equal(desc[[4]]$idx, c(4, 8))

  # All indices covered exactly once
  all_idx <- unlist(lapply(desc$shards, function(s) s$idx))
  expect_equal(sort(all_idx), 1:10)
})

test_that("shards() handles uneven division", {
  desc <- shards(10, block_size = 3, strategy = "contiguous")

  expect_equal(desc$num_shards, 4)
  expect_equal(desc[[1]]$idx, 1:3)
  expect_equal(desc[[2]]$idx, 4:6)
  expect_equal(desc[[3]]$idx, 7:9)
  expect_equal(desc[[4]]$idx, 10L)  # Last shard has only 1 item

  all_idx <- unlist(lapply(desc$shards, function(s) s$idx))
  expect_equal(sort(all_idx), 1:10)
})

test_that("shards() autotuning creates reasonable block sizes", {
  desc <- shards(1000, block_size = "auto", workers = 4)

  # With 4 workers, default 4-64 shards per worker = 16-256 shards
  expect_gte(desc$num_shards, 16)
  expect_lte(desc$num_shards, 256)

  # All indices covered
  all_idx <- unlist(lapply(desc$shards, function(s) s$idx))
  expect_equal(sort(all_idx), 1:1000)
})

test_that("shards() respects memory constraints", {
  desc <- shards(
    10000,
    block_size = "auto",
    workers = 4,
    scratch_bytes_per_item = 1024,  # 1KB per item
    scratch_budget = "1MB"  # 1MB total
  )

  # Memory constraint: 1MB / 4 workers / 1KB per item = 256 items max per shard
  expect_lte(desc$block_size, 256)

  all_idx <- unlist(lapply(desc$shards, function(s) s$idx))
  expect_equal(sort(all_idx), 1:10000)
})

test_that("shards() parses count strings", {
  desc <- shards(10000, block_size = "1K")
  expect_equal(desc$block_size, 1000)

  desc <- shards(10000, block_size = "500")
  expect_equal(desc$block_size, 500)
})

test_that("shards() validates input", {
  expect_error(shards(0), "positive integer")
  expect_error(shards(-1), "positive integer")
  expect_error(shards(100, block_size = 0), "positive integer")
  expect_error(shards(100, block_size = -1), "positive integer")
})

test_that("shards() handles edge cases", {
  # Single item
  desc <- shards(1, block_size = 10)
  expect_equal(desc$num_shards, 1)
  expect_equal(desc[[1]]$idx, 1L)

  # Block size larger than n
  desc <- shards(5, block_size = 100)
  expect_equal(desc$num_shards, 1)
  expect_equal(desc[[1]]$idx, 1:5)

  # Single shard exactly
  desc <- shards(10, block_size = 10)
  expect_equal(desc$num_shards, 1)
  expect_equal(desc[[1]]$idx, 1:10)
})

test_that("shard_descriptor print method works", {
  desc <- shards(100, block_size = 25)
  expect_output(print(desc), "shard descriptor")
  expect_output(print(desc), "Items: 100")
  expect_output(print(desc), "Shards: 4")
})

test_that("shard_descriptor subsetting works", {
  desc <- shards(100, block_size = 25)

  # Single bracket returns list of shards
  subset <- desc[1:2]
  expect_type(subset, "list")
  expect_length(subset, 2)

  # Double bracket returns single shard
  shard <- desc[[1]]
  expect_type(shard, "list")
  expect_equal(shard$id, 1)
})

test_that("coverage is complete for all n and block_size combinations", {
  for (n in c(1, 2, 7, 10, 100, 1000)) {
    for (bs in c(1, 2, 3, 7, 10, 100)) {
      desc_contig <- shards(n, block_size = bs, strategy = "contiguous")
      desc_strided <- shards(n, block_size = bs, strategy = "strided")

      # Contiguous coverage
      all_idx <- unlist(lapply(desc_contig$shards, function(s) s$idx))
      expect_equal(sort(all_idx), 1:n,
        info = sprintf("Contiguous failed for n=%d, bs=%d", n, bs))

      # Strided coverage
      all_idx <- unlist(lapply(desc_strided$shards, function(s) s$idx))
      expect_equal(sort(all_idx), 1:n,
        info = sprintf("Strided failed for n=%d, bs=%d", n, bs))
    }
  }
})

test_that("recommendations() suggests shm_queue / chunking for tiny rpc tasks", {
  skip_on_cran()

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  # Create a real shard_result, then pin diagnostics to a deterministic "tiny task"
  # scenario (avoid timing flakiness in CI).
  res <- shard_map(
    shards(10, workers = 2),
    fun = function(sh) NULL,
    workers = 2,
    chunk_size = 1,
    diagnostics = TRUE
  )

  # Force a "tiny task" shape.
  res$diagnostics$dispatch_mode <- "rpc_chunked"
  res$diagnostics$shards_processed <- 10000L
  res$diagnostics$chunks_dispatched <- 10000L
  res$diagnostics$duration <- 1.0

  recs <- recommendations(res)
  expect_true(any(grepl("shm_queue", recs, fixed = TRUE) | grepl("chunk_size", recs, fixed = TRUE)))
})

test_that("recommendations() suggests increasing block_size when shm_queue n_tasks is huge", {
  skip_on_cran()

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  res <- shard_map(
    shards(10, workers = 2),
    fun = function(sh) NULL,
    workers = 2,
    chunk_size = 1,
    diagnostics = TRUE
  )

  res$diagnostics$dispatch_mode <- "shm_queue"
  res$diagnostics$shm_queue <- list(n_tasks = 2000001L, done = 2000001L, failed = 0L)

  recs <- recommendations(res)
  expect_true(any(grepl("dispatch_opts\\$block_size", recs)))
})

# Streaming operations over on-disk table outputs (row-groups / datasets).
#
# Goal: bounded-memory post-processing without forcing collect()/as_tibble().

#' Stream over row-groups/datasets and reduce
#'
#' Applies \code{f()} to each partition (row-group) and combines results with
#' \code{combine()} into a single accumulator. This keeps peak memory bounded by the
#' largest single partition (plus your accumulator).
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @param f Function \code{(chunk, ...) -> value} producing a per-partition value.
#' @param init Initial accumulator value.
#' @param combine Function \code{(acc, value) -> acc} to update the accumulator.
#' @param ... Passed to \code{f()}.
#' @return The final accumulator value after processing all partitions.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L, data.frame(x = rnorm(5)))
#' rg <- table_finalize(sink)
#' total <- stream_reduce(rg, f = nrow, init = 0L, combine = `+`)
#' }
stream_reduce <- function(x, f, init, combine, ...) {
  UseMethod("stream_reduce")
}

#' @rdname stream_reduce
#' @export
stream_reduce.shard_row_groups <- function(x, f, init, combine, ...) {
  if (!is.function(f)) stop("f must be a function", call. = FALSE)
  if (!is.function(combine)) stop("combine must be a function", call. = FALSE)

  it <- iterate_row_groups(x)
  acc <- init
  repeat {
    chunk <- it()
    if (is.null(chunk)) break
    val <- f(chunk, ...)
    acc <- combine(acc, val)
  }
  acc
}

#' @rdname stream_reduce
#' @export
stream_reduce.shard_dataset <- function(x, f, init, combine, ...) {
  stream_reduce.shard_row_groups(structure(unclass(x), class = "shard_row_groups"),
                                 f = f, init = init, combine = combine, ...)
}

#' Stream over row-groups/datasets and map
#'
#' Applies \code{f()} to each partition and returns the list of per-partition results.
#' This is still much cheaper than collecting the full dataset when \code{f()} returns
#' a small summary per partition.
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @param f Function \code{(chunk, ...) -> value}.
#' @param ... Passed to \code{f()}.
#' @return A list of per-partition values, one element per row-group file.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L, data.frame(x = rnorm(5)))
#' rg <- table_finalize(sink)
#' nrows <- stream_map(rg, nrow)
#' }
stream_map <- function(x, f, ...) {
  UseMethod("stream_map")
}

#' @rdname stream_map
#' @export
stream_map.shard_row_groups <- function(x, f, ...) {
  if (!is.function(f)) stop("f must be a function", call. = FALSE)
  it <- iterate_row_groups(x)
  out <- list()
  i <- 0L
  repeat {
    chunk <- it()
    if (is.null(chunk)) break
    i <- i + 1L
    out[[i]] <- f(chunk, ...)
  }
  out
}

#' @rdname stream_map
#' @export
stream_map.shard_dataset <- function(x, f, ...) {
  stream_map.shard_row_groups(structure(unclass(x), class = "shard_row_groups"), f = f, ...)
}

#' Stream row count
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @return A single integer giving the total number of rows across all partitions.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L, data.frame(x = rnorm(5)))
#' rg <- table_finalize(sink)
#' stream_count(rg)
#' }
stream_count <- function(x) {
  stream_reduce(
    x,
    f = function(chunk) nrow(chunk),
    init = 0L,
    combine = function(acc, n) acc + as.integer(n)
  )
}

#' Stream-filter a dataset/row-groups into a new partitioned dataset
#'
#' Reads each partition, filters rows, and writes a new partitioned dataset.
#' Output is written as one partition per input partition (empty partitions are
#' allowed). This avoids materializing all results.
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @param predicate Function \code{(chunk, ...) -> logical} row mask (length == nrow(chunk)).
#' @param path Output directory. If NULL, a temp dir is created.
#' @param ... Passed to \code{predicate()}.
#' @return A \code{shard_dataset} handle pointing to the filtered partitions.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L, data.frame(x = c(1.0, 2.0, 3.0)))
#' rg <- table_finalize(sink)
#' filtered <- stream_filter(rg, predicate = function(chunk) chunk$x > 1.5)
#' }
stream_filter <- function(x, predicate, path = NULL, ...) {
  if (!(inherits(x, "shard_row_groups") || inherits(x, "shard_dataset"))) {
    stop("x must be a shard_row_groups or shard_dataset handle", call. = FALSE)
  }
  if (!is.function(predicate)) stop("predicate must be a function", call. = FALSE)

  schema <- x$schema
  if (is.null(schema) || !inherits(schema, "shard_schema")) {
    stop("x must contain a schema to filter", call. = FALSE)
  }

  if (is.null(path)) {
    path <- file.path(tempdir(), paste0("shard_stream_filter_", unique_id()))
  }

  sink <- table_sink(schema, mode = "partitioned", path = path)

  it <- iterate_row_groups(x)
  shard_id <- 0L
  repeat {
    chunk <- it()
    if (is.null(chunk)) break
    shard_id <- shard_id + 1L
    keep <- predicate(chunk, ...)
    keep <- as.logical(keep)
    if (length(keep) != nrow(chunk)) {
      stop("predicate must return a logical vector with length == nrow(chunk)", call. = FALSE)
    }
    out <- chunk[keep, , drop = FALSE]
    table_write(sink, shard_id, out)
  }

  table_finalize(sink, materialize = "never")
}

#' Stream sum of a numeric column
#'
#' Computes the sum of \code{col} across all partitions without collecting the full
#' dataset. When partitions are native-encoded, this avoids decoding string
#' columns entirely.
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @param col Column name to sum.
#' @param na_rm Logical; drop NAs (default TRUE).
#' @return A single numeric value giving the sum of the column across all partitions.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L, data.frame(x = c(1.0, 2.0, 3.0)))
#' rg <- table_finalize(sink)
#' stream_sum(rg, "x")
#' }
stream_sum <- function(x, col, na_rm = TRUE) {
  if (!(inherits(x, "shard_row_groups") || inherits(x, "shard_dataset"))) {
    stop("x must be a shard_row_groups or shard_dataset handle", call. = FALSE)
  }
  col <- as.character(col)
  if (!nzchar(col)) stop("col must be a non-empty column name", call. = FALSE)
  sch <- x$schema$columns %||% list()
  ct <- sch[[col]] %||% NULL
  if (is.null(ct)) stop("Unknown column '", col, "'", call. = FALSE)
  if (ct$kind %in% c("string", "factor", "raw")) {
    stop("stream_sum() requires a numeric column", call. = FALSE)
  }

  it <- iterate_row_groups(x, decode = FALSE)
  acc <- 0
  repeat {
    chunk <- it()
    if (is.null(chunk)) break
    if (inherits(chunk, "shard_table_part_native")) {
      v <- chunk$columns[[col]]
      # Factor codes are integers; sum on codes is rarely meaningful and is
      # likely user error.
      if (inherits(v, "shard_string_blob")) {
        stop("stream_sum() cannot sum a string column", call. = FALSE)
      }
      if (isTRUE(na_rm)) v <- v[!is.na(v)]
      acc <- acc + sum(as.double(v))
    } else {
      v <- chunk[[col]]
      if (isTRUE(na_rm)) v <- v[!is.na(v)]
      acc <- acc + sum(as.double(v))
    }
  }
  acc
}

#' Stream top-k rows by a numeric column
#'
#' Finds the top \code{k} rows by \code{col} without collecting the full dataset.
#'
#' For native-encoded partitions, this selects candidate rows using the numeric
#' column without decoding strings, then decodes only the chosen rows for the
#' returned result.
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @param col Column name to rank by.
#' @param k Number of rows to keep.
#' @param decreasing Logical; TRUE for largest values (default TRUE).
#' @param na_drop Logical; drop rows where \code{col} is NA (default TRUE).
#' @return A data.frame (or tibble if the \code{tibble} package is installed)
#'   with at most \code{k} rows ordered by \code{col}.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L, data.frame(x = c(3.0, 1.0, 2.0)))
#' rg <- table_finalize(sink)
#' stream_top_k(rg, "x", k = 2L)
#' }
stream_top_k <- function(x, col, k = 10L, decreasing = TRUE, na_drop = TRUE) {
  if (!(inherits(x, "shard_row_groups") || inherits(x, "shard_dataset"))) {
    stop("x must be a shard_row_groups or shard_dataset handle", call. = FALSE)
  }
  col <- as.character(col)
  k <- as.integer(k)
  if (is.na(k) || k < 1L) stop("k must be >= 1", call. = FALSE)

  schema <- x$schema
  sch_cols <- schema$columns %||% list()
  ct <- sch_cols[[col]] %||% NULL
  if (is.null(ct)) stop("Unknown column '", col, "'", call. = FALSE)
  if (!(ct$kind %in% c("int32", "float64"))) {
    stop("stream_top_k() requires an int32() or float64() column", call. = FALSE)
  }

  it <- iterate_row_groups(x, decode = FALSE)
  best <- NULL

  keep_best <- function(df) {
    if (is.null(best) || nrow(best) == 0) {
      best <<- df
    } else {
      best <<- rbind(best, df)
    }
    # Keep only top-k in memory.
    v <- best[[col]]
    if (isTRUE(na_drop)) {
      ok <- !is.na(v)
      best <<- best[ok, , drop = FALSE]
      v <- best[[col]]
    }
    ord <- order(v, decreasing = isTRUE(decreasing))
    if (length(ord) > k) ord <- ord[seq_len(k)]
    best <<- best[ord, , drop = FALSE]
  }

  repeat {
    chunk <- it()
    if (is.null(chunk)) break

    if (inherits(chunk, "shard_table_part_native")) {
      v <- chunk$columns[[col]]
      if (inherits(v, "shard_string_blob")) stop("stream_top_k() requires a numeric column", call. = FALSE)
      v <- as.double(v)
      if (isTRUE(na_drop)) {
        ok <- !is.na(v)
        v2 <- v[ok]
        if (length(v2) == 0) next
        idx_all <- which(ok)
      } else {
        v2 <- v
        idx_all <- seq_along(v)
      }

      ord <- order(v2, decreasing = isTRUE(decreasing))
      if (length(ord) > k) ord <- ord[seq_len(k)]
      rows <- idx_all[ord]
      df <- .table_part_native_decode_rows(chunk, schema, rows)
      keep_best(df)
    } else {
      v <- chunk[[col]]
      v <- as.double(v)
      if (isTRUE(na_drop)) {
        ok <- !is.na(v)
        if (!any(ok)) next
        chunk2 <- chunk[ok, , drop = FALSE]
      } else {
        chunk2 <- chunk
      }
      ord <- order(as.double(chunk2[[col]]), decreasing = isTRUE(decreasing))
      if (length(ord) > k) ord <- ord[seq_len(k)]
      keep_best(chunk2[ord, , drop = FALSE])
    }
  }

  if (is.null(best)) {
    df <- data.frame()
    if (pkg_available("tibble")) return(tibble::as_tibble(df))
    return(df)
  }

  if (pkg_available("tibble")) return(tibble::as_tibble(best))
  best
}

#' Stream group-wise sum
#'
#' Computes sum(value) by group across partitions without collecting. This is
#' optimized for factor groups (\code{factor_col()}).
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @param group Group column name (recommended: \code{factor_col()}).
#' @param value Numeric column name to sum.
#' @param na_rm Logical; drop rows where value is NA (default TRUE).
#' @return A data.frame with columns \code{group} (factor) and \code{sum} (numeric).
#' @export
#' @examples
#' \donttest{
#' s <- schema(g = factor_col(c("a", "b")), x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L,
#'   data.frame(g = factor(c("a", "b", "a"), levels = c("a", "b")), x = c(1, 2, 3)))
#' rg <- table_finalize(sink)
#' stream_group_sum(rg, "g", "x")
#' }
stream_group_sum <- function(x, group, value, na_rm = TRUE) {
  if (!(inherits(x, "shard_row_groups") || inherits(x, "shard_dataset"))) {
    stop("x must be a shard_row_groups or shard_dataset handle", call. = FALSE)
  }
  group <- as.character(group)
  value <- as.character(value)
  schema <- x$schema
  ct_g <- schema$columns[[group]] %||% NULL
  ct_v <- schema$columns[[value]] %||% NULL
  if (is.null(ct_g)) stop("Unknown group column '", group, "'", call. = FALSE)
  if (is.null(ct_v)) stop("Unknown value column '", value, "'", call. = FALSE)

  if (ct_g$kind != "factor") {
    stop("stream_group_sum() currently requires a factor_col() group column", call. = FALSE)
  }
  if (!(ct_v$kind %in% c("int32", "float64"))) {
    stop("stream_group_sum() requires an int32() or float64() value column", call. = FALSE)
  }

  lev <- ct_g$levels %||% character(0)
  acc <- numeric(length(lev))

  it <- iterate_row_groups(x, decode = FALSE)
  repeat {
    chunk <- it()
    if (is.null(chunk)) break

    if (inherits(chunk, "shard_table_part_native")) {
      codes <- as.integer(chunk$columns[[group]])
      vals <- as.double(chunk$columns[[value]])
      ok <- !is.na(codes) & codes >= 1L & codes <= length(lev)
      if (isTRUE(na_rm)) ok <- ok & !is.na(vals)
      if (!any(ok)) next
      rs <- rowsum(vals[ok], group = factor(codes[ok], levels = seq_len(length(lev))), reorder = FALSE)
      acc <- acc + as.double(rs[, 1])
    } else {
      g <- chunk[[group]]
      v <- as.double(chunk[[value]])
      ok <- !is.na(g)
      if (isTRUE(na_rm)) ok <- ok & !is.na(v)
      if (!any(ok)) next
      rs <- rowsum(v[ok], group = factor(g[ok], levels = lev), reorder = FALSE)
      acc <- acc + as.double(rs[, 1])
    }
  }

  data.frame(group = factor(lev, levels = lev), sum = acc, stringsAsFactors = FALSE)
}

#' Stream group-wise count
#'
#' Counts rows per group across partitions without collecting. Optimized for
#' factor groups (\code{factor_col()}).
#'
#' @param x A \code{shard_row_groups} or \code{shard_dataset} handle.
#' @param group Group column name (recommended: \code{factor_col()}).
#' @return A data.frame with columns \code{group} (factor) and \code{n} (integer).
#' @export
#' @examples
#' \donttest{
#' s <- schema(g = factor_col(c("a", "b")), x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L,
#'   data.frame(g = factor(c("a", "b", "a"), levels = c("a", "b")), x = c(1, 2, 3)))
#' rg <- table_finalize(sink)
#' stream_group_count(rg, "g")
#' }
stream_group_count <- function(x, group) {
  if (!(inherits(x, "shard_row_groups") || inherits(x, "shard_dataset"))) {
    stop("x must be a shard_row_groups or shard_dataset handle", call. = FALSE)
  }
  group <- as.character(group)
  schema <- x$schema
  ct_g <- schema$columns[[group]] %||% NULL
  if (is.null(ct_g)) stop("Unknown group column '", group, "'", call. = FALSE)

  if (ct_g$kind != "factor") {
    stop("stream_group_count() currently requires a factor_col() group column", call. = FALSE)
  }

  lev <- ct_g$levels %||% character(0)
  acc <- integer(length(lev))

  it <- iterate_row_groups(x, decode = FALSE)
  repeat {
    chunk <- it()
    if (is.null(chunk)) break

    if (inherits(chunk, "shard_table_part_native")) {
      codes <- as.integer(chunk$columns[[group]])
      ok <- !is.na(codes) & codes >= 1L & codes <= length(lev)
      if (!any(ok)) next
      acc <- acc + tabulate(codes[ok], nbins = length(lev))
    } else {
      g <- chunk[[group]]
      idx <- match(as.character(g), lev)
      idx <- idx[!is.na(idx)]
      if (length(idx) == 0) next
      acc <- acc + tabulate(idx, nbins = length(lev))
    }
  }

  data.frame(group = factor(lev, levels = lev), n = acc, stringsAsFactors = FALSE)
}

#' @title Task Dispatch Engine
#' @description Orchestrates chunk dispatch with worker supervision and failure handling.
#' @name dispatch
NULL

#' Dispatch Chunks to Worker Pool
#'
#' Executes a function over chunks using the worker pool with supervision.
#' Handles worker death and recycling transparently by requeuing failed chunks.
#'
#' @param chunks List of chunk descriptors. Each chunk will be passed to `fun`.
#' @param fun Function to execute. Receives (chunk, ...) as arguments.
#' @param ... Additional arguments passed to `fun`.
#' @param pool A `shard_pool` object. If NULL, uses the current pool.
#' @param health_check_interval Integer. Check pool health every N chunks (default 10).
#' @param max_retries Integer. Maximum retries per chunk before permanent failure (default 3).
#' @param timeout Numeric. Seconds to wait for each chunk (default 3600).
#' @param scheduler_policy Optional list of scheduling hints (advanced). Currently:
#'   - `max_huge_concurrency`: cap concurrent chunks with `footprint_class=="huge"`.
#' @param on_result Optional callback (advanced). If provided, called on the
#'   master process as `on_result(tag, value, worker_id)` for each successful
#'   chunk completion. Used by [shard_reduce()] to stream reductions.
#' @param store_results Logical (advanced). If FALSE, successful chunk values are
#'   not retained in the returned `results` list (streaming use cases).
#' @param retain_chunks Logical (advanced). If FALSE, completed chunk descriptors
#'   are stored minimally (avoids retaining large shard lists in memory).
#'
#' @return A \code{shard_dispatch_result} object with results and diagnostics.
#' @export
#' @examples
#' \donttest{
#' pool_create(2)
#' chunks <- list(list(id = 1L, x = 1), list(id = 2L, x = 2))
#' result <- dispatch_chunks(chunks, function(chunk) chunk$x * 2, pool = pool_get())
#' pool_stop()
#' }
dispatch_chunks <- function(chunks, fun, ...,
                            pool = NULL,
                            health_check_interval = 10L,
                            max_retries = 3L,
                            timeout = 3600,
                            scheduler_policy = NULL,
                            on_result = NULL,
                            store_results = TRUE,
                            retain_chunks = TRUE) {
  if (is.null(pool)) {
    pool <- pool_get()
  }

  if (is.null(pool)) {
    stop("No active pool. Call pool_create() first.", call. = FALSE)
  }

  # Capture additional arguments as a list
  extra_args <- list(...)

  # Avoid `parallel:::` usage (R CMD check will warn). We still rely on
  # parallel's internal protocol, but obtain the function dynamically.
  parallel_sendCall <- utils::getFromNamespace("sendCall", "parallel")

  # Create queue
  queue <- queue_create(chunks)

  # Track diagnostics
  diag <- list(
    start_time = Sys.time(),
    health_checks = list(),
    worker_assignments = list(),
    scheduler = list(throttle_events = 0L)
  )

  # Export fun + args to workers once to avoid per-chunk re-export overhead.
  # Each worker runs a small wrapper via sendCall/recv that references these globals.
  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    if (!is.null(w) && worker_is_alive(w)) {
      export_env <- new.env(parent = emptyenv())
      export_env$.shard_dispatch_fun <- fun
      export_env$.shard_dispatch_args <- extra_args
      parallel::clusterExport(w$cluster, c(".shard_dispatch_fun", ".shard_dispatch_args"),
                              envir = export_env)
    }
  }

  dispatch_wrapper <- function(chunk) {
    vd0 <- tryCatch(view_diagnostics(), error = function(e) NULL)
    cd0 <- tryCatch(buffer_diagnostics(), error = function(e) NULL)
    td0 <- tryCatch(table_diagnostics(), error = function(e) NULL)
    sd0 <- tryCatch(scratch_diagnostics(), error = function(e) NULL)
    res <- tryCatch(
      {
        f <- get(".shard_dispatch_fun", envir = globalenv(), inherits = FALSE)
        a <- get(".shard_dispatch_args", envir = globalenv(), inherits = FALSE)
        list(ok = TRUE, value = do.call(f, c(list(chunk), a)))
      },
      error = function(e) list(ok = FALSE, error = conditionMessage(e))
    )
    vd1 <- tryCatch(view_diagnostics(), error = function(e) NULL)
    cd1 <- tryCatch(buffer_diagnostics(), error = function(e) NULL)
    td1 <- tryCatch(table_diagnostics(), error = function(e) NULL)
    sd1 <- tryCatch(scratch_diagnostics(), error = function(e) NULL)

    if (is.list(vd0) && is.list(vd1)) {
      res$view_delta <- list(
        created = (vd1$created %||% 0L) - (vd0$created %||% 0L),
        materialized = (vd1$materialized %||% 0L) - (vd0$materialized %||% 0L),
        materialized_bytes = (vd1$materialized_bytes %||% 0) - (vd0$materialized_bytes %||% 0),
        packed = (vd1$packed %||% 0L) - (vd0$packed %||% 0L),
        packed_bytes = (vd1$packed_bytes %||% 0) - (vd0$packed_bytes %||% 0)
      )
    }

    # Best-effort attribution: only compute hotspot deltas when materialization occurred.
    if (is.list(res$view_delta) && (res$view_delta$materialized %||% 0L) > 0L) {
      hs1 <- tryCatch(view_materialize_hotspots_snapshot_(), error = function(e) NULL)
      if (is.list(hs1)) {
        hs0 <- if (exists(".shard_view_hotspot_snapshot", envir = .shard_worker_env, inherits = FALSE)) {
          get(".shard_view_hotspot_snapshot", envir = .shard_worker_env)
        } else {
          list()
        }

        delta <- list()
        keys <- unique(c(names(hs0), names(hs1)))
        for (k in keys) {
          v1 <- hs1[[k]] %||% list()
          v0 <- hs0[[k]] %||% list()
          b1 <- v1$bytes %||% 0
          c1 <- v1$count %||% 0L
          b0 <- v0$bytes %||% 0
          c0 <- v0$count %||% 0L
          db <- as.double(b1) - as.double(b0)
          dc <- as.integer(c1) - as.integer(c0)
          if (is.finite(db) && db > 0) {
            delta[[k]] <- list(bytes = db, count = dc)
          }
        }

        assign(".shard_view_hotspot_snapshot", hs1, envir = .shard_worker_env)
        if (length(delta) > 0) res$view_hotspots_delta <- delta
      }
    }

    if (is.list(cd0) && is.list(cd1)) {
      res$copy_delta <- list(
        buffer_writes = (cd1$writes %||% 0L) - (cd0$writes %||% 0L),
        buffer_bytes = (cd1$bytes %||% 0) - (cd0$bytes %||% 0)
      )
    }

    if (is.list(td0) && is.list(td1)) {
      res$table_delta <- list(
        table_writes = (td1$writes %||% 0L) - (td0$writes %||% 0L),
        table_rows = (td1$rows %||% 0L) - (td0$rows %||% 0L),
        table_bytes = (td1$bytes %||% 0) - (td0$bytes %||% 0)
      )
    }

    if (is.list(sd0) && is.list(sd1)) {
      res$scratch_delta <- list(
        hits = (sd1$hits %||% 0L) - (sd0$hits %||% 0L),
        misses = (sd1$misses %||% 0L) - (sd0$misses %||% 0L),
        high_water = sd1$high_water %||% 0
      )
      res$scratch_needs_recycle <- isTRUE(sd1$needs_recycle)
    }

    res
  }

  # Async scheduling state
  idle <- rep(TRUE, pool$n)
  inflight <- vector("list", pool$n) # worker_id -> {chunk_id,start_time}
  inflight_class <- rep(NA_character_, pool$n)

  chunks_processed <- 0L
  view_stats <- list(created = 0L, materialized = 0L, materialized_bytes = 0,
                     packed = 0L, packed_bytes = 0)
  view_hotspots <- list()
  copy_stats <- list(borrow_exports = 0L, borrow_bytes = 0, buffer_writes = 0L, buffer_bytes = 0)
  table_stats <- list(writes = 0L, rows = 0L, bytes = 0)
  scratch_stats <- list(hits = 0L, misses = 0L, high_water = 0)

  policy <- scheduler_policy %||% list()
  max_huge <- as.integer(policy$max_huge_concurrency %||% NA_integer_)
  if (!is.na(max_huge) && max_huge < 1L) max_huge <- 1L

  chunk_class <- function(ch) {
    (ch$footprint_class %||% "tiny")
  }

  huge_inflight <- function() {
    sum(inflight_class == "huge", na.rm = TRUE)
  }

  # Helper: receive a single result (non-blocking with small timeout) from any worker.
  recv_one <- function(timeout_sec = 0.1) {
    socklist <- list()
    idx_map <- integer(0)
    for (i in seq_along(pool$workers)) {
      w <- pool$workers[[i]]
      if (!is.null(w) && worker_is_alive(w)) {
        n <- w$cluster[[1]]
        if (!is.null(n) && !is.null(n$con)) {
          if (isOpen(n$con)) {
            socklist[[length(socklist) + 1L]] <- n$con
            idx_map <- c(idx_map, i)
          }
        }
      }
    }

    if (length(socklist) == 0) return(NULL)

    # socketSelect() is provided by base R (works for socket connections).
    ready <- socketSelect(socklist, timeout = timeout_sec)
    if (length(ready) == 0 || !any(ready)) return(NULL)

    # socketSelect is documented to return a logical vector (like parallel uses),
    # but be defensive if it returns indices on some platforms.
    if (is.logical(ready)) {
      n_local <- which.max(ready)
    } else if (is.numeric(ready)) {
      # Some platforms return a 0/1 numeric vector; others may return indices.
      if (length(ready) == length(socklist)) {
        n_local <- which.max(ready)
      } else {
        n_local <- as.integer(ready[1])
      }
    } else {
      return(NULL)
    }

    if (is.na(n_local) || n_local < 1L || n_local > length(socklist)) {
      return(NULL)
    }
    worker_id <- idx_map[n_local]

    msg <- tryCatch(
      unserialize(socklist[[n_local]]),
      error = function(e) list(tag = NA_character_, value = list(ok = FALSE, error = conditionMessage(e)),
                               .recv_error = TRUE)
    )
    list(worker_id = worker_id, tag = msg$tag, value = msg$value, recv_error = isTRUE(msg$.recv_error))
  }

  # Main dispatch loop: keep all workers busy and collect results as they finish.
  while (!queue_is_done(queue)) {
    # Health check at intervals (based on completions)
    if (chunks_processed > 0 && chunks_processed %% health_check_interval == 0) {
      busy_ids <- which(!idle)
      health <- pool_health_check(pool, busy_workers = busy_ids)
      diag$health_checks <- c(diag$health_checks, list(health))

      # Refresh pool reference after potential recycle/restart.
      pool <- pool_get()

      # Requeue any in-flight chunks from recycled/restarted workers
      for (action in health$worker_actions) {
        if (action$action %in% c("restart", "recycle")) {
          requeued <- queue_requeue_worker(queue, action$worker_id)
          if (requeued > 0) {
            message(sprintf("Requeued %d chunks from worker %d (%s)",
                            requeued, action$worker_id, action$reason))
          }
          idle[action$worker_id] <- TRUE
          inflight[action$worker_id] <- list(NULL)
          inflight_class[action$worker_id] <- NA_character_

          # Ensure the replacement worker has the dispatch globals.
          w_new <- pool$workers[[action$worker_id]]
          if (!is.null(w_new)) {
            export_env <- new.env(parent = emptyenv())
            export_env$.shard_dispatch_fun <- fun
            export_env$.shard_dispatch_args <- extra_args
            parallel::clusterExport(w_new$cluster,
                                    c(".shard_dispatch_fun", ".shard_dispatch_args"),
                                    envir = export_env)
          }
        }
      }
    }

    # Fill idle workers
    for (worker_id in which(idle)) {
      if (!queue_has_pending(queue)) break

      pred <- NULL
      if (!is.na(max_huge) && huge_inflight() >= max_huge) {
        pred <- function(ch) chunk_class(ch) != "huge"
      }

      chunk <- queue_next_where(queue, worker_id, pred)
      if (is.null(chunk)) {
        # If we couldn't find a non-huge chunk under throttling, leave the
        # worker idle and wait for an in-flight huge task to finish.
        if (!is.null(pred)) {
          diag$scheduler$throttle_events <- diag$scheduler$throttle_events + 1L
        }
        next
      }

      w <- pool$workers[[worker_id]]
      if (is.null(w) || !worker_is_alive(w)) {
        pool <- pool_restart_worker_(pool, worker_id)
        w <- pool$workers[[worker_id]]

        # Re-export globals to the restarted worker
        export_env <- new.env(parent = emptyenv())
        export_env$.shard_dispatch_fun <- fun
        export_env$.shard_dispatch_args <- extra_args
        parallel::clusterExport(w$cluster, c(".shard_dispatch_fun", ".shard_dispatch_args"),
                                envir = export_env)
      }

      # If this worker was flagged for recycle while busy, do it now (safe point).
      if (isTRUE(w$needs_recycle)) {
        min_age <- pool$min_recycle_interval %||% 0
        age <- as.numeric(difftime(Sys.time(), w$spawned_at %||% Sys.time(), units = "secs"))
        if (!is.na(age) && is.finite(min_age) && age < min_age) {
          # Cooldown: avoid thrashing PSOCK creation/destruction under extremely
          # tight rss_limit settings.
          # We'll keep processing and recycle on a later safe point.
          w$needs_recycle <- TRUE
        } else {
          pool$workers[[worker_id]] <- worker_recycle(w, pool$init_expr, pool$packages, dev_path = pool$dev_path)
          pool$workers[[worker_id]]$rss_baseline <- worker_rss(pool$workers[[worker_id]])
          pool$workers[[worker_id]]$needs_recycle <- FALSE
          pool$stats$total_recycles <- pool$stats$total_recycles + 1L
          .pool_env$pool <- pool
          w <- pool$workers[[worker_id]]

          export_env <- new.env(parent = emptyenv())
          export_env$.shard_dispatch_fun <- fun
          export_env$.shard_dispatch_args <- extra_args
          parallel::clusterExport(w$cluster, c(".shard_dispatch_fun", ".shard_dispatch_args"),
                                  envir = export_env)
        }
      }

      # Async send
      parallel_sendCall(w$cluster[[1]], fun = dispatch_wrapper, args = list(chunk),
                        return = TRUE, tag = as.character(chunk$id))
      idle[worker_id] <- FALSE
      inflight[[worker_id]] <- list(chunk_id = as.character(chunk$id), start_time = Sys.time())
      inflight_class[worker_id] <- chunk_class(chunk)
      pool$stats$total_tasks <- pool$stats$total_tasks + 1L
      .pool_env$pool <- pool
    }

    # Handle timeouts for in-flight chunks
    if (any(!idle)) {
      now <- Sys.time()
      for (worker_id in which(!idle)) {
        meta <- inflight[[worker_id]]
        if (is.null(meta)) next
        age <- as.numeric(difftime(now, meta$start_time, units = "secs"))
        if (!is.na(age) && age > timeout) {
          # Kill and restart the worker; requeue or permanently fail the chunk.
          chunk_id <- meta$chunk_id
          ch <- queue$env$in_flight[[chunk_id]]
          retry_count <- (ch$retry_count %||% 0L)
          requeue <- retry_count < max_retries
          queue_fail(queue, chunk_id, error = "timeout", requeue = requeue)
          if (!requeue) {
            warning(sprintf("Chunk %s permanently failed after %d retries: %s",
                            chunk_id, retry_count, "timeout"))
          }

          pool <- pool_restart_worker_(pool, worker_id, graceful = FALSE)

          # Re-export globals to the restarted worker
          export_env <- new.env(parent = emptyenv())
          export_env$.shard_dispatch_fun <- fun
          export_env$.shard_dispatch_args <- extra_args
          parallel::clusterExport(pool$workers[[worker_id]]$cluster,
                                  c(".shard_dispatch_fun", ".shard_dispatch_args"),
                                  envir = export_env)

          idle[worker_id] <- TRUE
          inflight[worker_id] <- list(NULL)
          inflight_class[worker_id] <- NA_character_
        }
      }
    }

    # Receive at most one completed chunk per loop iteration.
    rec <- recv_one(timeout_sec = 0.1)
    if (is.null(rec)) {
      if (!queue_has_pending(queue) && any(!idle)) {
        Sys.sleep(0.01)
      }
      next
    }

    worker_id <- rec$worker_id
    tag <- rec$tag
    payload <- rec$value

    if (isTRUE(rec$recv_error)) {
      # Treat a receive/unserialize failure as a worker death while running.
      meta <- inflight[[worker_id]]
      chunk_id <- meta$chunk_id %||% as.character(tag)
      if (!is.null(chunk_id) && !is.na(chunk_id)) {
        ch <- queue$env$in_flight[[as.character(chunk_id)]]
        retry_count <- (ch$retry_count %||% 0L)
        requeue <- retry_count < max_retries
        queue_fail(queue, as.character(chunk_id), error = payload$error %||% "recv error", requeue = requeue)
        if (!requeue) {
          warning(sprintf("Chunk %s permanently failed after %d retries: %s",
                          as.character(chunk_id), retry_count, payload$error %||% "recv error"))
        }
      }

      pool <- pool_restart_worker_(pool, worker_id, graceful = FALSE)

      export_env <- new.env(parent = emptyenv())
      export_env$.shard_dispatch_fun <- fun
      export_env$.shard_dispatch_args <- extra_args
      parallel::clusterExport(pool$workers[[worker_id]]$cluster,
                              c(".shard_dispatch_fun", ".shard_dispatch_args"),
                              envir = export_env)

      idle[worker_id] <- TRUE
      inflight[worker_id] <- list(NULL)
      inflight_class[worker_id] <- NA_character_
      next
    }

    idle[worker_id] <- TRUE
    inflight[worker_id] <- list(NULL)
    inflight_class[worker_id] <- NA_character_

    if (is.list(payload) && isTRUE(payload$ok)) {
      val <- payload$value
      if (!is.null(on_result)) {
        if (!is.function(on_result)) stop("on_result must be a function or NULL", call. = FALSE)
        # If the reducer throws, fail fast: continuing would silently corrupt results.
        on_result(tag = tag, value = val, worker_id = worker_id)
      }
      queue_complete(queue, tag, if (isTRUE(store_results)) val else NULL, retain = retain_chunks)
    } else {
      err <- if (is.list(payload)) payload$error else "unknown worker error"
      ch <- queue$env$in_flight[[as.character(tag)]]
      retry_count <- (ch$retry_count %||% 0L)
      if (retry_count >= max_retries) {
        queue_fail(queue, tag, error = err, requeue = FALSE)
        warning(sprintf("Chunk %s permanently failed after %d retries: %s",
                        tag, retry_count, err))
      } else {
        queue_fail(queue, tag, error = err, requeue = TRUE)
      }
    }

    if (is.list(payload) && is.list(payload$view_delta)) {
      vd <- payload$view_delta
      view_stats$created <- view_stats$created + (vd$created %||% 0L)
      view_stats$materialized <- view_stats$materialized + (vd$materialized %||% 0L)
      view_stats$materialized_bytes <- view_stats$materialized_bytes + (vd$materialized_bytes %||% 0)
      view_stats$packed <- view_stats$packed + (vd$packed %||% 0L)
      view_stats$packed_bytes <- view_stats$packed_bytes + (vd$packed_bytes %||% 0)
    }

    if (is.list(payload) && is.list(payload$view_hotspots_delta)) {
      hsd <- payload$view_hotspots_delta
      for (k in names(hsd)) {
        cur <- view_hotspots[[k]] %||% list(bytes = 0, count = 0L)
        cur$bytes <- (cur$bytes %||% 0) + (hsd[[k]]$bytes %||% 0)
        cur$count <- as.integer((cur$count %||% 0L) + (hsd[[k]]$count %||% 0L))
        view_hotspots[[k]] <- cur
      }
    }

    if (is.list(payload) && is.list(payload$copy_delta)) {
      cd <- payload$copy_delta
      copy_stats$buffer_writes <- copy_stats$buffer_writes + (cd$buffer_writes %||% 0L)
      copy_stats$buffer_bytes <- copy_stats$buffer_bytes + (cd$buffer_bytes %||% 0)
    }

    if (is.list(payload) && is.list(payload$table_delta)) {
      td <- payload$table_delta
      table_stats$writes <- table_stats$writes + (td$table_writes %||% 0L)
      table_stats$rows <- table_stats$rows + (td$table_rows %||% 0L)
      table_stats$bytes <- table_stats$bytes + (td$table_bytes %||% 0)
    }

    if (is.list(payload) && is.list(payload$scratch_delta)) {
      sd <- payload$scratch_delta
      scratch_stats$hits <- scratch_stats$hits + (sd$hits %||% 0L)
      scratch_stats$misses <- scratch_stats$misses + (sd$misses %||% 0L)
      scratch_stats$high_water <- max(as.double(scratch_stats$high_water), as.double(sd$high_water %||% 0))
    }

    if (isTRUE(payload$scratch_needs_recycle)) {
      # Request a recycle at the next safe point (before the worker is reused).
      w <- pool$workers[[worker_id]]
      if (!is.null(w)) {
        w$needs_recycle <- TRUE
        pool$workers[[worker_id]] <- w
        .pool_env$pool <- pool
      }
    }

    chunks_processed <- chunks_processed + 1L
  }

  diag$end_time <- Sys.time()
  diag$duration <- as.numeric(difftime(diag$end_time, diag$start_time, units = "secs"))
  diag$view_stats <- view_stats
  if (length(view_hotspots) > 0) {
    ord <- order(vapply(view_hotspots, function(x) as.double(x$bytes %||% 0), numeric(1)), decreasing = TRUE)
    view_hotspots <- view_hotspots[ord]
    if (length(view_hotspots) > 20) view_hotspots <- view_hotspots[seq_len(20)]
    diag$view_hotspots <- view_hotspots
  } else {
    diag$view_hotspots <- list()
  }
  diag$copy_stats <- copy_stats
  diag$table_stats <- table_stats
  diag$scratch_stats <- scratch_stats

  structure(
    list(
      results = queue_results(queue),
      failures = queue_failures(queue),
      queue_status = queue_status(queue),
      diagnostics = diag,
      pool_stats = pool_get()$stats
    ),
    class = "shard_dispatch_result"
  )
}

#' Parallel Dispatch with Async Workers
#'
#' An alternative dispatch that uses parallel::parLapply-style execution
#' but with supervision. This is a simpler interface for basic parallel apply.
#'
#' @param X List or vector to iterate over.
#' @param FUN Function to apply to each element.
#' @param ... Additional arguments to FUN.
#' @param pool A `shard_pool` object. If NULL, uses current pool.
#' @param chunk_size Integer. Elements per chunk (default 1).
#'
#' @return A list of results.
#' @export
#' @examples
#' \donttest{
#' pool_create(2)
#' result <- pool_lapply(1:4, function(x) x^2, pool = pool_get())
#' pool_stop()
#' }
pool_lapply <- function(X, FUN, ..., pool = NULL, chunk_size = 1L) {
  if (is.null(pool)) {
    pool <- pool_get()
  }

  if (is.null(pool)) {
    stop("No active pool. Call pool_create() first.", call. = FALSE)
  }

  # Capture extra args
  extra_args <- list(...)

  # Create chunks with the actual function embedded
  n <- length(X)
  chunks <- lapply(seq_len(ceiling(n / chunk_size)), function(i) {
    start <- (i - 1L) * chunk_size + 1L
    end <- min(i * chunk_size, n)
    list(
      id = i,
      indices = start:end,
      elements = X[start:end],
      FUN = FUN,
      extra_args = extra_args
    )
  })

  # Wrapper function that processes a chunk
  chunk_fun <- function(chunk) {
    FUN <- chunk$FUN
    extra_args <- chunk$extra_args
    lapply(chunk$elements, function(x) do.call(FUN, c(list(x), extra_args)))
  }

  # Dispatch
  result <- dispatch_chunks(chunks, chunk_fun, pool = pool)

  # Flatten results
  unlist(result$results, recursive = FALSE)
}

#' Parallel sapply with Supervision
#'
#' @param X List or vector to iterate over.
#' @param FUN Function to apply.
#' @param ... Additional arguments to FUN.
#' @param simplify Logical. Simplify result to vector/matrix?
#' @param pool A `shard_pool` object. If NULL, uses current pool.
#'
#' @return Simplified result if possible, otherwise a list.
#' @export
#' @examples
#' \donttest{
#' pool_create(2)
#' result <- pool_sapply(1:4, function(x) x^2, pool = pool_get())
#' pool_stop()
#' }
pool_sapply <- function(X, FUN, ..., simplify = TRUE, pool = NULL) {
  result <- pool_lapply(X, FUN, ..., pool = pool)

  if (simplify) {
    tryCatch(
      simplify2array(result),
      error = function(e) result
    )
  } else {
    result
  }
}

#' Print a shard_dispatch_result Object
#'
#' @param x A \code{shard_dispatch_result} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @export
#' @examples
#' \donttest{
#' pool_create(2)
#' chunks <- list(list(id = 1L, x = 1), list(id = 2L, x = 2))
#' result <- dispatch_chunks(chunks, function(chunk) chunk$x, pool = pool_get())
#' print(result)
#' pool_stop()
#' }
print.shard_dispatch_result <- function(x, ...) {
  cat("shard dispatch result\n")
  cat("  Duration:", sprintf("%.2f seconds", x$diagnostics$duration), "\n")

  status <- x$queue_status
  cat("  Chunks: ", status$completed, " completed",
      if (status$failed > 0) paste0(", ", status$failed, " failed") else "",
      if (status$total_retries > 0) paste0(" (", status$total_retries, " retries)") else "",
      "\n", sep = "")

  cat("  Pool stats: ", x$pool_stats$total_recycles, " recycles, ",
      x$pool_stats$total_deaths, " deaths\n", sep = "")

  if (length(x$failures) > 0) {
    cat("  Failed chunks:", paste(names(x$failures), collapse = ", "), "\n")
  }

  invisible(x)
}

test_that("dispatch_chunks processes all chunks", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  chunks <- lapply(1:5, function(i) list(id = i, value = i * 10))

  result <- dispatch_chunks(
    chunks,
    fun = function(chunk) chunk$value + 1
  )

  expect_s3_class(result, "shard_dispatch_result")
  expect_equal(result$queue_status$completed, 5L)
  expect_equal(result$queue_status$failed, 0L)

  # Check results (ignore names)
  results <- unname(unlist(result$results))
  expect_equal(sort(results), c(11, 21, 31, 41, 51))
})

test_that("pool_lapply works like lapply", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  x <- 1:10
  result <- pool_lapply(x, function(i) i^2)

  # Compare values, ignoring names
  expect_equal(unname(result), unname(as.list(x^2)))
})

test_that("pool_sapply simplifies results", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  x <- 1:5
  result <- pool_sapply(x, function(i) i * 2)

  # Compare values, ignoring names
  expect_equal(unname(result), unname(x * 2))
})

test_that("dispatch handles actual worker death by restarting and requeueing work", {
  skip_on_cran()
  skip_if_conn_exhausted()
  if (is_windows()) skip("signal-based worker tests require POSIX")

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  kill_flag <- tempfile("shard-dispatch-kill-")
  on.exit(unlink(kill_flag), add = TRUE)

  # Create chunks, one of which will kill the worker
  chunks <- list(
    list(id = 1, action = "normal"),
    list(id = 2, action = "normal"),
    list(id = 3, action = "kill"),  # This one kills the worker
    list(id = 4, action = "normal"),
    list(id = 5, action = "normal")
  )

  result <- suppressWarnings(
    dispatch_chunks(
      chunks,
      fun = function(chunk, flag) {
        if (chunk$action == "kill" && !file.exists(flag)) {
          file.create(flag)
          tools::pskill(Sys.getpid(), signal = 9L)
        }
        chunk$id
      },
      flag = kill_flag,
      health_check_interval = 1L,
      max_retries = 2L,
      timeout = 1
    )
  )

  expect_equal(result$queue_status$completed, 5L)
  expect_equal(result$queue_status$failed, 0L)
  expect_equal(sort(unname(unlist(result$results))), 1:5)
  expect_gte(pool_get()$stats$total_deaths, 1L)
})

test_that("dispatch respects max_retries", {
  skip_on_cran()

  pool <- pool_create(n = 1)
  on.exit(pool_stop())

  # Chunk that always fails
  chunks <- list(list(id = 1))

  result <- suppressWarnings(
    dispatch_chunks(
      chunks,
      fun = function(chunk) stop("always_fails"),
      max_retries = 2
    )
  )

  # Should be marked as failed after retries exhausted
  expect_equal(result$queue_status$failed, 1L)
  expect_equal(result$queue_status$completed, 0L)
  expect_gte(result$queue_status$total_retries, 2L)
})

test_that("print.shard_dispatch_result produces output", {
  skip_on_cran()

  pool <- pool_create(n = 1)
  on.exit(pool_stop())

  chunks <- list(list(id = 1, val = 42))
  result <- dispatch_chunks(chunks, fun = function(chunk) chunk$val)

  output <- capture.output(print(result))
  expect_true(any(grepl("shard dispatch result", output)))
  expect_true(any(grepl("Duration", output)))
})

#' @title Worker Pool Management
#' @description Spawn and supervise persistent R worker processes with RSS monitoring.
#' @name pool
NULL

# Module-level pool state (package environment)
.pool_env <- new.env(parent = emptyenv())
.pool_env$pool <- NULL
.pool_env$dev_path <- NULL

#' Create a Worker Pool
#'
#' Spawns N R worker processes that persist across multiple `shard_map()` calls.
#' Workers are supervised and recycled when RSS drift exceeds thresholds.
#'
#' @param n Integer. Number of worker processes to spawn.
#' @param rss_limit Numeric or character. Maximum RSS per worker before recycling.
#'   Can be bytes (numeric) or human-readable (e.g., "2GB"). Default is "2GB".
#' @param rss_drift_threshold Numeric. Fraction of RSS increase from baseline
#'   that triggers recycling (default 0.5 = 50% growth).
#' @param heartbeat_interval Numeric. Seconds between health checks (default 5).
#' @param min_recycle_interval Numeric. Minimum time in seconds between recycling
#'   the same worker (default 1.0). This prevents thrashing PSOCK worker creation
#'   under extremely tight RSS limits.
#' @param init_expr Expression to evaluate in each worker on startup.
#' @param packages Character vector. Packages to load in workers.
#'
#' @return A `shard_pool` object (invisibly). The pool is also stored in the
#'   package environment for reuse.
#'
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' pool_stop(p)
#' }
#' @export
pool_create <- function(n = parallel::detectCores() - 1L,
                        rss_limit = "2GB",
                        rss_drift_threshold = 0.5,
                        heartbeat_interval = 5,
                        min_recycle_interval = 1.0,
                        init_expr = NULL,
                        packages = NULL) {
  # Validate inputs
  n <- as.integer(n)
  if (is.na(n)) n <- 1L
  if (n < 1L) {
    stop("pool_create: n must be >= 1", call. = FALSE)
  }

  rss_limit_bytes <- parse_bytes(rss_limit)

  # Stop existing pool if any
  if (!is.null(.pool_env$pool)) {
    pool_stop()
  }

  # Create pool structure
  dev_path <- tryCatch(getNamespaceInfo("shard", "path"), error = function(e) NULL)
  if (!is.null(dev_path) && isTRUE(file.exists(file.path(dev_path, ".git")))) {
    .pool_env$dev_path <- dev_path
  } else {
    # Best-effort dev mode: if we're in the *shard* repo root, use it as
    # dev_path so workers can load the in-tree R code without requiring build
    # tools.
    repo_root <- tryCatch(normalizePath(".", winslash = "/", mustWork = TRUE), error = function(e) NULL)
    if (!is.null(repo_root) &&
        isTRUE(file.exists(file.path(repo_root, ".git"))) &&
        isTRUE(file.exists(file.path(repo_root, "DESCRIPTION")))) {
      desc <- tryCatch(readLines(file.path(repo_root, "DESCRIPTION"), warn = FALSE), error = function(e) character())
      is_shard_pkg <- any(grepl("^Package:\\s*shard\\s*$", desc))
      looks_like_shard_repo <- isTRUE(file.exists(file.path(repo_root, "R", "pool.R"))) &&
        isTRUE(file.exists(file.path(repo_root, "R", "worker.R")))
      if (!isTRUE(is_shard_pkg && looks_like_shard_repo)) {
        .pool_env$dev_path <- NULL
        dev_path <- NULL
      } else {
        .pool_env$dev_path <- repo_root
        dev_path <- repo_root
      }
    } else {
      .pool_env$dev_path <- NULL
      dev_path <- NULL
    }
  }

  pool <- structure(
    list(
      workers = vector("list", n),
      n = n,
      rss_limit_bytes = rss_limit_bytes,
      rss_drift_threshold = rss_drift_threshold,
      heartbeat_interval = heartbeat_interval,
      min_recycle_interval = as.double(min_recycle_interval),
      dev_path = dev_path,
      init_expr = init_expr,
      packages = packages,
      created_at = Sys.time(),
      stats = list(
        total_recycles = 0L,
        total_deaths = 0L,
        total_tasks = 0L
      )
    ),
    class = "shard_pool"
  )

  # Spawn workers
  for (i in seq_len(n)) {
    pool$workers[[i]] <- worker_spawn(
      id = i,
      init_expr = init_expr,
      packages = packages,
      dev_path = dev_path
    )
  }

  # Record baseline RSS for each worker
  for (i in seq_len(n)) {
    pool$workers[[i]]$rss_baseline <- worker_rss(pool$workers[[i]])
  }

  .pool_env$pool <- pool
  invisible(pool)
}

#' Get the Current Worker Pool
#'
#' Returns the active worker pool, or NULL if none exists.
#'
#' @return A `shard_pool` object or NULL.
#' @examples
#' p <- pool_get()
#' is.null(p)
#' @export
pool_get <- function() {
  .pool_env$pool
}

pool_restart_worker_ <- function(pool, worker_id, graceful = TRUE) {
  old_worker <- pool$workers[[worker_id]]
  if (!is.null(old_worker)) {
    tryCatch(worker_kill(old_worker, graceful = graceful), error = function(e) NULL)
  }

  pool$workers[[worker_id]] <- worker_spawn(
    id = worker_id,
    init_expr = pool$init_expr,
    packages = pool$packages,
    dev_path = pool$dev_path
  )
  pool$workers[[worker_id]]$needs_recycle <- FALSE
  pool$workers[[worker_id]]$rss_baseline <- worker_rss(pool$workers[[worker_id]])
  pool$stats$total_deaths <- pool$stats$total_deaths + 1L
  .pool_env$pool <- pool

  pool
}

#' Stop the Worker Pool
#'
#' Terminates all worker processes and releases resources. Waits for workers
#' to actually terminate before returning.
#'
#' @param pool A `shard_pool` object. If NULL, uses the current pool.
#' @param timeout Numeric. Seconds to wait for workers to terminate (default 5).
#'   Returns after timeout even if workers are still alive.
#' @return NULL (invisibly).
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' pool_stop(p)
#' }
#' @export
pool_stop <- function(pool = NULL, timeout = 5) {
  if (is.null(pool)) {
    pool <- .pool_env$pool
  }

  if (is.null(pool)) {
    return(invisible(NULL))
  }

  # Collect PIDs before killing
  pids <- vapply(pool$workers, function(w) {
    if (is.null(w)) NA_integer_ else w$pid
  }, integer(1))
  alive_pids <- pids[!is.na(pids)]

  # Fast path: if all workers already dead, just clean up
  if (length(alive_pids) > 0) {
    alive_check <- vapply(alive_pids, pid_is_alive, logical(1))
    if (!any(alive_check)) {
      for (w in pool$workers) {
        if (!is.null(w)) {
          tryCatch(worker_close_connection_(w), error = function(e) NULL)
        }
      }
      .pool_env$pool <- NULL
      return(invisible(NULL))
    }
  }

  # Kill all workers
  for (w in pool$workers) {
    if (!is.null(w)) {
      tryCatch(worker_kill(w), error = function(e) NULL)
    }
  }

  # Wait for workers to terminate with timeout
  if (length(alive_pids) > 0 && timeout > 0) {
    deadline <- Sys.time() + timeout
    while (Sys.time() < deadline) {
      still_alive <- vapply(alive_pids, pid_is_alive, logical(1))
      if (!any(still_alive)) {
        break
      }
      Sys.sleep(0.1)
    }
  }

  .pool_env$pool <- NULL
  invisible(NULL)
}

#' Check Pool Health
#'
#' Monitors all workers, recycling those with excessive RSS drift or that have died.
#'
#' @param pool A `shard_pool` object. If NULL, uses the current pool.
#' @param busy_workers Optional integer vector of worker ids that are currently
#'   running tasks (used internally by the dispatcher to avoid recycling a worker
#'   while a result is in flight).
#' @return A list with health status per worker and actions taken.
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' pool_health_check(p)
#' pool_stop(p)
#' }
#' @export
pool_health_check <- function(pool = NULL, busy_workers = NULL) {
  if (is.null(pool)) {
    pool <- .pool_env$pool
  }

  if (is.null(pool)) {
    stop("No active pool. Call pool_create() first.", call. = FALSE)
  }

  actions <- vector("list", pool$n)
  busy <- rep(FALSE, pool$n)
  if (!is.null(busy_workers)) {
    busy_workers <- as.integer(busy_workers)
    busy_workers <- busy_workers[busy_workers >= 1L & busy_workers <= pool$n]
    busy[busy_workers] <- TRUE
  }

  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    action <- list(worker_id = i, action = "none", reason = NA_character_)

    if (is.null(w) || !worker_is_alive(w)) {
      # Worker died - restart it
      action$action <- "restart"
      action$reason <- "worker_dead"
      pool <- pool_restart_worker_(pool, i)
    } else {
      # Check RSS drift
      current_rss <- worker_rss(w)
      baseline <- w$rss_baseline %||% current_rss

      drift <- (current_rss - baseline) / max(baseline, 1)
      exceeds_limit <- current_rss > pool$rss_limit_bytes
      exceeds_drift <- drift > pool$rss_drift_threshold

      if (exceeds_limit || exceeds_drift) {
        action$reason <- if (exceeds_limit) "rss_limit" else "rss_drift"
        action$rss_before <- current_rss

        min_age <- pool$min_recycle_interval %||% 0
        age <- as.numeric(difftime(Sys.time(), w$spawned_at %||% Sys.time(), units = "secs"))

        if (isTRUE(busy[i]) || (!is.na(age) && is.finite(min_age) && age < min_age)) {
          # Avoid recycling a worker while it has an in-flight task. Recycling
          # closes the PSOCK connection and can race with recv/unserialize.
          w$needs_recycle <- TRUE
          pool$workers[[i]] <- w
          action$action <- "defer_recycle"
        } else {
          action$action <- "recycle"
          pool$workers[[i]] <- worker_recycle(w, pool$init_expr, pool$packages, dev_path = pool$dev_path)
          pool$workers[[i]]$needs_recycle <- FALSE
          pool$workers[[i]]$rss_baseline <- worker_rss(pool$workers[[i]])
          pool$stats$total_recycles <- pool$stats$total_recycles + 1L
        }
      } else {
        w$needs_recycle <- isTRUE(w$needs_recycle)
        pool$workers[[i]] <- w
        action$rss_current <- current_rss
        action$rss_drift <- drift
      }
    }

    actions[[i]] <- action
  }

  # Update pool in environment
  .pool_env$pool <- pool

  structure(
    list(
      timestamp = Sys.time(),
      pool_stats = pool$stats,
      worker_actions = actions
    ),
    class = "shard_health_report"
  )
}

#' Get Pool Status
#'
#' Returns current status of all workers in the pool.
#'
#' @param pool A `shard_pool` object. If NULL, uses the current pool.
#' @return A data frame with worker status information.
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' pool_status(p)
#' pool_stop(p)
#' }
#' @export
pool_status <- function(pool = NULL) {
  if (is.null(pool)) {
    pool <- .pool_env$pool
  }

  if (is.null(pool)) {
    return(data.frame(
      worker_id = integer(0),
      pid = integer(0),
      status = character(0),
      rss_bytes = numeric(0),
      rss_baseline = numeric(0),
      rss_drift = numeric(0),
      recycle_count = integer(0)
    ))
  }

  rows <- lapply(seq_len(pool$n), function(i) {
    w <- pool$workers[[i]]
    if (is.null(w)) {
      return(data.frame(
        worker_id = i,
        pid = NA_integer_,
        status = "missing",
        rss_bytes = NA_real_,
        rss_baseline = NA_real_,
        rss_drift = NA_real_,
        recycle_count = 0L
      ))
    }

    alive <- worker_is_alive(w)
    rss <- if (alive) worker_rss(w) else NA_real_
    baseline <- w$rss_baseline %||% NA_real_
    drift <- if (!is.na(rss) && !is.na(baseline) && baseline > 0) {
      (rss - baseline) / baseline
    } else {
      NA_real_
    }

    data.frame(
      worker_id = i,
      pid = w$pid,
      status = if (alive) "ok" else "dead",
      rss_bytes = rss,
      rss_baseline = baseline,
      rss_drift = drift,
      recycle_count = w$recycle_count %||% 0L
    )
  })

  do.call(rbind, rows)
}

#' Dispatch Task to Worker
#'
#' Sends a task to a specific worker and waits for the result.
#'
#' @param worker_id Integer. Worker to dispatch to.
#' @param expr Expression to evaluate.
#' @param envir Environment containing variables needed by expr.
#' @param pool A `shard_pool` object. If NULL, uses the current pool.
#' @param timeout Numeric. Seconds to wait for result (default 3600).
#'
#' @return The result of evaluating expr in the worker.
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' pool_dispatch(1, quote(1 + 1), pool = p)
#' pool_stop(p)
#' }
#' @export
pool_dispatch <- function(worker_id, expr, envir = parent.frame(),
                          pool = NULL, timeout = 3600) {
  if (is.null(pool)) {
    pool <- .pool_env$pool
  }

  if (is.null(pool)) {
    stop("No active pool. Call pool_create() first.", call. = FALSE)
  }

  if (worker_id < 1L || worker_id > pool$n) {
    stop("Invalid worker_id: ", worker_id, call. = FALSE)
  }

  w <- pool$workers[[worker_id]]
  if (is.null(w) || !worker_is_alive(w)) {
    # Restart dead worker before dispatch
    pool <- pool_restart_worker_(pool, worker_id)
    w <- pool$workers[[worker_id]]
  }

  pool$stats$total_tasks <- pool$stats$total_tasks + 1L
  .pool_env$pool <- pool

  worker_eval(w, expr, envir, timeout)
}

#' Print a shard_pool Object
#'
#' @param x A \code{shard_pool} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' print(p)
#' pool_stop(p)
#' }
#' @export
print.shard_pool <- function(x, ...) {
  cat("shard worker pool\n")
  cat("  Workers:", x$n, "\n")
  cat("  RSS limit:", format_bytes(x$rss_limit_bytes), "\n")
  cat("  Drift threshold:", sprintf("%.0f%%", x$rss_drift_threshold * 100), "\n")
  cat("  Created:", format(x$created_at), "\n")
  cat("  Stats:\n")
  cat("    Total recycles:", x$stats$total_recycles, "\n")
  cat("    Total deaths:", x$stats$total_deaths, "\n")
  cat("    Total tasks:", x$stats$total_tasks, "\n")
  invisible(x)
}

#' Print a shard_health_report Object
#'
#' @param x A \code{shard_health_report} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' r <- pool_health_check(p)
#' print(r)
#' pool_stop(p)
#' }
#' @export
print.shard_health_report <- function(x, ...) {
  cat("Pool health check at", format(x$timestamp), "\n")
  cat("Stats: recycles =", x$pool_stats$total_recycles,
      ", deaths =", x$pool_stats$total_deaths,
      ", tasks =", x$pool_stats$total_tasks, "\n")

  actions_summary <- vapply(x$worker_actions, function(a) a$action, character(1))
  if (all(actions_summary == "none")) {
    cat("All workers healthy\n")
  } else {
    cat("Actions taken:\n")
    for (a in x$worker_actions) {
      if (a$action != "none") {
        cat("  Worker", a$worker_id, ":", a$action, "(", a$reason, ")\n")
      }
    }
  }
  invisible(x)
}

test_that("pool_create creates workers", {
  skip_on_cran()

  pool <- pool_create(n = 2, rss_limit = "1GB")
  on.exit(pool_stop())

  expect_s3_class(pool, "shard_pool")
  expect_equal(pool$n, 2L)

  status <- pool_status()
  expect_equal(nrow(status), 2L)
  expect_true(all(status$status == "ok"))
})

test_that("pool_status shows worker information", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  status <- pool_status()

  expect_true("worker_id" %in% names(status))
  expect_true("pid" %in% names(status))
  expect_true("status" %in% names(status))
  expect_true("rss_bytes" %in% names(status))

  # PIDs should be valid
  expect_true(all(!is.na(status$pid)))
})

test_that("pool_dispatch executes code on workers", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  # Simple evaluation
  result <- pool_dispatch(1, quote(1 + 1))
  expect_equal(result, 2)

  # Get worker PID (should be different from main process)
  worker_pid <- pool_dispatch(1, quote(Sys.getpid()))
  expect_true(worker_pid != Sys.getpid())
})

test_that("pool_health_check detects dead workers", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  # Kill a worker
  worker_pid <- pool_dispatch(1, quote(Sys.getpid()))
  tools::pskill(worker_pid)
  Sys.sleep(0.2)  # Give it time to die

  # Health check should detect and restart
  health <- pool_health_check()

  # Worker 1 should have been restarted
  actions <- vapply(health$worker_actions, function(a) a$action, character(1))
  expect_true("restart" %in% actions)

  # Pool should have recorded a death
  pool <- pool_get()
  expect_gte(pool$stats$total_deaths, 1L)

  # Worker should be alive again
  status <- pool_status()
  expect_true(all(status$status == "ok"))
})

test_that("pool_stop clears the pool", {
  skip_on_cran()

  pool <- pool_create(n = 2)

  # Pool should exist
  expect_false(is.null(pool_get()))

  pool_stop()

  # Pool should be NULL after stop
  expect_null(pool_get())

  # Can create a new pool after stopping
  pool2 <- pool_create(n = 1)
  on.exit(pool_stop())
  expect_false(is.null(pool_get()))
})

test_that("pool_stop waits for workers to terminate", {
  skip_on_cran()

  pool <- pool_create(n = 2)

  # Get worker PIDs before stopping
  pids <- vapply(pool$workers, function(w) w$pid, integer(1))
  expect_true(all(!is.na(pids)))

  # Workers should be alive
  alive_before <- vapply(pids, pid_is_alive, logical(1))
  expect_true(all(alive_before))

  pool_stop()

  # Workers should be dead after pool_stop returns
  # Allow small buffer for OS process cleanup
  Sys.sleep(0.3)
  alive_after <- vapply(pids, pid_is_alive, logical(1))
  expect_true(all(!alive_after))
})

test_that("pool_stop returns immediately when workers already dead", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  pids <- vapply(pool$workers, function(w) w$pid, integer(1))

  # Kill workers manually first
  for (pid in pids) {
    tools::pskill(pid, signal = 9L)
  }
  Sys.sleep(0.3)  # Wait for processes to die

  # Verify workers are dead
  expect_true(all(!vapply(pids, pid_is_alive, logical(1))))

  # pool_stop should return quickly (fast path)
  start_time <- Sys.time()
  pool_stop()
  elapsed <- as.numeric(Sys.time() - start_time, units = "secs")

  # Should complete in well under the default 5s timeout
  expect_lt(elapsed, 1)
  expect_null(pool_get())
})

test_that("pool_stop respects timeout parameter", {
  skip_on_cran()

  pool <- pool_create(n = 1)

  # Stop with explicit timeout
  pool_stop(timeout = 2)
  expect_null(pool_get())
})

test_that("pool_get returns current pool", {
  # Initially NULL
  pool_stop()  # Ensure clean state
  expect_null(pool_get())

  skip_on_cran()

  # After creation, returns pool
  pool <- pool_create(n = 1)
  on.exit(pool_stop())

  expect_identical(pool_get(), pool)
})

test_that("pool creates workers with packages loaded", {
  skip_on_cran()

  pool <- pool_create(n = 1, packages = c("stats"))
  on.exit(pool_stop())

  # stats::sd should be available
  result <- pool_dispatch(1, quote(sd(1:10)))
  expect_equal(result, sd(1:10))
})

test_that("print.shard_pool produces output", {
  skip_on_cran()

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  output <- capture.output(print(pool))
  expect_true(any(grepl("shard worker pool", output)))
  expect_true(any(grepl("Workers: 2", output)))
})

test_that("shard_reduce combines results without gathering per-shard values", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  res <- shard_reduce(
    shards(100, block_size = 1),
    map = function(s) sum(s$idx),
    combine = function(acc, x) acc + x,
    init = 0,
    workers = 2,
    chunk_size = 10,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))
  expect_equal(res$value, sum(1:100))

  # Reduction diagnostics should be present and bounded.
  rd <- res$diagnostics$reduce %||% list()
  expect_true((rd$partials %||% 0L) > 0L)
  expect_true((rd$partial_max_bytes %||% 0) > 0)
})

test_that("shard_reduce accepts integer N as shards", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Pass integer directly instead of shard_descriptor
  res <- shard_reduce(
    50L,
    map = function(s) length(s$idx),
    combine = `+`,
    init = 0L,
    workers = 2
  )

  expect_true(succeeded(res))
  expect_equal(res$value, 50L)
})

test_that("shard_reduce print method works", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  res <- shard_reduce(
    20L,
    map = function(s) 1,
    combine = `+`,
    init = 0,
    workers = 2,
    diagnostics = TRUE
  )

  out <- capture.output(print(res))
  expect_true(any(grepl("shard_reduce result", out)))
  expect_true(any(grepl("Duration:", out)))
  expect_true(any(grepl("Partials:", out)))
})

test_that("shard_reduce validates map is a function", {
  expect_error(
    shard_reduce(10, map = "not_a_function", combine = `+`, init = 0),
    "map must be a function"
  )
})

test_that("shard_reduce validates combine is a function", {
  expect_error(
    shard_reduce(10, map = function(s) 1, combine = "not_a_function", init = 0),
    "combine must be a function"
  )
})

test_that("shard_reduce validates shards argument", {
  expect_error(
    shard_reduce("invalid", map = function(s) 1, combine = `+`, init = 0),
    "shards must be a shard_descriptor or integer"
  )
})

test_that("shard_reduce works with list combine operations", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Collect unique items using list append
  res <- shard_reduce(
    shards(10, block_size = 2),
    map = function(s) list(shard_id = s$id),
    combine = function(acc, x) c(acc, list(x)),
    init = list(),
    workers = 2
  )

  expect_true(succeeded(res))
  expect_true(is.list(res$value))
  expect_length(res$value, 5)  # 5 shards from 10 items with block_size 2
})

test_that("shard_reduce with borrow passes shared data", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  M <- share(matrix(1:100, nrow = 10))

  res <- shard_reduce(
    shards(10, block_size = 1),
    map = function(s, M) sum(M[s$idx, ]),
    combine = `+`,
    init = 0,
    borrow = list(M = M),
    workers = 2
  )

  expect_true(succeeded(res))
  expect_equal(res$value, sum(M))
})

test_that("shard_reduce with out= writes to buffer", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  out_buf <- buffer("double", dim = 10, init = 0)

  res <- shard_reduce(
    shards(10, block_size = 1),
    map = function(s, out) {
      out[s$id] <- s$id * 10
      s$id
    },
    combine = `+`,
    init = 0,
    out = list(out = out_buf),
    workers = 2
  )

  expect_true(succeeded(res))
  expect_equal(res$value, sum(1:10))
  expect_equal(as.numeric(out_buf[]), (1:10) * 10)
})

test_that("shard_reduce diagnostics can be disabled", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  res <- shard_reduce(
    20L,
    map = function(s) 1,
    combine = `+`,
    init = 0,
    workers = 2,
    diagnostics = FALSE
  )

  expect_true(succeeded(res))
  expect_null(res$diagnostics)
})

test_that("shard_reduce returns proper structure", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  res <- shard_reduce(
    shards(20, block_size = 5),
    map = function(s) 1,
    combine = `+`,
    init = 0,
    workers = 2,
    profile = "default",
    cow = "deny"
  )

  expect_s3_class(res, "shard_reduce_result")
  expect_true("value" %in% names(res))
  expect_true("failures" %in% names(res))
  expect_true("shards" %in% names(res))
  expect_true("queue_status" %in% names(res))
  expect_true("pool_stats" %in% names(res))
  expect_equal(res$cow_policy, "deny")
  expect_equal(res$profile, "default")
})

test_that("shard_reduce print shows failures when present", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Force a failure
  res <- shard_reduce(
    shards(10, block_size = 1),
    map = function(s) {
      if (s$id == 5) stop("forced error")
      1
    },
    combine = `+`,
    init = 0,
    workers = 2,
    max_retries = 0
  )

  out <- capture.output(print(res))
  expect_true(any(grepl("Failed chunks:", out)))
})

test_that("shard_map borrows language objects without evaluating them", {
  skip_on_cran()

  pool_create(n = 1, rss_limit = "2GB", rss_drift_threshold = 1)
  on.exit(pool_stop(), add = TRUE)

  ex <- quote({ i * 2 })

  res <- shard_map(
    shards = shards(1, workers = 1, block_size = 1),
    borrow = list(ex = ex),
    fun = function(shard, ex) {
      is.language(ex) && grepl("i\\s*\\*\\s*2", paste(deparse(ex), collapse = " "))
    },
    workers = 1,
    diagnostics = FALSE
  )

  expect_true(isTRUE(res$results[[1]][[1]]))
})

count_localhost_connections <- function() {
  con <- showConnections(all = TRUE)
  if (is.null(dim(con)) || nrow(con) == 0L) {
    return(0L)
  }

  sum(grepl("localhost", con[, "description"], fixed = TRUE))
}

wait_for_pid_exit <- function(pid, timeout = 2) {
  deadline <- Sys.time() + timeout
  while (Sys.time() < deadline) {
    if (!pid_is_alive(pid)) {
      return(TRUE)
    }
    Sys.sleep(0.05)
  }

  !pid_is_alive(pid)
}

test_that("pool_stop closes sockets for already-dead workers", {
  skip_on_cran()
  skip_if_conn_exhausted()
  if (is_windows()) skip("signal-based worker tests require POSIX")

  baseline <- count_localhost_connections()
  pool_create(n = 1)

  worker <- pool_get()$workers[[1]]
  worker_pid <- worker$pid

  expect_equal(count_localhost_connections(), baseline + 1L)

  tools::pskill(worker_pid, signal = 9L)
  expect_true(wait_for_pid_exit(worker_pid))

  pool_stop()

  expect_null(pool_get())
  expect_equal(count_localhost_connections(), baseline)
})

test_that("pool_health_check restarts dead workers without leaking sockets", {
  skip_on_cran()
  skip_if_conn_exhausted()
  if (is_windows()) skip("signal-based worker tests require POSIX")

  baseline <- count_localhost_connections()
  pool_create(n = 1)
  on.exit(pool_stop(), add = TRUE)

  old_worker <- pool_get()$workers[[1]]
  old_pid <- old_worker$pid

  tools::pskill(old_pid, signal = 9L)
  expect_true(wait_for_pid_exit(old_pid))

  health <- pool_health_check()
  new_worker <- pool_get()$workers[[1]]

  expect_equal(health$worker_actions[[1]]$action, "restart")
  expect_true(worker_is_alive(new_worker))
  expect_false(identical(new_worker$pid, old_pid))
  expect_equal(count_localhost_connections(), baseline + 1L)
})

test_that("pool_dispatch restarts dead workers without leaking sockets", {
  skip_on_cran()
  skip_if_conn_exhausted()
  if (is_windows()) skip("signal-based worker tests require POSIX")

  baseline <- count_localhost_connections()
  pool_create(n = 1)
  on.exit(pool_stop(), add = TRUE)

  old_worker <- pool_get()$workers[[1]]
  old_pid <- old_worker$pid

  tools::pskill(old_pid, signal = 9L)
  expect_true(wait_for_pid_exit(old_pid))

  expect_equal(pool_dispatch(1, quote(21 + 21)), 42)
  expect_true(worker_is_alive(pool_get()$workers[[1]]))
  expect_false(identical(pool_get()$workers[[1]]$pid, old_pid))
  expect_equal(count_localhost_connections(), baseline + 1L)
})

test_that("pool_dispatch timeout kills the hung worker and leaves the pool reusable", {
  skip_on_cran()
  skip_if_conn_exhausted()

  baseline <- count_localhost_connections()
  pool_create(n = 1)
  on.exit(pool_stop(), add = TRUE)

  start_time <- Sys.time()
  expect_error(
    pool_dispatch(1, quote(Sys.sleep(2)), timeout = 0.2),
    "timed out"
  )
  elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  expect_lt(elapsed, 1)
  expect_false(worker_is_alive(pool_get()$workers[[1]]))
  expect_equal(count_localhost_connections(), baseline)

  expect_equal(pool_dispatch(1, quote(1 + 1)), 2)
  expect_true(worker_is_alive(pool_get()$workers[[1]]))
  expect_equal(count_localhost_connections(), baseline + 1L)
})

test_that("shard_crossprod computes correct result for small matrices", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Use larger matrices to work with autotune, or specify explicit block sizes
  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(100), nrow = 10, ncol = 10)

  # Specify block sizes explicitly for small matrices
  res <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                         materialize = "always")

  expect_true(succeeded(res$run))
  expect_true(!is.null(res$value))
  expect_equal(dim(res$value), c(ncol(X), ncol(Y)))
  expect_equal(res$value, crossprod(X, Y), tolerance = 1e-10)
})

test_that("shard_crossprod with materialize='never' returns buffer only", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(100), nrow = 10, ncol = 10)

  res <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                         materialize = "never")

  expect_null(res$value)
  expect_s3_class(res$buffer, "shard_buffer")
  expect_equal(as.matrix(res$buffer), crossprod(X, Y), tolerance = 1e-10)
})

test_that("shard_crossprod with materialize='auto' respects threshold", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(100), nrow = 10, ncol = 10)

  # Very low threshold - should not materialize
  res1 <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                          materialize = "auto", materialize_max_bytes = 1)
  expect_null(res1$value)

  # Very high threshold - should materialize
  res2 <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                          materialize = "auto", materialize_max_bytes = 1e9)
  expect_true(!is.null(res2$value))
})

test_that("shard_crossprod accepts pre-shared matrices", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- share(matrix(rnorm(100), nrow = 10, ncol = 10))
  Y <- share(matrix(rnorm(100), nrow = 10, ncol = 10))

  res <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                         materialize = "always")

  expect_true(succeeded(res$run))
  expect_equal(res$value, crossprod(X, Y), tolerance = 1e-10)
})

test_that("shard_crossprod validates input types", {
  expect_error(shard_crossprod(1:10, matrix(1:10, nrow = 10)),
               "X and Y must be matrices")
  expect_error(shard_crossprod(matrix(1:10, nrow = 10), 1:10),
               "X and Y must be matrices")
})

test_that("shard_crossprod validates double type", {
  X <- matrix(1L:100L, nrow = 10)
  Y <- matrix(1L:100L, nrow = 10)

  expect_error(shard_crossprod(X, Y),
               "shard_crossprod currently supports double matrices only")
})

test_that("shard_crossprod validates row count match", {
  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(50), nrow = 5, ncol = 10)

  expect_error(shard_crossprod(X, Y),
               "X and Y must have the same number of rows")
})

test_that("shard_crossprod returns tile sizes", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(100), nrow = 10, ncol = 10)

  res <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5)

  expect_true("tile" %in% names(res))
  expect_equal(res$tile[["block_x"]], 5L)
  expect_equal(res$tile[["block_y"]], 5L)
})

test_that("shard_crossprod returns shard_map result", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(100), nrow = 10, ncol = 10)

  res <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                         diagnostics = TRUE)

  expect_true("run" %in% names(res))
  expect_s3_class(res$run, "shard_result")
})

test_that("shard_crossprod validates block sizes", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(100), nrow = 10, ncol = 10)

  expect_error(shard_crossprod(X, Y, block_x = 0, block_y = 5),
               "block_x must be >= 1")
  expect_error(shard_crossprod(X, Y, block_x = 5, block_y = 0),
               "block_y must be >= 1")
})

test_that("shard_crossprod supports different backing types", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(rnorm(100), nrow = 10, ncol = 10)
  Y <- matrix(rnorm(100), nrow = 10, ncol = 10)
  expected <- crossprod(X, Y)

  res_mmap <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                               backing = "mmap", materialize = "always")
  expect_equal(res_mmap$value, expected, tolerance = 1e-10)

  # shm backing might not be available everywhere
  tryCatch({
    res_shm <- shard_crossprod(X, Y, workers = 2, block_x = 5, block_y = 5,
                               backing = "shm", materialize = "always")
    expect_equal(res_shm$value, expected, tolerance = 1e-10)
  }, error = function(e) {
    # shm not supported, skip
  })
})

test_that("shard_crossprod handles rectangular matrices", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # More cols than rows - use explicit block sizes since autotune
  # has issues with small dimensions < 16
  X <- matrix(rnorm(200), nrow = 10, ncol = 20)
  Y <- matrix(rnorm(150), nrow = 10, ncol = 15)

  res <- shard_crossprod(X, Y, workers = 2, block_x = 10, block_y = 8,
                         materialize = "always")

  expect_equal(dim(res$value), c(20, 15))
  expect_equal(res$value, crossprod(X, Y), tolerance = 1e-10)
})

test_that("autotune_crossprod_tiles returns valid tile sizes for large matrices", {
  # Use larger matrices so candidates pass filtering
  X <- share(matrix(rnorm(25600), nrow = 100, ncol = 256))
  Y <- share(matrix(rnorm(25600), nrow = 100, ncol = 256))

  tiles <- shard:::.autotune_crossprod_tiles(X, Y)

  expect_true(is.list(tiles))
  expect_true("block_x" %in% names(tiles))
  expect_true("block_y" %in% names(tiles))
  expect_true(tiles$block_x >= 1 && tiles$block_x <= ncol(X))
  expect_true(tiles$block_y >= 1 && tiles$block_y <= ncol(Y))
})

test_that("autotune_crossprod_tiles handles small matrices", {
  # Small matrices where v >= 16 (minimum candidate in cand_y)
  # For matrices with ncol < 16, autotune currently has a bug returning -Inf
  X <- share(matrix(rnorm(640), nrow = 10, ncol = 64))
  Y <- share(matrix(rnorm(320), nrow = 10, ncol = 32))

  tiles <- shard:::.autotune_crossprod_tiles(X, Y)

  expect_true(is.list(tiles))
  expect_true(tiles$block_x >= 1)
  expect_true(tiles$block_y >= 1)
})

test_that("autotune_crossprod_tiles respects fixed dimensions", {
  # Use larger matrices so at least one candidate passes
  X <- share(matrix(rnorm(25600), nrow = 100, ncol = 256))
  Y <- share(matrix(rnorm(25600), nrow = 100, ncol = 256))

  tiles <- shard:::.autotune_crossprod_tiles(X, Y, block_x = 4, block_y = "auto")

  expect_equal(tiles$block_x, 4L)
  expect_true(tiles$block_y >= 1)
})

test_that("shard_crossprod with auto block sizes on large matrices", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Large enough for auto-tuning to work
  X <- matrix(rnorm(25600), nrow = 100, ncol = 256)
  Y <- matrix(rnorm(25600), nrow = 100, ncol = 256)

  res <- shard_crossprod(X, Y, workers = 2, materialize = "always")

  expect_true(succeeded(res$run))
  expect_equal(res$value, crossprod(X, Y), tolerance = 1e-10)
})

# Stress tests for worker recycling and supervision under load
#
# These tests exercise:
# - Worker recycling under memory pressure
# - Worker crash recovery mid-task
# - Correctness verification despite recycling/deaths
# - RSS bounds enforcement
#
# Expected runtime: 30-60 seconds per test due to high shard counts

skip_if_not_stress <- function() {
  if (!identical(tolower(Sys.getenv("SHARD_RUN_STRESS", "false")), "true")) {
    skip("Set SHARD_RUN_STRESS=true to run stress tests")
  }
}

test_that("stress: 1000+ shards with aggressive mem_cap triggers recycling", {
  skip_on_cran()
  skip_if_not_stress()
  # Expected runtime: ~30 seconds

  # Use very low mem_cap to force recycling
  pool <- pool_create(
    n = 2,
    rss_limit = "50MB",
    rss_drift_threshold = 0.1  # 10% growth triggers recycle
  )
  on.exit(pool_stop())

  # Run 1000 shards - each allocates memory to push RSS
  blocks <- shards(1000, block_size = 10, workers = 2)
  expected_bonus <- sum(matrix(seq_len(1e4), nrow = 100)[1, ])

  result <- shard_map(blocks,
    fun = function(shard) {
      # Allocate some memory to push RSS
      junk <- matrix(seq_len(1e4), nrow = 100)
      sum(shard$idx) + sum(junk[1, ])
    },
    mem_cap = "50MB",
    profile = "memory",  # Aggressive recycling
    workers = 2,
    health_check_interval = 5L
  )

  # Verify all shards completed successfully
  expect_true(succeeded(result))
  expect_equal(length(results(result)), 100)  # 1000/10 = 100 chunks

  # Verify at least some recycling occurred
  pool <- pool_get()
  expect_gte(pool$stats$total_recycles, 1L)

  # Verify results are correct sums (not corrupted by recycling)
  res <- unname(unlist(results(result)))
  expected <- vapply(seq_len(100), function(i) {
    start <- (i - 1L) * 10L + 1L
    sum(start:(start + 9L)) + expected_bonus
  }, numeric(1))
  expect_equal(res, expected)
})

test_that("stress: worker crashes mid-task are recovered", {
  skip_on_cran()
  skip_if_not_stress()
  if (is_windows()) skip("signal-based worker tests require POSIX")
  # Expected runtime: ~20 seconds

  pool <- pool_create(n = 2)
  on.exit(pool_stop())
  kill_flag <- tempfile("shard-stress-kill-")
  on.exit(unlink(kill_flag), add = TRUE)

  blocks <- shards(100, block_size = 10, workers = 2)

  result <- shard_map(blocks,
    fun = function(shard, flag) {
      if (shard$id == 5L && !file.exists(flag)) {
        file.create(flag)
        tools::pskill(Sys.getpid(), signal = 9L)
      }
      sum(shard$idx)
    },
    borrow = list(flag = kill_flag),
    workers = 2,
    max_retries = 5L
  )

  # Should succeed
  expect_true(succeeded(result))
  expect_equal(length(results(result)), 10)  # 100/10 = 10 chunks

  # Verify results are correct (using unname to strip names)
  res <- unname(unlist(results(result)))
  expected_first <- sum(1:10)  # First chunk
  expect_equal(res[1], expected_first)
})

test_that("stress: deliberate worker kill is recovered", {
  skip_on_cran()
  skip_if_not_stress()
  # Expected runtime: ~15 seconds

  pool <- pool_create(n = 2)
  on.exit(pool_stop())

  # Kill a worker before running shards
  worker_pid <- pool_get()$workers[[1]]$pid

  blocks <- shards(50, block_size = 5, workers = 2)

  # Kill worker 1 right before we start
  tools::pskill(worker_pid, signal = 9L)  # SIGKILL
  Sys.sleep(0.2)  # Let the death register

  result <- shard_map(blocks,
    fun = function(shard) {
      sum(shard$idx)
    },
    workers = 2,
    max_retries = 3L
  )

  # Should succeed - pool_dispatch auto-restarts dead workers
  expect_true(succeeded(result))

  # Verify no chunks were lost - all results present
  res <- results(result)
  expect_equal(length(res), 10)  # 50/5 = 10 chunks
  expect_true(all(!is.na(unlist(res))))

  # Verify correct results despite kill (sum of 1:5 for first chunk)
  first_sum <- unname(unlist(res))[1]
  expect_equal(first_sum, sum(1:5))
})

test_that("stress: verify no chunks lost during recycling", {
  skip_on_cran()
  skip_if_not_stress()
  # Expected runtime: ~30 seconds

  pool <- pool_create(
    n = 2,
    rss_limit = "30MB",
    rss_drift_threshold = 0.05  # Very aggressive recycling
  )
  on.exit(pool_stop())

  # Run many shards with frequent recycling
  n_items <- 500
  blocks <- shards(n_items, block_size = 5, workers = 2)

  result <- shard_map(blocks,
    fun = function(shard) {
      # Allocate memory to trigger recycling
      junk <- rnorm(5e4)
      list(
        id = shard$id,
        idx = shard$idx,
        sum = sum(shard$idx)
      )
    },
    mem_cap = "30MB",
    profile = "memory",
    workers = 2,
    health_check_interval = 3L
  )

  expect_true(succeeded(result))

  # Flatten and verify all shards present
  res <- results(result, flatten = FALSE)
  all_results <- unlist(res, recursive = FALSE)

  # Verify we have all 100 chunk results (500/5 = 100)
  expect_equal(length(all_results), 100)

  # Verify all shard IDs are present (1 to 100)
  ids <- vapply(all_results, function(x) x$id, integer(1))
  expect_true(setequal(ids, 1:100))

  # Verify sums are correct (use unname to avoid name comparison)
  sums <- unname(vapply(all_results, function(x) x$sum, numeric(1)))
  expected_sums <- vapply(1:100, function(i) {
    start <- (i - 1L) * 5L + 1L
    end <- i * 5L
    sum(start:end)
  }, numeric(1))
  expect_equal(sort(sums), sort(expected_sums))
})

test_that("stress: RSS stays bounded under load", {
  skip_on_cran()
  skip_if_not_stress()
  # Expected runtime: ~45 seconds

  rss_limit_bytes <- 100 * 1024^2  # 100MB limit

  pool <- pool_create(
    n = 2,
    rss_limit = "100MB",
    rss_drift_threshold = 0.3
  )
  on.exit(pool_stop())

  blocks <- shards(200, block_size = 10, workers = 2)

  result <- shard_map(blocks,
    fun = function(shard) {
      # Allocate significant memory
      big <- matrix(rnorm(1e5), nrow = 100)
      sum(shard$idx) + mean(big)
    },
    mem_cap = "100MB",
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(result))

  # Sample RSS after execution
  status <- pool_status()

  # Verify current RSS is below limit (or NA if unavailable)
  for (i in seq_len(nrow(status))) {
    rss <- status$rss_bytes[i]
    if (!is.na(rss)) {
      # Allow some slack (1.5x limit) due to GC timing
      expect_lte(rss, rss_limit_bytes * 1.5,
        label = sprintf("Worker %d RSS %.1fMB",
          status$worker_id[i], rss / 1024^2))
    }
  }

  # Verify recycling happened to enforce bounds
  pool <- pool_get()
  expect_gte(pool$stats$total_recycles + pool$stats$total_deaths, 1L)
})

test_that("stress: high shard count completes correctly", {
  skip_on_cran()
  skip_if_not_stress()
  # Expected runtime: ~30 seconds

  # Run without aggressive recycling to test correctness under load
  pool <- pool_create(n = 3, rss_limit = "500MB")
  on.exit(pool_stop())

  # 2000 shards - verify all complete correctly
  n <- 2000
  blocks <- shards(n, block_size = 20, workers = 3)

  result <- shard_map(blocks,
    fun = function(shard) {
      # Simple computation - verify shard indices
      list(
        id = shard$id,
        sum = sum(shard$idx),
        len = length(shard$idx)
      )
    },
    workers = 3
  )

  expect_true(succeeded(result))

  # Verify all chunks processed
  res <- results(result, flatten = FALSE)
  all_results <- unlist(res, recursive = FALSE)
  expect_equal(length(all_results), 100)  # 2000/20 = 100 chunks

  # Verify total sum matches expected
  total_sum <- sum(vapply(all_results, function(x) x$sum, numeric(1)))
  expected_sum <- sum(1:n)
  expect_equal(total_sum, expected_sum)
})

test_that("shard_map() executes simple function", {
  skip_on_cran()

  # Simple computation
  blocks <- shards(100, block_size = 25, workers = 2)

  result <- shard_map(blocks, function(shard) {
    sum(shard$idx)
  }, workers = 2)

  expect_s3_class(result, "shard_result")
  expect_true(succeeded(result))

  # Results should be correct sums
  res <- results(result)
  expect_equal(res[[1]], sum(1:25))
  expect_equal(res[[2]], sum(26:50))
  expect_equal(res[[3]], sum(51:75))
  expect_equal(res[[4]], sum(76:100))

  # Cleanup
  pool_stop()
})

test_that("shard_map() handles borrowed inputs", {
  skip_on_cran()

  # Create test data
  X <- matrix(1:100, nrow = 10, ncol = 10)

  blocks <- shards(ncol(X), block_size = 2, workers = 2)

  result <- shard_map(blocks,
    borrow = list(X = X),
    fun = function(shard, X) {
      colSums(X[, shard$idx, drop = FALSE])
    },
    workers = 2
  )

  expect_true(succeeded(result))

  # Results should be column sums
  res <- unlist(results(result))
  expect_equal(unname(res), unname(colSums(X)))

  pool_stop()
})

test_that("shard_map() accepts integer n directly", {
  skip_on_cran()

  result <- shard_map(10, function(shard) {
    length(shard$idx)
  }, workers = 2)

  expect_s3_class(result, "shard_result")
  expect_true(succeeded(result))

  pool_stop()
})

test_that("shard_map() respects chunk_size", {
  skip_on_cran()

  blocks <- shards(100, block_size = 10, workers = 2)

  # With chunk_size = 5, we dispatch 2 chunks (each with 5 shards)
  result <- shard_map(blocks,
    fun = function(shard) {
      shard$id
    },
    chunk_size = 5,
    workers = 2
  )

  expect_true(succeeded(result))
  expect_equal(result$diagnostics$chunks_dispatched, 2)

  pool_stop()
})

test_that("shard_map() collects diagnostics", {
  skip_on_cran()

  blocks <- shards(20, block_size = 5, workers = 2)

  result <- shard_map(blocks,
    fun = function(shard) {
      Sys.sleep(0.01)
      sum(shard$idx)
    },
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(result))
  expect_false(is.null(result$diagnostics))
  expect_gt(result$diagnostics$duration, 0)
  expect_equal(result$diagnostics$shards_processed, 4)

  pool_stop()
})

test_that("shard_map() uses existing pool", {
  skip_on_cran()

  # Create pool explicitly
  pool_create(n = 2)

  blocks <- shards(10, block_size = 5)

  result <- shard_map(blocks,
    fun = function(shard) {
      sum(shard$idx)
    }
    # Not specifying workers - should use existing pool
  )

  expect_true(succeeded(result))
  expect_equal(pool_get()$n, 2)

  pool_stop()
})

test_that("shard_map() handles profile settings", {
  skip_on_cran()

  blocks <- shards(10, block_size = 5, workers = 2)

  # Memory profile - more aggressive recycling
  result <- shard_map(blocks,
    fun = function(shard) sum(shard$idx),
    profile = "memory",
    workers = 2
  )

  expect_true(succeeded(result))
  expect_equal(result$profile, "memory")

  # Speed profile - less recycling
  result <- shard_map(blocks,
    fun = function(shard) sum(shard$idx),
    profile = "speed",
    workers = 2
  )

  expect_true(succeeded(result))
  expect_equal(result$profile, "speed")

  pool_stop()
})

test_that("shard_map() validates input", {
  skip_on_cran()

  # Invalid shards
  expect_error(
    shard_map("not a shard", function(x) x),
    "shard_descriptor"
  )

  # Invalid borrow
  expect_error(
    shard_map(10, function(x) x, borrow = 1:10),
    "named list"
  )

  # Unnamed borrow
  expect_error(
    shard_map(10, function(x) x, borrow = list(1:10)),
    "named"
  )

  pool_stop()
})

test_that("shard_map() print method works", {
  skip_on_cran()

  blocks <- shards(10, block_size = 5, workers = 2)
  result <- shard_map(blocks,
    fun = function(shard) sum(shard$idx),
    workers = 2
  )

  expect_output(print(result), "shard_map result")
  expect_output(print(result), "Duration")
  expect_output(print(result), "Completed")

  pool_stop()
})

test_that("results() extracts results correctly", {
  skip_on_cran()

  blocks <- shards(10, block_size = 5, workers = 2)
  result <- shard_map(blocks,
    fun = function(shard) list(sum = sum(shard$idx), len = length(shard$idx)),
    workers = 2
  )

  res <- results(result)
  expect_type(res, "list")
  expect_length(res, 2)
  expect_equal(res[[1]]$sum, sum(1:5))
  expect_equal(res[[2]]$sum, sum(6:10))

  pool_stop()
})

test_that("succeeded() reports correctly", {
  skip_on_cran()

  blocks <- shards(10, block_size = 5, workers = 2)

  # Successful run
  result <- shard_map(blocks,
    fun = function(shard) sum(shard$idx),
    workers = 2
  )
  expect_true(succeeded(result))

  pool_stop()
})

test_that("shard_map() with multiple borrowed inputs", {
  skip_on_cran()

  X <- matrix(1:20, nrow = 4)
  Y <- 1:5
  Z <- 100

  blocks <- shards(ncol(X), block_size = 2, workers = 2)

  result <- shard_map(blocks,
    borrow = list(X = X, Y = Y, Z = Z),
    fun = function(shard, X, Y, Z) {
      idx <- shard$idx
      sum(X[, idx]) + sum(Y[idx]) + Z
    },
    workers = 2
  )

  expect_true(succeeded(result))

  pool_stop()
})

test_that("shard_map() respects recycle setting", {
  skip_on_cran()

  blocks <- shards(10, block_size = 5, workers = 2)

  # Disable recycling
  result <- shard_map(blocks,
    fun = function(shard) sum(shard$idx),
    recycle = FALSE,
    workers = 2
  )

  expect_true(succeeded(result))

  # Custom recycle threshold
  result <- shard_map(blocks,
    fun = function(shard) sum(shard$idx),
    recycle = 0.25,  # 25% drift threshold
    workers = 2
  )

  expect_true(succeeded(result))

  pool_stop()
})

test_that("shard_map() handles COW policies", {
  skip_on_cran()

  blocks <- shards(10, block_size = 5, workers = 2)

  for (policy in c("deny", "audit", "allow")) {
    result <- shard_map(blocks,
      fun = function(shard) sum(shard$idx),
      cow = policy,
      workers = 2
    )

    expect_true(succeeded(result))
    expect_equal(result$cow_policy, policy)
  }

  pool_stop()
})

test_that("shard_map() handles single-element input", {
  skip_on_cran()

  # Single element with n=1
  result <- shard_map(1, function(shard) {
    sum(shard$idx)
  }, workers = 2)

  expect_s3_class(result, "shard_result")
  expect_true(succeeded(result))

  res <- results(result)
  expect_equal(res[[1]], 1)

  pool_stop()
})

test_that("shard_map() handles single shard", {
  skip_on_cran()

  # Single shard with block_size larger than n
  blocks <- shards(5, block_size = 100, workers = 2)
  expect_equal(blocks$num_shards, 1)

  result <- shard_map(blocks, function(shard) {
    sum(shard$idx)
  }, workers = 2)

  expect_true(succeeded(result))
  expect_equal(results(result)[[1]], sum(1:5))

  pool_stop()
})

test_that("shard_map() handles block_size equal to n", {
  skip_on_cran()

  # Exactly one shard
  blocks <- shards(10, block_size = 10, workers = 2)
  expect_equal(blocks$num_shards, 1)

  result <- shard_map(blocks, function(shard) {
    length(shard$idx)
  }, workers = 2)

  expect_true(succeeded(result))
  expect_equal(results(result)[[1]], 10)

  pool_stop()
})

test_that("shard_map() with borrowed input handles single element", {
  skip_on_cran()

  X <- matrix(1:10, nrow = 10, ncol = 1)

  blocks <- shards(ncol(X), block_size = 10, workers = 2)
  expect_equal(blocks$num_shards, 1)

  result <- shard_map(blocks,
    borrow = list(X = X),
    fun = function(shard, X) {
      sum(X[, shard$idx, drop = FALSE])
    },
    workers = 2
  )

  expect_true(succeeded(result))
  expect_equal(results(result)[[1]], sum(1:10))

  pool_stop()
})

#' @title Shared Memory Buffers
#' @name buffer
#' @description Create typed writable output buffers backed by shared memory
#'   for cross-process writes during parallel execution.
#'
#' @details
#' Buffers provide an explicit output mechanism for \code{\link{shard_map}}.
#' Instead of returning results from workers (which requires serialization
#' and memory copying), workers write directly to shared buffers.
#'
#' Supported types:
#' \itemize{
#'   \item \code{"double"}: 8-byte floating point (default)
#'   \item \code{"integer"}: 4-byte signed integer
#'   \item \code{"logical"}: 4-byte logical (stored as integer)
#'   \item \code{"raw"}: 1-byte raw data
#' }
#'
#' Buffers support slice assignment using standard R indexing:
#' \code{buf[1:100] <- values}
#'
#' @seealso \code{\link{segment_create}} for low-level segment operations,
#'   \code{\link{share}} for read-only shared inputs
NULL

# Type sizes in bytes
.buffer_type_size <- function(type) {
    switch(type,
        "double"  = 8L,
        "integer" = 4L,
        "logical" = 4L,
        "raw"     = 1L,
        stop("Unsupported buffer type: ", type)
    )
}

# R type code (SEXPTYPE) for type
.buffer_sexptype <- function(type) {
    switch(type,
        "double"  = 14L,  # REALSXP
        "integer" = 13L,  # INTSXP
        "logical" = 10L,  # LGLSXP
        "raw"     = 24L,  # RAWSXP
        stop("Unsupported buffer type: ", type)
    )
}

# Buffer write diagnostics (per-process). These counters are used to attribute
# output-buffer writes to shard_map runs via per-chunk deltas in the dispatcher.
.buffer_diag_env <- new.env(parent = emptyenv())
.buffer_diag_env$writes <- 0L
.buffer_diag_env$bytes <- 0

#' Buffer Diagnostics
#'
#' Returns per-process counters for shard buffer writes. shard_map uses these
#' internally to report write volume/operations in copy_report().
#'
#' @return A list with elements \code{writes} (integer count) and \code{bytes}
#'   (total bytes written) accumulated in the current process.
#' @export
#' @examples
#' buffer_diagnostics()
buffer_diagnostics <- function() {
    list(
        writes = .buffer_diag_env$writes,
        bytes = .buffer_diag_env$bytes
    )
}

# Reset counters (internal; tests may use shard:::buffer_reset_diagnostics()).
buffer_reset_diagnostics <- function() {
    .buffer_diag_env$writes <- 0L
    .buffer_diag_env$bytes <- 0
    invisible(NULL)
}

#' Create a Shared Memory Buffer
#'
#' Creates a typed output buffer backed by shared memory that can be written
#' to by parallel workers using slice assignment.
#'
#' @param type Character. Data type: "double" (default), "integer", "logical",
#'   or "raw".
#' @param dim Integer vector. Dimensions of the buffer. For a vector, specify
#'   the length. For a matrix, specify \code{c(nrow, ncol)}. For arrays,
#'   specify all dimensions.
#' @param init Initial value to fill the buffer. Default is type-appropriate
#'   zero (\code{0}, \code{0L}, \code{FALSE}, or \code{raw(0)}).
#' @param backing Backing type for shared memory: "auto" (default), "mmap",
#'   or "shm".
#'
#' @return An S3 object of class "shard_buffer" that supports:
#' \itemize{
#'   \item Slice assignment: \code{buf[idx] <- values}
#'   \item Slice reading: \code{buf[idx]}
#'   \item Full extraction: \code{buf[]}
#'   \item Conversion to R vector: \code{as.vector(buf)}, \code{as.double(buf)}, etc.
#' }
#'
#' @export
#' @examples
#' \donttest{
#' out <- buffer("double", dim = 100)
#' out[1:10] <- rnorm(10)
#' result <- out[]
#' }
buffer <- function(type = c("double", "integer", "logical", "raw"),
                   dim,
                   init = NULL,
                   backing = c("auto", "mmap", "shm")) {
    type <- match.arg(type)
    backing <- match.arg(backing)

    # Validate and normalize dim
    dim <- as.integer(dim)
    if (any(dim <= 0)) {
        stop("All dimensions must be positive")
    }
    n <- prod(dim)

    # Calculate byte size
    elem_size <- .buffer_type_size(type)
    byte_size <- as.double(n) * elem_size

    # Create underlying segment
    seg <- segment_create(byte_size, backing = backing)

    # Initialize buffer contents
    if (is.null(init)) {
        init <- switch(type,
            "double"  = 0,
            "integer" = 0L,
            "logical" = FALSE,
            "raw"     = raw(1)
        )
    }

    # Write initial value
    init_data <- switch(type,
        "double"  = rep(as.double(init), n),
        "integer" = rep(as.integer(init), n),
        "logical" = rep(as.logical(init), n),
        "raw"     = rep(as.raw(init), n)
    )
    segment_write(seg, init_data, offset = 0)

    structure(
        list(
            segment = seg,
            type = type,
            dim = dim,
            n = n,
            elem_size = elem_size
        ),
        class = "shard_buffer"
    )
}

#' Open an Existing Buffer
#'
#' Opens a shared memory buffer that was created in another process.
#' Used by workers to attach to the parent's output buffer.
#'
#' @param path Path or shm name of the buffer's segment.
#' @param type Character. Data type of the buffer.
#' @param dim Integer vector. Dimensions of the buffer.
#' @param backing Backing type: "mmap" or "shm".
#' @param readonly Logical. Open as read-only? Default FALSE for workers.
#'
#' @return A \code{shard_buffer} object attached to the existing segment.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 10)
#' path <- buffer_path(buf)
#' buf2 <- buffer_open(path, type = "double", dim = 10, backing = "mmap")
#' buffer_close(buf2, unlink = FALSE)
#' buffer_close(buf)
#' }
buffer_open <- function(path, type, dim,
                        backing = c("mmap", "shm"),
                        readonly = FALSE) {
    backing <- match.arg(backing)

    dim <- as.integer(dim)
    n <- prod(dim)
    elem_size <- .buffer_type_size(type)

    seg <- segment_open(path, backing = backing, readonly = readonly)

    # Verify size matches
    expected_size <- as.double(n) * elem_size
    actual_size <- segment_size(seg)
    if (actual_size < expected_size) {
        segment_close(seg)
        stop("Segment size mismatch: expected ", expected_size,
             " bytes, got ", actual_size)
    }

    structure(
        list(
            segment = seg,
            type = type,
            dim = dim,
            n = n,
            elem_size = elem_size
        ),
        class = "shard_buffer"
    )
}

#' Get Buffer Path
#'
#' Returns the path or name of the buffer's underlying segment.
#' Use this to pass buffer location to workers.
#'
#' @param x A shard_buffer object.
#' @return A character string with the path or name of the segment, or
#'   \code{NULL} if the segment is anonymous.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 10)
#' buffer_path(buf)
#' buffer_close(buf)
#' }
buffer_path <- function(x) {
    stopifnot(inherits(x, "shard_buffer"))
    segment_path(x$segment)
}

#' Get Buffer Info
#'
#' Returns information about a buffer.
#'
#' @param x A shard_buffer object.
#' @return A named list with buffer properties: \code{type}, \code{dim},
#'   \code{n}, \code{bytes}, \code{backing}, \code{path}, and \code{readonly}.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("integer", dim = c(5, 5))
#' buffer_info(buf)
#' buffer_close(buf)
#' }
buffer_info <- function(x) {
    stopifnot(inherits(x, "shard_buffer"))
    seg_info <- segment_info(x$segment)
    list(
        type = x$type,
        dim = x$dim,
        n = x$n,
        bytes = as.double(x$n) * x$elem_size,
        backing = seg_info$backing,
        path = seg_info$path,
        readonly = seg_info$readonly
    )
}

#' Close a Buffer
#'
#' Closes the buffer and releases the underlying shared memory.
#'
#' @param x A shard_buffer object.
#' @param unlink Whether to unlink the underlying segment.
#' @return \code{NULL}, invisibly.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 10)
#' buffer_close(buf)
#' }
buffer_close <- function(x, unlink = NULL) {
    stopifnot(inherits(x, "shard_buffer"))
    segment_close(x$segment, unlink = unlink)
    invisible(NULL)
}

#' Print a Shared Memory Buffer
#'
#' @param x A \code{shard_buffer} object.
#' @param ... Ignored.
#' @return The input \code{x}, invisibly.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 10)
#' print(buf)
#' buffer_close(buf)
#' }
print.shard_buffer <- function(x, ...) {
    info <- buffer_info(x)
    cat("<shard_buffer>\n")
    cat("  Type:", x$type, "\n")
    if (length(x$dim) == 1) {
        cat("  Length:", format(x$n, big.mark = ","), "\n")
    } else {
        cat("  Dim:", paste(x$dim, collapse = " x "), "\n")
    }
    cat("  Size:", format(info$bytes, big.mark = ","), "bytes\n")
    cat("  Backing:", info$backing, "\n")
    if (!is.null(info$path)) {
        cat("  Path:", info$path, "\n")
    }
    invisible(x)
}

#' Length of a Shared Memory Buffer
#'
#' @param x A \code{shard_buffer} object.
#' @return An integer scalar giving the total number of elements.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 20)
#' length(buf)
#' buffer_close(buf)
#' }
length.shard_buffer <- function(x) {
    x$n
}

#' Dimensions of a Shared Memory Buffer
#'
#' @param x A \code{shard_buffer} object.
#' @return An integer vector of dimensions, or \code{NULL} for 1-D buffers.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = c(4, 5))
#' dim(buf)
#' buffer_close(buf)
#' }
dim.shard_buffer <- function(x) {
    if (length(x$dim) == 1) NULL else x$dim
}

# Internal: read a range of elements from buffer
.buffer_read_range <- function(x, start, count) {
    offset <- (start - 1L) * x$elem_size
    nbytes <- count * x$elem_size

    raw_data <- segment_read(x$segment, offset = offset, size = nbytes)

    # Convert raw to appropriate type
    switch(x$type,
        "double"  = readBin(raw_data, "double", n = count),
        "integer" = readBin(raw_data, "integer", n = count),
        "logical" = as.logical(readBin(raw_data, "integer", n = count)),
        "raw"     = raw_data
    )
}

# Internal: write a range of elements to buffer
.buffer_write_range <- function(x, start, values) {
    count <- length(values)
    offset <- (start - 1L) * x$elem_size

    # Convert values to appropriate type and write
    typed_values <- switch(x$type,
        "double"  = as.double(values),
        "integer" = as.integer(values),
        "logical" = as.integer(as.logical(values)),  # Store as int
        "raw"     = as.raw(values)
    )

    segment_write(x$segment, typed_values, offset = offset)

    # Update per-process diagnostics after the write completes.
    .buffer_diag_env$writes <- .buffer_diag_env$writes + 1L
    .buffer_diag_env$bytes <- .buffer_diag_env$bytes + (count * x$elem_size)
}

# Internal: write arbitrary indices safely without read/modify/write.
# This is critical for disjoint parallel writes where indices are not one single
# contiguous run (e.g., singleton writes when block_size=1).
.buffer_write_indices <- function(x, idx, values) {
    idx <- as.integer(idx)
    if (length(idx) == 0) return(invisible(NULL))

    # If indices are strictly increasing, coalesce adjacent runs and write each
    # run as one contiguous segment write.
    if (length(idx) == 1L || (all(!is.na(idx)) && all(diff(idx) >= 0L) && !anyDuplicated(idx))) {
        run_start_pos <- 1L
        for (k in 2L:(length(idx) + 1L)) {
            is_break <- (k > length(idx)) || (idx[k] != (idx[k - 1L] + 1L))
            if (is_break) {
                start <- idx[run_start_pos]
                end_pos <- k - 1L
                .buffer_write_range(x, start, values[run_start_pos:end_pos])
                run_start_pos <- k
            }
        }
        return(invisible(NULL))
    }

    # Fallback: preserve R assignment semantics for duplicates/out-of-order by
    # performing writes in the given order (last write wins).
    for (k in seq_along(idx)) {
        .buffer_write_range(x, idx[k], values[k])
    }
    invisible(NULL)
}

#' Extract Buffer Elements
#'
#' @param x A shard_buffer object.
#' @param i Index or indices.
#' @param j Optional second index (for matrices).
#' @param ... Additional indices (for arrays).
#' @param drop Whether to drop dimensions.
#' @return A vector or array of values read from the buffer.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 10)
#' buf[1:5] <- 1:5
#' buf[1:3]
#' buffer_close(buf)
#' }
`[.shard_buffer` <- function(x, i, j, ..., drop = TRUE) {
    # Determine if this is matrix-style indexing by checking nargs()
    # nargs() > 2 means we have x[i, ...] style (at least one comma)
    is_matrix_style <- nargs() > 2

    # Handle empty index (return all)
    if (missing(i) && missing(j) && length(list(...)) == 0) {
        result <- .buffer_read_range(x, 1L, x$n)
        if (length(x$dim) > 1) {
            dim(result) <- x$dim
        }
        return(result)
    }

    # For 1D buffers or linear indexing (no comma used)
    if (length(x$dim) == 1 || !is_matrix_style) {
        if (missing(i)) {
            i <- seq_len(x$n)
        }
        i <- as.integer(i)

        # Check bounds
        if (any(i < 1 | i > x$n, na.rm = TRUE)) {
            stop("Index out of bounds")
        }

        # Optimize for contiguous ranges
        if (length(i) > 1 && all(diff(i) == 1L)) {
            result <- .buffer_read_range(x, i[1], length(i))
        } else {
            # Non-contiguous: read all and subset
            # TODO: optimize for sparse reads
            all_data <- .buffer_read_range(x, 1L, x$n)
            result <- all_data[i]
        }
        return(result)
    }

    # Multi-dimensional indexing
    if (length(x$dim) == 2) {
        # Matrix case
        nrow <- x$dim[1]
        ncol <- x$dim[2]

        if (missing(i)) i <- seq_len(nrow)
        if (missing(j)) j <- seq_len(ncol)

        i <- as.integer(i)
        j <- as.integer(j)

        # Convert to linear indices (column-major)
        linear_idx <- as.vector(outer(i, (j - 1L) * nrow, `+`))

        result <- .buffer_read_range(x, 1L, x$n)[linear_idx]

        if (!drop || (length(i) > 1 && length(j) > 1)) {
            dim(result) <- c(length(i), length(j))
        }
        return(result)
    }

    # General array case
    args <- c(list(i), if (!missing(j)) list(j), list(...))
    # Read all data and use standard array subsetting
    all_data <- .buffer_read_range(x, 1L, x$n)
    dim(all_data) <- x$dim
    do.call(`[`, c(list(all_data), args, list(drop = drop)))
}

#' Assign to Buffer Elements
#'
#' @param x A shard_buffer object.
#' @param i Index or indices.
#' @param j Optional second index (for matrices).
#' @param ... Additional indices (for arrays).
#' @param value Values to assign.
#' @return The modified \code{shard_buffer} object, invisibly.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 10)
#' buf[1:5] <- rnorm(5)
#' buffer_close(buf)
#' }
`[<-.shard_buffer` <- function(x, i, j, ..., value) {
    # Determine if this is matrix-style indexing by checking nargs()
    # nargs() > 3 means we have x[i, ...] <- value style (at least one comma)
    is_matrix_style <- nargs() > 3

    # For 1D buffers or linear indexing (no comma used)
    if (length(x$dim) == 1 || !is_matrix_style) {
        if (missing(i)) {
            i <- seq_len(x$n)
        }
        i <- as.integer(i)

        # Check bounds
        if (any(i < 1 | i > x$n, na.rm = TRUE)) {
            stop("Index out of bounds")
        }

        # Recycle value if needed
        if (length(value) != length(i)) {
            value <- rep_len(value, length(i))
        }

        # Optimize for contiguous ranges
        if (length(i) > 1 && all(diff(i) == 1L)) {
            .buffer_write_range(x, i[1], value)
        } else {
            # Non-contiguous: write indices directly to avoid read/modify/write.
            # This keeps disjoint parallel writes correct (lock-free) and avoids
            # clobbering other workers' updates.
            .buffer_write_indices(x, i, value)
        }
        return(invisible(x))
    }

    # Multi-dimensional indexing
    if (length(x$dim) == 2) {
      nrow <- x$dim[1]
      ncol <- x$dim[2]

      if (missing(i)) i <- seq_len(nrow)
      if (missing(j)) j <- seq_len(ncol)

      i <- as.integer(i)
      j <- as.integer(j)

      # Fast path for contiguous block assignment to avoid read/modify/write
      # of the entire buffer (critical for parallel disjoint writes).
      is_contig <- function(idx) length(idx) <= 1L || all(diff(idx) == 1L)

      if (length(i) > 0 && length(j) > 0 && is_contig(i) && is_contig(j) &&
          is.matrix(value) && identical(dim(value), c(length(i), length(j)))) {
        # If writing full rows for contiguous columns, write one contiguous span.
        if (identical(i, seq_len(nrow))) {
          start_lin <- (j[1] - 1L) * nrow + 1L
          .buffer_write_range(x, start_lin, as.vector(value))
          return(invisible(x))
        }

        # Otherwise write each column slice separately.
        for (k in seq_along(j)) {
          col <- j[k]
          start_lin <- (col - 1L) * nrow + i[1]
          .buffer_write_range(x, start_lin, value[, k])
        }
        return(invisible(x))
      }

      # Fallback: read all, modify, write back (slow; avoid in hot paths).
      all_data <- .buffer_read_range(x, 1L, x$n)
      dim(all_data) <- x$dim
      all_data[i, j] <- value
      dim(all_data) <- NULL
      .buffer_write_range(x, 1L, all_data)
      return(invisible(x))
    }

    # General array case
    all_data <- .buffer_read_range(x, 1L, x$n)
    dim(all_data) <- x$dim
    args <- c(list(i), if (!missing(j)) list(j), list(...))
    all_data <- do.call(`[<-`, c(list(all_data), args, list(value = value)))
    dim(all_data) <- NULL
    .buffer_write_range(x, 1L, all_data)
    invisible(x)
}

#' Coerce a Shared Memory Buffer to a Vector
#'
#' @param x A \code{shard_buffer} object.
#' @param mode Storage mode passed to \code{\link{as.vector}}.
#' @return A vector of the buffer's type (or coerced to \code{mode}).
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 5)
#' buf[1:5] <- 1:5
#' as.vector(buf)
#' buffer_close(buf)
#' }
as.vector.shard_buffer <- function(x, mode = "any") {
    result <- .buffer_read_range(x, 1L, x$n)
    if (mode != "any") {
        result <- as.vector(result, mode = mode)
    }
    result
}

#' Coerce a Shared Memory Buffer to Double
#'
#' @param x A \code{shard_buffer} object.
#' @param ... Ignored.
#' @return A double vector with the buffer contents.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 5)
#' as.double(buf)
#' buffer_close(buf)
#' }
as.double.shard_buffer <- function(x, ...) {
    as.double(.buffer_read_range(x, 1L, x$n))
}

#' Coerce a Shared Memory Buffer to Integer
#'
#' @param x A \code{shard_buffer} object.
#' @param ... Ignored.
#' @return An integer vector with the buffer contents.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("integer", dim = 5)
#' as.integer(buf)
#' buffer_close(buf)
#' }
as.integer.shard_buffer <- function(x, ...) {
    as.integer(.buffer_read_range(x, 1L, x$n))
}

#' Coerce a Shared Memory Buffer to Logical
#'
#' @param x A \code{shard_buffer} object.
#' @param ... Ignored.
#' @return A logical vector with the buffer contents.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("logical", dim = 5)
#' as.logical(buf)
#' buffer_close(buf)
#' }
as.logical.shard_buffer <- function(x, ...) {
    as.logical(.buffer_read_range(x, 1L, x$n))
}

#' Coerce a Shared Memory Buffer to Raw
#'
#' @param x A \code{shard_buffer} object.
#' @param ... Ignored.
#' @return A raw vector with the buffer contents.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("raw", dim = 5)
#' as.raw(buf)
#' buffer_close(buf)
#' }
as.raw.shard_buffer <- function(x, ...) {
    .buffer_read_range(x, 1L, x$n)
}

#' Coerce a Shared Memory Buffer to Matrix
#'
#' @param x A \code{shard_buffer} object (must be 2-dimensional).
#' @param ... Ignored.
#' @return A matrix with the buffer contents and the buffer's dimensions.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = c(3, 4))
#' as.matrix(buf)
#' buffer_close(buf)
#' }
as.matrix.shard_buffer <- function(x, ...) {
    if (length(x$dim) != 2) {
        stop("Buffer is not 2-dimensional")
    }
    result <- .buffer_read_range(x, 1L, x$n)
    dim(result) <- x$dim
    result
}

#' Coerce a Shared Memory Buffer to Array
#'
#' @param x A \code{shard_buffer} object.
#' @param ... Ignored.
#' @return An array with the buffer contents and the buffer's dimensions, or a
#'   plain vector for 1-D buffers.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = c(2, 3, 4))
#' as.array(buf)
#' buffer_close(buf)
#' }
as.array.shard_buffer <- function(x, ...) {
    result <- .buffer_read_range(x, 1L, x$n)
    if (length(x$dim) > 1) {
        dim(result) <- x$dim
    }
    result
}

# Tests for buffer() - typed writable output buffers

test_that("buffer creates double buffer", {
    buf <- buffer("double", dim = 100)
    expect_s3_class(buf, "shard_buffer")
    expect_equal(buf$type, "double")
    expect_equal(buf$n, 100)
    expect_equal(length(buf), 100)

    # Check initial values are zero
    expect_equal(buf[], rep(0, 100))

    buffer_close(buf)
})

test_that("buffer creates integer buffer", {
    buf <- buffer("integer", dim = 50)
    expect_equal(buf$type, "integer")
    expect_equal(buf[], rep(0L, 50))
    buffer_close(buf)
})

test_that("buffer creates logical buffer", {
    buf <- buffer("logical", dim = 25)
    expect_equal(buf$type, "logical")
    expect_equal(buf[], rep(FALSE, 25))
    buffer_close(buf)
})

test_that("buffer creates raw buffer", {
    buf <- buffer("raw", dim = 10)
    expect_equal(buf$type, "raw")
    expect_equal(buf[], raw(10))
    buffer_close(buf)
})

test_that("buffer with custom init value", {
    buf <- buffer("double", dim = 10, init = 42)
    expect_equal(buf[], rep(42, 10))
    buffer_close(buf)
})

test_that("buffer slice assignment works", {
    buf <- buffer("double", dim = 100)

    # Write to a slice
    buf[1:10] <- 1:10
    expect_equal(buf[1:10], as.double(1:10))
    expect_equal(buf[11], 0)

    # Write to non-contiguous indices
    buf[c(20, 30, 40)] <- c(200, 300, 400)
    expect_equal(buf[c(20, 30, 40)], c(200, 300, 400))

    buffer_close(buf)
})

test_that("buffer slice reading works", {
    buf <- buffer("integer", dim = 100)
    buf[1:100] <- 1:100

    expect_equal(buf[1:10], 1L:10L)
    expect_equal(buf[50], 50L)
    expect_equal(buf[c(1, 50, 100)], c(1L, 50L, 100L))

    buffer_close(buf)
})

test_that("buffer full extraction with [] works", {
    buf <- buffer("double", dim = 10)
    buf[1:5] <- 1:5
    buf[6:10] <- 6:10

    result <- buf[]
    expect_equal(result, as.double(1:10))

    buffer_close(buf)
})

test_that("buffer matrix operations work", {
    buf <- buffer("double", dim = c(10, 5))

    expect_equal(dim(buf), c(10, 5))
    expect_equal(length(buf), 50)

    # Write a column
    buf[, 1] <- 1:10
    expect_equal(buf[, 1], as.double(1:10))

    # Write a row
    buf[1, ] <- 101:105
    expect_equal(buf[1, ], as.double(101:105))

    # Write a submatrix
    buf[2:4, 2:3] <- matrix(1:6, nrow = 3)
    expect_equal(buf[2:4, 2:3], matrix(as.double(1:6), nrow = 3))

    buffer_close(buf)
})

test_that("buffer as.* conversions work", {
    buf <- buffer("double", dim = 10)
    buf[] <- 1:10

    expect_equal(as.vector(buf), as.double(1:10))
    expect_equal(as.double(buf), as.double(1:10))
    expect_equal(as.integer(buf), 1L:10L)

    buffer_close(buf)
})

test_that("buffer matrix conversion works", {
    buf <- buffer("double", dim = c(3, 4))
    buf[] <- 1:12

    mat <- as.matrix(buf)
    expect_true(is.matrix(mat))
    expect_equal(dim(mat), c(3, 4))
    expect_equal(mat[1, 1], 1)
    expect_equal(mat[3, 4], 12)

    buffer_close(buf)
})

test_that("buffer info returns correct information", {
    buf <- buffer("double", dim = c(100, 50))

    info <- buffer_info(buf)
    expect_equal(info$type, "double")
    expect_equal(info$dim, c(100L, 50L))
    expect_equal(info$n, 5000)
    expect_equal(info$bytes, 5000 * 8)

    buffer_close(buf)
})

test_that("buffer path returns path", {
    buf <- buffer("double", dim = 100)
    path <- buffer_path(buf)
    expect_true(!is.null(path))
    expect_true(is.character(path))
    buffer_close(buf)
})

test_that("buffer print works", {
    buf <- buffer("double", dim = c(100, 50))
    expect_output(print(buf), "shard_buffer")
    expect_output(print(buf), "double")
    expect_output(print(buf), "100 x 50")
    buffer_close(buf)
})

test_that("buffer rejects invalid dimensions", {
    expect_error(buffer("double", dim = 0))
    expect_error(buffer("double", dim = -1))
    expect_error(buffer("double", dim = c(10, 0)))
})

test_that("buffer rejects out of bounds indices", {
    buf <- buffer("double", dim = 10)
    expect_error(buf[0])
    expect_error(buf[11])
    expect_error(buf[1:11])
    expect_error(buf[-1] <- 1)
    buffer_close(buf)
})

test_that("buffer works with mmap backing", {
    buf <- buffer("double", dim = 100, backing = "mmap")
    buf[1:50] <- 1:50
    expect_equal(buf[1:50], as.double(1:50))
    buffer_close(buf)
})

test_that("buffer_open attaches to existing buffer", {
    # Create original buffer
    buf1 <- buffer("double", dim = 100)
    buf1[1:10] <- 1:10
    path <- buffer_path(buf1)
    info <- buffer_info(buf1)

    # Open from another "process" (same process but tests API)
    buf2 <- buffer_open(path, type = "double", dim = 100,
                        backing = info$backing)

    # Read what buf1 wrote
    expect_equal(buf2[1:10], as.double(1:10))

    # Write from buf2
    buf2[50:60] <- 50:60

    # Read from buf1
    expect_equal(buf1[50:60], as.double(50:60))

    buffer_close(buf1)
    buffer_close(buf2, unlink = FALSE)  # Don't unlink since buf1 owns it
})

test_that("buffer value recycling works", {
    buf <- buffer("double", dim = 10)

    # Single value recycled to fill indices
    buf[1:5] <- 42
    expect_equal(buf[1:5], rep(42, 5))

    buffer_close(buf)
})

test_that("buffer integer type preserves values", {
    buf <- buffer("integer", dim = 100)
    buf[1:10] <- c(-5L, -4L, -3L, -2L, -1L, 0L, 1L, 2L, 3L, 4L)
    expect_equal(buf[1:10], c(-5L, -4L, -3L, -2L, -1L, 0L, 1L, 2L, 3L, 4L))
    buffer_close(buf)
})

test_that("buffer logical type works", {
    buf <- buffer("logical", dim = 10)
    buf[c(1, 3, 5, 7, 9)] <- TRUE
    expect_equal(buf[c(1, 3, 5, 7, 9)], rep(TRUE, 5))
    expect_equal(buf[c(2, 4, 6, 8, 10)], rep(FALSE, 5))
    buffer_close(buf)
})

test_that("buffer handles single-element dimension", {
    # Single-element double buffer
    buf <- buffer("double", dim = 1)
    expect_equal(buf$n, 1)
    expect_equal(length(buf), 1)
    expect_equal(buf[], 0)

    buf[1] <- 42
    expect_equal(buf[1], 42)
    expect_equal(buf[], 42)
    buffer_close(buf)

    # Single-element integer buffer
    buf <- buffer("integer", dim = 1)
    buf[1] <- 99L
    expect_equal(buf[1], 99L)
    buffer_close(buf)

    # Single-element logical buffer
    buf <- buffer("logical", dim = 1)
    expect_equal(buf[1], FALSE)
    buf[1] <- TRUE
    expect_equal(buf[1], TRUE)
    buffer_close(buf)
})

test_that("buffer handles 1x1 matrix dimension", {
    buf <- buffer("double", dim = c(1, 1))
    expect_equal(dim(buf), c(1, 1))
    expect_equal(length(buf), 1)

    buf[1, 1] <- 42
    expect_equal(buf[1, 1], 42)
    expect_equal(buf[], matrix(42, 1, 1))

    buffer_close(buf)
})

test_that("tiled crossprod writes into an output buffer without view materialization", {
  skip_on_cran()

  pool_stop()
  shard:::view_reset_diagnostics()
  shard:::buffer_reset_diagnostics()

  set.seed(1)
  n <- 200L
  p <- 12L
  v <- 18L

  X <- matrix(rnorm(n * p), nrow = n)
  Y <- matrix(rnorm(n * v), nrow = n)

  colnames(X) <- paste0("x", seq_len(ncol(X)))
  colnames(Y) <- paste0("y", seq_len(ncol(Y)))

  Xsh <- share(X, backing = "mmap")
  Ysh <- share(Y, backing = "mmap")

  # Output is a shared buffer; each tile writes to a disjoint slice.
  Z <- buffer("double", dim = c(p, v), init = 0, backing = "mmap")

  tiles <- shard:::tiles2d(n_x = p, n_y = v, block_x = 4L, block_y = 6L)

  expect_true("crossprod_tile" %in% list_kernels())

  res <- shard_map(
    tiles,
    borrow = list(X = Xsh, Y = Ysh),
    out = list(Z = Z),
    kernel = "crossprod_tile",
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))

  cr <- copy_report(res)
  # This kernel writes tiles directly into the output buffer; it should not
  # materialize any view.
  expect_equal(cr$view_materialized %||% 0L, 0L)
  expect_equal(cr$view_materialized_bytes %||% 0, 0)

  # Each output element is written exactly once across tiles.
  expect_equal(cr$buffer_bytes %||% 0, as.double(p * v * 8L))

  # Result gathering stays tiny (NULLs only).
  expect_true(all(vapply(results(res), is.null, logical(1))))

  out <- as.matrix(Z)
  # Buffers do not currently preserve dimnames; compare values only.
  expect_equal(unname(out), unname(crossprod(X, Y)), tolerance = 1e-10)

  pool_stop()
})

test_that("tiled crossprod re-exports output buffers between calls", {
  skip_on_cran()

  pool_stop()

  set.seed(1)
  n <- 200L
  p <- 12L
  v <- 18L

  X <- matrix(rnorm(n * p), nrow = n)
  Y <- matrix(rnorm(n * v), nrow = n)

  Xsh <- share(X, backing = "mmap")
  Ysh <- share(Y, backing = "mmap")

  tiles <- shard:::tiles2d(n_x = p, n_y = v, block_x = 4L, block_y = 6L)
  expected <- unname(crossprod(X, Y))

  Z1 <- buffer("double", dim = c(p, v), init = 0, backing = "mmap")
  Z2 <- buffer("double", dim = c(p, v), init = 0, backing = "mmap")

  expect_false(identical(buffer_info(Z1)$path, buffer_info(Z2)$path))

  res1 <- shard_map(
    tiles,
    borrow = list(X = Xsh, Y = Ysh),
    out = list(Z = Z1),
    kernel = "crossprod_tile",
    workers = 2,
    diagnostics = FALSE
  )
  expect_true(succeeded(res1))
  expect_equal(unname(as.matrix(Z1)), expected, tolerance = 1e-10)

  # Critical: do NOT pool_stop() here. This ensures workers must switch to the
  # new output buffer descriptor rather than reusing cached file handles.
  res2 <- shard_map(
    tiles,
    borrow = list(X = Xsh, Y = Ysh),
    out = list(Z = Z2),
    kernel = "crossprod_tile",
    workers = 2,
    diagnostics = FALSE
  )
  expect_true(succeeded(res2))
  expect_equal(unname(as.matrix(Z2)), expected, tolerance = 1e-10)

  pool_stop()
})

#' @aliases shard-package
#' @section Core API:
#' * [shard_map()] - Primary parallel execution entry point
#' * [shards()] - Create shard descriptors with autotuning
#' * [results()] - Extract results from a shard_map run
#' * [succeeded()] - Check if shard_map completed without failures
#'
#' @section Zero-Copy Shared Data:
#' * [share()] - Share an R object for parallel access
#' * [fetch()] - Retrieve data from a shared object
#' * [materialize()] - Alias for fetch()
#' * [is_shared()] - Check if an object is shared
#' * [shared_info()] - Get information about a shared object
#'
#' @section Output Buffers:
#' * [buffer()] - Create typed writable output buffer
#' * [buffer_open()] - Open existing buffer from another process
#' * [buffer_path()] - Get buffer path for cross-process sharing
#' * [buffer_info()] - Get buffer information
#' * [buffer_close()] - Close and release buffer
#'
#' @section Worker Pool Management:
#' * [pool_create()] - Create a supervised worker pool
#' * [pool_stop()] - Stop the worker pool
#' * [pool_status()] - Check worker status and RSS
#' * [pool_health_check()] - Monitor and recycle workers
#'
#' @section Task Dispatch:
#' * [dispatch_chunks()] - Execute chunks with supervision
#' * [pool_lapply()] - Parallel lapply with supervision
#' * [pool_sapply()] - Parallel sapply with supervision
#'
#' @keywords internal
"_PACKAGE"

# On package load, ensure clean state
.onLoad <- function(libname, pkgname) {
  # Initialize pool environment
  .pool_env$pool <- NULL
  .pool_env$dev_path <- NULL

  # Register a finalizer that runs at R exit (onexit = TRUE).

  # .onUnload only fires on explicit unloadNamespace(), NOT on normal R exit.
  # Without this, parallel's PSOCK cluster finalizers fire first during
  # shutdown GC and fail with "Error in unserialize(node$con)" because
  # the worker connections are already being torn down.
  reg.finalizer(.pool_env, function(e) {
    tryCatch(pool_stop(), error = function(e2) NULL)
  }, onexit = TRUE)
}

# On package unload, stop any running pool
.onUnload <- function(libpath) {
  tryCatch(pool_stop(), error = function(e) NULL)
}

# Package load/unload hooks
# Currently empty - using string-based .Call with PACKAGE = "shard"

# Package-level environment for worker-side caching.
# Used instead of globalenv() to avoid R CMD check NOTEs.
.shard_worker_env <- new.env(parent = emptyenv())

#' @title Adapter Registry for Class-Specific Deep Sharing
#' @name adapter
#' @description Register custom traversal logic for specific classes during
#'   deep sharing operations. Adapters allow fine-grained control over how
#'   objects are decomposed and reconstructed.
#'
#' @details
#' The adapter registry provides a way to customize how specific classes are
#' handled during deep sharing. Instead of generic slot traversal for S4 objects
#' or element-wise traversal for lists, you can provide custom functions to:
#'
#' \enumerate{
#'   \item Extract the shareable children from an object (\code{children})
#'   \item Reconstruct the object from shared children (\code{replace})
#' }
#'
#' This is useful for:
#' \itemize{
#'   \item Complex S4 objects where only certain slots should be shared
#'   \item S3 objects with internal structure that differs from list structure
#'   \item Objects with accessors that should be used instead of direct slot access
#' }
#'
#' @seealso \code{\link{share}} for the main sharing function that uses adapters.
NULL

# Internal registry: environment storing adapters keyed by class name
.adapter_registry <- new.env(parent = emptyenv())

#' Validate Adapter Structure
#'
#' Internal function to validate an adapter has the required structure.
#'
#' @param adapter A list with adapter functions.
#' @param class_name The class name being registered.
#' @return TRUE if valid, throws error otherwise.
#' @keywords internal
#' @noRd
validate_adapter <- function(adapter, class_name) {
    if (!is.list(adapter)) {
        stop("Adapter must be a list.", call. = FALSE)
    }

    # Required: class
    if (is.null(adapter$class) || !is.character(adapter$class) ||
        length(adapter$class) != 1) {
        stop("Adapter must have a 'class' element (single character string).",
             call. = FALSE)
    }

    if (adapter$class != class_name) {
        stop("Adapter class '", adapter$class, "' does not match registered class '",
             class_name, "'.", call. = FALSE)
    }

    # Required: children function
    if (is.null(adapter$children) || !is.function(adapter$children)) {
        stop("Adapter must have a 'children' function: function(x) -> named list.",
             call. = FALSE)
    }

    # Required: replace function
    if (is.null(adapter$replace) || !is.function(adapter$replace)) {
        stop("Adapter must have a 'replace' function: function(x, children) -> object.",
             call. = FALSE)
    }

    # Optional: path_prefix (character or NULL)
    if (!is.null(adapter$path_prefix) &&
        (!is.character(adapter$path_prefix) || length(adapter$path_prefix) != 1)) {
        stop("Adapter 'path_prefix' must be a single character string or NULL.",
             call. = FALSE)
    }

    invisible(TRUE)
}


#' Register an Adapter for Class-Specific Traversal
#'
#' Registers a custom adapter for a specific class. When deep sharing encounters
#' an object of this class, it will use the adapter's \code{children()} function
#' to extract shareable components instead of generic traversal.
#'
#' @param class A character string naming the class to register the adapter for.
#' @param adapter A list containing:
#'   \describe{
#'     \item{class}{Character string matching the \code{class} parameter.}
#'     \item{children}{Function taking an object and returning a named list of
#'       child objects to traverse.}
#'     \item{replace}{Function taking the original object and a named list of
#'       (potentially shared) children, returning a reconstructed object.}
#'     \item{path_prefix}{Optional character string prefix for child paths in
#'       the sharing plan (default: class name).}
#'   }
#'
#' @return Invisibly returns the previous adapter for this class (if any),
#'   or NULL if no adapter was registered.
#'
#' @export
#' @examples
#' shard_list_adapters()
shard_register_adapter <- function(class, adapter) {
    stopifnot(is.character(class), length(class) == 1, nzchar(class))

    validate_adapter(adapter, class)

    # Store previous adapter for return
    previous <- if (exists(class, envir = .adapter_registry, inherits = FALSE)) {
        get(class, envir = .adapter_registry, inherits = FALSE)
    } else {
        NULL
    }

    # Register the adapter
    assign(class, adapter, envir = .adapter_registry)

    invisible(previous)
}


#' Unregister an Adapter
#'
#' Removes a previously registered adapter for a class. After unregistration,
#' objects of this class will use default traversal behavior during deep sharing.
#'
#' @param class A character string naming the class to unregister.
#'
#' @return Invisibly returns the removed adapter, or NULL if no adapter was
#'   registered for this class.
#'
#' @export
#' @examples
#' shard_list_adapters()
shard_unregister_adapter <- function(class) {
    stopifnot(is.character(class), length(class) == 1, nzchar(class))

    previous <- if (exists(class, envir = .adapter_registry, inherits = FALSE)) {
        adapter <- get(class, envir = .adapter_registry, inherits = FALSE)
        rm(list = class, envir = .adapter_registry)
        adapter
    } else {
        NULL
    }

    invisible(previous)
}


#' Get Adapter for an Object
#'
#' Retrieves the registered adapter for an object's class. Checks all classes
#' in the object's class hierarchy, returning the first matching adapter.
#'
#' @param x An R object.
#'
#' @return The adapter list if one is registered for any of the object's classes,
#'   or NULL if no adapter is registered.
#'
#' @export
#' @examples
#' shard_get_adapter(1:10)
shard_get_adapter <- function(x) {
    # Check each class in the class hierarchy
    classes <- class(x)

    for (cls in classes) {
        if (exists(cls, envir = .adapter_registry, inherits = FALSE)) {
            return(get(cls, envir = .adapter_registry, inherits = FALSE))
        }
    }

    NULL
}


#' List Registered Adapters
#'
#' Returns a character vector of all classes with registered adapters.
#'
#' @return Character vector of class names with registered adapters.
#'
#' @export
#' @examples
#' shard_list_adapters()
shard_list_adapters <- function() {
    ls(envir = .adapter_registry)
}


#' Clear All Registered Adapters
#'
#' Removes all registered adapters. Primarily useful for testing.
#'
#' @return Invisibly returns the number of adapters cleared.
#'
#' @keywords internal
#' @noRd
shard_clear_adapters <- function() {
    n <- length(ls(envir = .adapter_registry))
    rm(list = ls(envir = .adapter_registry), envir = .adapter_registry)
    invisible(n)
}

# Tests for adapter registry and class-specific traversal

test_that("shard_register_adapter registers adapter correctly", {
    # Clean up any existing adapters
    shard:::shard_clear_adapters()

    adapter <- list(
        class = "TestClass",
        children = function(x) list(data = x$data),
        replace = function(x, children) {
            x$data <- children$data
            x
        }
    )

    result <- shard_register_adapter("TestClass", adapter)

    # First registration returns NULL
    expect_null(result)

    # Check adapter is registered
    expect_true("TestClass" %in% shard_list_adapters())

    # Clean up
    shard:::shard_clear_adapters()
})

test_that("shard_register_adapter returns previous adapter on re-registration", {
    shard:::shard_clear_adapters()

    adapter1 <- list(
        class = "TestClass",
        children = function(x) list(a = 1),
        replace = function(x, children) x
    )

    adapter2 <- list(
        class = "TestClass",
        children = function(x) list(b = 2),
        replace = function(x, children) x
    )

    shard_register_adapter("TestClass", adapter1)
    previous <- shard_register_adapter("TestClass", adapter2)

    expect_identical(previous$class, "TestClass")
    expect_identical(previous$children(NULL), list(a = 1))

    shard:::shard_clear_adapters()
})

test_that("shard_unregister_adapter removes adapter and returns it", {
    shard:::shard_clear_adapters()

    adapter <- list(
        class = "TestClass",
        children = function(x) list(),
        replace = function(x, children) x
    )

    shard_register_adapter("TestClass", adapter)
    expect_true("TestClass" %in% shard_list_adapters())

    removed <- shard_unregister_adapter("TestClass")

    expect_identical(removed$class, "TestClass")
    expect_false("TestClass" %in% shard_list_adapters())

    shard:::shard_clear_adapters()
})

test_that("shard_unregister_adapter returns NULL for non-existent adapter", {
    shard:::shard_clear_adapters()

    result <- shard_unregister_adapter("NonExistentClass")
    expect_null(result)
})

test_that("shard_get_adapter returns adapter for object's class", {
    shard:::shard_clear_adapters()

    adapter <- list(
        class = "myclass",
        children = function(x) list(val = x$value),
        replace = function(x, children) {
            x$value <- children$val
            x
        }
    )

    shard_register_adapter("myclass", adapter)

    obj <- structure(list(value = 42), class = "myclass")
    found <- shard_get_adapter(obj)

    expect_identical(found$class, "myclass")

    shard:::shard_clear_adapters()
})

test_that("shard_get_adapter returns NULL for unregistered class", {
    shard:::shard_clear_adapters()

    obj <- structure(list(value = 42), class = "unregistered")
    found <- shard_get_adapter(obj)

    expect_null(found)
})

test_that("shard_get_adapter checks class hierarchy", {
    shard:::shard_clear_adapters()

    adapter <- list(
        class = "parentclass",
        children = function(x) list(val = x$value),
        replace = function(x, children) {
            x$value <- children$val
            x
        }
    )

    shard_register_adapter("parentclass", adapter)

    # Object with multiple classes - parentclass is second
    obj <- structure(list(value = 42), class = c("childclass", "parentclass"))
    found <- shard_get_adapter(obj)

    expect_identical(found$class, "parentclass")

    shard:::shard_clear_adapters()
})

test_that("adapter validation rejects invalid adapters", {
    shard:::shard_clear_adapters()

    # Missing class
    expect_error(
        shard_register_adapter("Test", list(
            children = function(x) list(),
            replace = function(x, ch) x
        )),
        "class"
    )

    # Class mismatch
    expect_error(
        shard_register_adapter("Test", list(
            class = "Other",
            children = function(x) list(),
            replace = function(x, ch) x
        )),
        "does not match"
    )

    # Missing children
    expect_error(
        shard_register_adapter("Test", list(
            class = "Test",
            replace = function(x, ch) x
        )),
        "children"
    )

    # Missing replace
    expect_error(
        shard_register_adapter("Test", list(
            class = "Test",
            children = function(x) list()
        )),
        "replace"
    )

    # Invalid path_prefix
    expect_error(
        shard_register_adapter("Test", list(
            class = "Test",
            children = function(x) list(),
            replace = function(x, ch) x,
            path_prefix = c("a", "b")  # Should be single string
        )),
        "path_prefix"
    )
})

test_that("deep share uses adapter's children function", {
    shard:::shard_clear_adapters()

    # Create a simple S3 class with internal structure
    create_myobj <- function(data, meta) {
        structure(list(data = data, meta = meta), class = "myobj")
    }

    # Register adapter that only exposes 'data'
    shard_register_adapter("myobj", list(
        class = "myobj",
        children = function(x) list(data = x$data),
        replace = function(x, children) {
            x$data <- children$data
            x
        }
    ))

    big_data <- rnorm(10000)
    obj <- create_myobj(data = big_data, meta = list(name = "test"))

    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    # Only the data should be shared (meta is not exposed by adapter)
    expect_equal(shared$summary$shared_count, 1)

    recovered <- fetch(shared)
    expect_equal(recovered$data, big_data)
    expect_equal(recovered$meta, list(name = "test"))  # meta preserved

    close(shared)
    shard:::shard_clear_adapters()
})

test_that("deep share uses adapter's replace function for reconstruction", {
    shard:::shard_clear_adapters()

    # Track calls to replace
    replace_called <- FALSE

    shard_register_adapter("tracked", list(
        class = "tracked",
        children = function(x) list(val = x$value),
        replace = function(x, children) {
            replace_called <<- TRUE
            x$value <- children$val
            x$reconstructed <- TRUE
            x
        }
    ))

    obj <- structure(list(value = rnorm(10000)), class = "tracked")
    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    recovered <- fetch(shared)

    expect_true(replace_called)
    expect_true(recovered$reconstructed)

    close(shared)
    shard:::shard_clear_adapters()
})

test_that("adapter with path_prefix uses custom path", {
    shard:::shard_clear_adapters()

    shard_register_adapter("prefixed", list(
        class = "prefixed",
        children = function(x) list(data = x$data),
        replace = function(x, children) {
            x$data <- children$data
            x
        },
        path_prefix = "/custom"
    ))

    obj <- structure(list(data = rnorm(10000)), class = "prefixed")
    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    # Check that the path includes the custom prefix
    expect_true(grepl("/custom", shared$tree$children$data$path))

    close(shared)
    shard:::shard_clear_adapters()
})

test_that("no adapter falls back to S4 slot traversal", {
    skip_if_not_installed("methods")
    shard:::shard_clear_adapters()

    # Define a simple S4 class
    setClass("TestS4", slots = c(mat = "matrix", info = "character"))

    big_mat <- matrix(rnorm(10000), nrow = 100)
    obj <- new("TestS4", mat = big_mat, info = "test")

    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")
    expect_equal(shared$summary$shared_count, 1)  # matrix should be shared

    recovered <- fetch(shared)
    expect_s4_class(recovered, "TestS4")
    expect_equal(recovered@mat, big_mat)
    expect_equal(recovered@info, "test")

    close(shared)
    removeClass("TestS4")
})

test_that("S4 adapter takes precedence over generic slot traversal", {
    skip_if_not_installed("methods")
    shard:::shard_clear_adapters()

    setClass("AdaptedS4", slots = c(data = "numeric", private = "numeric"))

    # Adapter only exposes 'data', not 'private'
    shard_register_adapter("AdaptedS4", list(
        class = "AdaptedS4",
        children = function(x) list(data = x@data),
        replace = function(x, children) {
            x@data <- children$data
            x
        }
    ))

    obj <- new("AdaptedS4", data = rnorm(10000), private = 1:100)
    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    # Only data should be traversed (via adapter), not private
    expect_equal(shared$summary$shared_count, 1)

    recovered <- fetch(shared)
    expect_equal(recovered@data, obj@data)
    expect_equal(recovered@private, 1:100)  # preserved unchanged

    close(shared)
    shard:::shard_clear_adapters()
    removeClass("AdaptedS4")
})

test_that("shard_list_adapters returns registered classes", {
    shard:::shard_clear_adapters()

    expect_length(shard_list_adapters(), 0)

    shard_register_adapter("ClassA", list(
        class = "ClassA",
        children = function(x) list(),
        replace = function(x, ch) x
    ))

    shard_register_adapter("ClassB", list(
        class = "ClassB",
        children = function(x) list(),
        replace = function(x, ch) x
    ))

    adapters <- shard_list_adapters()
    expect_length(adapters, 2)
    expect_true("ClassA" %in% adapters)
    expect_true("ClassB" %in% adapters)

    shard:::shard_clear_adapters()
})

test_that("shard_clear_adapters removes all adapters", {
    shard:::shard_clear_adapters()

    shard_register_adapter("A", list(class = "A", children = function(x) list(), replace = function(x, ch) x))
    shard_register_adapter("B", list(class = "B", children = function(x) list(), replace = function(x, ch) x))
    shard_register_adapter("C", list(class = "C", children = function(x) list(), replace = function(x, ch) x))

    expect_length(shard_list_adapters(), 3)

    n <- shard:::shard_clear_adapters()

    expect_equal(n, 3)
    expect_length(shard_list_adapters(), 0)
})

#' @title Ergonomic Apply/Lapply Wrappers
#' @description
#' Convenience wrappers that provide apply/lapply-style ergonomics while
#' preserving shard's core contract: shared immutable inputs, supervised
#' execution, and diagnostics.
#'
#' These functions are intentionally thin wrappers around [shard_map()] and
#' related primitives.
#' @name ergonomics
NULL

.shard_is_scalar <- function(x) {
  is.atomic(x) && length(x) == 1L && !is.null(x)
}

.shard_fetch_if_needed <- function(x) {
  if (inherits(x, c("shard_shared", "shard_deep_shared"))) {
    return(fetch(x))
  }
  x
}

.shard_strip_shared_attrs <- function(x) {
  # Some operations can accidentally propagate shard bookkeeping attributes
  # (e.g., when arithmetic uses a shared scalar as the longer operand).
  # For ergonomic wrappers, we return plain R objects by default.
  if (is.null(x)) return(x)
  for (nm in c("shard_cow", "shard_readonly")) {
    if (!is.null(attr(x, nm, exact = TRUE))) attr(x, nm) <- NULL
  }
  if (inherits(x, "shard_shared_vector")) {
    class(x) <- setdiff(class(x), "shard_shared_vector")
    if (length(class(x)) == 0) class(x) <- NULL
  }
  x
}

.shard_auto_share_one <- function(x, min_bytes, readonly = TRUE, backing = "auto") {
  # Keep existing shared objects.
  if (is_shared(x) || is_shared_vector(x)) return(x)

  # Only auto-share atomic vectors/matrices/arrays; otherwise keep as-is.
  # Deep-sharing is powerful but should remain an explicit choice.
  if (!is.atomic(x) || is.null(x)) return(x)

  sz <- as.numeric(utils::object.size(x))
  if (is.na(sz) || sz < min_bytes) return(x)

  share(x, backing = backing, readonly = readonly)
}

#' Apply Wrapper Policy
#'
#' Centralizes safe defaults and guardrails for apply/lapply convenience wrappers.
#'
#' @param auto_share_min_bytes Minimum object size for auto-sharing (default "1MB").
#' @param max_gather_bytes Maximum estimated gathered result bytes before refusing
#'   to run (default "256MB").
#' @param cow Copy-on-write policy for borrowed inputs. One of `"deny"`, `"audit"`,
#'   or `"allow"`. Default `"deny"`.
#' @param profile Execution profile passed through to [shard_map()]. One of
#'   `"default"`, `"memory"`, or `"speed"`. Default `"default"`.
#' @param block_size Shard block size for apply-style workloads. Default `"auto"`.
#' @param backing Backing type used when auto-sharing (`"auto"`, `"mmap"`, `"shm"`).
#' @return An object of class `shard_apply_policy`.
#' @export
#' @examples
#' cfg <- shard_apply_policy()
#' cfg
shard_apply_policy <- function(auto_share_min_bytes = "1MB",
                               max_gather_bytes = "256MB",
                               cow = c("deny", "audit", "allow"),
                               profile = c("default", "memory", "speed"),
                               block_size = "auto",
                               backing = c("auto", "mmap", "shm")) {
  cow <- match.arg(cow)
  profile <- match.arg(profile)
  backing <- match.arg(backing)

  structure(
    list(
      auto_share_min_bytes = parse_bytes(auto_share_min_bytes),
      max_gather_bytes = parse_bytes(max_gather_bytes),
      cow = cow,
      profile = profile,
      block_size = block_size,
      backing = backing
    ),
    class = "shard_apply_policy"
  )
}

#' Print a shard_apply_policy Object
#'
#' @param x A \code{shard_apply_policy} object.
#' @param ... Ignored.
#' @return The input \code{x}, invisibly.
#' @export
print.shard_apply_policy <- function(x, ...) {
  cat("shard_apply_policy\n")
  cat("  auto_share_min_bytes:", format_bytes(x$auto_share_min_bytes), "\n")
  cat("  max_gather_bytes:", format_bytes(x$max_gather_bytes), "\n")
  cat("  cow:", x$cow, "\n")
  cat("  profile:", x$profile, "\n")
  cat("  block_size:", as.character(x$block_size), "\n")
  cat("  backing:", x$backing, "\n")
  invisible(x)
}

#' Apply a Function Over Matrix Columns with Shared Inputs
#'
#' A convenience wrapper for the common "per-column apply" pattern. The matrix
#' is shared once and each worker receives a zero-copy column view when possible.
#'
#' Current limitation: `MARGIN` must be 2 (columns). Row-wise apply would require
#' strided/gather slicing and is intentionally explicit in shard via views/kernels.
#'
#' @param X A numeric/integer/logical matrix (or a shared matrix created by [share()]).
#' @param MARGIN Must be 2 (columns).
#' @param FUN Function of the form `function(v, ...)` returning a scalar atomic.
#' @param VARS Optional named list of extra variables. Large atomic VARS are
#'   auto-shared based on `policy$auto_share_min_bytes`.
#' @param workers Number of workers (passed to [shard_map()]).
#' @param ... Additional arguments forwarded to `FUN`.
#' @param policy A [shard_apply_policy()] object.
#' @return An atomic vector of length `ncol(X)` with the results.
#' @export
#' @examples
#' \donttest{
#' X <- matrix(rnorm(400), 20, 20)
#' shard_apply_matrix(X, MARGIN = 2, FUN = mean)
#' pool_stop()
#' }
shard_apply_matrix <- function(X,
                               MARGIN = 2,
                               FUN,
                               VARS = NULL,
                               workers = NULL,
                               ...,
                               policy = shard_apply_policy()) {
  if (!is.function(FUN)) stop("FUN must be a function", call. = FALSE)
  if (length(MARGIN) != 1 || is.na(MARGIN) || as.integer(MARGIN) != 2L) {
    stop("Only MARGIN=2 (columns) is supported; use shard_map()+views for other patterns.", call. = FALSE)
  }
  if (!is.matrix(X)) stop("X must be a matrix", call. = FALSE)

  # Ensure shared backing (ALTREP vector with dim) so workers can receive a
  # reopenable descriptor rather than a full copy.
  Xsh <- if (is_shared_vector(X) || is_shared(X)) X else share(X, readonly = TRUE, backing = policy$backing)
  if (!is_shared_vector(Xsh) && is_shared(Xsh)) {
    # share() for some objects yields a handle; fetch into a usable matrix
    # (still reopens in workers).
    Xsh <- fetch(Xsh)
  }
  if (!is.matrix(Xsh)) stop("Failed to obtain shared matrix representation", call. = FALSE)
  if (!is_shared_vector(Xsh)) {
    stop("Expected X to be represented as a shared vector-backed matrix; got: ", paste(class(Xsh), collapse = ", "), call. = FALSE)
  }

  vars <- VARS %||% list()
  if (!is.list(vars)) stop("VARS must be a named list (or NULL)", call. = FALSE)
  if (length(vars) > 0 && (is.null(names(vars)) || any(!nzchar(names(vars))))) {
    stop("VARS must be a named list (or NULL)", call. = FALSE)
  }
  extra_args <- list(...)

  # Auto-share large atomic vars to avoid duplicating them into each worker.
  if (length(vars) > 0) {
    for (nm in names(vars)) {
      vars[[nm]] <- .shard_auto_share_one(
        vars[[nm]],
        min_bytes = policy$auto_share_min_bytes,
        readonly = TRUE,
        backing = policy$backing
      )
    }
  }

  p <- ncol(Xsh)
  blocks <- shards(p, workers = workers, block_size = policy$block_size)
  out <- buffer("double", dim = p, backing = policy$backing, init = NA_real_)

  run <- shard_map(
    blocks,
    borrow = c(list(X = Xsh), vars),
    out = list(out = out),
    workers = workers,
    profile = policy$profile,
    cow = policy$cow,
    fun = function(shard, X, out, ...) {
      # Capture user args via parent env serialization.
      FUN <- get("FUN", inherits = TRUE)
      extra_args <- get("extra_args", inherits = TRUE)
      extra_vars <- list(...)

      # We cannot assume base slicing is copy-free; build explicit column views.
      nr <- dim(X)[1]
      for (j in shard$idx) {
        start <- (j - 1L) * nr + 1L
        v <- shared_view(X, start = start, length = nr)

        args <- c(list(v), extra_vars, extra_args)
        res <- do.call(FUN, args, quote = TRUE)
        if (!.shard_is_scalar(res)) {
          stop("FUN must return a scalar atomic value for shard_apply_matrix()", call. = FALSE)
        }
        out[j] <- as.double(res)
      }
      NULL
    }
  )

  # If failures exist, surface them.
  if (length(run$failures %||% list()) > 0) {
    fails <- run$failures %||% list()
    # Best-effort to surface the original worker error message for usability.
    first <- fails[[1]]
    msg <- NULL
    if (is.list(first)) {
      msg <- first$last_error %||% first$error %||% first$message %||% NULL
    }
    msg <- as.character(msg %||% "shard_apply_matrix() failed for some shards")
    stop(msg, call. = FALSE)
  }

  as.double(out[])
}

#' Apply a Function Over a List with Optional Auto-Sharing
#'
#' A convenience wrapper for list workloads that need supervision and shared
#' inputs. Large atomic list elements are auto-shared based on policy.
#'
#' This wrapper enforces guardrails to avoid accidental huge gathers: it
#' estimates the total gathered result size from a probe call and refuses
#' to run if it exceeds `policy$max_gather_bytes`.
#'
#' @param x A list.
#' @param FUN Function of the form `function(el, ...)`.
#' @param VARS Optional named list of extra variables (auto-shared when large).
#' @param workers Number of workers (passed to [shard_map()]).
#' @param ... Additional arguments forwarded to `FUN`.
#' @param policy A [shard_apply_policy()] object.
#' @return A list of results, one per element of \code{x}.
#' @export
#' @examples
#' \donttest{
#' res <- shard_lapply_shared(as.list(1:4), function(x) x^2)
#' pool_stop()
#' res
#' }
shard_lapply_shared <- function(x,
                                FUN,
                                VARS = NULL,
                                workers = NULL,
                                ...,
                                policy = shard_apply_policy()) {
  if (!is.list(x)) stop("x must be a list", call. = FALSE)
  if (!is.function(FUN)) stop("FUN must be a function", call. = FALSE)

  vars <- VARS %||% list()
  if (!is.list(vars)) stop("VARS must be a named list (or NULL)", call. = FALSE)
  if (length(vars) > 0 && (is.null(names(vars)) || any(!nzchar(names(vars))))) {
    stop("VARS must be a named list (or NULL)", call. = FALSE)
  }
  extra_args <- list(...)

  # Probe to estimate gathered result size.
  if (length(x) > 0) {
    probe <- do.call(FUN, c(list(x[[1]]), vars, extra_args), quote = TRUE)
    est <- as.numeric(utils::object.size(probe)) * length(x)
    if (is.finite(est) && est > policy$max_gather_bytes) {
      stop(
        "Refusing to gather an estimated ", format_bytes(est), " of results.\n",
        "  Hint: use shard_map()+buffer()/table_sink() or shard_reduce() for large outputs,\n",
        "  or raise policy$max_gather_bytes explicitly.",
        call. = FALSE
      )
    }
  }

  # Auto-share large atomic vars.
  if (length(vars) > 0) {
    for (nm in names(vars)) {
      vars[[nm]] <- .shard_auto_share_one(
        vars[[nm]],
        min_bytes = policy$auto_share_min_bytes,
        readonly = TRUE,
        backing = policy$backing
      )
    }
  }

  # Auto-share large atomic list elements; keep the list of handles small.
  shared_list <- x
  if (length(shared_list) > 0) {
    for (i in seq_along(shared_list)) {
      shared_list[[i]] <- .shard_auto_share_one(
        shared_list[[i]],
        min_bytes = policy$auto_share_min_bytes,
        readonly = TRUE,
        backing = policy$backing
      )
    }
  }

  n <- length(shared_list)
  if (n == 0) return(list())
  blocks <- shards(n, workers = workers, block_size = policy$block_size)

  # We gather results as a list; this wrapper is intentionally for moderate outputs.
  run <- shard_map(
    blocks,
    borrow = c(list(x = shared_list), vars),
    workers = workers,
    profile = policy$profile,
    cow = policy$cow,
    fun = function(shard, x, ...) {
      FUN <- get("FUN", inherits = TRUE)
      extra_args <- get("extra_args", inherits = TRUE)
      extra_vars <- list(...)

      out <- vector("list", length(shard$idx))
      for (k in seq_along(shard$idx)) {
        i <- shard$idx[[k]]
        el <- .shard_fetch_if_needed(x[[i]])
        out[[k]] <- do.call(FUN, c(list(el), extra_vars, extra_args), quote = TRUE)
      }
      names(out) <- as.character(shard$idx)
      out
    }
  )

  if (length(run$failures %||% list()) > 0) {
    stop("shard_lapply_shared() failed for some shards; see report(run) / task_report(run).", call. = FALSE)
  }

  # run$results is list of chunk results; each chunk result is a list of
  # per-shard return values.
  pieces <- run$results %||% list()
  res <- vector("list", n)
  for (chunk in pieces) {
    if (!is.list(chunk)) next
    for (shard_val in chunk) {
      if (!is.list(shard_val) || is.null(names(shard_val))) next
      for (nm in names(shard_val)) {
        idx <- as.integer(nm)
        if (!is.na(idx) && idx >= 1L && idx <= n) res[[idx]] <- shard_val[[nm]]
      }
    }
  }

  # Ergonomics: return plain R objects, not shard shared wrappers.
  for (i in seq_along(res)) {
    v <- res[[i]]
    if (is_shared_vector(v)) {
      res[[i]] <- .shard_strip_shared_attrs(materialize(v))
    } else if (inherits(v, c("shard_shared", "shard_deep_shared"))) {
      res[[i]] <- .shard_strip_shared_attrs(fetch(v))
    } else {
      res[[i]] <- .shard_strip_shared_attrs(v)
    }
  }

  res
}

test_that("shard_apply_matrix matches col-wise apply for scalar returns", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  set.seed(1)
  X <- matrix(rnorm(50), nrow = 10, ncol = 5)

  res <- shard_apply_matrix(
    X,
    MARGIN = 2,
    FUN = function(v) sum(v),
    workers = 2,
    policy = shard_apply_policy(backing = "mmap", block_size = 2)
  )

  expect_equal(res, as.numeric(colSums(X)))
})

test_that("shard_apply_matrix passes VARS through to FUN", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(1:12, nrow = 3, ncol = 4)
  res <- shard_apply_matrix(
    X,
    FUN = function(v, add) sum(v) + add,
    VARS = list(add = 10),
    workers = 2,
    policy = shard_apply_policy(backing = "mmap", block_size = 2)
  )

  expect_equal(res, as.numeric(colSums(X) + 10))
})

test_that("shard_apply_matrix works when X is already shared", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  set.seed(1)
  X <- matrix(rnorm(200), nrow = 20, ncol = 10)
  Xsh <- share(X, backing = "mmap", readonly = TRUE)

  res <- shard_apply_matrix(
    Xsh,
    FUN = function(v) mean(v),
    workers = 2,
    policy = shard_apply_policy(backing = "mmap", block_size = 3)
  )

  expect_equal(res, as.numeric(colMeans(X)))
})

test_that("shard_apply_matrix errors if FUN returns a non-scalar", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  X <- matrix(1:12, nrow = 3, ncol = 4)
  expect_error(
    suppressWarnings(
      shard_apply_matrix(
        X,
        FUN = function(v) v,
        workers = 2,
        policy = shard_apply_policy(backing = "mmap", block_size = 2)
      )
    ),
    "scalar atomic"
  )
})

test_that("shard_lapply_shared enforces max_gather_bytes guardrail", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  x <- as.list(1:4)
  pol <- shard_apply_policy(max_gather_bytes = "1MB", backing = "mmap", block_size = 2)

  expect_error(
    shard_lapply_shared(
      x,
      FUN = function(el) raw(1024 * 1024), # ~1MB per element => ~4MB gathered
      workers = 2,
      policy = pol
    ),
    "Refusing to gather"
  )
})

test_that("shard_lapply_shared behaves like lapply for small outputs", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  x <- list(1:3, 4:5, integer())
  res <- shard_lapply_shared(
    x,
    FUN = function(el) length(el),
    workers = 2,
    policy = shard_apply_policy(backing = "mmap", block_size = 2)
  )

  expect_equal(res, lapply(x, length))
})

test_that("shard_lapply_shared exercises auto-share path for list elements", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  x <- list(1:3, 4:5)
  pol <- shard_apply_policy(auto_share_min_bytes = 0, backing = "mmap", block_size = 2)

  res <- shard_lapply_shared(
    x,
    FUN = function(el) {
      length(el)
    },
    workers = 2,
    policy = pol
  )

  expect_equal(res, lapply(x, length))
})

test_that("shard_lapply_shared passes VARS through to FUN", {
  skip_on_cran()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  x <- list(1:3, 4:5, integer())
  res <- shard_lapply_shared(
    x,
    FUN = function(el, add) length(el) + add,
    VARS = list(add = 7),
    workers = 2,
    policy = shard_apply_policy(backing = "mmap", block_size = 2)
  )

  expect_equal(res, lapply(x, function(el) length(el) + 7))
})

test_that("shards() and shard_map() fall back to one worker when detectCores() returns NA", {
  skip_on_cran()
  skip_if_conn_exhausted()
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  testthat::local_mocked_bindings(
    detectCores = function(...) NA_integer_,
    .package = "parallel"
  )

  blocks <- shards(100, block_size = "auto", workers = NULL)
  expect_true(inherits(blocks, "shard_descriptor"))
  expect_true(blocks$num_shards >= 1L)

  res <- shard_map(10, fun = function(shard) sum(shard$idx), workers = NULL)
  expect_true(succeeded(res))
  expect_equal(pool_get()$n, 1L)
})

#' @title Chunk Queue Management
#' @description Queue management for dispatching chunks to workers with requeue support.
#' @name queue
NULL

#' Create a Chunk Queue
#'
#' Creates a queue for managing chunk dispatch with support for requeuing
#' failed chunks when workers die or are recycled.
#'
#' @param chunks List of chunk descriptors. Each chunk should have an `id` field.
#' @return A `shard_queue` object with queue management methods.
#' @keywords internal
#' @noRd
queue_create <- function(chunks) {
  # Ensure each chunk has an ID
  for (i in seq_along(chunks)) {
    if (is.null(chunks[[i]]$id)) {
      chunks[[i]]$id <- i
    }
  }

  env <- new.env(parent = emptyenv())
  env$pending <- chunks
  env$in_flight <- list()
  env$completed <- list()
  env$failed <- list()
  env$assignment <- list()  # chunk_id -> worker_id mapping
  env$order <- vapply(chunks, function(x) as.character(x$id), character(1))

  structure(
    list(
      env = env,
      total = length(chunks)
    ),
    class = "shard_queue"
  )
}

#' Get Next Chunk from Queue
#'
#' @param queue A `shard_queue` object.
#' @param worker_id Worker requesting the chunk.
#' @return A chunk descriptor or NULL if queue is empty.
#' @keywords internal
#' @noRd
queue_next <- function(queue, worker_id) {
  env <- queue$env

  if (length(env$pending) == 0) {
    return(NULL)
  }

  chunk <- env$pending[[1]]
  env$pending <- env$pending[-1]

  # Track assignment
  chunk_id <- as.character(chunk$id)
  env$in_flight[[chunk_id]] <- chunk
  env$assignment[[chunk_id]] <- worker_id

  chunk
}

queue_next_where <- function(queue, worker_id, predicate = NULL) {
  env <- queue$env
  if (length(env$pending) == 0) return(NULL)
  if (is.null(predicate)) return(queue_next(queue, worker_id))
  if (!is.function(predicate)) stop("predicate must be a function or NULL", call. = FALSE)

  idx <- NA_integer_
  for (i in seq_along(env$pending)) {
    ch <- env$pending[[i]]
    ok <- FALSE
    ok <- tryCatch(isTRUE(predicate(ch)), error = function(e) FALSE)
    if (ok) {
      idx <- i
      break
    }
  }
  if (is.na(idx)) return(NULL)

  chunk <- env$pending[[idx]]
  env$pending <- env$pending[-idx]

  chunk_id <- as.character(chunk$id)
  env$in_flight[[chunk_id]] <- chunk
  env$assignment[[chunk_id]] <- worker_id
  chunk
}

#' Mark Chunk as Completed
#'
#' @param queue A `shard_queue` object.
#' @param chunk_id Chunk identifier.
#' @param result The result from processing.
#' @return NULL (invisibly).
#' @keywords internal
#' @noRd
queue_complete <- function(queue, chunk_id, result = NULL, retain = TRUE) {
  env <- queue$env
  chunk_id <- as.character(chunk_id)

  if (!is.null(env$in_flight[[chunk_id]])) {
    chunk <- env$in_flight[[chunk_id]]
    if (isTRUE(retain)) {
      chunk$result <- result
      chunk$completed_at <- Sys.time()
      env$completed[[chunk_id]] <- chunk
    } else {
      # Keep only minimal completion metadata to avoid retaining large shard lists
      # when callers are doing streaming reductions.
      env$completed[[chunk_id]] <- list(
        id = chunk$id,
        completed_at = Sys.time(),
        retry_count = chunk$retry_count %||% 0L,
        result = result
      )
    }
    env$in_flight[[chunk_id]] <- NULL
    env$assignment[[chunk_id]] <- NULL
  }

  invisible(NULL)
}

#' Mark Chunk as Failed
#'
#' @param queue A `shard_queue` object.
#' @param chunk_id Chunk identifier.
#' @param error The error that occurred.
#' @param requeue Logical. Whether to requeue for retry.
#' @return NULL (invisibly).
#' @keywords internal
#' @noRd
queue_fail <- function(queue, chunk_id, error = NULL, requeue = TRUE) {
  env <- queue$env
  chunk_id <- as.character(chunk_id)

  chunk <- env$in_flight[[chunk_id]]
  if (is.null(chunk)) {
    return(invisible(NULL))
  }

  env$in_flight[[chunk_id]] <- NULL
  env$assignment[[chunk_id]] <- NULL

  # Track retry count
  chunk$retry_count <- (chunk$retry_count %||% 0L) + 1L
  chunk$last_error <- error

  if (requeue) {
    # Put back in pending queue for retry
    env$pending <- c(env$pending, list(chunk))
  } else {
    # Mark as permanently failed
    chunk$failed_at <- Sys.time()
    env$failed[[chunk_id]] <- chunk
  }

  invisible(NULL)
}

#' Requeue All Chunks for a Worker
#'
#' When a worker dies or is recycled, requeue all its in-flight chunks.
#'
#' @param queue A `shard_queue` object.
#' @param worker_id Worker identifier.
#' @return Integer. Number of chunks requeued.
#' @keywords internal
#' @noRd
queue_requeue_worker <- function(queue, worker_id) {
  env <- queue$env
  requeued <- 0L

  chunk_ids <- names(env$assignment)
  for (chunk_id in chunk_ids) {
    if (identical(env$assignment[[chunk_id]], worker_id)) {
      queue_fail(queue, chunk_id, error = "worker_recycled", requeue = TRUE)
      requeued <- requeued + 1L
    }
  }

  requeued
}

#' Check if Queue is Done
#'
#' @param queue A `shard_queue` object.
#' @return Logical. TRUE if all chunks are completed or failed.
#' @keywords internal
#' @noRd
queue_is_done <- function(queue) {
  env <- queue$env
  length(env$pending) == 0 && length(env$in_flight) == 0
}

#' Check if Queue Has Work
#'
#' @param queue A `shard_queue` object.
#' @return Logical. TRUE if there are pending chunks.
#' @keywords internal
#' @noRd
queue_has_pending <- function(queue) {
  length(queue$env$pending) > 0
}

#' Get Queue Status
#'
#' @param queue A `shard_queue` object.
#' @return A list with queue statistics.
#' @keywords internal
#' @noRd
queue_status <- function(queue) {
  env <- queue$env

  # Count retries
  total_retries <- sum(vapply(
    c(env$completed, env$pending, env$in_flight, env$failed),
    function(x) x$retry_count %||% 0L,
    integer(1)
  ))

  list(
    total = queue$total,
    pending = length(env$pending),
    in_flight = length(env$in_flight),
    completed = length(env$completed),
    failed = length(env$failed),
    total_retries = total_retries
  )
}

#' Get Queue Results
#'
#' @param queue A `shard_queue` object.
#' @return List of results from completed chunks.
#' @keywords internal
#' @noRd
queue_results <- function(queue) {
  env <- queue$env
  completed <- env$completed
  ids <- names(completed)
  if (length(ids) == 0) return(list())

  # Prefer numeric id ordering for predictable, lapply/sapply-like behavior.
  num_ids <- suppressWarnings(as.integer(ids))
  if (!anyNA(num_ids)) {
    completed <- completed[order(num_ids)]
  } else if (!is.null(env$order)) {
    ord_ids <- intersect(env$order, ids)
    completed <- completed[ord_ids]
  }

  lapply(completed, function(x) x$result)
}

#' Get Failed Chunks
#'
#' @param queue A `shard_queue` object.
#' @return List of failed chunk descriptors.
#' @keywords internal
#' @noRd
queue_failures <- function(queue) {
  queue$env$failed
}

#' @param x A \code{shard_queue} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @export
#' @noRd
print.shard_queue <- function(x, ...) {
  status <- queue_status(x)
  cat("shard chunk queue\n")
  cat("  Total chunks:", status$total, "\n")
  cat("  Pending:", status$pending, "\n")
  cat("  In flight:", status$in_flight, "\n")
  cat("  Completed:", status$completed, "\n")
  cat("  Failed:", status$failed, "\n")
  cat("  Total retries:", status$total_retries, "\n")
  invisible(x)
}

test_that("queue_create initializes correctly", {
  chunks <- list(list(data = 1), list(data = 2), list(data = 3))
  q <- queue_create(chunks)

  expect_s3_class(q, "shard_queue")
  expect_equal(q$total, 3L)

  status <- queue_status(q)
  expect_equal(status$total, 3L)
  expect_equal(status$pending, 3L)
  expect_equal(status$in_flight, 0L)
  expect_equal(status$completed, 0L)
})

test_that("queue_next returns chunks and tracks in-flight", {
  chunks <- list(list(data = "a"), list(data = "b"))
  q <- queue_create(chunks)

  chunk1 <- queue_next(q, worker_id = 1)
  expect_equal(chunk1$data, "a")

  status <- queue_status(q)
  expect_equal(status$pending, 1L)
  expect_equal(status$in_flight, 1L)

  chunk2 <- queue_next(q, worker_id = 2)
  expect_equal(chunk2$data, "b")

  status <- queue_status(q)
  expect_equal(status$pending, 0L)
  expect_equal(status$in_flight, 2L)

  # No more chunks
  chunk3 <- queue_next(q, worker_id = 1)
  expect_null(chunk3)
})

test_that("queue_complete marks chunks done", {
  chunks <- list(list(data = 1))
  q <- queue_create(chunks)

  chunk <- queue_next(q, worker_id = 1)
  queue_complete(q, chunk$id, result = "done")

  status <- queue_status(q)
  expect_equal(status$completed, 1L)
  expect_equal(status$in_flight, 0L)

  results <- queue_results(q)
  expect_equal(results[[1]], "done")
})

test_that("queue_fail with requeue puts chunk back", {
  chunks <- list(list(data = 1))
  q <- queue_create(chunks)

  chunk <- queue_next(q, worker_id = 1)
  queue_fail(q, chunk$id, error = "oops", requeue = TRUE)

  status <- queue_status(q)
  expect_equal(status$pending, 1L)
  expect_equal(status$in_flight, 0L)
  expect_equal(status$total_retries, 1L)

  # Can get the chunk again
  chunk2 <- queue_next(q, worker_id = 2)
  expect_equal(chunk2$data, 1)
  expect_equal(chunk2$retry_count, 1L)
})

test_that("queue_fail without requeue marks permanent failure", {
  chunks <- list(list(data = 1))
  q <- queue_create(chunks)

  chunk <- queue_next(q, worker_id = 1)
  queue_fail(q, chunk$id, error = "fatal", requeue = FALSE)

  status <- queue_status(q)
  expect_equal(status$failed, 1L)
  expect_equal(status$pending, 0L)

  failures <- queue_failures(q)
  expect_length(failures, 1L)
})

test_that("queue_requeue_worker requeues worker's chunks", {
  chunks <- list(list(data = 1), list(data = 2), list(data = 3))
  q <- queue_create(chunks)

  # Assign chunks to workers
  queue_next(q, worker_id = 1)  # chunk 1 -> worker 1
  queue_next(q, worker_id = 2)  # chunk 2 -> worker 2
  queue_next(q, worker_id = 1)  # chunk 3 -> worker 1

  # Requeue all chunks from worker 1
  requeued <- queue_requeue_worker(q, worker_id = 1)
  expect_equal(requeued, 2L)

  status <- queue_status(q)
  expect_equal(status$pending, 2L)  # chunks 1 and 3 requeued
  expect_equal(status$in_flight, 1L)  # chunk 2 still in flight
})

test_that("queue_is_done reflects completion state", {
  chunks <- list(list(data = 1))
  q <- queue_create(chunks)

  expect_false(queue_is_done(q))

  chunk <- queue_next(q, worker_id = 1)
  expect_false(queue_is_done(q))  # Still in flight

  queue_complete(q, chunk$id)
  expect_true(queue_is_done(q))
})

test_that("print.shard_queue produces output", {
  chunks <- list(list(data = 1), list(data = 2))
  q <- queue_create(chunks)

  output <- capture.output(print(q))
  expect_true(any(grepl("shard chunk queue", output)))
  expect_true(any(grepl("Total chunks: 2", output)))
})

# High-level "low ceremony" helpers for common shard-aware kernels.

#' Parallel crossprod() using shard views + output buffers
#'
#' Computes `crossprod(X, Y)` (i.e. `t(X) %*% Y`) using:
#' - shared/mmap-backed inputs (one copy),
#' - block views (no slice materialization),
#' - BLAS-3 dgemm in each tile,
#' - an explicit shared output buffer (no gather/bind spikes).
#'
#' This is intended as an ergonomic entry point for the "wow" path: users
#' shouldn't have to manually call `share()`, `view_block()`, `buffer()`,
#' `tiles2d()`, and `shard_map()` for common patterns.
#'
#' @param X,Y Double matrices with the same number of rows.
#' @param workers Number of worker processes.
#' @param block_x,block_y Tile sizes over `ncol(X)` and `ncol(Y)`. Use `"auto"`
#'   (default) to autotune on the current machine.
#' @param backing Backing for shared inputs and output buffer (`"mmap"` or `"shm"`).
#' @param materialize Whether to return the result as a standard R matrix:
#'   `"never"` (return buffer handle), `"always"`, or `"auto"` (materialize if
#'   estimated output size is below `materialize_max_bytes`).
#' @param materialize_max_bytes Threshold for `"auto"` materialization.
#' @param diagnostics Whether to collect shard_map diagnostics.
#'
#' @return A list with:
#'   - `buffer`: shard_buffer for the result (p x v)
#'   - `value`: materialized matrix if requested, otherwise NULL
#'   - `run`: the underlying shard_result from shard_map
#'   - `tile`: chosen tile sizes
#' @export
#' @examples
#' \donttest{
#' X <- matrix(rnorm(2000), 100, 20)
#' Y <- matrix(rnorm(2000), 100, 20)
#' res <- shard_crossprod(X, Y, block_x = 50, block_y = 10, workers = 2)
#' pool_stop()
#' res$value
#' }
shard_crossprod <- function(X,
                            Y,
                            workers = NULL,
                            block_x = "auto",
                            block_y = "auto",
                            backing = c("mmap", "shm"),
                            materialize = c("auto", "never", "always"),
                            materialize_max_bytes = 512 * 1024^2,
                            diagnostics = TRUE) {
  backing <- match.arg(backing)
  materialize <- match.arg(materialize)

  if (!is.matrix(X) || !is.matrix(Y)) stop("X and Y must be matrices", call. = FALSE)
  if (!identical(typeof(X), "double") || !identical(typeof(Y), "double")) {
    stop("shard_crossprod currently supports double matrices only", call. = FALSE)
  }
  if (nrow(X) != nrow(Y)) stop("X and Y must have the same number of rows", call. = FALSE)

  Xsh <- if (is_shared_vector(X)) X else share(X, backing = backing)
  Ysh <- if (is_shared_vector(Y)) Y else share(Y, backing = backing)

  p <- ncol(Xsh)
  v <- ncol(Ysh)

  if (identical(block_x, "auto") || identical(block_y, "auto")) {
    tuned <- .autotune_crossprod_tiles(Xsh, Ysh,
                                      block_x = block_x, block_y = block_y)
    block_x <- tuned$block_x
    block_y <- tuned$block_y
  } else {
    block_x <- as.integer(block_x)
    block_y <- as.integer(block_y)
  }

  if (is.na(block_x) || block_x < 1L) stop("block_x must be >= 1", call. = FALSE)
  if (is.na(block_y) || block_y < 1L) stop("block_y must be >= 1", call. = FALSE)
  block_x <- min(block_x, p)
  block_y <- min(block_y, v)

  Z <- buffer("double", dim = c(p, v), init = 0, backing = backing)
  tiles <- tiles2d(n_x = p, n_y = v, block_x = block_x, block_y = block_y)

  run <- shard_map(
    tiles,
    borrow = list(X = Xsh, Y = Ysh),
    out = list(Z = Z),
    kernel = "crossprod_tile",
    workers = workers,
    diagnostics = diagnostics
  )

  out_bytes <- as.double(p) * as.double(v) * 8
  do_mat <- switch(materialize,
    "never" = FALSE,
    "always" = TRUE,
    "auto" = isTRUE(out_bytes <= materialize_max_bytes)
  )

  list(
    buffer = Z,
    value = if (do_mat) as.matrix(Z) else NULL,
    run = run,
    tile = c(block_x = block_x, block_y = block_y)
  )
}

.autotune_crossprod_tiles <- function(Xsh, Ysh, block_x = "auto", block_y = "auto") {
  p <- ncol(Xsh)
  v <- ncol(Ysh)

  # Candidate sizes are chosen to hit common BLAS sweet spots without creating
  # huge temporary matrices (tile result is kx x ky).
  cand_x <- c(8L, 16L, 32L, 64L)
  cand_y <- c(16L, 32L, 64L, 128L)
  cand_x <- cand_x[cand_x <= p]
  cand_y <- cand_y[cand_y <= v]

  if (!identical(block_x, "auto")) cand_x <- as.integer(block_x)
  if (!identical(block_y, "auto")) cand_y <- as.integer(block_y)

  # For small problems, just take the max that fits.
  if (p <= 64L && v <= 128L) {
    return(list(block_x = min(max(cand_x), p), block_y = min(max(cand_y), v)))
  }

  best <- list(score = -Inf, block_x = min(max(cand_x), p), block_y = min(max(cand_y), v))
  trials <- 3L

  for (bx in cand_x) {
    for (by in cand_y) {
      if (bx > p || by > v) next

      times <- numeric(trials)
      for (t in seq_len(trials)) {
        xs <- if (p == bx) 1L else sample.int(p - bx + 1L, 1L)
        ys <- if (v == by) 1L else sample.int(v - by + 1L, 1L)

        vX <- view_block(Xsh, cols = idx_range(xs, xs + bx - 1L))
        vY <- view_block(Ysh, cols = idx_range(ys, ys + by - 1L))

        t0 <- proc.time()[["elapsed"]]
        view_crossprod(vX, vY)
        times[t] <- proc.time()[["elapsed"]] - t0
      }

      med <- stats::median(times)
      if (!is.finite(med) || med <= 0) next

      # Throughput proxy: output elements per second.
      score <- (as.double(bx) * as.double(by)) / med
      if (score > best$score) best <- list(score = score, block_x = bx, block_y = by)
    }
  }

  list(block_x = best$block_x, block_y = best$block_y)
}

# Shared-memory task queue (opt-in).
#
# This is a "fast dispatch" mode intended for tiny tasks where per-task socket
# overhead dominates. It is conservative and intentionally limited: tasks are
# identified by integer ids and results are not gathered.

taskq_create <- function(n_tasks,
                         backing = c("auto", "mmap", "shm")) {
  backing <- match.arg(backing)
  n_tasks <- as.integer(n_tasks)
  if (is.na(n_tasks) || n_tasks < 1L) stop("n_tasks must be >= 1", call. = FALSE)

  # Header + per-task records; size computed in C as well.
  # Allocate with a small safety margin.
  # Over-allocate: C validates and will error if too small.
  bytes_per_task <- 64L
  header_bytes <- 128L
  size <- as.double(header_bytes + as.double(n_tasks) * bytes_per_task)

  seg <- segment_create(size, backing = backing)
  .Call("C_shard_taskq_init", seg$ptr, n_tasks, PACKAGE = "shard")

  desc <- list(
    path = segment_path(seg),
    backing = seg$backing,
    size = segment_size(seg),
    n_tasks = n_tasks
  )
  # Keep the owning segment alive in the creating process; do not attempt to
  # serialize the xptr to workers.
  list(desc = desc, owner = seg)
}

taskq_supported <- function() {
  isTRUE(.Call("C_shard_taskq_supported", PACKAGE = "shard"))
}

taskq_open <- function(desc, readonly = FALSE) {
  if (!is.list(desc) || is.null(desc$path) || is.null(desc$backing)) {
    stop("desc must be a task queue descriptor", call. = FALSE)
  }
  segment_open(desc$path, backing = desc$backing, readonly = readonly)
}

taskq_stats <- function(seg) {
  stopifnot(inherits(seg, "shard_segment"))
  .Call("C_shard_taskq_stats", seg$ptr, PACKAGE = "shard")
}

taskq_claim <- function(seg, worker_id) {
  stopifnot(inherits(seg, "shard_segment"))
  .Call("C_shard_taskq_claim", seg$ptr, as.integer(worker_id), PACKAGE = "shard")
}

taskq_done <- function(seg, task_id) {
  stopifnot(inherits(seg, "shard_segment"))
  .Call("C_shard_taskq_mark_done", seg$ptr, as.integer(task_id), PACKAGE = "shard")
}

taskq_error <- function(seg, task_id, max_retries) {
  stopifnot(inherits(seg, "shard_segment"))
  .Call("C_shard_taskq_mark_error", seg$ptr, as.integer(task_id), as.integer(max_retries), PACKAGE = "shard")
}

taskq_reset_claims <- function(seg, worker_id) {
  stopifnot(inherits(seg, "shard_segment"))
  .Call("C_shard_taskq_reset_claims", seg$ptr, as.integer(worker_id), PACKAGE = "shard")
}

taskq_failures <- function(seg) {
  stopifnot(inherits(seg, "shard_segment"))
  .Call("C_shard_taskq_failures", seg$ptr, PACKAGE = "shard")
}

test_that("taskq_supported returns logical", {
  result <- shard:::taskq_supported()
  expect_true(is.logical(result))
  expect_length(result, 1)
})

test_that("taskq_create creates valid task queue", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(10)

  expect_true(is.list(q))
  expect_true("desc" %in% names(q))
  expect_true("owner" %in% names(q))

  desc <- q$desc
  expect_equal(desc$n_tasks, 10L)
  expect_true(!is.null(desc$path))
  expect_true(!is.null(desc$backing))
  expect_true(!is.null(desc$size))
})

test_that("taskq_create validates n_tasks", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  expect_error(shard:::taskq_create(0), "n_tasks must be >= 1")
  expect_error(shard:::taskq_create(-1), "n_tasks must be >= 1")
})

test_that("taskq_open opens existing queue", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(10, backing = "mmap")
  seg <- shard:::taskq_open(q$desc, readonly = FALSE)

  expect_s3_class(seg, "shard_segment")
})

test_that("taskq_open validates descriptor", {
  expect_error(shard:::taskq_open(NULL), "desc must be a task queue descriptor")
  expect_error(shard:::taskq_open(list()), "desc must be a task queue descriptor")
  expect_error(shard:::taskq_open(list(path = NULL, backing = "mmap")),
               "desc must be a task queue descriptor")
})

test_that("taskq_stats returns queue statistics", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(10, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  stats <- shard:::taskq_stats(seg)

  expect_true(is.list(stats))
  expect_true("n_tasks" %in% names(stats))
  expect_true("done" %in% names(stats))
  expect_true("failed" %in% names(stats))
  expect_true("retries" %in% names(stats))

  # Initially no tasks done or failed
  expect_equal(stats$n_tasks, 10L)
  expect_equal(stats$done, 0L)
  expect_equal(stats$failed, 0L)
})

test_that("taskq_claim claims a pending task", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(5, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  # Worker 1 claims a task
  task_id <- shard:::taskq_claim(seg, worker_id = 1L)

  # Should get a valid task id (1-5)
  expect_true(task_id >= 1L && task_id <= 5L)
})

test_that("taskq_claim returns 0 when no tasks available", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(2, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  # Claim both tasks
  t1 <- shard:::taskq_claim(seg, worker_id = 1L)
  t2 <- shard:::taskq_claim(seg, worker_id = 1L)

  # Mark both as done
  shard:::taskq_done(seg, t1)
  shard:::taskq_done(seg, t2)

  # No more tasks available - returns 0
  t3 <- shard:::taskq_claim(seg, worker_id = 1L)
  expect_equal(t3, 0L)
})

test_that("taskq_done marks task as completed", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(3, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  task_id <- shard:::taskq_claim(seg, worker_id = 1L)
  shard:::taskq_done(seg, task_id)

  stats <- shard:::taskq_stats(seg)
  expect_equal(stats$done, 1L)
})

test_that("taskq_error handles retry logic", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(3, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  task_id <- shard:::taskq_claim(seg, worker_id = 1L)

  # First error - returns logical indicating if task was requeued
  result <- shard:::taskq_error(seg, task_id, max_retries = 2L)

  expect_true(is.logical(result))
})

test_that("taskq_failures returns failed task info", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(3, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  # Claim and fail a task with no retries
  task_id <- shard:::taskq_claim(seg, worker_id = 1L)
  shard:::taskq_error(seg, task_id, max_retries = 0L)

  failures <- shard:::taskq_failures(seg)

  # Returns list structure
  expect_true(is.list(failures))
  expect_true("task_id" %in% names(failures))
  expect_true("retry_count" %in% names(failures))
})

test_that("taskq_reset_claims works without error", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(5, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  # Worker 1 claims tasks but "dies" before completing
  t1 <- shard:::taskq_claim(seg, worker_id = 1L)
  t2 <- shard:::taskq_claim(seg, worker_id = 1L)

  # Reset claims from worker 1 - should not error
  expect_no_error(shard:::taskq_reset_claims(seg, worker_id = 1L))
})

test_that("taskq with mmap backing", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  # mmap backing
  q_mmap <- shard:::taskq_create(5, backing = "mmap")
  expect_equal(q_mmap$desc$backing, "mmap")
})

test_that("taskq handles many tasks", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  n <- 1000L
  q <- shard:::taskq_create(n, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  stats <- shard:::taskq_stats(seg)
  expect_equal(stats$n_tasks, n)
})

test_that("taskq multiple workers can claim tasks", {
  if (!shard:::taskq_supported()) skip("taskq not supported on this platform")

  q <- shard:::taskq_create(10, backing = "mmap")
  seg <- shard:::taskq_open(q$desc)

  # Different workers claim tasks
  t1 <- shard:::taskq_claim(seg, worker_id = 1L)
  t2 <- shard:::taskq_claim(seg, worker_id = 2L)
  t3 <- shard:::taskq_claim(seg, worker_id = 1L)

  # All should be valid task ids
  expect_true(all(c(t1, t2, t3) >= 1L))

  # Mark all done
  shard:::taskq_done(seg, t1)
  shard:::taskq_done(seg, t2)
  shard:::taskq_done(seg, t3)

  stats <- shard:::taskq_stats(seg)
  expect_equal(stats$done, 3L)
})

#' @title Diagnostics API
#' @name diagnostics
#' @description Comprehensive diagnostics for shard parallel execution, providing
#'   insights into memory usage, worker status, task execution, and shared memory
#'   segments.
#'
#' @details
#' The diagnostics API provides multiple views into shard's runtime behavior:
#' \itemize{
#'   \item \code{report()}: Primary entry point with configurable detail levels
#'   \item \code{mem_report()}: Memory usage across workers
#'   \item \code{cow_report()}: Copy-on-write policy tracking
#'   \item \code{copy_report()}: Data transfer statistics
#'   \item \code{task_report()}: Task/chunk execution statistics
#'   \item \code{segment_report()}: Shared memory segment information
#' }
#'
#' All functions return S3 \code{shard_report} objects with appropriate print
#' methods for human-readable output.
NULL

#' Generate Shard Runtime Report
#'
#' Primary entry point for shard diagnostics. Generates a comprehensive report
#' of the current runtime state including pool status, memory usage, and
#' execution statistics.
#'
#' @param level Character. Detail level for the report:
#'   \itemize{
#'     \item \code{"summary"}: High-level overview (default)
#'     \item \code{"workers"}: Include per-worker details
#'     \item \code{"tasks"}: Include task execution history
#'     \item \code{"segments"}: Include shared memory segment details
#'   }
#' @param result Optional. A \code{shard_result} object from \code{\link{shard_map}}
#'   to include execution diagnostics from.
#'
#' @return An S3 object of class \code{shard_report} containing:
#'   \itemize{
#'     \item \code{level}: The requested detail level
#'     \item \code{timestamp}: When the report was generated
#'     \item \code{pool}: Pool status information (if pool exists)
#'     \item \code{memory}: Memory usage summary
#'     \item \code{workers}: Per-worker details (if level includes workers)
#'     \item \code{tasks}: Task execution details (if level includes tasks)
#'     \item \code{segments}: Segment details (if level includes segments)
#'     \item \code{result_diagnostics}: Diagnostics from shard_result (if provided)
#'   }
#'
#' @export
#' @examples
#' \donttest{
#' res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
#' pool_stop()
#' report(result = res)
#' }
report <- function(level = c("summary", "workers", "tasks", "segments"),
                   result = NULL) {
  level <- match.arg(level)

  # Build report based on level
  rpt <- list(
    level = level,
    timestamp = Sys.time(),
    pool = NULL,
    memory = NULL,
    workers = NULL,
    tasks = NULL,
    segments = NULL,
    result_diagnostics = NULL
  )

  # Get pool information
  pool <- pool_get()
  if (!is.null(pool)) {
    rpt$pool <- list(
      n_workers = pool$n,
      rss_limit = pool$rss_limit_bytes,
      drift_threshold = pool$rss_drift_threshold,
      created_at = pool$created_at,
      stats = pool$stats
    )
  }

  # Memory summary (always included)
  rpt$memory <- mem_report()

  # Worker details (for workers level and above)
  if (level %in% c("workers", "tasks", "segments")) {
    if (!is.null(pool)) {
      rpt$workers <- lapply(seq_len(pool$n), function(i) {
        w <- pool$workers[[i]]
        if (is.null(w)) {
          return(list(id = i, status = "missing"))
        }
        worker_metrics(w)
      })
    }
  }

  # Task details (for tasks level and above)
  if (level %in% c("tasks", "segments")) {
    if (!is.null(result) && inherits(result, "shard_result")) {
      rpt$tasks <- task_report(result)
      rpt$result_diagnostics <- result$diagnostics
    }
  }

  # Performance recommendations (only when a result is provided).
  if (!is.null(result) && inherits(result, "shard_result")) {
    rpt$recommendations <- recommendations(result)
  }

  # Segment details (for segments level only)
  if (level == "segments") {
    rpt$segments <- segment_report()
  }

  structure(rpt, class = "shard_report")
}

#' Performance Recommendations
#'
#' Uses run telemetry (copy/materialization stats, packing volume, buffer/table
#' writes, scratch pool stats) to produce actionable recommendations.
#'
#' @param result A `shard_result` from [shard_map()].
#' @return A character vector of recommendations (possibly empty).
#' @export
#' @examples
#' \donttest{
#' res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
#' pool_stop()
#' recommendations(res)
#' }
recommendations <- function(result) {
  if (is.null(result) || !inherits(result, "shard_result")) {
    stop("result must be a shard_result", call. = FALSE)
  }

  recs <- character(0)

  cr <- tryCatch(copy_report(result), error = function(e) NULL)
  if (is.list(cr)) {
    mb <- as.double(cr$view_materialized_bytes %||% 0)
    pb <- as.double(cr$view_packed_bytes %||% 0)
    if (mb > 0) {
      recs <- c(recs, sprintf(
        "Views were materialized (%.1f MB). Prefer view_block()/view_gather() and shard-aware kernels to avoid slice copies.",
        mb / 1024^2
      ))
      hs <- result$diagnostics$view_hotspots %||% list()
      if (is.list(hs) && length(hs) > 0) {
        keys <- names(hs)
        if (length(keys) > 3) keys <- keys[seq_len(3)]
        parts <- vapply(keys, function(k) {
          b <- as.double(hs[[k]]$bytes %||% 0)
          paste0(format_bytes(b), " @ ", k)
        }, character(1))
        recs <- c(recs, paste0("Top view materialization hotspots: ", paste(parts, collapse = " ; ")))
      }
    }
    if (pb > 0) {
      recs <- c(recs, sprintf(
        "Views required packing (%.1f MB). Prefer contiguous/block shards when possible, or use gather-aware kernels to keep packing bounded.",
        pb / 1024^2
      ))
    }
    bb <- as.double(cr$buffer_bytes %||% 0)
    tb <- as.double(cr$table_bytes %||% 0)
    if (bb == 0 && tb == 0 && (mb > 0 || pb > 0)) {
      recs <- c(recs, "Consider writing large outputs into explicit buffers (buffer()/table_buffer()/table_sink()) to avoid master-side gather/concat.")
    }
  }

  d <- result$diagnostics %||% list()

  # Dispatch overhead guidance (tiny tasks).
  if (is.list(d)) {
    dm <- as.character(d$dispatch_mode %||% "rpc_chunked")
    shards_done <- as.double(d$shards_processed %||% NA_real_)
    chunks_done <- as.double(d$chunks_dispatched %||% NA_real_)
    dur <- as.double(d$duration %||% NA_real_)

    # Heuristic: 1 shard per chunk + very small per-shard time means RPC overhead
    # likely dominates.
    if (identical(dm, "rpc_chunked") &&
        is.finite(shards_done) && is.finite(chunks_done) && is.finite(dur) &&
        shards_done > 0 && chunks_done > 0 &&
        chunks_done == shards_done) {
      mean_shard_time <- dur / shards_done
      if (is.finite(mean_shard_time) && mean_shard_time > 0 && mean_shard_time < 0.02) {
        recs <- c(recs, sprintf(
          "Per-shard work looks tiny (%.3f ms/shard). Consider increasing chunk_size, or use profile='speed' / dispatch_mode='shm_queue' for out-buffer workflows to reduce dispatch overhead.",
          mean_shard_time * 1000
        ))
      }
    }

    # shm_queue mode returns a placeholder list of length n_tasks; for extremely
    # large n_tasks, recommend increasing block_size to reduce bookkeeping.
    if (identical(dm, "shm_queue") && is.list(d$shm_queue)) {
      nt <- as.double(d$shm_queue$n_tasks %||% NA_real_)
      if (is.finite(nt) && nt > 1e6) {
        recs <- c(recs, sprintf(
          "shm_queue is tracking %.0f tasks. If bookkeeping dominates, increase dispatch_opts$block_size to reduce n_tasks (you can keep work-per-task small without going all the way to 1).",
          nt
        ))
      }
    }
  }

  if (is.list(d$scratch_stats)) {
    hw <- as.double(d$scratch_stats$high_water %||% 0)
    if (hw > 0) {
      recs <- c(recs, sprintf(
        "Scratch pool high-water mark: %.1f MB. If allocations churn, consider using scratch_matrix() (and set scratch_pool_config(max_bytes=...)).",
        hw / 1024^2
      ))
    }
  }

  if (is.list(d$scheduler) && is.finite(as.double(d$scheduler$throttle_events %||% 0)) &&
      as.double(d$scheduler$throttle_events %||% 0) > 0) {
    recs <- c(recs, sprintf(
      "Scheduler throttled heavy chunks %d times. Consider tuning scheduler_policy (e.g., max_huge_concurrency) or reducing per-shard footprint.",
      as.integer(d$scheduler$throttle_events)
    ))
  }

  unique(recs)
}

#' Memory Usage Report
#'
#' Generates a report of memory usage across all workers in the pool.
#'
#' @param pool Optional. A \code{shard_pool} object. If NULL, uses the current pool.
#'
#' @return An S3 object of class \code{shard_report} with type \code{"memory"}
#'   containing:
#'   \itemize{
#'     \item \code{type}: "memory"
#'     \item \code{timestamp}: When the report was generated
#'     \item \code{pool_active}: Whether a pool exists
#'     \item \code{n_workers}: Number of workers
#'     \item \code{rss_limit}: RSS limit per worker (bytes)
#'     \item \code{total_rss}: Sum of RSS across all workers
#'     \item \code{peak_rss}: Highest RSS among workers
#'     \item \code{mean_rss}: Mean RSS across workers
#'     \item \code{workers}: Per-worker RSS details
#'   }
#'
#' @export
#' @examples
#' \donttest{
#' p <- pool_create(2)
#' mem_report(p)
#' pool_stop(p)
#' }
mem_report <- function(pool = NULL) {
  if (is.null(pool)) {
    pool <- pool_get()
  }

  rpt <- list(
    type = "memory",
    timestamp = Sys.time(),
    pool_active = !is.null(pool),
    n_workers = 0L,
    rss_limit = NA_real_,
    total_rss = 0,
    peak_rss = NA_real_,
    mean_rss = NA_real_,
    workers = list()
  )

  if (is.null(pool)) {
    return(structure(rpt, class = "shard_report"))
  }

  rpt$n_workers <- pool$n
  rpt$rss_limit <- pool$rss_limit_bytes

  # Collect RSS for each worker
  worker_stats <- lapply(seq_len(pool$n), function(i) {
    w <- pool$workers[[i]]
    if (is.null(w)) {
      return(list(
        id = i,
        pid = NA_integer_,
        status = "missing",
        rss = NA_real_,
        rss_baseline = NA_real_,
        drift = NA_real_
      ))
    }

    alive <- worker_is_alive(w)
    rss <- if (alive) worker_rss(w) else NA_real_
    baseline <- w$rss_baseline %||% NA_real_
    drift <- if (!is.na(rss) && !is.na(baseline) && baseline > 0) {
      (rss - baseline) / baseline
    } else {
      NA_real_
    }

    list(
      id = i,
      pid = w$pid,
      status = if (alive) "alive" else "dead",
      rss = rss,
      rss_baseline = baseline,
      drift = drift,
      recycle_count = w$recycle_count %||% 0L
    )
  })

  rpt$workers <- worker_stats

  # Compute aggregates
  rss_values <- vapply(worker_stats, function(x) x$rss %||% NA_real_, numeric(1))
  rss_valid <- rss_values[!is.na(rss_values)]

  if (length(rss_valid) > 0) {
    rpt$total_rss <- sum(rss_valid)
    rpt$peak_rss <- max(rss_valid)
    rpt$mean_rss <- mean(rss_valid)
  }

  structure(rpt, class = "shard_report")
}

#' Copy-on-Write Policy Report
#'
#' Generates a report of copy-on-write behavior for borrowed inputs.
#'
#' @param result Optional. A \code{shard_result} object to extract COW stats from.
#'
#' @return An S3 object of class \code{shard_report} with type \code{"cow"}
#'   containing:
#'   \itemize{
#'     \item \code{type}: "cow"
#'     \item \code{timestamp}: When the report was generated
#'     \item \code{policy}: The COW policy used ("deny", "audit", "allow")
#'     \item \code{violations}: Count of COW violations detected (audit mode)
#'     \item \code{copies_triggered}: Estimated copies triggered by mutations
#'   }
#'
#' @export
#' @examples
#' \donttest{
#' res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
#' pool_stop()
#' cow_report(res)
#' }
cow_report <- function(result = NULL) {
  rpt <- list(
    type = "cow",
    timestamp = Sys.time(),
    policy = NA_character_,
    violations = 0L,
    copies_triggered = 0L,
    details = list()
  )

  if (!is.null(result) && inherits(result, "shard_result")) {
    rpt$policy <- result$cow_policy %||% NA_character_

    # Extract COW diagnostics if available
    if (!is.null(result$diagnostics) && !is.null(result$diagnostics$cow_stats)) {
      cow_stats <- result$diagnostics$cow_stats
      rpt$violations <- cow_stats$violations %||% 0L
      rpt$copies_triggered <- cow_stats$copies %||% 0L
      rpt$details <- cow_stats$details %||% list()
    }
  }

  structure(rpt, class = "shard_report")
}

#' Data Copy Report
#'
#' Generates a report of data transfer and copy statistics during parallel
#' execution.
#'
#' @param result Optional. A \code{shard_result} object to extract copy stats from.
#'
#' @return An S3 object of class \code{shard_report} with type \code{"copy"}
#'   containing:
#'   \itemize{
#'     \item \code{type}: "copy"
#'     \item \code{timestamp}: When the report was generated
#'     \item \code{borrow_exports}: Number of borrowed input exports
#'     \item \code{borrow_bytes}: Total bytes in borrowed inputs
#'     \item \code{result_imports}: Number of result imports
#'     \item \code{result_bytes}: Estimated bytes in results
#'     \item \code{buffer_writes}: Number of buffer write operations
#'     \item \code{buffer_bytes}: Total bytes written to buffers
#'   }
#'
#' @export
#' @examples
#' \donttest{
#' res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
#' pool_stop()
#' copy_report(res)
#' }
copy_report <- function(result = NULL) {
  rpt <- list(
    type = "copy",
    timestamp = Sys.time(),
    borrow_exports = 0L,
    borrow_bytes = 0,
    result_imports = 0L,
    result_bytes = 0,
    buffer_writes = 0L,
    buffer_bytes = 0,
    table_writes = 0L,
    table_rows = 0L,
    table_bytes = 0,
    view_created = 0L,
    view_materialized = 0L,
    view_materialized_bytes = 0,
    view_materialize_hotspots = list(),
    view_packed = 0L,
    view_packed_bytes = 0
  )

  if (!is.null(result) && inherits(result, "shard_result")) {
    diag <- result$diagnostics

    # Count results
    if (!is.null(result$results)) {
      if (inherits(result$results, c("shard_results_placeholder", "shard_row_groups", "shard_dataset", "shard_table_handle"))) {
        # shm_queue results are not gathered; avoid reporting placeholder size.
        rpt$result_imports <- 0L
        rpt$result_bytes <- 0
      } else {
        rpt$result_imports <- length(result$results)
        # Estimate result size
        rpt$result_bytes <- tryCatch(
          as.numeric(utils::object.size(result$results)),
          error = function(e) 0
        )
      }
    }

    # Extract copy diagnostics if available
    if (!is.null(diag) && !is.null(diag$copy_stats)) {
      copy_stats <- diag$copy_stats
      rpt$borrow_exports <- copy_stats$borrow_exports %||% 0L
      rpt$borrow_bytes <- copy_stats$borrow_bytes %||% 0
      rpt$buffer_writes <- copy_stats$buffer_writes %||% 0L
      rpt$buffer_bytes <- copy_stats$buffer_bytes %||% 0
    }

    if (!is.null(diag) && !is.null(diag$table_stats)) {
      ts <- diag$table_stats
      rpt$table_writes <- ts$writes %||% 0L
      rpt$table_rows <- ts$rows %||% 0L
      rpt$table_bytes <- ts$bytes %||% 0
    }
  }

  # Prefer run-level view stats when present (aggregated from workers).
  if (!is.null(result) && inherits(result, "shard_result")) {
    diag <- result$diagnostics
    if (!is.null(diag) && !is.null(diag$view_stats)) {
      vs <- diag$view_stats
      rpt$view_created <- vs$created %||% 0L
      rpt$view_materialized <- vs$materialized %||% 0L
      rpt$view_materialized_bytes <- vs$materialized_bytes %||% 0
      rpt$view_packed <- vs$packed %||% 0L
      rpt$view_packed_bytes <- vs$packed_bytes %||% 0
    }
    if (!is.null(diag) && is.list(diag$view_hotspots)) {
      rpt$view_materialize_hotspots <- diag$view_hotspots %||% list()
    }
  }

  if (is.null(rpt$view_created) || is.null(rpt$view_materialized)) {
    # Fallback to global view diagnostics when run-level aggregation isn't available.
    vd <- tryCatch(view_diagnostics(), error = function(e) NULL)
    if (is.list(vd)) {
      rpt$view_created <- vd$created %||% 0L
      rpt$view_materialized <- vd$materialized %||% 0L
      rpt$view_materialized_bytes <- vd$materialized_bytes %||% 0
      rpt$view_packed <- vd$packed %||% 0L
      rpt$view_packed_bytes <- vd$packed_bytes %||% 0
    }
  }

  structure(rpt, class = "shard_report")
}

#' Task Execution Report
#'
#' Generates a report of task/chunk execution statistics from a shard_map result.
#'
#' @param result A \code{shard_result} object from \code{\link{shard_map}}.
#'
#' @return An S3 object of class \code{shard_report} with type \code{"task"}
#'   containing:
#'   \itemize{
#'     \item \code{type}: "task"
#'     \item \code{timestamp}: When the report was generated
#'     \item \code{shards_total}: Total number of shards
#'     \item \code{shards_processed}: Number of shards successfully processed
#'     \item \code{shards_failed}: Number of permanently failed shards
#'     \item \code{chunks_dispatched}: Number of chunk batches dispatched
#'     \item \code{total_retries}: Total number of retry attempts
#'     \item \code{duration}: Total execution duration (seconds)
#'     \item \code{throughput}: Shards processed per second
#'     \item \code{queue_status}: Final queue status
#'   }
#'
#' @export
#' @examples
#' \donttest{
#' res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
#' pool_stop()
#' task_report(res)
#' }
task_report <- function(result = NULL) {
  rpt <- list(
    type = "task",
    timestamp = Sys.time(),
    shards_total = 0L,
    shards_processed = 0L,
    shards_failed = 0L,
    chunks_dispatched = 0L,
    total_retries = 0L,
    duration = NA_real_,
    throughput = NA_real_,
    queue_status = NULL,
    health_checks = list()
  )

  if (is.null(result) || !inherits(result, "shard_result")) {
    return(structure(rpt, class = "shard_report"))
  }

  # Extract from shards descriptor
  if (!is.null(result$shards)) {
    rpt$shards_total <- result$shards$num_shards %||% 0L
  }

  # Extract from diagnostics
  if (!is.null(result$diagnostics)) {
    diag <- result$diagnostics
    rpt$shards_processed <- diag$shards_processed %||% 0L
    rpt$chunks_dispatched <- diag$chunks_dispatched %||% 0L
    rpt$duration <- diag$duration %||% NA_real_
    rpt$health_checks <- diag$health_checks %||% list()
  }

  # Extract from queue status
  if (!is.null(result$queue_status)) {
    qs <- result$queue_status
    rpt$queue_status <- qs
    rpt$shards_failed <- qs$failed %||% 0L
    rpt$total_retries <- qs$total_retries %||% 0L
  }

  # Extract from failures
  if (!is.null(result$failures)) {
    rpt$shards_failed <- max(rpt$shards_failed, length(result$failures))
  }

  # Compute throughput
  if (!is.na(rpt$duration) && rpt$duration > 0 && rpt$shards_processed > 0) {
    rpt$throughput <- rpt$shards_processed / rpt$duration
  }

  structure(rpt, class = "shard_report")
}

#' Shared Memory Segment Report
#'
#' Generates a report of active shared memory segments in the current session.
#'
#' @return An S3 object of class \code{shard_report} with type \code{"segment"}
#'   containing:
#'   \itemize{
#'     \item \code{type}: "segment"
#'     \item \code{timestamp}: When the report was generated
#'     \item \code{n_segments}: Number of tracked segments
#'     \item \code{total_bytes}: Total bytes across all segments
#'     \item \code{segments}: List of segment details
#'   }
#'
#' @details
#' This function reports on segments that are currently accessible. Note that
#' segments are automatically cleaned up when their R objects are garbage
#' collected, so this only shows segments with live references.
#'
#' @export
#' @examples
#' \donttest{
#' segment_report()
#' }
segment_report <- function() {
  rpt <- list(
    type = "segment",
    timestamp = Sys.time(),
    n_segments = 0L,
    total_bytes = 0,
    segments = list(),
    backing_summary = list()
  )

  # Get available backings
  available <- tryCatch(
    available_backings(),
    error = function(e) c("mmap")
  )
  rpt$backing_summary$available <- available

  # Note: We can't enumerate all segments because they're tracked by

# individual R objects. This provides platform information instead.
  rpt$backing_summary$platform <- .Platform$OS.type
  rpt$backing_summary$is_windows <- tryCatch(
    is_windows(),
    error = function(e) .Platform$OS.type == "windows"
  )

  structure(rpt, class = "shard_report")
}

#' Print a shard_report Object
#'
#' @param x A \code{shard_report} object.
#' @param ... Ignored.
#' @return The input \code{x}, invisibly.
#' @examples
#' \donttest{
#' res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
#' pool_stop()
#' rpt <- report(result = res)
#' print(rpt)
#' }
#' @export
print.shard_report <- function(x, ...) {
  # Dispatch based on report type
  if (!is.null(x$type)) {
    switch(x$type,
      "memory" = print_mem_report(x),
      "cow" = print_cow_report(x),
      "copy" = print_copy_report(x),
      "task" = print_task_report(x),
      "segment" = print_segment_report(x),
      print_main_report(x)
    )
  } else {
    print_main_report(x)
  }

  invisible(x)
}

#' Print Main Report
#' @param x A shard_report object.
#' @keywords internal
#' @noRd
print_main_report <- function(x) {
  cat("shard_report (", x$level, ")\n", sep = "")
  cat("Generated:", format(x$timestamp), "\n")

  # Pool summary
  if (!is.null(x$pool)) {
    cat("\nPool:\n")
    cat("  Workers:", x$pool$n_workers, "\n")
    cat("  RSS limit:", format_bytes(x$pool$rss_limit), "\n")
    cat("  Drift threshold:", sprintf("%.0f%%", x$pool$drift_threshold * 100), "\n")
    if (!is.null(x$pool$stats)) {
      cat("  Stats: ", x$pool$stats$total_tasks, " tasks, ",
          x$pool$stats$total_recycles, " recycles, ",
          x$pool$stats$total_deaths, " deaths\n", sep = "")
    }
  } else {
    cat("\nPool: (not active)\n")
  }

  # Memory summary
  if (!is.null(x$memory) && inherits(x$memory, "shard_report")) {
    cat("\nMemory:\n")
    if (x$memory$pool_active) {
      cat("  Total RSS:", format_bytes(x$memory$total_rss), "\n")
      cat("  Peak RSS:", format_bytes(x$memory$peak_rss), "\n")
      cat("  Mean RSS:", format_bytes(x$memory$mean_rss), "\n")
    } else {
      cat("  (no pool active)\n")
    }
  }

  # Worker details (if included)
  if (!is.null(x$workers) && length(x$workers) > 0) {
    cat("\nWorkers:\n")
    for (w in x$workers) {
      status_icon <- switch(w$status %||% "unknown",
        "alive" = "+",
        "ok" = "+",
        "dead" = "x",
        "missing" = "?",
        "?"
      )
      cat(sprintf("  [%s] %d: pid=%s, rss=%s, drift=%.1f%%, recycles=%d\n",
        status_icon,
        w$id %||% w$worker_id %||% NA,
        w$pid %||% "NA",
        format_bytes(w$rss %||% w$rss_bytes %||% NA_real_),
        (w$drift %||% 0) * 100,
        w$recycle_count %||% 0
      ))
    }
  }

  # Task details (if included)
  if (!is.null(x$tasks) && inherits(x$tasks, "shard_report")) {
    cat("\nTasks:\n")
    cat("  Total:", x$tasks$shards_total, "\n")
    cat("  Processed:", x$tasks$shards_processed, "\n")
    if (x$tasks$shards_failed > 0) {
      cat("  Failed:", x$tasks$shards_failed, "\n")
    }
    if (!is.na(x$tasks$duration)) {
      cat("  Duration:", sprintf("%.2f seconds", x$tasks$duration), "\n")
    }
    if (!is.na(x$tasks$throughput)) {
      cat("  Throughput:", sprintf("%.1f shards/sec", x$tasks$throughput), "\n")
    }
  }

  # Segment details (if included)
  if (!is.null(x$segments) && inherits(x$segments, "shard_report")) {
    cat("\nSegments:\n")
    cat("  Available backings:", paste(x$segments$backing_summary$available, collapse = ", "), "\n")
    cat("  Platform:", x$segments$backing_summary$platform, "\n")
  }

  if (!is.null(x$recommendations) && length(x$recommendations) > 0) {
    cat("\nRecommendations:\n")
    for (r in x$recommendations) {
      cat("  - ", r, "\n", sep = "")
    }
  }
}

#' Print Memory Report
#' @param x A shard_report object with type "memory".
#' @keywords internal
#' @noRd
print_mem_report <- function(x) {
  cat("shard memory report\n")
  cat("Generated:", format(x$timestamp), "\n")

  if (!x$pool_active) {
    cat("\nNo pool active.\n")
    return(invisible(x))
  }

  cat("\nPool: ", x$n_workers, " workers\n", sep = "")
  cat("RSS limit:", format_bytes(x$rss_limit), "\n")

  cat("\nAggregate:\n")
  cat("  Total:", format_bytes(x$total_rss), "\n")
  cat("  Peak:", format_bytes(x$peak_rss), "\n")
  cat("  Mean:", format_bytes(x$mean_rss), "\n")

  cat("\nPer-worker:\n")
  for (w in x$workers) {
    status_icon <- switch(w$status %||% "unknown",
      "alive" = "+",
      "ok" = "+",
      "dead" = "x",
      "missing" = "?",
      "?"
    )
    cat(sprintf("  [%s] %d: rss=%s, baseline=%s, drift=%.1f%%\n",
      status_icon,
      w$id,
      format_bytes(w$rss %||% NA_real_),
      format_bytes(w$rss_baseline %||% NA_real_),
      (w$drift %||% 0) * 100
    ))
  }
}

#' Print COW Report
#' @param x A shard_report object with type "cow".
#' @keywords internal
#' @noRd
print_cow_report <- function(x) {
  cat("shard copy-on-write report\n")
  cat("Generated:", format(x$timestamp), "\n")

  cat("\nPolicy:", x$policy %||% "(unknown)", "\n")
  cat("Violations:", x$violations, "\n")
  cat("Copies triggered:", x$copies_triggered, "\n")

  if (length(x$details) > 0) {
    cat("\nDetails:\n")
    for (d in x$details) {
      cat("  -", d, "\n")
    }
  }
}

#' Print Copy Report
#' @param x A shard_report object with type "copy".
#' @keywords internal
#' @noRd
print_copy_report <- function(x) {
  cat("shard data copy report\n")
  cat("Generated:", format(x$timestamp), "\n")

  cat("\nBorrowed inputs:\n")
  cat("  Exports:", x$borrow_exports, "\n")
  cat("  Bytes:", format_bytes(x$borrow_bytes), "\n")

  cat("\nResults:\n")
  cat("  Imports:", x$result_imports, "\n")
  cat("  Bytes:", format_bytes(x$result_bytes), "\n")

  cat("\nBuffers:\n")
  cat("  Writes:", x$buffer_writes, "\n")
  cat("  Bytes:", format_bytes(x$buffer_bytes), "\n")

  if (!is.null(x$table_writes) || !is.null(x$table_rows)) {
    cat("\nTables:\n")
    cat("  Writes:", x$table_writes %||% 0L, "\n")
    cat("  Rows:", x$table_rows %||% 0L, "\n")
    cat("  Bytes:", format_bytes(x$table_bytes %||% 0), "\n")
  }

  if (!is.null(x$view_created) || !is.null(x$view_materialized)) {
    cat("\nViews:\n")
    cat("  Created:", x$view_created %||% 0L, "\n")
    cat("  Materialized:", x$view_materialized %||% 0L, "\n")
    cat("  Materialized bytes:", format_bytes(x$view_materialized_bytes %||% 0), "\n")
    if (!is.null(x$view_packed) || !is.null(x$view_packed_bytes)) {
      cat("  Packed:", x$view_packed %||% 0L, "\n")
      cat("  Packed bytes:", format_bytes(x$view_packed_bytes %||% 0), "\n")
    }
  }
}

#' Print Task Report
#' @param x A shard_report object with type "task".
#' @keywords internal
#' @noRd
print_task_report <- function(x) {
  cat("shard task report\n")
  cat("Generated:", format(x$timestamp), "\n")

  cat("\nExecution:\n")
  cat("  Total shards:", x$shards_total, "\n")
  cat("  Processed:", x$shards_processed, "\n")
  cat("  Failed:", x$shards_failed, "\n")
  cat("  Chunks dispatched:", x$chunks_dispatched, "\n")

  if (x$total_retries > 0) {
    cat("  Retries:", x$total_retries, "\n")
  }

  cat("\nTiming:\n")
  if (!is.na(x$duration)) {
    cat("  Duration:", sprintf("%.2f seconds", x$duration), "\n")
  }
  if (!is.na(x$throughput)) {
    cat("  Throughput:", sprintf("%.1f shards/sec", x$throughput), "\n")
  }

  if (length(x$health_checks) > 0) {
    cat("\nHealth checks:", length(x$health_checks), "performed\n")
  }
}

#' Print Segment Report
#' @param x A shard_report object with type "segment".
#' @keywords internal
#' @noRd
print_segment_report <- function(x) {
  cat("shard segment report\n")
  cat("Generated:", format(x$timestamp), "\n")

  cat("\nBacking types:\n")
  cat("  Available:", paste(x$backing_summary$available, collapse = ", "), "\n")
  cat("  Platform:", x$backing_summary$platform, "\n")
  cat("  Windows:", x$backing_summary$is_windows, "\n")
}

# Tests for diagnostics API

test_that("report() returns shard_report with correct structure", {
  # Without pool
  rpt <- report()
  expect_s3_class(rpt, "shard_report")
  expect_equal(rpt$level, "summary")
  expect_true(!is.null(rpt$timestamp))
  expect_true(!is.null(rpt$memory))

  # With pool
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  rpt <- report()
  expect_s3_class(rpt, "shard_report")
  expect_true(!is.null(rpt$pool))
  expect_equal(rpt$pool$n_workers, 2L)
})

test_that("report() respects detail levels", {
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Summary level - no workers
  rpt_summary <- report("summary")
  expect_null(rpt_summary$workers)
  expect_null(rpt_summary$segments)

  # Workers level - has workers
  rpt_workers <- report("workers")
  expect_true(!is.null(rpt_workers$workers))
  expect_length(rpt_workers$workers, 2)

  # Segments level - has segments
  rpt_segments <- report("segments")
  expect_true(!is.null(rpt_segments$segments))
  expect_true(!is.null(rpt_segments$workers))
})

test_that("report() includes shard_result diagnostics", {
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Use explicit block_size=1 to get exact shard count
  result <- shard_map(shards(10, block_size = 1), function(s) sum(s$idx), workers = 2)

  rpt <- report("tasks", result = result)
  expect_true(!is.null(rpt$tasks))
  expect_s3_class(rpt$tasks, "shard_report")
  expect_equal(rpt$tasks$shards_total, 10)
  expect_true(!is.null(rpt$result_diagnostics))
})

test_that("mem_report() returns memory statistics", {
  # Without pool
  rpt <- mem_report()
  expect_s3_class(rpt, "shard_report")
  expect_equal(rpt$type, "memory")
  expect_false(rpt$pool_active)

  # With pool
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  rpt <- mem_report()
  expect_true(rpt$pool_active)
  expect_equal(rpt$n_workers, 2)
  expect_length(rpt$workers, 2)
  expect_true(all(vapply(rpt$workers, function(w) !is.null(w$id), logical(1))))
})

test_that("mem_report() computes aggregates correctly", {
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  rpt <- mem_report()

  # Should have computed total, peak, mean from worker RSS values
  expect_true(is.numeric(rpt$total_rss))
  expect_true(is.numeric(rpt$peak_rss))
  expect_true(is.numeric(rpt$mean_rss))

  # Total should be sum of workers
  worker_rss <- vapply(rpt$workers, function(w) w$rss %||% 0, numeric(1))
  valid_rss <- worker_rss[!is.na(worker_rss)]
  if (length(valid_rss) > 0) {
    expect_equal(rpt$total_rss, sum(valid_rss))
  }
})

test_that("cow_report() returns COW statistics", {
  rpt <- cow_report()
  expect_s3_class(rpt, "shard_report")
  expect_equal(rpt$type, "cow")
  expect_true(is.na(rpt$policy))

  # With result
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  result <- shard_map(shards(5), function(s) s$idx, cow = "audit", workers = 2)
  rpt <- cow_report(result)
  expect_equal(rpt$policy, "audit")
})

test_that("copy_report() returns data copy statistics", {
  rpt <- copy_report()
  expect_s3_class(rpt, "shard_report")
  expect_equal(rpt$type, "copy")
  expect_equal(rpt$borrow_exports, 0L)

  # With result
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  result <- shard_map(shards(5), function(s) sum(s$idx), workers = 2)
  rpt <- copy_report(result)
  expect_equal(rpt$result_imports, 5)  # 5 shards = 5 results
  expect_true(rpt$result_bytes > 0)  # Results have some size
})

test_that("task_report() returns task execution statistics", {
  rpt <- task_report()
  expect_s3_class(rpt, "shard_report")
  expect_equal(rpt$type, "task")
  expect_equal(rpt$shards_total, 0L)

  # With result - use explicit block_size=1 for exact shard count
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  result <- shard_map(shards(10, block_size = 1), function(s) sum(s$idx), workers = 2)
  rpt <- task_report(result)

  expect_equal(rpt$shards_total, 10)
  expect_equal(rpt$shards_processed, 10)
  expect_equal(rpt$shards_failed, 0L)
  expect_true(!is.na(rpt$duration))
  expect_true(rpt$duration > 0)
  expect_true(!is.na(rpt$throughput))
  expect_true(rpt$throughput > 0)
})

test_that("task_report() reports failures correctly", {
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Create a task that fails for some shards - use block_size=1 for predictable shards
  result <- suppressWarnings(
    shard_map(shards(5, block_size = 1), function(s) {
      if (s$id == 3) stop("deliberate failure")
      sum(s$idx)
    }, workers = 2, max_retries = 1L)
  )

  rpt <- task_report(result)
  expect_true(rpt$shards_failed >= 1)  # At least one failure
})

test_that("segment_report() returns segment information", {
  rpt <- segment_report()
  expect_s3_class(rpt, "shard_report")
  expect_equal(rpt$type, "segment")
  expect_true(!is.null(rpt$backing_summary))
  expect_true(!is.null(rpt$backing_summary$available))
  expect_true(!is.null(rpt$backing_summary$platform))
})

test_that("print methods work without error", {
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  result <- shard_map(shards(5), function(s) sum(s$idx), workers = 2)

  # Test all print methods
  expect_output(print(report()))
  expect_output(print(report("workers")))
  expect_output(print(report("tasks", result = result)))
  expect_output(print(report("segments")))
  expect_output(print(mem_report()))
  expect_output(print(cow_report()))
  expect_output(print(cow_report(result)))
  expect_output(print(copy_report()))
  expect_output(print(copy_report(result)))
  expect_output(print(task_report()))
  expect_output(print(task_report(result)))
  expect_output(print(segment_report()))
})

test_that("report() handles missing/dead workers", {
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Get report with live workers
  rpt <- report("workers")

  for (w in rpt$workers) {
    expect_true(w$status %in% c("alive", "ok", "missing", "dead"))
  }
})

test_that("recommendations() is deterministic given the same run telemetry", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  X <- share(matrix(rnorm(200), nrow = 20), backing = "mmap")
  blocks <- shards(10, block_size = 1)

  # Force view materialization inside the task.
  result <- shard_map(blocks, borrow = list(X = X), fun = function(s, X) {
    v <- view_block(X, cols = idx_range(1, 5))
    sum(materialize(v))
  }, workers = 2, diagnostics = TRUE)

  r1 <- recommendations(result)
  r2 <- recommendations(result)
  expect_identical(r1, r2)
  expect_true(any(grepl("Views were materialized", r1, fixed = TRUE)))
})

test_that("reports have consistent timestamp format", {
  rpt1 <- report()
  rpt2 <- mem_report()
  rpt3 <- cow_report()
  rpt4 <- copy_report()
  rpt5 <- task_report()
  rpt6 <- segment_report()

  expect_s3_class(rpt1$timestamp, "POSIXct")
  expect_s3_class(rpt2$timestamp, "POSIXct")
  expect_s3_class(rpt3$timestamp, "POSIXct")
  expect_s3_class(rpt4$timestamp, "POSIXct")
  expect_s3_class(rpt5$timestamp, "POSIXct")
  expect_s3_class(rpt6$timestamp, "POSIXct")
})

#' @title Parallel Execution with shard_map
#' @description Core parallel execution engine with supervision, shared inputs, and output buffers.
#' @name shard_map
NULL

#' Parallel Shard Execution
#'
#' Executes a function over shards in parallel with worker supervision,
#' shared inputs, and explicit output buffers. This is the primary entry
#' point for shard's parallel execution model.
#'
#' @param shards A `shard_descriptor` from [shards()], or an integer N to
#'   auto-generate shards.
#' @param fun Function to execute per shard. Receives the shard descriptor
#'   as first argument, followed by borrowed inputs and outputs. You can also
#'   select a registered kernel via `kernel=` instead of providing `fun=`.
#' @param borrow Named list of shared inputs. These are exported to workers
#'   once and reused across shards. Treated as read-only by default.
#' @param out Named list of output buffers (from `buffer()`). Workers write
#'   results directly to these buffers.
#' @param kernel Optional. Name of a registered kernel (see [list_kernels()]).
#'   If provided, `fun` must be NULL.
#' @param scheduler_policy Optional list of scheduling hints (advanced). Currently:
#'   - `max_huge_concurrency`: cap concurrent chunks whose kernel footprint is
#'     classified as `"huge"` (see [register_kernel()]).
#' @param autotune Optional. Online autotuning for scalar-N sharding (advanced).
#'   When `shards` is an integer `N`, shard_map can adjust shard block sizes over
#'   time based on observed wall time and worker RSS.
#'
#'   Accepted values:
#'   - `NULL` (default): enable online autotuning for `shard_map(N, ...)`, off for
#'     precomputed shard descriptors.
#'   - `TRUE` / `"online"`: force online autotuning (only applies when `shards` is
#'     an integer `N`).
#'   - `FALSE` / `"none"`: disable autotuning.
#'   - a list: `list(mode="online", max_rounds=..., probe_shards_per_worker=..., min_shard_time=...)`
#' @param dispatch_mode Dispatch mode (advanced). `"rpc_chunked"` is the default
#'   supervised socket-based dispatcher. `"shm_queue"` is an opt-in fast mode
#'   that uses a shared-memory task queue to reduce per-task overhead for tiny
#'   tasks. In v1, `"shm_queue"` is only supported for `shard_map(N, ...)` with
#'   `chunk_size=1` and is intended for out-buffer/sink workflows (results are
#'   not gathered).
#' @param dispatch_opts Optional list of dispatch-mode specific knobs (advanced).
#'   Currently:
#'   - For `dispatch_mode="rpc_chunked"`:
#'     - `auto_table`: logical. If TRUE, shard_map treats data.frame/tibble return
#'       values as row-group outputs and writes them to a table sink
#'       automatically (one partition per shard id). This avoids building a large
#'       list of tibbles and calling bind_rows() on the master. Requires `out=`
#'       to be empty (use explicit `out=list(sink=table_sink(...))` otherwise).
#'     - `auto_table_materialize`: `"never"`, `"auto"`, or `"always"` (default `"auto"`).
#'     - `auto_table_max_bytes`: numeric/integer. For `"auto"`, materialize only
#'       if estimated output size <= this threshold (default 256MB).
#'     - `auto_table_mode`: `"row_groups"` (default) or `"partitioned"`.
#'     - `auto_table_path`: optional output directory (default tempdir()).
#'     - `auto_table_format`: `"auto"`, `"rds"` (default), or `"native"`.
#'     - `auto_table_schema`: optional `shard_schema` for validation/native encoding.
#'   - For `dispatch_mode="shm_queue"`:
#'     - `block_size`: integer. If provided, overrides the default heuristic for
#'       contiguous shard block sizing.
#'     - `queue_backing`: one of `"mmap"` or `"shm"` (default `"mmap"`).
#'     - `error_log`: logical. If TRUE, workers write a bounded per-worker error
#'       log to disk to aid debugging failed tasks (default FALSE).
#'     - `error_log_max_lines`: integer. Maximum lines per worker in the error
#'       log (default 100).
#' @param workers Integer. Number of worker processes. If NULL, uses existing
#'   pool or creates one with `detectCores() - 1`.
#' @param chunk_size Integer. Shards to batch per worker dispatch (default 1).
#'   Higher values reduce RPC overhead but may hurt load balancing.
#' @param profile Execution profile: `"default"`, `"memory"` (aggressive recycling),
#'   or `"speed"` (minimal overhead). With `profile="speed"`, shard_map will
#'   automatically enable `dispatch_mode="shm_queue"` when possible for
#'   `shard_map(N, ...)` out-buffer workflows (scalar `N`, `chunk_size=1`),
#'   unless `dispatch_mode` is explicitly specified.
#' @param mem_cap Memory cap per worker (e.g., "2GB"). Workers exceeding this
#'   are recycled.
#' @param recycle Logical or numeric. If TRUE, recycle workers on RSS drift.
#'   If numeric, specifies drift threshold (default 0.5 = 50% growth).
#' @param cow Copy-on-write policy for borrowed inputs: `"deny"` (error on mutation),
#'   `"audit"` (detect and flag), or `"allow"` (permit with tracking).
#' @param seed Integer. RNG seed for reproducibility. If NULL, no seed is set.
#' @param diagnostics Logical. Collect detailed diagnostics (default TRUE).
#' @param packages Character vector. Additional packages to load in workers.
#' @param init_expr Expression to evaluate in each worker on startup.
#' @param timeout Numeric. Seconds to wait for each shard (default 3600).
#' @param max_retries Integer. Maximum retries per shard on failure (default 3).
#' @param health_check_interval Integer. Check worker health every N shards (default 10).
#'
#' @return A `shard_result` object containing:
#'   - `results`: List of results from each shard (if fun returns values)
#'   - `failures`: Any permanently failed shards
#'   - `diagnostics`: Timing, memory, and worker statistics
#'   - `pool_stats`: Pool-level statistics
#'
#' @export
#' @examples
#' \donttest{
#' blocks <- shards(1000, workers = 2)
#' result <- shard_map(blocks, function(shard) {
#'   sum(shard$idx^2)
#' }, workers = 2)
#' pool_stop()
#' }
shard_map <- function(shards,
                      fun = NULL,
                      borrow = list(),
                      out = list(),
                      kernel = NULL,
                      scheduler_policy = NULL,
                      autotune = NULL,
                      dispatch_mode = c("rpc_chunked", "shm_queue"),
                      dispatch_opts = NULL,
                      workers = NULL,
                      chunk_size = 1L,
                      profile = c("default", "memory", "speed"),
                      mem_cap = "2GB",
                      recycle = TRUE,
                      cow = c("deny", "audit", "allow"),
                      seed = NULL,
                      diagnostics = TRUE,
                      packages = NULL,
                      init_expr = NULL,
                      timeout = 3600,
                      max_retries = 3L,
                      health_check_interval = 10L) {
  profile <- match.arg(profile)
  cow <- match.arg(cow)
  dispatch_mode_user_provided <- !missing(dispatch_mode)
  dispatch_mode <- if (dispatch_mode_user_provided) match.arg(dispatch_mode) else "rpc_chunked"
  if (is.null(dispatch_opts)) dispatch_opts <- list()
  if (!is.list(dispatch_opts)) stop("dispatch_opts must be NULL or a list", call. = FALSE)

  kernel_meta <- NULL
  if (!is.null(kernel)) {
    kernel <- as.character(kernel)
    km <- get_kernel(kernel)
    if (is.null(km)) stop("Unknown kernel: ", kernel, call. = FALSE)
    if (!is.null(fun)) stop("Provide either fun= or kernel=, not both", call. = FALSE)
    fun <- km$impl
    kernel_meta <- km
  }
  if (!is.function(fun)) stop("fun must be a function (or specify kernel=)", call. = FALSE)

  # Start timing
  start_time <- Sys.time()
  diag <- if (diagnostics) {
    list(
      start_time = start_time,
      health_checks = list(),
      shard_times = list(),
      worker_usage = list(),
      kernel = kernel %||% NULL,
      autotune = NULL,
      dispatch_mode = dispatch_mode
    )
  } else {
    NULL
  }

  # If the user passed an integer N, we can optionally do online autotuning
  # while generating shards in phases (no up-front huge shard list required).
  shards_is_scalar_n <- is.numeric(shards) && length(shards) == 1
  n_items <- if (shards_is_scalar_n) as.integer(shards) else NA_integer_

  # Determine worker count
  if (is.null(workers)) {
    pool <- pool_get()
    if (!is.null(pool)) {
      workers <- pool$n
    } else {
      dc <- parallel::detectCores()
      if (is.na(dc) || dc < 1L) dc <- 1L
      workers <- max(dc - 1L, 1L)
    }
  }
  workers <- as.integer(workers)
  if (is.na(workers) || workers < 1L) workers <- 1L

  # Apply profile settings
  profile_settings <- get_profile_settings(profile, mem_cap, recycle)
  mem_cap <- profile_settings$mem_cap
  rss_drift_threshold <- profile_settings$rss_drift_threshold
  health_check_interval <- profile_settings$health_check_interval

  # Convert integer to shard_descriptor if needed (after worker/profile resolution).
  if (shards_is_scalar_n) {
    if (is.na(n_items) || n_items < 1L) stop("shards must be >= 1", call. = FALSE)
  } else {
    if (!inherits(shards, "shard_descriptor")) {
      stop("shards must be a shard_descriptor or integer", call. = FALSE)
    }
  }

  # Validate inputs before expensive pool creation
  borrow <- validate_borrow(borrow, cow)
  auto_shared_names <- attr(borrow, "auto_shared")
  if (length(auto_shared_names) > 0) {
    on.exit({
      for (nm in auto_shared_names) {
        tryCatch(close(borrow[[nm]]), error = function(e) NULL)
      }
    }, add = TRUE)
  }
  out <- validate_out(out)

  # Ensure pool exists with correct worker count
 pool <- ensure_pool(
    workers = workers,
    mem_cap = mem_cap,
    rss_drift_threshold = rss_drift_threshold,
    packages = packages,
    init_expr = init_expr
  )

  auto_table <- isTRUE(dispatch_opts$auto_table %||% FALSE)
  auto_table_sink <- NULL
  if (auto_table) {
    if (length(out) > 0) {
      stop("dispatch_opts$auto_table=TRUE requires out= to be empty; use out=list(sink=table_sink(...)) for explicit table outputs.", call. = FALSE)
    }

    auto_table_mode <- as.character(dispatch_opts$auto_table_mode %||% "row_groups")
    if (!auto_table_mode %in% c("row_groups", "partitioned")) {
      stop("dispatch_opts$auto_table_mode must be 'row_groups' or 'partitioned'", call. = FALSE)
    }
    auto_table_path <- dispatch_opts$auto_table_path %||% NULL
    auto_table_format <- as.character(dispatch_opts$auto_table_format %||% "rds")
    if (!auto_table_format %in% c("auto", "rds", "native")) {
      stop("dispatch_opts$auto_table_format must be 'auto', 'rds', or 'native'", call. = FALSE)
    }

    # Schema-less by default to keep ceremony low; users can supply a shard_schema
    # via dispatch_opts$auto_table_schema for native encoding + strict validation.
    auto_table_schema <- dispatch_opts$auto_table_schema %||% NULL
    if (!is.null(auto_table_schema) && !inherits(auto_table_schema, "shard_schema")) {
      stop("dispatch_opts$auto_table_schema must be a shard_schema (or NULL)", call. = FALSE)
    }

    auto_table_sink <- table_sink(
      schema = auto_table_schema,
      mode = auto_table_mode,
      path = auto_table_path,
      format = auto_table_format
    )
  }

  # Low-ceremony fast path: profile="speed" will automatically use shm_queue
  # for scalar-N, chunk_size=1 out-buffer workflows unless dispatch_mode was
  # explicitly set by the user.
  if (!dispatch_mode_user_provided &&
      identical(profile, "speed") &&
      shards_is_scalar_n &&
      as.integer(chunk_size) == 1L &&
      length(out) > 0 &&
      taskq_supported()) {
    dispatch_mode <- "shm_queue"
    if (diagnostics) diag$dispatch_mode <- dispatch_mode
  }

  # Set seed in workers if specified
  if (!is.null(seed)) {
    # When shards are auto-generated online, we don't know num_shards yet.
    # Use total items as a stable substream spacing.
    set_worker_seeds(pool, seed, if (shards_is_scalar_n) n_items else shards$num_shards)
  }

  # Export borrowed inputs to workers (once, not per shard)
  export_borrow_to_workers(pool, borrow)

  # Export output buffer references if any
  if (length(out) > 0) {
    export_out_to_workers(pool, out)
  }

  # Optional: auto table sink for tibble/data.frame return values.
  if (!is.null(auto_table_sink)) {
    export_auto_table_sink_to_workers(pool, auto_table_sink)
  }

  if (isTRUE(diagnostics)) {
    reset_worker_diagnostics_(pool)
  }

  # shm_queue fast mode: chunk_size=1, fire-and-forget (no gathered results).
  if (identical(dispatch_mode, "shm_queue")) {
    if (!taskq_supported()) {
      warning("dispatch_mode='shm_queue' not supported on this platform; falling back to rpc_chunked", call. = FALSE)
      dispatch_mode <- "rpc_chunked"
    } else {
    if (isTRUE(auto_table)) {
      stop("dispatch_opts$auto_table is not supported in dispatch_mode='shm_queue' (use rpc_chunked or explicit out=table_sink())", call. = FALSE)
    }
    if (as.integer(chunk_size) != 1L) {
      stop("dispatch_mode='shm_queue' currently requires chunk_size=1", call. = FALSE)
    }
    if (length(out) == 0) {
      warning("dispatch_mode='shm_queue' does not gather results; prefer using out= buffers/sinks.", call. = FALSE)
    }

    queue_backing <- dispatch_opts$queue_backing %||% "mmap"

    if (shards_is_scalar_n) {
      block_size <- dispatch_opts$block_size %||% autotune_block_size(
          n = n_items,
          workers = workers,
          min_shards_per_worker = 4L,
          max_shards_per_worker = 64L
        )
      block_size <- as.integer(block_size)
      if (is.na(block_size) || block_size < 1L) stop("dispatch_opts$block_size must be >= 1", call. = FALSE)

      shards <- shards_lazy(n_items, block_size = block_size)

      dispatch_result <- dispatch_shards_shm_queue_(
        n = n_items,
        block_size = block_size,
        shards = NULL,
        fun = fun,
        borrow = borrow,
        out = out,
        pool = pool,
        max_retries = max_retries,
        timeout = timeout,
        queue_backing = queue_backing,
        error_log = isTRUE(dispatch_opts$error_log %||% FALSE),
        error_log_max_lines = dispatch_opts$error_log_max_lines %||% 100L
      )
    } else {
      if (!inherits(shards, "shard_descriptor")) {
        stop("dispatch_mode='shm_queue' requires shard_map(N, ...) or a shard_descriptor", call. = FALSE)
      }

      dispatch_result <- dispatch_shards_shm_queue_(
        n = as.integer(shards$num_shards %||% length(shards$shards)),
        block_size = 1L,
        shards = shards,
        fun = fun,
        borrow = borrow,
        out = out,
        pool = pool,
        max_retries = max_retries,
        timeout = timeout,
        queue_backing = queue_backing,
        error_log = isTRUE(dispatch_opts$error_log %||% FALSE),
        error_log_max_lines = dispatch_opts$error_log_max_lines %||% 100L
      )
    }

    results <- dispatch_result$results

    if (diagnostics) {
      diag$end_time <- Sys.time()
      diag$duration <- as.numeric(difftime(diag$end_time, diag$start_time, units = "secs"))
      diag$health_checks <- dispatch_result$diagnostics$health_checks %||% list()
      diag$shards_processed <- shards$num_shards
      diag$chunks_dispatched <- shards$num_shards
      diag$pool_stats <- dispatch_result$pool_stats
      diag$view_stats <- dispatch_result$diagnostics$view_stats %||% NULL
      diag$copy_stats <- dispatch_result$diagnostics$copy_stats %||% NULL
      diag$table_stats <- dispatch_result$diagnostics$table_stats %||% NULL
      diag$scratch_stats <- dispatch_result$diagnostics$scratch_stats %||% NULL
      diag$error_logs <- dispatch_result$diagnostics$error_logs %||% list()
      diag$shm_queue <- dispatch_result$diagnostics$taskq %||% NULL
    }

    return(structure(
      list(
        results = results,
        failures = dispatch_result$failures,
        shards = shards,
        diagnostics = diag,
        queue_status = dispatch_result$queue_status,
        pool_stats = dispatch_result$pool_stats,
        cow_policy = cow,
        profile = profile
      ),
      class = "shard_result"
    ))
  }
  }

  # Create self-contained executor function for workers
  chunk_executor <- make_chunk_executor(auto_table = auto_table)

  # Optional: online shard sizing autotune for scalar-N sharding. This is opt-in
  # by default for shard_map(N, ...) (low ceremony), and off for precomputed
  # shard descriptors.
  autotune_mode <- NULL
  autotune_cfg <- NULL
  if (is.list(autotune)) {
    autotune_mode <- as.character(autotune$mode %||% "online")
    autotune_cfg <- autotune
  } else if (!is.null(autotune)) {
    autotune_mode <- if (isTRUE(autotune)) "online" else as.character(autotune)
  } else if (shards_is_scalar_n) {
    autotune_mode <- "online"
  } else {
    autotune_mode <- "none"
  }
  if (!nzchar(autotune_mode)) autotune_mode <- "none"

  if (shards_is_scalar_n && identical(autotune_mode, "online")) {
    tuned <- shard_map_online_(
      n = n_items,
      fun = fun,
      borrow = borrow,
      out = out,
      kernel_meta = kernel_meta,
      chunk_executor = chunk_executor,
      auto_table = auto_table,
      auto_table_sink = auto_table_sink,
      auto_table_materialize = dispatch_opts$auto_table_materialize %||% "auto",
      auto_table_max_bytes = dispatch_opts$auto_table_max_bytes %||% (256 * 1024^2),
      pool = pool,
      workers = workers,
      mem_cap = mem_cap,
      chunk_size = chunk_size,
      autotune_cfg = autotune_cfg,
      profile = profile,
      diagnostics = diagnostics,
      diag = diag,
      health_check_interval = health_check_interval,
      max_retries = max_retries,
      timeout = timeout,
      scheduler_policy = scheduler_policy
    )
    shards <- tuned$shards
    dispatch_result <- tuned$dispatch_result
    if (diagnostics) diag <- tuned$diag
    results <- tuned$results
  } else {
    # Create chunk batches if chunk_size > 1
    if (shards_is_scalar_n) {
      shards <- shards(n_items, workers = workers)
    }
    chunks <- create_shard_chunks(shards, chunk_size, fun, borrow, out, kernel_meta = kernel_meta)

    # Dispatch chunks to workers with supervision
    dispatch_result <- dispatch_chunks(
      chunks = chunks,
      fun = chunk_executor,
      pool = pool,
      health_check_interval = health_check_interval,
      max_retries = max_retries,
      timeout = timeout,
      scheduler_policy = scheduler_policy,
      store_results = !auto_table
    )

    if (auto_table) {
      mat <- dispatch_opts$auto_table_materialize %||% "auto"
      mx <- dispatch_opts$auto_table_max_bytes %||% (256 * 1024^2)
      results <- table_finalize(auto_table_sink, materialize = mat, max_bytes = mx)
    } else {
      # Flatten results if chunk_size > 1
      results <- if (chunk_size > 1L) {
        unlist(dispatch_result$results, recursive = FALSE)
      } else {
        dispatch_result$results
      }
    }

    # Collect diagnostics
    if (diagnostics) {
      diag$end_time <- Sys.time()
      diag$duration <- as.numeric(difftime(diag$end_time, diag$start_time, units = "secs"))
      diag$health_checks <- dispatch_result$diagnostics$health_checks
      diag$shards_processed <- shards$num_shards
      diag$chunks_dispatched <- length(chunks)
      diag$pool_stats <- dispatch_result$pool_stats
      diag$view_stats <- dispatch_result$diagnostics$view_stats %||% NULL
      diag$view_hotspots <- dispatch_result$diagnostics$view_hotspots %||% list()
      diag$copy_stats <- dispatch_result$diagnostics$copy_stats %||% NULL
      diag$table_stats <- dispatch_result$diagnostics$table_stats %||% NULL
      diag$scratch_stats <- dispatch_result$diagnostics$scratch_stats %||% NULL
      diag$scheduler <- dispatch_result$diagnostics$scheduler %||% NULL
    }
  }

  # Build result object
  structure(
    list(
      results = results,
      failures = dispatch_result$failures,
      shards = shards,
      diagnostics = diag,
      queue_status = dispatch_result$queue_status,
      pool_stats = dispatch_result$pool_stats,
      cow_policy = cow,
      profile = profile
    ),
    class = "shard_result"
  )
}

# Online shard sizing for scalar-N shard_map.
#
# This runs a few small phases to pick a reasonable block_size using observed
# wall time and worker RSS, then processes the remainder with the chosen size.
shard_map_online_ <- function(n,
                              fun,
                              borrow,
                              out,
                              kernel_meta,
                              chunk_executor,
                              auto_table = FALSE,
                              auto_table_sink = NULL,
                              auto_table_materialize = "auto",
                              auto_table_max_bytes = 256 * 1024^2,
                              pool,
                              workers,
                              mem_cap,
                              chunk_size,
                              autotune_cfg = NULL,
                              profile,
                              diagnostics,
                              diag,
                              health_check_interval,
                              max_retries,
                              timeout,
                              scheduler_policy) {
  n <- as.integer(n)
  if (is.na(n) || n < 1L) stop("n must be >= 1", call. = FALSE)

  # Conservative defaults; keep user-facing ceremony low by being predictable.
  cfg <- list(
    max_rounds = 3L,
    probe_shards_per_worker = 4L,
    min_shard_time = 0.02, # seconds; below this, overhead dominates -> grow block
    grow_factor = 2.0,
    shrink_factor = 0.5,
    rss_hi = 0.85,
    rss_lo = 0.50
  )
  if (is.list(autotune_cfg)) {
    for (nm in names(cfg)) {
      if (!is.null(autotune_cfg[[nm]])) cfg[[nm]] <- autotune_cfg[[nm]]
    }
  }

  # Initial block size from the existing heuristic.
  bs <- autotune_block_size(
    n = n,
    workers = workers,
    min_shards_per_worker = 4L,
    max_shards_per_worker = 64L,
    scratch_bytes_per_item = 0,
    scratch_budget = 0
  )

  cursor <- 1L
  shard_id <- 1L
  all_shards <- list()
  all_results <- list()
  all_failures <- list()

  # Aggregate dispatch diagnostics across phases.
  agg_diag <- list(
    health_checks = list(),
    view_stats = list(created = 0L, materialized = 0L, materialized_bytes = 0, packed = 0L, packed_bytes = 0),
    view_hotspots = list(),
    copy_stats = list(borrow_exports = 0L, borrow_bytes = 0, buffer_writes = 0L, buffer_bytes = 0),
    table_stats = list(writes = 0L, rows = 0L, bytes = 0),
    scratch_stats = list(hits = 0L, misses = 0L, high_water = 0),
    scheduler = list(throttle_events = 0L),
    chunks_dispatched = 0L
  )

  hist <- list()
  rounds <- 0L

  # Use chunk_size=1 for probe phases so timing per shard is meaningful.
  probe_chunk_size <- 1L

  while (cursor <= n) {
    # Probe in early rounds; afterward, use the run's requested chunk_size.
    is_probe <- rounds < cfg$max_rounds
    use_chunk_size <- if (is_probe) probe_chunk_size else chunk_size

    # Probe only a small prefix; after tuning, take bigger bites.
    target_shards <- if (is_probe) workers * cfg$probe_shards_per_worker else workers * 32L
    phase_items <- min(n - cursor + 1L, as.integer(bs) * as.integer(max(target_shards, 1L)))
    if (phase_items < 1L) phase_items <- 1L
    phase_end <- min(cursor + phase_items - 1L, n)

    phase_shards <- create_contiguous_shards_window_(
      start = cursor,
      end = phase_end,
      block_size = bs,
      start_id = shard_id
    )
    shard_id <- shard_id + length(phase_shards)

    all_shards <- c(all_shards, phase_shards)
    phase_desc <- structure(
      list(
        n = phase_end - cursor + 1L,
        block_size = as.integer(bs),
        strategy = "contiguous",
        num_shards = length(phase_shards),
        shards = phase_shards
      ),
      class = "shard_descriptor"
    )

    chunks <- create_shard_chunks(phase_desc, use_chunk_size, fun, borrow, out, kernel_meta = kernel_meta)

    rss_before <- tryCatch(mem_report(pool)$peak_rss, error = function(e) NA_real_)
    t0 <- proc.time()[["elapsed"]]
    dr <- dispatch_chunks(
      chunks = chunks,
      fun = chunk_executor,
      pool = pool,
      health_check_interval = health_check_interval,
      max_retries = max_retries,
      timeout = timeout,
      scheduler_policy = scheduler_policy,
      store_results = !isTRUE(auto_table)
    )
    t1 <- proc.time()[["elapsed"]]
    rss_after <- tryCatch(mem_report(pool)$peak_rss, error = function(e) NA_real_)

    if (!isTRUE(auto_table)) {
      # Flatten phase results into per-shard results and append.
      phase_res <- if (use_chunk_size > 1L) unlist(dr$results, recursive = FALSE) else dr$results
      all_results <- c(all_results, phase_res)
    }
    if (length(dr$failures)) all_failures <- c(all_failures, dr$failures)

    # Update aggregate diagnostics.
    agg_diag$health_checks <- c(agg_diag$health_checks, dr$diagnostics$health_checks %||% list())
    agg_diag$chunks_dispatched <- agg_diag$chunks_dispatched + length(chunks)
    if (is.list(dr$diagnostics$view_stats)) {
      for (k in names(agg_diag$view_stats)) agg_diag$view_stats[[k]] <- (agg_diag$view_stats[[k]] %||% 0) + (dr$diagnostics$view_stats[[k]] %||% 0)
    }
    if (is.list(dr$diagnostics$view_hotspots) && length(dr$diagnostics$view_hotspots) > 0) {
      for (k in names(dr$diagnostics$view_hotspots)) {
        cur <- agg_diag$view_hotspots[[k]] %||% list(bytes = 0, count = 0L)
        cur$bytes <- (cur$bytes %||% 0) + (dr$diagnostics$view_hotspots[[k]]$bytes %||% 0)
        cur$count <- as.integer((cur$count %||% 0L) + (dr$diagnostics$view_hotspots[[k]]$count %||% 0L))
        agg_diag$view_hotspots[[k]] <- cur
      }
    }
    if (is.list(dr$diagnostics$copy_stats)) {
      for (k in names(agg_diag$copy_stats)) agg_diag$copy_stats[[k]] <- (agg_diag$copy_stats[[k]] %||% 0) + (dr$diagnostics$copy_stats[[k]] %||% 0)
    }
    if (is.list(dr$diagnostics$table_stats)) {
      for (k in names(agg_diag$table_stats)) agg_diag$table_stats[[k]] <- (agg_diag$table_stats[[k]] %||% 0) + (dr$diagnostics$table_stats[[k]] %||% 0)
    }
    if (is.list(dr$diagnostics$scratch_stats)) {
      agg_diag$scratch_stats$hits <- (agg_diag$scratch_stats$hits %||% 0L) + (dr$diagnostics$scratch_stats$hits %||% 0L)
      agg_diag$scratch_stats$misses <- (agg_diag$scratch_stats$misses %||% 0L) + (dr$diagnostics$scratch_stats$misses %||% 0L)
      agg_diag$scratch_stats$high_water <- max(as.double(agg_diag$scratch_stats$high_water %||% 0), as.double(dr$diagnostics$scratch_stats$high_water %||% 0))
    }
    if (is.list(dr$diagnostics$scheduler)) {
      agg_diag$scheduler$throttle_events <- (agg_diag$scheduler$throttle_events %||% 0L) + as.integer(dr$diagnostics$scheduler$throttle_events %||% 0L)
    }

    # Phase metrics
    elapsed <- as.double(t1 - t0)
    items_done <- as.integer(phase_end - cursor + 1L)
    shards_done <- length(phase_shards)
    throughput <- if (elapsed > 0) as.double(items_done) / elapsed else NA_real_
    shard_time <- if (shards_done > 0) elapsed / as.double(shards_done) else NA_real_

    rss_peak <- suppressWarnings(max(c(rss_before, rss_after), na.rm = TRUE))
    rss_frac <- if (is.finite(rss_peak) && is.finite(mem_cap) && mem_cap > 0) rss_peak / mem_cap else NA_real_

    if (is_probe) rounds <- rounds + 1L
    hist[[length(hist) + 1L]] <- list(
      round = rounds,
      start = cursor,
      end = phase_end,
      block_size = as.integer(bs),
      chunk_size = as.integer(use_chunk_size),
      elapsed_sec = elapsed,
      items = items_done,
      shards = shards_done,
      throughput_items_per_sec = throughput,
      shard_time_sec = shard_time,
      rss_peak = rss_peak,
      rss_fraction_of_mem_cap = rss_frac
    )

    cursor <- phase_end + 1L

    # Update block_size for next probe phase using simple, safe heuristics.
    if (is_probe && cursor <= n) {
      bs_next <- bs
      reason <- "keep"

      # If we get close to mem_cap, shrink.
      if (is.finite(rss_frac) && rss_frac >= cfg$rss_hi) {
        bs_next <- max(as.integer(floor(as.double(bs) * cfg$shrink_factor)), 1L)
        reason <- "shrink_rss"
      } else if (is.finite(shard_time) && shard_time < cfg$min_shard_time) {
        # If shards are too tiny (overhead dominates), grow.
        bs_next <- as.integer(ceiling(as.double(bs) * cfg$grow_factor))
        bs_next <- min(bs_next, n)
        reason <- "grow_overhead"
      } else if (is.finite(rss_frac) && rss_frac <= cfg$rss_lo && is.finite(shard_time) && shard_time < (cfg$min_shard_time * 0.5)) {
        # Extra nudge: very low RSS and very small shard time.
        bs_next <- as.integer(ceiling(as.double(bs) * cfg$grow_factor))
        bs_next <- min(bs_next, n)
        reason <- "grow_low_rss"
      }

      # Record decision.
      hist[[length(hist)]]$decision <- reason
      hist[[length(hist)]]$next_block_size <- as.integer(bs_next)
      bs <- bs_next
    }
  }

  full_desc <- structure(
    list(
      n = n,
      block_size = NA_integer_,
      strategy = "contiguous",
      num_shards = length(all_shards),
      shards = all_shards
    ),
    class = "shard_descriptor"
  )

  # Produce a unified dispatch_result-like payload.
  vh <- agg_diag$view_hotspots %||% list()
  if (length(vh) > 0) {
    ord <- order(vapply(vh, function(x) as.double(x$bytes %||% 0), numeric(1)), decreasing = TRUE)
    vh <- vh[ord]
    if (length(vh) > 20) vh <- vh[seq_len(20)]
  }

  dispatch_result <- structure(
    list(
      results = all_results,
      failures = all_failures,
      queue_status = list(
        total = agg_diag$chunks_dispatched,
        pending = 0L,
        in_flight = 0L,
        completed = agg_diag$chunks_dispatched - length(all_failures),
        failed = length(all_failures),
        total_retries = sum(vapply(all_failures, function(x) x$retry_count %||% 0L, integer(1)), na.rm = TRUE)
      ),
      diagnostics = list(
        health_checks = agg_diag$health_checks,
        view_stats = agg_diag$view_stats,
        view_hotspots = vh,
        copy_stats = agg_diag$copy_stats,
        table_stats = agg_diag$table_stats,
        scratch_stats = agg_diag$scratch_stats,
        scheduler = agg_diag$scheduler
      ),
      pool_stats = pool_get()$stats
    ),
    class = "shard_dispatch_result"
  )

  if (diagnostics) {
    diag$end_time <- Sys.time()
    diag$duration <- as.numeric(difftime(diag$end_time, diag$start_time, units = "secs"))
    diag$health_checks <- agg_diag$health_checks
    diag$shards_processed <- full_desc$num_shards
    diag$chunks_dispatched <- agg_diag$chunks_dispatched
    diag$pool_stats <- dispatch_result$pool_stats
    diag$view_stats <- agg_diag$view_stats
    diag$view_hotspots <- vh
    diag$copy_stats <- agg_diag$copy_stats
    diag$table_stats <- agg_diag$table_stats
    diag$scratch_stats <- agg_diag$scratch_stats
    diag$scheduler <- agg_diag$scheduler
    diag$autotune <- list(mode = "online", history = hist)
  }

  list(
    shards = full_desc,
    results = if (isTRUE(auto_table)) {
      if (is.null(auto_table_sink)) stop("auto_table enabled but auto_table_sink is NULL", call. = FALSE)
      table_finalize(auto_table_sink, materialize = auto_table_materialize, max_bytes = auto_table_max_bytes)
    } else {
      all_results
    },
    dispatch_result = dispatch_result,
    diag = diag
  )
}

#' Get Profile Settings
#'
#' Returns settings based on execution profile.
#'
#' @param profile Profile name.
#' @param mem_cap User-specified memory cap.
#' @param recycle User-specified recycle setting.
#' @return List of settings.
#' @keywords internal
#' @noRd
get_profile_settings <- function(profile, mem_cap, recycle) {
  settings <- list(
    mem_cap = parse_bytes(mem_cap),
    rss_drift_threshold = if (is.numeric(recycle)) recycle else 0.5,
    health_check_interval = 10L
  )

  switch(profile,
    "memory" = {
      settings$rss_drift_threshold <- 0.25  # More aggressive recycling
      settings$health_check_interval <- 5L
    },
    "speed" = {
      settings$rss_drift_threshold <- 1.0  # Less recycling
      settings$health_check_interval <- 50L
    }
  )

  if (isFALSE(recycle)) {
    settings$rss_drift_threshold <- Inf  # Disable recycling
  }

  settings
}

#' Ensure Pool Exists
#'
#' Creates or validates worker pool.
#'
#' @param workers Number of workers.
#' @param mem_cap Memory cap in bytes.
#' @param rss_drift_threshold Drift threshold.
#' @param packages Packages to load.
#' @param init_expr Init expression.
#' @return Pool object.
#' @keywords internal
#' @noRd
ensure_pool <- function(workers, mem_cap, rss_drift_threshold, packages, init_expr) {
  pool <- pool_get()

  # Check if existing pool is suitable
  if (!is.null(pool)) {
    if (pool$n == workers) {
      return(pool)
    }
    # Pool exists but wrong size - stop it
    pool_stop()
  }

  # Create new pool
  pool_create(
    n = workers,
    rss_limit = mem_cap,
    rss_drift_threshold = rss_drift_threshold,
    packages = packages,
    init_expr = init_expr
  )

  pool_get()
}

#' Validate Borrowed Inputs
#'
#' Validates and prepares borrowed inputs.
#'
#' @param borrow List of inputs.
#' @param cow COW policy.
#' @return Validated borrow list.
#' @keywords internal
#' @noRd
validate_borrow <- function(borrow, cow) {
  if (length(borrow) == 0) return(borrow)

  if (!is.list(borrow) || is.null(names(borrow))) {
    stop("borrow must be a named list", call. = FALSE)
  }

  if (any(names(borrow) == "")) {
    stop("All borrowed inputs must be named", call. = FALSE)
  }

  auto_shared <- character(0)

  # Auto-share large atomic inputs once in the main process so PSOCK workers
  # can receive a small descriptor (via ALTREP serialization) instead of full
  # data copies.
  for (name in names(borrow)) {
    x <- borrow[[name]]

    if (is.atomic(x) && !is.null(x) &&
        typeof(x) %in% c("double", "integer", "logical", "raw") &&
        !is_shared_vector(x)) {
      # Build with cow='allow' so we can attach attributes, then lock down to
      # the requested policy.
      shared <- as_shared(x, readonly = TRUE, backing = "auto", cow = "allow")

      # Preserve non-class attributes (dim, dimnames, names, tsp, etc).
      attrs <- attributes(x)
      x_class <- attr(x, "class")
      attrs$class <- NULL
      if (length(attrs)) {
        for (nm in names(attrs)) {
          attr(shared, nm) <- attrs[[nm]]
        }
      }
      class(shared) <- unique(c("shard_shared_vector", x_class))
      attr(shared, "shard_cow") <- cow

      borrow[[name]] <- shared
      x <- shared
      auto_shared <- c(auto_shared, name)
    }

    # Best-effort tag for downstream diagnostics.
    if (!is_shared_vector(x)) {
      attr(borrow[[name]], "shard_cow") <- cow
    } else {
      existing <- attr(x, "shard_cow", exact = TRUE)
      if (!is.null(existing) && is.character(existing) &&
          length(existing) == 1L && !identical(existing, cow)) {
        warning("Borrowed input '", name, "' has shard_cow='", existing,
                "' but shard_map(cow='", cow, "') was requested. Using '",
                existing, "'.", call. = FALSE)
      }
    }
  }

  attr(borrow, "auto_shared") <- auto_shared
  borrow
}

#' Validate Output Buffers
#'
#' Validates output buffer specifications.
#'
#' @param out List of output buffers.
#' @return Validated out list.
#' @keywords internal
#' @noRd
validate_out <- function(out) {
  if (length(out) == 0) return(out)

  if (!is.list(out) || is.null(names(out))) {
    stop("out must be a named list", call. = FALSE)
  }

  if (any(names(out) == "")) {
    stop("All output buffers must be named", call. = FALSE)
  }

  bad <- vapply(out, function(x) {
    !(inherits(x, "shard_buffer") ||
        inherits(x, "shard_table_buffer") ||
        inherits(x, "shard_table_sink"))
  }, logical(1))
  if (any(bad)) {
    stop("All outputs must be shard_buffer, shard_table_buffer, or shard_table_sink objects.",
         call. = FALSE)
  }

  out
}

#' Set Worker Seeds
#'
#' Sets reproducible RNG seeds in workers.
#'
#' @param pool Worker pool.
#' @param seed Base seed.
#' @param num_shards Number of shards for substream calculation.
#' @keywords internal
#' @noRd
set_worker_seeds <- function(pool, seed, num_shards) {
  for (i in seq_len(pool$n)) {
    worker_seed <- seed + (i - 1L) * num_shards
    tryCatch({
      parallel::clusterCall(pool$workers[[i]]$cluster, function(s) {
        set.seed(s)
      }, worker_seed)
    }, error = function(e) NULL)
  }
}

#' Export Borrowed Inputs to Workers
#'
#' Exports borrowed data to all workers (once, reused across shards).
#'
#' @param pool Worker pool.
#' @param borrow List of borrowed inputs.
#' @keywords internal
#' @noRd
export_borrow_to_workers <- function(pool, borrow) {
  if (length(borrow) == 0) return(invisible(NULL))

  # Create an environment with the borrowed data
  export_env <- new.env(parent = emptyenv())
  export_env$.shard_borrow <- borrow

  # Export to all workers
  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    if (!is.null(w) && worker_is_alive(w)) {
      tryCatch({
        parallel::clusterExport(w$cluster, ".shard_borrow", envir = export_env)
      }, error = function(e) {
        warning("Failed to export borrow to worker ", i, ": ", conditionMessage(e))
      })
    }
  }

  invisible(NULL)
}

#' Export Output Buffers to Workers
#'
#' Exports output buffer references to workers.
#'
#' @param pool Worker pool.
#' @param out List of output buffers.
#' @keywords internal
#' @noRd
export_out_to_workers <- function(pool, out) {
  if (length(out) == 0) return(invisible(NULL))

  # Export reopenable descriptors rather than shard_buffer objects. The raw
  # segment externalptr does not survive PSOCK serialization.
  describe_one <- function(obj) {
    if (inherits(obj, "shard_buffer")) {
      info <- buffer_info(obj)
      return(list(
        kind = "buffer",
        path = info$path,
        backing = info$backing,
        type = info$type,
        dim = info$dim
      ))
    }

    if (inherits(obj, "shard_table_buffer")) {
      cols <- lapply(obj$columns, function(buf) {
        info <- buffer_info(buf)
        list(path = info$path, backing = info$backing, type = info$type, dim = info$dim)
      })
      return(list(
        kind = "table_buffer",
        schema = obj$schema,
        nrow = obj$nrow,
        backing = obj$backing,
        columns = cols
      ))
    }

    if (inherits(obj, "shard_table_sink")) {
      return(list(
        kind = "table_sink",
        schema = obj$schema,
        mode = obj$mode,
        path = obj$path,
        format = obj$format
      ))
    }

    stop("Unsupported out object type", call. = FALSE)
  }

  out_desc <- lapply(out, describe_one)

  export_env <- new.env(parent = emptyenv())
  export_env$.shard_out <- out_desc

  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    if (!is.null(w) && worker_is_alive(w)) {
      tryCatch({
        parallel::clusterExport(w$cluster, ".shard_out", envir = export_env)
      }, error = function(e) {
        warning("Failed to export out to worker ", i, ": ", conditionMessage(e))
      })
    }
  }

  invisible(NULL)
}

export_auto_table_sink_to_workers <- function(pool, sink) {
  if (is.null(sink)) return(invisible(NULL))
  if (!inherits(sink, "shard_table_sink")) {
    stop("sink must be a shard_table_sink", call. = FALSE)
  }

  export_env <- new.env(parent = emptyenv())
  export_env$.shard_auto_table_sink <- list(
    schema = sink$schema,
    mode = sink$mode,
    path = sink$path,
    format = sink$format
  )

  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    if (!is.null(w) && worker_is_alive(w)) {
      tryCatch({
        parallel::clusterExport(w$cluster, ".shard_auto_table_sink", envir = export_env)
      }, error = function(e) {
        warning("Failed to export auto table sink to worker ", i, ": ", conditionMessage(e))
      })
    }
  }

  invisible(NULL)
}

reset_worker_diagnostics_ <- function(pool) {
  # Best-effort: reset per-process counters so run telemetry is clean and
  # attribution (e.g. view materialization hotspots) isn't polluted by previous runs.
  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    if (!is.null(w) && worker_is_alive(w)) {
      tryCatch({
        parallel::clusterCall(w$cluster, function() {
          f1 <- tryCatch(get("view_reset_diagnostics", asNamespace("shard")), error = function(e) NULL)
          f2 <- tryCatch(get("buffer_reset_diagnostics", asNamespace("shard")), error = function(e) NULL)
          f3 <- tryCatch(get("table_reset_diagnostics", asNamespace("shard")), error = function(e) NULL)
          f4 <- tryCatch(get("scratch_reset_diagnostics", asNamespace("shard")), error = function(e) NULL)
          if (is.function(f1)) tryCatch(f1(), error = function(e) NULL)
          if (is.function(f2)) tryCatch(f2(), error = function(e) NULL)
          if (is.function(f3)) tryCatch(f3(), error = function(e) NULL)
          if (is.function(f4)) tryCatch(f4(), error = function(e) NULL)
          if (exists(".shard_view_hotspot_snapshot", envir = .shard_worker_env, inherits = FALSE)) {
            rm(".shard_view_hotspot_snapshot", envir = .shard_worker_env)
          }
          NULL
        })
      }, error = function(e) NULL)
    }
  }
  invisible(NULL)
}

#' Create Shard Chunks
#'
#' Groups shards into chunks for dispatch. Each chunk contains a self-contained
#' executor function that can run in a worker without needing package functions.
#'
#' @param shards Shard descriptor.
#' @param chunk_size Shards per chunk.
#' @param fun User function.
#' @param borrow Borrowed inputs.
#' @param out Output buffers.
#' @return List of chunk descriptors.
#' @keywords internal
#' @noRd
create_shard_chunks <- function(shards, chunk_size, fun, borrow, out, kernel_meta = NULL) {
  chunk_size <- max(as.integer(chunk_size), 1L)
  num_chunks <- ceiling(shards$num_shards / chunk_size)

  chunks <- vector("list", num_chunks)

  borrow_names <- names(borrow)
  out_names <- names(out)

  classify_bytes <- function(bytes) {
    bytes <- as.double(bytes)
    if (!is.finite(bytes) || is.na(bytes)) return("tiny")
    if (bytes >= 64 * 1024^2) return("huge")
    if (bytes >= 8 * 1024^2) return("medium")
    "tiny"
  }

  for (i in seq_len(num_chunks)) {
    start_idx <- (i - 1L) * chunk_size + 1L
    end_idx <- min(i * chunk_size, shards$num_shards)

    chunk_shards <- shards$shards[start_idx:end_idx]

    # Optional footprint hint for memory-aware scheduling.
    fp_class <- NULL
    fp_bytes <- NULL
    if (!is.null(kernel_meta) && !is.null(kernel_meta$footprint)) {
      fp <- kernel_meta$footprint
      if (is.numeric(fp) && length(fp) == 1L) {
        fp_bytes <- as.double(fp)
        fp_class <- classify_bytes(fp_bytes)
      } else if (is.function(fp)) {
        vals <- lapply(chunk_shards, function(s) {
          tryCatch(fp(s), error = function(e) NULL)
        })
        # Accept either numeric bytes or list(class=..., bytes=...).
        bytes <- vapply(vals, function(v) {
          if (is.null(v)) return(NA_real_)
          if (is.numeric(v)) return(as.double(v[[1]]))
          if (is.list(v) && !is.null(v$bytes)) return(as.double(v$bytes))
          NA_real_
        }, numeric(1))
        fp_bytes <- suppressWarnings(max(bytes, na.rm = TRUE))
        if (!is.finite(fp_bytes)) fp_bytes <- NULL
        cls <- vapply(vals, function(v) {
          if (is.list(v) && !is.null(v$class)) as.character(v$class) else NA_character_
        }, character(1))
        cls <- cls[!is.na(cls) & nzchar(cls)]
        fp_class <- if (length(cls) > 0) cls[[1]] else if (!is.null(fp_bytes)) classify_bytes(fp_bytes) else NULL
      }
    }

    chunks[[i]] <- list(
      id = i,
      shard_ids = start_idx:end_idx,
      shards = chunk_shards,
      fun = fun,
      borrow_names = borrow_names,
      out_names = out_names,
      footprint_class = fp_class,
      footprint_bytes = fp_bytes
    )
  }

  chunks
}

#' Create Chunk Executor Function
#'
#' Creates a self-contained function that can execute a chunk in a worker.
#' This function is passed to dispatch_chunks and runs entirely within
#' the worker process.
#'
#' @return A function that executes chunks.
#' @keywords internal
#' @noRd
make_chunk_executor <- function(auto_table = FALSE) {
  out_desc_key_ <- function(d) {
    # A stable identifier for deciding whether a cached out handle can be reused.
    kind <- d$kind %||% "buffer"
    if (identical(kind, "buffer")) {
      return(paste0("buffer|", d$path, "|", d$backing, "|", d$type, "|", paste(d$dim, collapse = "x")))
    }
    if (identical(kind, "table_buffer")) {
      col_keys <- vapply(d$columns, function(cd) {
        paste0(cd$path, "|", cd$backing, "|", cd$type, "|", paste(cd$dim, collapse = "x"))
      }, character(1))
      return(paste0("table_buffer|", paste(names(col_keys), col_keys, sep = "=", collapse = ";")))
    }
    if (identical(kind, "table_sink")) {
      return(paste0("table_sink|", d$path, "|", d$format, "|", d$mode))
    }
    paste0("unknown|", kind)
  }

  open_out_one_ <- function(d) {
    kind <- d$kind %||% "buffer"
    if (identical(kind, "buffer")) {
      buf <- buffer_open(
        path = d$path,
        type = d$type,
        dim = d$dim,
        backing = d$backing,
        readonly = FALSE
      )
      return(buf)
    }
    if (identical(kind, "table_buffer")) {
      cols <- list()
      for (cn in names(d$columns)) {
        cd <- d$columns[[cn]]
        cols[[cn]] <- buffer_open(
          path = cd$path,
          type = cd$type,
          dim = cd$dim,
          backing = cd$backing,
          readonly = FALSE
        )
      }
      tb <- structure(
        list(schema = d$schema, nrow = as.integer(d$nrow), backing = d$backing, columns = cols),
        class = "shard_table_buffer"
      )
      return(tb)
    }
    if (identical(kind, "table_sink")) {
      ts <- structure(
        list(schema = d$schema, mode = d$mode, path = d$path, format = d$format),
        class = "shard_table_sink"
      )
      return(ts)
    }
    stop("Unsupported out descriptor kind: ", kind, call. = FALSE)
  }

  close_out_one_ <- function(obj) {
    if (inherits(obj, "shard_buffer")) {
      tryCatch(buffer_close(obj), error = function(e) NULL)
      return(invisible(NULL))
    }
    if (inherits(obj, "shard_table_buffer")) {
      if (!is.null(obj$columns) && is.list(obj$columns)) {
        for (col in obj$columns) {
          if (inherits(col, "shard_buffer")) {
            tryCatch(buffer_close(col), error = function(e) NULL)
          }
        }
      }
      return(invisible(NULL))
    }
    invisible(NULL)
  }

  # This function runs inside workers
  function(chunk) {
    # Get borrowed inputs and outputs from worker environment
    borrow <- if (exists(".shard_borrow", envir = globalenv())) {
      get(".shard_borrow", envir = globalenv())
    } else {
      list()
    }

    out_desc <- if (exists(".shard_out", envir = globalenv())) {
      get(".shard_out", envir = globalenv())
    } else {
      list()
    }

    # Lazily open output buffers once per worker process and cache them.
    out <- list()
    if (length(out_desc) > 0) {
      if (!exists(".shard_out_opened", envir = .shard_worker_env, inherits = FALSE)) {
        assign(".shard_out_opened", new.env(parent = emptyenv()), envir = .shard_worker_env)
      }
      opened <- get(".shard_out_opened", envir = .shard_worker_env)

      for (nm in names(out_desc)) {
        d <- out_desc[[nm]]
        want_key <- out_desc_key_(d)
        entry <- if (exists(nm, envir = opened, inherits = FALSE)) opened[[nm]] else NULL
        cur_key <- NULL
        cur_obj <- NULL
        if (!is.null(entry) && is.list(entry) && !is.null(entry$key) && !is.null(entry$obj)) {
          cur_key <- entry$key
          cur_obj <- entry$obj
        } else if (!is.null(entry)) {
          cur_obj <- entry
        }

        if (is.null(cur_obj) || !identical(cur_key, want_key)) {
          if (!is.null(cur_obj)) close_out_one_(cur_obj)
          new_obj <- open_out_one_(d)
          opened[[nm]] <- list(key = want_key, obj = new_obj)
        }

        out[[nm]] <- opened[[nm]]$obj
      }
    }

    # Get the user function from the chunk
    fun <- chunk$fun
    borrow_names <- chunk$borrow_names
    out_names <- chunk$out_names

    if (isTRUE(auto_table)) {
      # Low-ceremony table outputs: if the user function returns a data.frame,
      # write it as a row-group partition (one file per shard id). This avoids
      # building a giant list + bind_rows() on the master.
      if (!exists(".shard_auto_table_sink", envir = globalenv(), inherits = FALSE)) {
        stop("auto_table is enabled but no auto table sink is available in the worker", call. = FALSE)
      }
      d <- get(".shard_auto_table_sink", envir = globalenv(), inherits = FALSE)
      sink <- structure(
        list(schema = d$schema, mode = d$mode, path = d$path, format = d$format),
        class = "shard_table_sink"
      )

      for (shard in chunk$shards) {
        args <- list(shard)
        for (name in borrow_names) args[[name]] <- borrow[[name]]
        for (name in out_names) args[[name]] <- out[[name]]

        val <- do.call(fun, args, quote = TRUE)
        if (is.null(val)) next
        if (!is.data.frame(val)) {
          stop("auto_table requires fun() to return a data.frame/tibble (or NULL) for all shards", call. = FALSE)
        }
        sid <- as.integer(shard$id %||% NA_integer_)
        if (is.na(sid) || sid < 1L) stop("Invalid shard id for table_write()", call. = FALSE)
        table_write(sink, sid, val)
      }
      return(NULL)
    }

    # Execute for each shard in the chunk (return values gathered to master).
    lapply(chunk$shards, function(shard) {
      args <- list(shard)
      for (name in borrow_names) args[[name]] <- borrow[[name]]
      for (name in out_names) args[[name]] <- out[[name]]
      # `do.call()` has a sharp edge: if an argument value is a language object,
      # it will be spliced into the call and evaluated (surprising for "data"
      # being passed through borrow/out). Using quote=TRUE ensures language
      # objects are passed as values, not executed as code.
      do.call(fun, args, quote = TRUE)
    })
  }
}

#' Print a shard_result Object
#'
#' @param x A \code{shard_result} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @export
#' @examples
#' \donttest{
#' result <- shard_map(4L, function(shard) shard$idx, workers = 2)
#' pool_stop()
#' print(result)
#' }
print.shard_result <- function(x, ...) {
  cat("shard_map result\n")

  if (!is.null(x$diagnostics)) {
    cat("  Duration:", sprintf("%.2f seconds", x$diagnostics$duration), "\n")
    cat("  Shards:", x$diagnostics$shards_processed, "\n")
    cat("  Chunks:", x$diagnostics$chunks_dispatched, "\n")
  }

  status <- x$queue_status
  if (!is.null(status)) {
    cat("  Completed:", status$completed,
        if (status$failed > 0) paste0(" (", status$failed, " failed)") else "", "\n")
    if (status$total_retries > 0) {
      cat("  Retries:", status$total_retries, "\n")
    }
  }

  if (!is.null(x$pool_stats)) {
    cat("  Worker recycles:", x$pool_stats$total_recycles, "\n")
    cat("  Worker deaths:", x$pool_stats$total_deaths, "\n")
  }

  cat("  Profile:", x$profile, "\n")
  cat("  COW policy:", x$cow_policy, "\n")

  if (length(x$failures) > 0) {
    cat("  Failed shards:", length(x$failures), "\n")
  }

  invisible(x)
}

#' Extract Results from shard_map
#'
#' @param x A shard_result object.
#' @param flatten Logical. Flatten nested results?
#' @return List or vector of results.
#' @export
#' @examples
#' \donttest{
#' result <- shard_map(4L, function(shard) shard$idx[[1L]], workers = 2)
#' pool_stop()
#' results(result)
#' }
results <- function(x, flatten = TRUE) {
  if (!inherits(x, "shard_result")) {
    stop("x must be a shard_result object", call. = FALSE)
  }

  res <- x$results

  if (inherits(res, c("shard_row_groups", "shard_dataset", "shard_table_handle"))) {
    return(res)
  }
  if (is.data.frame(res)) {
    return(res)
  }

  if (inherits(res, "shard_results_placeholder")) {
    # Avoid unlist() on a placeholder (would allocate enormous objects).
    return(res)
  }

  if (!is.list(res)) {
    # Unusual but allowed (e.g., auto-materialized scalar results).
    return(res)
  }

  if (flatten && length(res) > 0) {
    # Try to simplify
    tryCatch(
      unlist(res, recursive = FALSE),
      error = function(e) res
    )
  } else {
    res
  }
}

#' Check if shard_map Succeeded
#'
#' @param x A shard_result object.
#' @return Logical. TRUE if no failures.
#' @export
#' @examples
#' \donttest{
#' result <- shard_map(4L, function(shard) shard$idx[[1L]], workers = 2)
#' pool_stop()
#' succeeded(result)
#' }
succeeded <- function(x) {
  if (inherits(x, "shard_result") || inherits(x, "shard_reduce_result")) {
    return(length(x$failures) == 0)
  }
  stop("x must be a shard_result or shard_reduce_result object", call. = FALSE)
}

test_that("shard_map gather views use a BLAS-3 packed path with zero view materialization", {
  skip_on_cran()

  pool_stop()
  shard:::view_reset_diagnostics()

  set.seed(1)
  n <- 250L
  p <- 10L
  v <- 60L

  X <- matrix(rnorm(n * p), nrow = n)
  Y <- matrix(rnorm(n * v), nrow = n)

  colnames(X) <- paste0("x", seq_len(ncol(X)))
  colnames(Y) <- paste0("y", seq_len(ncol(Y)))

  Xsh <- share(X, backing = "mmap")
  Ysh <- share(Y, backing = "mmap")

  # Each shard is a "searchlight": a small, non-contiguous set of columns.
  idxs <- replicate(12, sort(sample.int(ncol(Y), 7)), simplify = FALSE)
  shards_g <- shards_list(idxs)

  res <- shard_map(
    shards_g,
    borrow = list(X = Xsh, Y = Ysh),
    fun = function(shard, X, Y) {
      vY <- view(Y, cols = shard$idx, type = "auto")
      shard:::view_xTy(X, vY)
    },
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))

  cr <- copy_report(res)
  expect_gt(cr$view_created %||% 0L, 0L)
  expect_equal(cr$view_materialized %||% 0L, 0L)
  expect_equal(cr$view_materialized_bytes %||% 0, 0)
  expect_gt(cr$view_packed_bytes %||% 0, 0)

  # Gather packing reuses worker-local scratch buffers.
  expect_true(is.list(res$diagnostics$scratch_stats))
  expect_gt(res$diagnostics$scratch_stats$misses %||% 0L, 0L)
  expect_gt(res$diagnostics$scratch_stats$hits %||% 0L, 0L)

  mats <- results(res)
  for (i in seq_along(mats)) {
    idx <- idxs[[i]]
    expect_equal(mats[[i]], crossprod(X, Y[, idx, drop = FALSE]), tolerance = 1e-10)
  }

  pool_stop()
})

test_that("kernel registry supports dispatch via shard_map(kernel=)", {
  skip_on_cran()

  pool_stop()

  # Built-ins should be registered at load time.
  expect_true("crossprod_tile" %in% list_kernels())
  expect_true("col_means" %in% list_kernels())
  expect_true("col_vars" %in% list_kernels())

  # Register a tiny custom kernel.
  register_kernel(
    name = "test_kernel_id",
    impl = function(sh, ...) sh$id,
    footprint = function(sh) list(class = "tiny", bytes = 8),
    supports_views = FALSE
  )
  expect_true("test_kernel_id" %in% list_kernels())

  res <- shard_map(shards(10, block_size = 1), kernel = "test_kernel_id", workers = 2, diagnostics = TRUE)
  expect_true(succeeded(res))
  expect_equal(unname(unlist(results(res))), 1:10)

  # Diagnostics should record the chosen kernel.
  expect_equal(res$diagnostics$kernel %||% NULL, "test_kernel_id")

  pool_stop()
})

test_that("shard_map errors clearly on unknown kernel", {
  pool_stop()
  pool_create(1)
  on.exit(pool_stop(), add = TRUE)

  expect_error(
    shard_map(shards(2, block_size = 1), kernel = "no_such_kernel", workers = 1),
    "Unknown kernel"
  )
})

test_that("register_kernel validates name argument", {
  expect_error(register_kernel("", impl = function(x) x),
               "name must be a non-empty string")
})

test_that("register_kernel validates impl argument", {
  expect_error(register_kernel("test", impl = "not_a_function"),
               "impl must be a function")
})

test_that("register_kernel validates signature argument", {
  expect_error(register_kernel("test", impl = function(x) x, signature = 123),
               "signature must be a string or NULL")
})

test_that("register_kernel validates description argument", {
  expect_error(register_kernel("test", impl = function(x) x, description = 123),
               "description must be a string or NULL")
})

test_that("register_kernel validates footprint argument", {
  expect_error(register_kernel("test", impl = function(x) x, footprint = "invalid"),
               "footprint must be a function, numeric, or NULL")
})

test_that("register_kernel accepts valid footprint types", {
  # Numeric footprint
  meta1 <- register_kernel("test_numeric_fp", impl = function(x) x, footprint = 1024)
  expect_equal(meta1$footprint, 1024)

  # Function footprint
  fp_fn <- function(sh) list(class = "tiny", bytes = 8)
  meta2 <- register_kernel("test_fn_fp", impl = function(x) x, footprint = fp_fn)
  expect_true(is.function(meta2$footprint))

  # NULL footprint
  meta3 <- register_kernel("test_null_fp", impl = function(x) x, footprint = NULL)
  expect_null(meta3$footprint)
})

test_that("get_kernel returns registered kernel", {
  register_kernel("test_get", impl = function(x) x * 2, description = "test")

  k <- shard:::get_kernel("test_get")
  expect_true(is.list(k))
  expect_equal(k$name, "test_get")
  expect_true(is.function(k$impl))
  expect_equal(k$description, "test")
})

test_that("get_kernel returns NULL for unregistered kernel", {
  k <- shard:::get_kernel("definitely_not_registered_kernel_xyz")
  expect_null(k)
})

test_that("col_means kernel computes correct results", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Create test data
  n_rows <- 100
  n_cols <- 20
  X <- share(matrix(rnorm(n_rows * n_cols), nrow = n_rows, ncol = n_cols))
  out <- buffer("double", dim = n_cols, init = 0)

  # Use the col_means kernel
  res <- shard_map(
    shards(n_cols, block_size = 5),
    kernel = "col_means",
    borrow = list(X = X),
    out = list(out = out),
    workers = 2
  )

  expect_true(succeeded(res))
  expect_equal(as.numeric(out[]), colMeans(as.matrix(X)), tolerance = 1e-10)
})

test_that("col_vars kernel computes correct results", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Create test data
  n_rows <- 100
  n_cols <- 20
  X <- share(matrix(rnorm(n_rows * n_cols), nrow = n_rows, ncol = n_cols))
  out <- buffer("double", dim = n_cols, init = 0)

  # Use the col_vars kernel
  res <- shard_map(
    shards(n_cols, block_size = 5),
    kernel = "col_vars",
    borrow = list(X = X),
    out = list(out = out),
    workers = 2
  )

  expect_true(succeeded(res))

  # Calculate expected using apply
  expected_vars <- apply(as.matrix(X), 2, var)
  expect_equal(as.numeric(out[]), expected_vars, tolerance = 1e-10)
})

test_that("register_kernel records supports_views flag", {
  meta1 <- register_kernel("test_views_true", impl = function(x) x, supports_views = TRUE)
  expect_true(meta1$supports_views)

  meta2 <- register_kernel("test_views_false", impl = function(x) x, supports_views = FALSE)
  expect_false(meta2$supports_views)

  # Default should be TRUE
  meta3 <- register_kernel("test_views_default", impl = function(x) x)
  expect_true(meta3$supports_views)
})

# Tests for large matrix + multicore interaction
#
# These tests verify correctness when parallel workers operate on matrices
# large enough to exercise real-world memory/BLAS patterns that small toy
# matrices miss:
#   - shard_map with borrowed large matrices vs single-threaded baseline
#   - shard_crossprod on 1000+ column matrices
#   - shard_reduce streaming aggregation over large shared data
#   - shard_apply_matrix column-parallel correctness
#   - Workers=1 vs Workers=N equivalence (determinism check)
#   - Buffer output correctness under concurrent tile writes

# ---------------------------------------------------------------------------
# helpers
# ---------------------------------------------------------------------------
skip_if_not_large <- function() {
  if (!identical(tolower(Sys.getenv("SHARD_RUN_LARGE", "false")), "true")) {
    skip("Set SHARD_RUN_LARGE=true to run large-matrix multicore tests")
  }
}

# ---------------------------------------------------------------------------
# shard_map: large borrowed matrix, parallel vs serial baseline
# ---------------------------------------------------------------------------
test_that("shard_map: column sums on 1000x500 matrix match base R", {

  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(42)
  X <- matrix(rnorm(1000 * 500), nrow = 1000, ncol = 500)
  expected <- colSums(X)

  blocks <- shards(ncol(X), block_size = 50, workers = 2)

  result <- shard_map(blocks,
    borrow = list(X = X),
    fun = function(shard, X) {
      colSums(X[, shard$idx, drop = FALSE])
    },
    workers = 2
  )

  expect_true(succeeded(result))
  res <- unlist(results(result))
  expect_equal(unname(res), unname(expected), tolerance = 1e-12)
})

test_that("shard_map: row-wise crossprod slices on 2000x200 matrix match base R", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(7)
  n <- 2000L
  p <- 200L
  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  expected <- crossprod(X)  # p x p

  # Partition rows; each shard computes crossprod of its row slice
  blocks <- shards(n, block_size = 250, workers = 2)

  result <- shard_map(blocks,
    borrow = list(X = X),
    fun = function(shard, X) {
      crossprod(X[shard$idx, , drop = FALSE])
    },
    workers = 2
  )

  expect_true(succeeded(result))

  # Sum the partial crossprods
  partials <- results(result)
  accumulated <- Reduce(`+`, partials)
  expect_equal(accumulated, expected, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# shard_crossprod: larger dimensions
# ---------------------------------------------------------------------------
test_that("shard_crossprod: 500x300 x 500x200 matches base crossprod", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(99)
  n <- 500L; px <- 300L; py <- 200L
  X <- matrix(rnorm(n * px), nrow = n, ncol = px)
  Y <- matrix(rnorm(n * py), nrow = n, ncol = py)
  expected <- crossprod(X, Y)  # px x py

  res <- shard_crossprod(X, Y, workers = 2, block_x = 64, block_y = 64,
                         materialize = "always")

  expect_true(succeeded(res$run))
  expect_equal(dim(res$value), c(px, py))
  expect_equal(res$value, expected, tolerance = 1e-10)
})

test_that("shard_crossprod: 1000x512 square case matches base crossprod", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(3)
  on.exit(pool_stop(), add = TRUE)

  set.seed(123)
  n <- 1000L; p <- 512L
  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  expected <- crossprod(X)  # X'X, p x p

  res <- shard_crossprod(X, X, workers = 3, block_x = 128, block_y = 128,
                         materialize = "always")

  expect_true(succeeded(res$run))
  expect_equal(dim(res$value), c(p, p))
  expect_equal(res$value, expected, tolerance = 1e-10)
})

test_that("shard_crossprod: auto tile sizes on 800x256 x 800x384", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(11)
  n <- 800L; px <- 256L; py <- 384L
  X <- matrix(rnorm(n * px), nrow = n, ncol = px)
  Y <- matrix(rnorm(n * py), nrow = n, ncol = py)
  expected <- crossprod(X, Y)

  res <- shard_crossprod(X, Y, workers = 2, materialize = "always")

  expect_true(succeeded(res$run))
  expect_equal(res$value, expected, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# Workers=1 vs Workers=N: determinism / correctness equivalence
# ---------------------------------------------------------------------------
test_that("shard_crossprod: workers=1 vs workers=3 produce identical results", {
  skip_on_cran()
  skip_if_not_large()

  set.seed(77)
  n <- 600L; px <- 200L; py <- 150L
  X <- matrix(rnorm(n * px), nrow = n, ncol = px)
  Y <- matrix(rnorm(n * py), nrow = n, ncol = py)

  # Single worker
  pool_stop()
  pool_create(1)
  res1 <- shard_crossprod(X, Y, workers = 1, block_x = 50, block_y = 50,
                          materialize = "always")
  pool_stop()

  # Three workers
  pool_create(3)
  res3 <- shard_crossprod(X, Y, workers = 3, block_x = 50, block_y = 50,
                          materialize = "always")
  pool_stop()

  expect_true(succeeded(res1$run))
  expect_true(succeeded(res3$run))
  expect_equal(res1$value, res3$value, tolerance = 1e-12)
})

test_that("shard_map: workers=1 vs workers=2 give identical column sums", {
  skip_on_cran()
  skip_if_not_large()

  set.seed(55)
  X <- matrix(rnorm(800 * 300), nrow = 800, ncol = 300)

  run_colsums <- function(w) {
    pool_stop()
    pool_create(w)
    blocks <- shards(ncol(X), block_size = 30, workers = w)
    result <- shard_map(blocks,
      borrow = list(X = X),
      fun = function(shard, X) {
        colSums(X[, shard$idx, drop = FALSE])
      },
      workers = w
    )
    pool_stop()
    unname(unlist(results(result)))
  }

  res1 <- run_colsums(1)
  res2 <- run_colsums(2)

  expect_equal(res1, res2, tolerance = 1e-14)
})

# ---------------------------------------------------------------------------
# shard_reduce: streaming aggregation over large shared matrix
# ---------------------------------------------------------------------------
test_that("shard_reduce: streaming row-sum over 1500x400 shared matrix", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(33)
  n <- 1500L; p <- 400L
  M <- share(matrix(rnorm(n * p), nrow = n, ncol = p))
  expected <- sum(M)

  # Partition over rows; each shard sums its row block
  res <- shard_reduce(
    shards(n, block_size = 100),
    map = function(s, M) {
      sum(M[s$idx, ])
    },
    combine = `+`,
    init = 0,
    borrow = list(M = M),
    workers = 2
  )

  expect_true(succeeded(res))
  expect_equal(res$value, expected, tolerance = 1e-10)
})

test_that("shard_reduce: streaming crossprod accumulation matches base R", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(88)
  n <- 1000L; p <- 100L
  X <- share(matrix(rnorm(n * p), nrow = n, ncol = p))
  expected <- crossprod(X)  # p x p

  res <- shard_reduce(
    shards(n, block_size = 200),
    map = function(s, X) {
      crossprod(X[s$idx, , drop = FALSE])
    },
    combine = function(acc, x) acc + x,
    init = matrix(0, nrow = p, ncol = p),
    borrow = list(X = X),
    workers = 2
  )

  expect_true(succeeded(res))
  expect_equal(res$value, expected, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# shard_apply_matrix: large column-parallel apply
# ---------------------------------------------------------------------------
test_that("shard_apply_matrix: column norms on 1000x600 matrix", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  set.seed(22)
  X <- matrix(rnorm(1000 * 600), nrow = 1000, ncol = 600)
  expected <- apply(X, 2, function(v) sqrt(sum(v^2)))

  res <- shard_apply_matrix(X, MARGIN = 2,
    FUN = function(v) sqrt(sum(v^2)),
    workers = 2
  )

  expect_length(res, 600)
  expect_equal(res, expected, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# Buffer output correctness under concurrent tile writes
# ---------------------------------------------------------------------------
test_that("shard_crossprod buffer: concurrent tile writes produce correct result", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(3)
  on.exit(pool_stop(), add = TRUE)

  set.seed(44)
  n <- 700L; px <- 256L; py <- 192L
  X <- matrix(rnorm(n * px), nrow = n, ncol = px)
  Y <- matrix(rnorm(n * py), nrow = n, ncol = py)
  expected <- crossprod(X, Y)

  # Use materialize="never" to get buffer, then read back

  res <- shard_crossprod(X, Y, workers = 3, block_x = 32, block_y = 32,
                         materialize = "never")

  expect_s3_class(res$buffer, "shard_buffer")
  mat <- as.matrix(res$buffer)
  expect_equal(dim(mat), c(px, py))
  expect_equal(mat, expected, tolerance = 1e-10)
})

test_that("shard_crossprod buffer: small tiles (16x16) on large matrix", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(66)
  n <- 500L; px <- 128L; py <- 96L
  X <- matrix(rnorm(n * px), nrow = n, ncol = px)
  Y <- matrix(rnorm(n * py), nrow = n, ncol = py)
  expected <- crossprod(X, Y)

  # Many small tiles = many concurrent buffer writes
  res <- shard_crossprod(X, Y, workers = 2, block_x = 16, block_y = 16,
                         materialize = "always")

  expect_true(succeeded(res$run))
  expect_equal(res$value, expected, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# Pre-shared large matrices (zero-copy path)
# ---------------------------------------------------------------------------
test_that("shard_map with pre-shared 1200x400 matrix matches base R", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(101)
  X_raw <- matrix(rnorm(1200 * 400), nrow = 1200, ncol = 400)
  X <- share(X_raw)
  expected <- colMeans(X_raw)

  blocks <- shards(ncol(X), block_size = 40, workers = 2)

  result <- shard_map(blocks,
    borrow = list(X = X),
    fun = function(shard, X) {
      colMeans(X[, shard$idx, drop = FALSE])
    },
    workers = 2
  )

  expect_true(succeeded(result))
  res <- unlist(results(result))
  expect_equal(unname(res), unname(expected), tolerance = 1e-12)
})

test_that("shard_crossprod with pre-shared 800x256 matrices", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(202)
  n <- 800L; px <- 256L; py <- 128L
  X <- share(matrix(rnorm(n * px), nrow = n, ncol = px))
  Y <- share(matrix(rnorm(n * py), nrow = n, ncol = py))
  expected <- crossprod(X, Y)

  res <- shard_crossprod(X, Y, workers = 2, block_x = 64, block_y = 64,
                         materialize = "always")

  expect_true(succeeded(res$run))
  expect_equal(res$value, expected, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# Multiple large borrowed inputs
# ---------------------------------------------------------------------------
test_that("shard_map: two large borrowed matrices, element-wise combine", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(303)
  n <- 800L; p <- 300L
  A <- matrix(rnorm(n * p), nrow = n, ncol = p)
  B <- matrix(rnorm(n * p), nrow = n, ncol = p)
  expected <- colSums(A * B)

  blocks <- shards(p, block_size = 50, workers = 2)

  result <- shard_map(blocks,
    borrow = list(A = A, B = B),
    fun = function(shard, A, B) {
      colSums(A[, shard$idx, drop = FALSE] * B[, shard$idx, drop = FALSE])
    },
    workers = 2
  )

  expect_true(succeeded(result))
  res <- unlist(results(result))
  expect_equal(unname(res), unname(expected), tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# Buffer write with shard_map on large output
# ---------------------------------------------------------------------------
test_that("shard_map: write large output to shared buffer", {
  skip_on_cran()
  skip_if_not_large()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  set.seed(404)
  n <- 1000L; p <- 500L
  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  expected <- colSums(X)

  out_buf <- buffer("double", dim = p, init = 0)

  blocks <- shards(p, block_size = 50, workers = 2)

  result <- shard_map(blocks,
    borrow = list(X = X),
    out = list(out = out_buf),
    fun = function(shard, X, out) {
      for (j in shard$idx) {
        out[j] <- sum(X[, j])
      }
      NULL
    },
    workers = 2
  )

  expect_true(succeeded(result))
  expect_equal(as.numeric(out_buf[]), unname(expected), tolerance = 1e-12)
})

test_that("view materialization hotspots are attributed in diagnostics", {
  skip_on_cran()

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  set.seed(1)
  n <- 50L
  v <- 20L
  Y <- matrix(rnorm(n * v), nrow = n)
  Ysh <- share(Y, backing = "mmap")

  blocks <- shards(ncol(Ysh), workers = 2)
  res <- shard_map(
    blocks,
    borrow = list(Y = Ysh),
    fun = function(sh, Y) {
      vv <- view_block(Y, cols = idx_range(sh$start, sh$end))
      m <- materialize(vv)
      sum(m)
    },
    workers = 2,
    chunk_size = 1L,
    autotune = FALSE,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))

  cr <- copy_report(res)
  expect_gt(cr$view_materialized_bytes %||% 0, 0)
  hs <- cr$view_materialize_hotspots %||% list()
  expect_true(is.list(hs))
  expect_gt(length(hs), 0L)
})

test_that("shard_map re-exports borrow/out after scratch-triggered recycle", {
  skip_on_cran()

  pool_stop()

  n_workers <- 2L
  mats <- lapply(1:8, function(i) matrix(rnorm(10 * 10), 10, 10))
  expected <- vapply(mats, function(m) sqrt(sum(m^2)), numeric(1))

  out <- buffer("double", dim = length(mats))
  blocks <- shards(length(mats), block_size = 1, workers = n_workers)

  res <- suppressWarnings(
    shard_map(
      blocks,
      borrow = list(mats = mats),
      out = list(out = out),
      fun = function(sh, mats, out) {
        # Force a recycle once mid-run. This ensures any replacement worker
        # still has the shard_map borrow/out context re-exported.
        if (isTRUE(sh$id == 1L)) {
          scratch_pool_config(max_bytes = 1)
          scratch_matrix(200, 200, key = "boom")
        }

        i <- sh$idx[[1]]
        out[i] <- sqrt(sum(mats[[i]]^2))
        NULL
      },
      workers = n_workers,
      diagnostics = TRUE
    )
  )

  expect_true(succeeded(res))
  expect_equal(as.numeric(out[]), expected, tolerance = 1e-10)

  pool_stop()
})

test_that("shard_map(N) can run in shm_queue mode for out-buffer workflows", {
  skip_on_cran()
  if (!shard:::taskq_supported()) skip("shm_queue not supported (no atomics)")

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 100L
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")

  res <- shard_map(
    n,
    out = list(out = out),
    fun = function(sh, out) {
      out[sh$idx] <- sh$idx
      NULL
    },
    workers = 2,
    chunk_size = 1,
    dispatch_mode = "shm_queue",
    diagnostics = TRUE
  )

  expect_true(succeeded(res))
  expect_equal(res$diagnostics$dispatch_mode %||% NULL, "shm_queue")

  # Results are not gathered in shm_queue mode; expect NULL placeholders.
  rr <- results(res)
  expect_true(inherits(rr, "shard_results_placeholder"))
  expect_true(all(vapply(rr, is.null, logical(1))))

  expect_equal(as.integer(out[]), 1:n)
})

test_that("shm_queue dispatch_opts can override block_size", {
  skip_on_cran()
  if (!shard:::taskq_supported()) skip("shm_queue not supported (no atomics)")

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 32L
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")

  res <- shard_map(
    n,
    out = list(out = out),
    fun = function(sh, out) {
      out[sh$idx] <- sh$idx
      NULL
    },
    workers = 2,
    chunk_size = 1,
    dispatch_mode = "shm_queue",
    dispatch_opts = list(block_size = 1L),
    diagnostics = TRUE
  )

  expect_true(succeeded(res))
  expect_equal(res$shards$num_shards, n)
  expect_equal(as.integer(out[]), 1:n)
})

test_that("profile='speed' auto-enables shm_queue for scalar-N out-buffer workflows", {
  skip_on_cran()
  if (!shard:::taskq_supported()) skip("shm_queue not supported (no atomics)")

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 100L
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")

  res <- shard_map(
    n,
    out = list(out = out),
    fun = function(sh, out) {
      out[sh$idx] <- sh$idx
      NULL
    },
    workers = 2,
    chunk_size = 1,
    profile = "speed",
    diagnostics = TRUE
  )

  expect_true(succeeded(res))
  expect_equal(res$diagnostics$dispatch_mode %||% NULL, "shm_queue")
  expect_equal(as.integer(out[]), 1:n)
})

test_that("shard_map(shard_descriptor) can run in shm_queue mode for out-buffer workflows", {
  skip_on_cran()
  if (!shard:::taskq_supported()) skip("shm_queue not supported (no atomics)")

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 100L
  sh <- shards(n, block_size = 10L, workers = 2)
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")

  res <- shard_map(
    sh,
    out = list(out = out),
    fun = function(shard, out) {
      out[shard$idx] <- shard$idx
      NULL
    },
    workers = 2,
    chunk_size = 1,
    dispatch_mode = "shm_queue",
    diagnostics = TRUE
  )

  expect_true(succeeded(res))
  expect_equal(res$diagnostics$dispatch_mode %||% NULL, "shm_queue")
  expect_equal(as.integer(out[]), 1:n)
})

test_that("shm_queue reports retry accounting and per-task retry_count for failures", {
  skip_on_cran()
  if (!shard:::taskq_supported()) skip("shm_queue not supported (no atomics)")

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 10L
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")

  res <- shard_map(
    n,
    out = list(out = out),
    fun = function(sh, out) {
      if (identical(sh$idx, 3L)) stop("boom")
      out[sh$idx] <- sh$idx
      NULL
    },
    workers = 2,
    chunk_size = 1,
    max_retries = 1,
    dispatch_mode = "shm_queue",
    dispatch_opts = list(block_size = 1L),
    diagnostics = TRUE
  )

  expect_false(succeeded(res))
  expect_equal(res$queue_status$failed %||% NULL, 1L)
  expect_equal(res$queue_status$total_retries %||% NULL, 1L)

  # Task 3 errors twice: first schedules a retry, second marks as failed.
  expect_true("3" %in% names(res$failures))
  expect_equal(res$failures[["3"]]$id %||% NULL, 3L)
  expect_equal(res$failures[["3"]]$retry_count %||% NULL, 2L)

  # Only task 3 fails; others write through.
  expect_equal(as.integer(out[])[-3], (1:n)[-3])
  expect_equal(as.integer(out[])[3], 0L)
})

test_that("shm_queue can write bounded per-worker error logs when enabled", {
  skip_on_cran()
  if (!shard:::taskq_supported()) skip("shm_queue not supported (no atomics)")

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 10L
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")

  res <- shard_map(
    n,
    out = list(out = out),
    fun = function(sh, out) {
      if (identical(sh$idx, 3L)) stop("boom")
      out[sh$idx] <- sh$idx
      NULL
    },
    workers = 2,
    chunk_size = 1,
    max_retries = 1,
    dispatch_mode = "shm_queue",
    dispatch_opts = list(block_size = 1L, error_log = TRUE, error_log_max_lines = 10L),
    diagnostics = TRUE
  )

  expect_false(succeeded(res))
  logs <- res$diagnostics$error_logs %||% list()
  expect_true(length(logs) >= 1L)

  paths <- vapply(logs, function(x) x$path %||% NA_character_, character(1))
  expect_true(all(file.exists(paths)))

  lines <- unlist(lapply(paths, function(p) readLines(p, warn = FALSE)), use.names = FALSE)
  expect_true(any(grepl("^3\\tboom$", lines)))
})

test_that("shm_queue timeout cleans up workers and leaves the pool reusable", {
  skip_on_cran()
  skip_if_conn_exhausted()
  if (!shard:::taskq_supported()) skip("shm_queue not supported (no atomics)")

  pool_stop()
  pool_create(n = 1)
  on.exit(pool_stop(), add = TRUE)

  out <- buffer("integer", dim = 2L, init = 0L, backing = "mmap")

  expect_error(
    shard_map(
      2L,
      out = list(out = out),
      fun = function(sh, out) {
        Sys.sleep(0.2)
        out[sh$idx] <- sh$idx
        NULL
      },
      chunk_size = 1L,
      dispatch_mode = "shm_queue",
      dispatch_opts = list(block_size = 1L),
      timeout = 0.05
    ),
    "timed out"
  )

  follow_up <- shard_map(
    shards(2L, block_size = 1L, workers = 1L),
    fun = function(shard) shard$id
  )

  expect_true(succeeded(follow_up))
  expect_equal(unname(unlist(results(follow_up))), c(1, 2))
})

test_that("dispatch_opts$auto_table streams data.frame returns into a table sink", {
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 100L
  blocks <- shards(n, workers = 2)

  res <- shard_map(
    blocks,
    fun = function(sh) {
      data.frame(i = sh$idx, v = sh$idx * 2L)
    },
    workers = 2,
    chunk_size = 1L,
    autotune = FALSE,
    dispatch_mode = "rpc_chunked",
    dispatch_opts = list(auto_table = TRUE, auto_table_materialize = "never"),
    diagnostics = TRUE
  )

  expect_true(succeeded(res))
  handle <- results(res)
  expect_true(inherits(handle, "shard_row_groups"))

  expect_equal(stream_count(handle), n)

  df <- collect(handle)
  expect_equal(nrow(df), n)
  expect_equal(sort(df$i), 1:n)
  expect_equal(df$v[order(df$i)], (1:n) * 2L)
})

test_that("auto_table requires out= to be empty", {
  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  n <- 10L
  blocks <- shards(n, workers = 2)
  out <- buffer("integer", dim = n, init = 0L, backing = "mmap")

  expect_error(
    shard_map(
      blocks,
      out = list(out = out),
      fun = function(sh, out) {
        out[sh$idx] <- sh$idx
        data.frame(i = sh$idx)
      },
      workers = 2,
      chunk_size = 1L,
      autotune = FALSE,
      dispatch_mode = "rpc_chunked",
      dispatch_opts = list(auto_table = TRUE),
      diagnostics = TRUE
    ),
    "requires out="
  )
})

test_that("shard_map(N) can autotune shard block sizes online and records history", {
  skip_on_cran()

  pool_stop()
  on.exit(pool_stop(), add = TRUE)

  # Force growth decisions by making min_shard_time unrealistically high.
  res <- shard_map(
    200,
    fun = function(sh) sum(sh$idx),
    workers = 1,
    chunk_size = 1,
    diagnostics = TRUE,
    autotune = list(mode = "online", max_rounds = 2L, probe_shards_per_worker = 2L, min_shard_time = 10)
  )

  expect_true(succeeded(res))

  # We should see multiple rounds and a block_size change.
  h <- res$diagnostics$autotune$history %||% list()
  expect_true(length(h) >= 2L)
  bss <- vapply(h, function(x) x$block_size %||% NA_integer_, integer(1))
  expect_true(length(unique(bss[!is.na(bss)])) >= 2L)

  # Results are correct for a plausible tuning path (50 -> 100).
  r <- results(res, flatten = FALSE)
  expect_equal(unname(unlist(r)), c(sum(1:50), sum(51:100), sum(101:200)))
})

test_that("scheduler_policy can throttle 'huge' chunks", {
  skip_on_cran()

  pool_stop()

  # A kernel that just sleeps. The footprint marks every chunk as huge so the
  # scheduler must cap concurrency when max_huge_concurrency is set.
  register_kernel(
    name = "sleep_huge",
    impl = function(sh, ...) {
      Sys.sleep(0.02)
      sh$id
    },
    footprint = function(sh) list(class = "huge", bytes = 128 * 1024^2),
    supports_views = FALSE
  )

  res <- shard_map(
    shards(8, block_size = 1),
    kernel = "sleep_huge",
    workers = 2,
    scheduler_policy = list(max_huge_concurrency = 1L),
    diagnostics = TRUE
  )

  expect_true(succeeded(res))
  expect_equal(length(results(res)), 8L)
  expect_true((res$diagnostics$scheduler$throttle_events %||% 0L) > 0L)

  pool_stop()
})

test_that("scheduler throttle clears huge in-flight state after timeouts", {
  skip_on_cran()
  skip_if_conn_exhausted()

  pool_stop()
  on.exit({
    setTimeLimit(cpu = Inf, elapsed = Inf, transient = FALSE)
    pool_stop()
  }, add = TRUE)
  setTimeLimit(cpu = Inf, elapsed = 5, transient = TRUE)

  pool_create(n = 1)
  chunks <- list(
    list(id = 1L, footprint_class = "huge"),
    list(id = 2L, footprint_class = "huge")
  )

  res <- suppressWarnings(
    dispatch_chunks(
      chunks,
      fun = function(chunk) {
        Sys.sleep(0.2)
        chunk$id
      },
      timeout = 0.05,
      max_retries = 0L,
      health_check_interval = 1L,
      scheduler_policy = list(max_huge_concurrency = 1L)
    )
  )

  expect_equal(res$queue_status$completed, 0L)
  expect_equal(res$queue_status$failed, 2L)
})

#' @title Utility Functions
#' @description Internal utilities for shard package.
#' @name utils
NULL

#' Null-coalescing Operator
#'
#' @param x Value to check.
#' @param y Default value if x is NULL.
#' @return x if not NULL, otherwise y.
#' @keywords internal
#' @noRd
`%||%` <- function(x, y) {

  if (is.null(x)) y else x
}

#' Parse Human-Readable Byte Strings
#'
#' Converts strings like "2GB", "512MB", "1.5GB" to bytes.
#'
#' @param x Character or numeric. If numeric, returned as-is.
#' @return Numeric value in bytes.
#' @keywords internal
#' @noRd
parse_bytes <- function(x) {
  if (is.numeric(x)) {
    return(as.numeric(x))
  }

  x <- toupper(trimws(x))

  # Extract number and unit
  match <- regexec("^([0-9.]+)\\s*(B|KB|MB|GB|TB|K|M|G|T)?$", x)
  parts <- regmatches(x, match)[[1]]

  if (length(parts) < 2) {
    stop("Cannot parse byte string: ", x, call. = FALSE)
  }

  value <- as.numeric(parts[2])
  unit <- if (length(parts) >= 3) parts[3] else "B"

  multiplier <- switch(
    unit,
    "B" = 1,
    "K" = , "KB" = 1024,
    "M" = , "MB" = 1024^2,
    "G" = , "GB" = 1024^3,
    "T" = , "TB" = 1024^4,
    1
  )

  value * multiplier
}

#' Format Bytes as Human-Readable String
#'
#' @param bytes Numeric value in bytes.
#' @param digits Integer. Decimal places (default 1).
#' @return Character string like "1.5 GB".
#' @keywords internal
#' @noRd
format_bytes <- function(bytes, digits = 1) {
  if (is.na(bytes)) {
    return("NA")
  }

  if (bytes < 1024) {
    return(paste(bytes, "B"))
  }

  units <- c("B", "KB", "MB", "GB", "TB", "PB")
  exp <- min(floor(log(bytes, 1024)), length(units) - 1)
  value <- bytes / (1024^exp)

  sprintf(paste0("%.", digits, "f %s"), value, units[exp + 1])
}

#' Safe Divide
#'
#' Division that returns NA instead of Inf/NaN for zero divisors.
#'
#' @param x Numerator.
#' @param y Denominator.
#' @return x/y or NA if y is zero.
#' @keywords internal
#' @noRd
safe_div <- function(x, y) {
  ifelse(y == 0, NA_real_, x / y)
}

#' Retry with Backoff
#'
#' Retries a function call with exponential backoff.
#'
#' @param f Function to call.
#' @param max_attempts Integer. Maximum retry attempts.
#' @param initial_delay Numeric. Initial delay in seconds.
#' @param max_delay Numeric. Maximum delay in seconds.
#' @param on_error Function. Called on each error with (attempt, error).
#' @return Result of f() or error on final failure.
#' @keywords internal
#' @noRd
retry_with_backoff <- function(f, max_attempts = 3, initial_delay = 0.1,
                               max_delay = 5, on_error = NULL) {
  delay <- initial_delay

  for (attempt in seq_len(max_attempts)) {
    result <- tryCatch(
      list(success = TRUE, value = f()),
      error = function(e) list(success = FALSE, error = e)
    )

    if (result$success) {
      return(result$value)
    }

    if (!is.null(on_error)) {
      on_error(attempt, result$error)
    }

    if (attempt < max_attempts) {
      Sys.sleep(delay)
      delay <- min(delay * 2, max_delay)
    }
  }

  stop(result$error)
}

#' Check Package Availability
#'
#' @param pkg Package name.
#' @return Logical. TRUE if package is available.
#' @keywords internal
#' @noRd
pkg_available <- function(pkg) {
  requireNamespace(pkg, quietly = TRUE)
}

#' Generate Unique ID
#'
#' Creates a short unique identifier.
#'
#' @param prefix Character. Prefix for the ID.
#' @return Character string like "prefix_a1b2c3".
#' @keywords internal
#' @noRd
unique_id <- function(prefix = "") {
  # Use high-resolution time + random component for uniqueness
  timestamp <- as.numeric(Sys.time()) * 1e6
  random_part <- paste(sample(c(letters, 0:9), 8, replace = TRUE), collapse = "")
  hash <- sprintf("%s_%s", format(timestamp, scientific = FALSE), random_part)

  if (nchar(prefix) > 0) {
    paste0(prefix, "_", hash)
  } else {
    hash
  }
}

#' Validate Positive Integer
#'
#' @param x Value to validate.
#' @param name Name for error messages.
#' @return Integer value.
#' @keywords internal
#' @noRd
assert_positive_int <- function(x, name = "value") {
  x <- as.integer(x)
  if (is.na(x) || x < 1L) {
    stop(name, " must be a positive integer", call. = FALSE)
  }
  x
}

#' Capture Time
#'
#' Simple timing wrapper.
#'
#' @param expr Expression to time.
#' @return List with result and elapsed time.
#' @keywords internal
#' @noRd
capture_time <- function(expr) {
  start <- Sys.time()
  result <- force(expr)
  elapsed <- as.numeric(difftime(Sys.time(), start, units = "secs"))
  list(result = result, elapsed = elapsed)
}

test_that("parse_bytes handles numeric input", {
  expect_equal(parse_bytes(1024), 1024)
  expect_equal(parse_bytes(1.5e9), 1.5e9)
})

test_that("parse_bytes handles byte strings", {
  expect_equal(parse_bytes("1024"), 1024)
  expect_equal(parse_bytes("1KB"), 1024)
  expect_equal(parse_bytes("1 KB"), 1024)
  expect_equal(parse_bytes("1K"), 1024)
  expect_equal(parse_bytes("1MB"), 1024^2)
  expect_equal(parse_bytes("1.5MB"), 1.5 * 1024^2)
  expect_equal(parse_bytes("1GB"), 1024^3)
  expect_equal(parse_bytes("2gb"), 2 * 1024^3)
  expect_equal(parse_bytes("1TB"), 1024^4)
})

test_that("parse_bytes errors on invalid input", {
  expect_error(parse_bytes("invalid"))
  expect_error(parse_bytes(""))
})

test_that("format_bytes produces readable strings", {
  expect_equal(format_bytes(500), "500 B")
  expect_equal(format_bytes(1024), "1.0 KB")
  expect_equal(format_bytes(1536), "1.5 KB")
  expect_equal(format_bytes(1024^2), "1.0 MB")
  expect_equal(format_bytes(1.5 * 1024^3), "1.5 GB")
  expect_equal(format_bytes(NA), "NA")
})

test_that("null coalescing operator works", {
  expect_equal(NULL %||% 5, 5)
  expect_equal(3 %||% 5, 3)
  expect_equal(NA %||% 5, NA)  # NA is not NULL
})

test_that("safe_div handles zero divisors", {
  expect_equal(safe_div(10, 2), 5)
  expect_true(is.na(safe_div(10, 0)))
  expect_equal(safe_div(c(10, 20), c(2, 0)), c(5, NA))
})

test_that("assert_positive_int validates correctly", {
  expect_equal(assert_positive_int(5), 5L)
  expect_equal(assert_positive_int(5.9), 5L)  # Truncates
  expect_error(assert_positive_int(0))
  expect_error(assert_positive_int(-1))
  expect_error(assert_positive_int(NA))
})

test_that("capture_time measures elapsed time", {
  result <- capture_time({
    Sys.sleep(0.1)
    42
  })

  expect_equal(result$result, 42)
  expect_gte(result$elapsed, 0.1)
  expect_lt(result$elapsed, 0.5)  # Should not take too long
})

test_that("unique_id generates unique IDs", {
  id1 <- unique_id("test")
  id2 <- unique_id("test")

  expect_true(startsWith(id1, "test_"))
  expect_true(startsWith(id2, "test_"))
  expect_false(id1 == id2)

  # Without prefix
  id3 <- unique_id()
  expect_false(startsWith(id3, "_"))
})

# Lazy shard descriptor for contiguous shards.
#
# Used by shm_queue dispatch mode to avoid materializing a giant list of shard
# descriptors on the master.

shards_lazy <- function(n, block_size) {
  n <- as.integer(n)
  block_size <- as.integer(block_size)
  if (is.na(n) || n < 1L) stop("n must be >= 1", call. = FALSE)
  if (is.na(block_size) || block_size < 1L) stop("block_size must be >= 1", call. = FALSE)

  num_shards <- as.integer(ceiling(n / block_size))

  structure(
    list(
      n = n,
      block_size = block_size,
      strategy = "contiguous_lazy",
      num_shards = num_shards
    ),
    class = c("shard_descriptor_lazy", "shard_descriptor")
  )
}

.lazy_shard_at <- function(x, i) {
  i <- as.integer(i)
  if (is.na(i) || i < 1L || i > x$num_shards) stop("shard index out of range", call. = FALSE)
  start <- (i - 1L) * x$block_size + 1L
  end <- min(i * x$block_size, x$n)
  list(id = i, start = start, end = end, idx = start:end, len = end - start + 1L)
}

#' Print a shard_descriptor_lazy Object
#'
#' @param x A \code{shard_descriptor_lazy} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @examples
#' sh <- shards(100, block_size = 25)
#' print(sh)
#' @export
print.shard_descriptor_lazy <- function(x, ...) {
  cat("shard descriptor (lazy)\n")
  cat("  Items:", format(x$n, big.mark = ","), "\n")
  cat("  Block size:", format(x$block_size, big.mark = ","), "\n")
  cat("  Strategy:", x$strategy, "\n")
  cat("  Shards:", x$num_shards, "\n")
  invisible(x)
}

#' Length of a shard_descriptor_lazy Object
#'
#' @param x A \code{shard_descriptor_lazy} object.
#' @return An integer scalar giving the number of shards.
#' @examples
#' sh <- shards(100, block_size = 25)
#' length(sh)
#' @export
length.shard_descriptor_lazy <- function(x) {
  x$num_shards
}

#' Subset a shard_descriptor_lazy Object
#'
#' @param x A \code{shard_descriptor_lazy} object.
#' @param i Index or indices.
#' @return A subset of the object.
#' @examples
#' sh <- shards(100, block_size = 25)
#' sh[1:2]
#' @export
`[.shard_descriptor_lazy` <- function(x, i) {
  lapply(i, function(j) .lazy_shard_at(x, j))
}

#' Extract a Single Shard from a shard_descriptor_lazy Object
#'
#' @param x A \code{shard_descriptor_lazy} object.
#' @param i Index.
#' @return A subset of the object.
#' @examples
#' sh <- shards(100, block_size = 25)
#' sh[[1]]
#' @export
`[[.shard_descriptor_lazy` <- function(x, i) {
  .lazy_shard_at(x, i)
}

test_that("shards_lazy creates valid lazy shard descriptor", {
  x <- shard:::shards_lazy(100, block_size = 10)

  expect_s3_class(x, "shard_descriptor_lazy")
  expect_s3_class(x, "shard_descriptor")
  expect_equal(x$n, 100L)
  expect_equal(x$block_size, 10L)
  expect_equal(x$strategy, "contiguous_lazy")
  expect_equal(x$num_shards, 10L)
})

test_that("shards_lazy handles non-evenly-divisible sizes", {
  x <- shard:::shards_lazy(95, block_size = 10)

  expect_equal(x$num_shards, 10L)  # ceiling(95/10)
  expect_equal(length(x), 10L)

  # Last shard should have fewer elements
  last <- x[[10]]
  expect_equal(last$start, 91L)
  expect_equal(last$end, 95L)
  expect_equal(last$len, 5L)
})

test_that("shards_lazy length method works", {
  x <- shard:::shards_lazy(100, block_size = 25)
  expect_equal(length(x), 4L)

  x2 <- shard:::shards_lazy(1000, block_size = 1)
  expect_equal(length(x2), 1000L)
})

test_that("shards_lazy [[ indexing returns correct shard", {
  x <- shard:::shards_lazy(50, block_size = 10)

  # First shard
  s1 <- x[[1]]
  expect_equal(s1$id, 1L)
  expect_equal(s1$start, 1L)
  expect_equal(s1$end, 10L)
  expect_equal(s1$len, 10L)
  expect_equal(s1$idx, 1:10)

  # Middle shard
  s3 <- x[[3]]
  expect_equal(s3$id, 3L)
  expect_equal(s3$start, 21L)
  expect_equal(s3$end, 30L)
  expect_equal(s3$len, 10L)
  expect_equal(s3$idx, 21:30)

  # Last shard
  s5 <- x[[5]]
  expect_equal(s5$id, 5L)
  expect_equal(s5$start, 41L)
  expect_equal(s5$end, 50L)
  expect_equal(s5$len, 10L)
  expect_equal(s5$idx, 41:50)
})

test_that("shards_lazy [ indexing returns list of shards", {
  x <- shard:::shards_lazy(50, block_size = 10)

  # Single index
  s <- x[2]
  expect_true(is.list(s))
  expect_length(s, 1)
  expect_equal(s[[1]]$id, 2L)

  # Multiple indices
  ss <- x[c(1, 3, 5)]
  expect_length(ss, 3)
  expect_equal(ss[[1]]$id, 1L)
  expect_equal(ss[[2]]$id, 3L)
  expect_equal(ss[[3]]$id, 5L)
})

test_that("shards_lazy print method works", {
  x <- shard:::shards_lazy(1000000, block_size = 100)

  out <- capture.output(print(x))
  expect_true(any(grepl("shard descriptor \\(lazy\\)", out)))
  expect_true(any(grepl("Items:", out)))
  expect_true(any(grepl("Block size:", out)))
  expect_true(any(grepl("Strategy:", out)))
  expect_true(any(grepl("Shards:", out)))
})

test_that("shards_lazy validates inputs", {
  expect_error(shard:::shards_lazy(0, block_size = 10), "n must be >= 1")
  expect_error(shard:::shards_lazy(-1, block_size = 10), "n must be >= 1")
  expect_error(shard:::shards_lazy(100, block_size = 0), "block_size must be >= 1")
  expect_error(shard:::shards_lazy(100, block_size = -1), "block_size must be >= 1")
})

test_that("shards_lazy index out of range errors", {
  x <- shard:::shards_lazy(50, block_size = 10)

  expect_error(x[[0]], "shard index out of range")
  expect_error(x[[6]], "shard index out of range")
  expect_error(x[[-1]], "shard index out of range")
})

test_that("shards_lazy works with block_size = 1", {
  x <- shard:::shards_lazy(5, block_size = 1)

  expect_equal(length(x), 5L)

  for (i in 1:5) {
    s <- x[[i]]
    expect_equal(s$id, i)
    expect_equal(s$start, i)
    expect_equal(s$end, i)
    expect_equal(s$len, 1L)
    expect_equal(s$idx, i)
  }
})

test_that("shards_lazy works with single shard", {
  x <- shard:::shards_lazy(10, block_size = 100)

  expect_equal(length(x), 1L)

  s <- x[[1]]
  expect_equal(s$id, 1L)
  expect_equal(s$start, 1L)
  expect_equal(s$end, 10L)
  expect_equal(s$len, 10L)
})

# Kernel registry for shard_map().
#
# This provides a low-ceremony way to select known high-performance kernels
# without forcing users to write the same boilerplate fun() repeatedly.

.kernel_env <- new.env(parent = emptyenv())
.kernel_env$registry <- new.env(parent = emptyenv())

#' Register a shard kernel
#'
#' Registers a named kernel implementation that can be selected by
#' `shard_map(..., kernel = "name")`.
#'
#' A "kernel" is just a function that shard_map can call for each shard. The
#' registry lets shard_map attach additional metadata (footprint hints,
#' supports_views) for scheduling/autotuning.
#'
#' @param name Kernel name (string).
#' @param impl Function implementing the kernel. It must accept the shard
#'   descriptor as its first argument.
#' @param signature Optional short signature string for documentation.
#' @param footprint Optional footprint hint. Either a constant (bytes) or a
#'   function `(shard, ...) -> list(class='tiny'|'medium'|'huge', bytes=...)`.
#' @param supports_views Logical. Whether the kernel is intended to operate on
#'   shard views without slice materialization.
#' @param description Optional human-readable description.
#' @return Invisibly, the registered kernel metadata.
#' @export
#' @examples
#' list_kernels()
register_kernel <- function(name,
                            impl,
                            signature = NULL,
                            footprint = NULL,
                            supports_views = TRUE,
                            description = NULL) {
  name <- as.character(name)
  if (!nzchar(name)) stop("name must be a non-empty string", call. = FALSE)
  if (!is.function(impl)) stop("impl must be a function", call. = FALSE)
  if (!is.null(signature) && !is.character(signature)) stop("signature must be a string or NULL", call. = FALSE)
  if (!is.null(description) && !is.character(description)) stop("description must be a string or NULL", call. = FALSE)
  if (!is.null(footprint) && !(is.function(footprint) || is.numeric(footprint))) {
    stop("footprint must be a function, numeric, or NULL", call. = FALSE)
  }

  meta <- list(
    name = name,
    impl = impl,
    signature = signature,
    footprint = footprint,
    supports_views = isTRUE(supports_views),
    description = description
  )

  assign(name, meta, envir = .kernel_env$registry)
  invisible(meta)
}

#' List registered kernels
#'
#' @return A character vector of registered kernel names.
#' @export
#' @examples
#' list_kernels()
list_kernels <- function() {
  ls(.kernel_env$registry, all.names = TRUE)
}

get_kernel <- function(name) {
  name <- as.character(name)
  if (!exists(name, envir = .kernel_env$registry, inherits = FALSE)) return(NULL)
  get(name, envir = .kernel_env$registry, inherits = FALSE)
}

# Built-in kernel: crossprod tile (t(X_block) %*% Y_block) into an output buffer.
.kernel_crossprod_tile <- function(tile, X, Y, Z) {
  # Fast path: call into BLAS and write directly into the output buffer segment.
  # This avoids allocating a temporary (kx x ky) R matrix per tile and then
  # copying it into Z.
  .Call(
    "C_shard_mat_crossprod_block_into",
    X,
    Y,
    1L,
    as.integer(nrow(X)),
    as.integer(tile$x_start),
    as.integer(tile$x_end),
    as.integer(tile$y_start),
    as.integer(tile$y_end),
    Z$segment$ptr,
    PACKAGE = "shard"
  )

  # Maintain per-process buffer write diagnostics for copy_report().
  # This matches the semantics of writing exactly once per output element.
  kx <- as.integer(tile$x_end - tile$x_start + 1L)
  ky <- as.integer(tile$y_end - tile$y_start + 1L)
  if (!is.na(kx) && !is.na(ky) && kx > 0L && ky > 0L) {
    .buffer_diag_env$writes <- .buffer_diag_env$writes + 1L
    .buffer_diag_env$bytes <- .buffer_diag_env$bytes + (as.double(kx) * as.double(ky) * 8)
  }
  NULL
}

register_kernel(
  name = "crossprod_tile",
  impl = .kernel_crossprod_tile,
  signature = "(tile, X, Y, Z) -> NULL",
  footprint = function(tile) {
    bx <- as.integer(tile$x_end - tile$x_start + 1L)
    by <- as.integer(tile$y_end - tile$y_start + 1L)
    bytes <- as.double(bx) * as.double(by) * 8
    cls <- if (bytes >= 64 * 1024^2) "huge" else if (bytes >= 8 * 1024^2) "medium" else "tiny"
    list(class = cls, bytes = bytes)
  },
  supports_views = TRUE,
  description = "Compute a crossprod tile using block views and BLAS-3."
)

.kernel_col_means <- function(shard, X, out) {
  if (is.null(shard$start) || is.null(shard$end)) {
    stop("col_means kernel requires contiguous shard descriptors (start/end)", call. = FALSE)
  }
  cs <- as.integer(shard$start)
  ce <- as.integer(shard$end)
  if (is.na(cs) || is.na(ce) || cs < 1L || ce < cs) stop("Invalid shard start/end", call. = FALSE)
  v <- view_block(X, cols = idx_range(cs, ce))
  out[cs:ce] <- view_col_sums(v) / as.double(nrow(X))
  NULL
}

register_kernel(
  name = "col_means",
  impl = .kernel_col_means,
  signature = "(shard, X, out) -> NULL",
  supports_views = TRUE,
  description = "Compute column means for a contiguous shard using view_col_sums()."
)

.kernel_col_vars <- function(shard, X, out) {
  if (is.null(shard$start) || is.null(shard$end)) {
    stop("col_vars kernel requires contiguous shard descriptors (start/end)", call. = FALSE)
  }
  cs <- as.integer(shard$start)
  ce <- as.integer(shard$end)
  if (is.na(cs) || is.na(ce) || cs < 1L || ce < cs) stop("Invalid shard start/end", call. = FALSE)
  v <- view_block(X, cols = idx_range(cs, ce))
  out[cs:ce] <- view_col_vars(v)
  NULL
}

register_kernel(
  name = "col_vars",
  impl = .kernel_col_vars,
  signature = "(shard, X, out) -> NULL",
  supports_views = TRUE,
  description = "Compute sample column variances for a contiguous shard using view_col_vars()."
)

#' @title Streaming Reductions over Shards
#' @description Reduce shard results without gathering all per-shard returns on the master.
#' @name shard_reduce
NULL

#' Reduce over shards without gathering
#'
#' `shard_reduce()` executes `map()` over shards in parallel and combines results
#' using an associative `combine()` function. Unlike `shard_map()`, it does not
#' accumulate all per-shard results on the master; it streams partials as chunks
#' complete.
#'
#' For performance and memory efficiency, reduction is performed in two stages:
#' 1) per-chunk partial reduction inside each worker, and
#' 2) streaming combine of partials on the master.
#'
#' @param shards A `shard_descriptor` from [shards()], or an integer N.
#' @param map Function executed per shard. Receives shard descriptor as first
#'   argument, followed by borrowed inputs and outputs.
#' @param combine Function `(acc, value) -> acc` used to combine results. Should
#'   be associative for deterministic behavior under chunking.
#' @param init Initial accumulator value.
#' @param borrow Named list of shared inputs (same semantics as [shard_map()]).
#' @param out Named list of output buffers/sinks (same semantics as [shard_map()]).
#' @param workers Number of worker processes.
#' @param chunk_size Shards to batch per worker dispatch (default 1).
#' @param profile Execution profile (same semantics as [shard_map()]).
#' @param mem_cap Memory cap per worker (same semantics as [shard_map()]).
#' @param recycle Worker recycling policy (same semantics as [shard_map()]).
#' @param cow Copy-on-write policy for borrowed inputs (same semantics as [shard_map()]).
#' @param seed RNG seed for reproducibility.
#' @param diagnostics Logical; collect diagnostics (default TRUE).
#' @param packages Additional packages to load in workers.
#' @param init_expr Expression to evaluate in each worker on startup.
#' @param timeout Seconds to wait for each chunk.
#' @param max_retries Maximum retries per chunk.
#' @param health_check_interval Check worker health every N completions.
#'
#' @return A `shard_reduce_result` with fields:
#'   - `value`: final accumulator
#'   - `failures`: any permanently failed chunks
#'   - `diagnostics`: run telemetry including reduction stats
#'   - `queue_status`, `pool_stats`
#' @export
#' @examples
#' \donttest{
#' res <- shard_reduce(
#'   100L,
#'   map = function(s) sum(s$idx),
#'   combine = function(acc, x) acc + x,
#'   init = 0,
#'   workers = 2
#' )
#' pool_stop()
#' res$value
#' }
shard_reduce <- function(shards,
                         map,
                         combine,
                         init,
                         borrow = list(),
                         out = list(),
                         workers = NULL,
                         chunk_size = 1L,
                         profile = c("default", "memory", "speed"),
                         mem_cap = "2GB",
                         recycle = TRUE,
                         cow = c("deny", "audit", "allow"),
                         seed = NULL,
                         diagnostics = TRUE,
                         packages = NULL,
                         init_expr = NULL,
                         timeout = 3600,
                         max_retries = 3L,
                         health_check_interval = 10L) {
  profile <- match.arg(profile)
  cow <- match.arg(cow)

  if (!is.function(map)) stop("map must be a function", call. = FALSE)
  if (!is.function(combine)) stop("combine must be a function", call. = FALSE)

  start_time <- Sys.time()
  diag <- if (diagnostics) {
    list(
      start_time = start_time,
      kernel = NULL,
      reduce = list(partials = 0L, combine_steps = 0L, partial_max_bytes = 0)
    )
  } else NULL

  # Convert integer to shard_descriptor if needed
  if (is.numeric(shards) && length(shards) == 1) {
    shards <- shards(as.integer(shards), workers = workers)
  }
  if (!inherits(shards, "shard_descriptor")) {
    stop("shards must be a shard_descriptor or integer", call. = FALSE)
  }

  # Determine worker count
  if (is.null(workers)) {
    pool <- pool_get()
    workers <- if (!is.null(pool)) pool$n else max(parallel::detectCores() - 1L, 1L)
  }
  workers <- max(as.integer(workers), 1L)

  # Profile settings
  settings <- get_profile_settings(profile, mem_cap, recycle)

  # Ensure pool exists
  pool <- ensure_pool(workers, settings$mem_cap, settings$rss_drift_threshold, packages, init_expr)

  # Validate and export borrow/out
  borrow <- validate_borrow(borrow, cow)
  auto_shared_names <- attr(borrow, "auto_shared")
  if (length(auto_shared_names) > 0) {
    on.exit({
      for (nm in auto_shared_names) {
        tryCatch(close(borrow[[nm]]), error = function(e) NULL)
      }
    }, add = TRUE)
  }
  out <- validate_out(out)
  export_borrow_to_workers(pool, borrow)
  export_out_to_workers(pool, out)

  if (!is.null(seed)) {
    set_worker_seeds(pool, as.integer(seed), shards$num_shards)
  }

  # Create chunk descriptors. We reuse the same chunk format as shard_map, but
  # the executor will reduce inside the worker and return a single partial.
  chunks <- create_shard_chunks(shards, chunk_size, fun = map, borrow = borrow, out = out, kernel_meta = NULL)

  # Worker-side chunk reducer.
  chunk_reducer <- local({
    combine_fun <- combine
    init_val <- init
    function(chunk) {
      borrow <- if (exists(".shard_borrow", envir = globalenv())) {
        get(".shard_borrow", envir = globalenv())
      } else {
        list()
      }

      out_desc <- if (exists(".shard_out", envir = globalenv())) {
        get(".shard_out", envir = globalenv())
      } else {
        list()
      }

      out <- list()
      if (length(out_desc) > 0) {
        if (!exists(".shard_out_opened", envir = .shard_worker_env, inherits = FALSE)) {
          assign(".shard_out_opened", new.env(parent = emptyenv()), envir = .shard_worker_env)
        }
        opened <- get(".shard_out_opened", envir = .shard_worker_env)
        for (nm in names(out_desc)) {
          if (!exists(nm, envir = opened, inherits = FALSE)) {
            d <- out_desc[[nm]]
            if (is.null(d$kind) || identical(d$kind, "buffer")) {
              opened[[nm]] <- buffer_open(
                path = d$path,
                type = d$type,
                dim = d$dim,
                backing = d$backing,
                readonly = FALSE
              )
            } else if (identical(d$kind, "table_buffer")) {
              cols <- list()
              for (cn in names(d$columns)) {
                cd <- d$columns[[cn]]
                cols[[cn]] <- buffer_open(
                  path = cd$path,
                  type = cd$type,
                  dim = cd$dim,
                  backing = cd$backing,
                  readonly = FALSE
                )
              }
              opened[[nm]] <- structure(
                list(schema = d$schema, nrow = as.integer(d$nrow), backing = d$backing, columns = cols),
                class = "shard_table_buffer"
              )
            } else if (identical(d$kind, "table_sink")) {
              opened[[nm]] <- structure(
                list(schema = d$schema, mode = d$mode, path = d$path, format = d$format),
                class = "shard_table_sink"
              )
            } else {
              stop("Unsupported out descriptor kind: ", d$kind, call. = FALSE)
            }
          }
          out[[nm]] <- opened[[nm]]
        }
      }

      map_fun <- chunk$fun
      borrow_names <- chunk$borrow_names
      out_names <- chunk$out_names

      acc <- init_val
      for (shard in chunk$shards) {
        args <- list(shard)
        for (name in borrow_names) args[[name]] <- borrow[[name]]
        for (name in out_names) args[[name]] <- out[[name]]
        val <- do.call(map_fun, args, quote = TRUE)
        acc <- combine_fun(acc, val)
      }
      acc
    }
  })

  # Master-side streaming reducer: combine partials as chunks complete.
  acc <- init
  partial_bytes_max <- 0
  partial_count <- 0L
  combine_steps <- 0L

  on_partial <- function(tag, value, worker_id) {
    partial_count <<- partial_count + 1L
    combine_steps <<- combine_steps + 1L
    bs <- tryCatch(as.double(utils::object.size(value)), error = function(e) 0)
    if (is.finite(bs) && bs > partial_bytes_max) partial_bytes_max <<- bs
    acc <<- combine(acc, value)
    invisible(NULL)
  }

  dispatch_result <- dispatch_chunks(
    chunks,
    chunk_reducer,
    pool = pool,
    health_check_interval = health_check_interval,
    max_retries = max_retries,
    timeout = timeout,
    on_result = on_partial,
    store_results = FALSE,
    retain_chunks = FALSE
  )

  if (diagnostics) {
    diag$end_time <- Sys.time()
    diag$duration <- as.numeric(difftime(diag$end_time, diag$start_time, units = "secs"))
    diag$reduce$partials <- partial_count
    diag$reduce$combine_steps <- combine_steps
    diag$reduce$partial_max_bytes <- partial_bytes_max
    # Plumb runtime stats (views/scratch/scheduler/etc)
    diag$view_stats <- dispatch_result$diagnostics$view_stats %||% NULL
    diag$copy_stats <- dispatch_result$diagnostics$copy_stats %||% NULL
    diag$table_stats <- dispatch_result$diagnostics$table_stats %||% NULL
    diag$scratch_stats <- dispatch_result$diagnostics$scratch_stats %||% NULL
    diag$scheduler <- dispatch_result$diagnostics$scheduler %||% NULL
  }

  structure(
    list(
      value = acc,
      failures = dispatch_result$failures,
      shards = shards,
      diagnostics = diag,
      queue_status = dispatch_result$queue_status,
      pool_stats = dispatch_result$pool_stats,
      cow_policy = cow,
      profile = profile
    ),
    class = "shard_reduce_result"
  )
}

#' Print a shard_reduce_result Object
#'
#' @param x A \code{shard_reduce_result} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @export
#' @examples
#' \donttest{
#' res <- shard_reduce(4L, map = function(s) sum(s$idx),
#'   combine = `+`, init = 0, workers = 2)
#' pool_stop()
#' print(res)
#' }
print.shard_reduce_result <- function(x, ...) {
  cat("shard_reduce result\n")
  if (!is.null(x$diagnostics)) {
    cat("  Duration:", sprintf("%.2f seconds", x$diagnostics$duration %||% NA_real_), "\n")
    r <- x$diagnostics$reduce %||% list()
    cat("  Partials:", r$partials %||% 0L, "\n")
    cat("  Partial max bytes:", format_bytes(r$partial_max_bytes %||% 0), "\n")
  }
  if (length(x$failures) > 0) {
    cat("  Failed chunks:", length(x$failures), "\n")
  }
  invisible(x)
}

#' @title Individual Worker Control
#' @description Spawn, monitor, and control individual R worker processes.
#' @name worker
NULL

#' Spawn a Worker Process
#'
#' Creates a new R worker process using parallel sockets.
#'
#' @param id Integer. Worker identifier.
#' @param init_expr Expression to evaluate on startup.
#' @param packages Character vector. Packages to load.
#'
#' @return A `shard_worker` object.
#' @keywords internal
#' @noRd
worker_spawn <- function(id, init_expr = NULL, packages = NULL, dev_path = NULL) {
  # Use makeCluster to spawn a single worker
  # This creates a socket connection to an R process
  # Use --vanilla to avoid user/site profiles that may pre-load packages and
  # interfere with deterministic worker setup (notably ALTREP class methods).
  cl <- tryCatch(
    parallel::makeCluster(1L, type = "PSOCK", outfile = "", rscript_args = "--vanilla"),
    error = function(e) {
      msg <- conditionMessage(e)
      if (grepl("all.*connections are in use", msg, ignore.case = TRUE)) {
        cnd <- simpleError(
          paste0("Cannot spawn worker: ", msg,
                 ". Consider reducing workers or calling pool_stop() to free connections."),
          call = sys.call(-1L)
        )
        class(cnd) <- c("shard_connections_exhausted", class(cnd))
        stop(cnd)
      }
      stop(e)
    }
  )
  node <- cl[[1]]

  # Get the PID of the worker
  pid <- tryCatch(
    parallel::clusterCall(cl, function() Sys.getpid())[[1]],
    error = function(e) NA_integer_
  )

  # Ensure workers can load the same library paths as the master process.
  # This is critical for ALTREP class registration and for loading this
  # in-development package when tests install to a temp lib.
  master_libpaths <- .libPaths()
  parallel::clusterCall(cl, function(paths, dev_path) {
    .libPaths(paths)

    # If shard was loaded via user/site profiles (or other bootstrapping),
    # unload it so we re-load from the intended library path.
    if ("package:shard" %in% search()) {
      try(detach("package:shard", unload = TRUE, character.only = TRUE), silent = TRUE)
    }
    if ("shard" %in% loadedNamespaces()) {
      try(unloadNamespace("shard"), silent = TRUE)
    }

    if (!is.null(dev_path) && nzchar(dev_path)) {
      # In dev/test runs, ensure workers run the same in-tree code as the master.
      #
      # compile = FALSE avoids requiring system build tools as long as an
      # up-to-date `src/shard.so` is present (e.g. from a prior build).
      if (requireNamespace("pkgload", quietly = TRUE)) {
        pkgload::load_all(dev_path, quiet = TRUE, compile = FALSE, recompile = FALSE)
      } else {
        suppressPackageStartupMessages(library(shard))
      }
    } else {
      suppressPackageStartupMessages(library(shard))
    }
    TRUE
  }, master_libpaths, dev_path)

  # Load packages if specified
  if (length(packages) > 0) {
    parallel::clusterCall(cl, function(pkgs) {
      for (pkg in pkgs) {
        if (identical(pkg, "shard")) next
        suppressPackageStartupMessages(library(pkg, character.only = TRUE))
      }
    }, packages)
  }

  # Run init expression if specified
  if (!is.null(init_expr)) {
    parallel::clusterCall(cl, function(expr) {
      eval(expr, envir = globalenv())
    }, init_expr)
  }

  structure(
    list(
      id = id,
      pid = pid,
      cluster = cl,
      node = node,
      spawned_at = Sys.time(),
      rss_baseline = NA_real_,
      recycle_count = 0L,
      needs_recycle = FALSE,
      status = "ok"
    ),
    class = "shard_worker"
  )
}

#' Check if Worker is Alive
#'
#' Tests whether the worker process is still running.
#'
#' @param worker A `shard_worker` object.
#' @return Logical. TRUE if worker is alive.
#' @keywords internal
#' @noRd
worker_is_alive <- function(worker) {
  if (is.null(worker) || is.null(worker$cluster)) {
    return(FALSE)
  }

  # Prefer a PID-level check: cluster round-trips can fail when the worker is
  # legitimately busy (e.g. during async sendCall/recv dispatch).
  if (!is.na(worker$pid) && isTRUE(pid_is_alive(worker$pid))) {
    return(TRUE)
  }

  # Fallback: try a simple ping.
  tryCatch({
    result <- parallel::clusterCall(worker$cluster, function() TRUE)
    isTRUE(result[[1]])
  }, error = function(e) FALSE)
}

#' Get Worker RSS (Resident Set Size)
#'
#' Queries the worker's memory usage via the ps package if available,
#' otherwise falls back to /proc on Linux or ps command.
#'
#' @param worker A `shard_worker` object.
#' @return Numeric. RSS in bytes, or NA if unavailable.
#' @keywords internal
#' @noRd
worker_rss <- function(worker) {
  if (is.null(worker) || is.na(worker$pid)) {
    return(NA_real_)
  }

  rss_get_pid(worker$pid)
}

worker_connection_ <- function(worker) {
  if (is.null(worker) || is.null(worker$cluster) || length(worker$cluster) < 1L) {
    return(NULL)
  }

  node <- worker$cluster[[1]]
  if (is.null(node)) {
    return(NULL)
  }

  node$con %||% NULL
}

worker_close_connection_ <- function(worker) {
  con <- worker_connection_(worker)
  if (!is.null(con)) {
    try(close(con), silent = TRUE)
  }

  invisible(NULL)
}

#' Kill a Worker Process
#'
#' Terminates the worker process and closes connections.
#'
#' @param worker A `shard_worker` object.
#' @param graceful Logical scalar. If `TRUE`, request an orderly cluster
#'   shutdown before falling back to a process kill. If `FALSE`, skip the
#'   cluster stop step and force termination directly.
#' @return NULL (invisibly).
#' @keywords internal
#' @noRd
worker_kill <- function(worker, graceful = TRUE) {
  if (is.null(worker)) {
    return(invisible(NULL))
  }

  wait_dead <- function(pid, timeout = 2) {
    if (is.na(pid) || !isTRUE(pid_is_alive(pid))) return(invisible(NULL))
    deadline <- Sys.time() + timeout
    while (Sys.time() < deadline) {
      if (!isTRUE(pid_is_alive(pid))) break
      Sys.sleep(0.05)
    }
    invisible(NULL)
  }

  pid <- worker$pid %||% NA_integer_
  alive <- !is.na(pid) && isTRUE(pid_is_alive(pid))

  # Best-effort shutdown via parallel, but always fall back to PID kill and
  # always close the socket connection.
  if (alive && isTRUE(graceful)) {
    tryCatch(parallel::stopCluster(worker$cluster), error = function(e) NULL)
  }

  if (alive && isTRUE(pid_is_alive(pid))) {
    tryCatch(tools::pskill(pid, signal = 15L), error = function(e) NULL)
    wait_dead(pid, timeout = 1)
    if (isTRUE(pid_is_alive(pid))) {
      tryCatch(tools::pskill(pid, signal = 9L), error = function(e) NULL)
      wait_dead(pid, timeout = 1)
    }
  }

  worker_close_connection_(worker)

  invisible(NULL)
}

#' Recycle a Worker
#'
#' Kills the current worker and spawns a fresh replacement.
#' The new worker inherits the same ID but has a fresh R process.
#'
#' @param worker A `shard_worker` object.
#' @param init_expr Expression to evaluate on startup.
#' @param packages Character vector. Packages to load.
#' @return A new `shard_worker` object.
#' @keywords internal
#' @noRd
worker_recycle <- function(worker, init_expr = NULL, packages = NULL, dev_path = NULL) {
  id <- worker$id
  old_recycle_count <- worker$recycle_count %||% 0L

  # Kill the old worker

  worker_kill(worker)

  # Spawn a fresh one
  new_worker <- worker_spawn(id, init_expr, packages, dev_path = dev_path)
  new_worker$recycle_count <- old_recycle_count + 1L

  new_worker
}

#' Evaluate Expression in Worker
#'
#' Sends an expression to the worker for evaluation.
#'
#' @param worker A `shard_worker` object.
#' @param expr Expression to evaluate.
#' @param envir Environment containing variables needed by expr.
#' @param timeout Numeric. Seconds to wait for result.
#' @return The result of evaluation.
#' @keywords internal
#' @noRd
worker_eval <- function(worker, expr, envir = parent.frame(), timeout = 3600) {
  if (!worker_is_alive(worker)) {
    stop("Worker ", worker$id, " is not alive", call. = FALSE)
  }

  timeout <- as.double(timeout)
  if (is.na(timeout) || timeout < 0) {
    timeout <- 3600
  }

  # Capture variables from environment that are referenced in expr
  expr_vars <- all.vars(expr)
  export_env <- new.env(parent = emptyenv())
  for (v in expr_vars) {
    if (exists(v, envir = envir, inherits = TRUE)) {
      export_env[[v]] <- get(v, envir = envir, inherits = TRUE)
    }
  }

  # Send to worker
  tryCatch({
    # Export variables
    if (length(ls(export_env)) > 0) {
      parallel::clusterExport(worker$cluster, ls(export_env), envir = export_env)
    }

    con <- worker_connection_(worker)
    if (is.null(con) || !isOpen(con)) {
      stop("worker connection is not open", call. = FALSE)
    }

    parallel_sendCall <- utils::getFromNamespace("sendCall", "parallel")
    parallel_sendCall(
      worker$cluster[[1]],
      fun = function(e) {
        tryCatch(
          list(ok = TRUE, value = eval(e, envir = globalenv())),
          error = function(err) list(ok = FALSE, error = conditionMessage(err))
        )
      },
      args = list(expr),
      return = TRUE,
      tag = sprintf("worker_eval_%d", worker$id)
    )

    ready <- tryCatch(socketSelect(list(con), timeout = timeout), error = function(e) FALSE)
    if (!isTRUE(any(ready))) {
      worker_kill(worker, graceful = FALSE)
      stop(
        sprintf("Worker %d evaluation timed out after %.3f seconds", worker$id, timeout),
        call. = FALSE
      )
    }

    recv_result <- utils::getFromNamespace("recvResult", "parallel")
    payload <- tryCatch(
      recv_result(worker$cluster[[1]]),
      error = function(e) {
        worker_kill(worker)
        stop("recv failed: ", conditionMessage(e), call. = FALSE)
      }
    )

    if (is.list(payload) && isTRUE(payload$ok)) {
      return(payload$value)
    }

    err_msg <- if (is.list(payload)) payload$error %||% "unknown worker error" else "invalid worker response"
    stop(err_msg, call. = FALSE)
  }, error = function(e) {
    stop("Worker ", worker$id, " evaluation failed: ", conditionMessage(e), call. = FALSE)
  })
}

#' Get Worker Metrics
#'
#' Returns current metrics for a worker.
#'
#' @param worker A `shard_worker` object.
#' @return A list of worker metrics.
#' @keywords internal
#' @noRd
worker_metrics <- function(worker) {
  alive <- worker_is_alive(worker)
  rss <- if (alive) worker_rss(worker) else NA_real_

  list(
    worker_id = worker$id,
    pid = worker$pid,
    host = "localhost",
    status = if (alive) "ok" else "dead",
    rss_bytes = rss,
    rss_baseline = worker$rss_baseline,
    recycle_count = worker$recycle_count %||% 0L,
    spawned_at = worker$spawned_at
  )
}

#' Print a shard_worker Object
#'
#' @param x A \code{shard_worker} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @examples
#' \donttest{
#' p <- pool_create(1)
#' print(p$workers[[1]])
#' pool_stop(p)
#' }
#' @export
print.shard_worker <- function(x, ...) {
  cat("shard worker [", x$id, "]\n", sep = "")
  cat("  PID:", x$pid, "\n")
  cat("  Status:", if (worker_is_alive(x)) "alive" else "dead", "\n")
  cat("  RSS baseline:", format_bytes(x$rss_baseline), "\n")
  cat("  Recycles:", x$recycle_count %||% 0L, "\n")
  cat("  Spawned:", format(x$spawned_at), "\n")
  invisible(x)
}

#' Check if PID is Alive
#'
#' @param pid Process ID to check.
#' @return Logical.
#' @keywords internal
#' @noRd
pid_is_alive <- function(pid) {
  if (is.na(pid)) return(FALSE)

  # Use ps package if available
  if (requireNamespace("ps", quietly = TRUE)) {
    tryCatch({
      p <- ps::ps_handle(pid)
      ps::ps_is_running(p)
    }, error = function(e) FALSE)
  } else {
    # Fallback: try to send signal 0
    tryCatch({
      tools::pskill(pid, signal = 0L)
      TRUE
    }, error = function(e) FALSE)
  }
}

# Shared-memory dispatch mode (shm_queue).
#
# This avoids per-task RPC by running a single long-lived loop per worker that
# pulls task ids from a shared segment. This mode is intentionally limited:
# - tasks are identified by integer ids only
# - results are not gathered (use explicit output buffers)

open_out_from_desc_ <- function(out_desc) {
  # Lazily open output buffers once per worker process and cache them.
  out <- list()
  if (length(out_desc) == 0) return(out)

  out_desc_key_ <- function(d) {
    kind <- d$kind %||% "buffer"
    if (identical(kind, "buffer")) {
      return(paste0("buffer|", d$path, "|", d$backing, "|", d$type, "|", paste(d$dim, collapse = "x")))
    }
    if (identical(kind, "table_buffer")) {
      col_keys <- vapply(d$columns, function(cd) {
        paste0(cd$path, "|", cd$backing, "|", cd$type, "|", paste(cd$dim, collapse = "x"))
      }, character(1))
      return(paste0("table_buffer|", paste(names(col_keys), col_keys, sep = "=", collapse = ";")))
    }
    if (identical(kind, "table_sink")) {
      return(paste0("table_sink|", d$path, "|", d$format, "|", d$mode))
    }
    paste0("unknown|", kind)
  }

  open_one_ <- function(d) {
    kind <- d$kind %||% "buffer"
    if (identical(kind, "buffer")) {
      buf <- buffer_open(
        path = d$path,
        type = d$type,
        dim = d$dim,
        backing = d$backing,
        readonly = FALSE
      )
      return(buf)
    }
    if (identical(kind, "table_buffer")) {
      cols <- list()
      for (cn in names(d$columns)) {
        cd <- d$columns[[cn]]
        cols[[cn]] <- buffer_open(
          path = cd$path,
          type = cd$type,
          dim = cd$dim,
          backing = cd$backing,
          readonly = FALSE
        )
      }
      tb <- structure(
        list(schema = d$schema, nrow = as.integer(d$nrow), backing = d$backing, columns = cols),
        class = "shard_table_buffer"
      )
      return(tb)
    }
    if (identical(kind, "table_sink")) {
      ts <- structure(
        list(schema = d$schema, mode = d$mode, path = d$path, format = d$format),
        class = "shard_table_sink"
      )
      return(ts)
    }
    stop("Unsupported out descriptor kind: ", kind, call. = FALSE)
  }

  close_one_ <- function(obj) {
    if (inherits(obj, "shard_buffer")) {
      tryCatch(buffer_close(obj), error = function(e) NULL)
      return(invisible(NULL))
    }
    if (inherits(obj, "shard_table_buffer")) {
      if (!is.null(obj$columns) && is.list(obj$columns)) {
        for (col in obj$columns) {
          if (inherits(col, "shard_buffer")) {
            tryCatch(buffer_close(col), error = function(e) NULL)
          }
        }
      }
      return(invisible(NULL))
    }
    invisible(NULL)
  }

  if (!exists(".shard_out_opened", envir = .shard_worker_env, inherits = FALSE)) {
    assign(".shard_out_opened", new.env(parent = emptyenv()), envir = .shard_worker_env)
  }
  opened <- get(".shard_out_opened", envir = .shard_worker_env)

  for (nm in names(out_desc)) {
    d <- out_desc[[nm]]
    want_key <- out_desc_key_(d)
    entry <- if (exists(nm, envir = opened, inherits = FALSE)) opened[[nm]] else NULL
    cur_key <- NULL
    cur_obj <- NULL
    if (!is.null(entry) && is.list(entry) && !is.null(entry$key) && !is.null(entry$obj)) {
      cur_key <- entry$key
      cur_obj <- entry$obj
    } else if (!is.null(entry)) {
      cur_obj <- entry
    }

    if (is.null(cur_obj) || !identical(cur_key, want_key)) {
      if (!is.null(cur_obj)) close_one_(cur_obj)
      new_obj <- open_one_(d)
      opened[[nm]] <- list(key = want_key, obj = new_obj)
    }

    out[[nm]] <- opened[[nm]]$obj
  }

  out
}

worker_shm_queue_loop_ <- function(queue_desc,
                                  worker_id,
                                  mode,
                                  n,
                                  block_size,
                                  fun,
                                  borrow_names,
                                  out_names,
                                  max_retries,
                                  error_log = FALSE,
                                  error_log_max_lines = 100L,
                                  poll_sleep = 0.0005) {
  # Open queue segment (writable).
  seg <- segment_open(queue_desc$path, backing = queue_desc$backing, readonly = FALSE)
  on.exit(try(segment_close(seg, unlink = FALSE), silent = TRUE), add = TRUE)

  # Get borrow/out descriptors from worker environment.
  borrow <- if (exists(".shard_borrow", envir = globalenv())) {
    get(".shard_borrow", envir = globalenv())
  } else {
    list()
  }
  out_desc <- if (exists(".shard_out", envir = globalenv())) {
    get(".shard_out", envir = globalenv())
  } else {
    list()
  }
  out <- open_out_from_desc_(out_desc)

  mode <- as.character(mode %||% "scalar_n")
  block_size <- as.integer(block_size)
  n <- as.integer(n)

  error_log <- isTRUE(error_log)
  error_log_max_lines <- as.integer(error_log_max_lines)
  if (is.na(error_log_max_lines) || error_log_max_lines < 0L) error_log_max_lines <- 0L

  err_con <- NULL
  err_path <- NULL
  err_written <- 0L
  if (error_log && error_log_max_lines > 0L) {
    err_path <- file.path(tempdir(), sprintf("shard_shm_queue_errors_pid%d_w%d.log", Sys.getpid(), worker_id))
    err_con <- file(err_path, open = "a", encoding = "UTF-8")
    on.exit(try(close(err_con), silent = TRUE), add = TRUE)

    assign(".shard_shm_queue_error_log_path", err_path, envir = .shard_worker_env)
    assign(".shard_shm_queue_error_count", 0L, envir = .shard_worker_env)
  }

  repeat {
    st <- taskq_stats(seg)
    if ((st$done %||% 0L) + (st$failed %||% 0L) >= (st$n_tasks %||% 0L)) break

    task_id <- as.integer(taskq_claim(seg, worker_id))
    if (is.na(task_id) || task_id < 1L) {
      Sys.sleep(poll_sleep)
      next
    }

    shard <- NULL
    if (identical(mode, "scalar_n")) {
      start <- (task_id - 1L) * block_size + 1L
      end <- min(task_id * block_size, n)
      shard <- list(id = task_id, start = start, end = end, idx = start:end, len = end - start + 1L)
    } else if (identical(mode, "descriptor")) {
      # Fetch shard descriptors once per worker process.
      shards_key <- if (exists(".shard_shm_queue_shards_key", envir = globalenv(), inherits = FALSE)) {
        get(".shard_shm_queue_shards_key", envir = globalenv())
      } else {
        NULL
      }
      fetched_key <- if (exists(".shard_shm_queue_shards_fetched_key", envir = .shard_worker_env, inherits = FALSE)) {
        get(".shard_shm_queue_shards_fetched_key", envir = .shard_worker_env)
      } else {
        NULL
      }
      if (!exists(".shard_shm_queue_shards_fetched", envir = .shard_worker_env, inherits = FALSE) ||
          !identical(fetched_key, shards_key)) {
        shards_handle <- if (exists(".shard_shm_queue_shards", envir = globalenv(), inherits = FALSE)) {
          get(".shard_shm_queue_shards", envir = globalenv())
        } else {
          NULL
        }
        if (is.null(shards_handle)) stop("shm_queue descriptor mode missing .shard_shm_queue_shards", call. = FALSE)
        assign(".shard_shm_queue_shards_fetched", fetch(shards_handle), envir = .shard_worker_env)
        assign(".shard_shm_queue_shards_fetched_key", shards_key, envir = .shard_worker_env)
      }

      shards_list <- get(".shard_shm_queue_shards_fetched", envir = .shard_worker_env)
      shard <- shards_list[[task_id]]
      if (is.null(shard$id)) shard$id <- task_id
      if (is.null(shard$len)) {
        if (!is.null(shard$idx)) shard$len <- length(shard$idx)
        else if (!is.null(shard$start) && !is.null(shard$end)) shard$len <- as.integer(shard$end - shard$start + 1L)
      }
      if (is.null(shard$idx) && !is.null(shard$start) && !is.null(shard$end)) {
        shard$idx <- as.integer(shard$start):as.integer(shard$end)
      }
    } else {
      stop("Unsupported shm_queue mode: ", mode, call. = FALSE)
    }

    args <- list(shard)
    for (nm in borrow_names) args[[nm]] <- borrow[[nm]]
    for (nm in out_names) args[[nm]] <- out[[nm]]

    ok <- TRUE
    err_msg <- NULL
    tryCatch(
      do.call(fun, args, quote = TRUE),
      error = function(e) {
        ok <<- FALSE
        err_msg <<- conditionMessage(e)
        invisible(NULL)
      }
    )

    if (ok) {
      taskq_done(seg, task_id)
    } else {
      if (!is.null(err_con) && err_written < error_log_max_lines) {
        # Keep logs compact and parseable; avoid timestamps to keep tests stable.
        writeLines(paste0(task_id, "\t", err_msg %||% "error"), con = err_con, sep = "\n")
        err_written <- err_written + 1L
        assign(".shard_shm_queue_error_count", err_written, envir = .shard_worker_env)
      }
      taskq_error(seg, task_id, max_retries = max_retries)
    }
  }

  # Return final queue stats (small).
  taskq_stats(seg)
}

dispatch_shards_shm_queue_ <- function(n,
                                      block_size,
                                      shards = NULL,
                                      fun,
                                      borrow,
                                      out,
                                      pool,
                                      max_retries,
                                      timeout = 3600,
                                      queue_backing = c("mmap", "shm"),
                                      error_log = FALSE,
                                      error_log_max_lines = 100L) {
  n <- as.integer(n)
  block_size <- as.integer(block_size)
  if (is.na(n) || n < 1L) stop("n must be >= 1", call. = FALSE)
  if (is.na(block_size) || block_size < 1L) stop("block_size must be >= 1", call. = FALSE)
  queue_backing <- match.arg(queue_backing)
  error_log <- isTRUE(error_log)
  error_log_max_lines <- as.integer(error_log_max_lines)
  if (is.na(error_log_max_lines) || error_log_max_lines < 0L) error_log_max_lines <- 0L

  mode <- if (is.null(shards)) "scalar_n" else "descriptor"
  if (!is.null(shards) && !inherits(shards, "shard_descriptor")) {
    stop("shards must be a shard_descriptor (or NULL)", call. = FALSE)
  }

  n_tasks <- if (identical(mode, "scalar_n")) {
    as.integer(ceiling(n / block_size))
  } else {
    as.integer(shards$num_shards %||% length(shards$shards))
  }

  q <- taskq_create(n_tasks, backing = queue_backing)
  qdesc <- q$desc
  seg_master <- q$owner
  on.exit(try(segment_close(seg_master, unlink = TRUE), silent = TRUE), add = TRUE)

  borrow_names <- names(borrow)
  out_names <- names(out)

  shards_shared <- NULL
  shards_key <- NULL
  if (identical(mode, "descriptor")) {
    shards_shared <- share(shards$shards, backing = queue_backing)
    shards_key <- shards_shared$path %||% NULL
  }

  # Start one long-lived loop per worker. If a worker dies, requeue its claims
  # and restart the loop on a fresh worker.
  parallel_sendCall <- utils::getFromNamespace("sendCall", "parallel")

  idle <- rep(TRUE, pool$n)
  started <- rep(FALSE, pool$n)
  inflight <- vector("list", pool$n)

  start_loop <- function(worker_id) {
    w <- pool$workers[[worker_id]]
    if (is.null(w) || !worker_is_alive(w)) {
      pool <<- pool_restart_worker_(pool, worker_id)
      w <- pool$workers[[worker_id]]

      # Re-export borrow/out to the restarted worker.
      export_borrow_to_workers(pool, borrow)
      if (length(out) > 0) export_out_to_workers(pool, out)
      if (identical(mode, "descriptor")) {
        export_env <- new.env(parent = emptyenv())
        export_env$.shard_shm_queue_shards <- shards_shared
        export_env$.shard_shm_queue_shards_key <- shards_key
        tryCatch(parallel::clusterExport(w$cluster, c(".shard_shm_queue_shards", ".shard_shm_queue_shards_key"), envir = export_env),
                 error = function(e) NULL)
      }
    }

    if (identical(mode, "descriptor")) {
      export_env <- new.env(parent = emptyenv())
      export_env$.shard_shm_queue_shards <- shards_shared
      export_env$.shard_shm_queue_shards_key <- shards_key
      tryCatch(parallel::clusterExport(w$cluster, c(".shard_shm_queue_shards", ".shard_shm_queue_shards_key"), envir = export_env),
               error = function(e) NULL)
    }

    parallel_sendCall(
      w$cluster[[1]],
      fun = worker_shm_queue_loop_,
      args = list(qdesc, worker_id, mode, n, block_size, fun, borrow_names, out_names, max_retries, error_log, error_log_max_lines),
      return = TRUE,
      tag = paste0("shm_queue_", worker_id)
    )
    started[worker_id] <<- TRUE
    idle[worker_id] <<- FALSE
    inflight[[worker_id]] <<- list(start_time = Sys.time())
  }

  # Kick off loops.
  for (i in seq_len(pool$n)) start_loop(i)

  # Wait for completion. Use queue stats to decide done.
  deadline <- Sys.time() + timeout
  repeat {
    st <- taskq_stats(seg_master)
    if ((st$done %||% 0L) + (st$failed %||% 0L) >= (st$n_tasks %||% 0L)) break
    if (Sys.time() > deadline) {
      for (i in seq_len(pool$n)) {
        if (isTRUE(started[i])) {
          tryCatch(taskq_reset_claims(seg_master, i), error = function(e) NULL)
        }
        w <- pool$workers[[i]]
        if (!is.null(w)) {
          tryCatch(worker_kill(w, graceful = FALSE), error = function(e) NULL)
        }
        idle[i] <- TRUE
        started[i] <- FALSE
        inflight[[i]] <- NULL
      }
      .pool_env$pool <- pool
      stop("shm_queue dispatch timed out", call. = FALSE)
    }

    # Restart dead workers.
    for (i in seq_len(pool$n)) {
      w <- pool$workers[[i]]
      if (!is.null(w) && started[i] && !worker_is_alive(w)) {
        # Requeue tasks claimed by this worker (best-effort).
        taskq_reset_claims(seg_master, i)
        idle[i] <- TRUE
        started[i] <- FALSE
        inflight[[i]] <- NULL
        start_loop(i)
      }
    }

    Sys.sleep(0.01)
  }

  # Drain the single return value from each worker loop to keep PSOCK
  # connections clean (avoid "unused connection" warnings).
  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    if (is.null(w) || is.null(w$cluster) || length(w$cluster) < 1L) next
    n <- w$cluster[[1]]
    con <- n$con %||% NULL
    if (is.null(con) || !isOpen(con)) next
    ready <- tryCatch(socketSelect(list(con), timeout = 0.1), error = function(e) FALSE)
    if (isTRUE(ready)) {
      tryCatch(unserialize(con), error = function(e) NULL)
    }
  }

  # Aggregate per-worker counters once (cheap).
  view_stats <- list(created = 0L, materialized = 0L, materialized_bytes = 0, packed = 0L, packed_bytes = 0)
  copy_stats <- list(borrow_exports = 0L, borrow_bytes = 0, buffer_writes = 0L, buffer_bytes = 0)
  table_stats <- list(writes = 0L, rows = 0L, bytes = 0)
  scratch_stats <- list(hits = 0L, misses = 0L, high_water = 0)
  error_logs <- list()

  for (i in seq_len(pool$n)) {
    w <- pool$workers[[i]]
    if (is.null(w) || !worker_is_alive(w)) next
    vals <- tryCatch(
      parallel::clusterCall(w$cluster, function() {
        list(
          view = tryCatch(view_diagnostics(), error = function(e) NULL),
          buf = tryCatch(buffer_diagnostics(), error = function(e) NULL),
          table = tryCatch(table_diagnostics(), error = function(e) NULL),
          scratch = tryCatch(scratch_diagnostics(), error = function(e) NULL),
          err_path = if (exists(".shard_shm_queue_error_log_path", envir = .shard_worker_env, inherits = FALSE)) {
            get(".shard_shm_queue_error_log_path", envir = .shard_worker_env)
          } else {
            NULL
          },
          err_count = if (exists(".shard_shm_queue_error_count", envir = .shard_worker_env, inherits = FALSE)) {
            get(".shard_shm_queue_error_count", envir = .shard_worker_env)
          } else {
            NULL
          }
        )
      })[[1]],
      error = function(e) NULL
    )
    if (is.null(vals)) next
    if (is.list(vals$view)) {
      view_stats$created <- view_stats$created + (vals$view$created %||% 0L)
      view_stats$materialized <- view_stats$materialized + (vals$view$materialized %||% 0L)
      view_stats$materialized_bytes <- view_stats$materialized_bytes + (vals$view$materialized_bytes %||% 0)
      view_stats$packed <- view_stats$packed + (vals$view$packed %||% 0L)
      view_stats$packed_bytes <- view_stats$packed_bytes + (vals$view$packed_bytes %||% 0)
    }
    if (is.list(vals$buf)) {
      copy_stats$buffer_writes <- copy_stats$buffer_writes + (vals$buf$writes %||% 0L)
      copy_stats$buffer_bytes <- copy_stats$buffer_bytes + (vals$buf$bytes %||% 0)
    }
    if (is.list(vals$table)) {
      table_stats$writes <- table_stats$writes + (vals$table$writes %||% 0L)
      table_stats$rows <- table_stats$rows + (vals$table$rows %||% 0L)
      table_stats$bytes <- table_stats$bytes + (vals$table$bytes %||% 0)
    }
    if (is.list(vals$scratch)) {
      scratch_stats$hits <- scratch_stats$hits + (vals$scratch$hits %||% 0L)
      scratch_stats$misses <- scratch_stats$misses + (vals$scratch$misses %||% 0L)
      scratch_stats$high_water <- max(as.double(scratch_stats$high_water), as.double(vals$scratch$high_water %||% 0))
    }
    if (error_log && !is.null(vals$err_path) && is.character(vals$err_path) && length(vals$err_path) == 1L) {
      ec <- as.integer(vals$err_count %||% 0L)
      if (!is.na(ec) && ec > 0L) {
        error_logs[[length(error_logs) + 1L]] <- list(worker_id = i, path = vals$err_path, errors = ec)
      }
    }
  }

  # Collect final results (not gathered) + failures.
  fail <- taskq_failures(seg_master)
  failures <- list()
  if (is.list(fail) && !is.null(fail$task_id)) {
    ids <- as.integer(fail$task_id)
    rcs <- as.integer(fail$retry_count %||% rep.int(NA_integer_, length(ids)))
    for (i in seq_along(ids)) {
      id <- ids[[i]]
      if (is.na(id)) next
      failures[[as.character(id)]] <- list(
        id = id,
        last_error = "shm_queue_failed",
        retry_count = rcs[[i]] %||% NA_integer_
      )
    }
  }

  st_final <- taskq_stats(seg_master)
  total_retries <- as.integer(st_final$retries %||% 0L)

  list(
    # Fire-and-forget: do not allocate a giant placeholder list for large runs.
    # Use a lightweight results placeholder that behaves like a list of NULLs.
    results = structure(list(n = as.integer(n_tasks)), class = "shard_results_placeholder"),
    failures = failures,
    queue_status = list(
      total = n_tasks,
      pending = 0L,
      in_flight = 0L,
      completed = as.integer(st_final$done %||% 0L),
      failed = as.integer(st_final$failed %||% 0L),
      total_retries = total_retries
    ),
    diagnostics = list(
      taskq = st_final,
      view_stats = view_stats,
      copy_stats = copy_stats,
      table_stats = table_stats,
      scratch_stats = scratch_stats,
      error_logs = error_logs
    ),
    pool_stats = pool_get()$stats
  )
}

# Lightweight "no-results" placeholder used by shm_queue dispatch.
#
# This avoids allocating a giant list when results are intentionally not
# gathered (out-buffer workflows). It behaves like a list of NULLs for basic
# indexing and length().

.results_placeholder <- function(n) {
  structure(list(n = as.integer(n)), class = "shard_results_placeholder")
}

#' @param x A \code{shard_results_placeholder} object.
#' @param ... Further arguments (ignored).
#' @return The input \code{x}, invisibly.
#' @export
#' @noRd
print.shard_results_placeholder <- function(x, ...) {
  cat("<shard_results_placeholder>\n")
  cat("  count:", format(x$n, big.mark = ","), "\n")
  cat("  note: results are not gathered in shm_queue mode; use out= buffers/sinks.\n")
  invisible(x)
}

#' @param x A \code{shard_results_placeholder} object.
#' @return An integer scalar giving the number of placeholder results.
#' @export
#' @noRd
length.shard_results_placeholder <- function(x) {
  as.integer(x$n %||% 0L)
}

#' @param x A \code{shard_results_placeholder} object.
#' @param i Index.
#' @param ... Further arguments (ignored).
#' @return Always \code{NULL}; results are not gathered in shm_queue mode.
#' @export
#' @noRd
`[[.shard_results_placeholder` <- function(x, i, ...) {
  # Always NULL; tasks do not return gathered results in shm_queue mode.
  NULL
}

#' @param x A \code{shard_results_placeholder} object.
#' @param i Index or indices.
#' @param ... Further arguments (ignored).
#' @return A subset of the object (a list of \code{NULL} values).
#' @export
#' @noRd
`[.shard_results_placeholder` <- function(x, i, ...) {
  if (missing(i)) i <- seq_len(length(x))
  rep(list(NULL), length(i))
}

test_that("results_placeholder creates valid placeholder", {
  p <- shard:::.results_placeholder(100)

  expect_s3_class(p, "shard_results_placeholder")
  expect_equal(p$n, 100L)
})

test_that("results_placeholder length method works", {
  p <- shard:::.results_placeholder(50)
  expect_equal(length(p), 50L)

  p2 <- shard:::.results_placeholder(0)
  expect_equal(length(p2), 0L)

  p3 <- shard:::.results_placeholder(1000000)
  expect_equal(length(p3), 1000000L)
})

test_that("results_placeholder [[ returns NULL", {
  p <- shard:::.results_placeholder(10)

  expect_null(p[[1]])
  expect_null(p[[5]])
  expect_null(p[[10]])

  # Out of range still returns NULL (by design)
  expect_null(p[[100]])
})

test_that("results_placeholder [ returns list of NULLs", {
  p <- shard:::.results_placeholder(10)

  # Single index
  r <- p[1]
  expect_true(is.list(r))
  expect_length(r, 1)
  expect_null(r[[1]])

  # Multiple indices
  r2 <- p[c(1, 3, 5)]
  expect_true(is.list(r2))
  expect_length(r2, 3)
  expect_true(all(vapply(r2, is.null, logical(1))))

  # All indices
  r3 <- p[1:10]
  expect_length(r3, 10)
  expect_true(all(vapply(r3, is.null, logical(1))))
})

test_that("results_placeholder [ with missing index returns all NULLs", {
  p <- shard:::.results_placeholder(5)

  r <- p[]
  expect_true(is.list(r))
  expect_length(r, 5)
  expect_true(all(vapply(r, is.null, logical(1))))
})

test_that("results_placeholder print method works", {
  p <- shard:::.results_placeholder(1000)

  out <- capture.output(print(p))
  expect_true(any(grepl("<shard_results_placeholder>", out)))
  expect_true(any(grepl("count:", out)))
  expect_true(any(grepl("1,000", out)))  # formatted with comma
  expect_true(any(grepl("shm_queue mode", out)))
})

test_that("results_placeholder handles zero count", {
  p <- shard:::.results_placeholder(0)

  expect_equal(length(p), 0L)

  r <- p[]
  expect_true(is.list(r))
  expect_length(r, 0)
})

test_that("results_placeholder handles NULL count gracefully", {
  # Edge case: NULL n field
  p <- structure(list(n = NULL), class = "shard_results_placeholder")

  expect_equal(length(p), 0L)
})

#' @title ALTREP Shared Vectors
#' @name altrep
#' @description ALTREP-backed zero-copy vectors for shared memory.
#'
#' @details
#' These functions create ALTREP (Alternative Representation) vectors that
#' are backed by shared memory segments. The key benefits are:
#'
#' \itemize{
#'   \item \strong{Zero-copy subsetting}: Contiguous subsets return views
#'     into the same shared memory, not copies.
#'   \item \strong{Diagnostics}: Track when data pointers are accessed or
#'     when vectors are materialized (copied to standard R vectors).
#'   \item \strong{Read-only protection}: Optionally prevent write access
#'     to protect shared data.
#' }
#'
#' Supported types: integer, double/numeric, logical, raw.
#'
#' @useDynLib shard, .registration = TRUE
NULL

#' Create a shared vector from a segment
#'
#' @param segment A shard_segment object
#' @param type Vector type: "integer", "double"/"numeric", "logical", or "raw"
#' @param offset Byte offset into segment (default: 0)
#' @param length Number of elements. If NULL, calculated from segment size.
#' @param readonly If TRUE, prevent write access via DATAPTR (default: TRUE)
#' @param cow Copy-on-write policy for mutation attempts. One of
#'   `"deny"`, `"audit"`, or `"allow"`. If NULL, defaults to `"deny"` when
#'   `readonly=TRUE` and `"allow"` otherwise.
#' @return An ALTREP vector backed by shared memory
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(400)
#' segment_write(seg, 1:100, offset = 0)
#'
#' x <- shared_vector(seg, "integer", length = 100)
#' x[1:10]
#'
#' shared_diagnostics(x)
#' }
shared_vector <- function(segment,
                          type = c("double", "integer", "logical", "raw"),
                          offset = 0,
                          length = NULL,
                          readonly = TRUE,
                          cow = NULL) {
    stopifnot(inherits(segment, "shard_segment"))
    type <- match.arg(type)

    if (is.null(cow)) {
        cow <- if (isTRUE(readonly)) "deny" else "allow"
    }
    cow <- match.arg(cow, c("deny", "audit", "allow"))

    # Calculate element size
    elem_size <- switch(type,
        "integer" = 4L,
        "double"  = 8L,
        "numeric" = 8L,
        "logical" = 4L,  # R logicals are stored as int
        "raw"     = 1L
    )

    # Calculate length if not provided
    if (is.null(length)) {
        seg_size <- segment_size(segment)
        available <- seg_size - offset
        length <- floor(available / elem_size)
    }

    if (length < 0) {
        stop("length must be non-negative")
    }

    .Call("C_shard_altrep_create",
          segment$ptr,
          type,
          as.double(offset),
          as.double(length),
          readonly,
          cow,
          PACKAGE = "shard")
}

#' Create a view (subset) of a shared vector
#'
#' @param x A shard ALTREP vector
#' @param start Start index (1-based, like R)
#' @param length Number of elements
#' @return An ALTREP view into the same shared memory
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(800)
#' segment_write(seg, 1:100, offset = 0)
#' x <- shared_vector(seg, "integer", length = 100)
#'
#' y <- shared_view(x, start = 10, length = 11)
#' y[1]
#' }
shared_view <- function(x, start, length) {
    if (!is_shared_vector(x)) {
        stop("x must be a shard ALTREP vector")
    }

    # Convert to 0-based
    start_0 <- as.double(start - 1)

    .Call("C_shard_altrep_view", x, start_0, as.double(length),
          PACKAGE = "shard")
}

#' Get diagnostics for a shared vector
#'
#' @param x A shard ALTREP vector
#' @return A list with diagnostic information:
#'   \describe{
#'     \item{dataptr_calls}{Number of times DATAPTR was accessed}
#'     \item{materialize_calls}{Number of times vector was copied to standard R vector}
#'     \item{length}{Number of elements}
#'     \item{offset}{Byte offset into underlying segment}
#'     \item{readonly}{Whether write access is prevented}
#'     \item{type}{R type of the vector}
#'   }
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(400)
#' segment_write(seg, 1:100, offset = 0)
#' x <- shared_vector(seg, "integer", length = 100)
#'
#' sum(x)
#'
#' shared_diagnostics(x)
#' }
shared_diagnostics <- function(x) {
    if (!is_shared_vector(x)) {
        stop("x must be a shard ALTREP vector")
    }

    .Call("C_shard_altrep_diagnostics", x, PACKAGE = "shard")
}

#' Check if an object is a shared vector
#'
#' @param x Any R object
#' @return TRUE if x is a shard ALTREP vector, FALSE otherwise
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(400)
#' segment_write(seg, 1:100, offset = 0)
#' x <- shared_vector(seg, "integer", length = 100)
#'
#' is_shared_vector(x)
#' is_shared_vector(1:10)
#' }
is_shared_vector <- function(x) {
    .Call("C_is_shard_altrep", x, PACKAGE = "shard")
}

#' Get the underlying segment from a shared vector
#'
#' @param x A shard ALTREP vector
#' @return A `shard_segment` S3 object wrapping the underlying segment
#' @export
#' @examples
#' \donttest{
#' x <- as_shared(1:100)
#' shared_segment(x)
#' }
shared_segment <- function(x) {
    if (!is_shared_vector(x)) {
        stop("x must be a shard ALTREP vector")
    }

    ptr <- .Call("C_shard_altrep_segment", x, PACKAGE = "shard")
    info <- .Call("C_shard_segment_info", ptr, PACKAGE = "shard")

    structure(
        list(
            ptr = ptr,
            size = info$size %||% NA_real_,
            backing = info$backing %||% "unknown",
            readonly = info$readonly %||% FALSE
        ),
        class = "shard_segment"
    )
}

#' Reset diagnostic counters for a shared vector
#'
#' @param x A shard ALTREP vector
#' @return x (invisibly)
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(400)
#' segment_write(seg, 1:100, offset = 0)
#' x <- shared_vector(seg, "integer", length = 100)
#'
#' sum(x)
#' shared_diagnostics(x)$dataptr_calls
#'
#' shared_reset_diagnostics(x)
#' shared_diagnostics(x)$dataptr_calls
#' }
shared_reset_diagnostics <- function(x) {
    if (!is_shared_vector(x)) {
        stop("x must be a shard ALTREP vector")
    }

    .Call("C_shard_altrep_reset_diagnostics", x, PACKAGE = "shard")
    invisible(x)
}

#' Create a shared vector from an existing R vector
#'
#' Convenience function that creates a segment, writes the data,
#' and returns an ALTREP view.
#'
#' @param x An atomic vector (integer, double, logical, or raw)
#' @param readonly If TRUE, prevent write access (default: TRUE)
#' @param backing Backing type for the segment: "auto", "mmap", or "shm"
#' @param cow Copy-on-write policy for the resulting shared vector. One of
#'   `"deny"`, `"audit"`, or `"allow"`. If NULL, defaults based on `readonly`.
#' @return An ALTREP vector backed by shared memory
#' @export
#' @examples
#' \donttest{
#' x <- as_shared(1:100)
#' is_shared_vector(x)
#'
#' y <- x[1:10]
#' is_shared_vector(y)
#' }
as_shared <- function(x, readonly = TRUE, backing = "auto", cow = NULL) {
    if (!is.atomic(x) || is.null(x)) {
        stop("x must be an atomic vector")
    }

    # Determine type
    type <- switch(typeof(x),
        "integer" = "integer",
        "double"  = "double",
        "logical" = "logical",
        "raw"     = "raw",
        stop("Unsupported type: ", typeof(x))
    )

    # Calculate size
    elem_size <- switch(type,
        "integer" = 4L,
        "double"  = 8L,
        "logical" = 4L,
        "raw"     = 1L
    )
    size <- length(x) * elem_size
    # mmap(2) of length 0 is invalid; allocate a 1-byte segment for empty vectors.
    alloc_size <- max(size, 1L)

    # Create segment and write data
    seg <- segment_create(alloc_size, backing = backing)
    if (size > 0) {
        segment_write(seg, x, offset = 0)
    }

    # Optionally protect segment
    if (readonly) {
        segment_protect(seg)
    }

    # Create ALTREP view
    shared_vector(seg, type = type, offset = 0, length = length(x),
                  readonly = readonly, cow = cow)
}

# Tests for ALTREP shared vectors

test_that("shared_vector creates ALTREP vectors for different types", {
    # Integer
    seg_int <- segment_create(400)
    segment_write(seg_int, 1:100, offset = 0)
    x_int <- shared_vector(seg_int, "integer", length = 100)

    expect_true(is_shared_vector(x_int))
    expect_equal(length(x_int), 100)
    expect_equal(as.integer(x_int[1]), 1L)
    expect_equal(as.integer(x_int[100]), 100L)

    # Double
    seg_dbl <- segment_create(800)
    segment_write(seg_dbl, as.double(1:100), offset = 0)
    x_dbl <- shared_vector(seg_dbl, "double", length = 100)

    expect_true(is_shared_vector(x_dbl))
    expect_equal(length(x_dbl), 100)
    expect_equal(as.double(x_dbl[1]), 1.0)
    expect_equal(as.double(x_dbl[100]), 100.0)

    # Logical
    seg_lgl <- segment_create(400)
    lgl_data <- rep(c(TRUE, FALSE), 50)
    segment_write(seg_lgl, lgl_data, offset = 0)
    x_lgl <- shared_vector(seg_lgl, "logical", length = 100)

    expect_true(is_shared_vector(x_lgl))
    expect_equal(length(x_lgl), 100)
    expect_true(as.logical(x_lgl[1]))
    expect_false(as.logical(x_lgl[2]))

    # Raw
    seg_raw <- segment_create(100)
    segment_write(seg_raw, as.raw(1:100), offset = 0)
    x_raw <- shared_vector(seg_raw, "raw", length = 100)

    expect_true(is_shared_vector(x_raw))
    expect_equal(length(x_raw), 100)
    expect_equal(as.raw(x_raw[1]), as.raw(1))
})

test_that("shared_vector respects offset parameter", {
    seg <- segment_create(800)
    segment_write(seg, 1:100, offset = 0)
    segment_write(seg, 101:200, offset = 400)  # After first 100 integers

    # Read from offset
    x <- shared_vector(seg, "integer", offset = 400, length = 100)

    expect_equal(length(x), 100)
    expect_equal(as.integer(x[1]), 101L)
    expect_equal(as.integer(x[100]), 200L)
})

test_that("contiguous subsetting returns views, not copies", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)
    x <- shared_vector(seg, "integer", length = 100)

    # Reset diagnostics
    shared_reset_diagnostics(x)

    # Contiguous subset
    y <- x[1:10]

    # For contiguous indices, we should get a view (also ALTREP)
    # Note: R's subsetting might materialize in some cases
    expect_equal(length(y), 10)
    expect_equal(as.integer(y[1]), 1L)
    expect_equal(as.integer(y[10]), 10L)
})

test_that("shared_view creates views explicitly", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)
    x <- shared_vector(seg, "integer", length = 100)

    # Create explicit view
    y <- shared_view(x, start = 11, length = 10)

    expect_true(is_shared_vector(y))
    expect_equal(length(y), 10)
    expect_equal(as.integer(y[1]), 11L)  # x[11]
    expect_equal(as.integer(y[10]), 20L) # x[20]
})

test_that("shared_diagnostics tracks dataptr and materialize calls", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)
    x <- shared_vector(seg, "integer", length = 100)

    # Get initial diagnostics
    diag1 <- shared_diagnostics(x)
    expect_equal(diag1$length, 100)
    expect_equal(diag1$offset, 0)
    expect_true(diag1$readonly)
    expect_equal(diag1$type, "integer")

    # Reset and verify
    shared_reset_diagnostics(x)
    diag2 <- shared_diagnostics(x)
    expect_equal(diag2$dataptr_calls, 0)
    expect_equal(diag2$materialize_calls, 0)

    # Access data (may increment counters depending on operation)
    sum_val <- sum(x)
    expect_equal(sum_val, sum(1:100))
})

test_that("cow='deny' prevents mutation via replacement functions", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)

    # Create readonly vector on UNPROTECTED segment
    x <- shared_vector(seg, "integer", length = 100, readonly = TRUE, cow = "deny")

    # Reading should work
    expect_equal(as.integer(x[1]), 1L)
    expect_equal(as.integer(x[50]), 50L)

    # Attempt to write should error and leave x unchanged.
    expect_error(x[1] <- 999L, "cow='deny'")
    expect_equal(as.integer(x[1]), 1L)

    # x remains a shard shared object (it may be materialized by R internals).
    expect_true(inherits(x, "shard_shared_vector"))
})

test_that("is_shared_vector correctly identifies ALTREP vectors", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)
    x <- shared_vector(seg, "integer", length = 100)

    expect_true(is_shared_vector(x))
    expect_false(is_shared_vector(1:100))
    expect_false(is_shared_vector(c(1.0, 2.0)))
    expect_false(is_shared_vector("hello"))
    expect_false(is_shared_vector(list(a = 1)))
})

test_that("as_shared converts standard vectors to shared", {
    # Integer
    x_int <- as_shared(1:100)
    expect_true(is_shared_vector(x_int))
    expect_equal(length(x_int), 100)
    expect_equal(as.integer(x_int[50]), 50L)

    # Double
    x_dbl <- as_shared(as.double(1:100))
    expect_true(is_shared_vector(x_dbl))
    expect_equal(as.double(x_dbl[50]), 50.0)

    # Logical
    x_lgl <- as_shared(c(TRUE, FALSE, TRUE))
    expect_true(is_shared_vector(x_lgl))
    expect_true(as.logical(x_lgl[1]))
    expect_false(as.logical(x_lgl[2]))

    # Raw
    x_raw <- as_shared(as.raw(1:10))
    expect_true(is_shared_vector(x_raw))
    expect_equal(as.raw(x_raw[5]), as.raw(5))
})

test_that("shared_segment returns the underlying segment", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)
    x <- shared_vector(seg, "integer", length = 100)

    seg <- shared_segment(x)
    expect_true(inherits(seg, "shard_segment"))
    expect_true(inherits(seg$ptr, "externalptr"))
})

test_that("vector operations work on shared vectors", {
    x <- as_shared(1:100)

    # Sum
    expect_equal(sum(x), sum(1:100))

    # Mean
    expect_equal(mean(x), mean(1:100))

    # Range
    expect_equal(as.integer(range(x)), c(1L, 100L))

    # Comparison
    expect_equal(sum(x > 50), 50)
})

test_that("shared vectors work with double precision", {
    vals <- seq(0.1, 10.0, by = 0.1)
    x <- as_shared(vals)

    expect_true(is_shared_vector(x))
    expect_equal(length(x), 100)
    expect_equal(as.double(x[1]), 0.1)
    expect_equal(sum(x), sum(vals))
})

test_that("views share the same underlying memory", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)

    x <- shared_vector(seg, "integer", length = 100)
    y <- shared_view(x, start = 1, length = 50)
    z <- shared_view(x, start = 51, length = 50)

    # Views should have same segment
    expect_equal(shared_segment(x), shared_segment(y))
    expect_equal(shared_segment(x), shared_segment(z))

    # Data should match
    expect_equal(as.integer(y[1]), 1L)
    expect_equal(as.integer(z[1]), 51L)
})

test_that("writing through a view triggers private copy at the correct offset", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)

    # Allow copy-on-write so we can verify the "view write" path is correct.
    x <- shared_vector(seg, "integer", length = 100, readonly = TRUE, cow = "allow")
    v <- shared_view(x, start = 11, length = 10)  # maps x[11:20]

    # Write should materialize v privately, and must not mutate the shared segment.
    v[1] <- 999L

    expect_equal(as.integer(v[1]), 999L)
    expect_equal(as.integer(x[11]), 11L)
    expect_equal(as.integer(x[1]), 1L)

    # A fresh view of the same segment sees the original data.
    y <- shared_vector(seg, "integer", length = 100, readonly = TRUE, cow = "allow")
    expect_equal(as.integer(y[11]), 11L)
})

test_that("multiple views can be created", {
    x <- as_shared(1:1000)

    # Create multiple overlapping views
    v1 <- shared_view(x, start = 1, length = 100)
    v2 <- shared_view(x, start = 50, length = 100)
    v3 <- shared_view(x, start = 100, length = 100)

    expect_equal(length(v1), 100)
    expect_equal(length(v2), 100)
    expect_equal(length(v3), 100)

    # Overlapping region should have same values
    expect_equal(v1[50:99], v2[1:50])
    expect_equal(v2[51:100], v3[1:50])
})

test_that("error handling for invalid inputs", {
    seg <- segment_create(400)
    segment_write(seg, 1:100, offset = 0)
    x <- shared_vector(seg, "integer", length = 100)

    # View start out of bounds
    expect_error(shared_view(x, start = 101, length = 10))

    # View extends beyond bounds
    expect_error(shared_view(x, start = 95, length = 10))

    # Non-ALTREP input
    expect_error(shared_view(1:100, start = 1, length = 10))
    expect_error(shared_diagnostics(1:100))
    expect_error(shared_segment(1:100))
})

#' Zero-copy Views
#'
#' Views are explicit slice descriptors over shared arrays/matrices. They avoid
#' creating slice-sized allocations (e.g. `Y[, a:b]`) by carrying only metadata
#' plus a reference to the shared backing.
#'
#' This is a low-level optimization handle: arbitrary base R operations may
#' materialize a view; use `materialize()` explicitly when you want a standard
#' matrix/array.
#'
#' @name views
NULL

.views_env <- new.env(parent = emptyenv())
.views_env$created <- 0L
.views_env$materialized <- 0L
.views_env$materialized_bytes <- 0
.views_env$packed <- 0L
.views_env$packed_bytes <- 0
.views_env$materialize_hotspots <- new.env(parent = emptyenv())

elem_size_bytes <- function(x) {
  switch(
    typeof(x),
    "integer" = 4L,
    "double" = 8L,
    "logical" = 4L,
    "raw" = 1L,
    NA_integer_
  )
}

is_idx_range <- function(x) inherits(x, "shard_idx_range")

idx_range_validate <- function(start, end) {
  if (!is.numeric(start) || length(start) != 1L || is.na(start)) {
    stop("start must be a single, non-NA number", call. = FALSE)
  }
  if (!is.numeric(end) || length(end) != 1L || is.na(end)) {
    stop("end must be a single, non-NA number", call. = FALSE)
  }
  start <- as.integer(start)
  end <- as.integer(end)
  if (start < 1L) stop("start must be >= 1", call. = FALSE)
  if (end < start) stop("end must be >= start", call. = FALSE)
  list(start = start, end = end)
}

#' Contiguous index range
#'
#' Creates a compact, serializable range descriptor for contiguous indices.
#' This avoids allocating an explicit index vector for large slices.
#'
#' @param start Integer. Start index (1-based, inclusive).
#' @param end Integer. End index (1-based, inclusive).
#' @return An object of class \code{shard_idx_range}.
#' @export
#' @examples
#' r <- idx_range(1, 100)
#' r
idx_range <- function(start, end) {
  r <- idx_range_validate(start, end)
  structure(r, class = "shard_idx_range")
}

idx_range_resolve <- function(r, n, name = "range") {
  if (is.null(r)) return(NULL)
  if (!is_idx_range(r)) stop(name, " must be NULL or idx_range()", call. = FALSE)
  if (r$end > n) stop(name, " end exceeds dimension (", n, ")", call. = FALSE)
  r$start:r$end
}

idx_range_len <- function(r, n) {
  if (is.null(r)) return(as.integer(n))
  if (!is_idx_range(r)) return(NA_integer_)
  as.integer(r$end - r$start + 1L)
}

idx_range_print <- function(r) {
  if (is.null(r)) return("(all)")
  paste0("[", r$start, ":", r$end, "]")
}

idx_vec_validate <- function(idx, n, name = "idx") {
  if (!is.numeric(idx) || is.null(idx)) {
    stop(name, " must be a non-empty numeric/integer vector", call. = FALSE)
  }
  idx <- as.integer(idx)
  if (anyNA(idx)) stop(name, " must not contain NA", call. = FALSE)
  if (any(idx < 1L) || any(idx > n)) {
    stop(name, " contains out-of-range indices (1..", n, ")", call. = FALSE)
  }
  idx
}

idx_print <- function(x) {
  if (is.null(x)) return("(all)")
  if (is_idx_range(x)) return(idx_range_print(x))
  if (is.integer(x)) return(paste0("[gather n=", length(x), "]"))
  "(unknown)"
}

.view_hotspot_key_ <- function() {
  # Best-effort attribution: try to find the first non-materialize call above
  # materialize() in the stack. Keep strings short/stable for reporting.
  calls <- sys.calls()
  if (length(calls) == 0) return("(unknown)")

  skip <- c(
    "materialize",
    "UseMethod",
    "materialize.shard_view_block",
    "materialize.shard_view_gather",
    "materialize.default"
  )

  for (i in rev(seq_along(calls))) {
    cl <- calls[[i]]
    if (!is.call(cl)) next
    fn <- tryCatch(as.character(cl[[1]]), error = function(e) "")
    if (!nzchar(fn) || fn %in% skip) next
    txt <- paste(deparse(cl, nlines = 1L, width.cutoff = 200L), collapse = " ")
    if (nzchar(txt)) return(txt)
  }

  txt <- paste(deparse(calls[[length(calls)]], nlines = 1L, width.cutoff = 200L), collapse = " ")
  if (nzchar(txt)) return(txt)
  "(unknown)"
}

.view_record_materialization_ <- function(nbytes) {
  nbytes <- as.double(nbytes)
  if (!is.finite(nbytes) || is.na(nbytes) || nbytes <= 0) return(invisible(NULL))

  key <- .view_hotspot_key_()
  if (!exists(key, envir = .views_env$materialize_hotspots, inherits = FALSE)) {
    .views_env$materialize_hotspots[[key]] <- list(bytes = 0, count = 0L)
  }
  cur <- .views_env$materialize_hotspots[[key]]
  cur$bytes <- (cur$bytes %||% 0) + nbytes
  cur$count <- as.integer((cur$count %||% 0L) + 1L)
  .views_env$materialize_hotspots[[key]] <- cur
  invisible(NULL)
}

view_materialize_hotspots_snapshot_ <- function() {
  keys <- ls(.views_env$materialize_hotspots, all.names = TRUE)
  if (length(keys) == 0) return(list())
  out <- list()
  for (k in keys) {
    v <- get(k, envir = .views_env$materialize_hotspots, inherits = FALSE)
    out[[k]] <- list(bytes = as.double(v$bytes %||% 0), count = as.integer(v$count %||% 0L))
  }
  out
}

view_layout <- function(x_dim, rows, cols) {
  if (length(x_dim) != 2L) return("unsupported")
  if (is.null(rows) && is.null(cols)) return("full")

  # Column-major matrices: contiguous column blocks are contiguous in memory.
  if (is.null(rows) && is_idx_range(cols)) return("col_block_contiguous")

  # Row blocks are strided (typically require packing for BLAS).
  if (is_idx_range(rows) && is.null(cols)) return("row_block_strided")

  if (is_idx_range(rows) && is_idx_range(cols)) return("submatrix_strided")

  if ((is.null(rows) || is_idx_range(rows)) && is.integer(cols)) return("col_gather")

  "unsupported"
}

is_shared_matrix_like <- function(x) {
  d <- dim(x)
  if (is.null(d) || length(d) != 2L) return(FALSE)
  is_shared_vector(x) && typeof(x) %in% c("double", "integer", "logical", "raw")
}

view_validate_dims <- function(x) {
  if (!is_shared_matrix_like(x)) {
    stop(
      "x must be a shared (share()d) atomic matrix with supported type.\n",
      "  Supported types: double/integer/logical/raw.\n",
      "  Hint: use share(matrix(...)) to create a shared input first.",
      call. = FALSE
    )
  }
  d <- dim(x)
  if (length(d) != 2L || any(is.na(d)) || any(d < 0)) {
    stop("Invalid matrix dimensions on shared object", call. = FALSE)
  }
  invisible(d)
}

view_block_validate_selector <- function(sel, which = c("rows", "cols")) {
  which <- match.arg(which)
  if (is.null(sel)) return(invisible(NULL))
  if (is_idx_range(sel)) return(invisible(NULL))

  stop(
    which, " must be NULL or idx_range() for block views.\n",
    "  Gather/indexed views are not implemented yet.",
    call. = FALSE
  )
}

view_block_build <- function(x, rows, cols) {
  d <- dim(x)
  nrow <- d[1]
  ncol <- d[2]

  view_block_validate_selector(rows, "rows")
  view_block_validate_selector(cols, "cols")

  if (!is.null(rows) && rows$end > nrow) stop("rows end exceeds nrow", call. = FALSE)
  if (!is.null(cols) && cols$end > ncol) stop("cols end exceeds ncol", call. = FALSE)

  layout <- view_layout(d, rows, cols)
  if (layout == "unsupported") {
    stop("Unsupported view layout for block view", call. = FALSE)
  }

  # Slice dims
  rlen <- idx_range_len(rows, nrow)
  clen <- idx_range_len(cols, ncol)
  slice_dim <- c(rlen, clen)

  # Estimate payload bytes (what materialize() would copy).
  es <- elem_size_bytes(x)
  nbytes <- if (!is.na(es)) as.double(prod(slice_dim)) * as.double(es) else NA_real_

  .views_env$created <- .views_env$created + 1L

  structure(
    list(
      base = x,
      dim = d,
      rows = rows,
      cols = cols,
      dtype = typeof(x),
      layout = layout,
      slice_dim = slice_dim,
      nbytes_est = nbytes
    ),
    class = c("shard_view_block", "shard_view")
  )
}

#' Create a gather (indexed) view over a shared matrix
#'
#' Gather views describe non-contiguous column (or row) subsets without
#' allocating a slice-sized matrix. shard-aware kernels can then choose to
#' pack the requested indices into scratch explicitly (bounded and reportable)
#' or run gather-aware compute paths.
#'
#' v1 note: only column-gather views are implemented (rows may be NULL or
#' idx_range()).
#'
#' @param x A shared (share()d) atomic matrix (double/integer/logical/raw).
#' @param rows Row selector. NULL (all rows) or idx_range().
#' @param cols Integer vector of column indices (1-based).
#' @return A \code{shard_view_gather} object describing the indexed column view.
#' @export
#' @examples
#' \donttest{
#' m <- share(matrix(1:20, nrow = 4))
#' v <- view_gather(m, cols = c(1L, 3L))
#' }
view_gather <- function(x, rows = NULL, cols) {
  view_validate_dims(x)
  d <- dim(x)
  nrow <- d[1]
  ncol <- d[2]

  if (!is.null(rows) && !is_idx_range(rows)) {
    stop("rows must be NULL or idx_range() for gather views", call. = FALSE)
  }
  if (!is.null(rows) && rows$end > nrow) stop("rows end exceeds nrow", call. = FALSE)

  cols <- idx_vec_validate(cols, ncol, name = "cols")

  slice_dim <- c(idx_range_len(rows, nrow), length(cols))
  es <- elem_size_bytes(x)
  nbytes <- if (!is.na(es)) as.double(prod(slice_dim)) * as.double(es) else NA_real_

  .views_env$created <- .views_env$created + 1L

  structure(
    list(
      base = x,
      dim = d,
      rows = rows,
      cols = cols,
      dtype = typeof(x),
      layout = "col_gather",
      slice_dim = slice_dim,
      nbytes_est = nbytes
    ),
    class = c("shard_view_gather", "shard_view")
  )
}

#' Create a view over a shared matrix
#'
#' @param x A shared (share()d) atomic matrix (double/integer/logical/raw).
#' @param rows Row selector. NULL (all rows) or idx_range().
#' @param cols Column selector. NULL (all cols) or idx_range().
#' @param type View type. \code{"block"} or \code{"gather"} (or \code{"auto"}).
#' @return A \code{shard_view_block} or \code{shard_view_gather} object depending
#'   on the selectors provided.
#' @export
#' @examples
#' \donttest{
#' m <- share(matrix(1:20, nrow = 4))
#' v <- view(m, cols = idx_range(1, 2))
#' }
view <- function(x, rows = NULL, cols = NULL, type = c("auto", "block", "gather")) {
  type <- match.arg(type)
  view_validate_dims(x)

  if (type == "gather") {
    if (is.null(cols)) stop("gather view requires cols", call. = FALSE)
    return(view_gather(x, rows = rows, cols = cols))
  }
  if (type == "auto") {
    # Auto-select block for idx_range selectors, gather for integer selectors.
    if (!is.null(cols) && is.numeric(cols) && !is_idx_range(cols)) {
      return(view_gather(x, rows = rows, cols = cols))
    }
    if ((!is.null(rows) && !is_idx_range(rows)) || (!is.null(cols) && !is_idx_range(cols))) {
      stop("Unsupported selectors for auto view; use idx_range() or integer indices.", call. = FALSE)
    }
    return(view_block_build(x, rows = rows, cols = cols))
  }

  view_block_build(x, rows = rows, cols = cols)
}

#' Create a contiguous block view
#'
#' @param x A shared (share()d) atomic matrix.
#' @param rows NULL or idx_range().
#' @param cols NULL or idx_range().
#' @return A \code{shard_view_block} object representing the contiguous block slice.
#' @export
#' @examples
#' \donttest{
#' m <- share(matrix(1:20, nrow = 4))
#' v <- view_block(m, cols = idx_range(1, 2))
#' }
view_block <- function(x, rows = NULL, cols = NULL) {
  view_validate_dims(x)
  view_block_build(x, rows = rows, cols = cols)
}

#' Introspection for a view
#'
#' Returns metadata about a view without forcing materialization.
#'
#' @param v A shard view.
#' @return A named list with fields: \code{dtype}, \code{dim}, \code{slice_dim},
#'   \code{rows}, \code{cols}, \code{layout}, \code{fast_path},
#'   \code{nbytes_est}, and \code{base_is_shared}.
#' @export
#' @examples
#' \donttest{
#' m <- share(matrix(1:20, nrow = 4))
#' v <- view_block(m, cols = idx_range(1, 2))
#' view_info(v)
#' }
view_info <- function(v) {
  if (!inherits(v, "shard_view")) stop("v must be a shard view", call. = FALSE)
  base <- v$base
  d <- v$dim

  fast_path <- isTRUE(v$layout == "col_block_contiguous") && identical(v$dtype, "double")

  list(
    dtype = v$dtype,
    dim = d,
    slice_dim = v$slice_dim,
    rows = v$rows,
    cols = v$cols,
    rows_print = idx_print(v$rows),
    cols_print = idx_print(v$cols),
    layout = v$layout,
    fast_path = fast_path,
    nbytes_est = v$nbytes_est,
    base_is_shared = is_shared_vector(base)
  )
}

#' View Predicates
#'
#' @param x An object.
#' @return Logical. \code{TRUE} if \code{x} is a shard view (or block view).
#' @export
#' @examples
#' \donttest{
#' m <- share(matrix(1:20, nrow = 4))
#' v <- view_block(m, cols = idx_range(1, 2))
#' is_view(v)
#' is_block_view(v)
#' }
is_view <- function(x) inherits(x, "shard_view")

#' @rdname is_view
#' @export
is_block_view <- function(x) inherits(x, "shard_view_block")

#' Print a shard_idx_range object
#'
#' @param x A \code{shard_idx_range} object.
#' @param ... Additional arguments (ignored).
#' @return Returns \code{x} invisibly.
#' @export
#' @examples
#' r <- idx_range(1, 10)
#' print(r)
print.shard_idx_range <- function(x, ...) {
  cat("idx_range", idx_range_print(x), "\n")
  invisible(x)
}

#' Print a shard_view_block object
#'
#' @param x A \code{shard_view_block} object.
#' @param ... Additional arguments (ignored).
#' @return Returns \code{x} invisibly.
#' @export
print.shard_view_block <- function(x, ...) {
  info <- view_info(x)
  cat("shard_view_block\n")
  cat("  dtype:", info$dtype, "\n")
  cat("  dim:", paste(info$dim, collapse = "x"), "\n")
  cat("  slice:", paste(info$slice_dim, collapse = "x"),
      " rows=", info$rows_print, " cols=", info$cols_print, "\n", sep = "")
  cat("  layout:", info$layout, " fast_path:", info$fast_path, "\n")
  invisible(x)
}

#' Print a shard_view_gather object
#'
#' @param x A \code{shard_view_gather} object.
#' @param ... Additional arguments (ignored).
#' @return Returns \code{x} invisibly.
#' @export
print.shard_view_gather <- function(x, ...) {
  info <- view_info(x)
  cat("shard_view_gather\n")
  cat("  dtype:", info$dtype, "\n")
  cat("  dim:", paste(info$dim, collapse = "x"), "\n")
  cat("  slice:", paste(info$slice_dim, collapse = "x"),
      " rows=", info$rows_print, " cols=", info$cols_print, "\n", sep = "")
  cat("  layout:", info$layout, " fast_path:", info$fast_path, "\n")
  invisible(x)
}

#' Materialize a block view into an R matrix
#'
#' @param x A \code{shard_view_block} object.
#' @return A standard R matrix containing the selected rows and columns.
#' @export
materialize.shard_view_block <- function(x) {
  # Materialization is explicit and counted.
  base <- x$base
  d <- x$dim
  rows <- idx_range_resolve(x$rows, d[1], name = "rows")
  cols <- idx_range_resolve(x$cols, d[2], name = "cols")

  .views_env$materialized <- .views_env$materialized + 1L
  if (!is.null(x$nbytes_est) && is.finite(x$nbytes_est)) {
    .views_env$materialized_bytes <- .views_env$materialized_bytes + x$nbytes_est
    .view_record_materialization_(x$nbytes_est)
  }

  if (is.null(rows)) rows <- seq_len(d[1])
  if (is.null(cols)) cols <- seq_len(d[2])

  base[rows, cols, drop = FALSE]
}

#' Materialize a gather view into an R matrix
#'
#' @param x A \code{shard_view_gather} object.
#' @return A standard R matrix containing the gathered columns.
#' @export
materialize.shard_view_gather <- function(x) {
  base <- x$base
  d <- x$dim
  rows <- idx_range_resolve(x$rows, d[1], name = "rows")
  cols <- x$cols

  .views_env$materialized <- .views_env$materialized + 1L
  if (!is.null(x$nbytes_est) && is.finite(x$nbytes_est)) {
    .views_env$materialized_bytes <- .views_env$materialized_bytes + x$nbytes_est
    .view_record_materialization_(x$nbytes_est)
  }

  if (is.null(rows)) rows <- seq_len(d[1])
  base[rows, cols, drop = FALSE]
}

#' View diagnostics
#'
#' Returns global counters for view creation/materialization. This is a simple
#' first step; in future this should be integrated into shard_map run-level
#' diagnostics.
#'
#' @return A list with counters.
#' @export
view_diagnostics <- function() {
  list(
    created = .views_env$created,
    materialized = .views_env$materialized,
    materialized_bytes = .views_env$materialized_bytes,
    packed = .views_env$packed,
    packed_bytes = .views_env$packed_bytes
  )
}

view_reset_diagnostics <- function() {
  .views_env$created <- 0L
  .views_env$materialized <- 0L
  .views_env$materialized_bytes <- 0
  .views_env$packed <- 0L
  .views_env$packed_bytes <- 0
  rm(list = ls(.views_env$materialize_hotspots, all.names = TRUE), envir = .views_env$materialize_hotspots)
  invisible(NULL)
}

view_col_sums <- function(v) {
  if (!inherits(v, "shard_view_block")) {
    stop("v must be a shard_view_block", call. = FALSE)
  }
  base <- v$base
  d <- v$dim
  if (length(d) != 2L) stop("View base must be a matrix", call. = FALSE)

  rs <- if (is.null(v$rows)) c(1L, d[1]) else c(v$rows$start, v$rows$end)
  cs <- if (is.null(v$cols)) c(1L, d[2]) else c(v$cols$start, v$cols$end)

  out <- .Call(
    "C_shard_mat_block_col_sums",
    base,
    as.integer(rs[1]),
    as.integer(rs[2]),
    as.integer(cs[1]),
    as.integer(cs[2]),
    PACKAGE = "shard"
  )

  dns <- dimnames(base)
  if (!is.null(dns) && length(dns) == 2L && !is.null(dns[[2]])) {
    nms <- dns[[2]][cs[1]:cs[2]]
    if (!is.null(nms)) names(out) <- nms
  }

  out
}

view_col_vars <- function(v) {
  if (!inherits(v, "shard_view_block")) {
    stop("v must be a shard_view_block", call. = FALSE)
  }
  base <- v$base
  d <- v$dim
  if (length(d) != 2L) stop("View base must be a matrix", call. = FALSE)

  rs <- if (is.null(v$rows)) c(1L, d[1]) else c(v$rows$start, v$rows$end)
  cs <- if (is.null(v$cols)) c(1L, d[2]) else c(v$cols$start, v$cols$end)

  out <- .Call(
    "C_shard_mat_block_col_vars",
    base,
    as.integer(rs[1]),
    as.integer(rs[2]),
    as.integer(cs[1]),
    as.integer(cs[2]),
    PACKAGE = "shard"
  )

  dns <- dimnames(base)
  if (!is.null(dns) && length(dns) == 2L && !is.null(dns[[2]])) {
    nms <- dns[[2]][cs[1]:cs[2]]
    if (!is.null(nms)) names(out) <- nms
  }

  out
}

view_xTy <- function(x, y_view, x_cols = NULL) {
  if (!inherits(y_view, "shard_view")) {
    stop("y_view must be a shard view", call. = FALSE)
  }
  if (!is.matrix(x) && !(is_shared_vector(x) && !is.null(dim(x)) && length(dim(x)) == 2L)) {
    stop("x must be a matrix or a shared matrix backing", call. = FALSE)
  }

  y_base <- y_view$base
  dY <- y_view$dim
  dX <- dim(x)
  if (length(dY) != 2L || length(dX) != 2L) stop("x and y must be matrices", call. = FALSE)
  if (dX[1] != dY[1]) stop("x and y must have the same number of rows", call. = FALSE)

  rs <- if (is.null(y_view$rows)) c(1L, dY[1]) else c(y_view$rows$start, y_view$rows$end)

  if (is.null(x_cols)) {
    xcs <- 1L
    xce <- dX[2]
  } else {
    if (!is_idx_range(x_cols)) stop("x_cols must be NULL or idx_range()", call. = FALSE)
    xcs <- x_cols$start
    xce <- x_cols$end
    if (xce > dX[2]) stop("x_cols end exceeds ncol(x)", call. = FALSE)
  }

  if (!identical(y_view$dtype, "double") || !identical(typeof(x), "double")) {
    # Keep this strict while the kernels are double-only.
    stop("view_xTy currently supports double matrices only", call. = FALSE)
  }

  if (inherits(y_view, "shard_view_block")) {
    if (is.null(y_view$cols)) stop("y_view must specify cols for block views", call. = FALSE)
    ycs <- y_view$cols$start
    yce <- y_view$cols$end

    out <- .Call(
      "C_shard_mat_crossprod_block",
      x,
      y_base,
      as.integer(rs[1]),
      as.integer(rs[2]),
      as.integer(xcs),
      as.integer(xce),
      as.integer(ycs),
      as.integer(yce),
      PACKAGE = "shard"
    )

    dnX <- dimnames(x)
    dnY <- dimnames(y_base)
    if (!is.null(dnX) || !is.null(dnY)) {
      rn <- if (!is.null(dnX) && length(dnX) == 2L) dnX[[2]] else NULL
      cn <- if (!is.null(dnY) && length(dnY) == 2L) dnY[[2]] else NULL
      if (!is.null(rn)) rn <- rn[xcs:xce]
      if (!is.null(cn)) cn <- cn[ycs:yce]
      dimnames(out) <- list(rn, cn)
    }

    return(out)
  }

  if (!inherits(y_view, "shard_view_gather")) {
    stop("Unsupported view type for view_xTy", call. = FALSE)
  }
  if (is.null(y_view$cols) || !is.integer(y_view$cols)) {
    stop("gather view must provide integer cols", call. = FALSE)
  }

  y_cols <- y_view$cols
  nr <- as.integer(rs[2] - rs[1] + 1L)
  ky <- length(y_cols)
  # Reuse worker-local packing buffers to avoid repeated allocation churn.
  B <- scratch_matrix(nr, ky, key = paste0("gather_pack_", nr, "x", ky))
  out <- .Call(
    "C_shard_mat_crossprod_gather_scratch",
    x,
    y_base,
    as.integer(rs[1]),
    as.integer(rs[2]),
    as.integer(xcs),
    as.integer(xce),
    as.integer(y_cols),
    B,
    PACKAGE = "shard"
  )

  # Record controlled packing volume (scratch), distinct from view materialization.
  .views_env$packed <- .views_env$packed + 1L
  .views_env$packed_bytes <- .views_env$packed_bytes + (as.double(nr) * as.double(length(y_cols)) * 8)

  dnX <- dimnames(x)
  dnY <- dimnames(y_base)
  rn <- if (!is.null(dnX) && length(dnX) == 2L) dnX[[2]] else NULL
  cn <- if (!is.null(dnY) && length(dnY) == 2L) dnY[[2]] else NULL
  if (!is.null(rn)) rn <- rn[xcs:xce]
  if (!is.null(cn)) cn <- cn[y_cols]
  if (!is.null(rn) || !is.null(cn)) dimnames(out) <- list(rn, cn)

  out
}

view_crossprod <- function(x_view, y_view) {
  if (!inherits(x_view, "shard_view_block")) {
    stop("x_view must be a shard_view_block", call. = FALSE)
  }
  if (!inherits(y_view, "shard_view_block")) {
    stop("y_view must be a shard_view_block", call. = FALSE)
  }

  xb <- x_view$base
  yb <- y_view$base
  dX <- x_view$dim
  dY <- y_view$dim
  if (length(dX) != 2L || length(dY) != 2L) stop("views must be over matrices", call. = FALSE)
  if (dX[1] != dY[1]) stop("x and y must have the same number of rows", call. = FALSE)

  if (!identical(x_view$dtype, "double") || !identical(y_view$dtype, "double")) {
    stop("view_crossprod currently supports double matrices only", call. = FALSE)
  }

  # Require compatible row ranges (or NULL meaning full). Keep behavior strict
  # for now so the kernel can be reasoned about easily.
  if (!is.null(x_view$rows) && !is.null(y_view$rows) &&
      (!is_idx_range(x_view$rows) || !is_idx_range(y_view$rows) ||
       x_view$rows$start != y_view$rows$start || x_view$rows$end != y_view$rows$end)) {
    stop("x_view and y_view must have identical row ranges (or NULL)", call. = FALSE)
  }

  rs <- if (is.null(x_view$rows) && is.null(y_view$rows)) c(1L, dX[1]) else {
    r <- if (!is.null(x_view$rows)) x_view$rows else y_view$rows
    c(r$start, r$end)
  }

  xcs <- if (is.null(x_view$cols)) 1L else x_view$cols$start
  xce <- if (is.null(x_view$cols)) dX[2] else x_view$cols$end
  ycs <- if (is.null(y_view$cols)) 1L else y_view$cols$start
  yce <- if (is.null(y_view$cols)) dY[2] else y_view$cols$end

  out <- .Call(
    "C_shard_mat_crossprod_block",
    xb,
    yb,
    as.integer(rs[1]),
    as.integer(rs[2]),
    as.integer(xcs),
    as.integer(xce),
    as.integer(ycs),
    as.integer(yce),
    PACKAGE = "shard"
  )

  dnX <- dimnames(xb)
  dnY <- dimnames(yb)
  rn <- if (!is.null(dnX) && length(dnX) == 2L) dnX[[2]] else NULL
  cn <- if (!is.null(dnY) && length(dnY) == 2L) dnY[[2]] else NULL
  if (!is.null(rn)) rn <- rn[xcs:xce]
  if (!is.null(cn)) cn <- cn[ycs:yce]
  if (!is.null(rn) || !is.null(cn)) dimnames(out) <- list(rn, cn)

  out
}

tiles2d <- function(n_x, n_y, block_x, block_y) {
  n_x <- as.integer(n_x)
  n_y <- as.integer(n_y)
  block_x <- as.integer(block_x)
  block_y <- as.integer(block_y)

  if (is.na(n_x) || n_x < 1L) stop("n_x must be a positive integer", call. = FALSE)
  if (is.na(n_y) || n_y < 1L) stop("n_y must be a positive integer", call. = FALSE)
  if (is.na(block_x) || block_x < 1L) stop("block_x must be a positive integer", call. = FALSE)
  if (is.na(block_y) || block_y < 1L) stop("block_y must be a positive integer", call. = FALSE)

  x_blocks <- ceiling(n_x / block_x)
  y_blocks <- ceiling(n_y / block_y)
  num_tiles <- x_blocks * y_blocks

  shards <- vector("list", num_tiles)
  id <- 1L
  for (xb in seq_len(x_blocks)) {
    x_start <- (xb - 1L) * block_x + 1L
    x_end <- min(xb * block_x, n_x)
    for (yb in seq_len(y_blocks)) {
      y_start <- (yb - 1L) * block_y + 1L
      y_end <- min(yb * block_y, n_y)

      shards[[id]] <- list(
        id = id,
        x_start = x_start,
        x_end = x_end,
        y_start = y_start,
        y_end = y_end,
        x_len = x_end - x_start + 1L,
        y_len = y_end - y_start + 1L,
        len = (x_end - x_start + 1L) * (y_end - y_start + 1L)
      )
      id <- id + 1L
    }
  }

  structure(
    list(
      n = num_tiles,
      block_size = NA_integer_,
      strategy = "tiles2d",
      num_shards = length(shards),
      shards = shards,
      x_dim = n_x,
      y_dim = n_y,
      x_block = block_x,
      y_block = block_y
    ),
    class = c("shard_tiles", "shard_descriptor")
  )
}

#' Print a shard_tiles object
#'
#' @param x A \code{shard_tiles} object.
#' @param ... Additional arguments (ignored).
#' @return Returns \code{x} invisibly.
#' @export
print.shard_tiles <- function(x, ...) {
  cat("shard tiles (2D)\n")
  cat("  Tiles:", x$num_shards, "\n")
  cat("  X dim:", x$x_dim, " block:", x$x_block, "\n")
  cat("  Y dim:", x$y_dim, " block:", x$y_block, "\n")
  invisible(x)
}

# Tests for shard views (block views)

test_that("block views over shared matrices materialize correctly", {
  shard:::view_reset_diagnostics()
  X <- matrix(as.double(1:30), nrow = 5)
  Xsh <- share(X, backing = "mmap")

  expect_true(is_shared_vector(Xsh))
  expect_equal(dim(Xsh), dim(X))

  v <- view_block(Xsh, cols = idx_range(2, 4))
  expect_true(is_view(v))
  expect_true(is_block_view(v))

  info <- view_info(v)
  expect_equal(info$layout, "col_block_contiguous")
  expect_equal(info$slice_dim, c(5L, 3L))
  expect_true(isTRUE(info$base_is_shared))

  Xm <- materialize(v)
  expect_equal(Xm, X[, 2:4, drop = FALSE])
})

test_that("views are serializable and remain usable after unserialize", {
  shard:::view_reset_diagnostics()
  X <- matrix(as.double(1:60), nrow = 6)
  Xsh <- share(X, backing = "mmap")
  v <- view_block(Xsh, cols = idx_range(3, 7))

  raw <- serialize(v, connection = NULL, xdr = FALSE)
  v2 <- unserialize(raw)

  expect_true(is_block_view(v2))
  expect_equal(materialize(v2), X[, 3:7, drop = FALSE])
})

test_that("block views reject non-range selectors; view() auto-selects gather", {
  shard:::view_reset_diagnostics()
  X <- matrix(as.double(1:20), nrow = 4)
  Xsh <- share(X, backing = "mmap")

  expect_error(view_block(Xsh, cols = 1:3))

  v1 <- view(Xsh, cols = 1:3, type = "auto")
  expect_true(is_view(v1))
  expect_false(is_block_view(v1))
  expect_true(inherits(v1, "shard_view_gather"))
  expect_equal(materialize(v1), X[, 1:3, drop = FALSE])

  v2 <- view(Xsh, cols = 1:3, type = "gather")
  expect_true(inherits(v2, "shard_view_gather"))
})

test_that("view_col_sums runs without materializing views", {
  shard:::view_reset_diagnostics()

  X <- matrix(rnorm(1000), nrow = 50)
  colnames(X) <- paste0("c", seq_len(ncol(X)))
  Xsh <- share(X, backing = "mmap")

  v <- view_block(Xsh, cols = idx_range(3, 12))
  sums <- shard:::view_col_sums(v)

  expect_equal(unname(sums), unname(colSums(X[, 3:12, drop = FALSE])))
  expect_equal(names(sums), colnames(X)[3:12])

  vd <- view_diagnostics()
  expect_equal(vd$materialized, 0L)
  expect_equal(vd$materialized_bytes, 0)
})

test_that("view_xTy uses BLAS path without materializing views", {
  shard:::view_reset_diagnostics()

  X <- matrix(rnorm(2000), nrow = 100)
  Y <- matrix(rnorm(3000), nrow = 100)
  colnames(X) <- paste0("x", seq_len(ncol(X)))
  colnames(Y) <- paste0("y", seq_len(ncol(Y)))

  Xsh <- share(X, backing = "mmap")
  Ysh <- share(Y, backing = "mmap")

  vY <- view_block(Ysh, cols = idx_range(4, 11))
  out <- shard:::view_xTy(Xsh, vY)

  expect_equal(out, crossprod(X, Y[, 4:11, drop = FALSE]))
  expect_equal(rownames(out), colnames(X))
  expect_equal(colnames(out), colnames(Y)[4:11])

  vd <- view_diagnostics()
  expect_equal(vd$materialized, 0L)
  expect_equal(vd$materialized_bytes, 0)
})

test_that("view_crossprod consumes two views without materializing", {
  shard:::view_reset_diagnostics()

  X <- matrix(rnorm(1500), nrow = 100)
  Y <- matrix(rnorm(2200), nrow = 100)
  colnames(X) <- paste0("x", seq_len(ncol(X)))
  colnames(Y) <- paste0("y", seq_len(ncol(Y)))

  Xsh <- share(X, backing = "mmap")
  Ysh <- share(Y, backing = "mmap")

  vX <- view_block(Xsh, cols = idx_range(2, 9))
  vY <- view_block(Ysh, cols = idx_range(5, 11))

  out <- shard:::view_crossprod(vX, vY)

  expect_equal(out, crossprod(X[, 2:9, drop = FALSE], Y[, 5:11, drop = FALSE]))
  expect_equal(rownames(out), colnames(X)[2:9])
  expect_equal(colnames(out), colnames(Y)[5:11])

  vd <- view_diagnostics()
  expect_equal(vd$materialized, 0L)
  expect_equal(vd$materialized_bytes, 0)
})

test_that("shard_map view-enabled xTy path has zero view materialization in hot path", {
  skip_on_cran()

  pool_stop()
  shard:::view_reset_diagnostics()

  n <- 200L
  p <- 8L
  v <- 24L
  X <- matrix(rnorm(n * p), nrow = n)
  Y <- matrix(rnorm(n * v), nrow = n)

  colnames(X) <- paste0("x", seq_len(ncol(X)))
  colnames(Y) <- paste0("y", seq_len(ncol(Y)))

  Xsh <- share(X, backing = "mmap")
  Ysh <- share(Y, backing = "mmap")

  blocks <- shards(ncol(Y), block_size = 6, workers = 2)

  result <- shard_map(
    blocks,
    borrow = list(X = Xsh, Y = Ysh),
    fun = function(shard, X, Y) {
      vY <- view_block(Y, cols = idx_range(shard$start, shard$end))
      shard:::view_xTy(X, vY)
    },
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(result))

  cr <- copy_report(result)
  expect_gt(cr$view_created %||% 0L, 0L)
  expect_equal(cr$view_materialized %||% 0L, 0L)
  expect_equal(cr$view_materialized_bytes %||% 0, 0)

  # Spot-check correctness by recombining blocks.
  mats <- results(result)
  out <- do.call(cbind, mats)
  expect_equal(out, crossprod(X, Y))

  pool_stop()
})

test_that("shard_map view-enabled crossprod(vX, vY) path has zero view materialization", {
  skip_on_cran()

  pool_stop()
  shard:::view_reset_diagnostics()

  n <- 300L
  p <- 16L
  v <- 40L
  X <- matrix(rnorm(n * p), nrow = n)
  Y <- matrix(rnorm(n * v), nrow = n)

  colnames(X) <- paste0("x", seq_len(ncol(X)))
  colnames(Y) <- paste0("y", seq_len(ncol(Y)))

  Xsh <- share(X, backing = "mmap")
  Ysh <- share(Y, backing = "mmap")

  blocks <- shards(ncol(Y), block_size = 10, workers = 2)

  res <- shard_map(
    blocks,
    borrow = list(X = Xsh, Y = Ysh),
    fun = function(shard, X, Y) {
      vX <- view_block(X, cols = idx_range(1, ncol(X)))
      vY <- view_block(Y, cols = idx_range(shard$start, shard$end))
      shard:::view_crossprod(vX, vY)
    },
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))

  cr <- copy_report(res)
  expect_gt(cr$view_created %||% 0L, 0L)
  expect_equal(cr$view_materialized %||% 0L, 0L)
  expect_equal(cr$view_materialized_bytes %||% 0, 0)

  mats <- results(res)
  out <- do.call(cbind, mats)
  expect_equal(out, crossprod(X, Y))

  pool_stop()
})

test_that("view_col_vars matches base var() for block views", {
  skip_on_cran()

  pool_stop()

  n <- 200L
  p <- 20L
  X <- matrix(rnorm(n * p), nrow = n)
  colnames(X) <- paste0("x", seq_len(ncol(X)))

  Xsh <- share(X, backing = "mmap")
  vX <- view_block(Xsh, cols = idx_range(5, 15))

  expected <- apply(X[, 5:15, drop = FALSE], 2, var)
  actual <- shard:::view_col_vars(vX)

  expect_equal(as.numeric(actual), as.numeric(expected), tolerance = 1e-10)
  expect_equal(names(actual), colnames(X)[5:15])

  pool_stop()
})

#' CPU Affinity + mmap Advice (Advanced)
#'
#' These controls are opt-in and best-effort. On unsupported platforms, they
#' safely no-op (returning FALSE).
#'
#' @name affinity
NULL

#' Check whether CPU affinity is supported
#'
#' Currently supported on Linux only.
#'
#' @return A logical scalar indicating platform support.
#' @export
#' @examples
#' affinity_supported()
affinity_supported <- function() {
  isTRUE(.Call("C_shard_affinity_supported", PACKAGE = "shard"))
}

#' Set CPU affinity for the current process
#'
#' Intended to be called inside a worker process (e.g., via `clusterCall()`).
#' On unsupported platforms, returns FALSE.
#'
#' @param cores Integer vector of 0-based CPU core ids.
#' @return A logical scalar; \code{TRUE} on success, \code{FALSE} if not supported.
#' @export
#' @examples
#' affinity_supported()
set_affinity <- function(cores) {
  cores <- as.integer(cores)
  isTRUE(.Call("C_shard_set_affinity", cores, PACKAGE = "shard"))
}

#' Pin shard workers to CPU cores
#'
#' Best-effort worker pinning to improve cache locality and reduce cross-core
#' migration. Currently supported on Linux only.
#'
#' @param pool Optional shard_pool. Defaults to current pool.
#' @param strategy "spread" assigns worker i -> core i mod ncores. "compact"
#'   assigns workers to the first cores.
#' @param cores Optional integer vector of available cores (0-based). If NULL,
#'   uses 0:(detectCores()-1).
#' @return Invisibly, a logical vector per worker indicating success.
#' @export
#' @examples
#' \donttest{
#' affinity_supported()
#' }
pin_workers <- function(pool = NULL,
                        strategy = c("spread", "compact"),
                        cores = NULL) {
  strategy <- match.arg(strategy)
  if (is.null(pool)) pool <- pool_get()
  if (is.null(pool)) stop("No active pool", call. = FALSE)

  if (!affinity_supported()) {
    return(invisible(rep(FALSE, pool$n)))
  }

  if (is.null(cores)) {
    ncores <- parallel::detectCores()
    cores <- seq.int(0L, as.integer(ncores) - 1L)
  }
  cores <- as.integer(cores)
  cores <- cores[!is.na(cores)]
  if (length(cores) < 1L) stop("cores must be non-empty", call. = FALSE)

  ok <- logical(pool$n)
  for (i in seq_len(pool$n)) {
    core <- switch(strategy,
      "spread" = cores[((i - 1L) %% length(cores)) + 1L],
      "compact" = cores[min(i, length(cores))]
    )
    ok[i] <- tryCatch(
      isTRUE(parallel::clusterCall(pool$workers[[i]]$cluster, function(c) shard::set_affinity(c), core)[[1]]),
      error = function(e) FALSE
    )
  }
  invisible(ok)
}

#' Advise OS about expected access pattern for a segment
#'
#' This calls `madvise()` on the segment mapping when available.
#'
#' @param seg A shard_segment.
#' @param advice One of "normal", "sequential", "random", "willneed", "dontneed".
#' @return A logical scalar; \code{TRUE} if the OS accepted the hint.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' segment_advise(seg, "sequential")
#' }
segment_advise <- function(seg,
                           advice = c("normal", "sequential", "random", "willneed", "dontneed")) {
  if (!(inherits(seg, "shard_segment") || identical(typeof(seg), "externalptr"))) {
    stop("seg must be a shard_segment or an external pointer", call. = FALSE)
  }
  advice <- match.arg(advice)
  adv_int <- switch(advice,
    "normal" = 0L,
    "sequential" = 1L,
    "random" = 2L,
    "willneed" = 3L,
    "dontneed" = 4L
  )
  ptr <- if (inherits(seg, "shard_segment")) seg$ptr else seg
  isTRUE(.Call("C_shard_segment_madvise", ptr, adv_int, PACKAGE = "shard"))
}

#' Advise access pattern for a buffer
#'
#' @param x A shard_buffer.
#' @param advice See [segment_advise()].
#' @return A logical scalar; \code{TRUE} if the OS accepted the hint.
#' @export
#' @examples
#' \donttest{
#' buf <- buffer("double", dim = 10L)
#' buffer_advise(buf, "sequential")
#' }
buffer_advise <- function(x,
                          advice = c("normal", "sequential", "random", "willneed", "dontneed")) {
  stopifnot(inherits(x, "shard_buffer"))
  segment_advise(x$segment, advice = advice)
}

#' Advise access pattern for a shared input vector/matrix
#'
#' @param x A shard shared vector (from [share()]).
#' @param advice See [segment_advise()].
#' @return A logical scalar; \code{TRUE} if the OS accepted the hint.
#' @export
#' @examples
#' \donttest{
#' x <- as_shared(1:100)
#' shared_advise(x, "sequential")
#' }
shared_advise <- function(x,
                          advice = c("normal", "sequential", "random", "willneed", "dontneed")) {
  if (!is_shared_vector(x)) stop("x must be a shared vector", call. = FALSE)
  segment_advise(shared_segment(x), advice = advice)
}

test_that("segment/buffer/shared advise functions are safe", {
  skip_on_cran()

  seg <- segment_create(4096, backing = "mmap")
  on.exit(segment_close(seg, unlink = TRUE), add = TRUE)

  ok <- segment_advise(seg, "sequential")
  expect_true(is.logical(ok))
  expect_length(ok, 1)

  buf <- buffer("double", dim = 16, backing = "mmap")
  ok2 <- buffer_advise(buf, "random")
  expect_true(is.logical(ok2))
  expect_length(ok2, 1)

  x <- share(matrix(rnorm(100), nrow = 10), backing = "mmap")
  ok3 <- shared_advise(x, "willneed")
  expect_true(is.logical(ok3))
  expect_length(ok3, 1)
})

test_that("affinity helpers are feature-gated", {
  skip_on_cran()

  sup <- affinity_supported()
  expect_true(is.logical(sup))
  expect_length(sup, 1)

  # set_affinity is a safe no-op on unsupported platforms.
  ok <- set_affinity(0L)
  expect_true(is.logical(ok))
  expect_length(ok, 1)
})

test_that("segment_advise supports all advice types", {
  skip_on_cran()

  seg <- segment_create(4096, backing = "mmap")
  on.exit(segment_close(seg, unlink = TRUE), add = TRUE)

  for (advice in c("normal", "sequential", "random", "willneed", "dontneed")) {
    ok <- segment_advise(seg, advice)
    expect_true(is.logical(ok))
    expect_length(ok, 1)
  }
})

test_that("segment_advise validates seg argument", {
  expect_error(segment_advise(NULL, "normal"),
               "seg must be a shard_segment or an external pointer")
  expect_error(segment_advise("not a segment", "normal"),
               "seg must be a shard_segment or an external pointer")
})

test_that("buffer_advise validates buffer argument", {
  expect_error(buffer_advise(NULL, "normal"))
  expect_error(buffer_advise("not a buffer", "normal"))
})

test_that("shared_advise validates shared vector argument", {
  expect_error(shared_advise(1:10, "normal"),
               "x must be a shared vector")
  expect_error(shared_advise(matrix(1:10), "normal"),
               "x must be a shared vector")
})

test_that("pin_workers requires active pool", {
  skip_on_cran()

  pool_stop()  # Ensure no pool is active

  expect_error(pin_workers(), "No active pool")
})

test_that("pin_workers returns logical vector for each worker", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # pin_workers returns FALSE on unsupported platforms (macOS)
  result <- pin_workers()

  expect_true(is.logical(result))
  expect_length(result, 2)

  # On non-Linux, should return FALSE for all workers
  if (!affinity_supported()) {
    expect_true(all(result == FALSE))
  }
})

test_that("pin_workers spread strategy distributes workers", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Test spread strategy
  result <- pin_workers(strategy = "spread")
  expect_true(is.logical(result))
  expect_length(result, 2)
})

test_that("pin_workers compact strategy assigns sequential cores", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Test compact strategy
  result <- pin_workers(strategy = "compact")
  expect_true(is.logical(result))
  expect_length(result, 2)
})

test_that("pin_workers accepts custom core list", {
  skip_on_cran()

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  # Specify specific cores
  result <- pin_workers(cores = c(0L, 1L))
  expect_true(is.logical(result))
  expect_length(result, 2)
})

test_that("pin_workers validates cores argument on supported platforms", {
  skip_on_cran()

  # On unsupported platforms, pin_workers returns early before checking cores
  if (!affinity_supported()) skip("affinity not supported on this platform")

  pool_stop()
  pool_create(2)
  on.exit(pool_stop(), add = TRUE)

  expect_error(pin_workers(cores = integer(0)), "cores must be non-empty")
})

test_that("set_affinity accepts vector of cores", {
  skip_on_cran()

  # Even on unsupported platforms, should not error
  ok1 <- set_affinity(0L)
  expect_true(is.logical(ok1))

  ok2 <- set_affinity(c(0L, 1L))
  expect_true(is.logical(ok2))
})

#' @title Zero-Copy Shared Objects
#' @name share
#' @description Create shared memory representations of R objects for efficient
#'   parallel access without duplication.
#'
#' @details
#' The `share()` function is the primary high-level API for creating zero-copy
#' shared inputs. When you share an object:
#'
#' \enumerate{
#'   \item The object is serialized into a shared memory segment
#'   \item The segment is marked read-only (protected)
#'   \item A lightweight handle is returned that can be passed to workers
#'   \item Workers attach to the segment and deserialize on demand
#' }
#'
#' This approach eliminates per-worker duplication of large inputs. The data
#' exists once in shared memory, and all workers read from the same location.
#'
#' \strong{Immutability Contract}: Shared objects are immutable by design.
#' Any attempt to modify shared data in a worker will fail. This guarantees
#' deterministic behavior and prevents accidental copy-on-write.
#'
#' @seealso \code{\link{segment_create}} for low-level segment operations,
#'   \code{\link{pool_create}} for worker pool management.
#'
#' @useDynLib shard, .registration = TRUE
NULL

# Internal helper: check for non-serializable objects
# Returns NULL on success, stops with error on failure
validate_serializable <- function(x, path = "x", seen = NULL) {
    # Guard against cycles/aliases during validation (environments can be truly
    # cyclic; lists can alias). We only need to ensure we don't recurse forever.
    if (is.null(seen)) {
        seen <- new.env(parent = emptyenv())
    }

    # Use address token for memoization; if this fails, fall back to best-effort.
    id <- tryCatch(object_identity(x), error = function(e) NULL)
    if (!is.null(id) && exists(id, envir = seen, inherits = FALSE)) {
        return(invisible(NULL))
    }
    if (!is.null(id)) {
        assign(id, TRUE, envir = seen)
    }

    # Check for functions (closures)
    if (is.function(x)) {
        stop("Cannot share functions (closures).\n",
             "  Found at: ", path, "\n",
             "  Hint: Extract the data you need and share that instead.",
             call. = FALSE)
    }

    # Check for external pointers
    if (typeof(x) == "externalptr") {
        stop("Cannot share external pointers.\n",
             "  Found at: ", path, "\n",
             "  External pointers reference memory that cannot be serialized.",
             call. = FALSE)
    }

    # Check for connections
    if (inherits(x, "connection")) {
        stop("Cannot share connection objects.\n",
             "  Found at: ", path, "\n",
             "  Connections cannot be serialized.",
             call. = FALSE)
    }

    # Recursively check environments
    if (is.environment(x)) {
        # Skip base/global/empty environments (these serialize fine)
        if (identical(x, baseenv()) ||
            identical(x, globalenv()) ||
            identical(x, emptyenv())) {
            return(invisible(NULL))
        }

        # Check contents of user environments
        for (nm in names(x)) {
            validate_serializable(x[[nm]], paste0(path, "$", nm), seen = seen)
        }
    }

    # Recursively check lists
    if (is.list(x) && !is.data.frame(x)) {
        nms <- names(x)
        for (i in seq_along(x)) {
            item_path <- if (!is.null(nms) && nzchar(nms[i])) {
                paste0(path, "$", nms[i])
            } else {
                paste0(path, "[[", i, "]]")
            }
            validate_serializable(x[[i]], item_path, seen = seen)
        }
    }

    # Recursively check S4 slots
    if (isS4(x)) {
        slot_names <- methods::slotNames(x)
        for (sn in slot_names) {
            validate_serializable(methods::slot(x, sn), paste0(path, "@", sn), seen = seen)
        }
    }

    invisible(NULL)
}

# Helper: get object identity (memory address) for alias detection
object_identity <- function(x) {
    # Prefer a stable, API-safe address token from C. This avoids using
    # .Internal(inspect()), which is not part of R's public API.
    tryCatch(
        .Call("C_shard_obj_address", x, PACKAGE = "shard"),
        error = function(e) paste0("obj_", sample.int(1e9, 1))
    )
}

# Helper: check if object is a shareable atomic type
is_shareable_atomic <- function(x) {
    # Shareable: numeric vectors (double, integer, complex), logical, raw
    # Not shareable: character, lists, environments, functions, etc.
    if (is.atomic(x) && !is.null(x)) {
        mode <- typeof(x)
        # Complex is intentionally excluded: shard ALTREP currently supports
        # double/integer/logical/raw only.
        return(mode %in% c("double", "integer", "logical", "raw"))
    }
    FALSE
}

#' Deep Sharing Hook for Custom Classes
#'
#' S3 generic that allows classes to customize deep sharing behavior.
#' Override this for your class to control which slots/elements are traversed,
#' force sharing of small objects, or transform objects before traversal.
#'
#' @param x The object being traversed during deep sharing.
#' @param ctx A context list containing:
#'   \describe{
#'     \item{path}{Current node path string (e.g., \code{"<root>$data@cache"})}
#'     \item{class}{class(x) - the object's class vector}
#'     \item{mode}{'strict' or 'balanced' - sharing mode}
#'     \item{min_bytes}{Minimum size threshold for sharing}
#'     \item{types}{Character vector of enabled types for sharing}
#'     \item{deep}{Logical, always TRUE when hook is called}
#'   }
#'
#' @return A list with optional fields:
#'   \describe{
#'     \item{skip_slots}{Character vector of S4 slot names to not traverse}
#'     \item{skip_paths}{Character vector of paths to not traverse}
#'     \item{force_share_paths}{Character vector of paths to force share (ignore min_bytes)}
#'     \item{rewrite}{Function(x) -> x to transform object before traversal}
#'   }
#'   Return an empty list for default behavior (no customization).
#'
#' @export
#' @examples
#' \donttest{
#' shard_share_hook.MyModelClass <- function(x, ctx) {
#'     list(
#'         skip_slots = "cache",
#'         force_share_paths = paste0(ctx$path, "@coefficients")
#'     )
#' }
#'
#' shard_share_hook.LazyData <- function(x, ctx) {
#'     list(
#'         rewrite = function(obj) {
#'             obj$data <- as.matrix(obj$data)
#'             obj
#'         }
#'     )
#' }
#' }
shard_share_hook <- function(x, ctx) {
    UseMethod("shard_share_hook")
}

#' @export
#' @rdname shard_share_hook
shard_share_hook.default <- function(x, ctx) {
    # Default: no customization, traverse normally
    list()
}

# Internal: build context object for hooks
build_hook_context <- function(x, path, mode, min_bytes, types, deep) {
    list(
        path = path,
        class = class(x),
        mode = mode,
        min_bytes = min_bytes,
        types = types,
        deep = deep
    )
}

# Internal: call hook safely with error handling based on mode
call_hook_safe <- function(x, ctx, mode) {
    result <- tryCatch(
        shard_share_hook(x, ctx),
        error = function(e) {
            if (mode == "strict") {
                stop("Hook error in strict mode.\n",
                     "  Path: ", ctx$path, "\n",
                     "  Class: ", paste(ctx$class, collapse = ", "), "\n",
                     "  Error: ", conditionMessage(e),
                     call. = FALSE)
            }
            # In balanced mode, return special marker indicating hook failure
            structure(
                list(hook_error = conditionMessage(e)),
                class = "shard_hook_error"
            )
        }
    )

    # Validate hook return value
    if (!is.list(result)) {
        if (mode == "strict") {
            stop("Hook must return a list.\n",
                 "  Path: ", ctx$path, "\n",
                 "  Class: ", paste(ctx$class, collapse = ", "), "\n",
                 "  Got: ", typeof(result),
                 call. = FALSE)
        }
        return(structure(
            list(hook_error = "Hook did not return a list"),
            class = "shard_hook_error"
        ))
    }

    result
}

# Internal: check if path should be skipped based on hook result
should_skip_path <- function(current_path, hook_result) {
    if (is.null(hook_result$skip_paths)) return(FALSE)
    current_path %in% hook_result$skip_paths
}

# Internal: check if path should be force-shared based on hook result
should_force_share_path <- function(current_path, hook_result) {
    if (is.null(hook_result$force_share_paths)) return(FALSE)
    current_path %in% hook_result$force_share_paths
}

# Internal: traverse object for deep sharing with memoization and hooks
# Returns a structure with shared segments and alias information
share_deep_traverse <- function(x,
                                env,
                                path = "<root>",
                                depth = 0,
                                min_bytes = 64 * 1024 * 1024,
                                backing = "auto",
                                readonly = TRUE,
                                max_depth = Inf,
                                cycle_policy = "error",
                                mode = "balanced",
                                types = c("double", "integer", "logical", "raw", "complex"),
                                hook_result = NULL) {
    env_has_self_cycle_ <- function(e) {
        # R lists don't form true cycles, but environments can via bindings or
        # parent.env(e) == e. We don't deep-traverse environments yet, but we
        # should fail fast on obvious self-cycles to avoid pathological cases.
        if (!is.environment(e)) return(FALSE)
        if (identical(parent.env(e), e)) return(TRUE)
        nms <- ls(envir = e, all.names = TRUE)
        if (length(nms) == 0) return(FALSE)
        vals <- mget(nms, envir = e, inherits = FALSE)
        for (v in vals) {
            if (is.environment(v) && identical(v, e)) return(TRUE)
        }
        FALSE
    }

    # Handle environment objects based on mode
    if (is.environment(x) && !identical(x, baseenv()) &&
        !identical(x, globalenv()) && !identical(x, emptyenv())) {
        if (mode == "strict") {
            stop("Cannot share: contains environment.\n",
                 "  Found at: ", path, "\n",
                 "  Hint: Use mode='balanced' to skip environment slots.",
                 call. = FALSE)
        }
        if (cycle_policy == "error" && env_has_self_cycle_(x)) {
            stop("Cycle detected during deep sharing.\n",
                 "  Path: ", path, "\n",
                 "  Cause: environment self-reference.\n",
                 "  Hint: Use cycle='skip' to skip cyclic references instead.",
                 call. = FALSE)
        }
        # In balanced mode, keep environments as-is
        identity <- object_identity(x)
        return(structure(
            list(
                type = "kept",
                path = path,
                node_id = env$next_id + 1,
                identity = identity,
                value = x
            ),
            class = "shard_deep_kept"
        ))
    }

    # Get object identity for memoization and cycle detection
    identity <- object_identity(x)

    # Check for cycle (currently in traversal stack)
    if (exists(identity, envir = env$in_progress, inherits = FALSE)) {
        if (cycle_policy == "error") {
            stop("Cycle detected during deep sharing.\n",
                 "  Path: ", path, "\n",
                 "  Object is self-referential.\n",
                 "  Use cycle='skip' to skip cyclic references instead.",
                 call. = FALSE)
        }
        # cycle='skip': return a marker indicating cycle
        return(structure(
            list(
                type = "cycle",
                path = path,
                identity = identity
            ),
            class = "shard_deep_cycle"
        ))
    }

    # Check for alias (already seen and shared)
    if (exists(identity, envir = env$seen, inherits = FALSE)) {
        original <- get(identity, envir = env$seen, inherits = FALSE)
        return(structure(
            list(
                type = "alias",
                path = path,
                identity = identity,
                alias_of = original$path,
                alias_node_id = original$node_id,
                shared = original$shared,
                value = original$value  # For aliases to kept values
            ),
            class = "shard_deep_alias"
        ))
    }

    # Mark as in-progress for cycle detection
    original_identity <- identity  # Save for cleanup
    assign(identity, TRUE, envir = env$in_progress)
    on.exit({
        if (exists(original_identity, envir = env$in_progress, inherits = FALSE)) {
            rm(list = original_identity, envir = env$in_progress)
        }
    }, add = TRUE)

    # Assign a node ID for tracking
    env$next_id <- env$next_id + 1
    node_id <- env$next_id

    # Call hook for this object
    # Build context and call hook
    ctx <- build_hook_context(x, path, mode, min_bytes, types, TRUE)
    current_hook_result <- call_hook_safe(x, ctx, mode)

    # Check for hook error
    if (inherits(current_hook_result, "shard_hook_error")) {
        # In balanced mode, continue with default behavior but track error
        env$hook_errors <- c(env$hook_errors, list(list(
            path = path,
            error = current_hook_result$hook_error
        )))
        current_hook_result <- list()  # Use empty hook result
    }

    # Store hook result in environment for child access
    if (length(current_hook_result) > 0) {
        env$active_hooks[[path]] <- current_hook_result
    }

    # Apply rewrite function if provided
    if (!is.null(current_hook_result$rewrite) && is.function(current_hook_result$rewrite)) {
        x <- tryCatch(
            current_hook_result$rewrite(x),
            error = function(e) {
                if (mode == "strict") {
                    stop("Rewrite function error in strict mode.\n",
                         "  Path: ", path, "\n",
                         "  Error: ", conditionMessage(e),
                         call. = FALSE)
                }
                env$hook_errors <- c(env$hook_errors, list(list(
                    path = path,
                    error = paste("rewrite error:", conditionMessage(e))
                )))
                x  # Return original on error in balanced mode
            }
        )
        # Re-get identity after rewrite (object may have changed)
        identity <- object_identity(x)
    }

    # Check force_share from parent hooks as well
    force_share <- should_force_share_path(path, current_hook_result)
    if (!force_share) {
        # Check parent hook results for force_share_paths
        for (hook_path in names(env$active_hooks)) {
            if (should_force_share_path(path, env$active_hooks[[hook_path]])) {
                force_share <- TRUE
                break
            }
        }
    }

    # Check for registered adapter before default traversal
    adapter <- shard_get_adapter(x)
    if (!is.null(adapter) && depth < max_depth) {
        # Use adapter to get children
        adapter_children <- adapter$children(x)

        if (!is.list(adapter_children)) {
            stop("Adapter children() must return a list for class '",
                 adapter$class, "'.", call. = FALSE)
        }

        # Process each child from the adapter
        nms <- names(adapter_children)
        children <- vector("list", length(adapter_children))
        names(children) <- nms

        path_prefix <- if (!is.null(adapter$path_prefix)) {
            adapter$path_prefix
        } else {
            paste0("@", adapter$class)
        }

        for (i in seq_along(adapter_children)) {
            child_path <- if (!is.null(nms) && nzchar(nms[i])) {
                paste0(path, path_prefix, "$", nms[i])
            } else {
                paste0(path, path_prefix, "[[", i, "]]")
            }
            children[[i]] <- share_deep_traverse(
                adapter_children[[i]], env, child_path, depth + 1,
                min_bytes, backing, readonly, max_depth, cycle_policy,
                mode, types, current_hook_result
            )
        }

        # Record in seen table
        assign(identity, list(
            path = path,
            node_id = node_id,
            shared = NULL
        ), envir = env$seen)

        return(structure(
            list(
                type = "container",
                container_type = "adapter",
                adapter_class = adapter$class,
                path = path,
                node_id = node_id,
                identity = identity,
                children = children,
                original_value = x,
                adapter = adapter,
                original_class = class(x),
                original_names = nms
            ),
            class = "shard_deep_container"
        ))
    }

    # Determine what to do with this object
    should_share <- is_shareable_atomic(x) &&
        (force_share || utils::object.size(x) >= min_bytes)

    if (should_share) {
        # Share this atomic object
        shared <- share(x, backing = backing, readonly = readonly)

        # Record in seen table for alias detection
        assign(identity, list(
            path = path,
            node_id = node_id,
            shared = shared
        ), envir = env$seen)

        # Track the shared segment for cleanup
        env$segments <- c(env$segments, list(shared))

        return(structure(
            list(
                type = "shared",
                path = path,
                node_id = node_id,
                identity = identity,
                shared = shared,
                forced = force_share
            ),
            class = "shard_deep_node"
        ))
    } else if (isS4(x) && depth < max_depth) {
        # Process S4 object slots
        all_slot_names <- methods::slotNames(x)

        # Get skipped slots from hook
        skipped_slots <- current_hook_result$skip_slots

        children <- vector("list", length(all_slot_names))
        names(children) <- all_slot_names

        for (sn in all_slot_names) {
            child_path <- paste0(path, "@", sn)
            slot_val <- methods::slot(x, sn)

            # Check if this slot should be skipped (either by skip_slots or skip_paths)
            slot_skipped <- sn %in% skipped_slots ||
                            should_skip_path(child_path, current_hook_result)

            if (slot_skipped) {
                # Keep the slot value as-is (don't traverse)
                children[[sn]] <- structure(
                    list(
                        type = "kept",
                        path = child_path,
                        node_id = env$next_id + 1,
                        identity = object_identity(slot_val),
                        value = slot_val,
                        skipped_by_hook = TRUE
                    ),
                    class = "shard_deep_kept"
                )
                env$next_id <- env$next_id + 1
            } else {
                children[[sn]] <- share_deep_traverse(
                    slot_val, env, child_path, depth + 1,
                    min_bytes, backing, readonly, max_depth, cycle_policy,
                    mode, types, current_hook_result
                )
            }
        }

        # Record in seen table
        assign(identity, list(
            path = path,
            node_id = node_id,
            shared = NULL
        ), envir = env$seen)

        return(structure(
            list(
                type = "container",
                container_type = "S4",
                path = path,
                node_id = node_id,
                identity = identity,
                children = children,
                original_class = class(x),
                skipped_slots = skipped_slots
            ),
            class = "shard_deep_container"
        ))
    } else if (is.list(x) && !is.data.frame(x) && depth < max_depth) {
        # Recursively process list elements
        nms <- names(x)
        children <- vector("list", length(x))
        names(children) <- nms

        for (i in seq_along(x)) {
            child_path <- if (!is.null(nms) && nzchar(nms[i])) {
                paste0(path, "$", nms[i])
            } else {
                paste0(path, "[[", i, "]]")
            }

            # Check if this path should be skipped
            if (should_skip_path(child_path, current_hook_result)) {
                children[[i]] <- structure(
                    list(
                        type = "kept",
                        path = child_path,
                        node_id = env$next_id + 1,
                        identity = object_identity(x[[i]]),
                        value = x[[i]],
                        skipped_by_hook = TRUE
                    ),
                    class = "shard_deep_kept"
                )
                env$next_id <- env$next_id + 1
            } else {
                children[[i]] <- share_deep_traverse(
                    x[[i]], env, child_path, depth + 1,
                    min_bytes, backing, readonly, max_depth, cycle_policy,
                    mode, types, current_hook_result
                )
            }
        }

        # Record in seen table
        assign(identity, list(
            path = path,
            node_id = node_id,
            shared = NULL
        ), envir = env$seen)

        return(structure(
            list(
                type = "container",
                container_type = "list",
                path = path,
                node_id = node_id,
                identity = identity,
                children = children,
                original_class = class(x),
                original_names = nms
            ),
            class = "shard_deep_container"
        ))
    } else if (is.data.frame(x) && depth < max_depth) {
        # Process data.frame columns
        children <- vector("list", ncol(x))
        names(children) <- names(x)

        for (i in seq_along(x)) {
            child_path <- paste0(path, "$", names(x)[i])

            # Check if this path should be skipped
            if (should_skip_path(child_path, current_hook_result)) {
                children[[i]] <- structure(
                    list(
                        type = "kept",
                        path = child_path,
                        node_id = env$next_id + 1,
                        identity = object_identity(x[[i]]),
                        value = x[[i]],
                        skipped_by_hook = TRUE
                    ),
                    class = "shard_deep_kept"
                )
                env$next_id <- env$next_id + 1
            } else {
                children[[i]] <- share_deep_traverse(
                    x[[i]], env, child_path, depth + 1,
                    min_bytes, backing, readonly, max_depth, cycle_policy,
                    mode, types, current_hook_result
                )
            }
        }

        # Record in seen table
        assign(identity, list(
            path = path,
            node_id = node_id,
            shared = NULL
        ), envir = env$seen)

        return(structure(
            list(
                type = "container",
                container_type = "data.frame",
                path = path,
                node_id = node_id,
                identity = identity,
                children = children,
                original_class = class(x),
                original_names = names(x),
                original_nrow = nrow(x),
                original_row_names = attr(x, "row.names")
            ),
            class = "shard_deep_container"
        ))
    } else if (isS4(x) && depth < max_depth) {
        # S4 object: generic slot traversal (no adapter registered)
        slot_names <- methods::slotNames(x)

        if (length(slot_names) > 0) {
            children <- vector("list", length(slot_names))
            names(children) <- slot_names

            for (i in seq_along(slot_names)) {
                sn <- slot_names[i]
                child_path <- paste0(path, "@", sn)
                children[[i]] <- share_deep_traverse(
                    methods::slot(x, sn), env, child_path, depth + 1,
                    min_bytes, backing, readonly, max_depth, cycle_policy
                )
            }

            # Record in seen table
            assign(identity, list(
                path = path,
                node_id = node_id,
                shared = NULL
            ), envir = env$seen)

            return(structure(
                list(
                    type = "container",
                    container_type = "s4",
                    path = path,
                    node_id = node_id,
                    identity = identity,
                    children = children,
                    original_value = x,
                    original_class = class(x),
                    original_names = slot_names
                ),
                class = "shard_deep_container"
            ))
        }
        # S4 with no slots falls through to kept
    }

    # Keep as-is (small atomic, unshareable type, max depth reached, or S4 with no slots)
    # Record in seen table (including value for alias reconstruction)
    assign(identity, list(
        path = path,
        node_id = node_id,
        shared = NULL,
        value = x
    ), envir = env$seen)

    return(structure(
        list(
            type = "kept",
            path = path,
            node_id = node_id,
            identity = identity,
            value = x
        ),
        class = "shard_deep_kept"
    ))
}

# Internal: reconstruct object from deep shared structure
fetch_deep_reconstruct <- function(node) {
    if (inherits(node, "shard_deep_node")) {
        # Shared leaf - fetch and materialize so reconstruction yields ordinary
        # R objects (important for S4 slot assignment and for equality checks
        # against the original, non-shared objects).
        return(materialize(fetch(node$shared)))
    } else if (inherits(node, "shard_deep_alias")) {
        # Alias - fetch from the original shared segment or return kept value
        if (!is.null(node$shared)) {
            return(materialize(fetch(node$shared)))
        } else {
            # Alias to a kept value - return the stored value
            return(node$value)
        }
    } else if (inherits(node, "shard_deep_cycle")) {
        # Cycle marker - return NULL or special marker
        # This shouldn't normally be reached if cycle='error'
        return(NULL)
    } else if (inherits(node, "shard_deep_kept")) {
        # Kept value - return as-is
        return(node$value)
    } else if (inherits(node, "shard_deep_container")) {
        # Container - reconstruct recursively
        children <- lapply(node$children, fetch_deep_reconstruct)

        if (node$container_type == "adapter") {
            # Use adapter's replace function to reconstruct
            result <- node$adapter$replace(node$original_value, children)
        } else if (node$container_type == "s4") {
            # Reconstruct S4 object by setting slots
            result <- node$original_value  # Start with a copy
            for (sn in names(children)) {
                methods::slot(result, sn) <- children[[sn]]
            }
        } else if (node$container_type == "data.frame") {
            # Reconstruct data.frame
            result <- structure(
                children,
                class = node$original_class,
                names = node$original_names,
                row.names = node$original_row_names
            )
        } else if (node$container_type == "S4") {
            # Reconstruct S4 object
            # Create new instance of the class
            result <- methods::new(node$original_class)
            # Set each slot
            for (sn in names(children)) {
                methods::slot(result, sn) <- children[[sn]]
            }
        } else {
            # Reconstruct list
            result <- children
            if (!is.null(node$original_names)) {
                names(result) <- node$original_names
            }
            if (!identical(node$original_class, "list")) {
                class(result) <- node$original_class
            }
        }
        return(result)
    } else {
        # Unknown type - return as-is
        return(node)
    }
}

#' Share an R Object for Parallel Access
#'
#' Creates a shared memory representation of an R object. The object is
#' serialized once and can be accessed by multiple worker processes without
#' copying.
#'
#' @param x An R object to share. Supports vectors, matrices, arrays, lists,
#'   data frames, and any object that can be serialized with \code{serialize()}.
#' @param backing Backing type: "auto" (default), "mmap", or "shm".
#'   \itemize{
#'     \item \code{"auto"}: Let the system choose the best option.
#'     \item \code{"mmap"}: File-backed memory mapping. Most portable.
#'     \item \code{"shm"}: POSIX shared memory or Windows named mapping.
#'   }
#' @param readonly Logical. If TRUE (default), the segment is protected after
#'   writing, making it read-only. Set to FALSE only if you need to modify
#'   the shared data (advanced use case).
#' @param name Optional name for the shared object. If NULL (default), a unique
#'   name is generated. Named shares can be opened by name in other processes.
#' @param deep Logical. If TRUE, recursively traverse lists and data.frames,
#'   sharing individual components that meet the size threshold. When FALSE
#'   (default), the entire object is serialized as one unit.
#' @param min_bytes Minimum size in bytes for an object to be shared when
#'   deep=TRUE. Objects smaller than this threshold are kept in-place.
#'   Default is 64MB (64 * 1024 * 1024).
#' @param cycle How to handle cyclic references when deep=TRUE. Either "error"
#'   (default) to stop with an error, or "skip" to skip cyclic references.
#' @param mode Sharing mode when deep=TRUE. Either "balanced" (default) to
#'   continue on hook errors and non-shareable types, or "strict" to error.
#'
#' @return A \code{shard_shared} object (when deep=FALSE) or
#'   \code{shard_deep_shared} object (when deep=TRUE) containing:
#'   \itemize{
#'     \item \code{path}: The path or name of the shared segment
#'     \item \code{backing}: The backing type used
#'     \item \code{size}: Total size in bytes
#'     \item \code{readonly}: Whether the segment is protected
#'     \item \code{class_info}: Original class information
#'   }
#'
#' @export
#' @examples
#' \donttest{
#' mat <- matrix(rnorm(1e4), nrow = 100)
#' shared_mat <- share(mat)
#' recovered <- fetch(shared_mat)
#' identical(mat, recovered)
#' close(shared_mat)
#' }
share <- function(x,
                  backing = c("auto", "mmap", "shm"),
                  readonly = TRUE,
                  name = NULL,
                  deep = FALSE,
                  min_bytes = 64 * 1024 * 1024,
                  cycle = c("error", "skip"),
                  mode = c("balanced", "strict")) {
    backing <- match.arg(backing)
    cycle <- match.arg(cycle)
    mode <- match.arg(mode)

    # Validate input is serializable
    validate_serializable(x)

    # Shareable types
    shareable_types <- c("double", "integer", "logical", "raw", "complex")

    # Deep sharing: traverse structure and share components individually
    if (deep) {
        # Create environment for memoization state
        env <- new.env(parent = emptyenv())
        env$seen <- new.env(parent = emptyenv())       # identity -> {path, node_id, shared}
        env$in_progress <- new.env(parent = emptyenv()) # identity -> TRUE (cycle detection)
        env$segments <- list()                          # All created segments for cleanup
        env$next_id <- 0                               # Node ID counter
        env$hook_errors <- list()                      # Track hook errors in balanced mode
        env$active_hooks <- list()                     # Active hook results by path

        # Traverse and share
        structure_tree <- share_deep_traverse(
            x, env, "<root>", 0,
            min_bytes, backing, readonly, Inf, cycle,
            mode, shareable_types, NULL
        )

        # Count shared vs aliased
        shared_count <- 0
        alias_count <- 0
        cycle_count <- 0
        kept_count <- 0
        total_shared_bytes <- 0

        count_nodes <- function(node) {
            if (inherits(node, "shard_deep_node")) {
                shared_count <<- shared_count + 1
                if (inherits(node$shared, "shard_shared")) {
                    total_shared_bytes <<- total_shared_bytes + node$shared$size
                } else if (is_shared_vector(node$shared)) {
                    elem_size <- switch(typeof(node$shared),
                        "integer" = 4L,
                        "double"  = 8L,
                        "logical" = 4L,
                        "raw"     = 1L,
                        NA_integer_
                    )
                    if (!is.na(elem_size)) {
                        total_shared_bytes <<- total_shared_bytes + length(node$shared) * elem_size
                    }
                }
            } else if (inherits(node, "shard_deep_alias")) {
                alias_count <<- alias_count + 1
            } else if (inherits(node, "shard_deep_cycle")) {
                cycle_count <<- cycle_count + 1
            } else if (inherits(node, "shard_deep_kept")) {
                kept_count <<- kept_count + 1
            } else if (inherits(node, "shard_deep_container")) {
                for (child in node$children) {
                    count_nodes(child)
                }
            }
        }
        count_nodes(structure_tree)

        return(structure(
            list(
                tree = structure_tree,
                segments = env$segments,
                backing = backing,
                readonly = readonly,
                mode = mode,
                summary = list(
                    shared_count = shared_count,
                    alias_count = alias_count,
                    cycle_count = cycle_count,
                    kept_count = kept_count,
                    total_shared_bytes = total_shared_bytes,
                    hook_error_count = length(env$hook_errors),
                    hook_errors = env$hook_errors
                ),
                class_info = list(
                    type = if (is.data.frame(x)) "data.frame"
                           else if (is.list(x)) "list"
                           else if (isS4(x)) "S4"
                           else "other",
                    deep = TRUE
                )
            ),
            class = "shard_deep_shared"
        ))
    }

    # Fast path: shareable atomic vectors/matrices/arrays become ALTREP-backed
    # shared vectors. This avoids per-worker serialization of large inputs.
    if (is.atomic(x) && !is.null(x) &&
        typeof(x) %in% c("double", "integer", "logical", "raw") &&
        !is_shared_vector(x)) {
        cow <- if (isTRUE(readonly)) "deny" else "allow"
        # Build with cow='allow' so we can attach attributes, then lock down
        # by setting shard_cow to the requested policy.
        shared <- as_shared(x, readonly = readonly, backing = backing, cow = "allow")

        # Preserve non-class attributes (dim, dimnames, names, tsp, etc).
        attrs <- attributes(x)
        x_class <- attr(x, "class")
        attrs$class <- NULL
        if (length(attrs)) {
            for (nm in names(attrs)) {
                attr(shared, nm) <- attrs[[nm]]
            }
        }

        # Preserve any underlying class (e.g., Date) behind the wrapper class.
        class(shared) <- unique(c("shard_shared_vector", x_class))
        attr(shared, "shard_cow") <- cow
        return(shared)
    }

    # Standard (non-deep) sharing: serialize entire object

    # Serialize the object (with tryCatch for edge cases)
    serialized <- tryCatch(
        serialize(x, connection = NULL, xdr = FALSE),
        error = function(e) {
            stop("Cannot share object: serialization failed.\n",
                 "Reason: ", conditionMessage(e), "\n",
                 "Objects containing closures, external pointers, or ",
                 "certain connection types cannot be shared.",
                 call. = FALSE)
        }
    )
    size <- length(serialized)

    # Create segment with enough space for the serialized data
    seg <- segment_create(size, backing = backing, path = name, readonly = FALSE)

    # Write serialized data
    segment_write(seg, serialized, offset = 0)

    # Protect if requested
    if (readonly) {
        segment_protect(seg)
    }

    # Get segment info for the handle
    info <- segment_info(seg)

    # Extract class information for better deserialization hints
    class_info <- if (is.matrix(x)) {
        list(type = "matrix", dim = dim(x), mode = mode(x))
    } else if (is.array(x)) {
        list(type = "array", dim = dim(x), mode = mode(x))
    } else if (is.data.frame(x)) {
        list(type = "data.frame", nrow = nrow(x), ncol = ncol(x))
    } else if (is.list(x)) {
        list(type = "list", length = length(x))
    } else {
        list(type = "vector", length = length(x), mode = mode(x))
    }

    structure(
        list(
            segment = seg,
            path = info$path,
            backing = info$backing,
            size = info$size,
            readonly = readonly,
            class_info = class_info
        ),
        class = "shard_shared"
    )
}

#' Fetch Data from a Shared Object
#'
#' Retrieves the R object from shared memory by deserializing it. This is the
#' primary way to access shared data in workers.
#'
#' @param x A \code{shard_shared} object.
#' @param ... Ignored.
#'
#' @return The original R object that was shared.
#'
#' @details
#' When called in the main process, this reads from the existing segment.
#' When called in a worker process, this opens the segment by path and
#' deserializes the data.
#'
#' The \code{fetch()} function is the primary way to access shared data.
#' It can also be called as \code{materialize()} for compatibility.
#'
#' @export
#' @examples
#' \donttest{
#' x <- 1:100
#' shared <- share(x)
#' recovered <- fetch(shared)
#' identical(x, recovered)
#' close(shared)
#' }
fetch <- function(x, ...) {
    UseMethod("fetch")
}

#' @rdname fetch
#' @export
fetch.shard_shared <- function(x, ...) {
    # Try using existing segment pointer (main process scenario)
    # Wrap in tryCatch because pointer becomes invalid after serialization to workers
    if (!is.null(x$segment) && inherits(x$segment, "shard_segment")) {
        result <- tryCatch({
            raw_data <- segment_read(x$segment, offset = 0, size = x$size)
            unserialize(raw_data)
        }, error = function(e) NULL)

        if (!is.null(result)) {
            return(result)
        }
        # Pointer was invalid (e.g., in worker process), fall through to reopen
    }

    # Open by path (worker process scenario)
    if (is.null(x$path)) {
        stop("Cannot fetch: no path available", call. = FALSE)
    }

    seg <- segment_open(x$path, backing = x$backing, readonly = TRUE)
    on.exit(segment_close(seg, unlink = FALSE))

    raw_data <- segment_read(seg, offset = 0, size = x$size)
    unserialize(raw_data)
}

#' @rdname fetch
#' @export
fetch.shard_deep_shared <- function(x, ...) {
    # Reconstruct the object from the shared structure tree
    fetch_deep_reconstruct(x$tree)
}

#' @rdname fetch
#' @export
fetch.default <- function(x, ...) {
    x
}

#' Materialize Shared Object
#'
#' Alias for \code{fetch()}. Retrieves the R object from shared memory.
#'
#' @param x A \code{shard_shared} object.
#' @return The original R object.
#' @export
#' @examples
#' \donttest{
#' shared <- share(1:100)
#' data <- materialize(shared)
#' close(shared)
#' }
materialize <- function(x) {
    UseMethod("materialize")
}

#' @rdname materialize
#' @export
materialize.shard_shared <- function(x) {
    fetch.shard_shared(x)
}

#' @rdname materialize
#' @export
materialize.default <- function(x) {
    # Handle shard ALTREP vectors (from shared_vector/as_shared)
    if (is_shared_vector(x)) {
        out <- .Call("C_shard_altrep_materialize", x, PACKAGE = "shard")

        # Restore non-shard attributes (dim, dimnames, names, and any original
        # underlying class like Date/POSIXct behind the wrapper class).
        attrs <- attributes(x)
        if (!is.null(attrs)) {
            attrs$shard_cow <- NULL
            attrs$shard_readonly <- NULL

            cls <- class(x)
            cls2 <- setdiff(cls, "shard_shared_vector")
            attrs$class <- if (length(cls2)) cls2 else NULL

            attributes(out) <- attrs
        }

        return(out)
    }
    x
}

#' Close a Shared Object
#'
#' Releases the shared memory segment. After closing, the shared object can
#' no longer be accessed.
#'
#' @param con A \code{shard_shared} object.
#' @param ... Ignored.
#'
#' @return NULL (invisibly).
#' @export
#' @method close shard_shared
close.shard_shared <- function(con, ...) {
    if (!is.null(con$segment)) {
        segment_close(con$segment, unlink = TRUE)
    }
    invisible(NULL)
}

#' @rdname close.shard_shared
#' @export
#' @method close shard_shared_vector
close.shard_shared_vector <- function(con, ...) {
    if (!is_shared_vector(con)) {
        stop("con must be a shard ALTREP vector", call. = FALSE)
    }

    seg <- shared_segment(con)
    segment_close(seg)
    invisible(NULL)
}

#' @rdname close.shard_shared
#' @export
#' @method close shard_deep_shared
close.shard_deep_shared <- function(con, ...) {
    # Close all shared segments
    for (seg in con$segments) {
        tryCatch(
            close(seg),
            error = function(e) NULL
        )
    }
    invisible(NULL)
}

#' Check if Object is Shared
#'
#' @param x An object to check.
#' @return A logical scalar: \code{TRUE} if \code{x} is a shared object,
#'   \code{FALSE} otherwise.
#' @export
#' @examples
#' is_shared(1:10)
#' \donttest{
#' shared <- share(1:10)
#' is_shared(shared)
#' close(shared)
#' }
is_shared <- function(x) {
    inherits(x, c("shard_shared", "shard_deep_shared", "shard_shared_vector")) ||
        (is.atomic(x) && is_shared_vector(x))
}

#' Get Information About a Shared Object
#'
#' @param x A \code{shard_shared} object.
#' @return A named list with fields \code{path}, \code{backing}, \code{size},
#'   \code{readonly}, \code{class_info}, and \code{segment_info}.
#' @export
#' @examples
#' \donttest{
#' shared <- share(1:100)
#' shared_info(shared)
#' close(shared)
#' }
shared_info <- function(x) {
    if (inherits(x, "shard_shared")) {
        return(list(
            path = x$path,
            backing = x$backing,
            size = x$size,
            readonly = x$readonly,
            class_info = x$class_info,
            segment_info = if (!is.null(x$segment)) segment_info(x$segment) else NULL
        ))
    }

    if (is.atomic(x) && is_shared_vector(x)) {
        seg <- shared_segment(x)
        sinfo <- .Call("C_shard_segment_info", seg$ptr, PACKAGE = "shard")

        class_info <- if (is.matrix(x)) {
            list(type = "matrix", dim = dim(x), mode = typeof(x))
        } else if (is.array(x) && length(dim(x)) > 1L) {
            list(type = "array", dim = dim(x), mode = typeof(x))
        } else {
            list(type = "vector", length = length(x), mode = typeof(x))
        }

        return(list(
            path = sinfo$path,
            backing = sinfo$backing,
            size = sinfo$size,
            readonly = sinfo$readonly,
            class_info = class_info,
            segment_info = sinfo
        ))
    }

    stop("shared_info() expects a shard_shared object or shard shared vector.",
         call. = FALSE)
}

#' Print a Shared Object
#'
#' @param x A \code{shard_shared} object.
#' @param ... Ignored.
#' @return The input \code{x}, invisibly.
#' @export
#' @examples
#' \donttest{
#' shared <- share(1:10)
#' print(shared)
#' close(shared)
#' }
print.shard_shared <- function(x, ...) {
    cat("<shard_shared>\n")
    cat("  Path:", x$path, "\n")
    cat("  Size:", format(x$size, big.mark = ","), "bytes\n")
    cat("  Backing:", x$backing, "\n")
    cat("  Read-only:", x$readonly, "\n")

    info <- x$class_info
    if (info$type == "matrix") {
        cat("  Original: matrix [", info$dim[1], " x ", info$dim[2], "], ",
            info$mode, "\n", sep = "")
    } else if (info$type == "array") {
        cat("  Original: array [", paste(info$dim, collapse = " x "), "], ",
            info$mode, "\n", sep = "")
    } else if (info$type == "data.frame") {
        cat("  Original: data.frame [", info$nrow, " x ", info$ncol, "]\n", sep = "")
    } else if (info$type == "list") {
        cat("  Original: list [", info$length, " elements]\n", sep = "")
    } else {
        cat("  Original: ", info$mode, " vector [", info$length, "]\n", sep = "")
    }

    invisible(x)
}

#' Print a Deep-Shared Object
#'
#' @param x A \code{shard_deep_shared} object.
#' @param ... Ignored.
#' @return The input \code{x}, invisibly.
#' @export
#' @examples
#' \donttest{
#' lst <- list(a = 1:10, b = 11:20)
#' shared <- share(lst, deep = TRUE, min_bytes = 1)
#' print(shared)
#' close(shared)
#' }
print.shard_deep_shared <- function(x, ...) {
    cat("<shard_deep_shared>\n")
    cat("  Type:", x$class_info$type, "\n")
    cat("  Backing:", x$backing, "\n")
    cat("  Read-only:", x$readonly, "\n")
    cat("\n")
    cat("  Summary:\n")
    cat("    Shared segments:", x$summary$shared_count, "\n")
    cat("    Aliased references:", x$summary$alias_count, "\n")
    if (x$summary$cycle_count > 0) {
        cat("    Cyclic references (skipped):", x$summary$cycle_count, "\n")
    }
    cat("    Kept in-place:", x$summary$kept_count, "\n")
    cat("    Total shared bytes:",
        format(x$summary$total_shared_bytes, big.mark = ","), "\n")

    invisible(x)
}

#' Open an Existing Shared Object by Path
#'
#' Opens a shared object that was created by another process. This is useful
#' for workers that need to attach to shared data without having the original
#' \code{shard_shared} object.
#'
#' @param path Path to the shared segment.
#' @param backing Backing type: "mmap" or "shm".
#' @param size Size of the segment in bytes. If NULL, attempts to detect.
#'
#' @return A \code{shard_shared} object attached to the existing segment.
#' @export
#' @examples
#' \donttest{
#' shared <- share(1:50)
#' info <- shared_info(shared)
#' reopened <- share_open(info$path, backing = "mmap")
#' close(reopened)
#' close(shared)
#' }
share_open <- function(path, backing = c("mmap", "shm"), size = NULL) {
    backing <- match.arg(backing)

    seg <- segment_open(path, backing = backing, readonly = TRUE)
    info <- segment_info(seg)

    if (is.null(size)) {
        size <- info$size
    }

    structure(
        list(
            segment = seg,
            path = info$path,
            backing = info$backing,
            size = size,
            readonly = TRUE,
            class_info = list(type = "unknown")
        ),
        class = "shard_shared"
    )
}

# Tests for copy-on-write enforcement on shared vectors.
# The COW policy is controlled via the readonly parameter of share():
#   readonly=TRUE -> cow='deny' (mutations blocked)
#   readonly=FALSE -> cow='allow' (mutations permitted)

test_that("cow='deny' (readonly=TRUE) blocks [<- assignment", {
  x <- share(1:10, readonly = TRUE)

  expect_error(x[1] <- 99L, "cow='deny'")
  expect_error(x[1:3] <- c(1L, 2L, 3L), "cow='deny'")
})

test_that("cow='deny' (readonly=TRUE) blocks [[<- assignment on vectors", {
  # Note: [[<- on vectors (single element access)
  x <- share(1:10, readonly = TRUE)

  expect_error(x[[1]] <- 99L, "cow='deny'")
})

test_that("cow='deny' (readonly=TRUE) blocks names<- assignment", {
  x <- share(c(a = 1, b = 2, c = 3), readonly = TRUE)

  expect_error(names(x) <- c("x", "y", "z"), "cow='deny'")
})

test_that("cow='deny' (readonly=TRUE) blocks dim<- assignment", {
  x <- share(1:12, readonly = TRUE)

  expect_error(dim(x) <- c(3, 4), "cow='deny'")
})

test_that("cow='deny' (readonly=TRUE) blocks dimnames<- assignment", {
  M <- share(matrix(1:6, nrow = 2), readonly = TRUE)

  expect_error(dimnames(M) <- list(c("a", "b"), c("x", "y", "z")), "cow='deny'")
})

# Note: attr<- and attributes<- methods are defined but R's primitive dispatch
# behavior may not always call them. The COW enforcement for these is best-effort.

test_that("cow='allow' (readonly=FALSE) permits all modifications", {
  x <- share(1:10, readonly = FALSE)

  # These should all succeed
  x[1] <- 99L
  expect_equal(as.integer(x[1]), 99L)

  names(x) <- letters[1:10]
  expect_equal(names(x)[1], "a")
})

test_that("default readonly=TRUE blocks modifications", {
  # Default is readonly=TRUE, which sets cow='deny'
  x <- share(1:10)

  expect_error(x[1] <- 99L, "cow='deny'")
})

test_that("print.shard_shared_vector works correctly", {
  x <- share(1:5)

  out <- capture.output(print(x))
  expect_true(length(out) > 0)
})

test_that("print.shard_shared_vector preserves underlying class", {
  # Date vector
  d <- share(as.Date(c("2024-01-01", "2024-01-02")))

  out <- capture.output(print(d))
  # Should print as dates, not as raw integers
  expect_true(any(grepl("2024", out)))
})

test_that(".shard_cow_policy extracts policy from attribute", {
  x <- share(1:5, readonly = TRUE)
  expect_equal(shard:::.shard_cow_policy(x), "deny")

  y <- share(1:5, readonly = FALSE)
  expect_equal(shard:::.shard_cow_policy(y), "allow")
})

test_that(".shard_cow_policy returns 'allow' for missing/invalid policy", {
  x <- 1:5  # Not shared, no policy
  expect_equal(shard:::.shard_cow_policy(x), "allow")

  # Shared but with NULL policy attr
  y <- share(1:5, readonly = FALSE)
  attr(y, "shard_cow") <- NULL
  expect_equal(shard:::.shard_cow_policy(y), "allow")
})

test_that("cow replacement methods preserve class", {
  x <- share(1:10, readonly = FALSE)
  original_class <- class(x)

  x[1] <- 99L
  expect_equal(class(x), original_class)

  names(x) <- letters[1:10]
  expect_equal(class(x), original_class)
})

test_that("cow='deny' works with matrices", {
  M <- share(matrix(1:12, nrow = 3), readonly = TRUE)

  expect_error(M[1, 1] <- 99L, "cow='deny'")
  expect_error(M[, 1] <- c(1L, 2L, 3L), "cow='deny'")
})

test_that("cow replacement preserves shard_shared_vector class after mutation", {
  x <- share(1:10, readonly = FALSE)

  # After mutation, class should still have shard_shared_vector
  x[1] <- 99L
  expect_true("shard_shared_vector" %in% class(x))

  # dim<- assignment
  y <- share(1:12, readonly = FALSE)
  dim(y) <- c(3, 4)
  expect_true("shard_shared_vector" %in% class(y))
  expect_equal(dim(y), c(3, 4))
})

# Tests for deep sharing with memoization, alias preservation, and cycle detection

test_that("deep share preserves aliases - same object creates one segment", {
    # Create a large matrix and reference it twice in a list
    big_mat <- matrix(rnorm(10000), nrow = 100)  # ~80KB

    lst <- list(a = big_mat, b = big_mat)

    # Share with deep=TRUE and low threshold to trigger sharing
    shared <- share(lst, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    # Should have exactly 1 shared segment (the matrix)
    # and 1 alias (the second reference to the same matrix)
    expect_equal(shared$summary$shared_count, 1)
    expect_equal(shared$summary$alias_count, 1)

    # Fetch and verify data is correct
    recovered <- fetch(shared)
    expect_identical(recovered$a, big_mat)
    expect_identical(recovered$b, big_mat)

    # Both should be the same object reference in the original
    # (after fetch they may be copies, but should have identical values)
    expect_identical(recovered$a, recovered$b)

    close(shared)
})

test_that("deep share handles lists with shareable and non-shareable components", {
    big_vec <- rnorm(10000)  # Large, shareable
    small_vec <- 1:10        # Small, kept
    char_vec <- letters      # Character, not shareable as atomic

    lst <- list(big = big_vec, small = small_vec, chars = char_vec)

    shared <- share(lst, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    # Only big_vec should be shared
    expect_equal(shared$summary$shared_count, 1)
    expect_equal(shared$summary$kept_count, 2)  # small + chars

    recovered <- fetch(shared)
    expect_equal(recovered$big, big_vec)
    expect_identical(recovered$small, small_vec)
    expect_identical(recovered$chars, char_vec)

    close(shared)
})

test_that("deep share detects cycles with environments and errors by default", {
    # R lists copy on assignment, so we need environments for true cycles
    env <- new.env()
    env$data <- 1:10
    env$self <- env  # True self-reference in environment
    expect_error(
        share(env, deep = TRUE, min_bytes = 100, cycle = "error"),
        "Cycle detected"
    )
})

test_that("deep share with cycle='skip' - list self-assignment creates copy not cycle", {
    # In R, lst$self <- lst creates a COPY, not a true cycle
    # So this actually creates an alias scenario, not a cycle
    lst <- list(a = 1:10)
    lst$self <- lst  # This creates a copy of lst at this point

    # This should work because there's no actual cycle
    shared <- share(lst, deep = TRUE, min_bytes = 100, cycle = "skip")

    expect_s3_class(shared, "shard_deep_shared")

    # No cycles should be detected (R lists don't create true cycles)
    expect_equal(shared$summary$cycle_count, 0)

    recovered <- fetch(shared)
    expect_identical(recovered$a, 1:10)
    # self contains the original structure (not a true cycle)
    expect_true(is.list(recovered$self))
    expect_identical(recovered$self$a, 1:10)

    close(shared)
})

test_that("deep share works with data.frames", {
    # Create a data.frame with large numeric columns
    df <- data.frame(
        x = rnorm(10000),
        y = rnorm(10000),
        z = letters[sample(26, 10000, replace = TRUE)]
    )

    shared <- share(df, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    # x and y should be shared (numeric, large enough)
    # z should be kept (character)
    expect_equal(shared$summary$shared_count, 2)

    recovered <- fetch(shared)
    expect_equal(recovered, df)
    expect_s3_class(recovered, "data.frame")

    close(shared)
})

test_that("deep share tracks total bytes correctly", {
    big_vec <- rnorm(10000)  # ~80KB (10000 * 8 bytes)

    lst <- list(a = big_vec)

    shared <- share(lst, deep = TRUE, min_bytes = 1000)

    # Total shared bytes should be approximately the size of big_vec
    expect_true(shared$summary$total_shared_bytes > 70000)
    expect_true(shared$summary$total_shared_bytes < 100000)

    close(shared)
})

test_that("deep share with aliased data.frame column", {
    # Same vector used as two columns
    shared_col <- rnorm(10000)
    df <- data.frame(a = shared_col, b = shared_col)

    shared <- share(df, deep = TRUE, min_bytes = 1000)

    # Should detect alias
    expect_equal(shared$summary$shared_count, 1)
    expect_equal(shared$summary$alias_count, 1)

    recovered <- fetch(shared)
    expect_equal(recovered$a, shared_col)
    expect_equal(recovered$b, shared_col)

    close(shared)
})

test_that("fetch and close work correctly for deep shared", {
    lst <- list(x = rnorm(10000))

    shared <- share(lst, deep = TRUE, min_bytes = 1000)

    # Multiple fetches should work
    r1 <- fetch(shared)
    r2 <- fetch(shared)
    expect_identical(r1, r2)

    # Close should not error
    expect_silent(close(shared))
})

test_that("print method works for deep shared", {
    lst <- list(x = rnorm(10000))

    shared <- share(lst, deep = TRUE, min_bytes = 1000)

    expect_output(print(shared), "shard_deep_shared")
    expect_output(print(shared), "Shared segments:")

    close(shared)
})

test_that("is_shared recognizes deep shared objects", {
    lst <- list(x = rnorm(10000))
    shared <- share(lst, deep = TRUE, min_bytes = 1000)

    expect_true(is_shared(shared))

    close(shared)
})

test_that("deep share respects min_bytes threshold", {
    small_vec <- 1:100  # Small
    medium_vec <- rnorm(1000)  # Medium
    large_vec <- rnorm(10000)  # Large

    # With high threshold, only large should be shared
    shared_high <- share(
        list(small = small_vec, medium = medium_vec, large = large_vec),
        deep = TRUE,
        min_bytes = 50000  # ~50KB threshold
    )
    expect_equal(shared_high$summary$shared_count, 1)
    close(shared_high)

    # With low threshold, medium and large should be shared
    shared_low <- share(
        list(small = small_vec, medium = medium_vec, large = large_vec),
        deep = TRUE,
        min_bytes = 5000  # ~5KB threshold
    )
    expect_equal(shared_low$summary$shared_count, 2)
    close(shared_low)
})

test_that("deep share preserves list names", {
    lst <- list(alpha = rnorm(10000), beta = rnorm(5000), gamma = 1:10)

    shared <- share(lst, deep = TRUE, min_bytes = 1000)
    recovered <- fetch(shared)

    expect_identical(names(recovered), names(lst))

    close(shared)
})

test_that("deep share preserves nested list structure", {
    nested <- list(
        level1 = list(
            a = rnorm(10000),
            b = 1:10
        ),
        top = rnorm(10000)
    )

    shared <- share(nested, deep = TRUE, min_bytes = 1000)
    recovered <- fetch(shared)

    expect_identical(names(recovered), c("level1", "top"))
    expect_identical(names(recovered$level1), c("a", "b"))
    expect_equal(recovered$level1$a, nested$level1$a)
    expect_identical(recovered$level1$b, nested$level1$b)
    expect_equal(recovered$top, nested$top)

    close(shared)
})

test_that("non-deep share ignores deep parameters", {
    x <- 1:100
    shared <- share(x, deep = FALSE)

    expect_true(is_shared_vector(shared) || inherits(shared, "shard_shared"))
    expect_false(inherits(shared, "shard_deep_shared"))

    close(shared)
})

# Tests for share() - zero-copy shared objects

test_that("share returns an ALTREP-backed shared vector for atomic inputs", {
    x <- 1:100
    shared <- share(x)

    expect_true(is_shared_vector(shared))
    expect_s3_class(shared, "shard_shared_vector")

    info <- shared_info(shared)
    expect_true(info$size > 0)
    expect_true(!is.null(info$path))
})

test_that("share returns a shard_shared handle for non-shareable types", {
    x <- letters
    shared <- share(x)
    expect_s3_class(shared, "shard_shared")
    expect_true(shared$size > 0)
    expect_true(!is.null(shared$path))
    close(shared)
})

test_that("share and local round-trip preserves data", {
    # Integer vector
    x <- 1:100
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)

    # Numeric vector
    x <- rnorm(50)
    shared <- share(x)
    recovered <- materialize(shared)
    expect_equal(x, recovered)

    # Character vector
    x <- letters
    shared <- share(x)
    recovered <- fetch(shared)
    expect_identical(x, recovered)
    close(shared)

    # Logical vector
    x <- c(TRUE, FALSE, TRUE, NA)
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)
})

test_that("share preserves matrix structure", {
    mat <- matrix(1:20, nrow = 4, ncol = 5)
    shared <- share(mat)

    info <- shared_info(shared)
    expect_equal(info$class_info$type, "matrix")
    expect_equal(info$class_info$dim, c(4L, 5L))

    recovered <- materialize(shared)
    expect_identical(mat, recovered)
})

test_that("share preserves array structure", {
    arr <- array(1:24, dim = c(2, 3, 4))
    shared <- share(arr)

    info <- shared_info(shared)
    expect_equal(info$class_info$type, "array")
    expect_equal(info$class_info$dim, c(2L, 3L, 4L))

    recovered <- materialize(shared)
    expect_identical(arr, recovered)
})

test_that("share preserves data.frame structure", {
    df <- data.frame(
        a = 1:5,
        b = letters[1:5],
        c = c(TRUE, FALSE, TRUE, FALSE, TRUE),
        stringsAsFactors = FALSE
    )
    shared <- share(df)

    expect_equal(shared$class_info$type, "data.frame")
    expect_equal(shared$class_info$nrow, 5)
    expect_equal(shared$class_info$ncol, 3)

    recovered <- fetch(shared)
    expect_identical(df, recovered)

    close(shared)
})

test_that("share preserves list structure", {
    lst <- list(
        a = 1:5,
        b = "hello",
        c = list(nested = TRUE)
    )
    shared <- share(lst)

    expect_equal(shared$class_info$type, "list")
    expect_equal(shared$class_info$length, 3)

    recovered <- fetch(shared)
    expect_identical(lst, recovered)

    close(shared)
})

test_that("share creates read-only segment by default", {
    x <- 1:10
    shared <- share(x)

    info <- shared_info(shared)
    expect_true(info$readonly)
})

test_that("share with readonly=FALSE creates writable segment", {
    x <- 1:10
    shared <- share(x, readonly = FALSE)

    info <- shared_info(shared)
    expect_false(info$readonly)
})

test_that("materialize is equivalent to fetch", {
    x <- letters
    shared <- share(x)

    via_fetch <- fetch(shared)
    via_materialize <- materialize(shared)

    expect_identical(via_fetch, via_materialize)

    close(shared)
})

test_that("materialize.default returns input unchanged", {
    x <- 1:10
    expect_identical(materialize(x), x)

    y <- list(a = 1, b = 2)
    expect_identical(materialize(y), y)
})

test_that("is_shared correctly identifies shared objects", {
    x <- 1:10
    shared <- share(x)

    expect_true(is_shared(shared))
    expect_false(is_shared(x))
    expect_false(is_shared(NULL))
    expect_false(is_shared(list()))
})

test_that("shared_info returns complete information", {
    x <- matrix(1:20, nrow = 4)
    shared <- share(x)

    info <- shared_info(shared)

    expect_true(is.list(info))
    expect_true(!is.null(info$path))
    expect_true(!is.null(info$backing))
    expect_true(!is.null(info$size))
    expect_true(!is.null(info$readonly))
    expect_true(!is.null(info$class_info))
    expect_true(!is.null(info$segment_info))
})

test_that("print.shard_shared works", {
    x <- letters
    shared <- share(x)

    expect_output(print(shared), "shard_shared")
    expect_output(print(shared), "Size:")

    close(shared)
})

test_that("share handles large objects", {
    # 1 million elements
    x <- rnorm(1e6)
    shared <- share(x)

    info <- shared_info(shared)
    expect_true(info$size >= 8e6)  # At least 8 bytes per double

    recovered <- materialize(shared)
    expect_equal(x, recovered)
})

test_that("share_open can reopen existing shared segment", {
    x <- letters
    shared1 <- share(x)

    path <- shared1$path
    backing <- shared1$backing
    size <- shared1$size

    skip_if(is.null(path), "Path is NULL")

    # Open the same segment
    shared2 <- share_open(path, backing = backing, size = size)

    # Read from reopened segment
    recovered <- fetch(shared2)
    expect_identical(x, recovered)

    # Close in order (shared2 doesn't unlink)
    close(shared2)
    close(shared1)
})

test_that("share respects backing type", {
    x <- letters

    # Test mmap backing
    shared_mmap <- share(x, backing = "mmap")
    expect_equal(shared_mmap$backing, "mmap")
    close(shared_mmap)

    # Test auto backing
    shared_auto <- share(x, backing = "auto")
    expect_s3_class(shared_auto, "shard_shared")
    close(shared_auto)
})

test_that("close releases resources", {
    x <- letters
    shared <- share(x)
    path <- shared$path

    # Close should work without error
    expect_silent(close(shared))

    # Segment should be gone (file-backed)
    if (!is.null(path) && shared$backing == "mmap") {
        expect_false(file.exists(path))
    }
})

test_that("share handles empty vectors", {
    x <- integer(0)
    shared <- share(x)

    recovered <- materialize(shared)
    expect_identical(x, recovered)
})

test_that("share handles zero-length vectors of all types", {
    # Double
    x <- double(0)
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)

    # Logical
    x <- logical(0)
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)

    # Character
    x <- character(0)
    shared <- share(x)
    recovered <- fetch(shared)
    expect_identical(x, recovered)
    close(shared)

    # Raw
    x <- raw(0)
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)
})

test_that("share handles single-element vectors", {
    # Single integer
    x <- 1L
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)

    # Single double
    x <- 3.14
    shared <- share(x)
    recovered <- materialize(shared)
    expect_equal(x, recovered)

    # Single character
    x <- "hello"
    shared <- share(x)
    recovered <- fetch(shared)
    expect_identical(x, recovered)
    close(shared)

    # Single logical
    x <- TRUE
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)

    # Single NA
    x <- NA
    shared <- share(x)
    recovered <- materialize(shared)
    expect_identical(x, recovered)
})

test_that("share handles single-element matrix and array", {
    # 1x1 matrix
    mat <- matrix(42, nrow = 1, ncol = 1)
    shared <- share(mat)
    recovered <- materialize(shared)
    expect_identical(mat, recovered)

    # 1x1x1 array
    arr <- array(42, dim = c(1, 1, 1))
    shared <- share(arr)
    recovered <- materialize(shared)
    expect_identical(arr, recovered)
})

test_that("share handles empty data.frame", {
    # Zero-row data.frame
    df <- data.frame(a = integer(0), b = character(0), stringsAsFactors = FALSE)
    shared <- share(df)
    recovered <- fetch(shared)
    expect_identical(df, recovered)
    close(shared)

    # Zero-column data.frame
    df <- data.frame()
    shared <- share(df)
    recovered <- fetch(shared)
    expect_identical(df, recovered)
    close(shared)
})

test_that("share handles empty list", {
    lst <- list()
    shared <- share(lst)
    recovered <- fetch(shared)
    expect_identical(lst, recovered)
    close(shared)
})

test_that("share handles NULL elements in lists", {
    lst <- list(a = 1, b = NULL, c = 3)
    shared <- share(lst)

    recovered <- fetch(shared)
    expect_identical(lst, recovered)

    close(shared)
})

test_that("share handles complex objects", {
    # Factor
    f <- factor(c("a", "b", "a", "c"))
    shared <- share(f)
    recovered <- materialize(shared)
    expect_identical(f, recovered)

    # POSIXct
    dt <- as.POSIXct("2024-01-01 12:00:00")
    shared <- share(dt)
    recovered <- materialize(shared)
    expect_equal(dt, recovered)

    # Formula (environment-sensitive)
    # Skip this as formulas capture environments
})

# Tests for input validation of non-serializable objects

test_that("share rejects functions with clear error message", {
    fn <- function(x) x + 1

    expect_error(
        share(fn),
        "Cannot share functions"
    )
    expect_error(
        share(fn),
        "closures"
    )
})
test_that("share rejects functions nested in lists", {
    lst <- list(a = 1, fn = function(x) x)

    expect_error(
        share(lst),
        "Cannot share functions"
    )
    expect_error(
        share(lst),
        "x\\$fn"  # Shows the path to the problem
    )
})

test_that("share rejects external pointers with clear error message", {
    # Create a simple external pointer for testing
    # We can use file() to get a connection which has an external pointer
    skip_if_not(exists(".Call"))

    # Actually creating an external pointer requires C code, so we'll test
    # the error message pattern by checking a list containing one
    # For now, verify the error message format for functions (same pattern)
    expect_error(
        share(function(x) x),
        "Extract the data you need"
    )
})

test_that("share rejects environments with external pointers", {
    # Create an environment with a function (which will trigger the validation)
    env <- new.env()
    env$fn <- function(x) x

    expect_error(
        share(env),
        "Cannot share functions"
    )
    expect_error(
        share(env),
        "x\\$fn"
    )
})

test_that("share accepts environments without problematic content", {
    env <- new.env()
    env$a <- 1:10
    env$b <- "hello"

    shared <- share(env)
    recovered <- fetch(shared)

    expect_equal(recovered$a, 1:10)
    expect_equal(recovered$b, "hello")

    close(shared)
})

test_that("share error message shows path to nested problems", {
    # Deeply nested function
    lst <- list(
        level1 = list(
            level2 = list(
                fn = function(x) x
            )
        )
    )

    expect_error(
        share(lst),
        "x\\$level1\\$level2\\$fn"
    )
})

test_that("share error message shows index for unnamed list elements", {
    lst <- list(1, 2, function(x) x)

    expect_error(
        share(lst),
        "x\\[\\[3\\]\\]"
    )
})

# Tests for deep sharing of S4 objects

# Define S4 classes for testing
# Use "ANY" type for slots that will hold reconstructed shared data
setClass("TestS4DeepMatrix",
    slots = c(
        data = "ANY",
        name = "character"
    )
)

setClass("TestS4DeepList",
    slots = c(
        items = "ANY",
        label = "character"
    )
)

setClass("TestS4DeepEnv",
    slots = c(
        data = "ANY",
        env = "environment"
    )
)

setClass("TestS4DeepNested",
    slots = c(
        inner = "ANY",
        value = "numeric"
    )
)

test_that("share with deep=TRUE processes S4 object slots", {
    # Create a large matrix that will be shared
    big_mat <- matrix(rnorm(1e7), nrow = 1000)  # ~80MB
    obj <- new("TestS4DeepMatrix", data = big_mat, name = "test")

    # Share with deep=TRUE
    shared <- share(obj, deep = TRUE, min_bytes = 1e6)

    # Should return a shard_deep_shared object
    expect_s3_class(shared, "shard_deep_shared")

    # Fetch should reconstruct the original
    recovered <- fetch(shared)
    expect_true(isS4(recovered))
    expect_s4_class(recovered, "TestS4DeepMatrix")
    expect_equal(slot(recovered, "data"), big_mat)
    expect_equal(slot(recovered, "name"), "test")

    close(shared)
})

test_that("share with deep=TRUE recursively shares S4 list slots", {
    big_vec <- rnorm(1e7)  # ~80MB
    obj <- new("TestS4DeepList",
        items = list(big = big_vec, small = 1:10),
        label = "test"
    )

    shared <- share(obj, deep = TRUE, min_bytes = 1e6)
    expect_s3_class(shared, "shard_deep_shared")

    # Summary should show shared nodes
    expect_true(shared$summary$shared_count > 0)

    # Fetch should reconstruct
    recovered <- fetch(shared)
    expect_s4_class(recovered, "TestS4DeepList")
    expect_equal(slot(recovered, "items")$big, big_vec)
    expect_equal(slot(recovered, "items")$small, 1:10)

    close(shared)
})

test_that("share with deep=TRUE and mode='balanced' skips S4 environment slots", {
    big_mat <- matrix(rnorm(1e6), nrow = 100)
    env <- new.env()
    env$foo <- "bar"
    obj <- new("TestS4DeepEnv", data = big_mat, env = env)

    # Should not error with mode='balanced' (default)
    shared <- share(obj, deep = TRUE, min_bytes = 1e5, mode = "balanced")
    expect_s3_class(shared, "shard_deep_shared")

    # Fetch should preserve the environment
    recovered <- fetch(shared)
    expect_true(is.environment(slot(recovered, "env")))

    close(shared)
})

test_that("share with deep=TRUE and mode='strict' errors on S4 environment slots", {
    big_mat <- matrix(rnorm(1e6), nrow = 100)
    env <- new.env()
    obj <- new("TestS4DeepEnv", data = big_mat, env = env)

    expect_error(
        share(obj, deep = TRUE, min_bytes = 1e3, mode = "strict"),
        "contains environment"
    )
})

test_that("share with deep=TRUE uses @slot path notation for S4", {
    setClass("TestS4WithFunc",
        slots = c(
            fn = "function"
        )
    )

    obj <- new("TestS4WithFunc", fn = function(x) x)

    # When validating S4 with function slot, should show @slot notation
    expect_error(
        share(obj),
        "@fn"
    )

    removeClass("TestS4WithFunc")
})

test_that("share with deep=TRUE handles nested S4 objects", {
    inner_mat <- matrix(rnorm(1e7), nrow = 1000)
    inner <- new("TestS4DeepMatrix", data = inner_mat, name = "inner")
    outer <- new("TestS4DeepNested", inner = inner, value = 42)

    shared <- share(outer, deep = TRUE, min_bytes = 1e6)
    expect_s3_class(shared, "shard_deep_shared")

    recovered <- fetch(shared)
    expect_s4_class(recovered, "TestS4DeepNested")
    expect_s4_class(slot(recovered, "inner"), "TestS4DeepMatrix")
    expect_equal(slot(slot(recovered, "inner"), "data"), inner_mat)
    expect_equal(slot(recovered, "value"), 42)

    close(shared)
})

test_that("share with deep=TRUE preserves S4 class after fetch", {
    big_mat <- matrix(rnorm(1e7), nrow = 1000)
    obj <- new("TestS4DeepMatrix", data = big_mat, name = "preserved")

    shared <- share(obj, deep = TRUE, min_bytes = 1e6)
    recovered <- fetch(shared)

    expect_identical(class(recovered), class(obj))
    expect_s4_class(recovered, "TestS4DeepMatrix")

    close(shared)
})

# Clean up test classes
removeClass("TestS4DeepMatrix")
removeClass("TestS4DeepList")
removeClass("TestS4DeepEnv")
removeClass("TestS4DeepNested")

# Tests for deep sharing hook system

.shard_test_register_hook <- function(class, fun) {
    nm <- paste0("shard_share_hook.", class)
    # S3 method lookup for UseMethod() will find methods in the global env,
    # and we avoid mutating the (locked) package namespace.
    assign(nm, fun, envir = globalenv())
    nm
}

.shard_test_unregister_hook <- function(name) {
    if (exists(name, envir = globalenv(), inherits = FALSE)) {
        rm(list = name, envir = globalenv())
    }
    invisible(NULL)
}

test_that("default hook returns empty list (no-op)", {
    x <- list(a = 1:10)
    ctx <- list(
        path = "<root>",
        class = class(x),
        mode = "balanced",
        min_bytes = 1000,
        types = c("double", "integer"),
        deep = TRUE
    )

    result <- shard_share_hook(x, ctx)

    expect_type(result, "list")
    expect_length(result, 0)
})

test_that("hook with skip_paths prevents traversal", {
    # Define a custom class with a hook
    setClass("SkipPathTest", slots = c(data = "numeric", cache = "list"))

    # Create hook that skips the cache
    hook_name <- .shard_test_register_hook("SkipPathTest", function(x, ctx) {
        list(skip_paths = paste0(ctx$path, "@cache"))
    })

    obj <- new("SkipPathTest",
               data = rnorm(10000),  # Large enough to share
               cache = list(temp = rnorm(10000)))

    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    # The data slot should be shared, but cache contents should be kept
    # (cache path was skipped, so its children weren't traversed deeply)
    expect_true(shared$summary$shared_count >= 1)

    recovered <- fetch(shared)
    expect_equal(recovered@data, obj@data)
    expect_equal(recovered@cache, obj@cache)

    close(shared)

    # Clean up
    .shard_test_unregister_hook(hook_name)
    removeClass("SkipPathTest")
})

test_that("hook with skip_slots prevents slot traversal for S4 objects", {
    # Define S4 class
    setClass("CacheModel", slots = c(
        coefficients = "numeric",
        cache = "list"
    ))

    # Hook that skips cache slot
    hook_name <- .shard_test_register_hook("CacheModel", function(x, ctx) {
        list(skip_slots = "cache")
    })

    obj <- new("CacheModel",
               coefficients = rnorm(10000),
               cache = list(fitted = rnorm(10000), residuals = rnorm(10000)))

    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    # Only coefficients should be shared (cache slot was skipped entirely)
    expect_equal(shared$summary$shared_count, 1)

    recovered <- fetch(shared)
    expect_equal(recovered@coefficients, obj@coefficients)
    # Cache should still be there but wasn't traversed/shared
    expect_equal(recovered@cache, obj@cache)

    close(shared)

    .shard_test_unregister_hook(hook_name)
    removeClass("CacheModel")
})

test_that("hook with force_share_paths shares small objects", {
    # Define class that forces sharing of small data
    setClass("ForceShareTest", slots = c(
        important_small = "numeric",
        unimportant_large = "numeric"
    ))

    hook_name <- .shard_test_register_hook("ForceShareTest", function(x, ctx) {
        list(force_share_paths = paste0(ctx$path, "@important_small"))
    })

    obj <- new("ForceShareTest",
               important_small = 1:10,  # Small - normally wouldn't be shared
               unimportant_large = rnorm(10000))  # Large - would be shared

    # Use high threshold so only forced paths are shared
    shared <- share(obj, deep = TRUE, min_bytes = 100000)

    expect_s3_class(shared, "shard_deep_shared")

    # important_small was forced, unimportant_large is below threshold
    # So only 1 segment should be shared (the forced one)
    expect_equal(shared$summary$shared_count, 1)

    recovered <- fetch(shared)
    expect_equal(recovered@important_small, obj@important_small)
    expect_equal(recovered@unimportant_large, obj@unimportant_large)

    close(shared)

    .shard_test_unregister_hook(hook_name)
    removeClass("ForceShareTest")
})

test_that("hook with rewrite transforms object before sharing", {
    # Class that has lazy data needing materialization
    setClass("LazyData", slots = c(data = "ANY", materialized = "logical"))

    hook_name <- .shard_test_register_hook("LazyData", function(x, ctx) {
        list(
            rewrite = function(obj) {
                # Transform the object - e.g., materialize lazy data
                obj@materialized <- TRUE
                obj
            }
        )
    })

    obj <- new("LazyData", data = rnorm(10000), materialized = FALSE)

    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    recovered <- fetch(shared)
    # The rewrite function should have set materialized to TRUE
    expect_true(recovered@materialized)
    expect_equal(recovered@data, obj@data)

    close(shared)

    .shard_test_unregister_hook(hook_name)
    removeClass("LazyData")
})

test_that("hook errors in strict mode propagate", {
    setClass("ErrorHook", slots = c(data = "numeric"))

    hook_name <- .shard_test_register_hook("ErrorHook", function(x, ctx) {
        stop("Deliberate hook error")
    })

    obj <- new("ErrorHook", data = rnorm(10000))

    expect_error(
        share(obj, deep = TRUE, min_bytes = 1000, mode = "strict"),
        "Hook error in strict mode"
    )

    .shard_test_unregister_hook(hook_name)
    removeClass("ErrorHook")
})

test_that("hook errors in balanced mode continue and track error", {
    setClass("ErrorHookBalanced", slots = c(data = "numeric"))

    hook_name <- .shard_test_register_hook("ErrorHookBalanced", function(x, ctx) {
        stop("Deliberate hook error for balanced mode")
    })

    obj <- new("ErrorHookBalanced", data = rnorm(10000))

    # Should NOT error in balanced mode
    shared <- share(obj, deep = TRUE, min_bytes = 1000, mode = "balanced")

    expect_s3_class(shared, "shard_deep_shared")

    # Should have recorded the hook error
    expect_true(shared$summary$hook_error_count > 0)
    expect_true(length(shared$summary$hook_errors) > 0)

    # Data should still be shared
    recovered <- fetch(shared)
    expect_equal(recovered@data, obj@data)

    close(shared)

    .shard_test_unregister_hook(hook_name)
    removeClass("ErrorHookBalanced")
})

test_that("custom class hook dispatches correctly", {
    # Define a custom S3 class
    my_object <- structure(
        list(
            large_data = rnorm(10000),
            small_data = 1:10
        ),
        class = c("MyCustomClass", "list")
    )

    # Define hook for this class
    hook_name <- .shard_test_register_hook("MyCustomClass", function(x, ctx) {
        # Force share small_data even though it's tiny
        list(force_share_paths = paste0(ctx$path, "$small_data"))
    })

    # With very high threshold, normally nothing would be shared
    shared <- share(my_object, deep = TRUE, min_bytes = 1000000)

    # But hook forced small_data to be shared
    expect_equal(shared$summary$shared_count, 1)

    recovered <- fetch(shared)
    expect_equal(recovered$large_data, my_object$large_data)
    expect_equal(recovered$small_data, my_object$small_data)

    close(shared)

    .shard_test_unregister_hook(hook_name)
})

test_that("hook context contains correct information", {
    captured_ctx <- NULL

    setClass("ContextCapture", slots = c(data = "numeric"))

    hook_name <- .shard_test_register_hook("ContextCapture", function(x, ctx) {
        captured_ctx <<- ctx
        list()
    })

    obj <- new("ContextCapture", data = rnorm(100))

    shared <- share(obj, deep = TRUE, min_bytes = 1000, mode = "balanced")

    # Check context fields
    expect_equal(captured_ctx$path, "<root>")
    expect_true("ContextCapture" %in% captured_ctx$class)
    expect_equal(captured_ctx$mode, "balanced")
    expect_equal(captured_ctx$min_bytes, 1000)
    expect_true("double" %in% captured_ctx$types)
    expect_true(captured_ctx$deep)

    close(shared)

    .shard_test_unregister_hook(hook_name)
    removeClass("ContextCapture")
})

test_that("hook rewrite error in strict mode propagates", {
    setClass("RewriteError", slots = c(data = "numeric"))

    hook_name <- .shard_test_register_hook("RewriteError", function(x, ctx) {
        list(
            rewrite = function(obj) {
                stop("Rewrite failed!")
            }
        )
    })

    obj <- new("RewriteError", data = rnorm(100))

    expect_error(
        share(obj, deep = TRUE, min_bytes = 1000, mode = "strict"),
        "Rewrite function error"
    )

    .shard_test_unregister_hook(hook_name)
    removeClass("RewriteError")
})

test_that("hook rewrite error in balanced mode continues", {
    setClass("RewriteErrorBalanced", slots = c(data = "numeric"))

    hook_name <- .shard_test_register_hook("RewriteErrorBalanced", function(x, ctx) {
        list(
            rewrite = function(obj) {
                stop("Rewrite failed in balanced!")
            }
        )
    })

    obj <- new("RewriteErrorBalanced", data = rnorm(10000))

    # Should not error
    shared <- share(obj, deep = TRUE, min_bytes = 1000, mode = "balanced")

    expect_s3_class(shared, "shard_deep_shared")

    # Rewrite error should be tracked
    expect_true(shared$summary$hook_error_count > 0)

    # Original data should still be present
    recovered <- fetch(shared)
    expect_equal(recovered@data, obj@data)

    close(shared)

    .shard_test_unregister_hook(hook_name)
    removeClass("RewriteErrorBalanced")
})

test_that("S4 object with list slot shares list contents", {
    setClass("S4WithList", slots = c(
        matrix_data = "matrix",
        list_data = "list"
    ))

    obj <- new("S4WithList",
               matrix_data = matrix(rnorm(10000), 100, 100),
               list_data = list(a = rnorm(10000), b = rnorm(10000)))

    shared <- share(obj, deep = TRUE, min_bytes = 1000)

    expect_s3_class(shared, "shard_deep_shared")

    # Matrix and both list elements should be shared
    expect_equal(shared$summary$shared_count, 3)

    recovered <- fetch(shared)
    expect_equal(recovered@matrix_data, obj@matrix_data)
    expect_equal(recovered@list_data$a, obj@list_data$a)
    expect_equal(recovered@list_data$b, obj@list_data$b)

    close(shared)
    removeClass("S4WithList")
})

try({
  if (requireNamespace("withr", quietly = TRUE)) {
    withr::defer({
      try(pool_stop(), silent = TRUE)
      try(closeAllConnections(), silent = TRUE)
    }, testthat::teardown_env())
  }
}, silent = TRUE)

# Test helpers loaded before all test files

#' Skip test when R socket connections are nearly exhausted.
#'
#' Each PSOCK worker needs at least 2 socket connections.
#' On shared CI nodes the per-process limit (default 128) can be
#' reached when many tests create/destroy worker pools.
skip_if_conn_exhausted <- function(n_needed = 6L) {
  n_open <- nrow(showConnections(all = TRUE))
  max_conn <- 128L
  if (n_open > max_conn - n_needed) {
    testthat::skip(
      sprintf("Socket connections nearly exhausted (%d/%d in use)", n_open, max_conn)
    )
  }
}

# Worker-local scratch pool (R-level).
#
# This is a simple reusable allocator for native/R kernels that want to avoid
# repeated malloc/free churn. It is intentionally conservative: the pool holds
# at most one buffer per key and can request worker recycle if it grows beyond a
# configured limit.

.scratch_env <- new.env(parent = emptyenv())
.scratch_env$pool <- new.env(parent = emptyenv())
.scratch_env$hits <- 0L
.scratch_env$misses <- 0L
.scratch_env$bytes <- 0
.scratch_env$high_water <- 0
.scratch_env$max_bytes <- Inf
.scratch_env$needs_recycle <- FALSE

#' Configure scratch pool limits
#'
#' @param max_bytes Maximum scratch pool bytes allowed in a worker. If exceeded,
#'   the worker is flagged for recycle at the next safe point.
#' @return \code{NULL}, invisibly.
#' @export
#' @examples
#' cfg <- scratch_pool_config(max_bytes = 100 * 1024^2)
scratch_pool_config <- function(max_bytes = Inf) {
  if (is.character(max_bytes)) max_bytes <- parse_bytes(max_bytes)
  max_bytes <- as.double(max_bytes)
  if (!is.finite(max_bytes) || max_bytes <= 0) stop("max_bytes must be positive", call. = FALSE)
  .scratch_env$max_bytes <- max_bytes
  invisible(NULL)
}

scratch_reset_diagnostics <- function() {
  .scratch_env$hits <- 0L
  .scratch_env$misses <- 0L
  .scratch_env$bytes <- 0
  .scratch_env$high_water <- 0
  .scratch_env$needs_recycle <- FALSE
  # Keep the pool; reset is for counters.
  invisible(NULL)
}

#' Scratch pool diagnostics
#'
#' @return A list with counters and current pool bytes.
#' @export
#' @examples
#' scratch_diagnostics()
scratch_diagnostics <- function() {
  list(
    hits = .scratch_env$hits,
    misses = .scratch_env$misses,
    bytes = .scratch_env$bytes,
    high_water = .scratch_env$high_water,
    max_bytes = .scratch_env$max_bytes,
    needs_recycle = isTRUE(.scratch_env$needs_recycle)
  )
}

.scratch_update_bytes <- function() {
  keys <- ls(.scratch_env$pool, all.names = TRUE)
  total <- 0
  for (k in keys) {
    x <- get(k, envir = .scratch_env$pool, inherits = FALSE)
    total <- total + as.double(utils::object.size(x))
  }
  .scratch_env$bytes <- total
  if (total > (.scratch_env$high_water %||% 0)) .scratch_env$high_water <- total
  if (is.finite(.scratch_env$max_bytes) && total > .scratch_env$max_bytes) {
    .scratch_env$needs_recycle <- TRUE
  }
  invisible(NULL)
}

.scratch_get_double <- function(n, key) {
  n <- as.integer(n)
  if (is.na(n) || n < 0L) stop("n must be >= 0", call. = FALSE)
  key <- as.character(key)
  if (!nzchar(key)) stop("key must be non-empty", call. = FALSE)

  if (exists(key, envir = .scratch_env$pool, inherits = FALSE)) {
    buf <- get(key, envir = .scratch_env$pool, inherits = FALSE)
    if (is.double(buf) && length(buf) >= n) {
      .scratch_env$hits <- .scratch_env$hits + 1L
      return(buf)
    }
  }

  .scratch_env$misses <- .scratch_env$misses + 1L
  buf <- double(n)
  assign(key, buf, envir = .scratch_env$pool)
  .scratch_update_bytes()
  buf
}

#' Get a scratch matrix
#'
#' Allocates (or reuses) a double matrix in the worker scratch pool.
#'
#' @param nrow,ncol Dimensions.
#' @param key Optional key to control reuse. Defaults to a shape-derived key.
#' @return A double matrix of dimensions \code{nrow} by \code{ncol}.
#' @export
#' @examples
#' m <- scratch_matrix(10, 5)
#' dim(m)
scratch_matrix <- function(nrow, ncol, key = NULL) {
  nrow <- as.integer(nrow)
  ncol <- as.integer(ncol)
  if (is.na(nrow) || nrow < 0L) stop("nrow must be >= 0", call. = FALSE)
  if (is.na(ncol) || ncol < 0L) stop("ncol must be >= 0", call. = FALSE)
  if (is.null(key)) key <- paste0("mat_", nrow, "x", ncol)
  buf <- .scratch_get_double(nrow * ncol, key = key)
  dim(buf) <- c(nrow, ncol)
  buf
}

test_that("scratch_matrix reuses allocations and can request worker recycle", {
  skip_on_cran()

  # Local-process behavior: reuse counters.
  shard:::scratch_reset_diagnostics()
  scratch_pool_config(max_bytes = "10MB")

  x1 <- scratch_matrix(10, 10)
  x1[] <- 1
  x2 <- scratch_matrix(10, 10)
  expect_true(is.matrix(x2))

  sd <- scratch_diagnostics()
  expect_true((sd$misses %||% 0L) >= 1L)
  expect_true((sd$hits %||% 0L) >= 1L)
  expect_false(isTRUE(sd$needs_recycle))

  # Worker recycle signaling: exceed max_bytes in a worker, then ensure the
  # worker is recycled at the next safe point.
  pool_stop()
  pool_create(1, min_recycle_interval = 0)
  on.exit(pool_stop(), add = TRUE)

  shard:::scratch_reset_diagnostics()

  # Two shards: first triggers needs_recycle, second provides a safe-point to recycle.
  res <- shard_map(shards(2, block_size = 1), fun = function(s) {
    scratch_pool_config(max_bytes = "32KB")
    if (s$id == 1L) {
      # ~320KB
      tmp <- scratch_matrix(200, 200, key = "big")
      tmp[] <- 0
    }
    s$id
  }, workers = 1, diagnostics = TRUE)

  expect_true(succeeded(res))
  expect_true((res$diagnostics$scratch_stats$high_water %||% 0) > 0)

  # Worker should have been recycled once (between shard 1 and 2).
  p <- pool_get()
  expect_true((p$workers[[1]]$recycle_count %||% 0L) >= 1L)
})

test_that("scratch_pool_config validates max_bytes", {
  expect_error(scratch_pool_config(max_bytes = 0), "max_bytes must be positive")
  expect_error(scratch_pool_config(max_bytes = -1), "max_bytes must be positive")
})

test_that("scratch_pool_config accepts string bytes", {
  expect_no_error(scratch_pool_config(max_bytes = "1GB"))
  expect_no_error(scratch_pool_config(max_bytes = "512MB"))
})

test_that("scratch_diagnostics returns complete structure", {
  shard:::scratch_reset_diagnostics()

  sd <- scratch_diagnostics()

  expect_true(is.list(sd))
  expect_true("hits" %in% names(sd))
  expect_true("misses" %in% names(sd))
  expect_true("bytes" %in% names(sd))
  expect_true("high_water" %in% names(sd))
  expect_true("max_bytes" %in% names(sd))
  expect_true("needs_recycle" %in% names(sd))
})

test_that("scratch_matrix validates dimensions", {
  expect_error(scratch_matrix(-1, 10), "nrow must be >= 0")
  expect_error(scratch_matrix(10, -1), "ncol must be >= 0")
})

test_that("scratch_matrix uses custom key when provided", {
  shard:::scratch_reset_diagnostics()

  # First allocation with custom key
  x1 <- scratch_matrix(5, 5, key = "custom_key")
  expect_true(is.matrix(x1))
  expect_equal(dim(x1), c(5, 5))

  # Second allocation with same key should hit cache
  x2 <- scratch_matrix(5, 5, key = "custom_key")

  sd <- scratch_diagnostics()
  expect_true(sd$hits >= 1)
})

test_that("scratch_matrix creates double matrices", {
  x <- scratch_matrix(3, 4)

  expect_true(is.matrix(x))
  expect_equal(dim(x), c(3, 4))
  expect_true(is.double(x))
})

test_that(".scratch_get_double validates inputs", {
  expect_error(shard:::.scratch_get_double(-1, "key"), "n must be >= 0")
  expect_error(shard:::.scratch_get_double(10, ""), "key must be non-empty")
})

test_that("scratch_reset_diagnostics clears counters", {
  # Create some activity
  scratch_matrix(10, 10)

  # Reset
  shard:::scratch_reset_diagnostics()

  sd <- scratch_diagnostics()
  expect_equal(sd$hits, 0L)
  expect_equal(sd$misses, 0L)
  expect_equal(sd$bytes, 0)
  expect_equal(sd$high_water, 0)
  expect_false(sd$needs_recycle)
})

# Schema-driven columnar table outputs.
#
# v1.0 scope: fixed-row table buffers backed by shard buffers, with strict
# validation and disjoint row-range writes.

# Per-process table write diagnostics (aggregated into shard_map via dispatch
# deltas, similar to views/buffers).
.table_diag_env <- new.env(parent = emptyenv())
.table_diag_env$writes <- 0L
.table_diag_env$rows <- 0L
.table_diag_env$bytes <- 0

#' Table Diagnostics
#'
#' Per-process counters for table writes (number of table_write calls, rows,
#' and bytes written). shard_map uses deltas of these counters to produce
#' run-level diagnostics in copy_report().
#'
#' @return A list with `writes`, `rows`, and `bytes`.
#' @export
table_diagnostics <- function() {
  list(
    writes = .table_diag_env$writes,
    rows = .table_diag_env$rows,
    bytes = .table_diag_env$bytes
  )
}

table_reset_diagnostics <- function() {
  .table_diag_env$writes <- 0L
  .table_diag_env$rows <- 0L
  .table_diag_env$bytes <- 0
  invisible(NULL)
}

.coltype <- function(kind, ...) {
  structure(c(list(kind = kind), list(...)), class = "shard_coltype")
}

#' Column Types
#'
#' Type constructors for schema-driven table outputs.
#'
#' @return A `shard_coltype` object.
#' @name coltypes
NULL

#' @rdname coltypes
#' @export
int32 <- function() .coltype("int32")

#' @rdname coltypes
#' @export
float64 <- function() .coltype("float64")

#' @rdname coltypes
#' @export
bool <- function() .coltype("bool")

#' @rdname coltypes
#' @export
raw_col <- function() .coltype("raw")

#' @rdname coltypes
#' @export
string_col <- function() .coltype("string")

#' Categorical column type
#'
#' Stores factors as int32 codes plus shared levels metadata.
#'
#' @param levels Character vector of allowed levels.
#' @return A `shard_coltype` object.
#' @export
factor_col <- function(levels) {
  if (!is.character(levels) || length(levels) < 1L || anyNA(levels)) {
    stop("levels must be a non-empty character vector without NA", call. = FALSE)
  }
  .coltype("factor", levels = unique(levels))
}

.coltype_bytes <- function(ct) {
  switch(ct$kind,
    "int32" = 4,
    "float64" = 8,
    "bool" = 1,   # logical payload size estimate for table bytes (not storage)
    "raw" = 1,
    "factor" = 4,
    "string" = NA_real_,
    NA_real_
  )
}

.coltype_buffer_type <- function(ct) {
  switch(ct$kind,
    "int32" = "integer",
    "float64" = "double",
    "bool" = "logical",
    "raw" = "raw",
    "factor" = "integer",
    "string" = stop("string columns are not supported in fixed table_buffer() (use table_sink())", call. = FALSE),
    stop("Unsupported column type: ", ct$kind, call. = FALSE)
  )
}

#' Define a table schema
#'
#' A schema is a named set of columns with explicit types. It is used to
#' allocate table buffers and validate writes.
#'
#' @param ... Named columns with type specs (e.g., \code{int32()}, \code{float64()}).
#' @return A \code{shard_schema} object.
#' @export
#' @examples
#' s <- schema(x = float64(), y = int32(), label = string_col())
#' s
schema <- function(...) {
  cols <- list(...)
  if (length(cols) == 0) stop("schema() requires at least one column", call. = FALSE)
  if (is.null(names(cols)) || any(names(cols) == "")) {
    stop("schema() requires named columns", call. = FALSE)
  }
  if (anyDuplicated(names(cols))) stop("schema() column names must be unique", call. = FALSE)

  bad <- vapply(cols, function(x) !inherits(x, "shard_coltype"), logical(1))
  if (any(bad)) {
    stop("schema() columns must be type specs (e.g. int32(), float64(), factor_col())", call. = FALSE)
  }

  structure(list(columns = cols), class = "shard_schema")
}

#' Allocate a fixed-row table buffer
#'
#' Allocates a columnar table output: one typed buffer per column, each of
#' length \code{nrow}. Intended for lock-free disjoint row-range writes in shard_map.
#'
#' @param schema A \code{shard_schema}.
#' @param nrow Total number of rows in the final table.
#' @param backing Backing type for buffers (\code{"auto"}, \code{"mmap"}, \code{"shm"}).
#' @return A \code{shard_table_buffer} object with one shared buffer per schema column.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64(), y = int32())
#' tb <- table_buffer(s, nrow = 100L)
#' }
table_buffer <- function(schema, nrow, backing = c("auto", "mmap", "shm")) {
  backing <- match.arg(backing)
  if (!inherits(schema, "shard_schema")) stop("schema must be a shard_schema", call. = FALSE)

  nrow <- as.integer(nrow)
  if (is.na(nrow) || nrow < 1L) stop("nrow must be a positive integer", call. = FALSE)

  bufs <- list()
  for (nm in names(schema$columns)) {
    ct <- schema$columns[[nm]]
    buf_type <- .coltype_buffer_type(ct)
    bufs[[nm]] <- buffer(buf_type, dim = nrow, init = NULL, backing = backing)
  }

  structure(
    list(schema = schema, nrow = nrow, backing = backing, columns = bufs),
    class = "shard_table_buffer"
  )
}

#' Row layout for fixed-row table outputs
#'
#' Computes disjoint row ranges for each shard via prefix-sum, enabling lock-free
#' writes where each shard writes to a unique region.
#'
#' @param shards A \code{shard_descriptor}.
#' @param rows_per_shard Either a scalar integer or a function(shard)->integer.
#' @return A named list mapping shard id (character) to an \code{idx_range(start, end)}.
#' @export
#' @examples
#' \donttest{
#' sh <- shards(100, block_size = 25)
#' layout <- row_layout(sh, rows_per_shard = 25L)
#' }
row_layout <- function(shards, rows_per_shard) {
  if (!inherits(shards, "shard_descriptor")) stop("shards must be a shard_descriptor", call. = FALSE)
  if (missing(rows_per_shard)) stop("rows_per_shard is required", call. = FALSE)

  sizes <- integer(shards$num_shards)
  if (is.function(rows_per_shard)) {
    for (i in seq_len(shards$num_shards)) {
      n <- rows_per_shard(shards$shards[[i]])
      n <- as.integer(n)
      if (is.na(n) || n < 0L) stop("rows_per_shard returned invalid size for shard ", i, call. = FALSE)
      sizes[i] <- n
    }
  } else {
    n <- as.integer(rows_per_shard)
    if (is.na(n) || n < 0L) stop("rows_per_shard must be >= 0", call. = FALSE)
    sizes[] <- n
  }

  layout <- vector("list", shards$num_shards)
  cur <- 1L
  for (i in seq_len(shards$num_shards)) {
    n <- sizes[i]
    if (n == 0L) {
      layout[[i]] <- NULL
      next
    }
    layout[[i]] <- idx_range(cur, cur + n - 1L)
    cur <- cur + n
  }
  names(layout) <- as.character(seq_len(shards$num_shards))
  layout
}

.table_cast <- function(ct, x) {
  switch(ct$kind,
    "int32" = {
      xi <- as.integer(x)
      xi
    },
    "float64" = as.double(x),
    "bool" = as.logical(x),
    "raw" = as.raw(x),
    "string" = as.character(x),
    "factor" = {
      lev <- ct$levels
      if (is.factor(x)) {
        # Ensure levels are compatible
        if (!setequal(levels(x), lev)) {
          stop("factor levels mismatch in write", call. = FALSE)
        }
        # Map to schema levels ordering
        match(as.character(x), lev)
      } else if (is.character(x)) {
        match(x, lev)
      } else if (is.integer(x) || is.numeric(x)) {
        as.integer(x)
      } else {
        stop("Unsupported factor column input type", call. = FALSE)
      }
    },
    stop("Unsupported column type: ", ct$kind, call. = FALSE)
  )
}

.table_rows_len <- function(rows) {
  if (is.null(rows)) return(0L)
  if (inherits(rows, "shard_idx_range")) return(as.integer(rows$end - rows$start + 1L))
  rows <- as.integer(rows)
  length(rows)
}

.table_rows_resolve <- function(rows) {
  if (is.null(rows)) return(NULL)
  if (inherits(rows, "shard_idx_range")) return(rows$start:rows$end)
  as.integer(rows)
}

#' Write tabular results into a table buffer or sink
#'
#' \code{table_write()} is the common write path for shard table outputs:
#' \itemize{
#'   \item For fixed-size outputs, write into a \code{shard_table_buffer} using a row selector.
#'   \item For variable-size outputs, write into a \code{shard_table_sink} using a shard id.
#' }
#'
#' @param target A \code{shard_table_buffer} or \code{shard_table_sink}.
#' @param rows_or_shard_id For buffers: row selector (idx_range or integer vector).
#'   For sinks: shard id (integer).
#' @param data A data.frame or named list matching the schema columns.
#' @param ... Reserved for future extensions.
#' @return \code{NULL} (invisibly).
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64(), y = int32())
#' tb <- table_buffer(s, nrow = 10L)
#' table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
#' }
table_write <- function(target, rows_or_shard_id, data, ...) {
  UseMethod("table_write")
}

.table_write_buffer_impl <- function(target, rows, data) {
  if (!inherits(target, "shard_table_buffer")) stop("target must be a shard_table_buffer", call. = FALSE)
  if (missing(rows)) stop("rows is required", call. = FALSE)

  row_idx <- .table_rows_resolve(rows)
  n <- .table_rows_len(rows)
  if (n < 1L) return(invisible(NULL))
  if (anyNA(row_idx) || any(row_idx < 1L) || any(row_idx > target$nrow)) {
    stop("rows contain out-of-range indices", call. = FALSE)
  }

  if (is.data.frame(data)) {
    data <- as.list(data)
  }
  if (!is.list(data) || is.null(names(data))) {
    stop("data must be a data.frame or named list", call. = FALSE)
  }

  sch <- target$schema$columns
  if (!setequal(names(data), names(sch))) {
    missing_cols <- setdiff(names(sch), names(data))
    extra_cols <- setdiff(names(data), names(sch))
    msg <- "table_write() data columns must match schema"
    if (length(missing_cols)) msg <- paste0(msg, "; missing: ", paste(missing_cols, collapse = ", "))
    if (length(extra_cols)) msg <- paste0(msg, "; extra: ", paste(extra_cols, collapse = ", "))
    stop(msg, call. = FALSE)
  }

  # Write each column slice.
  bytes <- 0
  for (nm in names(sch)) {
    ct <- sch[[nm]]
    col <- data[[nm]]
    if (length(col) != n) stop("Column '", nm, "' length mismatch (expected ", n, ")", call. = FALSE)
    cast <- .table_cast(ct, col)
    target$columns[[nm]][row_idx] <- cast

    b <- .coltype_bytes(ct)
    if (is.finite(b)) bytes <- bytes + as.double(n) * as.double(b)
  }

  .table_diag_env$writes <- .table_diag_env$writes + 1L
  .table_diag_env$rows <- .table_diag_env$rows + n
  .table_diag_env$bytes <- .table_diag_env$bytes + bytes

  invisible(NULL)
}

#' Write into a table buffer
#'
#' @param target A `shard_table_buffer`.
#' @param rows_or_shard_id Row selector (idx_range or integer vector).
#' @param data A data.frame or named list matching the schema columns.
#' @param ... Reserved for future extensions.
#' @return NULL (invisibly).
#' @export
table_write.shard_table_buffer <- function(target, rows_or_shard_id, data, ...) {
  .table_write_buffer_impl(target, rows = rows_or_shard_id, data = data)
}

#' Finalize a table buffer or sink
#'
#' For a \code{shard_table_buffer}, this returns a lightweight in-memory handle (or a
#' materialized data.frame/tibble, depending on \code{materialize}).
#'
#' For a \code{shard_table_sink}, this returns a row-group handle referencing the
#' written partitions (or materializes them if requested).
#'
#' @param target A \code{shard_table_buffer} or \code{shard_table_sink}.
#' @param materialize \code{"never"}, \code{"auto"}, or \code{"always"}.
#' @param max_bytes For \code{"auto"}, materialize only if estimated bytes <= max_bytes.
#' @param ... Reserved for future extensions.
#' @return A \code{shard_table_handle}, \code{shard_row_groups}, or materialized
#'   data.frame/tibble depending on \code{target} type and \code{materialize}.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64(), y = int32())
#' tb <- table_buffer(s, nrow = 5L)
#' table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
#' handle <- table_finalize(tb)
#' }
table_finalize <- function(target, materialize = c("never", "auto", "always"), max_bytes = 256 * 1024^2, ...) {
  UseMethod("table_finalize")
}

#' Finalize a table buffer
#'
#' @param target A `shard_table_buffer`.
#' @param materialize `"never"`, `"auto"`, or `"always"`.
#' @param max_bytes For `"auto"`, materialize only if estimated bytes <= max_bytes.
#' @param ... Reserved for future extensions.
#' @return A `shard_table_handle` or a materialized data.frame/tibble.
#' @export
table_finalize.shard_table_buffer <- function(target, materialize = c("never", "auto", "always"), max_bytes = 256 * 1024^2, ...) {
  materialize <- match.arg(materialize)

  handle <- structure(
    list(schema = target$schema, nrow = target$nrow, columns = target$columns),
    class = "shard_table_handle"
  )

  est <- 0
  for (nm in names(target$schema$columns)) {
    b <- .coltype_bytes(target$schema$columns[[nm]])
    if (is.finite(b)) est <- est + as.double(target$nrow) * as.double(b)
  }

  do_mat <- switch(materialize,
    "never" = FALSE,
    "always" = TRUE,
    "auto" = isTRUE(est <= max_bytes)
  )

  if (!do_mat) return(handle)
  as_tibble(handle, max_bytes = max_bytes)
}

#' Materialize a shard table handle as a data.frame/tibble
#'
#' @param x A shard table object.
#' @param max_bytes Warn if estimated payload exceeds this threshold.
#' @param ... Reserved for future extensions.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64(), y = int32())
#' tb <- table_buffer(s, nrow = 5L)
#' table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
#' df <- as_tibble(tb)
#' }
as_tibble <- function(x, max_bytes = 256 * 1024^2, ...) {
  UseMethod("as_tibble")
}

#' Materialize a fixed table handle or buffer
#'
#' Converts a `shard_table_handle` to an in-memory data.frame (or tibble if the
#' tibble package is installed).
#'
#' @param x A `shard_table_handle` or `shard_table_buffer`.
#' @param max_bytes Warn if estimated payload exceeds this threshold.
#' @param ... Reserved for future extensions.
#' @return A data.frame (or tibble).
#' @export
as_tibble.shard_table_buffer <- function(x, max_bytes = 256 * 1024^2, ...) {
  as_tibble(structure(list(schema = x$schema, nrow = x$nrow, columns = x$columns),
                      class = "shard_table_handle"),
            max_bytes = max_bytes)
}

#' Materialize a table handle into a data.frame/tibble
#'
#' @param x A \code{shard_table_handle}.
#' @param max_bytes Warn if estimated payload exceeds this threshold.
#' @param ... Reserved for future extensions.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
as_tibble.shard_table_handle <- function(x, max_bytes = 256 * 1024^2, ...) {
  sch <- x$schema$columns
  n <- x$nrow

  est <- 0
  for (nm in names(sch)) {
    b <- .coltype_bytes(sch[[nm]])
    if (is.finite(b)) est <- est + as.double(n) * as.double(b)
  }
  if (is.finite(est) && est > max_bytes) {
    warning("Materializing a large table (", format_bytes(est), "): consider writing to disk in table_sink().",
            call. = FALSE)
  }

  out <- list()
  for (nm in names(sch)) {
    ct <- sch[[nm]]
    buf <- x$columns[[nm]]
    vec <- buf[]  # full read
    if (ct$kind == "factor") {
      lev <- ct$levels
      codes <- as.integer(vec)
      chr <- rep(NA_character_, length(codes))
      ok <- !is.na(codes) & codes >= 1L & codes <= length(lev)
      chr[ok] <- lev[codes[ok]]
      vec <- factor(chr, levels = lev)
    } else if (ct$kind == "bool") {
      vec <- as.logical(vec)
    }
    out[[nm]] <- vec
  }
  df <- as.data.frame(out, stringsAsFactors = FALSE)

  if (pkg_available("tibble")) {
    return(tibble::as_tibble(df))
  }
  df
}

#' Create a table sink for row-group or partitioned outputs
#'
#' A table sink supports variable-sized outputs without returning large
#' data.frames to the master. Each shard writes a separate row-group file.
#'
#' v1.1 implementation notes:
#' \itemize{
#'   \item Storage format is per-shard RDS (portable, CRAN-friendly).
#'   \item This guarantees bounded master memory during execution; final collection
#'     may still be large if you materialize.
#' }
#'
#' @param schema A \code{shard_schema}. If NULL, a schema-less sink is created (RDS
#'   format only). This is primarily intended for doShard/foreach compatibility
#'   where output schemas may not be known in advance.
#' @param mode \code{"row_groups"} (temp, managed) or \code{"partitioned"} (persistent path).
#' @param path Directory to write row-group files. If NULL, a temp dir is created.
#' @param format Storage format for partitions: \code{"rds"} (data.frame RDS),
#'   \code{"native"} (columnar encoding with string offsets+bytes), or \code{"auto"}
#'   (selects \code{"native"} if the schema contains \code{string_col()}; otherwise \code{"rds"}).
#' @return A \code{shard_table_sink} object.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64(), label = string_col())
#' sink <- table_sink(s, mode = "row_groups")
#' }
table_sink <- function(schema,
                       mode = c("row_groups", "partitioned"),
                       path = NULL,
                       format = c("auto", "rds", "native")) {
  mode <- match.arg(mode)
  format <- match.arg(format)
  if (!is.null(schema) && !inherits(schema, "shard_schema")) {
    stop("schema must be a shard_schema (or NULL for schema-less RDS sinks)", call. = FALSE)
  }

  # Schema-less sinks are allowed only for the simple RDS format, primarily for
  # doShard/foreach compatibility. Schema-driven validation/encoding requires a
  # shard_schema.
  if (is.null(schema)) {
    if (!identical(format, "rds")) {
      format <- "rds"
    }
  }

  if (identical(format, "auto")) {
    has_string <- any(vapply(schema$columns, function(ct) identical(ct$kind, "string"), logical(1)))
    format <- if (has_string) "native" else "rds"
  }

  if (is.null(path)) {
    path <- file.path(tempdir(), paste0("shard_table_sink_", unique_id()))
  }
  if (!dir.exists(path)) dir.create(path, recursive = TRUE, showWarnings = FALSE)

  structure(
    list(schema = schema, mode = mode, path = path, format = format),
    class = "shard_table_sink"
  )
}

.sink_part_path <- function(sink, shard_id) {
  file.path(sink$path, sprintf("part-%06d.rds", as.integer(shard_id)))
}

.sink_manifest_rds_path <- function(sink) {
  file.path(sink$path, "manifest.rds")
}

.sink_manifest_json_path <- function(sink) {
  file.path(sink$path, "manifest.json")
}

.sink_atomic_save_rds <- function(obj, path) {
  tmp <- paste0(path, ".tmp_", Sys.getpid(), "_", sample.int(1e9, 1))
  saveRDS(obj, tmp)
  ok <- file.rename(tmp, path)
  if (!isTRUE(ok)) {
    unlink(tmp)
    stop("Failed to finalize write", call. = FALSE)
  }
  invisible(NULL)
}

.schema_to_plain_list <- function(schema) {
  if (!inherits(schema, "shard_schema")) {
    stop("schema must be a shard_schema", call. = FALSE)
  }
  cols <- lapply(schema$columns, function(ct) {
    # shard_coltype objects are simple lists; strip class for JSON stability.
    out <- as.list(ct)
    class(out) <- NULL
    out
  })
  list(
    version = schema$version %||% 1L,
    schema_hash = schema$schema_hash %||% NULL,
    columns = cols
  )
}

.sink_write_manifest <- function(sink, handle) {
  man_rds <- list(
    version = 1L,
    created = format(Sys.time(), "%Y-%m-%dT%H:%M:%SZ", tz = "UTC"),
    mode = sink$mode,
    format = sink$format,
    schema = sink$schema,
    files = basename(handle$files %||% character(0))
  )

  .sink_atomic_save_rds(man_rds, .sink_manifest_rds_path(sink))

  # Optional JSON manifest (Suggests: jsonlite). Keep it behind requireNamespace
  # so shard core stays CRAN-friendly.
  if (requireNamespace("jsonlite", quietly = TRUE)) {
    man_json <- man_rds
    if (!is.null(sink$schema)) {
      man_json$schema <- .schema_to_plain_list(sink$schema)
    }
    tmp <- paste0(.sink_manifest_json_path(sink), ".tmp_", Sys.getpid(), "_", sample.int(1e9, 1))
    jsonlite::write_json(man_json, tmp, auto_unbox = TRUE, pretty = TRUE)
    ok <- file.rename(tmp, .sink_manifest_json_path(sink))
    if (!isTRUE(ok)) unlink(tmp)
  }

  invisible(NULL)
}

.sink_atomic_write_rds <- function(obj, path) {
  tmp <- paste0(path, ".tmp_", Sys.getpid(), "_", sample.int(1e9, 1))
  saveRDS(obj, tmp)
  ok <- file.rename(tmp, path)
  if (!isTRUE(ok)) {
    unlink(tmp)
    stop("Failed to finalize row-group write", call. = FALSE)
  }
  invisible(NULL)
}

.table_part_native_encode_string <- function(x) {
  x <- as.character(x)
  n <- length(x)
  is_na <- is.na(x)
  x0 <- x
  x0[is_na] <- ""

  # Ensure stable byte encoding.
  x0 <- enc2utf8(x0)

  lens <- vapply(x0, function(s) length(charToRaw(s)), integer(1))
  total <- sum(lens)
  if (total > .Machine$integer.max) {
    stop("String payload too large for native encoding", call. = FALSE)
  }

  bytes <- raw(total)
  offsets <- integer(n + 1L)
  pos <- 0L
  offsets[1] <- 0L

  for (i in seq_len(n)) {
    l <- lens[[i]]
    if (l > 0L) {
      b <- charToRaw(x0[[i]])
      bytes[(pos + 1L):(pos + l)] <- b
      pos <- pos + l
    }
    offsets[i + 1L] <- pos
  }

  structure(
    list(offsets = offsets, bytes = bytes, is_na = is_na),
    class = "shard_string_blob"
  )
}

.table_part_native_decode_string <- function(blob) {
  if (!inherits(blob, "shard_string_blob")) stop("blob must be a shard_string_blob", call. = FALSE)
  offsets <- blob$offsets
  bytes <- blob$bytes
  is_na <- blob$is_na
  n <- length(is_na)

  out <- character(n)
  for (i in seq_len(n)) {
    if (isTRUE(is_na[[i]])) {
      out[[i]] <- NA_character_
      next
    }
    s <- offsets[[i]]
    e <- offsets[[i + 1L]]
    if (e <= s) {
      out[[i]] <- ""
    } else {
      out[[i]] <- rawToChar(bytes[(s + 1L):e])
    }
  }
  out
}

.table_part_native_decode_string_rows <- function(blob, rows) {
  if (!inherits(blob, "shard_string_blob")) stop("blob must be a shard_string_blob", call. = FALSE)
  rows <- as.integer(rows)
  rows <- rows[!is.na(rows) & rows >= 1L & rows <= length(blob$is_na)]
  if (length(rows) == 0) return(character(0))

  offsets <- blob$offsets
  bytes <- blob$bytes
  is_na <- blob$is_na

  out <- character(length(rows))
  for (j in seq_along(rows)) {
    i <- rows[[j]]
    if (isTRUE(is_na[[i]])) {
      out[[j]] <- NA_character_
      next
    }
    s <- offsets[[i]]
    e <- offsets[[i + 1L]]
    if (e <= s) {
      out[[j]] <- ""
    } else {
      out[[j]] <- rawToChar(bytes[(s + 1L):e])
    }
  }
  out
}

.table_part_native_encode <- function(df, schema) {
  if (!is.data.frame(df)) stop("df must be a data.frame", call. = FALSE)
  if (!inherits(schema, "shard_schema")) stop("schema must be a shard_schema", call. = FALSE)

  cols <- schema$columns
  out_cols <- list()
  for (nm in names(cols)) {
    ct <- cols[[nm]]
    v <- df[[nm]]
    if (ct$kind == "string") {
      out_cols[[nm]] <- .table_part_native_encode_string(v)
    } else if (ct$kind == "factor") {
      out_cols[[nm]] <- as.integer(v)
    } else {
      out_cols[[nm]] <- v
    }
  }
  structure(
    list(encoding = "native1", nrow = nrow(df), columns = out_cols),
    class = "shard_table_part_native"
  )
}

.table_part_native_decode <- function(part, schema) {
  if (!inherits(part, "shard_table_part_native")) stop("part must be shard_table_part_native", call. = FALSE)
  if (!inherits(schema, "shard_schema")) stop("schema must be a shard_schema", call. = FALSE)

  cols <- schema$columns
  out <- list()
  for (nm in names(cols)) {
    ct <- cols[[nm]]
    v <- part$columns[[nm]]
    if (ct$kind == "string") {
      out[[nm]] <- .table_part_native_decode_string(v)
    } else if (ct$kind == "factor") {
      lev <- ct$levels
      codes <- as.integer(v)
      chr <- rep(NA_character_, length(codes))
      ok <- !is.na(codes) & codes >= 1L & codes <= length(lev)
      chr[ok] <- lev[codes[ok]]
      out[[nm]] <- factor(chr, levels = lev)
    } else if (ct$kind == "bool") {
      out[[nm]] <- as.logical(v)
    } else {
      out[[nm]] <- v
    }
  }
  as.data.frame(out, stringsAsFactors = FALSE)
}

.table_part_native_decode_rows <- function(part, schema, rows) {
  if (!inherits(part, "shard_table_part_native")) stop("part must be shard_table_part_native", call. = FALSE)
  if (!inherits(schema, "shard_schema")) stop("schema must be a shard_schema", call. = FALSE)
  rows <- as.integer(rows)
  rows <- rows[!is.na(rows) & rows >= 1L & rows <= (part$nrow %||% 0L)]

  cols <- schema$columns
  out <- list()
  for (nm in names(cols)) {
    ct <- cols[[nm]]
    v <- part$columns[[nm]]
    if (ct$kind == "string") {
      out[[nm]] <- .table_part_native_decode_string_rows(v, rows)
    } else if (ct$kind == "factor") {
      lev <- ct$levels
      codes <- as.integer(v)[rows]
      chr <- rep(NA_character_, length(codes))
      ok <- !is.na(codes) & codes >= 1L & codes <= length(lev)
      chr[ok] <- lev[codes[ok]]
      out[[nm]] <- factor(chr, levels = lev)
    } else if (ct$kind == "bool") {
      out[[nm]] <- as.logical(v)[rows]
    } else {
      out[[nm]] <- v[rows]
    }
  }
  as.data.frame(out, stringsAsFactors = FALSE)
}

#' Write a shard's row-group output
#'
#' @param target A `shard_table_sink`.
#' @param rows_or_shard_id Integer shard id used to name the row-group file.
#' @param data A data.frame matching the sink schema.
#' @param ... Reserved for future extensions.
#' @return NULL (invisibly).
#' @export
table_write.shard_table_sink <- function(target, rows_or_shard_id, data, ...) {
  sink <- target

  shard_id <- rows_or_shard_id
  if (missing(shard_id)) stop("shard_id is required for table_sink writes", call. = FALSE)
  if (!dir.exists(sink$path)) dir.create(sink$path, recursive = TRUE, showWarnings = FALSE)
  if (is.data.frame(data)) {
    data <- as.list(data)
    data <- as.data.frame(data, stringsAsFactors = FALSE)
  }
  if (!is.data.frame(data)) stop("data must be a data.frame", call. = FALSE)

  if (!is.null(sink$schema)) {
    sch <- sink$schema$columns
    if (!setequal(names(data), names(sch))) {
      missing_cols <- setdiff(names(sch), names(data))
      extra_cols <- setdiff(names(data), names(sch))
      msg <- "table_write(sink, ...) data columns must match schema"
      if (length(missing_cols)) msg <- paste0(msg, "; missing: ", paste(missing_cols, collapse = ", "))
      if (length(extra_cols)) msg <- paste0(msg, "; extra: ", paste(extra_cols, collapse = ", "))
      stop(msg, call. = FALSE)
    }

    # Cast + validate each column.
    out <- data
    for (nm in names(sch)) {
      ct <- sch[[nm]]
      col <- out[[nm]]
      if (ct$kind == "factor") {
        lev <- ct$levels
        if (is.factor(col)) {
          chr <- as.character(col)
        } else if (is.character(col)) {
          chr <- col
        } else if (is.integer(col) || is.numeric(col)) {
          codes <- as.integer(col)
          chr <- rep(NA_character_, length(codes))
          ok <- !is.na(codes) & codes >= 1L & codes <= length(lev)
          chr[ok] <- lev[codes[ok]]
          if (any(!ok & !is.na(codes))) {
            stop("Invalid factor code in column '", nm, "'", call. = FALSE)
          }
        } else {
          stop("Unsupported factor column input type", call. = FALSE)
        }
        bad <- !is.na(chr) & !(chr %in% lev)
        if (any(bad)) stop("factor levels mismatch in write", call. = FALSE)
        out[[nm]] <- factor(chr, levels = lev)
      } else {
        out[[nm]] <- .table_cast(ct, col)
      }
    }
  } else {
    # Schema-less sink: accept the data.frame as-is and persist as RDS.
    out <- data
  }

  p <- .sink_part_path(sink, shard_id)
  if (identical(sink$format, "native")) {
    part <- .table_part_native_encode(out, sink$schema)
    .sink_atomic_write_rds(part, p)
  } else {
    .sink_atomic_write_rds(out, p)
  }

  n <- nrow(out)
  sz <- tryCatch(file.info(p)$size, error = function(e) NA_real_)

  .table_diag_env$writes <- .table_diag_env$writes + 1L
  .table_diag_env$rows <- .table_diag_env$rows + as.integer(n)
  if (is.finite(sz)) .table_diag_env$bytes <- .table_diag_env$bytes + as.double(sz)

  invisible(NULL)
}

#' Finalize a sink
#'
#' @param target A `shard_table_sink`.
#' @param materialize `"never"`, `"auto"`, or `"always"`.
#' @param max_bytes For `"auto"`, materialize only if estimated bytes <= max_bytes.
#' @param ... Reserved for future extensions.
#' @return A `shard_row_groups` handle (or a materialized data.frame/tibble).
#' @export
table_finalize.shard_table_sink <- function(target, materialize = c("never", "auto", "always"), max_bytes = 256 * 1024^2, ...) {
  materialize <- match.arg(materialize)

  files <- list.files(target$path, pattern = "^part-[0-9]+\\.rds$", full.names = TRUE)
  files <- sort(files)

  cls <- if (identical(target$mode, "partitioned")) "shard_dataset" else "shard_row_groups"
  handle <- structure(
    list(schema = target$schema, path = target$path, files = files, format = target$format, mode = target$mode),
    class = cls
  )

  # Always emit a manifest so users can inspect/ship outputs without reading
  # partitions, and so future backends can build on this directory layout.
  .sink_write_manifest(target, handle)

  # Determine whether to materialize by file sizes (if available).
  total <- sum(vapply(files, function(f) as.double(file.info(f)$size %||% 0), numeric(1)), na.rm = TRUE)
  do_mat <- switch(materialize,
    "never" = FALSE,
    "always" = TRUE,
    "auto" = isTRUE(total <= max_bytes)
  )

  if (!do_mat) return(handle)
  as_tibble(handle, max_bytes = max_bytes)
}

#' Iterate row groups
#'
#' @param x A \code{shard_row_groups} handle.
#' @param decode Logical. If TRUE (default), native-encoded partitions are
#'   decoded to data.frames. If FALSE, native partitions are returned as their
#'   internal representation (advanced).
#' @return A zero-argument iterator function that returns the next data.frame on
#'   each call, or \code{NULL} when exhausted.
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64())
#' sink <- table_sink(s, mode = "row_groups")
#' table_write(sink, 1L, data.frame(x = rnorm(5)))
#' rg <- table_finalize(sink)
#' it <- iterate_row_groups(rg)
#' chunk <- it()
#' }
iterate_row_groups <- function(x, decode = TRUE) {
  if (!(inherits(x, "shard_row_groups") || inherits(x, "shard_dataset"))) {
    stop("x must be a shard_row_groups or shard_dataset handle", call. = FALSE)
  }
  files <- x$files
  schema <- x$schema
  i <- 0L
  function() {
    i <<- i + 1L
    if (i > length(files)) return(NULL)
    obj <- readRDS(files[[i]])
    if (inherits(obj, "shard_table_part_native")) {
      if (isTRUE(decode)) return(.table_part_native_decode(obj, schema))
      return(obj)
    }
    obj
  }
}

# Extend as_tibble() to accept row-group handles.
#' Materialize a row-groups handle into a data.frame/tibble
#'
#' @param x A \code{shard_row_groups} handle.
#' @param max_bytes Accepted for API consistency; currently unused for row-groups.
#' @param ... Reserved for future extensions.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
as_tibble.shard_row_groups <- function(x, max_bytes = 256 * 1024^2, ...) {
  # max_bytes is accepted for API consistency; row-groups are size-variable and
  # are typically materialized explicitly by users.
  it <- iterate_row_groups(x)
  chunks <- list()
  repeat {
    d <- it()
    if (is.null(d)) break
    chunks[[length(chunks) + 1L]] <- d
  }
  if (length(chunks) == 0) {
    df <- data.frame()
    if (pkg_available("tibble")) return(tibble::as_tibble(df))
    return(df)
  }
  df <- do.call(rbind, chunks)
  if (pkg_available("tibble")) return(tibble::as_tibble(df))
  df
}

#' Materialize a dataset handle into a data.frame/tibble
#'
#' @param x A \code{shard_dataset} handle.
#' @param max_bytes Accepted for API consistency.
#' @param ... Reserved for future extensions.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
as_tibble.shard_dataset <- function(x, max_bytes = 256 * 1024^2, ...) {
  # A shard_dataset is currently a directory of per-shard RDS partitions plus a
  # manifest. Materialization is identical to row-groups for now.
  as_tibble(structure(unclass(x), class = "shard_row_groups"), max_bytes = max_bytes)
}

#' Collect a shard table into memory
#'
#' \code{collect()} is a convenience alias for \code{as_tibble()} for shard table outputs.
#'
#' @param x A shard table handle (\code{shard_row_groups}, \code{shard_dataset},
#'   or \code{shard_table_handle}).
#' @param ... Passed to \code{as_tibble()}.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
#' @examples
#' \donttest{
#' s <- schema(x = float64(), y = int32())
#' tb <- table_buffer(s, nrow = 5L)
#' table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
#' handle <- table_finalize(tb)
#' df <- collect(handle)
#' }
collect <- function(x, ...) {
  UseMethod("collect")
}

#' Collect a row-groups handle into memory
#'
#' @param x A \code{shard_row_groups} handle.
#' @param ... Passed to \code{as_tibble()}.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
collect.shard_row_groups <- function(x, ...) {
  as_tibble(x, ...)
}

#' Collect a dataset handle into memory
#'
#' @param x A \code{shard_dataset} handle.
#' @param ... Passed to \code{as_tibble()}.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
collect.shard_dataset <- function(x, ...) {
  as_tibble(x, ...)
}

#' Collect a table handle into memory
#'
#' @param x A \code{shard_table_handle}.
#' @param ... Passed to \code{as_tibble()}.
#' @return A data.frame (or tibble if the \code{tibble} package is installed).
#' @export
collect.shard_table_handle <- function(x, ...) {
  as_tibble(x, ...)
}

test_that("table_buffer can be used as an out= target in shard_map", {
  skip_on_cran()

  pool_stop()
  shard:::table_reset_diagnostics()

  sch <- schema(a = int32(), b = float64())
  s <- shards(10, block_size = 2, workers = 2)
  layout <- row_layout(s, rows_per_shard = function(sh) sh$len)

  tbl <- table_buffer(sch, nrow = 10, backing = "mmap")

  res <- shard_map(
    s,
    out = list(tbl = tbl),
    fun = function(shard, tbl) {
      rows <- layout[[as.character(shard$id)]]
      table_write(tbl, rows, data.frame(
        a = as.integer(shard$idx),
        b = as.double(shard$idx) * 0.5
      ))
      NULL
    },
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))

  cr <- copy_report(res)
  expect_equal(cr$table_rows %||% 0L, 10L)
  expect_gt(cr$table_bytes %||% 0, 0)

  df <- table_finalize(tbl, materialize = "always")
  expect_equal(df$a, 1:10)
  expect_equal(df$b, (1:10) * 0.5)

  pool_stop()
})

test_that("stream_* helpers work without collect()", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32(), msg = string_col(), val = float64())
  s <- shards(10, block_size = 3, workers = 2)

  out_path <- file.path(tempdir(), paste0("shard_test_stream_", shard:::unique_id()))
  sink <- table_sink(sch, mode = "partitioned", path = out_path)

  res <- shard_map(
    s,
    out = list(sink = sink),
    fun = function(shard, sink) {
      df <- data.frame(
        id = as.integer(shard$idx),
        msg = paste0("x", shard$idx),
        val = as.double(shard$idx),
        stringsAsFactors = FALSE
      )
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )

  expect_true(succeeded(res))

  ds <- table_finalize(sink, materialize = "never")
  expect_s3_class(ds, "shard_dataset")

  expect_equal(stream_count(ds), 10L)

  ssum <- stream_reduce(
    ds,
    f = function(chunk) sum(chunk$val),
    init = 0,
    combine = `+`
  )
  expect_equal(ssum, sum(as.double(1:10)))

  filtered_path <- file.path(tempdir(), paste0("shard_test_stream_filt_", shard:::unique_id()))
  ds2 <- stream_filter(ds, predicate = function(chunk) chunk$id %% 2L == 0L, path = filtered_path)
  expect_s3_class(ds2, "shard_dataset")

  df2 <- collect(ds2)
  df2 <- as.data.frame(df2, stringsAsFactors = FALSE)
  expect_equal(df2$id, as.integer(seq(2, 10, by = 2)))

  unlink(out_path, recursive = TRUE, force = TRUE)
  unlink(filtered_path, recursive = TRUE, force = TRUE)
  pool_stop()
})

test_that("stream_map applies function to each partition", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32(), val = float64())
  s <- shards(6, block_size = 2, workers = 2)

  out_path <- file.path(tempdir(), paste0("shard_test_stream_map_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)

  res <- shard_map(
    s,
    out = list(sink = sink),
    fun = function(shard, sink) {
      df <- data.frame(id = as.integer(shard$idx), val = as.double(shard$idx) * 2)
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  # stream_map returns list of per-partition results
  results <- stream_map(ds, f = function(chunk) nrow(chunk))
  expect_true(is.list(results))
  expect_equal(sum(unlist(results)), 6L)

  pool_stop()
})

test_that("stream_sum computes correct sum", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32(), val = float64())
  s <- shards(10, block_size = 2, workers = 2)

  out_path <- file.path(tempdir(), paste0("shard_test_stream_sum_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)

  res <- shard_map(
    s,
    out = list(sink = sink),
    fun = function(shard, sink) {
      df <- data.frame(id = as.integer(shard$idx), val = as.double(shard$idx))
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  # Sum of 1:10 = 55
  total <- stream_sum(ds, col = "val")
  expect_equal(total, sum(1:10))

  pool_stop()
})

test_that("stream_sum validates inputs", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32(), msg = string_col())
  out_path <- file.path(tempdir(), paste0("shard_test_stream_sum_err_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)
  res <- shard_map(
    shards(2, block_size = 1),
    out = list(sink = sink),
    fun = function(shard, sink) {
      df <- data.frame(id = shard$id, msg = paste0("x", shard$id))
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  # String column cannot be summed
  expect_error(stream_sum(ds, col = "msg"), "requires a numeric column")

  # Unknown column
  expect_error(stream_sum(ds, col = "nonexistent"), "Unknown column")

  pool_stop()
})

test_that("stream_reduce validates f and combine", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32())
  out_path <- file.path(tempdir(), paste0("shard_test_reduce_err_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)
  res <- shard_map(
    shards(2, block_size = 1),
    out = list(sink = sink),
    fun = function(shard, sink) {
      table_write(sink, shard$id, data.frame(id = shard$id))
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  expect_error(stream_reduce(ds, f = "not_a_function", init = 0, combine = `+`),
               "f must be a function")
  expect_error(stream_reduce(ds, f = nrow, init = 0, combine = "not_a_function"),
               "combine must be a function")

  pool_stop()
})

test_that("stream_filter validates predicate", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32())
  out_path <- file.path(tempdir(), paste0("shard_test_filter_err_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)
  res <- shard_map(
    shards(2, block_size = 1),
    out = list(sink = sink),
    fun = function(shard, sink) {
      table_write(sink, shard$id, data.frame(id = shard$id))
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  expect_error(stream_filter(ds, predicate = "not_a_function"),
               "predicate must be a function")

  pool_stop()
})

test_that("stream_map validates f", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32())
  out_path <- file.path(tempdir(), paste0("shard_test_map_err_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)
  res <- shard_map(
    shards(2, block_size = 1),
    out = list(sink = sink),
    fun = function(shard, sink) {
      table_write(sink, shard$id, data.frame(id = shard$id))
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  expect_error(stream_map(ds, f = "not_a_function"),
               "f must be a function")

  pool_stop()
})

test_that("stream_top_k finds top k rows", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32(), score = float64())
  out_path <- file.path(tempdir(), paste0("shard_test_topk_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)

  res <- shard_map(
    shards(10, block_size = 2),
    out = list(sink = sink),
    fun = function(shard, sink) {
      df <- data.frame(id = as.integer(shard$idx), score = as.double(shard$idx))
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  # Top 3 by score (decreasing)
  top3 <- stream_top_k(ds, col = "score", k = 3, decreasing = TRUE)
  expect_true(nrow(top3) == 3)
  expect_equal(as.integer(top3$id), c(10L, 9L, 8L))

  # Bottom 3 (ascending)
  bottom3 <- stream_top_k(ds, col = "score", k = 3, decreasing = FALSE)
  expect_true(nrow(bottom3) == 3)
  expect_equal(as.integer(bottom3$id), c(1L, 2L, 3L))

  pool_stop()
})

test_that("stream_top_k validates inputs", {
  skip_on_cran()

  pool_stop()

  sch <- schema(id = int32(), msg = string_col())
  out_path <- file.path(tempdir(), paste0("shard_test_topk_err_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)
  res <- shard_map(
    shards(2, block_size = 1),
    out = list(sink = sink),
    fun = function(shard, sink) {
      table_write(sink, shard$id, data.frame(id = shard$id, msg = paste0("x", shard$id)))
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  # k must be >= 1
  expect_error(stream_top_k(ds, col = "id", k = 0), "k must be >= 1")

  # Unknown column
  expect_error(stream_top_k(ds, col = "nonexistent", k = 3), "Unknown column")

  # String column not supported
  expect_error(stream_top_k(ds, col = "msg", k = 3), "requires an int32\\(\\) or float64\\(\\)")

  pool_stop()
})

test_that("stream_group_sum computes grouped sums", {
  skip_on_cran()

  pool_stop()

  sch <- schema(grp = factor_col(c("A", "B", "C")), val = float64())
  out_path <- file.path(tempdir(), paste0("shard_test_grpsum_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)

  res <- shard_map(
    shards(9, block_size = 3),
    out = list(sink = sink),
    fun = function(shard, sink) {
      # Create pattern: A=1, B=2, C=3, A=4, B=5, C=6, A=7, B=8, C=9
      grps <- c("A", "B", "C")[(shard$idx - 1L) %% 3L + 1L]
      df <- data.frame(
        grp = factor(grps, levels = c("A", "B", "C")),
        val = as.double(shard$idx)
      )
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  result <- stream_group_sum(ds, group = "grp", value = "val")

  expect_true(is.data.frame(result))
  expect_true("group" %in% names(result))
  expect_true("sum" %in% names(result))

  # A: 1+4+7=12, B: 2+5+8=15, C: 3+6+9=18
  expect_equal(result$sum[result$group == "A"], 12)
  expect_equal(result$sum[result$group == "B"], 15)
  expect_equal(result$sum[result$group == "C"], 18)

  pool_stop()
})

test_that("stream_group_count counts rows per group", {
  skip_on_cran()

  pool_stop()

  sch <- schema(grp = factor_col(c("X", "Y")), id = int32())
  out_path <- file.path(tempdir(), paste0("shard_test_grpcnt_", shard:::unique_id()))
  on.exit(unlink(out_path, recursive = TRUE, force = TRUE), add = TRUE)

  sink <- table_sink(sch, mode = "partitioned", path = out_path)

  res <- shard_map(
    shards(6, block_size = 2),
    out = list(sink = sink),
    fun = function(shard, sink) {
      # Alternating X, Y
      grps <- c("X", "Y")[(shard$idx - 1L) %% 2L + 1L]
      df <- data.frame(
        grp = factor(grps, levels = c("X", "Y")),
        id = as.integer(shard$idx)
      )
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )

  ds <- table_finalize(sink, materialize = "never")

  result <- stream_group_count(ds, group = "grp")

  expect_true(is.data.frame(result))
  expect_true("group" %in% names(result))
  expect_true("n" %in% names(result))

  # X: 1,3,5 (3 items), Y: 2,4,6 (3 items)
  expect_equal(result$n[result$group == "X"], 3L)
  expect_equal(result$n[result$group == "Y"], 3L)

  pool_stop()
})

test_that("table_sink(partitioned) writes a manifest and can be collected", {
  skip_on_cran()

  pool_stop()
  shard:::table_reset_diagnostics()

  sch <- schema(id = int32(), grp = factor_col(levels = c("odd", "even")), msg = string_col(), val = float64())
  s <- shards(12, block_size = 4, workers = 2)

  out_path <- file.path(tempdir(), paste0("shard_test_dataset_", shard:::unique_id()))
  sink <- table_sink(sch, mode = "partitioned", path = out_path)

  res <- shard_map(
    s,
    out = list(sink = sink),
    fun = function(shard, sink) {
      grp <- if ((shard$id %% 2L) == 0L) "even" else "odd"
      df <- data.frame(
        id = as.integer(shard$idx),
        grp = rep(grp, length(shard$idx)),
        msg = ifelse(shard$idx == 3L, "", paste0("x", shard$idx)),
        val = as.double(shard$idx) * 0.1,
        stringsAsFactors = FALSE
      )
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))

  handle <- table_finalize(sink, materialize = "never")
  expect_s3_class(handle, "shard_dataset")
  expect_true(file.exists(file.path(out_path, "manifest.rds")))
  expect_equal(length(handle$files), s$num_shards)

  df <- collect(handle)
  df <- as.data.frame(df, stringsAsFactors = FALSE)

  expect_equal(df$id, as.integer(1:12))
  expected_msg <- paste0("x", 1:12)
  expected_msg[3] <- ""
  expect_equal(df$msg, expected_msg)
  expect_equal(df$val, as.double(1:12) * 0.1)

  unlink(out_path, recursive = TRUE, force = TRUE)
  pool_stop()
})

test_that("schema/table_buffer/table_write materialize correctly", {
  shard:::table_reset_diagnostics()

  sch <- schema(
    a = int32(),
    b = float64(),
    ok = bool(),
    g = factor_col(levels = c("x", "y"))
  )

  tbl <- table_buffer(sch, nrow = 5, backing = "mmap")
  table_write(tbl, idx_range(1, 3), data.frame(
    a = c(1L, 2L, 3L),
    b = c(1.5, 2.5, 3.5),
    ok = c(TRUE, FALSE, TRUE),
    g = factor(c("x", "y", "x"), levels = c("x", "y"))
  ))

  df <- table_finalize(tbl, materialize = "always")
  expect_equal(df$a[1:3], 1:3)
  expect_equal(df$b[1:3], c(1.5, 2.5, 3.5))
  expect_equal(df$ok[1:3], c(TRUE, FALSE, TRUE))
  expect_equal(as.character(df$g[1:3]), c("x", "y", "x"))

  td <- table_diagnostics()
  expect_equal(td$writes, 1L)
  expect_equal(td$rows, 3L)
  expect_gt(td$bytes, 0)
})

test_that("row_layout computes disjoint ranges", {
  s <- shards(10, block_size = 4, workers = 2)
  layout <- row_layout(s, rows_per_shard = function(sh) sh$len)
  expect_equal(layout[[1]]$start, 1L)
  expect_equal(layout[[1]]$end, 4L)
  expect_equal(layout[[2]]$start, 5L)
  expect_equal(layout[[3]]$end, 10L)
})

test_that("table_sink(row_groups) can be used as an out= target in shard_map", {
  skip_on_cran()

  pool_stop()
  shard:::table_reset_diagnostics()

  sch <- schema(id = int32(), grp = factor_col(levels = c("odd", "even")), msg = string_col(), val = float64())
  s <- shards(10, block_size = 3, workers = 2)

  sink_path <- file.path(tempdir(), paste0("shard_test_sink_", shard:::unique_id()))
  sink <- table_sink(sch, mode = "row_groups", path = sink_path)

  res <- shard_map(
    s,
    out = list(sink = sink),
    fun = function(shard, sink) {
      if ((shard$id %% 2L) == 1L) {
        # Variable-sized output: some shards emit nothing.
        df <- data.frame(id = integer(0), grp = character(0), msg = character(0), val = numeric(0),
                         stringsAsFactors = FALSE)
      } else {
        df <- data.frame(
          id = as.integer(shard$idx),
          grp = rep("even", length(shard$idx)),
          msg = ifelse(shard$idx == 5L, NA_character_, paste0("x", shard$idx)),
          val = as.double(shard$idx) * 0.25,
          stringsAsFactors = FALSE
        )
      }

      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2,
    diagnostics = TRUE
  )

  expect_true(succeeded(res))

  cr <- copy_report(res)
  expect_gt(cr$table_writes %||% 0L, 0L)
  expect_gt(cr$table_bytes %||% 0, 0)

  handle <- table_finalize(sink, materialize = "never")
  expect_s3_class(handle, "shard_row_groups")
  expect_equal(length(handle$files), s$num_shards)

  df <- as_tibble(handle)
  df <- as.data.frame(df, stringsAsFactors = FALSE)

  expected_idx <- unlist(lapply(s$shards, function(sh) {
    if ((sh$id %% 2L) == 0L) sh$idx else integer(0)
  }))

  expect_equal(df$id, as.integer(expected_idx))
  expected_msg <- paste0("x", expected_idx)
  if (5L %in% expected_idx) {
    expected_msg[which(expected_idx == 5L)] <- NA_character_
  }
  expect_equal(df$msg, expected_msg)
  expect_equal(df$val, as.double(expected_idx) * 0.25)

  unlink(sink_path, recursive = TRUE, force = TRUE)
  pool_stop()
})

test_that("stream_top_k and stream_group_* work on native datasets", {
  skip_on_cran()

  pool_stop()

  sch <- schema(
    id = int32(),
    grp = factor_col(levels = c("odd", "even")),
    msg = string_col(),
    val = float64()
  )

  s <- shards(12, block_size = 4, workers = 2)
  out_path <- file.path(tempdir(), paste0("shard_test_wow_", shard:::unique_id()))
  sink <- table_sink(sch, mode = "partitioned", path = out_path) # format auto -> native

  res <- shard_map(
    s,
    out = list(sink = sink),
    fun = function(shard, sink) {
      df <- data.frame(
        id = as.integer(shard$idx),
        grp = ifelse((as.integer(shard$idx) %% 2L) == 0L, "even", "odd"),
        msg = ifelse(shard$idx == 9L, NA_character_, paste0("x", shard$idx)),
        val = as.double(shard$idx),
        stringsAsFactors = FALSE
      )
      table_write(sink, shard$id, df)
      NULL
    },
    workers = 2
  )
  expect_true(succeeded(res))

  ds <- table_finalize(sink, materialize = "never")
  expect_s3_class(ds, "shard_dataset")

  top <- stream_top_k(ds, col = "val", k = 3)
  top <- as.data.frame(top, stringsAsFactors = FALSE)
  expect_equal(top$id, as.integer(c(12, 11, 10)))
  expect_equal(top$val, as.double(c(12, 11, 10)))
  expect_equal(top$msg, c("x12", "x11", "x10"))

  # Group counts should be balanced (6 even, 6 odd).
  cnt <- stream_group_count(ds, group = "grp")
  cnt <- as.data.frame(cnt, stringsAsFactors = FALSE)
  expect_equal(as.integer(cnt$n), c(6L, 6L))

  # Group sums: sum(odd ids) and sum(even ids).
  sm <- stream_group_sum(ds, group = "grp", value = "val")
  sm <- as.data.frame(sm, stringsAsFactors = FALSE)
  expect_equal(round(sm$sum, 6), c(sum(seq(1, 12, by = 2)), sum(seq(2, 12, by = 2))))

  # Numeric sum should ignore strings entirely.
  expect_equal(stream_sum(ds, col = "val"), sum(as.double(1:12)))

  unlink(out_path, recursive = TRUE, force = TRUE)
  pool_stop()
})

#' @title Shared Memory Segment
#' @name segment
#' @description Low-level shared memory segment operations for cross-process
#'   data sharing. These functions provide the foundation for the higher-level
#'   \code{share()} and \code{buffer()} APIs.
#'
#' @details
#' Segments can be backed by:
#' \itemize{
#'   \item \code{"shm"}: POSIX shared memory (Linux/macOS) or named file
#'     mapping (Windows). Faster but may have size limitations.
#'   \item \code{"mmap"}: File-backed memory mapping. Works on all platforms
#'     and supports larger sizes.
#'   \item \code{"auto"}: Let the system choose the best option.
#' }
#'
#' All segments are created with secure permissions (0600 on Unix) and are
#' automatically cleaned up when the R object is garbage collected.
#'
#' @useDynLib shard, .registration = TRUE
NULL

#' Create a new shared memory segment
#'
#' @param size Size of the segment in bytes
#' @param backing Backing type: "auto", "mmap", or "shm"
#' @param path Optional file path for mmap backing (NULL for temp file)
#' @param readonly Create as read-only (after initial write)
#' @return A \code{shard_segment} object backed by shared memory.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024 * 1024)
#' segment_info(seg)
#' segment_close(seg)
#' }
segment_create <- function(size,
                           backing = c("auto", "mmap", "shm"),
                           path = NULL,
                           readonly = FALSE) {
    backing <- match.arg(backing)
    backing_int <- switch(backing,
        "auto" = 0L,
        "mmap" = 1L,
        "shm"  = 2L
    )

    size <- as.double(size)
    if (size <= 0) {
        stop("size must be positive")
    }

    ptr <- .Call("C_shard_segment_create", size, backing_int, path, readonly,
                 PACKAGE = "shard")

    structure(
        list(
            ptr = ptr,
            size = size,
            backing = backing,
            readonly = readonly
        ),
        class = "shard_segment"
    )
}

#' Open an existing shared memory segment
#'
#' @param path Path or shm name of the segment
#' @param backing Backing type: "mmap" or "shm"
#' @param readonly Open as read-only
#' @return A \code{shard_segment} object attached to the existing segment.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024, backing = "mmap")
#' path <- segment_path(seg)
#' seg2 <- segment_open(path, backing = "mmap", readonly = TRUE)
#' segment_close(seg2, unlink = FALSE)
#' segment_close(seg)
#' }
segment_open <- function(path,
                         backing = c("mmap", "shm"),
                         readonly = TRUE) {
    backing <- match.arg(backing)
    backing_int <- switch(backing,
        "mmap" = 1L,
        "shm"  = 2L
    )

    ptr <- .Call("C_shard_segment_open", path, backing_int, readonly,
                 PACKAGE = "shard")
    info <- .Call("C_shard_segment_info", ptr, PACKAGE = "shard")

    structure(
        list(
            ptr = ptr,
            size = info$size,
            backing = backing,
            readonly = readonly
        ),
        class = "shard_segment"
    )
}

#' Close a shared memory segment
#'
#' @param x A shard_segment object
#' @param unlink Whether to unlink the underlying file/shm (default: FALSE for
#'   opened segments, TRUE for owned segments)
#' @return \code{NULL}, invisibly.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' segment_close(seg)
#' }
segment_close <- function(x, unlink = NULL) {
    stopifnot(inherits(x, "shard_segment"))

    if (is.null(unlink)) {
        info <- .Call("C_shard_segment_info", x$ptr, PACKAGE = "shard")
        unlink <- info$owns
    }

    .Call("C_shard_segment_close", x$ptr, unlink, PACKAGE = "shard")
    invisible(NULL)
}

#' Get segment information
#'
#' @param x A shard_segment object
#' @return A named list with segment metadata including \code{size}, \code{backing},
#'   \code{path}, \code{readonly}, and \code{owns}.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' segment_info(seg)
#' segment_close(seg)
#' }
segment_info <- function(x) {
    stopifnot(inherits(x, "shard_segment"))
    .Call("C_shard_segment_info", x$ptr, PACKAGE = "shard")
}

#' Get the path or name of a segment
#'
#' @param x A shard_segment object
#' @return The path string, or \code{NULL} for anonymous segments.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024, backing = "mmap")
#' segment_path(seg)
#' segment_close(seg)
#' }
segment_path <- function(x) {
    stopifnot(inherits(x, "shard_segment"))
    .Call("C_shard_segment_path", x$ptr, PACKAGE = "shard")
}

#' Get the size of a segment
#'
#' @param x A shard_segment object
#' @return Size in bytes as a numeric scalar.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' segment_size(seg)
#' segment_close(seg)
#' }
segment_size <- function(x) {
    stopifnot(inherits(x, "shard_segment"))
    .Call("C_shard_segment_size", x$ptr, PACKAGE = "shard")
}

#' Write data to a segment
#'
#' @param x A shard_segment object
#' @param data Data to write (raw, numeric, integer, or logical vector)
#' @param offset Byte offset to start writing (0-based)
#' @return Number of bytes written, invisibly.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' segment_write(seg, as.integer(1:10), offset = 0)
#' segment_close(seg)
#' }
segment_write <- function(x, data, offset = 0) {
    stopifnot(inherits(x, "shard_segment"))
    bytes <- .Call("C_shard_segment_write_raw", x$ptr, data, as.double(offset),
                   PACKAGE = "shard")
    invisible(bytes)
}

#' Read raw data from a segment
#'
#' @param x A shard_segment object
#' @param offset Byte offset to start reading (0-based)
#' @param size Number of bytes to read
#' @return A raw vector containing the bytes read from the segment.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' segment_write(seg, as.integer(1:4), offset = 0)
#' segment_read(seg, offset = 0, size = 16)
#' segment_close(seg)
#' }
segment_read <- function(x, offset = 0, size = NULL) {
    stopifnot(inherits(x, "shard_segment"))

    if (is.null(size)) {
        size <- segment_size(x) - offset
    }

    .Call("C_shard_segment_read_raw", x$ptr, as.double(offset), as.double(size),
          PACKAGE = "shard")
}

#' Make a segment read-only
#'
#' @param x A shard_segment object
#' @return The \code{shard_segment} object, invisibly.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' segment_protect(seg)
#' segment_close(seg)
#' }
segment_protect <- function(x) {
    stopifnot(inherits(x, "shard_segment"))
    .Call("C_shard_segment_protect", x$ptr, PACKAGE = "shard")
    x$readonly <- TRUE
    invisible(x)
}

#' Print a Shared Memory Segment
#'
#' @param x A \code{shard_segment} object.
#' @param ... Ignored.
#' @return The input \code{x}, invisibly.
#' @export
#' @examples
#' \donttest{
#' seg <- segment_create(1024)
#' print(seg)
#' segment_close(seg)
#' }
print.shard_segment <- function(x, ...) {
    info <- segment_info(x)
    cat("<shard_segment>\n")
    cat("  Size:", format(info$size, big.mark = ","), "bytes\n")
    cat("  Backing:", info$backing, "\n")
    if (!is.null(info$path)) {
        cat("  Path:", info$path, "\n")
    }
    cat("  Read-only:", info$readonly, "\n")
    cat("  Owns segment:", info$owns, "\n")
    invisible(x)
}

#' Check if running on Windows
#'
#' @return A logical scalar: \code{TRUE} if running on Windows, \code{FALSE} otherwise.
#' @export
#' @examples
#' is_windows()
is_windows <- function() {
    .Call("C_shard_is_windows", PACKAGE = "shard")
}

#' Get available shared memory backing types
#'
#' @return A character vector of available backing types on the current platform.
#' @export
#' @examples
#' available_backings()
available_backings <- function() {
    .Call("C_shard_available_backings", PACKAGE = "shard")
}

# Tests for shared memory segment operations

test_that("segment_create creates a segment with correct size", {
    seg <- segment_create(1024)
    expect_s3_class(seg, "shard_segment")
    expect_equal(segment_size(seg), 1024)
    segment_close(seg)
})

test_that("segment_create respects backing type", {
    # Test mmap backing
    seg_mmap <- segment_create(1024, backing = "mmap")
    info <- segment_info(seg_mmap)
    expect_equal(info$backing, "mmap")
    segment_close(seg_mmap)

    # Test auto backing (should work on all platforms)
    seg_auto <- segment_create(1024, backing = "auto")
    expect_s3_class(seg_auto, "shard_segment")
    segment_close(seg_auto)
})

test_that("segment_write and segment_read work with raw data", {
    seg <- segment_create(1024)

    # Write raw data
    data <- as.raw(1:100)
    segment_write(seg, data, offset = 0)

    # Read it back
    result <- segment_read(seg, offset = 0, size = 100)
    expect_equal(result, data)

    segment_close(seg)
})

test_that("segment_write and segment_read work with numeric data", {
    seg <- segment_create(1024)

    # Write numeric data
    data <- c(1.5, 2.5, 3.5, 4.5)
    segment_write(seg, data, offset = 0)

    # Read as raw and convert back
    raw_result <- segment_read(seg, offset = 0, size = length(data) * 8)
    result <- readBin(raw_result, "double", n = length(data))
    expect_equal(result, data)

    segment_close(seg)
})

test_that("segment_write and segment_read work with integer data", {
    seg <- segment_create(1024)

    # Write integer data
    data <- 1L:10L
    segment_write(seg, data, offset = 0)

    # Read as raw and convert back
    raw_result <- segment_read(seg, offset = 0, size = length(data) * 4)
    result <- readBin(raw_result, "integer", n = length(data))
    expect_equal(result, data)

    segment_close(seg)
})

test_that("segment_info returns correct information", {
    seg <- segment_create(2048, backing = "mmap")
    info <- segment_info(seg)

    expect_true(is.list(info))
    expect_equal(info$size, 2048)
    expect_equal(info$backing, "mmap")
    expect_false(info$readonly)
    expect_true(info$owns)
    expect_true(!is.null(info$path))

    segment_close(seg)
})

test_that("segment_path returns the path", {
    seg <- segment_create(1024, backing = "mmap")
    path <- segment_path(seg)

    expect_true(is.character(path) || is.null(path))
    if (is.character(path)) {
        expect_true(nchar(path) > 0)
    }

    segment_close(seg)
})

test_that("segment_protect makes segment read-only", {
    seg <- segment_create(1024)

    # Write before protecting
    data <- as.raw(1:10)
    segment_write(seg, data, offset = 0)

    # Protect
    segment_protect(seg)
    info <- segment_info(seg)
    expect_true(info$readonly)

    segment_close(seg)
})

test_that("segment_open can reopen an existing segment", {
    # Create and write to a segment
    seg1 <- segment_create(1024, backing = "mmap")
    data <- as.raw(42:52)
    segment_write(seg1, data, offset = 0)

    # Get the path
    path <- segment_path(seg1)
    skip_if(is.null(path), "Path is NULL")

    # Open the same segment
    seg2 <- segment_open(path, backing = "mmap", readonly = TRUE)
    result <- segment_read(seg2, offset = 0, size = length(data))
    expect_equal(result, data)

    # Clean up (close seg2 first without unlinking)
    segment_close(seg2, unlink = FALSE)
    segment_close(seg1, unlink = TRUE)
})

test_that("is_windows returns logical", {
    result <- is_windows()
    expect_true(is.logical(result))
    expect_length(result, 1)
})

test_that("available_backings returns character vector", {
    backings <- available_backings()
    expect_true(is.character(backings))
    expect_true("mmap" %in% backings)
})

test_that("print.shard_segment works", {
    seg <- segment_create(1024)
    expect_output(print(seg), "shard_segment")
    expect_output(print(seg), "Size:")
    expect_output(print(seg), "Backing:")
    segment_close(seg)
})

test_that("segment handles large sizes", {
    # 10 MB segment
    size <- 10 * 1024 * 1024
    seg <- segment_create(size)
    expect_equal(segment_size(seg), size)
    segment_close(seg)
})

test_that("segment write at offset works", {
    seg <- segment_create(1024)

    # Write at offset
    data <- as.raw(1:10)
    segment_write(seg, data, offset = 100)

    # Read from that offset
    result <- segment_read(seg, offset = 100, size = 10)
    expect_equal(result, data)

    segment_close(seg)
})

#' @title Arena Semantic Scope
#' @name arena
#' @description Semantic scope for scratch memory that signals temporary data
#'   should not accumulate. Enables memory-conscious parallel execution.
#'
#' @details
#' The `arena()` function provides a semantic scope that signals "this code
#' produces scratch data that should not outlive the scope." It serves two
#' purposes:
#'
#' \enumerate{
#'   \item \strong{For compiled kernels}: When Rust-based kernels are available,
#'     arena() provides real scratch arenas backed by temporary shared memory
#'     segments that are automatically reclaimed.
#'   \item \strong{For arbitrary R code}: Triggers post-task memory checks to
#'     detect growth and potential memory leaks.
#' }
#'
#' The \code{strict} parameter controls escape detection:
#' \itemize{
#'   \item \code{strict = FALSE} (default): Returns results normally, logs
#'     diagnostics about memory growth.
#'   \item \code{strict = TRUE}: Warns or errors if large objects escape the
#'     scope, and triggers aggressive memory reclamation.
#' }
#'
#' @seealso \code{\link{shard_map}} for parallel execution,
#'   \code{\link{share}} for shared memory inputs.
NULL

# Default threshold for "large object" detection in strict mode (1 MB)
.arena_escape_threshold <- 1024 * 1024

# Environment to hold arena state for nested arenas
.arena_state <- new.env(parent = emptyenv())

#' Execute Code in a Scratch Arena
#'
#' Evaluates an expression in a semantic scope that signals scratch memory usage.
#' This enables memory-conscious execution where temporaries are expected to be
#' reclaimed after the scope exits.
#'
#' @param expr An expression to evaluate within the arena scope.
#' @param strict Logical. If TRUE, enables strict mode which:
#'   \itemize{
#'     \item Warns if large objects (> 1MB by default) escape the scope
#'     \item Triggers garbage collection after scope exit
#'     \item Tracks memory growth for diagnostics
#'   }
#'   Default is FALSE for compatibility and performance.
#' @param escape_threshold Numeric. Size in bytes above which returned objects
#'   trigger a warning in strict mode. Default is 1MB (1048576 bytes).
#'   Only used when \code{strict = TRUE}.
#' @param gc_after Logical. If TRUE, triggers garbage collection after the
#'   arena scope exits. Default is TRUE in strict mode, FALSE otherwise.
#' @param diagnostics Logical. If TRUE, returns diagnostics about memory usage
#'   along with the result. Default is FALSE.
#'
#' @return The result of evaluating \code{expr}. If \code{diagnostics = TRUE},
#'   returns an \code{arena_result} object with elements \code{result} and
#'   \code{diagnostics}.
#' @export
#' @examples
#' \donttest{
#' result <- arena({
#'   tmp <- matrix(rnorm(1e6), nrow = 1000)
#'   colMeans(tmp)
#' })
#'
#' info <- arena({
#'   x <- rnorm(1e5)
#'   sum(x)
#' }, diagnostics = TRUE)
#' info$diagnostics
#' }
arena <- function(expr,
                  strict = FALSE,
                  escape_threshold = .arena_escape_threshold,
                  gc_after = strict,
                  diagnostics = FALSE) {

    # Capture expression for evaluation
    expr <- substitute(expr)
    parent_env <- parent.frame()

    # Initialize arena state
    arena_id <- .arena_push()
    on.exit(.arena_pop(arena_id, gc_after), add = TRUE)

    # Capture initial memory state
    rss_before <- if (strict || diagnostics) rss_self() else NA_real_

    # Track GC state before
    gc_before <- if (diagnostics) gc(reset = FALSE, full = FALSE) else NULL

    # Evaluate the expression
    result <- tryCatch(
        eval(expr, envir = parent_env),
        error = function(e) {
            # Re-throw with arena context
            stop(conditionMessage(e), call. = FALSE)
        }
    )

    # Post-execution checks
    if (strict) {
        # Check for large escaping objects
        result_size <- .estimate_object_size(result)
        if (!is.na(result_size) && result_size > escape_threshold) {
            warning(
                sprintf(
                    "arena(strict=TRUE): Large object escaping (%.1f MB > %.1f MB threshold)\n",
                    result_size / 1024 / 1024,
                    escape_threshold / 1024 / 1024
                ),
                "Consider reducing result size or setting strict=FALSE.",
                call. = FALSE
            )
        }
    }

    # Collect diagnostics if requested
    if (diagnostics) {
        rss_after <- rss_self()
        gc_after_info <- gc(reset = FALSE, full = FALSE)

        diag_info <- list(
            arena_id = arena_id,
            strict = strict,
            rss_before = rss_before,
            rss_after = rss_after,
            rss_delta = if (!is.na(rss_before) && !is.na(rss_after)) {
                rss_after - rss_before
            } else {
                NA_real_
            },
            result_size = .estimate_object_size(result),
            gc_before = gc_before,
            gc_after = gc_after_info
        )

        return(structure(
            list(
                result = result,
                diagnostics = diag_info
            ),
            class = "arena_result"
        ))
    }

    result
}

#' Check if Currently Inside an Arena
#'
#' Returns TRUE if the current execution context is within an arena() scope.
#'
#' @return Logical indicating whether we are in an arena scope.
#' @export
#' @examples
#' in_arena()
#' arena({
#'   in_arena()
#' })
in_arena <- function() {
    .arena_depth() > 0
}

#' Get Current Arena Depth
#'
#' Returns the nesting depth of arena scopes. Useful for debugging.
#'
#' @return Integer count of nested arena scopes (0 if not in an arena).
#' @export
#' @examples
#' arena_depth()
arena_depth <- function() {
    .arena_depth()
}

# ============================================================================
# Internal functions
# ============================================================================

#' Push a new arena onto the stack
#' @return Arena ID for matching pop
#' @keywords internal
#' @noRd
.arena_push <- function() {
    depth <- .arena_depth()
    new_id <- paste0("arena_", depth + 1L, "_", format(Sys.time(), "%H%M%S%OS3"))

    # Store in arena state
    .arena_state$stack <- c(.arena_state$stack, new_id)

    new_id
}

#' Pop an arena from the stack
#' @param arena_id The ID to verify we're popping the right arena
#' @param run_gc Whether to trigger GC after pop
#' @keywords internal
#' @noRd
.arena_pop <- function(arena_id, run_gc = FALSE) {
    stack <- .arena_state$stack

    if (length(stack) == 0) {
        warning("arena_pop called with empty stack", call. = FALSE)
        return(invisible(NULL))
    }

    # Verify we're popping the expected arena
    top <- stack[length(stack)]
    if (top != arena_id) {
        warning(
            sprintf("Arena mismatch: expected %s, found %s", arena_id, top),
            call. = FALSE
        )
    }

    # Pop the stack
    .arena_state$stack <- stack[-length(stack)]

    # Run GC if requested
    if (run_gc) {
        gc(reset = FALSE, full = FALSE)
    }

    invisible(NULL)
}

#' Get current arena depth
#' @return Integer depth
#' @keywords internal
#' @noRd
.arena_depth <- function() {
    length(.arena_state$stack %||% character(0))
}

#' Estimate object size safely
#' @param x Object to measure
#' @return Size in bytes or NA if cannot determine
#' @keywords internal
#' @noRd
.estimate_object_size <- function(x) {
    tryCatch({
        as.numeric(utils::object.size(x))
    }, error = function(e) {
        NA_real_
    })
}

#' Print an arena_result object
#'
#' @param x An \code{arena_result} object.
#' @param ... Additional arguments passed to \code{print}.
#' @return Returns \code{x} invisibly.
#' @export
#' @examples
#' \donttest{
#' info <- arena({ sum(1:10) }, diagnostics = TRUE)
#' print(info)
#' }
print.arena_result <- function(x, ...) {
    cat("<arena_result>\n")

    diag <- x$diagnostics
    if (!is.null(diag)) {
        cat("  Arena ID:", diag$arena_id, "\n")
        cat("  Strict mode:", diag$strict, "\n")

        if (!is.na(diag$rss_before)) {
            cat("  RSS before:", format(diag$rss_before, big.mark = ","), "bytes\n")
        }
        if (!is.na(diag$rss_after)) {
            cat("  RSS after:", format(diag$rss_after, big.mark = ","), "bytes\n")
        }
        if (!is.na(diag$rss_delta)) {
            cat("  RSS delta:", format(diag$rss_delta, big.mark = ","), "bytes\n")
        }
        if (!is.na(diag$result_size)) {
            cat("  Result size:", format(diag$result_size, big.mark = ","), "bytes\n")
        }
    }

    cat("\nResult:\n")
    print(x$result, ...)

    invisible(x)
}

# Tests for arena() - semantic scope for scratch memory

test_that("arena evaluates expression and returns result", {
    result <- arena({
        x <- 1:10
        sum(x)
    })

    expect_equal(result, 55)
})

test_that("arena evaluates in parent environment", {
    y <- 100
    result <- arena({
        y + 1
    })

    expect_equal(result, 101)
})

test_that("arena can modify parent environment variables via assign", {
    env <- environment()
    z <- 0
    arena({
        assign("z", 42, envir = env)
    })

    expect_equal(z, 42)
})

test_that("in_arena returns FALSE outside arena", {
    expect_false(in_arena())
})

test_that("in_arena returns TRUE inside arena", {
    result <- arena({
        in_arena()
    })

    expect_true(result)
})

test_that("arena_depth is 0 outside arena", {
    expect_equal(arena_depth(), 0)
})

test_that("arena supports nesting with correct depth", {
    result <- arena({
        depth1 <- arena_depth()
        depth2 <- arena({
            arena_depth()
        })
        c(depth1, depth2)
    })

    expect_equal(result[1], 1)  # Outer arena
    expect_equal(result[2], 2)  # Inner arena
})

test_that("arena with diagnostics returns arena_result", {
    result <- arena({
        sum(1:100)
    }, diagnostics = TRUE)

    expect_s3_class(result, "arena_result")
    expect_equal(result$result, 5050)
    expect_true(!is.null(result$diagnostics))
    expect_true(!is.null(result$diagnostics$arena_id))
})

test_that("arena diagnostics includes result size", {
    result <- arena({
        rnorm(100)
    }, diagnostics = TRUE)

    expect_true(!is.na(result$diagnostics$result_size))
    expect_true(result$diagnostics$result_size > 0)
})

test_that("arena strict mode warns on large escaping objects", {
    # Create a large object that escapes
    expect_warning(
        arena({
            # Create ~1MB+ object
            rnorm(200000)  # ~1.6MB
        }, strict = TRUE, escape_threshold = 1024 * 1024),
        "Large object escaping"
    )
})

test_that("arena strict mode does not warn on small objects", {
    expect_silent(
        arena({
            1:10  # Tiny object
        }, strict = TRUE)
    )
})

test_that("arena gc_after triggers garbage collection", {
    # Hard to test GC directly, but at least verify it doesn't error
    expect_no_error(
        arena({
            x <- rnorm(1000)
            sum(x)
        }, gc_after = TRUE)
    )
})

test_that("arena handles errors in expression", {
    expect_error(
        arena({
            stop("test error")
        }),
        "test error"
    )

    # Stack should be clean after error
    expect_equal(arena_depth(), 0)
})

test_that("arena handles warnings in expression", {
    expect_warning(
        result <- arena({
            warning("test warning")
            42
        }),
        "test warning"
    )

    expect_equal(result, 42)
})

test_that("arena can return NULL", {
    result <- arena({
        NULL
    })

    expect_null(result)
})

test_that("arena can return complex objects", {
    result <- arena({
        list(
            a = 1:10,
            b = matrix(1:6, nrow = 2),
            c = data.frame(x = 1:3, y = letters[1:3])
        )
    })

    expect_true(is.list(result))
    expect_equal(names(result), c("a", "b", "c"))
    expect_equal(result$a, 1:10)
})

test_that("print.arena_result works", {
    result <- arena({
        sum(1:10)
    }, diagnostics = TRUE)

    expect_output(print(result), "arena_result")
    expect_output(print(result), "Arena ID:")
    expect_output(print(result), "Result:")
})

test_that("arena stack is properly maintained across multiple calls", {
    # Run multiple arenas sequentially
    for (i in 1:5) {
        result <- arena({
            arena_depth()
        })
        expect_equal(result, 1)
    }

    # Stack should be empty after all
    expect_equal(arena_depth(), 0)
})

test_that("arena respects custom escape_threshold", {
    # Should warn with very low threshold
    expect_warning(
        arena({
            1:100  # ~400 bytes
        }, strict = TRUE, escape_threshold = 100),
        "Large object escaping"
    )

    # Should not warn with high threshold
    expect_silent(
        arena({
            rnorm(1000)  # ~8KB
        }, strict = TRUE, escape_threshold = 10 * 1024 * 1024)
    )
})

test_that("arena works with zero-length results", {
    result <- arena({
        integer(0)
    })

    expect_equal(result, integer(0))
})

test_that("arena diagnostics tracks RSS when available", {
    result <- arena({
        x <- rnorm(10000)
        sum(x)
    }, diagnostics = TRUE)

    # RSS tracking may not be available on all platforms
    # Just check the fields exist
    diag <- result$diagnostics
    expect_true("rss_before" %in% names(diag))
    expect_true("rss_after" %in% names(diag))
    expect_true("rss_delta" %in% names(diag))
})

test_that("nested arenas track depth correctly", {
    depths <- arena({
        d1 <- arena_depth()
        d2 <- arena({
            arena_depth()
        })
        d3 <- arena({
            arena({
                arena_depth()
            })
        })
        list(outer = d1, inner = d2, double_nested = d3)
    })

    expect_equal(depths$outer, 1)
    expect_equal(depths$inner, 2)
    expect_equal(depths$double_nested, 3)
})

# Copy-on-write / immutability enforcement for shard shared vectors
#
# NOTE: We enforce cow='deny' at the R level (replacement methods), not by
# erroring on DATAPTR(writable=TRUE) in C. Some base R algorithms request a
# writable pointer even for logically read-only operations; erroring there
# would create false positives and break normal reads.

.shard_cow_policy <- function(x) {
    pol <- attr(x, "shard_cow", exact = TRUE)
    if (is.null(pol) || !is.character(pol) || length(pol) != 1L) "allow" else pol
}

.shard_stop_cow_deny <- function() {
    stop("Attempted to mutate a shared input with cow='deny'.", call. = FALSE)
}

#' Print a Shared Vector
#'
#' Print method for \code{shard_shared_vector} objects. Drops the wrapper class
#' and delegates to the underlying R print method.
#'
#' @param x A \code{shard_shared_vector}.
#' @param ... Additional arguments passed to \code{print}.
#' @return The input \code{x}, invisibly.
#' @export
print.shard_shared_vector <- function(x, ...) {
    # Drop only the wrapper class and preserve any underlying class (e.g.,
    # Date, POSIXct) so printing stays natural.
    cls <- class(x)
    cls2 <- setdiff(cls, "shard_shared_vector")
    y <- x
    class(y) <- if (length(cls2)) cls2 else NULL
    print(y, ...)
    invisible(x)
}

#' Subset-assign a Shared Vector
#'
#' Replacement method for \code{shard_shared_vector}. Raises an error if the
#' copy-on-write policy is \code{"deny"}.
#'
#' @param x A \code{shard_shared_vector}.
#' @param ... Indices.
#' @param value Replacement value.
#' @return The modified object \code{x}.
#' @export
`[<-.shard_shared_vector` <- function(x, ..., value) {
    if (identical(.shard_cow_policy(x), "deny")) .shard_stop_cow_deny()

    # NextMethod() is unreliable for primitive replacement functions when the
    # object only has a single custom class. Temporarily drop the class to force
    # base replacement semantics (while keeping shard_* attributes).
    cls <- class(x)
    class(x) <- NULL
    x <- .Primitive("[<-")(x, ..., value = value)
    class(x) <- cls
    x
}

#' Double-bracket Subset-assign a Shared Vector
#'
#' Replacement method for \code{shard_shared_vector}. Raises an error if the
#' copy-on-write policy is \code{"deny"}.
#'
#' @param x A \code{shard_shared_vector}.
#' @param ... Indices.
#' @param value Replacement value.
#' @return The modified object \code{x}.
#' @export
`[[<-.shard_shared_vector` <- function(x, ..., value) {
    if (identical(.shard_cow_policy(x), "deny")) .shard_stop_cow_deny()
    cls <- class(x)
    class(x) <- NULL
    x <- .Primitive("[[<-")(x, ..., value = value)
    class(x) <- cls
    x
}

# Attribute mutations are also considered violations under cow='deny'.

#' Set an Attribute on a Shared Vector
#'
#' Raises an error if the copy-on-write policy is \code{"deny"}.
#'
#' @param x A \code{shard_shared_vector}.
#' @param which Attribute name.
#' @param value Attribute value.
#' @return The modified object \code{x}.
#' @method attr<- shard_shared_vector
#' @export
`attr<-.shard_shared_vector` <- function(x, which, value) {
    if (identical(.shard_cow_policy(x), "deny")) .shard_stop_cow_deny()
    cls <- class(x)
    class(x) <- NULL
    x <- .Primitive("attr<-")(x, which, value)
    class(x) <- cls
    x
}

#' Set Attributes on a Shared Vector
#'
#' Raises an error if the copy-on-write policy is \code{"deny"}.
#'
#' @param x A \code{shard_shared_vector}.
#' @param value Named list of attributes.
#' @return The modified object \code{x}.
#' @method attributes<- shard_shared_vector
#' @export
`attributes<-.shard_shared_vector` <- function(x, value) {
    if (identical(.shard_cow_policy(x), "deny")) .shard_stop_cow_deny()
    cls <- class(x)
    class(x) <- NULL
    x <- .Primitive("attributes<-")(x, value)
    class(x) <- cls
    x
}

#' Set Names on a Shared Vector
#'
#' Raises an error if the copy-on-write policy is \code{"deny"}.
#'
#' @param x A \code{shard_shared_vector}.
#' @param value Character vector of names.
#' @return The modified object \code{x}.
#' @export
`names<-.shard_shared_vector` <- function(x, value) {
    if (identical(.shard_cow_policy(x), "deny")) .shard_stop_cow_deny()
    cls <- class(x)
    class(x) <- NULL
    x <- .Primitive("names<-")(x, value)
    class(x) <- cls
    x
}

#' Set dim on a Shared Vector
#'
#' Raises an error if the copy-on-write policy is \code{"deny"}.
#'
#' @param x A \code{shard_shared_vector}.
#' @param value Integer vector of dimensions.
#' @return The modified object \code{x}.
#' @export
`dim<-.shard_shared_vector` <- function(x, value) {
    if (identical(.shard_cow_policy(x), "deny")) .shard_stop_cow_deny()
    cls <- class(x)
    class(x) <- NULL
    x <- .Primitive("dim<-")(x, value)
    class(x) <- cls
    x
}

#' Set dimnames on a Shared Vector
#'
#' Raises an error if the copy-on-write policy is \code{"deny"}.
#'
#' @param x A \code{shard_shared_vector}.
#' @param value List of dimnames.
#' @return The modified object \code{x}.
#' @export
`dimnames<-.shard_shared_vector` <- function(x, value) {
    if (identical(.shard_cow_policy(x), "deny")) .shard_stop_cow_deny()
    cls <- class(x)
    class(x) <- NULL
    x <- .Primitive("dimnames<-")(x, value)
    class(x) <- cls
    x
}

# Benchmark Table Templates

Blog-post-ready benchmark table templates for shard performance documentation.

---

## Template A: Scaling (time vs workers)

```markdown
### Scaling benchmark: {suite_name}

**Dataset:** {dataset_id}
**Dims:** T={T}, P={P}, V={V}
**Machine:** {cpu}, {ram}GB, {os}, R {r_ver}, BLAS: {blas}

| Backend | Workers | Time (s) ↓ | Speedup ↑ | Peak RSS (GB) ↓ | End RSS (GB) ↓ | RSS drift (GB) ↓ | Hidden copies (MB) ↓ | Recycles |
|---|---:|---:|---:|---:|---:|---:|---:|---:|
| serial | 1 | __ | 1.00× | __ | __ | __ | __ | 0 |
| foreach + doParallel | 2 | __ | __× | __ | __ | __ | __ | 0 |
| foreach + doParallel | 4 | __ | __× | __ | __ | __ | __ | 0 |
| foreach + doParallel | 8 | __ | __× | __ | __ | __ | __ | 0 |
| shard | 2 | __ | __× | __ | __ | __ | __ | __ |
| shard | 4 | __ | __× | __ | __ | __ | __ | __ |
| shard | 8 | __ | __× | __ | __ | __ | __ | __ |
```

---

## Template B: Long-run stability (drift stress test)

```markdown
### Memory stability benchmark: {suite_name}

**Workload:** {task_count} tasks, {workers} workers
**Goal:** demonstrate flat RSS + deterministic cleanup

| Backend | Workers | Tasks | Peak RSS (GB) ↓ | End RSS (GB) ↓ | Drift (GB) ↓ | Failures | Recycles |
|---|---:|---:|---:|---:|---:|---:|---:|
| foreach + doParallel | __ | __ | __ | __ | __ | __ | 0 |
| future.apply | __ | __ | __ | __ | __ | __ | 0 |
| shard | __ | __ | __ | __ | __ | __ | __ |
```

---

## Suggested "first blog post" outline

If you want the post to basically write itself from the tables:

```markdown
# shard: Deterministic Parallelism in R Without Memory Surprises

## The problem (with one painful chart)
- peak RSS grows over time with long-running workers
- hidden copies multiply inputs

## shard's contract (3 bullets)
- shared immutable inputs
- explicit output buffers
- supervised workers with recycling

## Benchmarks
### 1) Many regressions (QR-once)
[Template A table]
- time scaling plot
- peak RSS plot

### 2) 10k-task drift test
[Template B table]
- RSS drift over task index plot

## Diagnostics: proving "no hidden copies"
- materialized_mb
- recycle events

## When to use shard (and when not to)
## Next: doShard + distributed v2
```

---
title: "Getting Started with shard"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with shard}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)
```

```{r load-shard}
library(shard)
```

R's parallel tools make it easy to fan out work, but they leave you to manage
the hard parts yourself: duplicated memory, runaway workers, invisible
copy-on-write. shard handles all of that so you can focus on the computation.

The core idea is simple: **share inputs once, write outputs to a buffer, let
shard supervise the workers.**

## A first example

Suppose you have a large matrix and want to compute column means in parallel.
With shard, you share the matrix, allocate an output buffer, and map over
column indices:

```{r first-example}
set.seed(42)
X <- matrix(rnorm(5000), nrow = 100, ncol = 50)

# Share the matrix (zero-copy for workers)
X_shared <- share(X)

# Allocate an output buffer
out <- buffer("double", dim = ncol(X))

# Define column shards and run
blocks <- shards(ncol(X), workers = 2)
run <- shard_map(
  blocks,
  borrow = list(X = X_shared),
  out    = list(out = out),
  workers = 2,
  fun = function(shard, X, out) {
    for (j in shard$idx) {
      out[j] <- mean(X[, j])
    }
  }
)

# Read results from the buffer
result <- out[]
head(result)
```

No serialization of the full matrix per worker. No list of return values to
reassemble. The workers wrote directly into `out`.

## The three core objects

shard's workflow revolves around three things:

| Object | Constructor | Purpose |
|:-------|:------------|:--------|
| Shared input | `share()` | Immutable, zero-copy data visible to all workers |
| Output buffer | `buffer()` | Writable shared memory that workers fill in |
| Shard descriptor | `shards()` | Index ranges that partition the work |

### Sharing inputs

`share()` places an R object into shared memory. Workers attach to the same
segment instead of receiving a copy:

```{r share-input}
X_shared <- share(X)
is_shared(X_shared)
shared_info(X_shared)
```

Shared objects are **read-only**. Any attempt to modify them in a worker
raises an error, which prevents silent copy-on-write bugs.

### Output buffers

`buffer()` creates typed shared memory that workers write to using standard
R indexing:

```{r buffer-demo}
buf <- buffer("double", dim = c(10, 5))
buf[1:5, 1] <- rnorm(5)
buf[6:10, 1] <- rnorm(5)
buf[, 1]
```

Buffers support `"double"`, `"integer"`, `"logical"`, and `"raw"` types.
For matrices and arrays, pass a `dim` vector:

```{r buffer-types}
int_buf <- buffer("integer", dim = 100)
mat_buf <- buffer("double", dim = c(50, 20))
```

### Shard descriptors

`shards()` partitions a range of indices into chunks for parallel execution.
It auto-tunes the block size based on the number of workers:

```{r shards-demo}
blocks <- shards(1000, workers = 4)
blocks
```

Each shard carries an `idx` field with its assigned indices:

```{r shard-indices}
blocks[[1]]$idx[1:10]  # first 10 indices of shard 1
```

## Running shard_map()

`shard_map()` is the engine. It dispatches shards to a supervised worker pool,
passes shared inputs, and collects diagnostics:

```{r shard-map-example}
set.seed(1)
X <- matrix(rnorm(2000), nrow = 100, ncol = 20)
X_shared <- share(X)
col_sds <- buffer("double", dim = ncol(X))

blocks <- shards(ncol(X), workers = 2)
run <- shard_map(
  blocks,
  borrow  = list(X = X_shared),
  out     = list(col_sds = col_sds),
  workers = 2,
  fun = function(shard, X, col_sds) {
    for (j in shard$idx) {
      col_sds[j] <- sd(X[, j])
    }
  }
)

# Results are already in the buffer
sd_values <- col_sds[]

# Verify against base R
all.equal(sd_values, apply(X, 2, sd))
```

### What if workers return values?

If your function returns a value (instead of writing to a buffer), shard
gathers the results:

```{r return-values}
blocks <- shards(10, workers = 2)
run <- shard_map(
  blocks,
  workers = 2,
  fun = function(shard) {
    sum(shard$idx)
  }
)
results(run)
```

Buffers are preferred for large outputs because they avoid serializing results
back to the main process. Use return values for small summaries.

## Convenience wrappers

For common patterns, shard provides wrappers that handle sharing, sharding,
and buffering automatically.

### Column-wise apply

`shard_apply_matrix()` applies a scalar function over each column of a
matrix:

```{r apply-matrix}
set.seed(1)
X <- matrix(rnorm(2000), nrow = 100, ncol = 20)
y <- rnorm(100)

# Correlate each column of X with y
cors <- shard_apply_matrix(
  X,
  MARGIN = 2,
  FUN = function(v, y) cor(v, y),
  VARS = list(y = y),
  workers = 2
)
head(cors)
```

The matrix is auto-shared, columns are dispatched as shards, and results
are collected into a vector.

### List lapply

`shard_lapply_shared()` is a parallel `lapply` with automatic sharing of
large list elements:

```{r lapply-shared}
chunks <- lapply(1:10, function(i) rnorm(100))

means <- shard_lapply_shared(
  chunks,
  FUN = mean,
  workers = 2
)
unlist(means)
```

## Diagnostics

Every `shard_map()` call records timing, memory, and worker statistics.
Use `report()` to inspect them:

```{r diagnostics}
report(result = run)
```

For focused views:

- `mem_report(run)` -- peak and baseline RSS per worker
- `copy_report(run)` -- bytes transferred through buffers
- `task_report(run)` -- per-chunk execution times and retry counts

## Worker pool management

By default, `shard_map()` creates a worker pool on first use and reuses it.
You can also manage the pool explicitly:

```{r pool-management, eval = FALSE}
# Create a pool with 4 workers and a 1GB memory cap
pool_create(n = 4, rss_limit = "1GB")

# Check pool health
pool_status()

# Run multiple shard_map() calls (reuses the same pool)
run1 <- shard_map(shards(1000), workers = 4, fun = function(s) sum(s$idx))
run2 <- shard_map(shards(500),  workers = 4, fun = function(s) mean(s$idx))

# Shut down workers when done
pool_stop()
```

Workers are supervised: if a worker's memory usage drifts beyond the
threshold, shard recycles it automatically.

## Copy-on-write protection

Shared inputs are immutable by default (`cow = "deny"`). This prevents a
common class of parallel bugs where a worker accidentally modifies shared
data, triggering a silent copy:

```{r cow-deny, eval = FALSE}
shard_map(
  shards(10),
  borrow = list(X = share(matrix(1:100, 10, 10))),
  workers = 2,
  cow = "deny",
  fun = function(shard, X) {
    X[1, 1] <- 999  # Error: mutation denied
  }
)
```

You can relax this with `cow = "audit"` (detect and report mutations) or
`cow = "allow"` (permit copy-on-write with tracking). See `?shard_map` for
details.

## Clean up

```{r cleanup, include = FALSE}
try(pool_stop(), silent = TRUE)
```

When you are done, stop the pool to release worker processes:

```{r cleanup-show, eval = FALSE}
pool_stop()
```

## Next steps

- `?shard_map` -- full reference for the parallel engine
- `?share` -- sharing options and backing types
- `?buffer` -- buffer types and matrix/array support
- `?report` -- diagnostic reports and recommendations
- `?shard_apply_matrix` -- column-wise parallel apply
- `?pool_create` -- pool configuration and memory limits

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_stream.R
\name{stream_filter}
\alias{stream_filter}
\title{Stream-filter a dataset/row-groups into a new partitioned dataset}
\usage{
stream_filter(x, predicate, path = NULL, ...)
}
\arguments{
\item{x}{A \code{shard_row_groups} or \code{shard_dataset} handle.}

\item{predicate}{Function \code{(chunk, ...) -> logical} row mask (length == nrow(chunk)).}

\item{path}{Output directory. If NULL, a temp dir is created.}

\item{...}{Passed to \code{predicate()}.}
}
\value{
A \code{shard_dataset} handle pointing to the filtered partitions.
}
\description{
Reads each partition, filters rows, and writes a new partitioned dataset.
Output is written as one partition per input partition (empty partitions are
allowed). This avoids materializing all results.
}
\examples{
\donttest{
s <- schema(x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L, data.frame(x = c(1.0, 2.0, 3.0)))
rg <- table_finalize(sink)
filtered <- stream_filter(rg, predicate = function(chunk) chunk$x > 1.5)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pool.R
\name{pool_dispatch}
\alias{pool_dispatch}
\title{Dispatch Task to Worker}
\usage{
pool_dispatch(
  worker_id,
  expr,
  envir = parent.frame(),
  pool = NULL,
  timeout = 3600
)
}
\arguments{
\item{worker_id}{Integer. Worker to dispatch to.}

\item{expr}{Expression to evaluate.}

\item{envir}{Environment containing variables needed by expr.}

\item{pool}{A \code{shard_pool} object. If NULL, uses the current pool.}

\item{timeout}{Numeric. Seconds to wait for result (default 3600).}
}
\value{
The result of evaluating expr in the worker.
}
\description{
Sends a task to a specific worker and waits for the result.
}
\examples{
\donttest{
p <- pool_create(2)
pool_dispatch(1, quote(1 + 1), pool = p)
pool_stop(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/worker.R
\name{print.shard_worker}
\alias{print.shard_worker}
\title{Print a shard_worker Object}
\usage{
\method{print}{shard_worker}(x, ...)
}
\arguments{
\item{x}{A \code{shard_worker} object.}

\item{...}{Further arguments (ignored).}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a shard_worker Object
}
\examples{
\donttest{
p <- pool_create(1)
print(p$workers[[1]])
pool_stop(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_stream.R
\name{stream_top_k}
\alias{stream_top_k}
\title{Stream top-k rows by a numeric column}
\usage{
stream_top_k(x, col, k = 10L, decreasing = TRUE, na_drop = TRUE)
}
\arguments{
\item{x}{A \code{shard_row_groups} or \code{shard_dataset} handle.}

\item{col}{Column name to rank by.}

\item{k}{Number of rows to keep.}

\item{decreasing}{Logical; TRUE for largest values (default TRUE).}

\item{na_drop}{Logical; drop rows where \code{col} is NA (default TRUE).}
}
\value{
A data.frame (or tibble if the \code{tibble} package is installed)
with at most \code{k} rows ordered by \code{col}.
}
\description{
Finds the top \code{k} rows by \code{col} without collecting the full dataset.
}
\details{
For native-encoded partitions, this selects candidate rows using the numeric
column without decoding strings, then decodes only the chosen rows for the
returned result.
}
\examples{
\donttest{
s <- schema(x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L, data.frame(x = c(3.0, 1.0, 2.0)))
rg <- table_finalize(sink)
stream_top_k(rg, "x", k = 2L)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{[.shard_buffer}
\alias{[.shard_buffer}
\title{Extract Buffer Elements}
\usage{
\method{[}{shard_buffer}(x, i, j, ..., drop = TRUE)
}
\arguments{
\item{x}{A shard_buffer object.}

\item{i}{Index or indices.}

\item{j}{Optional second index (for matrices).}

\item{...}{Additional indices (for arrays).}

\item{drop}{Whether to drop dimensions.}
}
\value{
A vector or array of values read from the buffer.
}
\description{
Extract Buffer Elements
}
\examples{
\donttest{
buf <- buffer("double", dim = 10)
buf[1:5] <- 1:5
buf[1:3]
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{buffer}
\alias{buffer}
\title{Shared Memory Buffers}
\usage{
buffer(
  type = c("double", "integer", "logical", "raw"),
  dim,
  init = NULL,
  backing = c("auto", "mmap", "shm")
)
}
\arguments{
\item{type}{Character. Data type: "double" (default), "integer", "logical",
or "raw".}

\item{dim}{Integer vector. Dimensions of the buffer. For a vector, specify
the length. For a matrix, specify \code{c(nrow, ncol)}. For arrays,
specify all dimensions.}

\item{init}{Initial value to fill the buffer. Default is type-appropriate
zero (\code{0}, \code{0L}, \code{FALSE}, or \code{raw(0)}).}

\item{backing}{Backing type for shared memory: "auto" (default), "mmap",
or "shm".}
}
\value{
An S3 object of class "shard_buffer" that supports:
\itemize{
\item Slice assignment: \code{buf[idx] <- values}
\item Slice reading: \code{buf[idx]}
\item Full extraction: \code{buf[]}
\item Conversion to R vector: \code{as.vector(buf)}, \code{as.double(buf)}, etc.
}
}
\description{
Create typed writable output buffers backed by shared memory
for cross-process writes during parallel execution.

Creates a typed output buffer backed by shared memory that can be written
to by parallel workers using slice assignment.
}
\details{
Buffers provide an explicit output mechanism for \code{\link{shard_map}}.
Instead of returning results from workers (which requires serialization
and memory copying), workers write directly to shared buffers.

Supported types:
\itemize{
\item \code{"double"}: 8-byte floating point (default)
\item \code{"integer"}: 4-byte signed integer
\item \code{"logical"}: 4-byte logical (stored as integer)
\item \code{"raw"}: 1-byte raw data
}

Buffers support slice assignment using standard R indexing:
\code{buf[1:100] <- values}
}
\examples{
\donttest{
out <- buffer("double", dim = 100)
out[1:10] <- rnorm(10)
result <- out[]
}
}
\seealso{
\code{\link{segment_create}} for low-level segment operations,
\code{\link{share}} for read-only shared inputs
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinity.R
\name{pin_workers}
\alias{pin_workers}
\title{Pin shard workers to CPU cores}
\usage{
pin_workers(pool = NULL, strategy = c("spread", "compact"), cores = NULL)
}
\arguments{
\item{pool}{Optional shard_pool. Defaults to current pool.}

\item{strategy}{"spread" assigns worker i -> core i mod ncores. "compact"
assigns workers to the first cores.}

\item{cores}{Optional integer vector of available cores (0-based). If NULL,
uses 0:(detectCores()-1).}
}
\value{
Invisibly, a logical vector per worker indicating success.
}
\description{
Best-effort worker pinning to improve cache locality and reduce cross-core
migration. Currently supported on Linux only.
}
\examples{
\donttest{
affinity_supported()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shards.R
\name{print.shard_descriptor}
\alias{print.shard_descriptor}
\title{Print a shard_descriptor Object}
\usage{
\method{print}{shard_descriptor}(x, ...)
}
\arguments{
\item{x}{A \code{shard_descriptor} object.}

\item{...}{Further arguments (ignored).}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a shard_descriptor Object
}
\examples{
sh <- shards(100, block_size = 25)
print(sh)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arena.R
\name{arena_depth}
\alias{arena_depth}
\title{Get Current Arena Depth}
\usage{
arena_depth()
}
\value{
Integer count of nested arena scopes (0 if not in an arena).
}
\description{
Returns the nesting depth of arena scopes. Useful for debugging.
}
\examples{
arena_depth()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{buffer_diagnostics}
\alias{buffer_diagnostics}
\title{Buffer Diagnostics}
\usage{
buffer_diagnostics()
}
\value{
A list with elements \code{writes} (integer count) and \code{bytes}
(total bytes written) accumulated in the current process.
}
\description{
Returns per-process counters for shard buffer writes. shard_map uses these
internally to report write volume/operations in copy_report().
}
\examples{
buffer_diagnostics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_write.shard_table_sink}
\alias{table_write.shard_table_sink}
\title{Write a shard's row-group output}
\usage{
\method{table_write}{shard_table_sink}(target, rows_or_shard_id, data, ...)
}
\arguments{
\item{target}{A \code{shard_table_sink}.}

\item{rows_or_shard_id}{Integer shard id used to name the row-group file.}

\item{data}{A data.frame matching the sink schema.}

\item{...}{Reserved for future extensions.}
}
\value{
NULL (invisibly).
}
\description{
Write a shard's row-group output
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_stream.R
\name{stream_map}
\alias{stream_map}
\alias{stream_map.shard_row_groups}
\alias{stream_map.shard_dataset}
\title{Stream over row-groups/datasets and map}
\usage{
stream_map(x, f, ...)

\method{stream_map}{shard_row_groups}(x, f, ...)

\method{stream_map}{shard_dataset}(x, f, ...)
}
\arguments{
\item{x}{A \code{shard_row_groups} or \code{shard_dataset} handle.}

\item{f}{Function \code{(chunk, ...) -> value}.}

\item{...}{Passed to \code{f()}.}
}
\value{
A list of per-partition values, one element per row-group file.
}
\description{
Applies \code{f()} to each partition and returns the list of per-partition results.
This is still much cheaper than collecting the full dataset when \code{f()} returns
a small summary per partition.
}
\examples{
\donttest{
s <- schema(x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L, data.frame(x = rnorm(5)))
rg <- table_finalize(sink)
nrows <- stream_map(rg, nrow)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapter.R
\name{shard_register_adapter}
\alias{shard_register_adapter}
\title{Register an Adapter for Class-Specific Traversal}
\usage{
shard_register_adapter(class, adapter)
}
\arguments{
\item{class}{A character string naming the class to register the adapter for.}

\item{adapter}{A list containing:
\describe{
\item{class}{Character string matching the \code{class} parameter.}
\item{children}{Function taking an object and returning a named list of
child objects to traverse.}
\item{replace}{Function taking the original object and a named list of
(potentially shared) children, returning a reconstructed object.}
\item{path_prefix}{Optional character string prefix for child paths in
the sharing plan (default: class name).}
}}
}
\value{
Invisibly returns the previous adapter for this class (if any),
or NULL if no adapter was registered.
}
\description{
Registers a custom adapter for a specific class. When deep sharing encounters
an object of this class, it will use the adapter's \code{children()} function
to extract shareable components instead of generic traversal.
}
\examples{
shard_list_adapters()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{shared_segment}
\alias{shared_segment}
\title{Get the underlying segment from a shared vector}
\usage{
shared_segment(x)
}
\arguments{
\item{x}{A shard ALTREP vector}
}
\value{
A \code{shard_segment} S3 object wrapping the underlying segment
}
\description{
Get the underlying segment from a shared vector
}
\examples{
\donttest{
x <- as_shared(1:100)
shared_segment(x)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ergonomics.R
\name{print.shard_apply_policy}
\alias{print.shard_apply_policy}
\title{Print a shard_apply_policy Object}
\usage{
\method{print}{shard_apply_policy}(x, ...)
}
\arguments{
\item{x}{A \code{shard_apply_policy} object.}

\item{...}{Ignored.}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a shard_apply_policy Object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{copy_report}
\alias{copy_report}
\title{Data Copy Report}
\usage{
copy_report(result = NULL)
}
\arguments{
\item{result}{Optional. A \code{shard_result} object to extract copy stats from.}
}
\value{
An S3 object of class \code{shard_report} with type \code{"copy"}
containing:
\itemize{
\item \code{type}: "copy"
\item \code{timestamp}: When the report was generated
\item \code{borrow_exports}: Number of borrowed input exports
\item \code{borrow_bytes}: Total bytes in borrowed inputs
\item \code{result_imports}: Number of result imports
\item \code{result_bytes}: Estimated bytes in results
\item \code{buffer_writes}: Number of buffer write operations
\item \code{buffer_bytes}: Total bytes written to buffers
}
}
\description{
Generates a report of data transfer and copy statistics during parallel
execution.
}
\examples{
\donttest{
res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
pool_stop()
copy_report(res)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{segment_report}
\alias{segment_report}
\title{Shared Memory Segment Report}
\usage{
segment_report()
}
\value{
An S3 object of class \code{shard_report} with type \code{"segment"}
containing:
\itemize{
\item \code{type}: "segment"
\item \code{timestamp}: When the report was generated
\item \code{n_segments}: Number of tracked segments
\item \code{total_bytes}: Total bytes across all segments
\item \code{segments}: List of segment details
}
}
\description{
Generates a report of active shared memory segments in the current session.
}
\details{
This function reports on segments that are currently accessible. Note that
segments are automatically cleaned up when their R objects are garbage
collected, so this only shows segments with live references.
}
\examples{
\donttest{
segment_report()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shard_reduce.R
\name{print.shard_reduce_result}
\alias{print.shard_reduce_result}
\title{Print a shard_reduce_result Object}
\usage{
\method{print}{shard_reduce_result}(x, ...)
}
\arguments{
\item{x}{A \code{shard_reduce_result} object.}

\item{...}{Further arguments (ignored).}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a shard_reduce_result Object
}
\examples{
\donttest{
res <- shard_reduce(4L, map = function(s) sum(s$idx),
  combine = `+`, init = 0, workers = 2)
pool_stop()
print(res)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dispatch.R
\name{dispatch_chunks}
\alias{dispatch_chunks}
\title{Dispatch Chunks to Worker Pool}
\usage{
dispatch_chunks(
  chunks,
  fun,
  ...,
  pool = NULL,
  health_check_interval = 10L,
  max_retries = 3L,
  timeout = 3600,
  scheduler_policy = NULL,
  on_result = NULL,
  store_results = TRUE,
  retain_chunks = TRUE
)
}
\arguments{
\item{chunks}{List of chunk descriptors. Each chunk will be passed to \code{fun}.}

\item{fun}{Function to execute. Receives (chunk, ...) as arguments.}

\item{...}{Additional arguments passed to \code{fun}.}

\item{pool}{A \code{shard_pool} object. If NULL, uses the current pool.}

\item{health_check_interval}{Integer. Check pool health every N chunks (default 10).}

\item{max_retries}{Integer. Maximum retries per chunk before permanent failure (default 3).}

\item{timeout}{Numeric. Seconds to wait for each chunk (default 3600).}

\item{scheduler_policy}{Optional list of scheduling hints (advanced). Currently:
\itemize{
\item \code{max_huge_concurrency}: cap concurrent chunks with \code{footprint_class=="huge"}.
}}

\item{on_result}{Optional callback (advanced). If provided, called on the
master process as \code{on_result(tag, value, worker_id)} for each successful
chunk completion. Used by \code{\link[=shard_reduce]{shard_reduce()}} to stream reductions.}

\item{store_results}{Logical (advanced). If FALSE, successful chunk values are
not retained in the returned \code{results} list (streaming use cases).}

\item{retain_chunks}{Logical (advanced). If FALSE, completed chunk descriptors
are stored minimally (avoids retaining large shard lists in memory).}
}
\value{
A \code{shard_dispatch_result} object with results and diagnostics.
}
\description{
Executes a function over chunks using the worker pool with supervision.
Handles worker death and recycling transparently by requeuing failed chunks.
}
\examples{
\donttest{
pool_create(2)
chunks <- list(list(id = 1L, x = 1), list(id = 2L, x = 2))
result <- dispatch_chunks(chunks, function(chunk) chunk$x * 2, pool = pool_get())
pool_stop()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{buffer_open}
\alias{buffer_open}
\title{Open an Existing Buffer}
\usage{
buffer_open(path, type, dim, backing = c("mmap", "shm"), readonly = FALSE)
}
\arguments{
\item{path}{Path or shm name of the buffer's segment.}

\item{type}{Character. Data type of the buffer.}

\item{dim}{Integer vector. Dimensions of the buffer.}

\item{backing}{Backing type: "mmap" or "shm".}

\item{readonly}{Logical. Open as read-only? Default FALSE for workers.}
}
\value{
A \code{shard_buffer} object attached to the existing segment.
}
\description{
Opens a shared memory buffer that was created in another process.
Used by workers to attach to the parent's output buffer.
}
\examples{
\donttest{
buf <- buffer("double", dim = 10)
path <- buffer_path(buf)
buf2 <- buffer_open(path, type = "double", dim = 10, backing = "mmap")
buffer_close(buf2, unlink = FALSE)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shard_reduce.R
\name{shard_reduce}
\alias{shard_reduce}
\title{Streaming Reductions over Shards}
\usage{
shard_reduce(
  shards,
  map,
  combine,
  init,
  borrow = list(),
  out = list(),
  workers = NULL,
  chunk_size = 1L,
  profile = c("default", "memory", "speed"),
  mem_cap = "2GB",
  recycle = TRUE,
  cow = c("deny", "audit", "allow"),
  seed = NULL,
  diagnostics = TRUE,
  packages = NULL,
  init_expr = NULL,
  timeout = 3600,
  max_retries = 3L,
  health_check_interval = 10L
)
}
\arguments{
\item{shards}{A \code{shard_descriptor} from \code{\link[=shards]{shards()}}, or an integer N.}

\item{map}{Function executed per shard. Receives shard descriptor as first
argument, followed by borrowed inputs and outputs.}

\item{combine}{Function \verb{(acc, value) -> acc} used to combine results. Should
be associative for deterministic behavior under chunking.}

\item{init}{Initial accumulator value.}

\item{borrow}{Named list of shared inputs (same semantics as \code{\link[=shard_map]{shard_map()}}).}

\item{out}{Named list of output buffers/sinks (same semantics as \code{\link[=shard_map]{shard_map()}}).}

\item{workers}{Number of worker processes.}

\item{chunk_size}{Shards to batch per worker dispatch (default 1).}

\item{profile}{Execution profile (same semantics as \code{\link[=shard_map]{shard_map()}}).}

\item{mem_cap}{Memory cap per worker (same semantics as \code{\link[=shard_map]{shard_map()}}).}

\item{recycle}{Worker recycling policy (same semantics as \code{\link[=shard_map]{shard_map()}}).}

\item{cow}{Copy-on-write policy for borrowed inputs (same semantics as \code{\link[=shard_map]{shard_map()}}).}

\item{seed}{RNG seed for reproducibility.}

\item{diagnostics}{Logical; collect diagnostics (default TRUE).}

\item{packages}{Additional packages to load in workers.}

\item{init_expr}{Expression to evaluate in each worker on startup.}

\item{timeout}{Seconds to wait for each chunk.}

\item{max_retries}{Maximum retries per chunk.}

\item{health_check_interval}{Check worker health every N completions.}
}
\value{
A \code{shard_reduce_result} with fields:
\itemize{
\item \code{value}: final accumulator
\item \code{failures}: any permanently failed chunks
\item \code{diagnostics}: run telemetry including reduction stats
\item \code{queue_status}, \code{pool_stats}
}
}
\description{
Reduce shard results without gathering all per-shard returns on the master.

\code{shard_reduce()} executes \code{map()} over shards in parallel and combines results
using an associative \code{combine()} function. Unlike \code{shard_map()}, it does not
accumulate all per-shard results on the master; it streams partials as chunks
complete.
}
\details{
For performance and memory efficiency, reduction is performed in two stages:
\enumerate{
\item per-chunk partial reduction inside each worker, and
\item streaming combine of partials on the master.
}
}
\examples{
\donttest{
res <- shard_reduce(
  100L,
  map = function(s) sum(s$idx),
  combine = function(acc, x) acc + x,
  init = 0,
  workers = 2
)
pool_stop()
res$value
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{row_layout}
\alias{row_layout}
\title{Row layout for fixed-row table outputs}
\usage{
row_layout(shards, rows_per_shard)
}
\arguments{
\item{shards}{A \code{shard_descriptor}.}

\item{rows_per_shard}{Either a scalar integer or a function(shard)->integer.}
}
\value{
A named list mapping shard id (character) to an \code{idx_range(start, end)}.
}
\description{
Computes disjoint row ranges for each shard via prefix-sum, enabling lock-free
writes where each shard writes to a unique region.
}
\examples{
\donttest{
sh <- shards(100, block_size = 25)
layout <- row_layout(sh, rows_per_shard = 25L)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_write}
\alias{segment_write}
\title{Write data to a segment}
\usage{
segment_write(x, data, offset = 0)
}
\arguments{
\item{x}{A shard_segment object}

\item{data}{Data to write (raw, numeric, integer, or logical vector)}

\item{offset}{Byte offset to start writing (0-based)}
}
\value{
Number of bytes written, invisibly.
}
\description{
Write data to a segment
}
\examples{
\donttest{
seg <- segment_create(1024)
segment_write(seg, as.integer(1:10), offset = 0)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{length.shard_buffer}
\alias{length.shard_buffer}
\title{Length of a Shared Memory Buffer}
\usage{
\method{length}{shard_buffer}(x)
}
\arguments{
\item{x}{A \code{shard_buffer} object.}
}
\value{
An integer scalar giving the total number of elements.
}
\description{
Length of a Shared Memory Buffer
}
\examples{
\donttest{
buf <- buffer("double", dim = 20)
length(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_sink}
\alias{table_sink}
\title{Create a table sink for row-group or partitioned outputs}
\usage{
table_sink(
  schema,
  mode = c("row_groups", "partitioned"),
  path = NULL,
  format = c("auto", "rds", "native")
)
}
\arguments{
\item{schema}{A \code{shard_schema}. If NULL, a schema-less sink is created (RDS
format only). This is primarily intended for doShard/foreach compatibility
where output schemas may not be known in advance.}

\item{mode}{\code{"row_groups"} (temp, managed) or \code{"partitioned"} (persistent path).}

\item{path}{Directory to write row-group files. If NULL, a temp dir is created.}

\item{format}{Storage format for partitions: \code{"rds"} (data.frame RDS),
\code{"native"} (columnar encoding with string offsets+bytes), or \code{"auto"}
(selects \code{"native"} if the schema contains \code{string_col()}; otherwise \code{"rds"}).}
}
\value{
A \code{shard_table_sink} object.
}
\description{
A table sink supports variable-sized outputs without returning large
data.frames to the master. Each shard writes a separate row-group file.
}
\details{
v1.1 implementation notes:
\itemize{
\item Storage format is per-shard RDS (portable, CRAN-friendly).
\item This guarantees bounded master memory during execution; final collection
may still be large if you materialize.
}
}
\examples{
\donttest{
s <- schema(x = float64(), label = string_col())
sink <- table_sink(s, mode = "row_groups")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{share}
\alias{share}
\title{Zero-Copy Shared Objects}
\usage{
share(
  x,
  backing = c("auto", "mmap", "shm"),
  readonly = TRUE,
  name = NULL,
  deep = FALSE,
  min_bytes = 64 * 1024 * 1024,
  cycle = c("error", "skip"),
  mode = c("balanced", "strict")
)
}
\arguments{
\item{x}{An R object to share. Supports vectors, matrices, arrays, lists,
data frames, and any object that can be serialized with \code{serialize()}.}

\item{backing}{Backing type: "auto" (default), "mmap", or "shm".
\itemize{
\item \code{"auto"}: Let the system choose the best option.
\item \code{"mmap"}: File-backed memory mapping. Most portable.
\item \code{"shm"}: POSIX shared memory or Windows named mapping.
}}

\item{readonly}{Logical. If TRUE (default), the segment is protected after
writing, making it read-only. Set to FALSE only if you need to modify
the shared data (advanced use case).}

\item{name}{Optional name for the shared object. If NULL (default), a unique
name is generated. Named shares can be opened by name in other processes.}

\item{deep}{Logical. If TRUE, recursively traverse lists and data.frames,
sharing individual components that meet the size threshold. When FALSE
(default), the entire object is serialized as one unit.}

\item{min_bytes}{Minimum size in bytes for an object to be shared when
deep=TRUE. Objects smaller than this threshold are kept in-place.
Default is 64MB (64 * 1024 * 1024).}

\item{cycle}{How to handle cyclic references when deep=TRUE. Either "error"
(default) to stop with an error, or "skip" to skip cyclic references.}

\item{mode}{Sharing mode when deep=TRUE. Either "balanced" (default) to
continue on hook errors and non-shareable types, or "strict" to error.}
}
\value{
A \code{shard_shared} object (when deep=FALSE) or
\code{shard_deep_shared} object (when deep=TRUE) containing:
\itemize{
\item \code{path}: The path or name of the shared segment
\item \code{backing}: The backing type used
\item \code{size}: Total size in bytes
\item \code{readonly}: Whether the segment is protected
\item \code{class_info}: Original class information
}
}
\description{
Create shared memory representations of R objects for efficient
parallel access without duplication.

Creates a shared memory representation of an R object. The object is
serialized once and can be accessed by multiple worker processes without
copying.
}
\details{
The \code{share()} function is the primary high-level API for creating zero-copy
shared inputs. When you share an object:

\enumerate{
\item The object is serialized into a shared memory segment
\item The segment is marked read-only (protected)
\item A lightweight handle is returned that can be passed to workers
\item Workers attach to the segment and deserialize on demand
}

This approach eliminates per-worker duplication of large inputs. The data
exists once in shared memory, and all workers read from the same location.

\strong{Immutability Contract}: Shared objects are immutable by design.
Any attempt to modify shared data in a worker will fail. This guarantees
deterministic behavior and prevents accidental copy-on-write.
}
\examples{
\donttest{
mat <- matrix(rnorm(1e4), nrow = 100)
shared_mat <- share(mat)
recovered <- fetch(shared_mat)
identical(mat, recovered)
close(shared_mat)
}
}
\seealso{
\code{\link{segment_create}} for low-level segment operations,
\code{\link{pool_create}} for worker pool management.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{buffer_close}
\alias{buffer_close}
\title{Close a Buffer}
\usage{
buffer_close(x, unlink = NULL)
}
\arguments{
\item{x}{A shard_buffer object.}

\item{unlink}{Whether to unlink the underlying segment.}
}
\value{
\code{NULL}, invisibly.
}
\description{
Closes the buffer and releases the underlying shared memory.
}
\examples{
\donttest{
buf <- buffer("double", dim = 10)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{shared_info}
\alias{shared_info}
\title{Get Information About a Shared Object}
\usage{
shared_info(x)
}
\arguments{
\item{x}{A \code{shard_shared} object.}
}
\value{
A named list with fields \code{path}, \code{backing}, \code{size},
\code{readonly}, \code{class_info}, and \code{segment_info}.
}
\description{
Get Information About a Shared Object
}
\examples{
\donttest{
shared <- share(1:100)
shared_info(shared)
close(shared)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rss.R
\name{rss}
\alias{rss}
\title{RSS Monitoring Utilities}
\description{
Cross-platform utilities for monitoring process memory usage.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{mem_report}
\alias{mem_report}
\title{Memory Usage Report}
\usage{
mem_report(pool = NULL)
}
\arguments{
\item{pool}{Optional. A \code{shard_pool} object. If NULL, uses the current pool.}
}
\value{
An S3 object of class \code{shard_report} with type \code{"memory"}
containing:
\itemize{
\item \code{type}: "memory"
\item \code{timestamp}: When the report was generated
\item \code{pool_active}: Whether a pool exists
\item \code{n_workers}: Number of workers
\item \code{rss_limit}: RSS limit per worker (bytes)
\item \code{total_rss}: Sum of RSS across all workers
\item \code{peak_rss}: Highest RSS among workers
\item \code{mean_rss}: Mean RSS across workers
\item \code{workers}: Per-worker RSS details
}
}
\description{
Generates a report of memory usage across all workers in the pool.
}
\examples{
\donttest{
p <- pool_create(2)
mem_report(p)
pool_stop(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinity.R
\name{buffer_advise}
\alias{buffer_advise}
\title{Advise access pattern for a buffer}
\usage{
buffer_advise(
  x,
  advice = c("normal", "sequential", "random", "willneed", "dontneed")
)
}
\arguments{
\item{x}{A shard_buffer.}

\item{advice}{See \code{\link[=segment_advise]{segment_advise()}}.}
}
\value{
A logical scalar; \code{TRUE} if the OS accepted the hint.
}
\description{
Advise access pattern for a buffer
}
\examples{
\donttest{
buf <- buffer("double", dim = 10L)
buffer_advise(buf, "sequential")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_write}
\alias{table_write}
\title{Write tabular results into a table buffer or sink}
\usage{
table_write(target, rows_or_shard_id, data, ...)
}
\arguments{
\item{target}{A \code{shard_table_buffer} or \code{shard_table_sink}.}

\item{rows_or_shard_id}{For buffers: row selector (idx_range or integer vector).
For sinks: shard id (integer).}

\item{data}{A data.frame or named list matching the schema columns.}

\item{...}{Reserved for future extensions.}
}
\value{
\code{NULL} (invisibly).
}
\description{
\code{table_write()} is the common write path for shard table outputs:
\itemize{
\item For fixed-size outputs, write into a \code{shard_table_buffer} using a row selector.
\item For variable-size outputs, write into a \code{shard_table_sink} using a shard id.
}
}
\examples{
\donttest{
s <- schema(x = float64(), y = int32())
tb <- table_buffer(s, nrow = 10L)
table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dispatch.R
\name{dispatch}
\alias{dispatch}
\title{Task Dispatch Engine}
\description{
Orchestrates chunk dispatch with worker supervision and failure handling.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{as_tibble.shard_table_handle}
\alias{as_tibble.shard_table_handle}
\title{Materialize a table handle into a data.frame/tibble}
\usage{
\method{as_tibble}{shard_table_handle}(x, max_bytes = 256 * 1024^2, ...)
}
\arguments{
\item{x}{A \code{shard_table_handle}.}

\item{max_bytes}{Warn if estimated payload exceeds this threshold.}

\item{...}{Reserved for future extensions.}
}
\value{
A data.frame (or tibble if the \code{tibble} package is installed).
}
\description{
Materialize a table handle into a data.frame/tibble
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_stream.R
\name{stream_sum}
\alias{stream_sum}
\title{Stream sum of a numeric column}
\usage{
stream_sum(x, col, na_rm = TRUE)
}
\arguments{
\item{x}{A \code{shard_row_groups} or \code{shard_dataset} handle.}

\item{col}{Column name to sum.}

\item{na_rm}{Logical; drop NAs (default TRUE).}
}
\value{
A single numeric value giving the sum of the column across all partitions.
}
\description{
Computes the sum of \code{col} across all partitions without collecting the full
dataset. When partitions are native-encoded, this avoids decoding string
columns entirely.
}
\examples{
\donttest{
s <- schema(x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L, data.frame(x = c(1.0, 2.0, 3.0)))
rg <- table_finalize(sink)
stream_sum(rg, "x")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{print.shard_buffer}
\alias{print.shard_buffer}
\title{Print a Shared Memory Buffer}
\usage{
\method{print}{shard_buffer}(x, ...)
}
\arguments{
\item{x}{A \code{shard_buffer} object.}

\item{...}{Ignored.}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a Shared Memory Buffer
}
\examples{
\donttest{
buf <- buffer("double", dim = 10)
print(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{print.shard_report}
\alias{print.shard_report}
\title{Print a shard_report Object}
\usage{
\method{print}{shard_report}(x, ...)
}
\arguments{
\item{x}{A \code{shard_report} object.}

\item{...}{Ignored.}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a shard_report Object
}
\examples{
\donttest{
res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
pool_stop()
rpt <- report(result = res)
print(rpt)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{shared_reset_diagnostics}
\alias{shared_reset_diagnostics}
\title{Reset diagnostic counters for a shared vector}
\usage{
shared_reset_diagnostics(x)
}
\arguments{
\item{x}{A shard ALTREP vector}
}
\value{
x (invisibly)
}
\description{
Reset diagnostic counters for a shared vector
}
\examples{
\donttest{
seg <- segment_create(400)
segment_write(seg, 1:100, offset = 0)
x <- shared_vector(seg, "integer", length = 100)

sum(x)
shared_diagnostics(x)$dataptr_calls

shared_reset_diagnostics(x)
shared_diagnostics(x)$dataptr_calls
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{is_windows}
\alias{is_windows}
\title{Check if running on Windows}
\usage{
is_windows()
}
\value{
A logical scalar: \code{TRUE} if running on Windows, \code{FALSE} otherwise.
}
\description{
Check if running on Windows
}
\examples{
is_windows()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{share_open}
\alias{share_open}
\title{Open an Existing Shared Object by Path}
\usage{
share_open(path, backing = c("mmap", "shm"), size = NULL)
}
\arguments{
\item{path}{Path to the shared segment.}

\item{backing}{Backing type: "mmap" or "shm".}

\item{size}{Size of the segment in bytes. If NULL, attempts to detect.}
}
\value{
A \code{shard_shared} object attached to the existing segment.
}
\description{
Opens a shared object that was created by another process. This is useful
for workers that need to attach to shared data without having the original
\code{shard_shared} object.
}
\examples{
\donttest{
shared <- share(1:50)
info <- shared_info(shared)
reopened <- share_open(info$path, backing = "mmap")
close(reopened)
close(shared)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_close}
\alias{segment_close}
\title{Close a shared memory segment}
\usage{
segment_close(x, unlink = NULL)
}
\arguments{
\item{x}{A shard_segment object}

\item{unlink}{Whether to unlink the underlying file/shm (default: FALSE for
opened segments, TRUE for owned segments)}
}
\value{
\code{NULL}, invisibly.
}
\description{
Close a shared memory segment
}
\examples{
\donttest{
seg <- segment_create(1024)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{shared_diagnostics}
\alias{shared_diagnostics}
\title{Get diagnostics for a shared vector}
\usage{
shared_diagnostics(x)
}
\arguments{
\item{x}{A shard ALTREP vector}
}
\value{
A list with diagnostic information:
\describe{
\item{dataptr_calls}{Number of times DATAPTR was accessed}
\item{materialize_calls}{Number of times vector was copied to standard R vector}
\item{length}{Number of elements}
\item{offset}{Byte offset into underlying segment}
\item{readonly}{Whether write access is prevented}
\item{type}{R type of the vector}
}
}
\description{
Get diagnostics for a shared vector
}
\examples{
\donttest{
seg <- segment_create(400)
segment_write(seg, 1:100, offset = 0)
x <- shared_vector(seg, "integer", length = 100)

sum(x)

shared_diagnostics(x)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_diagnostics}
\alias{table_diagnostics}
\title{Table Diagnostics}
\usage{
table_diagnostics()
}
\value{
A list with \code{writes}, \code{rows}, and \code{bytes}.
}
\description{
Per-process counters for table writes (number of table_write calls, rows,
and bytes written). shard_map uses deltas of these counters to produce
run-level diagnostics in copy_report().
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{available_backings}
\alias{available_backings}
\title{Get available shared memory backing types}
\usage{
available_backings()
}
\value{
A character vector of available backing types on the current platform.
}
\description{
Get available shared memory backing types
}
\examples{
available_backings()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pool.R
\name{pool_stop}
\alias{pool_stop}
\title{Stop the Worker Pool}
\usage{
pool_stop(pool = NULL, timeout = 5)
}
\arguments{
\item{pool}{A \code{shard_pool} object. If NULL, uses the current pool.}

\item{timeout}{Numeric. Seconds to wait for workers to terminate (default 5).
Returns after timeout even if workers are still alive.}
}
\value{
NULL (invisibly).
}
\description{
Terminates all worker processes and releases resources. Waits for workers
to actually terminate before returning.
}
\examples{
\donttest{
p <- pool_create(2)
pool_stop(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{task_report}
\alias{task_report}
\title{Task Execution Report}
\usage{
task_report(result = NULL)
}
\arguments{
\item{result}{A \code{shard_result} object from \code{\link{shard_map}}.}
}
\value{
An S3 object of class \code{shard_report} with type \code{"task"}
containing:
\itemize{
\item \code{type}: "task"
\item \code{timestamp}: When the report was generated
\item \code{shards_total}: Total number of shards
\item \code{shards_processed}: Number of shards successfully processed
\item \code{shards_failed}: Number of permanently failed shards
\item \code{chunks_dispatched}: Number of chunk batches dispatched
\item \code{total_retries}: Total number of retry attempts
\item \code{duration}: Total execution duration (seconds)
\item \code{throughput}: Shards processed per second
\item \code{queue_status}: Final queue status
}
}
\description{
Generates a report of task/chunk execution statistics from a shard_map result.
}
\examples{
\donttest{
res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
pool_stop()
task_report(res)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment}
\alias{segment}
\title{Shared Memory Segment}
\description{
Low-level shared memory segment operations for cross-process
data sharing. These functions provide the foundation for the higher-level
\code{share()} and \code{buffer()} APIs.
}
\details{
Segments can be backed by:
\itemize{
\item \code{"shm"}: POSIX shared memory (Linux/macOS) or named file
mapping (Windows). Faster but may have size limitations.
\item \code{"mmap"}: File-backed memory mapping. Works on all platforms
and supports larger sizes.
\item \code{"auto"}: Let the system choose the best option.
}

All segments are created with secure permissions (0600 on Unix) and are
automatically cleaned up when the R object is garbage collected.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ergonomics.R
\name{shard_apply_matrix}
\alias{shard_apply_matrix}
\title{Apply a Function Over Matrix Columns with Shared Inputs}
\usage{
shard_apply_matrix(
  X,
  MARGIN = 2,
  FUN,
  VARS = NULL,
  workers = NULL,
  ...,
  policy = shard_apply_policy()
)
}
\arguments{
\item{X}{A numeric/integer/logical matrix (or a shared matrix created by \code{\link[=share]{share()}}).}

\item{MARGIN}{Must be 2 (columns).}

\item{FUN}{Function of the form \verb{function(v, ...)} returning a scalar atomic.}

\item{VARS}{Optional named list of extra variables. Large atomic VARS are
auto-shared based on \code{policy$auto_share_min_bytes}.}

\item{workers}{Number of workers (passed to \code{\link[=shard_map]{shard_map()}}).}

\item{...}{Additional arguments forwarded to \code{FUN}.}

\item{policy}{A \code{\link[=shard_apply_policy]{shard_apply_policy()}} object.}
}
\value{
An atomic vector of length \code{ncol(X)} with the results.
}
\description{
A convenience wrapper for the common "per-column apply" pattern. The matrix
is shared once and each worker receives a zero-copy column view when possible.
}
\details{
Current limitation: \code{MARGIN} must be 2 (columns). Row-wise apply would require
strided/gather slicing and is intentionally explicit in shard via views/kernels.
}
\examples{
\donttest{
X <- matrix(rnorm(400), 20, 20)
shard_apply_matrix(X, MARGIN = 2, FUN = mean)
pool_stop()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapter.R
\name{shard_list_adapters}
\alias{shard_list_adapters}
\title{List Registered Adapters}
\usage{
shard_list_adapters()
}
\value{
Character vector of class names with registered adapters.
}
\description{
Returns a character vector of all classes with registered adapters.
}
\examples{
shard_list_adapters()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{materialize}
\alias{materialize}
\alias{materialize.shard_shared}
\alias{materialize.default}
\title{Materialize Shared Object}
\usage{
materialize(x)

\method{materialize}{shard_shared}(x)

\method{materialize}{default}(x)
}
\arguments{
\item{x}{A \code{shard_shared} object.}
}
\value{
The original R object.
}
\description{
Alias for \code{fetch()}. Retrieves the R object from shared memory.
}
\examples{
\donttest{
shared <- share(1:100)
data <- materialize(shared)
close(shared)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{fetch}
\alias{fetch}
\alias{fetch.shard_shared}
\alias{fetch.shard_deep_shared}
\alias{fetch.default}
\title{Fetch Data from a Shared Object}
\usage{
fetch(x, ...)

\method{fetch}{shard_shared}(x, ...)

\method{fetch}{shard_deep_shared}(x, ...)

\method{fetch}{default}(x, ...)
}
\arguments{
\item{x}{A \code{shard_shared} object.}

\item{...}{Ignored.}
}
\value{
The original R object that was shared.
}
\description{
Retrieves the R object from shared memory by deserializing it. This is the
primary way to access shared data in workers.
}
\details{
When called in the main process, this reads from the existing segment.
When called in a worker process, this opens the segment by path and
deserializes the data.

The \code{fetch()} function is the primary way to access shared data.
It can also be called as \code{materialize()} for compatibility.
}
\examples{
\donttest{
x <- 1:100
shared <- share(x)
recovered <- fetch(shared)
identical(x, recovered)
close(shared)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{as_tibble}
\alias{as_tibble}
\title{Materialize a shard table handle as a data.frame/tibble}
\usage{
as_tibble(x, max_bytes = 256 * 1024^2, ...)
}
\arguments{
\item{x}{A shard table object.}

\item{max_bytes}{Warn if estimated payload exceeds this threshold.}

\item{...}{Reserved for future extensions.}
}
\value{
A data.frame (or tibble if the \code{tibble} package is installed).
}
\description{
Materialize a shard table handle as a data.frame/tibble
}
\examples{
\donttest{
s <- schema(x = float64(), y = int32())
tb <- table_buffer(s, nrow = 5L)
table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
df <- as_tibble(tb)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pool.R
\name{pool_get}
\alias{pool_get}
\title{Get the Current Worker Pool}
\usage{
pool_get()
}
\value{
A \code{shard_pool} object or NULL.
}
\description{
Returns the active worker pool, or NULL if none exists.
}
\examples{
p <- pool_get()
is.null(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{print.shard_shared}
\alias{print.shard_shared}
\title{Print a Shared Object}
\usage{
\method{print}{shard_shared}(x, ...)
}
\arguments{
\item{x}{A \code{shard_shared} object.}

\item{...}{Ignored.}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a Shared Object
}
\examples{
\donttest{
shared <- share(1:10)
print(shared)
close(shared)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{collect.shard_dataset}
\alias{collect.shard_dataset}
\title{Collect a dataset handle into memory}
\usage{
\method{collect}{shard_dataset}(x, ...)
}
\arguments{
\item{x}{A \code{shard_dataset} handle.}

\item{...}{Passed to \code{as_tibble()}.}
}
\value{
A data.frame (or tibble if the \code{tibble} package is installed).
}
\description{
Collect a dataset handle into memory
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{dim.shard_buffer}
\alias{dim.shard_buffer}
\title{Dimensions of a Shared Memory Buffer}
\usage{
\method{dim}{shard_buffer}(x)
}
\arguments{
\item{x}{A \code{shard_buffer} object.}
}
\value{
An integer vector of dimensions, or \code{NULL} for 1-D buffers.
}
\description{
Dimensions of a Shared Memory Buffer
}
\examples{
\donttest{
buf <- buffer("double", dim = c(4, 5))
dim(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_protect}
\alias{segment_protect}
\title{Make a segment read-only}
\usage{
segment_protect(x)
}
\arguments{
\item{x}{A shard_segment object}
}
\value{
The \code{shard_segment} object, invisibly.
}
\description{
Make a segment read-only
}
\examples{
\donttest{
seg <- segment_create(1024)
segment_protect(seg)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shard_map.R
\name{succeeded}
\alias{succeeded}
\title{Check if shard_map Succeeded}
\usage{
succeeded(x)
}
\arguments{
\item{x}{A shard_result object.}
}
\value{
Logical. TRUE if no failures.
}
\description{
Check if shard_map Succeeded
}
\examples{
\donttest{
result <- shard_map(4L, function(shard) shard$idx[[1L]], workers = 2)
pool_stop()
succeeded(result)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{iterate_row_groups}
\alias{iterate_row_groups}
\title{Iterate row groups}
\usage{
iterate_row_groups(x, decode = TRUE)
}
\arguments{
\item{x}{A \code{shard_row_groups} handle.}

\item{decode}{Logical. If TRUE (default), native-encoded partitions are
decoded to data.frames. If FALSE, native partitions are returned as their
internal representation (advanced).}
}
\value{
A zero-argument iterator function that returns the next data.frame on
each call, or \code{NULL} when exhausted.
}
\description{
Iterate row groups
}
\examples{
\donttest{
s <- schema(x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L, data.frame(x = rnorm(5)))
rg <- table_finalize(sink)
it <- iterate_row_groups(rg)
chunk <- it()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_stream.R
\name{stream_count}
\alias{stream_count}
\title{Stream row count}
\usage{
stream_count(x)
}
\arguments{
\item{x}{A \code{shard_row_groups} or \code{shard_dataset} handle.}
}
\value{
A single integer giving the total number of rows across all partitions.
}
\description{
Stream row count
}
\examples{
\donttest{
s <- schema(x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L, data.frame(x = rnorm(5)))
rg <- table_finalize(sink)
stream_count(rg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ergonomics.R
\name{shard_lapply_shared}
\alias{shard_lapply_shared}
\title{Apply a Function Over a List with Optional Auto-Sharing}
\usage{
shard_lapply_shared(
  x,
  FUN,
  VARS = NULL,
  workers = NULL,
  ...,
  policy = shard_apply_policy()
)
}
\arguments{
\item{x}{A list.}

\item{FUN}{Function of the form \verb{function(el, ...)}.}

\item{VARS}{Optional named list of extra variables (auto-shared when large).}

\item{workers}{Number of workers (passed to \code{\link[=shard_map]{shard_map()}}).}

\item{...}{Additional arguments forwarded to \code{FUN}.}

\item{policy}{A \code{\link[=shard_apply_policy]{shard_apply_policy()}} object.}
}
\value{
A list of results, one per element of \code{x}.
}
\description{
A convenience wrapper for list workloads that need supervision and shared
inputs. Large atomic list elements are auto-shared based on policy.
}
\details{
This wrapper enforces guardrails to avoid accidental huge gathers: it
estimates the total gathered result size from a probe call and refuses
to run if it exceeds \code{policy$max_gather_bytes}.
}
\examples{
\donttest{
res <- shard_lapply_shared(as.list(1:4), function(x) x^2)
pool_stop()
res
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapter.R
\name{adapter}
\alias{adapter}
\title{Adapter Registry for Class-Specific Deep Sharing}
\description{
Register custom traversal logic for specific classes during
deep sharing operations. Adapters allow fine-grained control over how
objects are decomposed and reconstructed.
}
\details{
The adapter registry provides a way to customize how specific classes are
handled during deep sharing. Instead of generic slot traversal for S4 objects
or element-wise traversal for lists, you can provide custom functions to:

\enumerate{
\item Extract the shareable children from an object (\code{children})
\item Reconstruct the object from shared children (\code{replace})
}

This is useful for:
\itemize{
\item Complex S4 objects where only certain slots should be shared
\item S3 objects with internal structure that differs from list structure
\item Objects with accessors that should be used instead of direct slot access
}
}
\seealso{
\code{\link{share}} for the main sharing function that uses adapters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scratch.R
\name{scratch_diagnostics}
\alias{scratch_diagnostics}
\title{Scratch pool diagnostics}
\usage{
scratch_diagnostics()
}
\value{
A list with counters and current pool bytes.
}
\description{
Scratch pool diagnostics
}
\examples{
scratch_diagnostics()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{is_view}
\alias{is_view}
\alias{is_block_view}
\title{View Predicates}
\usage{
is_view(x)

is_block_view(x)
}
\arguments{
\item{x}{An object.}
}
\value{
Logical. \code{TRUE} if \code{x} is a shard view (or block view).
}
\description{
View Predicates
}
\examples{
\donttest{
m <- share(matrix(1:20, nrow = 4))
v <- view_block(m, cols = idx_range(1, 2))
is_view(v)
is_block_view(v)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{view_gather}
\alias{view_gather}
\title{Create a gather (indexed) view over a shared matrix}
\usage{
view_gather(x, rows = NULL, cols)
}
\arguments{
\item{x}{A shared (share()d) atomic matrix (double/integer/logical/raw).}

\item{rows}{Row selector. NULL (all rows) or idx_range().}

\item{cols}{Integer vector of column indices (1-based).}
}
\value{
A \code{shard_view_gather} object describing the indexed column view.
}
\description{
Gather views describe non-contiguous column (or row) subsets without
allocating a slice-sized matrix. shard-aware kernels can then choose to
pack the requested indices into scratch explicitly (bounded and reportable)
or run gather-aware compute paths.
}
\details{
v1 note: only column-gather views are implemented (rows may be NULL or
idx_range()).
}
\examples{
\donttest{
m <- share(matrix(1:20, nrow = 4))
v <- view_gather(m, cols = c(1L, 3L))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{view_block}
\alias{view_block}
\title{Create a contiguous block view}
\usage{
view_block(x, rows = NULL, cols = NULL)
}
\arguments{
\item{x}{A shared (share()d) atomic matrix.}

\item{rows}{NULL or idx_range().}

\item{cols}{NULL or idx_range().}
}
\value{
A \code{shard_view_block} object representing the contiguous block slice.
}
\description{
Create a contiguous block view
}
\examples{
\donttest{
m <- share(matrix(1:20, nrow = 4))
v <- view_block(m, cols = idx_range(1, 2))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_finalize.shard_table_sink}
\alias{table_finalize.shard_table_sink}
\title{Finalize a sink}
\usage{
\method{table_finalize}{shard_table_sink}(
  target,
  materialize = c("never", "auto", "always"),
  max_bytes = 256 * 1024^2,
  ...
)
}
\arguments{
\item{target}{A \code{shard_table_sink}.}

\item{materialize}{\code{"never"}, \code{"auto"}, or \code{"always"}.}

\item{max_bytes}{For \code{"auto"}, materialize only if estimated bytes <= max_bytes.}

\item{...}{Reserved for future extensions.}
}
\value{
A \code{shard_row_groups} handle (or a materialized data.frame/tibble).
}
\description{
Finalize a sink
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shards_lazy.R
\name{[.shard_descriptor_lazy}
\alias{[.shard_descriptor_lazy}
\title{Subset a shard_descriptor_lazy Object}
\usage{
\method{[}{shard_descriptor_lazy}(x, i)
}
\arguments{
\item{x}{A \code{shard_descriptor_lazy} object.}

\item{i}{Index or indices.}
}
\value{
A subset of the object.
}
\description{
Subset a shard_descriptor_lazy Object
}
\examples{
sh <- shards(100, block_size = 25)
sh[1:2]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scratch.R
\name{scratch_matrix}
\alias{scratch_matrix}
\title{Get a scratch matrix}
\usage{
scratch_matrix(nrow, ncol, key = NULL)
}
\arguments{
\item{nrow, ncol}{Dimensions.}

\item{key}{Optional key to control reuse. Defaults to a shape-derived key.}
}
\value{
A double matrix of dimensions \code{nrow} by \code{ncol}.
}
\description{
Allocates (or reuses) a double matrix in the worker scratch pool.
}
\examples{
m <- scratch_matrix(10, 5)
dim(m)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{buffer_path}
\alias{buffer_path}
\title{Get Buffer Path}
\usage{
buffer_path(x)
}
\arguments{
\item{x}{A shard_buffer object.}
}
\value{
A character string with the path or name of the segment, or
\code{NULL} if the segment is anonymous.
}
\description{
Returns the path or name of the buffer's underlying segment.
Use this to pass buffer location to workers.
}
\examples{
\donttest{
buf <- buffer("double", dim = 10)
buffer_path(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_stream.R
\name{stream_reduce}
\alias{stream_reduce}
\alias{stream_reduce.shard_row_groups}
\alias{stream_reduce.shard_dataset}
\title{Stream over row-groups/datasets and reduce}
\usage{
stream_reduce(x, f, init, combine, ...)

\method{stream_reduce}{shard_row_groups}(x, f, init, combine, ...)

\method{stream_reduce}{shard_dataset}(x, f, init, combine, ...)
}
\arguments{
\item{x}{A \code{shard_row_groups} or \code{shard_dataset} handle.}

\item{f}{Function \code{(chunk, ...) -> value} producing a per-partition value.}

\item{init}{Initial accumulator value.}

\item{combine}{Function \code{(acc, value) -> acc} to update the accumulator.}

\item{...}{Passed to \code{f()}.}
}
\value{
The final accumulator value after processing all partitions.
}
\description{
Applies \code{f()} to each partition (row-group) and combines results with
\code{combine()} into a single accumulator. This keeps peak memory bounded by the
largest single partition (plus your accumulator).
}
\examples{
\donttest{
s <- schema(x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L, data.frame(x = rnorm(5)))
rg <- table_finalize(sink)
total <- stream_reduce(rg, f = nrow, init = 0L, combine = `+`)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{print.shard_deep_shared}
\alias{print.shard_deep_shared}
\title{Print a Deep-Shared Object}
\usage{
\method{print}{shard_deep_shared}(x, ...)
}
\arguments{
\item{x}{A \code{shard_deep_shared} object.}

\item{...}{Ignored.}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a Deep-Shared Object
}
\examples{
\donttest{
lst <- list(a = 1:10, b = 11:20)
shared <- share(lst, deep = TRUE, min_bytes = 1)
print(shared)
close(shared)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinity.R
\name{set_affinity}
\alias{set_affinity}
\title{Set CPU affinity for the current process}
\usage{
set_affinity(cores)
}
\arguments{
\item{cores}{Integer vector of 0-based CPU core ids.}
}
\value{
A logical scalar; \code{TRUE} on success, \code{FALSE} if not supported.
}
\description{
Intended to be called inside a worker process (e.g., via \code{clusterCall()}).
On unsupported platforms, returns FALSE.
}
\examples{
affinity_supported()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ergonomics.R
\name{ergonomics}
\alias{ergonomics}
\title{Ergonomic Apply/Lapply Wrappers}
\description{
Convenience wrappers that provide apply/lapply-style ergonomics while
preserving shard's core contract: shared immutable inputs, supervised
execution, and diagnostics.

These functions are intentionally thin wrappers around \code{\link[=shard_map]{shard_map()}} and
related primitives.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{materialize.shard_view_block}
\alias{materialize.shard_view_block}
\title{Materialize a block view into an R matrix}
\usage{
\method{materialize}{shard_view_block}(x)
}
\arguments{
\item{x}{A \code{shard_view_block} object.}
}
\value{
A standard R matrix containing the selected rows and columns.
}
\description{
Materialize a block view into an R matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{as_tibble.shard_table_buffer}
\alias{as_tibble.shard_table_buffer}
\title{Materialize a fixed table handle or buffer}
\usage{
\method{as_tibble}{shard_table_buffer}(x, max_bytes = 256 * 1024^2, ...)
}
\arguments{
\item{x}{A \code{shard_table_handle} or \code{shard_table_buffer}.}

\item{max_bytes}{Warn if estimated payload exceeds this threshold.}

\item{...}{Reserved for future extensions.}
}
\value{
A data.frame (or tibble).
}
\description{
Converts a \code{shard_table_handle} to an in-memory data.frame (or tibble if the
tibble package is installed).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cow.R
\name{print.shard_shared_vector}
\alias{print.shard_shared_vector}
\title{Print a Shared Vector}
\usage{
\method{print}{shard_shared_vector}(x, ...)
}
\arguments{
\item{x}{A \code{shard_shared_vector}.}

\item{...}{Additional arguments passed to \code{print}.}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print method for \code{shard_shared_vector} objects. Drops the wrapper class
and delegates to the underlying R print method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{print.shard_tiles}
\alias{print.shard_tiles}
\title{Print a shard_tiles object}
\usage{
\method{print}{shard_tiles}(x, ...)
}
\arguments{
\item{x}{A \code{shard_tiles} object.}

\item{...}{Additional arguments (ignored).}
}
\value{
Returns \code{x} invisibly.
}
\description{
Print a shard_tiles object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scratch.R
\name{scratch_pool_config}
\alias{scratch_pool_config}
\title{Configure scratch pool limits}
\usage{
scratch_pool_config(max_bytes = Inf)
}
\arguments{
\item{max_bytes}{Maximum scratch pool bytes allowed in a worker. If exceeded,
the worker is flagged for recycle at the next safe point.}
}
\value{
\code{NULL}, invisibly.
}
\description{
Configure scratch pool limits
}
\examples{
cfg <- scratch_pool_config(max_bytes = 100 * 1024^2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shard-package.R
\docType{package}
\name{shard-package}
\alias{shard}
\alias{shard-package}
\title{shard: Deterministic, Zero-Copy Parallel Execution for R}
\description{
Provides a parallel execution runtime for R that emphasizes deterministic memory behavior and efficient handling of large shared inputs. 'shard' enables zero-copy parallel reads via shared/memory-mapped segments, encourages explicit output buffers to avoid large result aggregation, and supervises worker processes to mitigate memory drift via controlled recycling. Diagnostics report peak memory usage, end-of-run memory return, and hidden copy/materialization events to support reproducible performance benchmarking.
}
\section{Core API}{

\itemize{
\item \code{\link[=shard_map]{shard_map()}} - Primary parallel execution entry point
\item \code{\link[=shards]{shards()}} - Create shard descriptors with autotuning
\item \code{\link[=results]{results()}} - Extract results from a shard_map run
\item \code{\link[=succeeded]{succeeded()}} - Check if shard_map completed without failures
}
}

\section{Zero-Copy Shared Data}{

\itemize{
\item \code{\link[=share]{share()}} - Share an R object for parallel access
\item \code{\link[=fetch]{fetch()}} - Retrieve data from a shared object
\item \code{\link[=materialize]{materialize()}} - Alias for fetch()
\item \code{\link[=is_shared]{is_shared()}} - Check if an object is shared
\item \code{\link[=shared_info]{shared_info()}} - Get information about a shared object
}
}

\section{Output Buffers}{

\itemize{
\item \code{\link[=buffer]{buffer()}} - Create typed writable output buffer
\item \code{\link[=buffer_open]{buffer_open()}} - Open existing buffer from another process
\item \code{\link[=buffer_path]{buffer_path()}} - Get buffer path for cross-process sharing
\item \code{\link[=buffer_info]{buffer_info()}} - Get buffer information
\item \code{\link[=buffer_close]{buffer_close()}} - Close and release buffer
}
}

\section{Worker Pool Management}{

\itemize{
\item \code{\link[=pool_create]{pool_create()}} - Create a supervised worker pool
\item \code{\link[=pool_stop]{pool_stop()}} - Stop the worker pool
\item \code{\link[=pool_status]{pool_status()}} - Check worker status and RSS
\item \code{\link[=pool_health_check]{pool_health_check()}} - Monitor and recycle workers
}
}

\section{Task Dispatch}{

\itemize{
\item \code{\link[=dispatch_chunks]{dispatch_chunks()}} - Execute chunks with supervision
\item \code{\link[=pool_lapply]{pool_lapply()}} - Parallel lapply with supervision
\item \code{\link[=pool_sapply]{pool_sapply()}} - Parallel sapply with supervision
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://bbuchsbaum.github.io/shard/}
  \item \url{https://github.com/bbuchsbaum/shard}
  \item Report bugs at \url{https://github.com/bbuchsbaum/shard/issues}
}

}
\author{
\strong{Maintainer}: Bradley Buchsbaum \email{brad.buchsbaum@gmail.com} [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shard_map.R
\name{print.shard_result}
\alias{print.shard_result}
\title{Print a shard_result Object}
\usage{
\method{print}{shard_result}(x, ...)
}
\arguments{
\item{x}{A \code{shard_result} object.}

\item{...}{Further arguments (ignored).}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a shard_result Object
}
\examples{
\donttest{
result <- shard_map(4L, function(shard) shard$idx, workers = 2)
pool_stop()
print(result)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shard_map.R
\name{results}
\alias{results}
\title{Extract Results from shard_map}
\usage{
results(x, flatten = TRUE)
}
\arguments{
\item{x}{A shard_result object.}

\item{flatten}{Logical. Flatten nested results?}
}
\value{
List or vector of results.
}
\description{
Extract Results from shard_map
}
\examples{
\donttest{
result <- shard_map(4L, function(shard) shard$idx[[1L]], workers = 2)
pool_stop()
results(result)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pool.R
\name{pool_health_check}
\alias{pool_health_check}
\title{Check Pool Health}
\usage{
pool_health_check(pool = NULL, busy_workers = NULL)
}
\arguments{
\item{pool}{A \code{shard_pool} object. If NULL, uses the current pool.}

\item{busy_workers}{Optional integer vector of worker ids that are currently
running tasks (used internally by the dispatcher to avoid recycling a worker
while a result is in flight).}
}
\value{
A list with health status per worker and actions taken.
}
\description{
Monitors all workers, recycling those with excessive RSS drift or that have died.
}
\examples{
\donttest{
p <- pool_create(2)
pool_health_check(p)
pool_stop(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{close.shard_shared}
\alias{close.shard_shared}
\alias{close.shard_shared_vector}
\alias{close.shard_deep_shared}
\title{Close a Shared Object}
\usage{
\method{close}{shard_shared}(con, ...)

\method{close}{shard_shared_vector}(con, ...)

\method{close}{shard_deep_shared}(con, ...)
}
\arguments{
\item{con}{A \code{shard_shared} object.}

\item{...}{Ignored.}
}
\value{
NULL (invisibly).
}
\description{
Releases the shared memory segment. After closing, the shared object can
no longer be accessed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_read}
\alias{segment_read}
\title{Read raw data from a segment}
\usage{
segment_read(x, offset = 0, size = NULL)
}
\arguments{
\item{x}{A shard_segment object}

\item{offset}{Byte offset to start reading (0-based)}

\item{size}{Number of bytes to read}
}
\value{
A raw vector containing the bytes read from the segment.
}
\description{
Read raw data from a segment
}
\examples{
\donttest{
seg <- segment_create(1024)
segment_write(seg, as.integer(1:4), offset = 0)
segment_read(seg, offset = 0, size = 16)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dispatch.R
\name{pool_lapply}
\alias{pool_lapply}
\title{Parallel Dispatch with Async Workers}
\usage{
pool_lapply(X, FUN, ..., pool = NULL, chunk_size = 1L)
}
\arguments{
\item{X}{List or vector to iterate over.}

\item{FUN}{Function to apply to each element.}

\item{...}{Additional arguments to FUN.}

\item{pool}{A \code{shard_pool} object. If NULL, uses current pool.}

\item{chunk_size}{Integer. Elements per chunk (default 1).}
}
\value{
A list of results.
}
\description{
An alternative dispatch that uses parallel::parLapply-style execution
but with supervision. This is a simpler interface for basic parallel apply.
}
\examples{
\donttest{
pool_create(2)
result <- pool_lapply(1:4, function(x) x^2, pool = pool_get())
pool_stop()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinity.R
\name{segment_advise}
\alias{segment_advise}
\title{Advise OS about expected access pattern for a segment}
\usage{
segment_advise(
  seg,
  advice = c("normal", "sequential", "random", "willneed", "dontneed")
)
}
\arguments{
\item{seg}{A shard_segment.}

\item{advice}{One of "normal", "sequential", "random", "willneed", "dontneed".}
}
\value{
A logical scalar; \code{TRUE} if the OS accepted the hint.
}
\description{
This calls \code{madvise()} on the segment mapping when available.
}
\examples{
\donttest{
seg <- segment_create(1024)
segment_advise(seg, "sequential")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ergonomics.R
\name{shard_apply_policy}
\alias{shard_apply_policy}
\title{Apply Wrapper Policy}
\usage{
shard_apply_policy(
  auto_share_min_bytes = "1MB",
  max_gather_bytes = "256MB",
  cow = c("deny", "audit", "allow"),
  profile = c("default", "memory", "speed"),
  block_size = "auto",
  backing = c("auto", "mmap", "shm")
)
}
\arguments{
\item{auto_share_min_bytes}{Minimum object size for auto-sharing (default "1MB").}

\item{max_gather_bytes}{Maximum estimated gathered result bytes before refusing
to run (default "256MB").}

\item{cow}{Copy-on-write policy for borrowed inputs. One of \code{"deny"}, \code{"audit"},
or \code{"allow"}. Default \code{"deny"}.}

\item{profile}{Execution profile passed through to \code{\link[=shard_map]{shard_map()}}. One of
\code{"default"}, \code{"memory"}, or \code{"speed"}. Default \code{"default"}.}

\item{block_size}{Shard block size for apply-style workloads. Default \code{"auto"}.}

\item{backing}{Backing type used when auto-sharing (\code{"auto"}, \code{"mmap"}, \code{"shm"}).}
}
\value{
An object of class \code{shard_apply_policy}.
}
\description{
Centralizes safe defaults and guardrails for apply/lapply convenience wrappers.
}
\examples{
cfg <- shard_apply_policy()
cfg
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{view}
\alias{view}
\title{Create a view over a shared matrix}
\usage{
view(x, rows = NULL, cols = NULL, type = c("auto", "block", "gather"))
}
\arguments{
\item{x}{A shared (share()d) atomic matrix (double/integer/logical/raw).}

\item{rows}{Row selector. NULL (all rows) or idx_range().}

\item{cols}{Column selector. NULL (all cols) or idx_range().}

\item{type}{View type. \code{"block"} or \code{"gather"} (or \code{"auto"}).}
}
\value{
A \code{shard_view_block} or \code{shard_view_gather} object depending
on the selectors provided.
}
\description{
Create a view over a shared matrix
}
\examples{
\donttest{
m <- share(matrix(1:20, nrow = 4))
v <- view(m, cols = idx_range(1, 2))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shard_map.R
\name{shard_map}
\alias{shard_map}
\title{Parallel Execution with shard_map}
\usage{
shard_map(
  shards,
  fun = NULL,
  borrow = list(),
  out = list(),
  kernel = NULL,
  scheduler_policy = NULL,
  autotune = NULL,
  dispatch_mode = c("rpc_chunked", "shm_queue"),
  dispatch_opts = NULL,
  workers = NULL,
  chunk_size = 1L,
  profile = c("default", "memory", "speed"),
  mem_cap = "2GB",
  recycle = TRUE,
  cow = c("deny", "audit", "allow"),
  seed = NULL,
  diagnostics = TRUE,
  packages = NULL,
  init_expr = NULL,
  timeout = 3600,
  max_retries = 3L,
  health_check_interval = 10L
)
}
\arguments{
\item{shards}{A \code{shard_descriptor} from \code{\link[=shards]{shards()}}, or an integer N to
auto-generate shards.}

\item{fun}{Function to execute per shard. Receives the shard descriptor
as first argument, followed by borrowed inputs and outputs. You can also
select a registered kernel via \verb{kernel=} instead of providing \verb{fun=}.}

\item{borrow}{Named list of shared inputs. These are exported to workers
once and reused across shards. Treated as read-only by default.}

\item{out}{Named list of output buffers (from \code{buffer()}). Workers write
results directly to these buffers.}

\item{kernel}{Optional. Name of a registered kernel (see \code{\link[=list_kernels]{list_kernels()}}).
If provided, \code{fun} must be NULL.}

\item{scheduler_policy}{Optional list of scheduling hints (advanced). Currently:
\itemize{
\item \code{max_huge_concurrency}: cap concurrent chunks whose kernel footprint is
classified as \code{"huge"} (see \code{\link[=register_kernel]{register_kernel()}}).
}}

\item{autotune}{Optional. Online autotuning for scalar-N sharding (advanced).
When \code{shards} is an integer \code{N}, shard_map can adjust shard block sizes over
time based on observed wall time and worker RSS.

Accepted values:
\itemize{
\item \code{NULL} (default): enable online autotuning for \code{shard_map(N, ...)}, off for
precomputed shard descriptors.
\item \code{TRUE} / \code{"online"}: force online autotuning (only applies when \code{shards} is
an integer \code{N}).
\item \code{FALSE} / \code{"none"}: disable autotuning.
\item a list: \code{list(mode="online", max_rounds=..., probe_shards_per_worker=..., min_shard_time=...)}
}}

\item{dispatch_mode}{Dispatch mode (advanced). \code{"rpc_chunked"} is the default
supervised socket-based dispatcher. \code{"shm_queue"} is an opt-in fast mode
that uses a shared-memory task queue to reduce per-task overhead for tiny
tasks. In v1, \code{"shm_queue"} is only supported for \code{shard_map(N, ...)} with
\code{chunk_size=1} and is intended for out-buffer/sink workflows (results are
not gathered).}

\item{dispatch_opts}{Optional list of dispatch-mode specific knobs (advanced).
Currently:
\itemize{
\item For \code{dispatch_mode="rpc_chunked"}:
\itemize{
\item \code{auto_table}: logical. If TRUE, shard_map treats data.frame/tibble return
values as row-group outputs and writes them to a table sink
automatically (one partition per shard id). This avoids building a large
list of tibbles and calling bind_rows() on the master. Requires \verb{out=}
to be empty (use explicit \code{out=list(sink=table_sink(...))} otherwise).
\item \code{auto_table_materialize}: \code{"never"}, \code{"auto"}, or \code{"always"} (default \code{"auto"}).
\item \code{auto_table_max_bytes}: numeric/integer. For \code{"auto"}, materialize only
if estimated output size <= this threshold (default 256MB).
\item \code{auto_table_mode}: \code{"row_groups"} (default) or \code{"partitioned"}.
\item \code{auto_table_path}: optional output directory (default tempdir()).
\item \code{auto_table_format}: \code{"auto"}, \code{"rds"} (default), or \code{"native"}.
\item \code{auto_table_schema}: optional \code{shard_schema} for validation/native encoding.
}
\item For \code{dispatch_mode="shm_queue"}:
\itemize{
\item \code{block_size}: integer. If provided, overrides the default heuristic for
contiguous shard block sizing.
\item \code{queue_backing}: one of \code{"mmap"} or \code{"shm"} (default \code{"mmap"}).
\item \code{error_log}: logical. If TRUE, workers write a bounded per-worker error
log to disk to aid debugging failed tasks (default FALSE).
\item \code{error_log_max_lines}: integer. Maximum lines per worker in the error
log (default 100).
}
}}

\item{workers}{Integer. Number of worker processes. If NULL, uses existing
pool or creates one with \code{detectCores() - 1}.}

\item{chunk_size}{Integer. Shards to batch per worker dispatch (default 1).
Higher values reduce RPC overhead but may hurt load balancing.}

\item{profile}{Execution profile: \code{"default"}, \code{"memory"} (aggressive recycling),
or \code{"speed"} (minimal overhead). With \code{profile="speed"}, shard_map will
automatically enable \code{dispatch_mode="shm_queue"} when possible for
\code{shard_map(N, ...)} out-buffer workflows (scalar \code{N}, \code{chunk_size=1}),
unless \code{dispatch_mode} is explicitly specified.}

\item{mem_cap}{Memory cap per worker (e.g., "2GB"). Workers exceeding this
are recycled.}

\item{recycle}{Logical or numeric. If TRUE, recycle workers on RSS drift.
If numeric, specifies drift threshold (default 0.5 = 50\% growth).}

\item{cow}{Copy-on-write policy for borrowed inputs: \code{"deny"} (error on mutation),
\code{"audit"} (detect and flag), or \code{"allow"} (permit with tracking).}

\item{seed}{Integer. RNG seed for reproducibility. If NULL, no seed is set.}

\item{diagnostics}{Logical. Collect detailed diagnostics (default TRUE).}

\item{packages}{Character vector. Additional packages to load in workers.}

\item{init_expr}{Expression to evaluate in each worker on startup.}

\item{timeout}{Numeric. Seconds to wait for each shard (default 3600).}

\item{max_retries}{Integer. Maximum retries per shard on failure (default 3).}

\item{health_check_interval}{Integer. Check worker health every N shards (default 10).}
}
\value{
A \code{shard_result} object containing:
\itemize{
\item \code{results}: List of results from each shard (if fun returns values)
\item \code{failures}: Any permanently failed shards
\item \code{diagnostics}: Timing, memory, and worker statistics
\item \code{pool_stats}: Pool-level statistics
}
}
\description{
Core parallel execution engine with supervision, shared inputs, and output buffers.

Executes a function over shards in parallel with worker supervision,
shared inputs, and explicit output buffers. This is the primary entry
point for shard's parallel execution model.
}
\examples{
\donttest{
blocks <- shards(1000, workers = 2)
result <- shard_map(blocks, function(shard) {
  sum(shard$idx^2)
}, workers = 2)
pool_stop()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shards.R
\name{shards_list}
\alias{shards_list}
\title{Create Shards from an Explicit Index List}
\usage{
shards_list(idxs)
}
\arguments{
\item{idxs}{List of integer vectors (1-based indices). Each element becomes
one shard with fields \code{id}, \code{idx}, and \code{len}.}
}
\value{
A \code{shard_descriptor} list describing the chunk layout.
}
\description{
Constructs a \code{shard_descriptor} from a user-supplied list of index vectors.
This is useful for non-contiguous workloads like searchlights/feature sets
where each shard operates on an arbitrary subset.
}
\examples{
sh <- shards_list(list(1:10, 11:20, 21:30))
length(sh)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapter.R
\name{shard_unregister_adapter}
\alias{shard_unregister_adapter}
\title{Unregister an Adapter}
\usage{
shard_unregister_adapter(class)
}
\arguments{
\item{class}{A character string naming the class to unregister.}
}
\value{
Invisibly returns the removed adapter, or NULL if no adapter was
registered for this class.
}
\description{
Removes a previously registered adapter for a class. After unregistration,
objects of this class will use default traversal behavior during deep sharing.
}
\examples{
shard_list_adapters()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{is_shared}
\alias{is_shared}
\title{Check if Object is Shared}
\usage{
is_shared(x)
}
\arguments{
\item{x}{An object to check.}
}
\value{
A logical scalar: \code{TRUE} if \code{x} is a shared object,
\code{FALSE} otherwise.
}
\description{
Check if Object is Shared
}
\examples{
is_shared(1:10)
\donttest{
shared <- share(1:10)
is_shared(shared)
close(shared)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/worker.R
\name{worker}
\alias{worker}
\title{Individual Worker Control}
\description{
Spawn, monitor, and control individual R worker processes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{as.logical.shard_buffer}
\alias{as.logical.shard_buffer}
\title{Coerce a Shared Memory Buffer to Logical}
\usage{
\method{as.logical}{shard_buffer}(x, ...)
}
\arguments{
\item{x}{A \code{shard_buffer} object.}

\item{...}{Ignored.}
}
\value{
A logical vector with the buffer contents.
}
\description{
Coerce a Shared Memory Buffer to Logical
}
\examples{
\donttest{
buf <- buffer("logical", dim = 5)
as.logical(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{buffer_info}
\alias{buffer_info}
\title{Get Buffer Info}
\usage{
buffer_info(x)
}
\arguments{
\item{x}{A shard_buffer object.}
}
\value{
A named list with buffer properties: \code{type}, \code{dim},
\code{n}, \code{bytes}, \code{backing}, \code{path}, and \code{readonly}.
}
\description{
Returns information about a buffer.
}
\examples{
\donttest{
buf <- buffer("integer", dim = c(5, 5))
buffer_info(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{diagnostics}
\alias{diagnostics}
\title{Diagnostics API}
\description{
Comprehensive diagnostics for shard parallel execution, providing
insights into memory usage, worker status, task execution, and shared memory
segments.
}
\details{
The diagnostics API provides multiple views into shard's runtime behavior:
\itemize{
\item \code{report()}: Primary entry point with configurable detail levels
\item \code{mem_report()}: Memory usage across workers
\item \code{cow_report()}: Copy-on-write policy tracking
\item \code{copy_report()}: Data transfer statistics
\item \code{task_report()}: Task/chunk execution statistics
\item \code{segment_report()}: Shared memory segment information
}

All functions return S3 \code{shard_report} objects with appropriate print
methods for human-readable output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinity.R
\name{shared_advise}
\alias{shared_advise}
\title{Advise access pattern for a shared input vector/matrix}
\usage{
shared_advise(
  x,
  advice = c("normal", "sequential", "random", "willneed", "dontneed")
)
}
\arguments{
\item{x}{A shard shared vector (from \code{\link[=share]{share()}}).}

\item{advice}{See \code{\link[=segment_advise]{segment_advise()}}.}
}
\value{
A logical scalar; \code{TRUE} if the OS accepted the hint.
}
\description{
Advise access pattern for a shared input vector/matrix
}
\examples{
\donttest{
x <- as_shared(1:100)
shared_advise(x, "sequential")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pool.R
\name{pool_status}
\alias{pool_status}
\title{Get Pool Status}
\usage{
pool_status(pool = NULL)
}
\arguments{
\item{pool}{A \code{shard_pool} object. If NULL, uses the current pool.}
}
\value{
A data frame with worker status information.
}
\description{
Returns current status of all workers in the pool.
}
\examples{
\donttest{
p <- pool_create(2)
pool_status(p)
pool_stop(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernels.R
\name{shard_crossprod}
\alias{shard_crossprod}
\title{Parallel crossprod() using shard views + output buffers}
\usage{
shard_crossprod(
  X,
  Y,
  workers = NULL,
  block_x = "auto",
  block_y = "auto",
  backing = c("mmap", "shm"),
  materialize = c("auto", "never", "always"),
  materialize_max_bytes = 512 * 1024^2,
  diagnostics = TRUE
)
}
\arguments{
\item{X, Y}{Double matrices with the same number of rows.}

\item{workers}{Number of worker processes.}

\item{block_x, block_y}{Tile sizes over \code{ncol(X)} and \code{ncol(Y)}. Use \code{"auto"}
(default) to autotune on the current machine.}

\item{backing}{Backing for shared inputs and output buffer (\code{"mmap"} or \code{"shm"}).}

\item{materialize}{Whether to return the result as a standard R matrix:
\code{"never"} (return buffer handle), \code{"always"}, or \code{"auto"} (materialize if
estimated output size is below \code{materialize_max_bytes}).}

\item{materialize_max_bytes}{Threshold for \code{"auto"} materialization.}

\item{diagnostics}{Whether to collect shard_map diagnostics.}
}
\value{
A list with:
\itemize{
\item \code{buffer}: shard_buffer for the result (p x v)
\item \code{value}: materialized matrix if requested, otherwise NULL
\item \code{run}: the underlying shard_result from shard_map
\item \code{tile}: chosen tile sizes
}
}
\description{
Computes \code{crossprod(X, Y)} (i.e. \code{t(X) \%*\% Y}) using:
\itemize{
\item shared/mmap-backed inputs (one copy),
\item block views (no slice materialization),
\item BLAS-3 dgemm in each tile,
\item an explicit shared output buffer (no gather/bind spikes).
}
}
\details{
This is intended as an ergonomic entry point for the "wow" path: users
shouldn't have to manually call \code{share()}, \code{view_block()}, \code{buffer()},
\code{tiles2d()}, and \code{shard_map()} for common patterns.
}
\examples{
\donttest{
X <- matrix(rnorm(2000), 100, 20)
Y <- matrix(rnorm(2000), 100, 20)
res <- shard_crossprod(X, Y, block_x = 50, block_y = 10, workers = 2)
pool_stop()
res$value
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{[<-.shard_buffer}
\alias{[<-.shard_buffer}
\title{Assign to Buffer Elements}
\usage{
\method{[}{shard_buffer}(x, i, j, ...) <- value
}
\arguments{
\item{x}{A shard_buffer object.}

\item{i}{Index or indices.}

\item{j}{Optional second index (for matrices).}

\item{...}{Additional indices (for arrays).}

\item{value}{Values to assign.}
}
\value{
The modified \code{shard_buffer} object, invisibly.
}
\description{
Assign to Buffer Elements
}
\examples{
\donttest{
buf <- buffer("double", dim = 10)
buf[1:5] <- rnorm(5)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{idx_range}
\alias{idx_range}
\title{Contiguous index range}
\usage{
idx_range(start, end)
}
\arguments{
\item{start}{Integer. Start index (1-based, inclusive).}

\item{end}{Integer. End index (1-based, inclusive).}
}
\value{
An object of class \code{shard_idx_range}.
}
\description{
Creates a compact, serializable range descriptor for contiguous indices.
This avoids allocating an explicit index vector for large slices.
}
\examples{
r <- idx_range(1, 100)
r
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dispatch.R
\name{pool_sapply}
\alias{pool_sapply}
\title{Parallel sapply with Supervision}
\usage{
pool_sapply(X, FUN, ..., simplify = TRUE, pool = NULL)
}
\arguments{
\item{X}{List or vector to iterate over.}

\item{FUN}{Function to apply.}

\item{...}{Additional arguments to FUN.}

\item{simplify}{Logical. Simplify result to vector/matrix?}

\item{pool}{A \code{shard_pool} object. If NULL, uses current pool.}
}
\value{
Simplified result if possible, otherwise a list.
}
\description{
Parallel sapply with Supervision
}
\examples{
\donttest{
pool_create(2)
result <- pool_sapply(1:4, function(x) x^2, pool = pool_get())
pool_stop()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{shared_vector}
\alias{shared_vector}
\title{Create a shared vector from a segment}
\usage{
shared_vector(
  segment,
  type = c("double", "integer", "logical", "raw"),
  offset = 0,
  length = NULL,
  readonly = TRUE,
  cow = NULL
)
}
\arguments{
\item{segment}{A shard_segment object}

\item{type}{Vector type: "integer", "double"/"numeric", "logical", or "raw"}

\item{offset}{Byte offset into segment (default: 0)}

\item{length}{Number of elements. If NULL, calculated from segment size.}

\item{readonly}{If TRUE, prevent write access via DATAPTR (default: TRUE)}

\item{cow}{Copy-on-write policy for mutation attempts. One of
\code{"deny"}, \code{"audit"}, or \code{"allow"}. If NULL, defaults to \code{"deny"} when
\code{readonly=TRUE} and \code{"allow"} otherwise.}
}
\value{
An ALTREP vector backed by shared memory
}
\description{
Create a shared vector from a segment
}
\examples{
\donttest{
seg <- segment_create(400)
segment_write(seg, 1:100, offset = 0)

x <- shared_vector(seg, "integer", length = 100)
x[1:10]

shared_diagnostics(x)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_info}
\alias{segment_info}
\title{Get segment information}
\usage{
segment_info(x)
}
\arguments{
\item{x}{A shard_segment object}
}
\value{
A named list with segment metadata including \code{size}, \code{backing},
\code{path}, \code{readonly}, and \code{owns}.
}
\description{
Get segment information
}
\examples{
\donttest{
seg <- segment_create(1024)
segment_info(seg)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/queue.R
\name{queue}
\alias{queue}
\title{Chunk Queue Management}
\description{
Queue management for dispatching chunks to workers with requeue support.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_registry.R
\name{register_kernel}
\alias{register_kernel}
\title{Register a shard kernel}
\usage{
register_kernel(
  name,
  impl,
  signature = NULL,
  footprint = NULL,
  supports_views = TRUE,
  description = NULL
)
}
\arguments{
\item{name}{Kernel name (string).}

\item{impl}{Function implementing the kernel. It must accept the shard
descriptor as its first argument.}

\item{signature}{Optional short signature string for documentation.}

\item{footprint}{Optional footprint hint. Either a constant (bytes) or a
function \verb{(shard, ...) -> list(class='tiny'|'medium'|'huge', bytes=...)}.}

\item{supports_views}{Logical. Whether the kernel is intended to operate on
shard views without slice materialization.}

\item{description}{Optional human-readable description.}
}
\value{
Invisibly, the registered kernel metadata.
}
\description{
Registers a named kernel implementation that can be selected by
\code{shard_map(..., kernel = "name")}.
}
\details{
A "kernel" is just a function that shard_map can call for each shard. The
registry lets shard_map attach additional metadata (footprint hints,
supports_views) for scheduling/autotuning.
}
\examples{
list_kernels()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{recommendations}
\alias{recommendations}
\title{Performance Recommendations}
\usage{
recommendations(result)
}
\arguments{
\item{result}{A \code{shard_result} from \code{\link[=shard_map]{shard_map()}}.}
}
\value{
A character vector of recommendations (possibly empty).
}
\description{
Uses run telemetry (copy/materialization stats, packing volume, buffer/table
writes, scratch pool stats) to produce actionable recommendations.
}
\examples{
\donttest{
res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
pool_stop()
recommendations(res)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_finalize.shard_table_buffer}
\alias{table_finalize.shard_table_buffer}
\title{Finalize a table buffer}
\usage{
\method{table_finalize}{shard_table_buffer}(
  target,
  materialize = c("never", "auto", "always"),
  max_bytes = 256 * 1024^2,
  ...
)
}
\arguments{
\item{target}{A \code{shard_table_buffer}.}

\item{materialize}{\code{"never"}, \code{"auto"}, or \code{"always"}.}

\item{max_bytes}{For \code{"auto"}, materialize only if estimated bytes <= max_bytes.}

\item{...}{Reserved for future extensions.}
}
\value{
A \code{shard_table_handle} or a materialized data.frame/tibble.
}
\description{
Finalize a table buffer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/share.R
\name{shard_share_hook}
\alias{shard_share_hook}
\alias{shard_share_hook.default}
\title{Deep Sharing Hook for Custom Classes}
\usage{
shard_share_hook(x, ctx)

\method{shard_share_hook}{default}(x, ctx)
}
\arguments{
\item{x}{The object being traversed during deep sharing.}

\item{ctx}{A context list containing:
\describe{
\item{path}{Current node path string (e.g., \code{"<root>$data@cache"})}
\item{class}{class(x) - the object's class vector}
\item{mode}{'strict' or 'balanced' - sharing mode}
\item{min_bytes}{Minimum size threshold for sharing}
\item{types}{Character vector of enabled types for sharing}
\item{deep}{Logical, always TRUE when hook is called}
}}
}
\value{
A list with optional fields:
\describe{
\item{skip_slots}{Character vector of S4 slot names to not traverse}
\item{skip_paths}{Character vector of paths to not traverse}
\item{force_share_paths}{Character vector of paths to force share (ignore min_bytes)}
\item{rewrite}{Function(x) -> x to transform object before traversal}
}
Return an empty list for default behavior (no customization).
}
\description{
S3 generic that allows classes to customize deep sharing behavior.
Override this for your class to control which slots/elements are traversed,
force sharing of small objects, or transform objects before traversal.
}
\examples{
\donttest{
shard_share_hook.MyModelClass <- function(x, ctx) {
    list(
        skip_slots = "cache",
        force_share_paths = paste0(ctx$path, "@coefficients")
    )
}

shard_share_hook.LazyData <- function(x, ctx) {
    list(
        rewrite = function(obj) {
            obj$data <- as.matrix(obj$data)
            obj
        }
    )
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{as_shared}
\alias{as_shared}
\title{Create a shared vector from an existing R vector}
\usage{
as_shared(x, readonly = TRUE, backing = "auto", cow = NULL)
}
\arguments{
\item{x}{An atomic vector (integer, double, logical, or raw)}

\item{readonly}{If TRUE, prevent write access (default: TRUE)}

\item{backing}{Backing type for the segment: "auto", "mmap", or "shm"}

\item{cow}{Copy-on-write policy for the resulting shared vector. One of
\code{"deny"}, \code{"audit"}, or \code{"allow"}. If NULL, defaults based on \code{readonly}.}
}
\value{
An ALTREP vector backed by shared memory
}
\description{
Convenience function that creates a segment, writes the data,
and returns an ALTREP view.
}
\examples{
\donttest{
x <- as_shared(1:100)
is_shared_vector(x)

y <- x[1:10]
is_shared_vector(y)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shards.R
\name{length.shard_descriptor}
\alias{length.shard_descriptor}
\title{Length of a shard_descriptor Object}
\usage{
\method{length}{shard_descriptor}(x)
}
\arguments{
\item{x}{A \code{shard_descriptor} object.}
}
\value{
An integer scalar giving the number of shards.
}
\description{
Length of a shard_descriptor Object
}
\examples{
sh <- shards(100, block_size = 25)
length(sh)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{factor_col}
\alias{factor_col}
\title{Categorical column type}
\usage{
factor_col(levels)
}
\arguments{
\item{levels}{Character vector of allowed levels.}
}
\value{
A \code{shard_coltype} object.
}
\description{
Stores factors as int32 codes plus shared levels metadata.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shards.R
\name{shards}
\alias{shards}
\title{Shard Descriptor Creation}
\usage{
shards(
  n,
  block_size = "auto",
  workers = NULL,
  strategy = c("contiguous", "strided"),
  min_shards_per_worker = 4L,
  max_shards_per_worker = 64L,
  scratch_bytes_per_item = 0,
  scratch_budget = 0
)
}
\arguments{
\item{n}{Integer. Total number of items to shard.}

\item{block_size}{Block size specification. Can be:
\itemize{
\item \code{"auto"} (default): Autotune based on worker count
\item Integer: Explicit number of items per shard
\item Character: Human-readable like \code{"1K"}, \code{"10K"}
}}

\item{workers}{Integer. Number of workers for autotuning (default: pool size or detectCores - 1).}

\item{strategy}{Sharding strategy: \code{"contiguous"} (default) or \code{"strided"}.}

\item{min_shards_per_worker}{Integer. Minimum shards per worker for load balancing (default 4).}

\item{max_shards_per_worker}{Integer. Maximum shards per worker to limit overhead (default 64).}

\item{scratch_bytes_per_item}{Numeric. Expected scratch memory per item for memory budgeting.}

\item{scratch_budget}{Character or numeric. Total scratch memory budget (e.g., "1GB").}
}
\value{
A \code{shard_descriptor} object containing:
\itemize{
\item \code{n}: Total items
\item \code{block_size}: Computed block size
\item \code{strategy}: Strategy used
\item \code{shards}: List of shard descriptors with \code{id}, \code{start}, \code{end}, \code{idx} fields
}
}
\description{
Create shard descriptors for parallel execution with autotuning.

Produces shard descriptors (index ranges) for use with \code{shard_map()}.
Supports autotuning based on worker count and memory constraints.
}
\examples{
blocks <- shards(1e6, workers = 8)
length(blocks$shards)

blocks <- shards(1000, block_size = 100)

blocks$shards[[1]]$idx
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_write.shard_table_buffer}
\alias{table_write.shard_table_buffer}
\title{Write into a table buffer}
\usage{
\method{table_write}{shard_table_buffer}(target, rows_or_shard_id, data, ...)
}
\arguments{
\item{target}{A \code{shard_table_buffer}.}

\item{rows_or_shard_id}{Row selector (idx_range or integer vector).}

\item{data}{A data.frame or named list matching the schema columns.}

\item{...}{Reserved for future extensions.}
}
\value{
NULL (invisibly).
}
\description{
Write into a table buffer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_buffer}
\alias{table_buffer}
\title{Allocate a fixed-row table buffer}
\usage{
table_buffer(schema, nrow, backing = c("auto", "mmap", "shm"))
}
\arguments{
\item{schema}{A \code{shard_schema}.}

\item{nrow}{Total number of rows in the final table.}

\item{backing}{Backing type for buffers (\code{"auto"}, \code{"mmap"}, \code{"shm"}).}
}
\value{
A \code{shard_table_buffer} object with one shared buffer per schema column.
}
\description{
Allocates a columnar table output: one typed buffer per column, each of
length \code{nrow}. Intended for lock-free disjoint row-range writes in shard_map.
}
\examples{
\donttest{
s <- schema(x = float64(), y = int32())
tb <- table_buffer(s, nrow = 100L)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arena.R
\name{print.arena_result}
\alias{print.arena_result}
\title{Print an arena_result object}
\usage{
\method{print}{arena_result}(x, ...)
}
\arguments{
\item{x}{An \code{arena_result} object.}

\item{...}{Additional arguments passed to \code{print}.}
}
\value{
Returns \code{x} invisibly.
}
\description{
Print an arena_result object
}
\examples{
\donttest{
info <- arena({ sum(1:10) }, diagnostics = TRUE)
print(info)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{as.vector.shard_buffer}
\alias{as.vector.shard_buffer}
\title{Coerce a Shared Memory Buffer to a Vector}
\usage{
\method{as.vector}{shard_buffer}(x, mode = "any")
}
\arguments{
\item{x}{A \code{shard_buffer} object.}

\item{mode}{Storage mode passed to \code{\link{as.vector}}.}
}
\value{
A vector of the buffer's type (or coerced to \code{mode}).
}
\description{
Coerce a Shared Memory Buffer to a Vector
}
\examples{
\donttest{
buf <- buffer("double", dim = 5)
buf[1:5] <- 1:5
as.vector(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_create}
\alias{segment_create}
\title{Create a new shared memory segment}
\usage{
segment_create(
  size,
  backing = c("auto", "mmap", "shm"),
  path = NULL,
  readonly = FALSE
)
}
\arguments{
\item{size}{Size of the segment in bytes}

\item{backing}{Backing type: "auto", "mmap", or "shm"}

\item{path}{Optional file path for mmap backing (NULL for temp file)}

\item{readonly}{Create as read-only (after initial write)}
}
\value{
A \code{shard_segment} object backed by shared memory.
}
\description{
Create a new shared memory segment
}
\examples{
\donttest{
seg <- segment_create(1024 * 1024)
segment_info(seg)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{shared_view}
\alias{shared_view}
\title{Create a view (subset) of a shared vector}
\usage{
shared_view(x, start, length)
}
\arguments{
\item{x}{A shard ALTREP vector}

\item{start}{Start index (1-based, like R)}

\item{length}{Number of elements}
}
\value{
An ALTREP view into the same shared memory
}
\description{
Create a view (subset) of a shared vector
}
\examples{
\donttest{
seg <- segment_create(800)
segment_write(seg, 1:100, offset = 0)
x <- shared_vector(seg, "integer", length = 100)

y <- shared_view(x, start = 10, length = 11)
y[1]
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arena.R
\name{in_arena}
\alias{in_arena}
\title{Check if Currently Inside an Arena}
\usage{
in_arena()
}
\value{
Logical indicating whether we are in an arena scope.
}
\description{
Returns TRUE if the current execution context is within an arena() scope.
}
\examples{
in_arena()
arena({
  in_arena()
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{coltypes}
\alias{coltypes}
\alias{int32}
\alias{float64}
\alias{bool}
\alias{raw_col}
\alias{string_col}
\title{Column Types}
\usage{
int32()

float64()

bool()

raw_col()

string_col()
}
\value{
A \code{shard_coltype} object.
}
\description{
Type constructors for schema-driven table outputs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{view_info}
\alias{view_info}
\title{Introspection for a view}
\usage{
view_info(v)
}
\arguments{
\item{v}{A shard view.}
}
\value{
A named list with fields: \code{dtype}, \code{dim}, \code{slice_dim},
\code{rows}, \code{cols}, \code{layout}, \code{fast_path},
\code{nbytes_est}, and \code{base_is_shared}.
}
\description{
Returns metadata about a view without forcing materialization.
}
\examples{
\donttest{
m <- share(matrix(1:20, nrow = 4))
v <- view_block(m, cols = idx_range(1, 2))
view_info(v)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{view_diagnostics}
\alias{view_diagnostics}
\title{View diagnostics}
\usage{
view_diagnostics()
}
\value{
A list with counters.
}
\description{
Returns global counters for view creation/materialization. This is a simple
first step; in future this should be integrated into shard_map run-level
diagnostics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_size}
\alias{segment_size}
\title{Get the size of a segment}
\usage{
segment_size(x)
}
\arguments{
\item{x}{A shard_segment object}
}
\value{
Size in bytes as a numeric scalar.
}
\description{
Get the size of a segment
}
\examples{
\donttest{
seg <- segment_create(1024)
segment_size(seg)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buffer.R
\name{as.array.shard_buffer}
\alias{as.array.shard_buffer}
\title{Coerce a Shared Memory Buffer to Array}
\usage{
\method{as.array}{shard_buffer}(x, ...)
}
\arguments{
\item{x}{A \code{shard_buffer} object.}

\item{...}{Ignored.}
}
\value{
An array with the buffer contents and the buffer's dimensions, or a
plain vector for 1-D buffers.
}
\description{
Coerce a Shared Memory Buffer to Array
}
\examples{
\donttest{
buf <- buffer("double", dim = c(2, 3, 4))
as.array(buf)
buffer_close(buf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapter.R
\name{shard_get_adapter}
\alias{shard_get_adapter}
\title{Get Adapter for an Object}
\usage{
shard_get_adapter(x)
}
\arguments{
\item{x}{An R object.}
}
\value{
The adapter list if one is registered for any of the object's classes,
or NULL if no adapter is registered.
}
\description{
Retrieves the registered adapter for an object's class. Checks all classes
in the object's class hierarchy, returning the first matching adapter.
}
\examples{
shard_get_adapter(1:10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinity.R
\name{affinity}
\alias{affinity}
\title{CPU Affinity + mmap Advice (Advanced)}
\description{
These controls are opt-in and best-effort. On unsupported platforms, they
safely no-op (returning FALSE).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dispatch.R
\name{print.shard_dispatch_result}
\alias{print.shard_dispatch_result}
\title{Print a shard_dispatch_result Object}
\usage{
\method{print}{shard_dispatch_result}(x, ...)
}
\arguments{
\item{x}{A \code{shard_dispatch_result} object.}

\item{...}{Further arguments (ignored).}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a shard_dispatch_result Object
}
\examples{
\donttest{
pool_create(2)
chunks <- list(list(id = 1L, x = 1), list(id = 2L, x = 2))
result <- dispatch_chunks(chunks, function(chunk) chunk$x, pool = pool_get())
print(result)
pool_stop()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{is_shared_vector}
\alias{is_shared_vector}
\title{Check if an object is a shared vector}
\usage{
is_shared_vector(x)
}
\arguments{
\item{x}{Any R object}
}
\value{
TRUE if x is a shard ALTREP vector, FALSE otherwise
}
\description{
Check if an object is a shared vector
}
\examples{
\donttest{
seg <- segment_create(400)
segment_write(seg, 1:100, offset = 0)
x <- shared_vector(seg, "integer", length = 100)

is_shared_vector(x)
is_shared_vector(1:10)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pool.R
\name{pool_create}
\alias{pool_create}
\title{Create a Worker Pool}
\usage{
pool_create(
  n = parallel::detectCores() - 1L,
  rss_limit = "2GB",
  rss_drift_threshold = 0.5,
  heartbeat_interval = 5,
  min_recycle_interval = 1,
  init_expr = NULL,
  packages = NULL
)
}
\arguments{
\item{n}{Integer. Number of worker processes to spawn.}

\item{rss_limit}{Numeric or character. Maximum RSS per worker before recycling.
Can be bytes (numeric) or human-readable (e.g., "2GB"). Default is "2GB".}

\item{rss_drift_threshold}{Numeric. Fraction of RSS increase from baseline
that triggers recycling (default 0.5 = 50\% growth).}

\item{heartbeat_interval}{Numeric. Seconds between health checks (default 5).}

\item{min_recycle_interval}{Numeric. Minimum time in seconds between recycling
the same worker (default 1.0). This prevents thrashing PSOCK worker creation
under extremely tight RSS limits.}

\item{init_expr}{Expression to evaluate in each worker on startup.}

\item{packages}{Character vector. Packages to load in workers.}
}
\value{
A \code{shard_pool} object (invisibly). The pool is also stored in the
package environment for reuse.
}
\description{
Spawns N R worker processes that persist across multiple \code{shard_map()} calls.
Workers are supervised and recycled when RSS drift exceeds thresholds.
}
\examples{
\donttest{
p <- pool_create(2)
pool_stop(p)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cow.R
\name{[<-.shard_shared_vector}
\alias{[<-.shard_shared_vector}
\title{Subset-assign a Shared Vector}
\usage{
\method{[}{shard_shared_vector}(x, ...) <- value
}
\arguments{
\item{x}{A \code{shard_shared_vector}.}

\item{...}{Indices.}

\item{value}{Replacement value.}
}
\value{
The modified object \code{x}.
}
\description{
Replacement method for \code{shard_shared_vector}. Raises an error if the
copy-on-write policy is \code{"deny"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_stream.R
\name{stream_group_sum}
\alias{stream_group_sum}
\title{Stream group-wise sum}
\usage{
stream_group_sum(x, group, value, na_rm = TRUE)
}
\arguments{
\item{x}{A \code{shard_row_groups} or \code{shard_dataset} handle.}

\item{group}{Group column name (recommended: \code{factor_col()}).}

\item{value}{Numeric column name to sum.}

\item{na_rm}{Logical; drop rows where value is NA (default TRUE).}
}
\value{
A data.frame with columns \code{group} (factor) and \code{sum} (numeric).
}
\description{
Computes sum(value) by group across partitions without collecting. This is
optimized for factor groups (\code{factor_col()}).
}
\examples{
\donttest{
s <- schema(g = factor_col(c("a", "b")), x = float64())
sink <- table_sink(s, mode = "row_groups")
table_write(sink, 1L,
  data.frame(g = factor(c("a", "b", "a"), levels = c("a", "b")), x = c(1, 2, 3)))
rg <- table_finalize(sink)
stream_group_sum(rg, "g", "x")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arena.R
\name{arena}
\alias{arena}
\title{Arena Semantic Scope}
\usage{
arena(
  expr,
  strict = FALSE,
  escape_threshold = .arena_escape_threshold,
  gc_after = strict,
  diagnostics = FALSE
)
}
\arguments{
\item{expr}{An expression to evaluate within the arena scope.}

\item{strict}{Logical. If TRUE, enables strict mode which:
\itemize{
\item Warns if large objects (> 1MB by default) escape the scope
\item Triggers garbage collection after scope exit
\item Tracks memory growth for diagnostics
}
Default is FALSE for compatibility and performance.}

\item{escape_threshold}{Numeric. Size in bytes above which returned objects
trigger a warning in strict mode. Default is 1MB (1048576 bytes).
Only used when \code{strict = TRUE}.}

\item{gc_after}{Logical. If TRUE, triggers garbage collection after the
arena scope exits. Default is TRUE in strict mode, FALSE otherwise.}

\item{diagnostics}{Logical. If TRUE, returns diagnostics about memory usage
along with the result. Default is FALSE.}
}
\value{
The result of evaluating \code{expr}. If \code{diagnostics = TRUE},
returns an \code{arena_result} object with elements \code{result} and
\code{diagnostics}.
}
\description{
Semantic scope for scratch memory that signals temporary data
should not accumulate. Enables memory-conscious parallel execution.

Evaluates an expression in a semantic scope that signals scratch memory usage.
This enables memory-conscious execution where temporaries are expected to be
reclaimed after the scope exits.
}
\details{
The \code{arena()} function provides a semantic scope that signals "this code
produces scratch data that should not outlive the scope." It serves two
purposes:

\enumerate{
\item \strong{For compiled kernels}: When Rust-based kernels are available,
arena() provides real scratch arenas backed by temporary shared memory
segments that are automatically reclaimed.
\item \strong{For arbitrary R code}: Triggers post-task memory checks to
detect growth and potential memory leaks.
}

The \code{strict} parameter controls escape detection:
\itemize{
\item \code{strict = FALSE} (default): Returns results normally, logs
diagnostics about memory growth.
\item \code{strict = TRUE}: Warns or errors if large objects escape the
scope, and triggers aggressive memory reclamation.
}
}
\examples{
\donttest{
result <- arena({
  tmp <- matrix(rnorm(1e6), nrow = 1000)
  colMeans(tmp)
})

info <- arena({
  x <- rnorm(1e5)
  sum(x)
}, diagnostics = TRUE)
info$diagnostics
}
}
\seealso{
\code{\link{shard_map}} for parallel execution,
\code{\link{share}} for shared memory inputs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pool.R
\name{pool}
\alias{pool}
\title{Worker Pool Management}
\description{
Spawn and supervise persistent R worker processes with RSS monitoring.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cow.R
\name{attr<-.shard_shared_vector}
\alias{attr<-.shard_shared_vector}
\title{Set an Attribute on a Shared Vector}
\usage{
\method{attr}{shard_shared_vector}(x, which) <- value
}
\arguments{
\item{x}{A \code{shard_shared_vector}.}

\item{which}{Attribute name.}

\item{value}{Attribute value.}
}
\value{
The modified object \code{x}.
}
\description{
Raises an error if the copy-on-write policy is \code{"deny"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{schema}
\alias{schema}
\title{Define a table schema}
\usage{
schema(...)
}
\arguments{
\item{...}{Named columns with type specs (e.g., \code{int32()}, \code{float64()}).}
}
\value{
A \code{shard_schema} object.
}
\description{
A schema is a named set of columns with explicit types. It is used to
allocate table buffers and validate writes.
}
\examples{
s <- schema(x = float64(), y = int32(), label = string_col())
s
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{cow_report}
\alias{cow_report}
\title{Copy-on-Write Policy Report}
\usage{
cow_report(result = NULL)
}
\arguments{
\item{result}{Optional. A \code{shard_result} object to extract COW stats from.}
}
\value{
An S3 object of class \code{shard_report} with type \code{"cow"}
containing:
\itemize{
\item \code{type}: "cow"
\item \code{timestamp}: When the report was generated
\item \code{policy}: The COW policy used ("deny", "audit", "allow")
\item \code{violations}: Count of COW violations detected (audit mode)
\item \code{copies_triggered}: Estimated copies triggered by mutations
}
}
\description{
Generates a report of copy-on-write behavior for borrowed inputs.
}
\examples{
\donttest{
res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
pool_stop()
cow_report(res)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{report}
\alias{report}
\title{Generate Shard Runtime Report}
\usage{
report(level = c("summary", "workers", "tasks", "segments"), result = NULL)
}
\arguments{
\item{level}{Character. Detail level for the report:
\itemize{
\item \code{"summary"}: High-level overview (default)
\item \code{"workers"}: Include per-worker details
\item \code{"tasks"}: Include task execution history
\item \code{"segments"}: Include shared memory segment details
}}

\item{result}{Optional. A \code{shard_result} object from \code{\link{shard_map}}
to include execution diagnostics from.}
}
\value{
An S3 object of class \code{shard_report} containing:
\itemize{
\item \code{level}: The requested detail level
\item \code{timestamp}: When the report was generated
\item \code{pool}: Pool status information (if pool exists)
\item \code{memory}: Memory usage summary
\item \code{workers}: Per-worker details (if level includes workers)
\item \code{tasks}: Task execution details (if level includes tasks)
\item \code{segments}: Segment details (if level includes segments)
\item \code{result_diagnostics}: Diagnostics from shard_result (if provided)
}
}
\description{
Primary entry point for shard diagnostics. Generates a comprehensive report
of the current runtime state including pool status, memory usage, and
execution statistics.
}
\examples{
\donttest{
res <- shard_map(shards(100, workers = 2), function(s) sum(s$idx), workers = 2)
pool_stop()
report(result = res)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{collect}
\alias{collect}
\title{Collect a shard table into memory}
\usage{
collect(x, ...)
}
\arguments{
\item{x}{A shard table handle (\code{shard_row_groups}, \code{shard_dataset},
or \code{shard_table_handle}).}

\item{...}{Passed to \code{as_tibble()}.}
}
\value{
A data.frame (or tibble if the \code{tibble} package is installed).
}
\description{
\code{collect()} is a convenience alias for \code{as_tibble()} for shard table outputs.
}
\examples{
\donttest{
s <- schema(x = float64(), y = int32())
tb <- table_buffer(s, nrow = 5L)
table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
handle <- table_finalize(tb)
df <- collect(handle)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{segment_open}
\alias{segment_open}
\title{Open an existing shared memory segment}
\usage{
segment_open(path, backing = c("mmap", "shm"), readonly = TRUE)
}
\arguments{
\item{path}{Path or shm name of the segment}

\item{backing}{Backing type: "mmap" or "shm"}

\item{readonly}{Open as read-only}
}
\value{
A \code{shard_segment} object attached to the existing segment.
}
\description{
Open an existing shared memory segment
}
\examples{
\donttest{
seg <- segment_create(1024, backing = "mmap")
path <- segment_path(seg)
seg2 <- segment_open(path, backing = "mmap", readonly = TRUE)
segment_close(seg2, unlink = FALSE)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/views.R
\name{views}
\alias{views}
\title{Zero-copy Views}
\description{
Views are explicit slice descriptors over shared arrays/matrices. They avoid
creating slice-sized allocations (e.g. \code{Y[, a:b]}) by carrying only metadata
plus a reference to the shared backing.
}
\details{
This is a low-level optimization handle: arbitrary base R operations may
materialize a view; use \code{materialize()} explicitly when you want a standard
matrix/array.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{print.shard_segment}
\alias{print.shard_segment}
\title{Print a Shared Memory Segment}
\usage{
\method{print}{shard_segment}(x, ...)
}
\arguments{
\item{x}{A \code{shard_segment} object.}

\item{...}{Ignored.}
}
\value{
The input \code{x}, invisibly.
}
\description{
Print a Shared Memory Segment
}
\examples{
\donttest{
seg <- segment_create(1024)
print(seg)
segment_close(seg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{table_finalize}
\alias{table_finalize}
\title{Finalize a table buffer or sink}
\usage{
table_finalize(
  target,
  materialize = c("never", "auto", "always"),
  max_bytes = 256 * 1024^2,
  ...
)
}
\arguments{
\item{target}{A \code{shard_table_buffer} or \code{shard_table_sink}.}

\item{materialize}{\code{"never"}, \code{"auto"}, or \code{"always"}.}

\item{max_bytes}{For \code{"auto"}, materialize only if estimated bytes <= max_bytes.}

\item{...}{Reserved for future extensions.}
}
\value{
A \code{shard_table_handle}, \code{shard_row_groups}, or materialized
data.frame/tibble depending on \code{target} type and \code{materialize}.
}
\description{
For a \code{shard_table_buffer}, this returns a lightweight in-memory handle (or a
materialized data.frame/tibble, depending on \code{materialize}).
}
\details{
For a \code{shard_table_sink}, this returns a row-group handle referencing the
written partitions (or materializes them if requested).
}
\examples{
\donttest{
s <- schema(x = float64(), y = int32())
tb <- table_buffer(s, nrow = 5L)
table_write(tb, idx_range(1, 5), data.frame(x = rnorm(5), y = 1:5))
handle <- table_finalize(tb)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{as_tibble.shard_row_groups}
\alias{as_tibble.shard_row_groups}
\title{Materialize a row-groups handle into a data.frame/tibble}
\usage{
\method{as_tibble}{shard_row_groups}(x, max_bytes = 256 * 1024^2, ...)
}
\arguments{
\item{x}{A \code{shard_row_groups} handle.}

\item{max_bytes}{Accepted for API consistency; currently unused for row-groups.}

\item{...}{Reserved for future extensions.}
}
\value{
A data.frame (or tibble if the \code{tibble} package is installed).
}
\description{
Materialize a row-groups handle into a data.frame/tibble
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kernel_registry.R
\name{list_kernels}
\alias{list_kernels}
\title{List registered kernels}
\usage{
list_kernels()
}
\value{
A character vector of registered kernel names.
}
\description{
List registered kernels
}
\examples{
list_kernels()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affinity.R
\name{affinity_supported}
\alias{affinity_supported}
\title{Check whether CPU affinity is supported}
\usage{
affinity_supported()
}
\value{
A logical scalar indicating platform support.
}
\description{
Currently supported on Linux only.
}
\examples{
affinity_supported()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/altrep.R
\name{altrep}
\alias{altrep}
\title{ALTREP Shared Vectors}
\description{
ALTREP-backed zero-copy vectors for shared memory.
}
\details{
These functions create ALTREP (Alternative Representation) vectors that
are backed by shared memory segments. The key benefits are:

\itemize{
\item \strong{Zero-copy subsetting}: Contiguous subsets return views
into the same shared memory, not copies.
\item \strong{Diagnostics}: Track when data pointers are accessed or
when vectors are materialized (copied to standard R vectors).
\item \strong{Read-only protection}: Optionally prevent write access
to protect shared data.
}

Supported types: integer, double/numeric, logical, raw.
}
